/* 2010-08-25: File added and changed by Sony Corporation */
/*
 *  linux/arch/arm/mach-mp200/clock.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/string.h>
#include <linux/clk.h>
#include <linux/mutex.h>

#include <asm/semaphore.h>
#include <asm/arch/pmu.h>
#include <asm/io.h>

#include "clock.h"

static LIST_HEAD(clocks);
static DEFINE_MUTEX(clocks_mutex);

struct clk *clk_get(struct device *dev, const char *id)
{
	struct clk *p, *clk = ERR_PTR(-ENOENT);

	mutex_lock(&clocks_mutex);
	list_for_each_entry(p, &clocks, node) {
		if (strcmp(id, p->name) == 0 && try_module_get(p->owner)) {
			clk = p;
			break;
		}
	}
	mutex_unlock(&clocks_mutex);

	return clk;
}

void clk_put(struct clk *clk)
{
	if (!clk)
		return;
	module_put(clk->owner);
}

int clk_enable(struct clk *clk)
{
	if (!clk)
		return -EINVAL;

	mp200_pmu_open_clockgate(clk->clk_bit);
	if (clk->clkctrl_bit) {
		mp200_pmu_clkctrl_on(clk->clkctrl_bit);
	}

	return 0;
}

void clk_disable(struct clk *clk)
{
	if (!clk)
		return;

	if (clk->clkctrl_bit) {
		mp200_pmu_clkctrl_off(clk->clkctrl_bit);
	}
	mp200_pmu_close_clockgate(clk->clk_bit);
}

unsigned long clk_get_rate(struct clk *clk)
{
	if (!clk)
		return -EINVAL;

	if (clk->div_reg) {
		return inl(clk->div_reg);
	} else {
		return -EINVAL;
	}
}

int clk_set_rate(struct clk *clk, unsigned long rate)
{
	if (!clk)
		return -EINVAL;

	if (clk->div_reg) {
		outl(rate, clk->div_reg);
		return 0;
	} else {
		return -EINVAL;
	}
}

int clk_register(struct clk *clk)
{
	if (!clk)
		return -EINVAL;

	mutex_lock(&clocks_mutex);
	list_add(&clk->node, &clocks);
	mutex_unlock(&clocks_mutex);
	return 0;
}


void clk_unregister(struct clk *clk)
{
	if (!clk)
		return;

	mutex_lock(&clocks_mutex);
	list_del(&clk->node);
	mutex_unlock(&clocks_mutex);
}


int __init clk_init(void)
{
	clk_register(&clk_dma_tclk);
	clk_register(&clk_dma_pch3);
	clk_register(&clk_dma_pch2);
	clk_register(&clk_dma_pch0);
	clk_register(&clk_dma_pclk);

	clk_register(&clk_imc_pclk);
	clk_register(&clk_imc);

	clk_register(&clk_ipu_ahb);
	clk_register(&clk_ipu_dma_pclk);
	clk_register(&clk_ipu_dma);
	clk_register(&clk_rot_pclk);
	clk_register(&clk_rot);
	clk_register(&clk_img_pclk);
	clk_register(&clk_img_2);
	clk_register(&clk_img_1);

	clk_register(&clk_pdma_pclk);
	clk_register(&clk_pdma_hclk);
	clk_register(&clk_pdma_aclk);

	clk_register(&clk_dsp_aclk);
	clk_register(&clk_dsp);

	clk_register(&clk_avc_hsclk);
	clk_register(&clk_avc_hmclk);
	clk_register(&clk_avc_clkm);
	clk_register(&clk_avc_clke);
	clk_register(&clk_avc_clkd);
	clk_register(&clk_avc_clkc);

	clk_register(&clk_dcv_pclk);
	clk_register(&clk_dcv);

	clk_register(&clk_lcd_cclk);
	clk_register(&clk_lcd_lclk);
	clk_register(&clk_lcd_pclk);
	clk_register(&clk_lcd);

	clk_register(&clk_dtv_pclk);
	clk_register(&clk_dtv);

	clk_register(&clk_nts_pclk);
	clk_register(&clk_nts);

	clk_register(&clk_nand_pclk);
	clk_register(&clk_nand_hclk);

	clk_register(&clk_mmm);
	clk_register(&clk_msp_rtck);
	clk_register(&clk_msp_sclk);
	clk_register(&clk_msp_pclk);

	clk_register(&clk_pcm1_sclk);
	clk_register(&clk_pcm0_sclk);
	clk_register(&clk_pcm_pclk);
	clk_register(&clk_pcm0_clk);
	clk_register(&clk_pcm1_clk);

	clk_register(&clk_usb);
	clk_register(&clk_sdia);
	clk_register(&clk_sdib);
	clk_register(&clk_sdic);

	clk_register(&clk_uart2);
	clk_register(&clk_uart1);
	clk_register(&clk_uart0);
	clk_register(&clk_uart);

	clk_register(&clk_i2c_sclk);
	clk_register(&clk_i2c2_sclk);
	clk_register(&clk_i2c);
	clk_register(&clk_i2c2);

	clk_register(&clk_efs);

	clk_register(&clk_pwm_pwclk1);
	clk_register(&clk_pwm_pwclk0);
	clk_register(&clk_pwm_pclk);

	clk_register(&clk_spi2_sclk);
	clk_register(&clk_spi2_pclk);
	clk_register(&clk_spi1_sclk);
	clk_register(&clk_spi1_pclk);
	clk_register(&clk_spi0_sclk);
	clk_register(&clk_spi0_pclk);

	clk_register(&clk_mwi_sclk);
	clk_register(&clk_mwi_pclk);

	clk_register(&clk_tim);
	clk_register(&clk_tw3);
	clk_register(&clk_tw2);
	clk_register(&clk_tw1);
	clk_register(&clk_tw0);
	clk_register(&clk_tg5);
	clk_register(&clk_tg4);
	clk_register(&clk_tg3);
	clk_register(&clk_tg2);
	clk_register(&clk_tg1);
	clk_register(&clk_tg0);
	clk_register(&clk_timer3);
	clk_register(&clk_timer2);
	clk_register(&clk_timer1);
	clk_register(&clk_timer0);


	return 0;
}

EXPORT_SYMBOL(clk_get);
EXPORT_SYMBOL(clk_put);
EXPORT_SYMBOL(clk_enable);
EXPORT_SYMBOL(clk_disable);
EXPORT_SYMBOL(clk_get_rate);
EXPORT_SYMBOL(clk_set_rate);
EXPORT_SYMBOL(clk_register);
EXPORT_SYMBOL(clk_unregister);
