/* 2010-08-25: File added and changed by Sony Corporation */
/*
 *  File Name       : linux/arch/arm/mach-mp200/gpio.c
 *  Function        : gpio
 *  Release Version : Ver 1.01
 *  Release Date    : 2009/02/04
 *
 *  Copyright (C) NEC Electronics Corporation 2009
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/delay.h>

#include <asm/mach/irq.h>
#include <asm/gpio.h>

//#define GPIO_DEBUG 1
#ifdef GPIO_DEBUG
#define DPRINT(FMT,ARGS...) printk(KERN_INFO "%s(): " FMT, __FUNCTION__,##ARGS)
#else
#define DPRINT(FMT,ARGS...)
#endif

#define PIN_MASK(X)	(1U << ((X) & 0x1f))	/* IIA/IEN/IDS/PLS/... */
#define PIN_INDEX(X)	((X) >> 5)

spinlock_t mp200_gio_lock = SPIN_LOCK_UNLOCKED;

static void mp200_gio_irq_ack(unsigned int irq);
static void mp200_gio_irq_mask(unsigned int irq);
static void mp200_gio_irq_unmask(unsigned int irq);
static int mp200_gio_set_irq_type(unsigned int irq, unsigned int type);

static struct irq_chip mp200_gio_chip_ack = {
	.name     = "GIO-edge",
	.ack      = mp200_gio_irq_ack,
	.mask     = mp200_gio_irq_mask,
	.unmask   = mp200_gio_irq_unmask,
	.set_type = mp200_gio_set_irq_type,
	.disable  = mp200_gio_irq_mask,
};

static struct irq_chip mp200_gio_chip = {
	.name     = "GIO-level",
	.ack      = mp200_gio_irq_mask,
	.mask     = mp200_gio_irq_mask,
	.unmask   = mp200_gio_irq_unmask,
	.set_type = mp200_gio_set_irq_type,
	.disable  = mp200_gio_irq_mask,
};

static uint32_t accept_bits[4] = { 0, 0, 0, 0};

static int idt_table[] =
	{
	  VA_GIO + GP_IDT0L,   VA_GIO + GP_IDT1L,   VA_GIO + GP_IDT2L,   VA_GIO + GP_IDT3L,
	  VA_GIO + GP_IDT0H,   VA_GIO + GP_IDT1H,   VA_GIO + GP_IDT2H,   VA_GIO + GP_IDT3H,
	  VA_GIO + GP_IDT0HH,  VA_GIO + GP_IDT1HH,  VA_GIO + GP_IDT2HH,  VA_GIO + GP_IDT3HH,
	  VA_GIO + GP_IDT0HHH, VA_GIO + GP_IDT1HHH, VA_GIO + GP_IDT2HHH, VA_GIO + GP_IDT3HHH,
	};


/* called from set_irq_type() */
static int mp200_gio_set_irq_type(unsigned int irq, unsigned int type)
{
	unsigned int pin = irq - INT_GPIO_BASE;
#ifdef CONFIG_PGPF_GPIO_USE_ASYNC_DETECT
	unsigned int mode = 0x8;
#else /* CONFIG_PGPF_GPIO_USE_ASYNC_DETECT */
	unsigned int mode = 0;
#endif /* CONFIG_PGPF_GPIO_USE_ASYNC_DETECT */
	unsigned long flag;
	unsigned int oiia;
	unsigned int x;
	unsigned int mask;
	unsigned int idtshift;
	unsigned int idt, iia, iir;

	DPRINT("Enter\n");

	if ((irq < INT_GPIO_BASE) || (INT_GPIO_LAST < irq)) {
		return -EINVAL;
	}

	switch (type & IRQ_TYPE_SENSE_MASK) {
	case IRQ_TYPE_EDGE_RISING:
		mode |= 0x00;
		set_irq_chip(irq, &mp200_gio_chip_ack);
		set_irq_handler(irq, handle_edge_irq);
		break;
	case IRQ_TYPE_EDGE_FALLING:
		mode |= 0x01;
		set_irq_chip(irq, &mp200_gio_chip_ack);
		set_irq_handler(irq, handle_edge_irq);
		break;
	case IRQ_TYPE_LEVEL_HIGH:
		mode |= 0x02;
		set_irq_chip(irq, &mp200_gio_chip);
		set_irq_handler(irq, handle_level_irq);
		break;
	case IRQ_TYPE_LEVEL_LOW:
		mode |= 0x03;
		set_irq_chip(irq, &mp200_gio_chip);
		set_irq_handler(irq, handle_level_irq);
		break;
	case IRQ_TYPE_EDGE_BOTH:
		mode |= 0x04;
		set_irq_chip(irq, &mp200_gio_chip_ack);
		set_irq_handler(irq, handle_edge_irq);
		break;
	default:
		return -EINVAL;
	}

	mask = PIN_MASK(pin);
	idtshift = (pin & 0x7) << 2;	/* IDT shift bit */

	spin_lock_irqsave(&mp200_gio_lock, flag);

	idt = idt_table[pin / 8];
	if (pin < 32) {
		iia = VA_GIO + GP_IIAL;
		iir = VA_GIO + GP_IIRL;
	} else if (pin < 64) {
		iia = VA_GIO + GP_IIAH;
		iir = VA_GIO + GP_IIRH;
	} else if (pin < 96) {
		iia = VA_GIO + GP_IIAHH;
		iir = VA_GIO + GP_IIRHH;
	} else {
		iia = VA_GIO + GP_IIAHHH;
		iir = VA_GIO + GP_IIRHHH;
	}

	/* IIA enable -> disable */
	oiia = __raw_readl(iia);
	if ((oiia & mask) != 0) {
		__raw_writel((oiia & ~(mask)), iia);
	}

	/* set IDT */
	x = __raw_readl(idt);
	if ((x & (0xfU << idtshift)) != (mode << idtshift)) {
		x &= ~(0xfU << idtshift);
		x |= (mode << idtshift);
		__raw_writel(x, idt);
	}
	DPRINT("PIN %d Set IDT, addr=0x%08x, data=0x%08x\n",
					pin, idt, __raw_readl(idt));

	/* Interrupt clear */
	__raw_writel(mask, iir);

	/* Restore if changed */
	if ((oiia & mask) != 0) {
		__raw_writel(oiia, iia);
	}

	spin_unlock_irqrestore(&mp200_gio_lock, flag);

	DPRINT("Exit\n");

	return 0;
}

static void mp200_gio_irq_handler(unsigned int irq, struct irq_desc *desc)
{
	struct irq_desc *d;
	uint32_t ist, gio_irq;

	DPRINT("Enter\n");

	if ( (irq != INT_GIO0) && (irq != INT_GIO1) &&
		 (irq != INT_GIO2) && (irq != INT_GIO3) &&
		 (irq != INT_GIO4) && (irq != INT_GIO5) &&
		 (irq != INT_GIO6) && (irq != INT_GIO7)) {
		return;
	}

	desc->chip->mask(irq);

	do {
		if (irq == INT_GIO0) {
			gio_irq = INT_GPIO_0;
			ist = __raw_readl(VA_GIO + GP_MSTL) & 0xffffU;
		} else if (irq == INT_GIO1) {
			gio_irq = INT_GPIO_16;
			ist = __raw_readl(VA_GIO + GP_MSTL) >> 16;
		} else if (irq == INT_GIO2) {
			gio_irq = INT_GPIO_32;
			ist = __raw_readl(VA_GIO + GP_MSTH) & 0xffffU;
		} else if (irq == INT_GIO3) {
			gio_irq = INT_GPIO_48;
			ist = __raw_readl(VA_GIO + GP_MSTH) >> 16;
		} else if (irq == INT_GIO4) {
			gio_irq = INT_GPIO_64;
			ist = __raw_readl(VA_GIO + GP_MSTHH) & 0xffffU;
		} else if (irq == INT_GIO5) {
			gio_irq = INT_GPIO_80;
			ist = __raw_readl(VA_GIO + GP_MSTHH) >> 16;
		} else if (irq == INT_GIO6) {
			gio_irq = INT_GPIO_96;
			ist = __raw_readl(VA_GIO + GP_MSTHHH) & 0xffffU;
		} else {
			gio_irq = INT_GPIO_112;
			ist = __raw_readl(VA_GIO + GP_MSTHHH) >> 16;
		}

		if (ist == 0)
			break;

		while (ist) {
			if ((ist & 1) != 0) {
				DPRINT("PIN %d interrupt handler\n",
						gio_irq - INT_GPIO_0);
				d = irq_desc + gio_irq;
				d->handle_irq(gio_irq, d);
			}
			ist >>= 1;
			gio_irq++;
		}
	} while (1);

	desc->chip->unmask(irq);

	DPRINT("Exit\n");
}

static void mp200_gio_irq_ack(unsigned int irq)
{
	unsigned int pin = irq - INT_GPIO_BASE;
	unsigned int mask = PIN_MASK(pin);

	DPRINT("Enter\n");

	if ((irq < INT_GPIO_BASE) || (INT_GPIO_LAST < irq)) {
		return;
	}

	if (pin < 32) {
		__raw_writel(mask, (VA_GIO + GP_IIRL));
	} else if (pin < 64) {
		__raw_writel(mask, (VA_GIO + GP_IIRH));
	} else if (pin < 96) {
		__raw_writel(mask, (VA_GIO + GP_IIRHH));
	} else {
		__raw_writel(mask, (VA_GIO + GP_IIRHHH));
	}

#ifdef GPIO_DEBUG
	{
		int iir_table[] = {
			VA_GIO + GP_IIRL,  VA_GIO + GP_IIRH,
			VA_GIO + GP_IIRHH, VA_GIO + GP_IIRHHH
		};

		DPRINT("PIN %d clear interrupt, addr=0x%08x, mask=0x%08x\n",
					pin, iir_table[pin/32], mask);
	}
#endif

	DPRINT("Exit\n");
}

static void mp200_gio_irq_mask(unsigned int irq)
{
	unsigned int pin = irq - INT_GPIO_BASE;
	unsigned int mask = PIN_MASK(pin);

	DPRINT("Enter\n");

	if ((irq < INT_GPIO_BASE) || (INT_GPIO_LAST < irq)) {
		return;
	}

	/* disable */
	if (pin < 32) {
		__raw_writel(mask, (VA_GIO + GP_IDSL));
	} else if (pin < 64) {
		__raw_writel(mask, (VA_GIO + GP_IDSH));
	} else if (pin < 96) {
		__raw_writel(mask, (VA_GIO + GP_IDSHH));
	} else {
		__raw_writel(mask, (VA_GIO + GP_IDSHHH));
	}

#ifdef GPIO_DEBUG
	{
		int ids_table[] = {
			VA_GIO + GP_IDSL,  VA_GIO + GP_IDSH,
			VA_GIO + GP_IDSHH, VA_GIO + GP_IDSHHH
		};

		DPRINT("PIN %d irq mask, addr=0x%08x, mask=0x%08x\n",
					pin, ids_table[pin/32], mask);
	}
#endif

	DPRINT("Exit\n");
}

static void mp200_gio_irq_unmask(unsigned int irq)
{
	unsigned int iia;
	unsigned int pin = irq - INT_GPIO_BASE;
	unsigned int mask = PIN_MASK(pin);
	unsigned long flag;
	unsigned int iia_addr, ien_addr;
	uint32_t *accept;

	DPRINT("Enter\n");

	if ((irq < INT_GPIO_BASE) || (INT_GPIO_LAST < irq)) {
		return;
	}

	/* IRQ enable */
	if (pin < 32) {
		accept = (accept_bits + 0);
		iia_addr = VA_GIO + GP_IIAL;
		ien_addr = VA_GIO + GP_IENL;
	} else if (pin < 64) {
		accept = (accept_bits + 1);
		iia_addr = VA_GIO + GP_IIAH;
		ien_addr = VA_GIO + GP_IENH;
	} else if (pin < 96) {
		accept = (accept_bits + 2);
		iia_addr = VA_GIO + GP_IIAHH;
		ien_addr = VA_GIO + GP_IENHH;
	} else {
		accept = (accept_bits + 3);
		iia_addr = VA_GIO + GP_IIAHHH;
		ien_addr = VA_GIO + GP_IENHHH;
	}

	spin_lock_irqsave(&mp200_gio_lock, flag);
	if ((*accept & mask) == 0) {
		iia = __raw_readl(iia_addr);
		__raw_writel((iia | mask), iia_addr);
		*accept |= mask;
	}
	spin_unlock_irqrestore(&mp200_gio_lock, flag);

	__raw_writel(mask, ien_addr);

	DPRINT("PIN %d irq unmask, addr=0x%08x, mask=0x%08x\n",
				pin, ien_addr, mask);

	DPRINT("Exit\n");
}

static int __init mp200_gio_init(void)
{
	int i, gpio_num = 0, level_irq, idt_val[16], idt;

	DPRINT("Enter\n");

	for (i = 0; i < 16; i++) {
		idt_val[i] = __raw_readl(idt_table[i]);
	}

	/* setup default GIO Interrupt modes */
	for (i = INT_GPIO_BASE; i <= INT_GPIO_LAST; i++, gpio_num++) {
		level_irq = 0;
		idt = idt_val[gpio_num / 8];
		if (idt & (2 << (4 * (gpio_num & 0x7)))) {
			level_irq = 1;
		}
		if (level_irq) {
			set_irq_chip(i, &mp200_gio_chip);
			set_irq_handler(i, handle_level_irq);
		} else {
			set_irq_chip(i, &mp200_gio_chip_ack);
			set_irq_handler(i, handle_edge_irq);
		}
		set_irq_flags(i, IRQF_VALID);
	}

	set_irq_chained_handler(INT_GIO0, mp200_gio_irq_handler);
	set_irq_chained_handler(INT_GIO1, mp200_gio_irq_handler);
	set_irq_chained_handler(INT_GIO2, mp200_gio_irq_handler);
	set_irq_chained_handler(INT_GIO3, mp200_gio_irq_handler);
	set_irq_chained_handler(INT_GIO4, mp200_gio_irq_handler);
	set_irq_chained_handler(INT_GIO5, mp200_gio_irq_handler);
	set_irq_chained_handler(INT_GIO6, mp200_gio_irq_handler);
	set_irq_chained_handler(INT_GIO7, mp200_gio_irq_handler);

	DPRINT("Exit\n");

	return 0;
}

arch_initcall(mp200_gio_init);
