/* 2010-08-25: File added and changed by Sony Corporation */
/*
 *  File Name	    : linux/arch/arm/mach-mp200/mp200_pdma.c
 *  Function	    : MP200 PCM Direct interface
 *  Release Version : Ver 1.10
 *  Release Date    : 2009/7/7
 *
 *  Copyright (C) NEC Electronics Corporation 2008, 2009
 *
 *
 *  This program is free software;you can redistribute it and/or modify itunder the terms of
 *  the GNU General Public License as published by Free Softwere Foundation;either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

/*============================================================*/
/* include header                                             */
/*============================================================*/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/fs.h>

#include <asm/arch/io.h>
#include <asm/arch/pm.h>
#include <asm/arch/pmu.h>

#include <asm/arch/pcm_irq.h>

/*============================================================*/
/* constant definition                                        */
/*============================================================*/
#define DRIVER_NAME	    "mp200_pdma"
#define DRIVER_MAJOR    251

#define	INTC_IT0_BASE   IO_ADDRESS(MP200_INTC_ACPU_BASE)
#define	INTC_IT0_RAW0   (INTC_IT0_BASE + 0x0010)
#define	INTC_IT_LIIS    (INTC_IT0_BASE + 0x0320)
#define	INTC_IT_LIIR    (INTC_IT0_BASE + 0x0324)

/* #define PDMA_DEBUG_LOG */
#ifdef PDMA_DEBUG_LOG
#define DPRINTK(format,args...) printk(format,##args)
#else
#define DPRINTK(format,args...)
#endif

/*============================================================*/
/* static function definition                                 */
/*============================================================*/
static int mp200_pdma_open( struct inode* inode, struct file* filp );
static int mp200_pdma_release( struct inode* inode, struct file* filp );
static irqreturn_t mp200_pdma_interrupt(int irq, void *dev_id);
static irqreturn_t mp200_pcm0_interrupt( int irq, void *dev_id );
static irqreturn_t mp200_sw_interrupt( int irq, void *dev_id );

/*============================================================*/
/* static variable definition                                 */
/*============================================================*/
static char *pdma_id = "PDMA_interrupt";
static char *pcm_id = "PCM_interrupt"; 

#ifdef PDMA_DEBUG_LOG
static int pdma_int_ct = 0;
static int pcm0_int_ct = 0;
static int sw_int_ct = 0;
#endif
static int mp200_pdma_drv_open = 0;

/*============================================================*/
/*  interrupt Handler                                         */
/*============================================================*/
static irqreturn_t mp200_pdma_interrupt(int irq, void *dev_id)
{
    /* disable PDMA/PCM0 interrupt */
    disable_irq( INT_PCM0 );
    disable_irq( INT_PDMA );

#ifdef PDMA_DEBUG_LOG
    pdma_int_ct ++;
    DPRINTK("** PDMA_INTERRUPT %d **\n", pdma_int_ct);
#endif	
    mp200_pm_pdma_suspend_disable();

    /* clear SW interrupt 2 */
    __raw_writel( 0x00000004, INTC_IT_LIIR );
    /* set SW interrupt 0 */
    __raw_writel( 0x00000001, INTC_IT_LIIS );

    return IRQ_HANDLED;
}

static irqreturn_t mp200_pcm0_interrupt( int irq, void *dev_id )
{
    /* disable PDMA/PCM0 interrupt */
    disable_irq( INT_PDMA );
    disable_irq( INT_PCM0 );

#ifdef PDMA_DEBUG_LOG
    pcm0_int_ct ++;
    DPRINTK("** PCM0_INTERRUPT %d **\n", pcm0_int_ct);
#endif
    mp200_pm_pdma_suspend_disable();

    /* clear SW interrupt 3 */
    __raw_writel( 0x00000008, INTC_IT_LIIR );
    /* set SW interrupt 1 */
    __raw_writel( 0x00000002, INTC_IT_LIIS );

    return IRQ_HANDLED;
}

static irqreturn_t mp200_sw_interrupt( int irq, void *dev_id )
{
#ifdef PDMA_DEBUG_LOG
    sw_int_ct ++;
    DPRINTK("** SW_INTERRUPT%d %d **\n", (irq - INT_SSD0), sw_int_ct);
#endif
    /* clear SW interrupt 2 or 3 */
    __raw_writel( (0x00000001 << (irq - INT_SSD0)), INTC_IT_LIIR );
    mp200_pm_pdma_suspend_enable();

    enable_irq( INT_PDMA );
    enable_irq( INT_PCM0 );

    return IRQ_HANDLED;
}

/*============================================================*/
/*  open/close function                                       */
/*============================================================*/
/* ----------------------------------------------------------
    FUNCTION : mp200_pdma_open
    IMPORT   : inode, filp
    RETURN   : 0      : success
               -EBUSY : fale
    NOTE     : none
    UPDATE   : 2009.7.7
   ---------------------------------------------------------- */
static int mp200_pdma_open( struct inode* inode, struct file* filp )
{
    int ret=0, err=0;

    DPRINTK( KERN_INFO "%s\n", __FUNCTION__);
    
    mp200_pm_pdma_suspend_disable();
    if( mp200_pdma_drv_open ) {
        DPRINTK( KERN_INFO "mp200_pdma is being opened.\n");
        return -EBUSY;
    }

    /* sw interrupt clear */
    __raw_writel( 0x0000000f, INTC_IT_LIIR );
    /* request irq */
    /* PDMA */
    ret = mp200_pdma_request_irq( mp200_pdma_interrupt, IRQF_DISABLED, pdma_id, NULL, MP200_PCMMODE_DSP);
    DPRINTK("-- request_irq (INT_PDMA) ret = 0x%08x\n", ret);
    if( !ret ) {
        /* PCM0 */
        ret = mp200_pcm0_request_irq( mp200_pcm0_interrupt, IRQF_DISABLED, pcm_id, NULL, MP200_PCMMODE_DSP);
        DPRINTK("-- request_irq (INT_PCM0) ret = 0x%08x\n", ret);
    } else {
        err = 1;
    }
    if( !ret ) {
        /*  */
        ret = request_irq( INT_SSD2, &mp200_sw_interrupt, 0, "SW_INT2_interrupt", NULL);
        DPRINTK("-- request_irq (INT_SSD2) ret = 0x%08x\n", ret);
    } else if( !err ) {
        err = 2;
    }
    if( !ret ) {
        ret = request_irq( INT_SSD3, &mp200_sw_interrupt, 0, "SW_INT3_interrupt", NULL);
        DPRINTK("-- request_irq (INT_SSD3) ret = 0x%08x\n", ret);
    } else if( !err ) {
        err = 3;
    }
    if( ret && !err ) {
        err = 4;
    }

    switch( err ) {
    case 4:   free_irq( INT_SSD2, NULL );
    case 3:   mp200_pcm0_free_irq(NULL, MP200_PCMMODE_DSP);
    case 2:   mp200_pdma_free_irq(NULL, MP200_PCMMODE_DSP);
    case 1:   ret = -EBUSY;
              break;
    default:  mp200_pdma_drv_open = 1;
#ifdef PDMA_DEBUG_LOG
              pdma_int_ct = 0;
              pcm0_int_ct = 0;
              sw_int_ct = 0;
#endif
              break;
    }

	return ret;
}


/* ----------------------------------------------------------
    FUNCTION : mp200_pdma_release
    IMPORT   : inode, filp
    RETURN   : 0 : success
    NOTE     : none
    UPDATE   : 2009.7.7
  ----------------------------------------------------------- */
static int mp200_pdma_release( struct inode* inode, struct file* filp )
{
    DPRINTK( KERN_INFO "%s\n", __FUNCTION__);
#ifdef PDMA_DEBUG_LOG
    DPRINTK( KERN_INFO "pdma_int = %d, pcm0_int = %d, sw_int = %d\n", pdma_int_ct, pcm0_int_ct, sw_int_ct);
#endif
    
    if( mp200_pdma_drv_open ) {
        mp200_pm_pdma_suspend_disable();
        /* sw interrupt clear */
        __raw_writel( 0x0000000f, INTC_IT_LIIR );
        /* free irq */
        mp200_pdma_free_irq(NULL, MP200_PCMMODE_DSP);
        mp200_pcm0_free_irq(NULL, MP200_PCMMODE_DSP);
        free_irq( INT_SSD2, NULL );
        free_irq( INT_SSD3, NULL );
        mp200_pdma_drv_open = 0;
    }

    return 0;
}


/*============================================================*/
/*  init/end function                                         */
/*============================================================*/

static struct file_operations mp200_pdma_dev = 
{
    owner   : THIS_MODULE,
    open    : mp200_pdma_open,
    release : mp200_pdma_release,
};

/* ----------------------------------------------------------
    FUNCTION : mp200_pdma_init
    IMPORT   : none
    RETURN   : 0 : success
    NOTE     : none
    UPDATE   : 2009.7.7
   ---------------------------------------------------------- */
static int __init mp200_pdma_init( void )
{
    if ( register_chrdev( DRIVER_MAJOR, DRIVER_NAME, &mp200_pdma_dev ) ) {
        printk( KERN_INFO "%s : register_chrdev failed\n", __FUNCTION__);
        return -EBUSY;
    }

    mp200_pdma_drv_open = 0;
    DPRINTK( KERN_INFO "%s : loaded  into kernel\n", __FUNCTION__);

    return 0;
}


/* ----------------------------------------------------------
    FUNCTION : mp200_pdma_exit
    IMPORT   : none
    RETURN   : none
    NOTE     : none
    UPDATE   : 2009.7.7
  ----------------------------------------------------------- */
static void __exit mp200_pdma_exit( void )
{
    unregister_chrdev( DRIVER_MAJOR, DRIVER_NAME );

    DPRINTK( KERN_INFO "%s : removed from kernel\n", __FUNCTION__);
}


/*============================================================*/
/*  module definition                                         */
/*============================================================*/
module_init( mp200_pdma_init );
module_exit( mp200_pdma_exit );

MODULE_DESCRIPTION("MP200 PDMA driver for PCM Direct");
MODULE_LICENSE("GPL");
MODULE_ALIAS(DRIVER_NAME);
MODULE_AUTHOR("NEC Electronics");
