/* 2010-08-25: File added by Sony Corporation */
/*
 *  File Name	    : linux/arch/arm/mach-mp200/mp200_pmmu.c
 *  Function	    : Peripheral MMU descriptor management functions.
 *  Release Version : Ver 1.00
 *  Release Date    : 2007/10/1
 *
 *  Copyright (C) NEC Electronics Corporation 2006
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/fs.h>	/* for ERR_PTR() */

/* test implementation */
#include <linux/vmalloc.h>
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#include <asm/arch/mp200_pmmu.h>

/*
 * Manage set of descriptors.
 */
#define	PMMU_BITMAPSZ	(PGDIR_SIZE / PAGE_SIZE / BITS_PER_LONG)
struct pmmudata {
	struct  list_head pmmul;/* queue node linked to `pmmushead'. */
	u_long *descriptors;	/* base address of descriptor registers. */
	u_long *indices;	/* base address of index registers. */
	int     nentry;		/* number of entries. */
	struct  list_head freeh;/* free descriptor list head. */
	int	nfree;		/* number of free descriptors. */
	struct  list_head desch;/* active descriptor list head. */
				/* sort by virtual address. */
	struct proc_dir_entry *proc_descs;
	struct descmap {
		struct  list_head descl;/* linked to free or inuse list. */
		unsigned long baseaddr;	/* base virtual address. */
		pid_t   pid;
		u_long *descr[2];		/* descriptor register address. */
		u_long *indexr[2];		/* index register address. */
#define	F_VLD_INVALID	0
#define	F_VLD_VALID	1
		u_long  map[PMMU_BITMAPSZ];	/* 1bit/4Kbyte page, for 512pages. */
	} descmap[1];
};


/*
 * Hold PMMU control structures as a list for debugging purpose.
 */
LIST_HEAD(pmmushead);

#ifdef CONFIG_PROC_FS
/*
 * /proc/driver/pmmu directory
 */
static struct proc_dir_entry *proc_pmmu;



/*
 * Dump bitmap through /proc/driver/pmmu/<address>/bitmap.
 */
static int
mp200_pmmu_proc_read_bitmap(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	int i;
	struct pmmudata *p = (struct pmmudata *)data;
	struct descmap *d = p->descmap;

	len += sprintf(page + len, "pmmudata %lx\n", (unsigned long)p);

	for (i = 0; i < p->nentry; i++) {
		int j;

		for (j = 0; j < PMMU_BITMAPSZ; j++) {
			len += sprintf(page + len, "%08lx ", d->map[j]);
		}
		*(page + len) = '\n';
		len++;
		d++;
	}

	return len;
}

/*
 * Dump descriptors through /proc/driver/pmmu/<address>/bitmap.
 */
static int
mp200_pmmu_proc_read_descs(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	int i;
	struct pmmudata *p = (struct pmmudata *)data;
	struct descmap *d = p->descmap;

	for (i = 0; i < p->nentry; i++) {
		len += sprintf(page + len, "%08lx %08lx %08lx\n",
				d->baseaddr, *d->descr[0], *d->indexr[0]);
		len += sprintf(page + len, "%08lx %08lx %08lx\n",
				d->baseaddr, *d->descr[1], *d->indexr[1]);
		d++;
	}
	return len;
}
#endif

/*
 * Prepare using Peripheral MMU.
 */
void *
mp200_pmmu_initiate(u_long *descr, u_long *idxr, int nentry)
{
	struct pmmudata *p;
	struct descmap *d;
	int n, off;
	char procname[9]; /* eight digits + NUL */

	if (descr == NULL || idxr == NULL || nentry <= 0)
		return ERR_PTR(-EINVAL);
	if (nentry % 2 != 0)
		return ERR_PTR(-EINVAL);
	nentry = nentry / 2 ;


	/* Allocate space for control structures. */
	p = (struct pmmudata *)kmalloc(sizeof(struct pmmudata) + sizeof(struct descmap) * (nentry - 1), GFP_KERNEL);
	if (p == NULL)
		return ERR_PTR(-ENOMEM);

	/* Link into global list for debugging purpose. */
	list_add_tail(&p->pmmul, &pmmushead);

#ifdef CONFIG_PROC_FS
	/* Create procfs entries. */
	sprintf(procname, "%08lx", (unsigned long)p);
	p->proc_descs = proc_mkdir(procname, proc_pmmu);
	if (p->proc_descs != NULL) {
		if (!create_proc_read_entry("bitmaps", 0, p->proc_descs, mp200_pmmu_proc_read_bitmap, p)) {
			printk(KERN_INFO "can't create proc entry\n");
			/* return -1; XXX HN2 */
		}
		if (!create_proc_read_entry("descs", 0, p->proc_descs, mp200_pmmu_proc_read_descs, p)) {
			printk(KERN_INFO "can't create proc entry\n");
			/* return -1; XXX HN2 */
		}
	} else {
		printk(KERN_INFO "can't create proc directory\n");
	}
#endif

	/* Initialize list heads. */
	INIT_LIST_HEAD(&p->freeh);
	INIT_LIST_HEAD(&p->desch);
	/* Hold register map */
	p->descriptors = descr;
	p->indices = idxr;
	p->nentry = nentry;

	/*
	 * Two type register placement exists.
	 * There are index and descriptor alternately, or
	 * continuous indices and continuous descriptors.
	 */
	off = 1;
	if (descr + 1 == idxr || descr - 1 == idxr)
		off = 2; /* need to skip 1 word. */

	/* Initialize descriptor management structures. */
	d = p->descmap;
	memzero(d, sizeof(struct descmap) * nentry);
	for (n = 0; n < nentry; n++) {
		/* Enqueue desciptor management structure to free list. */
		list_add_tail(&d->descl, &p->freeh);
		d->descr[0] = descr + off * n * 2;
		d->indexr[0] = idxr + off * n * 2;
		d->descr[1] = descr + off * (n * 2 + 1 );
		d->indexr[1] = idxr + off * (n * 2 + 1 );
		/* baseaddr, pid and map are cleared by memzero() */
		d++;
	}
	p->nfree = nentry;
	return (void *)p;
}

/*
 * Within map, set bits from off to end-1.
 */
static void
mp200_pmmu_setbit(u_long *map, u_long off, u_long end)
{
	int bpos, blast;

	bpos = off >> PAGE_SHIFT;
	blast = end >> PAGE_SHIFT;
	while (bpos < blast) {
		map[bpos / BITS_PER_LONG] |= 1 << bpos % BITS_PER_LONG;
		bpos++;
	}
}

/*
 * Within map, clear bits from off to end-1.
 * Return map status. 0 means that map is fully cleared.
 */
static u_long
mp200_pmmu_clrbit(u_long *map, u_long off, u_long end)
{
	int bpos, blast;
	int n;
	u_long m;

	bpos = off >> PAGE_SHIFT;
	blast = end >> PAGE_SHIFT;
	while (bpos < blast) {
		map[bpos / BITS_PER_LONG] &= ~(1 << bpos % BITS_PER_LONG);
		bpos++;
	}
	for (n = m = 0; n < PMMU_BITMAPSZ; n++)
		m |= map[n];
	return m;
}


#if 0
void
show_area_pte(pmd_t *pmd, unsigned long addr, unsigned long size)
{
	pte_t *pte;
	unsigned long end;

	if (pmd_none(*pmd))
		return;	/* Never happen on ARM */
	if (pmd_bad(*pmd)) {
		return;	/* Never happen on ARM */
	}
	pte = pte_offset_kernel(pmd, addr);
	addr &= ~PMD_MASK;
	end = addr + size;
	if (end > PMD_SIZE)
		end = PMD_SIZE;
	do {
		pte_t page;

		page = *pte;
		addr += PAGE_SIZE;
		pte++;
		if (pte_none(page)) {
			printk(KERN_INFO "PTE %x %x\n", addr, pte_val(page));
			continue;
		}
		if (pte_present(page)) {
			printk(KERN_INFO "PTE %x %x\n", addr, pte_val(page));
			continue;
		}
		printk(KERN_CRIT "Swapped out page\n");
	} while (addr < end);
}

/*
 * Search specified virtual address in the PGD.
 */
static int
mp200_pmmu_search(unsigned long addr, unsigned long size)
{
	pgd_t *pgdir;
	unsigned long end = addr + size;

	pgdir = pgd_offset_k(addr);
	do {
		pmd_t *pmd;
		unsigned long pgdoff, pmdend;
		if (pgd_none(*pgdir))
			return 1;	/* Never happen on ARM */
		if (pgd_bad(*pgdir))
			return 1;	/* Never happen on ARM */
		pmd = pmd_offset(pgdir, addr);
		pgdoff = addr & ~PGDIR_MASK;
		pmdend = pgdoff + end - addr;
		if (pmdend > PGDIR_SIZE)
			pmdend = PGDIR_SIZE;
		do {
			show_area_pte(pmd, pgdoff, pmdend - pgdoff);
			pgdoff = (pgdoff + PMD_SIZE) & PMD_MASK;
			pmd++;
		} while (pgdoff < pmdend);

		printk(KERN_INFO "PGD %x\n", pgd_val(*pgdir));
		addr = (addr + PGDIR_SIZE) & PGDIR_MASK;
		pgdir++;
	} while (addr && addr < end);

	return 0;
}

/* Can't access vmlist from loadable kernel module. */
/*
 * Search specified virtual address in the vmlist.
 */
static int
mp200_pmmu_search_vmlist(void *addr)
{
	struct vm_struct *vm;

	read_lock(&vmlist_lock);
	if (vmlist == NULL) {
		read_unlock(&vmlist_lock);
		return -1; /* XXX HN2 return pointer to PMD? */
	}
	for (vm = vmlist; vm != NULL; vm = vm->next) {
		if (vm->addr == addr) {
			printk(KERN_INFO "flags = %x addr = %x size = %x\n",
			vm->flags, vm->addr, vm->size);
		}
	}
	read_unlock(&vmlist_lock);

	return 0;
}
#endif

/*
 * Return 0 if empty page exists.
 * Else return 1.
 */
static int
check_area_pte(pmd_t *pmd, unsigned long addr, unsigned long size)
{
	pte_t *pte;
	unsigned long end;

	if (pmd_none(*pmd))
		return 0;	/* Never happen on ARM */
	if (pmd_bad(*pmd)) {
		return 0;	/* Never happen on ARM */
	}
	pte = pte_offset_kernel(pmd, addr);
	addr &= ~PMD_MASK;
	end = addr + size;
	if (end > PMD_SIZE)
	{
		end = PMD_SIZE;
	}
	do {
		pte_t page;

		page = *pte;
		addr += PAGE_SIZE;
		pte++;
		if (pte_none(page)) {
			return 0;
		}
		if (pte_present(page)) {
			continue;
		}
		return 0;	/* Swapped out page. */
	} while (addr < end);

	return 1;
}

/*
 * Check pages present for the region, or not.
 * Return 0 if region contains empty space (page not present).
 * This routine checks page table and ensure that
 * pages are present at that time. On the region allocated by
 * inapropriate method (e.g. mmap(2) without mlock(2)),
 * page may not be present in near future and PMMU will fault.
 */
static int
is_page_present(unsigned long addr, unsigned long size, pid_t pid)
{
	pgd_t *pgdir;
	unsigned long end = addr + size;

	pgdir = pgd_offset_k(addr);
	do {
		pmd_t *pmd;
		unsigned long pmdoff, pmdend;
		if (pgd_none(*pgdir))
			return 0;	/* Never happen on ARM */
		if (pgd_bad(*pgdir))
			return 0;	/* Never happen on ARM */
		pmd = pmd_offset(pgdir, addr);
		pmdoff = addr & ~PGDIR_MASK;
		pmdend = pmdoff + end - addr;
		if (pmdend > PGDIR_SIZE)
			pmdend = PGDIR_SIZE;
		do {
			if (!check_area_pte(pmd, pmdoff, pmdend - pmdoff)) {
				return 0; /* Error */
			}
			pmdoff = (pmdoff + PMD_SIZE) & PMD_MASK;
			pmd++;
		} while (pmdoff < pmdend);

		addr = (addr + PGDIR_SIZE) & PGDIR_MASK;
		pgdir++;
	} while (addr && addr < end);

	return 1;
}

/*
 * Map virtual address (allocate register if needed)
 */
int
mp200_pmmu_map(void *id, void *start, size_t length, pid_t pid)
{
	struct list_head *l;
	struct pmmudata *p;
	struct descmap *d= NULL;
	unsigned long ssec, esec, end;
	int mapped;

	if (id == NULL)
		return -EINVAL;
	if (start == NULL || VMALLOC_VMADDR(start) & ~PAGE_MASK || VMALLOC_VMADDR(start) < TASK_SIZE)
		return -EINVAL;
	if (length == 0 || start + length < start)
		return -EINVAL;


/* PGDIR and PMD are chaoticically used in Linux/ARM. */
/* Probably, Linux PGDIR are correspond to ARM SECTION. */
#define	SECTION_ALIGN(addr)	(((addr) + PGDIR_SIZE - 1) & PGDIR_MASK)

/* XXX HN2 should check context? */

	if (!is_page_present(VMALLOC_VMADDR(start), length, pid)) {
		/* Specified virtual address region contains bad page. */
		return -ENOENT;
	}

	p = (struct pmmudata *)id;
	/* Compute section aligned address that cover specfied region. */
	ssec = VMALLOC_VMADDR(start) & PGDIR_MASK;
	/* Section previous to escs contains end address of specified region. */
	esec = SECTION_ALIGN(VMALLOC_VMADDR(start) + length);
	end = VMALLOC_VMADDR(start) + length;
	/* XXX HN2 wrap around */

	/*
	 * Count how many sections are already mapped.
	 */
	l = p->desch.next;
	mapped = 0;

	while (ssec < esec) {
		while (l !=  &p->desch) {
			d = (struct descmap *)l;
			if (ssec <= d->baseaddr)
				break;
			l = l->next;
		}
		/* Don't need new descriptor when list is not empty and
		   address is match to existing one. */
		if (l !=  &p->desch && /* Ensure d was initialized */
		    d->baseaddr == ssec)
		{
			mapped++;
		}
		ssec += PGDIR_SIZE;
	}
	/* XXX rewind ssec HN2 */
	ssec = VMALLOC_VMADDR(start) & PGDIR_MASK;


	/*
	 * Check whether enough free descriptors exist or not.
	 */
	if (p->nfree < (esec - ssec) / PGDIR_SIZE - mapped) {
		return -ENFILE; /* There is not enough free descriptor. */
	}

	/*
	 * Map for region by every PGDIR_SIZE.
	 * There is enough free descriptors, following process will not fail.
	 */
	l = p->desch.next;
	while (ssec < esec) {
		unsigned long soff, eoff;
		/*
		 * New entry will link into previous to l.
		 * So, we can restart from l.
		 */
		while (l != &p->desch) {
			d = (struct descmap *)l;
			if (ssec <= d->baseaddr) {
				break;
			}
			l = l->next;
		}
		/*
		 * Found already mapped region or not.
		 */
		if (l == &p->desch || /* Ensure d was initialized */
		    d->baseaddr != ssec) {
			struct descmap *dmap;
			pgd_t *pgdir;

			/*
			 * Hunt new map management structure.
			 */
			if (list_empty(&p->freeh))
				return -ENFILE; /* No free descriptor, never happen */
			p->nfree--;
			dmap = (struct descmap *)p->freeh.next;
			list_del(p->freeh.next);
			/*
			 * Setup descriptor control structure.
			 *  - Hold translate base address.
			 *  - Insert into work queue.
			 */
			dmap->baseaddr = ssec;
			/*
			 * Add new section to the list.
			 * L points higher address already mapped.
			 * We should insert new map previous to that pointed by l.
			 */
			list_add(&dmap->descl, l->prev);
			/*
			 * Set descriptor and index register.
			 */
			pgdir = pgd_offset_k(ssec);
			*dmap->descr[0] = (*pgdir)[0];
			*dmap->descr[1] = (*pgdir)[1];
				/* XXX HN2 need mask off? */
			*dmap->indexr[0] = ssec | F_VLD_VALID;
			*dmap->indexr[1] = (ssec+ PGDIR_SIZE/2) | F_VLD_VALID;
			/*
			 * fall through.
			 */
			d = dmap;
		}
		/*
		 * Specified start and end address may not be aligned.
		 * Although middle of the region (if it's bigger than
		 * PGDIR_SIZE) can be processed by PGDIR_SIZE size,
		 * beginning and ending part of the region may be
		 * smaller than PGDIR_SIZE. We must compute size
		 * because of them.
		 */
		soff = max(VMALLOC_VMADDR(start), ssec) & (PGDIR_SIZE - 1);
		eoff = end < ssec + PGDIR_SIZE ? end & (PGDIR_SIZE - 1) : PGDIR_SIZE;
		mp200_pmmu_setbit(d->map, soff, eoff);

		/* Done until end of PGDIR */
		ssec += PGDIR_SIZE;
	}
	/*
	 * We don't need to flush cache now.
	 * _vmalloc_area_pages() has flushed cache using flush_cache_all(),
	 * so page table data is already written to memory.
	 *
	 * When we add support for mmap()ed region,
	 * it will need to verify that whether page tables for
	 * mmap()ed region are flushed or not.
	 */
	return 0;
}

/*
 * Unmap virtual address (release region, and descriptor register if possible)
 */
int
mp200_pmmu_unmap(void *id, void *start, size_t length, pid_t pid)
{
	struct pmmudata *p;
	unsigned long ssec, esec, end;
	struct list_head *l;
	struct descmap *d = NULL;

/* XXX HN2 should check context? */

	if (id == NULL)
		return -EINVAL;
	if (start == NULL || VMALLOC_VMADDR(start) & ~PAGE_MASK || VMALLOC_VMADDR(start) < TASK_SIZE)
		return -EINVAL;
	if (length == 0 || start + length < start)
		return -EINVAL;

	p = (struct pmmudata *)id;
	/* Compute section aligned address that cover specfied region. */
	ssec = VMALLOC_VMADDR(start) & PGDIR_MASK;
	/* Section previous to escs contains end address of specified region. */
	esec = SECTION_ALIGN(VMALLOC_VMADDR(start) + length);
	end = VMALLOC_VMADDR(start) + length;

	/* Check mapped sections by every PGDIR_SIZE size for region. */
	l = p->desch.next;
	while (ssec < esec) {
		unsigned long soff, eoff;

		/* Search until reaching to end of list or
		   finding matched entry. Keep pointer `l' at that place
		   so that restarting loop traverse. */
		while (l != &p->desch) {
			d = (struct descmap *)l;
			if (ssec <= d->baseaddr) {
				break;	/* found? */
			}
			l = l->next;
		}

		if (l == &p->desch || d->baseaddr != ssec) {
			/* This section is not mapped. Skip to next. */
			ssec += PGDIR_SIZE;
			continue;
		}
		/*
		 * Specified start and end address may not be aligned.
		 * Although middle of the region (if it's bigger than
		 * PGDIR_SIZE) can be processed by PGDIR_SIZE size,
		 * beginning and ending part of the region may be
		 * smaller than PGDIR_SIZE. We must compute start and end
		 * offset to PGDIR_SIZE because of them.
		 */
		soff = max(VMALLOC_VMADDR(start), ssec) & (PGDIR_SIZE - 1);
		eoff = end < ssec + PGDIR_SIZE ? end & (PGDIR_SIZE - 1) : PGDIR_SIZE;
		if (mp200_pmmu_clrbit(d->map, soff, eoff) == 0) {
			struct list_head *c;
			/*
			 * This section is no longer used.
			 * Invalidate descriptor in PMMU.
			 */
			*d->indexr[0] = F_VLD_INVALID;
			*d->indexr[1] = F_VLD_INVALID;
			/*
			 * Update next search starting point so that
			 * we can remove the entry from list.
			 */
			c = l;
			l = l->next;
			/*
			 * Free map management structure.
			 */
			list_del(c);
			list_add_tail(c, &p->freeh);
			/*
			 * map[] is cleard. There are not needed to clear
			 * baseaddr and pid, leave old value in them.
			 */
			p->nfree++;
		}

		ssec += PGDIR_SIZE;
	}
	return 0;
}

/*
 * End using descriptor register set.
 */
int
mp200_pmmu_terminate(void *id)
{
	struct pmmudata *p;
	struct descmap *d;
	char procname[9];
	int i;

	if (id == NULL)
		return -EINVAL;

	p = (struct pmmudata *)id;

	/* Invalidate all descriptors. */
	d = p->descmap;
	for (i = 0; i < p->nentry; i++) {
		*d->indexr[0] = F_VLD_INVALID;
		*d->indexr[0] = F_VLD_INVALID;
		d++;
	}

	/* Delete from global list. */
	list_del(&p->pmmul);

#ifdef CONFIG_PROC_FS
	/* remove proc entry */
	remove_proc_entry("bitmaps", p->proc_descs);
	remove_proc_entry("descs", p->proc_descs);
	sprintf(procname, "%08lx", (unsigned long)p);
	remove_proc_entry(procname, proc_pmmu);
#endif
	/* Release memory */
	kfree(id);

	return 0;
}

static int __init
mp200_pmmu_init(void)
{
#ifdef CONFIG_PROC_FS
	proc_pmmu = proc_mkdir("driver/pmmu", NULL);
	if (proc_pmmu == NULL) {
		printk(KERN_INFO "can't create proc directory\n");
		/* return -1; */
	}
#endif
	return 0;
}

static void __exit
mp200_pmmu_exit(void)
{
#ifdef CONFIG_PROC_FS
	remove_proc_entry("driver/pmmu", NULL); /* ??? XXX HN2 */
#endif
}


EXPORT_SYMBOL(mp200_pmmu_initiate);
EXPORT_SYMBOL(mp200_pmmu_map);
EXPORT_SYMBOL(mp200_pmmu_unmap);
EXPORT_SYMBOL(mp200_pmmu_terminate);

module_init(mp200_pmmu_init);
module_exit(mp200_pmmu_exit);
