/* 2010-08-25: File added and changed by Sony Corporation */
/*!
 * @file
 * @brief define file operation and interruption interface for Linux 2.4
 * @author NIS
 * @date 2007/9/10
 */

#include <linux/init.h> // __init
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mm.h>      // remap_page_range
//#include <linux/wrapper.h> // mem_map_reserve
#include <linux/slab.h> // kmalloc
#include <linux/poll.h> // POLLIN
#include <linux/interrupt.h> // tasklet
#include <linux/dma-mapping.h>
#include <linux/platform_device.h>
#include <linux/cdev.h>
#include <asm/semaphore.h>
#include <asm/delay.h>
#include <asm/io.h>  // ioremap
#include <asm/irq.h> // request_irq
#include <asm/hardware.h> // AVC base address
//#include <asm/arch/pm.h>  // mp200_pmu_frqctrl_on
#include <asm/arch/pmu.h> // clock
//#include <asm/arch/stm_n.h> // power management

#include "avc_info.h"
#include "avc_lib.h"
#include "avc_debug.h"

#if 1
static struct avc_info_t s_avc_info = {
    .regs = (struct avc_regs_t *)IO_ADDRESS(MP200_AVC_BASE),
};
#endif

#define AVC_MODNAME	"mp200_avc"
#define AVC_DEVNAME	"avc"

static char *devname = AVC_DEVNAME; //!< default device file name
static int devmajor = 125;    //!< default device major number
#if 1
struct avc_info_t *avc_info = &s_avc_info; //!< this driver information
#else
struct avc_info_t *avc_info = NULL; //!< this driver information
#endif

static void avc_tasklet_handler(unsigned long value);
static DECLARE_TASKLET(avc_tasklet, avc_tasklet_handler, 0);
static DECLARE_MUTEX(avc_mutex);
static DECLARE_WAIT_QUEUE_HEAD(readq);
static spinlock_t avc_lock = SPIN_LOCK_UNLOCKED;

extern int command_start(unsigned short command, unsigned short last_frame);
extern int avc_interruption_handler(unsigned int factor, unsigned int int_a);
extern int avc_create_proc(void);
extern int avc_remove_proc(void);
extern struct avc_driver_version_t avc_driver_version;

static void initialize_common_info(struct avc_common_info_t *common_info);

#if 1
#ifdef CONFIG_PM
static int avc_pf_suspend(struct platform_device *dev, pm_message_t state);
static int avc_pf_resume(struct platform_device *dev);
#endif
static int avc_pf_probe(struct platform_device *dev);
static int avc_pf_remove(struct platform_device *dev);

static void avc_pf_release(struct device *dev);

static struct class *avc_class;
static struct class_device *avc_class_device;

static struct platform_device avc_pf_device = {
    .name = AVC_MODNAME,
    .id = -1,
    .dev = {
      .release = avc_pf_release,
    },
};

static struct platform_driver avc_pf_driver = {
    .probe = avc_pf_probe,
    .remove = avc_pf_remove,
#ifdef CONFIG_PM
    .suspend = avc_pf_suspend,
    .resume = avc_pf_resume,
#endif
    .driver = {
	.name  = AVC_MODNAME,
	.owner = THIS_MODULE,
    },
};
#endif

static int avc_open_state = 0;

/* file operations ----------------------------------------------------------*/

/*!
 * read file operation
 * @param[in] filp
 * @param[out] buf to copy struct avc_interrupt_t
 * @param[in] count buf size
 * @param[in] offp
 * @retval size size of struct avc_interrupt_t
 * @retval -EINVAL count error
 */
static ssize_t avc_read(struct file *filp, char *buf,
                        size_t count, loff_t *offp)
{
    int minor = MINOR(filp->f_dentry->d_inode->i_rdev);
    unsigned long flags;
    struct avc_interrupt_t tmp_int;
    int ret;
    debug1("minor %d, count %d",minor,count);

    if (count < sizeof(avc_info->interrupt)) {
        debug1("count < sizeof(avc_info->interrupt)");
        return -EINVAL;
    }

    if (down_interruptible(&avc_mutex)) {
        debug0("down_interruptible failed");
        return -ERESTARTSYS;
    }

    if (wait_event_interruptible(readq,
                                 (
                                  (avc_info->interrupt.interrupt.l  != 0) ||
                                  (avc_info->interrupt.int_buf_full != 0))
                                 )
        ) {
        debug0("wait_event_interruptible failed");
        ret = -ERESTARTSYS;
	goto up_exit;
    }

    spin_lock_irqsave(&avc_lock, flags);
    tmp_int = avc_info->interrupt;
    memset(&avc_info->interrupt, 0, sizeof(avc_info->interrupt));// clear
    spin_unlock_irqrestore(&avc_lock, flags);

    if (copy_to_user(buf, &tmp_int, sizeof(tmp_int))) {
        debug0("copy_to_user failed");
        ret = -EFAULT;
	goto up_exit;
    }

    ret = sizeof(avc_info->interrupt);

 up_exit:
    up(&avc_mutex);
    return ret;
}

/*!
 * select file operation
 * @param[in] filp
 * @param[in] wait
 * @retval mask POLLIN | POLLRDNORM
 */
static unsigned int avc_poll(struct file *filp,
                             struct poll_table_struct *wait)
{
    unsigned int mask = 0;
    debug2("minor %d",MINOR(filp->f_dentry->d_inode->i_rdev));

    poll_wait(filp, &readq, wait);
    if (avc_info->interrupt.interrupt.l || avc_info->interrupt.int_buf_full) {
        mask |= POLLIN | POLLRDNORM;
    }

    debug2("mask 0x%X",mask);
    return mask;
}

/*!
 * ioctl file operation
 * @param[in] inode
 * @param[in] filp
 * @param[in] cmd AVC_IOCQBUFSIZE or AVC_IOCSREGCMD
 * @param[in] arg struct avc_command_t with AVC_IOCSREGCMD
 * @retval size for AVC_IOCQBUFSIZE
 * @retval 0 command successful
 * @retval -EINVAL command error
 */
static int avc_ioctl(struct inode *inode, struct file *filp,
                     unsigned int cmd, unsigned long arg)
{
    struct avc_command_t avc_command;
    struct avc_register_access_t ra;
    unsigned long flags;
    int ret = 0;

    switch (cmd) {

    case AVC_IOCQBUFSIZE:
        ret = PAGE_ALIGN(sizeof(struct avc_common_info_t));
        debug1("AVC_IOCQBUFSIZE %d",ret);
        break;

    case AVC_IOCSREGCMD:
        debug1("AVC_IOCSREGCMD 0x%08lX",arg);
        if (arg == 0) {
            ret = -EINVAL;
            break;
        }
        if (copy_from_user(&avc_command, (void *)arg, sizeof(avc_command))) {
            debug0("copy_from_user failed");
            ret = -EFAULT;
            break;
        }
        spin_lock_irqsave(&avc_lock, flags);
        if (command_start(avc_command.command, avc_command.last_frame) < 0) {
            debug1("command_start failed");
            ret = -EINVAL;
        }
        spin_unlock_irqrestore(&avc_lock, flags);
        break;

    case AVC_IOCSREGISTER:
    case AVC_IOCGREGISTER:
        debug1("AVC_IOCxREGISTER");
        if (arg == 0) {
            ret = -EINVAL;
            break;
        }
        if (copy_from_user(&ra, (void *)arg, sizeof(ra))) {
            debug0("copy_from_user failed");
            ret = -EFAULT;
            break;
        }
        {
            char *addr = (char *)avc_info->regs + ra.address;
            if (ra.address >= sizeof(struct avc_regs_t)) {
                ret = -EINVAL;
                break;
            }
            spin_lock_irqsave(&avc_lock, flags);
            if (ra.address >= offsetof(struct avc_regs_t, bar1)) {
                // 32bit access
                if (cmd == AVC_IOCSREGISTER) {
                    avc_writel(ra.value, addr);
                } else {
                    ra.value = readl(addr);
                }
            } else {
                // 16bit access
                if (cmd == AVC_IOCSREGISTER) {
                    avc_writew((unsigned short)ra.value, addr);
                } else {
                    ra.value = (unsigned int)readw(addr);
                }
            }
            spin_unlock_irqrestore(&avc_lock, flags);
            if (cmd == AVC_IOCGREGISTER &&
                copy_to_user((void *)arg, &ra, sizeof(ra))) {
                debug0("copy_to_user failed");
                ret = -EFAULT;
                break;
            }
            debug1(" address 0x%X",ra.address);
            debug1(" value   0x%X",ra.value);
        }
        break;

    case AVC_IOCGVERSION:
        debug1("AVC_IOCGVERSION");
        debug1(" v%d.%d.%d",avc_info->version.major,
               avc_info->version.minor, avc_info->version.local);
        if (arg == 0) {
            ret = -EINVAL;
            break;
        }
        if (copy_to_user((void *)arg, &avc_info->version,
                         sizeof(avc_info->version))) {
            debug0("copy_to_user failed");
            ret = -EFAULT;
            break;
        }
        break;

    case AVC_IOCSDEBUGLEVEL:
        debug1("AVC_IOCSDEBUGLEVEL %lu",arg);
        debug_level = (int)arg;
        break;

    default:
        debug1("unknown cmd 0x%08X, arg 0x%08lX",cmd,arg);
        ret = -EINVAL; //ENOTTY
        break;
    }

    debug1("ret %d",ret);
    return ret;
}

#if 0
/*!
 * allocate memory from free pages
 * @param[in] size
 * @retval va allocated kernel virtual address
 */
static void *alloc_from_free_pages(size_t size)
{
    int aligned_size = PAGE_ALIGN(size); // Align size to next page boundary
    unsigned long va; // kernel virtual address
    unsigned long pa; // page address

    debug1("size %d, aligned_size %d, order %d, PAGE_SIZE %d",
           size,aligned_size,get_order(aligned_size),PAGE_SIZE);

    // Attempt to allocate contiguous memory
    va = __get_free_pages(GFP_KERNEL, get_order(aligned_size));
    if (va == 0) {
        debug0("__get_free_pages() failed");
        return NULL;
    }

    debug1("allocated to 0x%08lX", va);

    // set PG_reserved bit for mmap
    for (pa=va; pa<(va+aligned_size); pa+=PAGE_SIZE) {
        mem_map_reserve(virt_to_page(pa));
    }

    return (void *)va;
}

/*!
 * free memory allocated by alloc_from_free_pages
 * @param[in] ptr address given by alloc_from_free_pages
 * @param[in] size the same value at alloc_from_free_pages
 * @return
 */
static void free_to_free_pages(void *ptr, size_t size)
{
    int aligned_size = PAGE_ALIGN(size);
    unsigned long va = (unsigned long)ptr;
    unsigned long pa;

    debug1("ptr 0x%08lX, size %d, aligned_size %d, order %d",
           va,size,aligned_size,get_order(aligned_size));

    // unset PG_reserved bit
    for (pa=va; pa<(va+aligned_size); pa+=PAGE_SIZE) {
        mem_map_unreserve(virt_to_page(pa));
    }

    free_pages(va, get_order(aligned_size));
}
#endif

/*!
 * mmap file operation
 * @param[in] filp
 * @param[in] vma
 * @retval 0 successful
 * @retval -EINVAL size error
 */
static int avc_mmap(struct file *filp, struct vm_area_struct *vma)
{
    unsigned long size = vma->vm_end - vma->vm_start;
    unsigned long info_size = PAGE_ALIGN(sizeof(struct avc_common_info_t));
#if 1
    unsigned long info_pa = (unsigned long)avc_info->pa_common_info;
#else
    unsigned long info_pa = __pa(avc_info->common_info);
#endif

    debug1("vm_start 0x%08lX", vma->vm_start);
    debug1("vm_end   0x%08lX", vma->vm_end);
    debug1("size     %lu", size);
    debug1("vm_pgoff 0x%08lX", vma->vm_pgoff);
    debug1("vm_page_prot 0x%08lX", pgprot_val(vma->vm_page_prot));

    if (size > info_size) {
        debug1("error request size %lu > allocated size %lu",
               size,info_size);
        return -EINVAL;
    }

#if 1
    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
    if (remap_pfn_range (vma, vma->vm_start, info_pa >> PAGE_SHIFT, size,
			 vma->vm_page_prot)) {
        debug1("remap_pfn_range failed");
        return -EAGAIN;
    }
#else
    if (remap_page_range(//vma,
                         vma->vm_start, info_pa, size, vma->vm_page_prot)) {
        debug1("remap_page_range failed");
        return -EAGAIN;
    }
#endif

    initialize_common_info(avc_info->common_info);
    debug1("mapped to 0x%08lX", info_pa);
    return 0;
}

static void print_clock_status(void)
{
    debug2("MP200_CLOCKGATE_AVC_HSCLK %d",
           mp200_pmu_get_clockgate(MP200_CLOCKGATE_AVC_HSCLK));
    debug2("MP200_CLOCKGATE_AVC_HMCLK %d",
           mp200_pmu_get_clockgate(MP200_CLOCKGATE_AVC_HMCLK));
    debug2("MP200_CLOCKGATE_AVC_CLKM %d",
           mp200_pmu_get_clockgate(MP200_CLOCKGATE_AVC_CLKM));
    debug2("MP200_CLOCKGATE_AVC_CLKE %d",
           mp200_pmu_get_clockgate(MP200_CLOCKGATE_AVC_CLKE));
    debug2("MP200_CLOCKGATE_AVC_CLKD %d",
           mp200_pmu_get_clockgate(MP200_CLOCKGATE_AVC_CLKD));
    debug2("MP200_CLOCKGATE_AVC_CLKC %d",
           mp200_pmu_get_clockgate(MP200_CLOCKGATE_AVC_CLKC));
    debug2("MP200_RESETDEVICE_AVC %d",
           mp200_pmu_get_reset_status(MP200_RESETDEVICE_AVC));
}

/*!
 * stop AVC module clock supply
 * @param void
 * @return void
 */
static void avc_clock_off(void)
{
    /* disable automatic clock supply control */
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCHSCK);
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCHMCK);
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCCKM);
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCCKE);
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCCKD);
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCCKC);

    /* reset */
    mp200_pmu_reset_device(MP200_RESETDEVICE_AVC);

    /* close all clock gate for AVC */
    mp200_pmu_close_clockgate(MP200_CLOCKGATE_AVC_HSCLK);
    mp200_pmu_close_clockgate(MP200_CLOCKGATE_AVC_HMCLK);
    mp200_pmu_close_clockgate(MP200_CLOCKGATE_AVC_CLKM);
    mp200_pmu_close_clockgate(MP200_CLOCKGATE_AVC_CLKE);
    mp200_pmu_close_clockgate(MP200_CLOCKGATE_AVC_CLKD);
    mp200_pmu_close_clockgate(MP200_CLOCKGATE_AVC_CLKC);

    print_clock_status();
}

#define SMU_L3_POWERSW_BUZ	(IO_ADDRESS(MP200_SMU_BASE) + 0x50C)
#define SMU_LOG3SW_ACTRLEN	(IO_ADDRESS(MP200_SMU_BASE) + 0x510)
#define SMU_LOG3SW_ACTRL	(IO_ADDRESS(MP200_SMU_BASE) + 0x514)

/*!
 * start AVC module clock supply
 * @param void
 * @return void
 */
static void avc_clock_on(void)
{
    /* disable automatic clock supply control (for reset). */
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCHSCK);
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCHMCK);
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCCKM);
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCCKE);
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCCKD);
    mp200_pmu_clkctrl_off(MP200_CLKCTRL_AVCCKC);

    /* open all clock gate for AVC */
    mp200_pmu_open_clockgate(MP200_CLOCKGATE_AVC_HSCLK);
    mp200_pmu_open_clockgate(MP200_CLOCKGATE_AVC_HMCLK);
    mp200_pmu_open_clockgate(MP200_CLOCKGATE_AVC_CLKM);
    mp200_pmu_open_clockgate(MP200_CLOCKGATE_AVC_CLKE);
    mp200_pmu_open_clockgate(MP200_CLOCKGATE_AVC_CLKD);
    mp200_pmu_open_clockgate(MP200_CLOCKGATE_AVC_CLKC);

    udelay(1);

    avc_writel(0x00010100, SMU_L3_POWERSW_BUZ);

    /* release reset */
    mp200_pmu_unreset_device(MP200_RESETDEVICE_AVC);

    /* mp200_pmu_unreset_device() waits apropriate period.
       Now, we can enable automatic clock control. */
    mp200_pmu_clkctrl_on(MP200_CLKCTRL_AVCHSCK);
    mp200_pmu_clkctrl_on(MP200_CLKCTRL_AVCHMCK);
    mp200_pmu_clkctrl_on(MP200_CLKCTRL_AVCCKM);
    mp200_pmu_clkctrl_on(MP200_CLKCTRL_AVCCKE);
    mp200_pmu_clkctrl_on(MP200_CLKCTRL_AVCCKD);
    mp200_pmu_clkctrl_on(MP200_CLKCTRL_AVCCKC);

    print_clock_status();
}


static void print_smu_reg(void)
{
    debug2("SMU_L3_POWERSW_BUZ 0x%X",readl(SMU_L3_POWERSW_BUZ));
    debug2("SMU_LOG3SW_ACTRLEN 0x%X",readl(SMU_LOG3SW_ACTRLEN));
    debug2("SMU_LOG3SW_ACTRL   0x%X",readl(SMU_LOG3SW_ACTRL));
}

/*!
 * start AVC module power supply
 * @param void
 * @retval 0 successful
 * @retval -1 failed
 */
static int avc_power_on(void)
{
//    unsigned long reg;
//    int i;

    avc_writel(0x0000007F, SMU_L3_POWERSW_BUZ);
    avc_writel(0x0000003F, SMU_L3_POWERSW_BUZ);
    avc_writel(0x0000001F, SMU_L3_POWERSW_BUZ);
    avc_writel(0x0000000F, SMU_L3_POWERSW_BUZ);
    avc_writel(0x00000007, SMU_L3_POWERSW_BUZ);
    avc_writel(0x00000003, SMU_L3_POWERSW_BUZ);
    avc_writel(0x00000001, SMU_L3_POWERSW_BUZ);
    avc_writel(0x00000000, SMU_L3_POWERSW_BUZ);

    return 0;
#if 0

    print_smu_reg();
    //avc_writel((1<<16)|(1<<8), SMU_L3_POWERSW_BUZ);
    reg = readl(SMU_LOG3SW_ACTRLEN);
    reg &= 0xFFFF;
    reg |= (0x53<<16);
    avc_writel(reg, SMU_LOG3SW_ACTRLEN); // LOG3_PWRSW_WAIT_PARAM[29:16]
    avc_writel(1, SMU_LOG3SW_ACTRL); // start on-sequence


    /* watch status */
    for (i=0; i<100; i++) {
        udelay(10);
        reg = readl(SMU_LOG3SW_ACTRL);
        if (reg & (1<<16)) { // L3_PWRSW_STATUS
            debug2("delay %d usec",i*10);
            print_smu_reg();
            return 0;
        }
    }

    print_smu_reg();
    return -1;

#endif
}

/*!
 * stop AVC module power supply
 * @param void
 * @retval 0 successful
 * @retval -1 failed
 */
static int avc_power_off(void)
{
//    unsigned long reg;
//    int i;

    avc_writel(0x00000100, SMU_LOG3SW_ACTRLEN);

    avc_writel(0x00000000, SMU_L3_POWERSW_BUZ);
    avc_writel(0x00000001, SMU_L3_POWERSW_BUZ);
    avc_writel(0x00000003, SMU_L3_POWERSW_BUZ);
    avc_writel(0x00000007, SMU_L3_POWERSW_BUZ);
    avc_writel(0x0000000F, SMU_L3_POWERSW_BUZ);
    avc_writel(0x0000001F, SMU_L3_POWERSW_BUZ);
    avc_writel(0x0000003F, SMU_L3_POWERSW_BUZ);
    avc_writel(0x0000007F, SMU_L3_POWERSW_BUZ);
    avc_writel(0x000000FF, SMU_L3_POWERSW_BUZ);

    udelay(2);

    return 0;

#if 0
    print_smu_reg();
    avc_writel(0, SMU_LOG3SW_ACTRL); // start off-sequence

    /* watch status */
    for (i=0; i<100; i++) {
        udelay(10);
        reg = readl(SMU_LOG3SW_ACTRL);
        if ((reg & (1 << 16)) == 0) { // SMU_LOG3SW_ACTRL
            debug2("delay %d usec",i*10);
            print_smu_reg();
            return 0;
        }
    }

    print_smu_reg();
    return -1;
#endif
}

/*!
 * open file operation
 * @param[in] inode
 * @param[in] filp
 * @retval 0 successful
 * @retval -EBUSY double request
 * @retval -FAULT power setting failure etc
 */
static int avc_open(struct inode *inode, struct file *filp)
{
//  int minor = MINOR(inode->i_rdev);
//  int ret;
//  debug1("minor %d, usecount %d",minor,GET_USE_COUNT(THIS_MODULE));

    if (avc_info->state != state_initial) {
        debug1("already opened");
        return -EBUSY;
    }

#if 0
    // energy-saving off
    status_ctrl_func(DRV_AVC, STAT_ON);

    // not to change frequency automatically
    ret = mp200_pmu_frqctrl_on(MP200_FRQCTRL_AVC);
    if (ret) {
        debug0("mp200_pmu_frqctrl_on failed %d",ret);
        status_ctrl_func(DRV_AVC, STAT_OFF);
        return -EFAULT;
    }
#endif

    // power on
    if (avc_power_on() < 0) {
        debug0("avc_power_on failed");
#if 0
        ret = mp200_pmu_frqctrl_off(MP200_FRQCTRL_AVC);
        if (ret) {
            debug0("mp200_pmu_frqctrl_off failed %d",ret);
        }
        status_ctrl_func(DRV_AVC, STAT_OFF);
#endif
        return -EFAULT;
    }

    avc_clock_on();

    avc_info->state = state_idle;

    avc_open_state = 1;

//  MOD_INC_USE_COUNT;

    return 0;
}

/*!
 * close file operation
 * @param[in] inode
 * @param[in] filp
 * @retval 0 successful
 */
static int avc_release(struct inode *inode, struct file *filp)
{
    int minor = MINOR(inode->i_rdev);
//  int ret;
    debug1("minor %d",minor);

    command_start(avc_command_reset, 0);
    avc_info->state = state_initial;

    avc_clock_off();
    if (avc_power_off() < 0) {
        debug0("avc_power_off failed");
    }
#if 0
    ret = mp200_pmu_frqctrl_off(MP200_FRQCTRL_AVC);
    if (ret) {
        debug0("mp200_pmu_frqctrl_off failed %d",ret);
    }
    status_ctrl_func(DRV_AVC, STAT_OFF);
#endif

    avc_open_state = 0;

//  MOD_DEC_USE_COUNT;

    return 0;
}

//! file operations for this driver
static struct file_operations avc_fops = {
    .owner	= THIS_MODULE,
    .llseek	= no_llseek,
    .open	= avc_open,
    .release	= avc_release,
    .read	= avc_read,
    .poll	= avc_poll,
    .ioctl	= avc_ioctl,
    .mmap	= avc_mmap,
};


/* interruption -------------------------------------------------------------*/

/*!
 * AVC interruption handler about stream buffer full
 * @param irq
 * @param dev_id
 * @param regs
 */
static irqreturn_t avc_interrupt_a(int irq, void *dev_id)
{
    struct avc_regs_t *avc_regs;
    unsigned short factor;

#if 0
    if (!avc_info) {
        debug1("not initialized yet");
        return IRQ_NONE;
    }
#endif

    avc_regs = avc_info->regs;
    factor = readw(&avc_regs->BINT1);
    if (factor) {
        avc_writew(factor, &avc_regs->BINT5); // clear interruption
#if 1
	if (avc_info->common_info) {
	    avc_info->int_a = 1;
	    tasklet_schedule(&avc_tasklet);
	}
#else
        avc_info->int_a = 1;
        tasklet_schedule(&avc_tasklet);
#endif
        return IRQ_HANDLED;
    }
    return IRQ_NONE;
}

/*!
 * AVC interruption handler
 * @param irq
 * @param dev_id
 * @param regs
 */
static irqreturn_t avc_interrupt_c(int irq, void *dev_id)
{
    struct avc_regs_t *avc_regs;
    unsigned short factor;

    if (!avc_info) {
        debug1("not initialized yet");
        return IRQ_NONE;
    }

    avc_regs = avc_info->regs;
    factor = readw(&avc_regs->INT1reg);
    if (factor) {
        avc_writew(factor, &avc_regs->INT5reg); // clear interruption
        avc_info->int_c |= factor;
        tasklet_schedule(&avc_tasklet);
        return IRQ_HANDLED;
    }
    return IRQ_NONE;
}

/*!
 * tasklet for interruption
 * @param[in] value unused
 * @return void
 */
void avc_tasklet_handler(unsigned long value)
{
    unsigned long flags;
    unsigned int int_a, int_c;

    while (1) {
        spin_lock_irqsave(&avc_lock, flags);
        int_c = avc_info->int_c;
        int_a = avc_info->int_a;
        avc_info->int_c = 0;
        avc_info->int_a = 0;
        spin_unlock_irqrestore(&avc_lock, flags);

        if (int_c == 0 && int_a == 0) break;
        avc_interruption_handler(int_c, int_a);
        avc_info->interrupt.interrupt.l  |= int_c;
        avc_info->interrupt.int_buf_full |= int_a;
    }

    wake_up_interruptible(&readq);
}


/* module functions ---------------------------------------------------------*/

static void initialize_common_info(struct avc_common_info_t *common_info)
{
    struct encode_info_t *ei = &common_info->encode_info;
    struct avc_user_set_parameter_t *usp = &common_info->avc_user_set_parameter;

    memset(common_info, 0, sizeof(*common_info));

    common_info->avc_buffer_address.change_endian = 1;
    ei->enc_image_value = 0x33;
    ei->enc_initial_image_value = 0x19;
    ei->enc_islice_image_mode = 0x08;
    ei->enc_min_qp_i = 1;
    ei->enc_min_qp_p = 1;

    usp->set_int_en = AVC_INT_ENC_END_EN | AVC_INT_DEC_END_EN
      | AVC_INT_DEC_ERR_EN | AVC_INT_ENC_PIC_END_EN | AVC_INT_DEC_PIC_END_EN
      | AVC_INT_HEAD_DEC_END_EN;

    usp->set_err_nal_en = AVC_ERR_NAL_END | AVC_ERR_FORBIDDEN_ZERO_BIT
      | AVC_ERR_REF_IDC | AVC_ERR_NAL_UNIT_TYPE;

    usp->set_err_sps_en = AVC_HRD_ERROR | AVC_VUI_ERROR
      | AVC_SPS_OTHER_ERROR | AVC_ERR_DIRECT_8X8_INFERENCE_FLAG
      | AVC_ERR_MB_ADAPTIVE_FRAME_FIELD_FLAG | AVC_ERR_PROFILE
      | AVC_ERR_LEVEL_IDC | AVC_ERR_SEQ_PARAMETER_SET_ID
      | AVC_ERR_LOG2_MAX_FRAME_NUM_MINUS4 | AVC_ERR_PIC_ORDER_CNT_TYPE
      | AVC_ERR_LOG2_MAX_PIC_ORDER_CNT_LSB_MINUS4 | AVC_ERR_NUM_REF_FRAMES
      | AVC_ERR_PIC_WIDTH_IN_MBS_MINUS1 | AVC_ERR_PIC_HEIGHT_IN_MAP_UNITS_MINUS1;

    usp->set_err_pps_en = AVC_ERR_PPS_PPS_ID | AVC_ERR_PPS_SPS_ID
      | AVC_ERR_PPS_ENTROPY_FLAG | AVC_ERR_PPS_NUM_SLICE
      | AVC_ERR_PPS_NUM_REF_IDX | AVC_ERR_PPS_WEIGHTED_FLAG
      | AVC_ERR_PPS_WEIGHTED_IDC | AVC_ERR_PPS_PIC_INIT_Q
      | AVC_ERR_PPS_CQP_IDX_OFS | AVC_ERR_PPS_REDUNDANT_FLAG;

    usp->set_err_sh_en = AVC_ERR_SH_END | AVC_ERR_SH_SLICE_TYPE
      | AVC_ERR_SH_PPS_ID | AVC_ERR_SH_IDR_PIC_ID
      | AVC_ERR_SH_NUM_REF_IDX_L0 | AVC_ERR_SH_DIS_DBF_IDC
      | AVC_ERR_SH_SLC_A_C0_OFS | AVC_ERR_SH_SLC_BETA_OFS
      | AVC_ERR_SH_ROPN_IDC | AVC_ERR_SH_MMCO;

    usp->set_err_sd_en = AVC_ERR_1ST_MB_IN_SLICE | AVC_ERR_SD_END
      | AVC_ERR_SD_MB_TYPE | AVC_ERR_SD_CBP
      | AVC_ERR_SD_MBP_REF_IDX | AVC_ERR_SD_MBP_ICPM
      | AVC_ERR_SD_SMB_REF_IDX | AVC_ERR_ZSL_LT_RUN
      | AVC_ERR_CV_COEFT | AVC_ERR_CV_LV_PREFIX
      | AVC_ERR_CV_TZEROS | AVC_ERR_CV_RUN_BEFORE;

    usp->set_header_end_en = AVC_SPS_END | AVC_PPS_END
      | AVC_SEI_END | AVC_AUD_END | AVC_EOSEQ_END
      | AVC_SH_END | AVC_END_OF_STREAM_END
      | AVC_FILLER_DATA_END;

    usp->set_ref_hw_ctrl = AVC_REF_HW_CTRL_HW;

    usp->set_dec_error_conceal = AVC_DEC_ERROR_CONCEAL_ON;

    usp->set_sei_raw_output = 0;

    usp->set_mem_num_wr = 0;

    usp->mem_num_ref_idx.set_ref_idx0 = 0;
    usp->mem_num_ref_idx.set_ref_idx1 = 0;
    usp->mem_num_ref_idx.set_ref_idx2 = 0;
    usp->mem_num_ref_idx.set_ref_idx3 = 0;
    usp->mem_num_ref_idx.set_ref_idx4 = 0;
    usp->mem_num_ref_idx.set_ref_idx5 = 0;
    usp->mem_num_ref_idx.set_ref_idx6 = 0;
    usp->mem_num_ref_idx.set_ref_idx7 = 0;
    usp->mem_num_ref_idx.set_ref_idx8 = 0;
    usp->mem_num_ref_idx.set_ref_idx9 = 0;
    usp->mem_num_ref_idx.set_ref_idx10 = 0;
    usp->mem_num_ref_idx.set_ref_idx11 = 0;
    usp->mem_num_ref_idx.set_ref_idx12 = 0;
    usp->mem_num_ref_idx.set_ref_idx13 = 0;
    usp->mem_num_ref_idx.set_ref_idx14 = 0;
    usp->mem_num_ref_idx.set_ref_idx15 = 0;

    usp->set_planar_sel = AVC_PLANAR_SEL_PLANAR;

    usp->set_bytelane = AVC_BYTELANE_Y3210 | AVC_BYTELANE_C3210;

}

#ifdef CONFIG_PM
static int avc_pf_suspend(struct platform_device *dev, pm_message_t state)
{
    if(avc_open_state){
      if(state.event == DEV_SUSPEND_IDLE_1){ /* PCM direct */
	return -EBUSY;
      }
      else{ /* state.event == PM_EVENT_SUSPEND */
	avc_clock_off();
	avc_power_off();
      }
    }
    return 0;
}

static int avc_pf_resume(struct platform_device *dev)
{
    if(avc_open_state){
	avc_power_on();
	avc_clock_on();
    }
    return 0;
}
#endif

static int avc_pf_probe(struct platform_device *dev)
{
    struct avc_common_info_t *common_info;
    dma_addr_t paddr;
    int ret;

    common_info = dma_alloc_coherent(0, sizeof *common_info,
				     &paddr, GFP_KERNEL|GFP_DMA);
    if (common_info) {
        initialize_common_info(common_info);
	avc_info->common_info = common_info;
	avc_info->pa_common_info = paddr;
    } else {
        debug0("dma_alloc_coherent avc_common_info failed");
        ret = -ENOMEM;
        goto error_return;
    }

    /* request_irq returns ENOMEM/EINVAL */
    ret = request_irq(INT_AVC_A, avc_interrupt_a,
                      IRQF_SHARED, AVC_MODNAME, avc_info);
    if (ret < 0) {
        debug0("request_irq(%d) failed %d", INT_AVC_A, ret);
        goto free_common_info;
    }

    ret = request_irq(INT_AVC_C, avc_interrupt_c,
                      IRQF_SHARED, AVC_MODNAME, avc_info);
    if (ret < 0) {
        debug0("request_irq(%d) failed %d", INT_AVC_C, ret);
        goto free_irq_inta;
    }

    /* register chrdev */
    debug1("register_chrdev %d, %s", devmajor, AVC_MODNAME);
    ret = register_chrdev(devmajor, AVC_MODNAME, &avc_fops);
    if (ret < 0) {
        debug0("register_chrdev %d, %s failed %d",devmajor,AVC_MODNAME,ret);
        ret = -EFAULT;
        goto free_irq_intc;
    }

    avc_class = class_create(THIS_MODULE, AVC_MODNAME);
    if (IS_ERR(avc_class)) {
        debug0("class_create failed %d", ret);
	ret = PTR_ERR(avc_class);
	goto free_chrdev;
    }

    avc_class_device = class_device_create(avc_class, 0, MKDEV(devmajor, 0),
					   &dev->dev, AVC_DEVNAME);
    if (IS_ERR(avc_class_device)) {
        debug0("class_device_create failed %s %d", AVC_DEVNAME, ret);
	class_destroy(avc_class);
	ret = PTR_ERR(avc_class_device);
	goto free_chrdev;
    }

    /* driver version information */
    avc_create_proc();

    avc_info->state = state_initial;
    memcpy(&avc_info->version, &avc_driver_version, sizeof(avc_info->version));

    avc_open_state = 0;
    debug1("success");
    return 0;

 free_chrdev:
    unregister_chrdev(devmajor, AVC_MODNAME);

 free_irq_intc:
    free_irq(INT_AVC_C, avc_info);

 free_irq_inta:
    free_irq(INT_AVC_A, avc_info);

 free_common_info:
    dma_free_coherent(0, sizeof *common_info, common_info, paddr);

 error_return:
    return ret;
}

static int avc_pf_remove(struct platform_device *dev)
{
    class_device_unregister(avc_class_device);
    class_destroy(avc_class);
    avc_remove_proc();
    debug1("unregister_chrdev %d, %s",devmajor, AVC_MODNAME);
    unregister_chrdev(devmajor, AVC_MODNAME);
    free_irq(INT_AVC_C, avc_info);
    free_irq(INT_AVC_A, avc_info);
    dma_free_coherent(0, sizeof *avc_info->common_info,
		      avc_info->common_info, avc_info->pa_common_info);
    return 0;
}

static void avc_pf_release(struct device *dev)
{
    /* none */
}

#if 1
/*!
 * initialize for insmod
 * @param void
 * @return void
 */
static int __init avc_init(void)
{
    int ret;

    ret = platform_device_register(&avc_pf_device);
    if (ret)
      return ret;
    return platform_driver_register(&avc_pf_driver);
}

/*!
 * finalize for rmmod
 * @param void
 * @return void
 */
static void __exit avc_exit(void)
{
  (void)platform_driver_unregister(&avc_pf_driver);
  (void)platform_device_unregister(&avc_pf_device);
}
#else
/*!
 * initialize for insmod
 * @param void
 * @return void
 */
static int __init avc_init(void)
{
    struct avc_regs_t *regs;
    struct avc_info_t *info;
    struct avc_common_info_t *common_info;
    int ret;

    /* allocate avc_info */
    info = kmalloc(sizeof(*info), GFP_KERNEL);
    if (info) {
        memset(info, 0, sizeof(*info));
    } else {
        debug0("kmalloc avc_info_t failed");
        ret = -ENOMEM;
        goto error_return;
    }

    /* allocate common_info */
    common_info = alloc_from_free_pages(sizeof(*common_info));
    if (common_info) {
        initialize_common_info(common_info);
    } else {
        debug0("kmalloc avc_common_info failed");
        ret = -ENOMEM;
        goto free_info;
    }

    /* base address */
    regs = (struct avc_regs_t *)IO_ADDRESS(MP200_AVC_BASE);

    /* request_irq returns ENOMEM/EINVAL */
    ret = request_irq(INT_AVC_A, avc_interrupt_a,
                      SA_SHARED, devname, info);
    if (ret < 0) {
        debug0("request_irq(%d) failed %d", INT_AVC_A, ret);
        goto free_common_info;
    }

    ret = request_irq(INT_AVC_C, avc_interrupt_c,
                      SA_SHARED, devname, info);
    if (ret < 0) {
        debug0("request_irq(%d) failed %d", INT_AVC_C, ret);
        goto free_irq_inta;
    }

    /* register chrdev */
    debug1("register_chrdev %d, %s",devmajor,devname);
    ret = register_chrdev(devmajor, devname, &avc_fops);
    if (ret < 0) {
        debug0("register_chrdev %d, %s failed %d",devmajor,devname,ret);
        ret = -EFAULT;
        goto free_irq_intc;
    }

    /* driver version information */
    avc_create_proc();

    info->common_info = common_info;
    info->regs = regs;
    avc_info = info;
    avc_info->state = state_initial;
    memcpy(&avc_info->version, &avc_driver_version, sizeof(avc_info->version));

    return 0;

 free_irq_intc:
    free_irq(INT_AVC_C, info);

 free_irq_inta:
    free_irq(INT_AVC_A, info);

 free_common_info:
    free_to_free_pages(common_info, sizeof(*common_info));

 free_info:
    kfree(info);
    avc_info = NULL;

 error_return:
    return ret;
}

/*!
 * finalize for rmmod
 * @param void
 * @return void
 */
static void __exit avc_exit(void)
{
    struct avc_common_info_t *common_info = avc_info->common_info;

    debug1("unregister_chrdev %d, %s",devmajor,devname);
    avc_remove_proc();
    unregister_chrdev(devmajor, devname);
    free_irq(INT_AVC_C, avc_info);
    free_irq(INT_AVC_A, avc_info);
    free_to_free_pages(common_info, sizeof(*common_info));
    kfree(avc_info);
    avc_info = NULL;
}
#endif

module_init(avc_init);
module_exit(avc_exit);
module_param(devname, charp, 0444);
module_param(devmajor, int, 0444);
module_param(debug_level, int, 0444);
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("H.264/AVC Driver for EM1");
MODULE_AUTHOR("NIS");
