/* 2009-09-01: File added by Sony Corporation */
/*
 *  drivers/block/snsc_loop_io_trace.c
 *
 *  Copyright 2009 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/loop.h>
#include <linux/fs.h>

#define LOOPTRACE_PFX "LOOP_TRACE: "

static void
print_filename_size(char *operation, struct dentry *backing_dentry,
		    unsigned int size, int ret)
{
	struct dentry *dent;

	dent = dget(backing_dentry);

	if (dent != NULL) {
		if (ret >= 0) {
			pr_info(LOOPTRACE_PFX "%s %s (%ubytes)\n",
				dent->d_name.name,
				operation,
				size);

		} else {
			pr_info(LOOPTRACE_PFX "%s %s error (errno = %d)\n",
				dent->d_name.name,
				operation,
				ret);
		}

	} else {
		pr_err(LOOPTRACE_PFX "backing file dentry is null\n");

	}

	dput(dent);
}

static void
send_probe(const struct marker *mdata, void *private, const char *fmt, ...)
{
	va_list args;
	struct loop_device *lo;
	struct bio *bio;
	int ret;

	va_start(args, fmt);
	lo  = va_arg(args, typeof(lo));
	bio = va_arg(args, typeof(bio));
	ret = va_arg(args, typeof(ret));
	va_end(args);

	/* sanity checks */
	if (lo == NULL || lo->lo_backing_file == NULL)  {
		pr_err(LOOPTRACE_PFX "backing file is unknown\n");
		return;
	}
	if (bio == NULL) {
		pr_err(LOOPTRACE_PFX "bio is null\n");
		return;
	}

	print_filename_size("write", lo->lo_backing_file->f_dentry,
			    bio->bi_size, ret);
}

struct probe_data {
	const char *name;
	const char *format;
	marker_probe_func *probe_func;
};

static struct probe_data probe_array[] =
{
	{
		.name       = "loop_send",
		.format     = "%p %p %d",
		.probe_func = send_probe,
	}
};

static int __init
loop_io_trace_init(void)
{
	int i, result, ret;

	ret = 0;

	for (i = 0; i < ARRAY_SIZE(probe_array); i++) {

		result = marker_probe_register(
				probe_array[i].name,
				probe_array[i].format,
				probe_array[i].probe_func,
				NULL);
		if (result) {
			pr_err(LOOPTRACE_PFX "unable to register probe %s\n",
			       probe_array[i].name);
			ret = result;
			continue;
		}

		result = marker_arm(probe_array[i].name);
		if (result) {
			pr_err(LOOPTRACE_PFX "unable to arm probe %s\n",
			       probe_array[i].name);
			ret = result;
		}
	}

	if (!ret)
		pr_info(LOOPTRACE_PFX "installed\n");
	else
		pr_err(LOOPTRACE_PFX "unable to install\n");

	return ret;
}

static void __exit
loop_io_trace_exit(void)
{
	int i, result;

	for (i = ARRAY_SIZE(probe_array) - 1; i >= 0; i--) {

		result = marker_disarm(probe_array[i].name);

		if (result) {
			pr_warning(LOOPTRACE_PFX "unable to unarm probe %s\n",
				   probe_array[i].name);
			continue;
		}

		marker_probe_unregister(probe_array[i].name);
	}
}

module_init(loop_io_trace_init);
module_exit(loop_io_trace_exit);

MODULE_LICENSE("GPL");
