/* 2010-08-25: File added by Sony Corporation */
/*
 *  File Name       : drivers/char/mp200-rtc.c
 *  Function        : MP200 Real Time Clock interface
 *  Release Version : Ver 1.00
 *  Release Date    : 2007/04/11
 *
 *  Copyright (C) NEC Electronics Corporation 2007
 *
 *
 *  This program is free software;you can redistribute it and/or modify itunder the terms of
 *  the GNU General Public License as published by Free Softwere Foundation;either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/rtc.h>
#include <linux/bcd.h>

#include <linux/log2.h>

#include <asm/uaccess.h>
#include <asm/rtc.h>
#include <asm/arch/pwc.h>

static DEFINE_SPINLOCK(mp200_rtc_lock);

/* RTC1 register bit define */
#define RTC1_WALE     0x80
#define RTC1_DALE     0x40
#define RTC1_CT_SEC   0x04
#define RTC1_CT_MIN   0x05
#define RTC1_CT_HOUR  0x06
#define RTC1_CT_MONTH  0x07
#define RTC1_CT_MASK  0x07

/* RTC2 register bit define */
#define RTC2_1224     0x20
#define RTC2_ADJ_XSTP 0x40
#define RTC2_CTFG     0x04
#define RTC2_WAFG     0x02
#define RTC2_DAFG     0x01

/* RTC register access macros: */
#define RTC_READ(addr, val) pwc_read((addr), (unsigned int *)(&(val)));
#define RTC_WRITE(val, addr) pwc_write((addr), (unsigned int)(val), 0xff);

static unsigned long epoch = 1900;	/* year corresponding to 0x00 */

static const unsigned char days_in_mo[] =
    { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

static int rtc_irq = 0;
static int p_freq = 10;
static unsigned int aie_irqen;
static unsigned int pie_irqen;

int get_rtc_time(struct rtc_time *rtc_tm)
{
	spin_lock_irq(&mp200_rtc_lock);
	RTC_READ(PWC_SEC_RD,   rtc_tm->tm_sec);
	RTC_READ(PWC_MIN_RD,   rtc_tm->tm_min);
	RTC_READ(PWC_HOUR_RD,  rtc_tm->tm_hour);
	RTC_READ(PWC_WEEK_RD,  rtc_tm->tm_wday);
	RTC_READ(PWC_DAY_RD,   rtc_tm->tm_mday);
	RTC_READ(PWC_MONTH_RD, rtc_tm->tm_mon);
	RTC_READ(PWC_YEAR_RD,  rtc_tm->tm_year);
	spin_unlock_irq(&mp200_rtc_lock);

	rtc_tm->tm_mon &= 0x1f;	/* Mask of D7(19/20) */

	rtc_tm->tm_sec  = BCD2BIN(rtc_tm->tm_sec);
	rtc_tm->tm_min  = BCD2BIN(rtc_tm->tm_min);
	rtc_tm->tm_hour = BCD2BIN(rtc_tm->tm_hour);
	rtc_tm->tm_mday = BCD2BIN(rtc_tm->tm_mday);
	rtc_tm->tm_mon  = BCD2BIN(rtc_tm->tm_mon);
	rtc_tm->tm_year = BCD2BIN(rtc_tm->tm_year);

	if ((rtc_tm->tm_year += (epoch - 1900)) <= 69) {
		rtc_tm->tm_year += 100;
	}
	rtc_tm->tm_mon--;

	return 0;
}

static int set_rtc_time(struct rtc_time *rtc_tm)
{
	unsigned char mon, day, hrs, min, sec, leap_yr, wday;
	unsigned int yrs;

	yrs = rtc_tm->tm_year + 1900;
	mon = rtc_tm->tm_mon + 1;	/* tm_mon starts at zero */
	day = rtc_tm->tm_mday;
	hrs = rtc_tm->tm_hour;
	min = rtc_tm->tm_min;
	sec = rtc_tm->tm_sec;

	if (yrs < 1970) {
		return -EINVAL;
	}

	/* setting day of week */
	if (mon < 3) {
		wday = ((yrs-1) + (yrs-1)/4 - (yrs-1)/100 + (yrs-1)/400 + (13*(mon+12)+8)/5 + day) % 7;
	} else {
		wday = (yrs + yrs/4 - yrs/100 + yrs/400 + (13*mon+8)/5 + day) % 7;
	}

	leap_yr = ((!(yrs % 4) && (yrs % 100)) || !(yrs % 400));

	if (wday >= 7) {
		return -EINVAL;
	}
	if ((mon > 12) || (day == 0)) {
		return -EINVAL;
	}
	if (day > (days_in_mo[mon] + ((mon == 2) && leap_yr))) {
		return -EINVAL;
	}
	if ((hrs >= 24) || (min >= 60) || (sec >= 60)) {
		return -EINVAL;
	}
	if ((yrs -= epoch) > 255) {
		return -EINVAL;
	}
	if (yrs > 169) {
		return -EINVAL;
	}
	if (yrs >= 100) {
		yrs -= 100;
	}

	sec = BIN2BCD(sec);
	min = BIN2BCD(min);
	hrs = BIN2BCD(hrs);
	day = BIN2BCD(day);
	mon = BIN2BCD(mon);
	yrs = BIN2BCD(yrs);

	spin_lock_irq(&mp200_rtc_lock);
	RTC_WRITE(yrs,  PWC_YEAR);
	RTC_WRITE(mon,  PWC_MONTH);
	RTC_WRITE(day,  PWC_DAY);
	RTC_WRITE(wday, PWC_WEEK);
	RTC_WRITE(hrs,  PWC_HOUR);
	RTC_WRITE(min,  PWC_MIN);
	RTC_WRITE(sec,  PWC_SEC);
	spin_unlock_irq(&mp200_rtc_lock);

	return 0;
}

static inline void mp200_rtc_setaie(unsigned int enable)
{
	unsigned char rtc1_val, rtc2_val;

	RTC_READ(PWC_RTC1_RD, rtc1_val);
	RTC_READ(PWC_RTC2_RD, rtc2_val);
	rtc1_val &= ~RTC1_WALE;
	rtc2_val &= ~RTC2_WAFG;
	if (enable) {
		rtc1_val |= RTC1_WALE;
	}
	RTC_WRITE(rtc2_val, PWC_RTC2);
	RTC_WRITE(rtc1_val, PWC_RTC1);

	if (enable && (aie_irqen == 0) && (pie_irqen == 0)) {
		enable_irq(rtc_irq);
	}

	if ((enable == 0) && (aie_irqen == 1) && (pie_irqen == 0)) {
		disable_irq(rtc_irq);
	}

	aie_irqen = enable;
}

static inline void mp200_rtc_setpie(unsigned int enable)
{
	unsigned char rtc1_val, rtc2_val;

	RTC_READ(PWC_RTC1_RD, rtc1_val);
	RTC_READ(PWC_RTC2_RD, rtc2_val);
	rtc1_val &= ~RTC1_CT_MASK;
	if (enable) {
		switch (p_freq) {
		case 1:
			rtc1_val |= RTC1_CT_SEC;
			break;
		case 60:
			rtc1_val |= RTC1_CT_MIN;
			break;
		case 60*60:
			rtc1_val |= RTC1_CT_HOUR;
			break;
		case 60*60*24*30:
			rtc1_val |= RTC1_CT_MONTH;
			break;
		}
	}

	rtc2_val &= ~RTC2_CTFG;
	RTC_WRITE(rtc2_val, PWC_RTC2);
	RTC_WRITE(rtc1_val, PWC_RTC1);
	if (enable && (aie_irqen == 0) && (pie_irqen == 0)) {
		enable_irq(rtc_irq);
	}

	if ((enable == 0) && (aie_irqen == 0) && (pie_irqen == 1)) {
		disable_irq(rtc_irq);
	}

	pie_irqen = enable;
}

static inline int mp200_rtc_setfreq(unsigned int freq)
{
	switch (freq) {
	case 1:
	case 60:
	case 60*60:
	case 60*60*24*30:
		break;
	default:
		return -EINVAL;
	}

	p_freq = freq;

	return 0;
}

static int mp200_rtc_getalarm(struct rtc_wkalrm *alrm)
{
	struct rtc_time *alm_tm = &alrm->time;

	spin_lock_irq(&mp200_rtc_lock);
	RTC_READ(PWC_ALM_WM_RD, alm_tm->tm_min);
	RTC_READ(PWC_ALM_WH_RD, alm_tm->tm_hour);
	RTC_READ(PWC_ALM_WW_RD, alm_tm->tm_wday);
	spin_unlock_irq(&mp200_rtc_lock);

	alm_tm->tm_min  = BCD2BIN(alm_tm->tm_min);
	alm_tm->tm_hour = BCD2BIN(alm_tm->tm_hour);

	alm_tm->tm_wday &= 0x7F;
	if(alm_tm->tm_wday) {
		if(alm_tm->tm_wday == 0x7F)
			alm_tm->tm_wday = -1;
		else
			alm_tm->tm_wday = ilog2(alm_tm->tm_wday);
	}

	return 0;
}

static int mp200_rtc_setalarm(struct rtc_wkalrm *alrm)
{
	struct rtc_time *alm_tm = &alrm->time;
	unsigned char hrs, min, wday;

	hrs = alm_tm->tm_hour;
	min = alm_tm->tm_min;
	if (alm_tm->tm_wday == -1) {
		wday = 0x7f;	/* All */
	} else {
		if (alm_tm->tm_wday > 6)
			return -EINVAL;
		wday = 1 << alm_tm->tm_wday;
	}

	if (hrs >= 24)
		return -EINVAL;
	if (min >= 60)
		return -EINVAL;

	min = BIN2BCD(min);
	hrs = BIN2BCD(hrs);

	spin_lock_irq(&mp200_rtc_lock);
	RTC_WRITE(hrs,  PWC_ALM_WH);
	RTC_WRITE(min,  PWC_ALM_WM);
	RTC_WRITE(wday, PWC_ALM_WW);
	spin_unlock_irq(&mp200_rtc_lock);

	if (alrm->enabled) {
		mp200_rtc_setaie(1);
	} else {
		mp200_rtc_setaie(0);
	}

	return 0;
}

static irqreturn_t mp200_rtc_interrupt(int irq, void *dev_id)
{
	unsigned long events = RTC_IRQF;
	unsigned char rtc2_val;

	RTC_READ(PWC_RTC2_RD, rtc2_val);

	if (!(rtc2_val & (RTC2_CTFG | RTC2_WAFG)))
		return IRQ_NONE;

	if (rtc2_val & RTC2_CTFG) {
		rtc2_val &= ~RTC2_CTFG;
		events |= RTC_PF;
	}
	if (rtc2_val & RTC2_WAFG) {
		rtc2_val &= ~RTC2_WAFG;
		events |= RTC_AF;
	}

	RTC_WRITE(rtc2_val, PWC_RTC2);

	rtc_update(1, events);

	return IRQ_HANDLED;
}


static int mp200_rtc_ioctl(unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case RTC_AIE_ON:
	case RTC_AIE_OFF:
		mp200_rtc_setaie((cmd == RTC_AIE_ON) ? 1 : 0);
		break;

	case RTC_PIE_ON:
	case RTC_PIE_OFF:
		mp200_rtc_setpie((cmd == RTC_PIE_ON) ? 1 : 0);
		break;

	case RTC_IRQP_READ:
		return  put_user(p_freq, (unsigned long __user *)arg);

	case RTC_IRQP_SET:
		if (!arg) {
			return -EINVAL;
		}
		return mp200_rtc_setfreq(arg);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static int mp200_rtc_proc(char *buf)
{
	char *p = buf;

	p += sprintf(p, "alarm_IRQ: %s\n", aie_irqen ? "Enable" : "Disable" );
	p += sprintf(p, "periodic_IRQ: %s\n", pie_irqen ? "Enable" : "Disable" );
	p += sprintf(p, "periodic_freq\t: %s\n",
		(p_freq == 1) ? "1sec" : ((p_freq == 60) ? "1min" :
			((p_freq == 60*60) ? "1hour" : ((p_freq == 60*60*24*30) ? "1year" : "not set"))));

	return p - buf;
}

static int mp200_rtc_open(void)
{
	int ret;

	ret = request_irq(rtc_irq, mp200_rtc_interrupt, IRQF_DISABLED, "rtc", NULL);
	if (ret < 0) {
		printk(KERN_ERR "rtc: request_irq error! (%d)\n", rtc_irq);
		return ret;
	}
	disable_irq(rtc_irq);

	return 0;
}

static void mp200_rtc_release(void)
{
	mp200_rtc_setaie(0);
	mp200_rtc_setpie(0);

	free_irq(rtc_irq, NULL);
}

static struct rtc_ops mp200_rtcops = {
	.owner		= THIS_MODULE,
	.open       = mp200_rtc_open,
	.release    = mp200_rtc_release,
	.ioctl      = mp200_rtc_ioctl,
	.read_time  = get_rtc_time,
	.set_time   = set_rtc_time,
	.read_alarm = mp200_rtc_getalarm,
	.set_alarm  = mp200_rtc_setalarm,
	.proc       = mp200_rtc_proc,
};

static int mp200_rtc_probe(struct platform_device *pdev)
{
	printk(KERN_DEBUG "Starting RTC.\n");

	rtc_irq = platform_get_irq(pdev, 0);
	if (rtc_irq < 0) {
		return -ENODEV;
	}

	register_rtc(&mp200_rtcops);

	aie_irqen = 0;
	pie_irqen = 0;

	RTC_WRITE(0x00, PWC_TRIM);
	RTC_WRITE(0x00, PWC_RTC1);
	RTC_WRITE(RTC2_1224, PWC_RTC2);

	return 0;
}

static int mp200_rtc_remove(struct platform_device *pdev)
{
	unregister_rtc(&mp200_rtcops);

	return 0;
}

static struct platform_driver mp200_rtcdrv = {
	.probe		= mp200_rtc_probe,
	.remove		= mp200_rtc_remove,
	.driver		= {
		.name	= "mp200-rtc",
		.owner	= THIS_MODULE,
	},
};

static int __init mp200_rtc_init(void)
{
	return platform_driver_register(&mp200_rtcdrv);
}

static void __exit mp200_rtc_exit(void)
{
	platform_driver_unregister(&mp200_rtcdrv);
}

module_init(mp200_rtc_init);
module_exit(mp200_rtc_exit);

MODULE_DESCRIPTION("MP200 series RTC Driver");
MODULE_LICENSE("GPL");
