/* 2010-08-25: File added by Sony Corporation */
/*
 *  File Name       : linux/drivers/char/mp200_spi_kern.c
 *  Function        : MP200 SPI interface
 *  Release Version : Ver 0.01
 *  Release Date    : 2006/06/02
 *
 *  Copyright (C) NEC Electronics Corporation 2005-2006
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>

#include <asm/io.h>
#include <asm/arch/smu.h>
#include <asm/arch/pmu.h>

#define SPI_DEBUG_PARAM

#include "mp200_spi_kern.h"

#ifdef SPI_DEBUG
#define DPRINTK(fmt, args...) printk(KERN_INFO "\n  %s: %s: " fmt,__FILE__,__FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif

#define RETRY_COUNT(x)  ((loops_per_jiffy * x)/(1000000/HZ)) /* x us */

static unsigned int spi_clock_table[] = {
	1433,		/* (1434KHz) */
	2867,		/* (2.87MHz) */
	5734,		/* (5.73MHz) */
	11468,		/* (11.47MHz) */
	22937,		/* (22.94MHz) */
	38229,		/* (38.23MHz) */
};

static unsigned int spi_plldiv_table[] = {
	SMU_DIV160,	/* 1500KHz(1434KHz) */
	SMU_DIV80,	/* 3MHz   (2.87MHz) */
	SMU_DIV40,	/* 6MHz   (5.73MHz) */
	SMU_DIV20,	/* 12MHz  (11.47MHz) */
	SMU_DIV10,	/* 24MHz  (22.94MHz) */
	SMU_DIV6,	/* 48MHz  (38.23MHz) */
};

static spi_data_t spi_private[] = {
	{
	 .regs = (spi_regs *)(SP0_ADDR),
	 .pol  = SPI_POL_VAL_SPI0,
	 .div_sclk = SMU_DIVSP0SCLK,
	 .lock = SPIN_LOCK_UNLOCKED,
	 },
	{
	 .regs = (spi_regs *)(SP1_ADDR),
	 .pol  = SPI_POL_VAL_SPI1,
	 .div_sclk = SMU_DIVSP1SCLK,
	 .lock = SPIN_LOCK_UNLOCKED,
	 },
	{
	 .regs = (spi_regs *)(SP2_ADDR),
	 .pol  = SPI_POL_VAL_SPI2,
	 .div_sclk = SMU_DIVSP2SCLK,
	 .lock = SPIN_LOCK_UNLOCKED,
	 },
};

static int spi_set_sclk_div(spi_data_t * spi, unsigned int sclk)
{
	switch (sclk) {
	case SPI_SCLK_1500KHZ:
	case SPI_SCLK_3MHZ:
	case SPI_SCLK_6MHZ:
	case SPI_SCLK_12MHZ:
	case SPI_SCLK_24MHZ:
	case SPI_SCLK_48MHZ:
		break;
	default:
		return -EINVAL;
	}

	outl(spi_plldiv_table[sclk], spi->div_sclk);

	return 0;
}

static inline int send_data(spi_regs *spi, unsigned int data, unsigned int cnt)
{
	spi->tx_data = data;
	spi->control = (SPx_CONTROL_WRT | SPx_CONTROL_START);
	while ((spi->control & SPx_CONTROL_START) && (--cnt > 0)){}
		;
	if(cnt == 0) {
		DPRINTK(" timeout! \n");
		return -1; /* timeout */
	}
	return 0;
}

int
spi_write_without_intr(const SPI_CONFIG *config, const char *buf, int rst_flag)
{
	spi_data_t *s;
	unsigned long flag;
	unsigned int count, time, nb = 0;
	int i, ret = 0;

	if (config == 0) return -EINVAL;

#ifdef SPI_DEBUG_PARAM
	/* dev check*/
	if (DEVNO_IS_INVALID(config->dev)) {
		return -EINVAL;
	}
#endif

	s = &spi_private[config->dev];
	nb = ((config->mode & SPx_MODE_NB_MASK) >> 8) + 1;

#ifdef SPI_DEBUG_PARAM
	/* m_s check */
	if (config->mode & SPx_MODE_M_S_MASK) { /* slave */
		return -EINVAL;
	}

	/* dma check */
	if (config->mode & SPx_MODE_DMA_MASK) { /* dma on */
		return -EINVAL;
	}

	/* nb check */
	if ((nb < SPI_NB_8BIT) || (nb > SPI_NB_32BIT)) {
		return -EINVAL;
	}

	/* sclk check */
	switch (config->sclk) {
	case SPI_SCLK_1500KHZ:
	case SPI_SCLK_3MHZ:
	case SPI_SCLK_6MHZ:
	case SPI_SCLK_12MHZ:
	case SPI_SCLK_24MHZ:
	case SPI_SCLK_48MHZ:
		break;
	default:
		return -EINVAL;
	}

	/* reset flag check */
	if ((rst_flag != SPI_RESET_OFF) && (rst_flag != SPI_RESET_ON)) {
		return -EINVAL;
	}
#endif

	count = config->size;
	time = RETRY_COUNT(
		(((100000 / spi_clock_table[config->sclk]) * nb) / 10) + 1);

	spin_lock_irqsave(&s->lock, flag); /* lock */
	spi_set_sclk_div(s, config->sclk);
	/* clear */
	s->regs->ffclr = SPx_FFCLR_ALL_CLR;
	s->regs->enclr = SPx_ENCLR_ALL_MASK;
	/* parameters */
	s->regs->mode	= config->mode;
	s->regs->pol	= config->pol;

	/* spi soft reset */
	if (rst_flag == SPI_RESET_ON) { /* reset on */
		s->regs->control |= SPx_CONTROL_RST;
		udelay((1000 * 4 / spi_clock_table[config->sclk]) + 1);
		s->regs->control &= ~SPx_CONTROL_RST;
	}
	if (buf == 0) {
		goto out;
	}

	/* send */
	if (nb > (SPI_NB_8BIT + SPI_NB_16BIT)) { /* 25~32 */
		unsigned int *data = (unsigned int *)buf;
		count = count/4;
		for (i = 0; i < count; i++) {
			if(send_data(s->regs, data[i], time)) {
				ret = -EBUSY;
				goto out;
			}
		}
	}
	else if (nb > SPI_NB_16BIT) { /* 17~24*/
		unsigned char *data = (unsigned char *)buf;
		count = (count / 3) * 3;
		for (i = 0; i < count; i+=3) {
			if(send_data(s->regs, ((data[i] << 16)
				| (data[i+1] << 8) | (data[i+2])), time)) {
				ret = -EBUSY;
				goto out;
			}
		}
	}
	else if ((nb > SPI_NB_8BIT) && (nb <= SPI_NB_16BIT)) { /* 9~16*/
		unsigned short *data = (unsigned short *)buf;
		count = count/2;
		for (i = 0; i < count; i++) {
			if(send_data(s->regs, data[i], time)) {
				ret = -EBUSY;
				goto out;
			}
		}
	}
	else if (nb == SPI_NB_8BIT) {  /* 8 */
		unsigned char *data = (unsigned char *)buf;
		for (i = 0; i < count; i++) {
			if(send_data(s->regs, data[i], time)) {
				ret = -EBUSY;
				goto out;
			}
		}
	}
out:
	/* end */
	spin_unlock_irqrestore(&s->lock, flag); /* unlock */

	return ret;
}

static int __init spi_init(void)
{
	spi_data_t *s0;
	spi_data_t *s1;
	spi_data_t *s2;

	printk(KERN_INFO "Starting %s.\n", SPI_NAME);

	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP0SCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP1SCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP2SCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP0PCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP1PCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP2PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SP0_PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SP1_PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SP2_PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SP0_SCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SP1_SCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SP2_SCLK);
	mp200_pmu_unreset_device(MP200_RESETDEVICE_SP0);
	mp200_pmu_unreset_device(MP200_RESETDEVICE_SP1);
	mp200_pmu_unreset_device(MP200_RESETDEVICE_SP2);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SP0PCLK);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SP1PCLK);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SP2PCLK);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SP0SCLK);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SP1SCLK);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SP2SCLK);

	s0 = &spi_private[SPI_DEV_SPI0];
	/* POL setting */
	s0->regs->pol = s0->pol;

	s1 = &spi_private[SPI_DEV_SPI1];
	/* POL setting */
	s1->regs->pol = s1->pol;

	s2 = &spi_private[SPI_DEV_SPI2];
	/* POL setting */
	s2->regs->pol = s2->pol;

	/* software reset */
	s0->regs->control |= SPx_CONTROL_RST;
	s1->regs->control |= SPx_CONTROL_RST;
	s2->regs->control |= SPx_CONTROL_RST;
	udelay(5);
	s0->regs->control &= ~SPx_CONTROL_RST;
	s1->regs->control &= ~SPx_CONTROL_RST;
	s2->regs->control &= ~SPx_CONTROL_RST;

	return 0;
}

/*
 * exit
 */
static void __exit spi_exit(void)
{

	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP0SCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP1SCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP2SCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP0PCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP1PCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SP2PCLK);
	mp200_pmu_reset_device(MP200_RESETDEVICE_SP0);
	mp200_pmu_reset_device(MP200_RESETDEVICE_SP1);
	mp200_pmu_reset_device(MP200_RESETDEVICE_SP2);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SP0_SCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SP1_SCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SP2_SCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SP0_PCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SP1_PCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SP2_PCLK);

	return;
}

module_init(spi_init);
module_exit(spi_exit);
MODULE_LICENSE("GPL");
EXPORT_SYMBOL(spi_write_without_intr);

