/* 2008-08-19: File added and changed by Sony Corporation */
/*
   -------------------------------------------------------------------------
   i2c-emma.c i2c-hw access for the I2C peripheral on the NEC EMMA
   -------------------------------------------------------------------------

    Shinya Kuribayashi, NEC Electronics Corporation
    shinya.kuribayashi@necel.com
    Takamitsu Honda, NEC Electronics Corporation
    takamitsu.honda@necel.com

    Copyright (C) NEC Electronics Corporation 2002-2005

    Changes made to support the I2C peripheral on the NEC EMMA

   -------------------------------------------------------------------------
    This file was highly leveraged from i2c-elektor.c, which was created
    by Simon G. Vogl and Hans Berglund:


     Copyright (C) 1995-97 Simon G. Vogl
                   1998-99 Hans Berglund

    With some changes from Kylkki <kmalkki@cc.hut.fi> and even
    Frodo Looijaard <frodol@dds.nl>

    Partialy rewriten by Oleg I. Vdovikin for mmapped support of
    for Alpha Processor Inc. UP-2000(+) boards


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   -------------------------------------------------------------------------
*/

#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <linux/interrupt.h>

#include <linux/i2c.h>
#include <linux/i2c-id.h>
#include <linux/i2c-algo-emma.h>
#include "../algos/i2c-emma.h"

struct i2c_emma {
	int base;
	int irq;
	int clock;
	int own;
	int index;
};

#ifdef CONFIG_EMMA2RH
#include <asm/emma2rh/emma2rh.h>

#define MAX_NUM_DEVS    3

struct i2c_emma adaps[MAX_NUM_DEVS] = {
	{
	 .base = KSEG1ADDR(EMMA_PIIC0_BASE),
	 .irq = EMMA_IRQ_PIIC0,
	 .clock = FAST397,
	 .own = 0x40,
	 .index = 0,
	 },
	{
	 .base = KSEG1ADDR(EMMA_PIIC1_BASE),
	 .irq = EMMA_IRQ_PIIC1,
	 .clock = FAST397,
	 .own = 0x44,
	 .index = 1,
	 },
	{
	 .base = KSEG1ADDR(EMMA_PIIC2_BASE),
	 .irq = EMMA_IRQ_PIIC2,
	 .clock = FAST397,
	 .own = 0x48,
	 .index = 2,
	 },
};
#elif defined(CONFIG_EMMA3_FAMILY)
#include <asm/emma3.h>

#if defined(CONFIG_EMMA3)
#define MAX_NUM_DEVS    3
#elif defined(CONFIG_EMMA3P) || defined(CONFIG_EMMA3PF) || defined(CONFIG_EMMA3R)
#define MAX_NUM_DEVS    2
#endif

struct i2c_emma adaps[MAX_NUM_DEVS] = {
	{
	.base = KSEG1ADDR(EMMA_PIIC_CH_BASE(0)),
	.irq = EMMA_PIIC_IRQ(0),
	 .clock = FAST397,
	 .own = 0x40,
	 .index = 0,
	 },
	{
	 .base = KSEG1ADDR(EMMA_PIIC_CH_BASE(1)),
	 .irq = EMMA_PIIC_IRQ(1),
	 .clock = FAST397,
	 .own = 0x44,
	 .index = 1,
	 },
#if MAX_NUM_DEVS > 2
	{
	 .base = KSEG1ADDR(EMMA_PIIC_CH_BASE(2)),
	 .irq = EMMA_PIIC_IRQ(2),
	 .clock = FAST397,
	 .own = 0x48,
	 .index = 2,
	 },
#endif
};
#else
#error unkown EMMA architecture
#endif

static struct i2c_emma *i2c_emma_adaps[MAX_NUM_DEVS] = {
	&adaps[0],
	&adaps[1],
#if MAX_NUM_DEVS > 2
	&adaps[2],
#endif
};

static struct i2c_algo_emma_data *emma_algo_data[MAX_NUM_DEVS];
static struct i2c_adapter *emma_ops[MAX_NUM_DEVS];

static int i2c_debug = 0;
static wait_queue_head_t emma_wait[MAX_NUM_DEVS];
static volatile int emma_pending[MAX_NUM_DEVS];;
static spinlock_t emma_lock[MAX_NUM_DEVS];

/* ----- global defines -----------------------------------------------	*/
#define DEB(x)	if (i2c_debug>=1) x
#define DEB2(x) if (i2c_debug>=2) x
#define DEB3(x) if (i2c_debug>=3) x
#define DEBE(x)	x		/* error messages                               */

/* ----- local functions ----------------------------------------------	*/

static void emma_pci_setbyte(void *data, int ctl, int val)
{
	int address = ((struct i2c_emma *)data)->base + ctl;

	DEB3(printk(KERN_DEBUG "emma_pci_setbyte: Write 0x%08x 0x%08x\n", address, val));
	writel(val, (unsigned int *)address);
}

static int emma_pci_getbyte(void *data, int ctl)
{
	int address = ((struct i2c_emma *)data)->base + ctl;
	int val = readl((unsigned int *)address);

	DEB3(printk(KERN_DEBUG "emma_pci_getbyte: Read 0x%08x 0x%08x\n", address, val));
	return (val);
}

static int emma_pci_getown(void *data)
{
	return (((struct i2c_emma *)data)->own);
}

static int emma_pci_getclock(void *data)
{
	return (((struct i2c_emma *)data)->clock);
}

static void emma_pci_waitforpin(void *data)
{
	int timeout = 2;
	struct i2c_emma *adap = (struct i2c_emma *)data;

	if (adap->irq > 0) {
		timeout = wait_event_timeout(emma_wait[adap->index],
					     emma_pending[adap->index], timeout * HZ);
		emma_pending[adap->index] = 0;
	} else
		udelay(100);
}

static irqreturn_t emma_pci_handler(int this_irq, void *dev_id)
{
	struct i2c_emma *adap = (struct i2c_emma *)dev_id;

	DEB2(printk("emma_pci_handler: status = 0x%08x\n",
		    readl((unsigned int *)(adap->base + I2C_EMMA_INT))));
	/* clear interrupt */
	writel(0, (unsigned int *)(adap->base + I2C_EMMA_INT));

	emma_pending[adap->index] = 1;
	wake_up(&emma_wait[adap->index]);
	return IRQ_HANDLED;
}

static int emma_pci_init(int instance)
{
	struct i2c_emma *adap = i2c_emma_adaps[instance];

	if (adap->irq > 0) {
		if (request_irq(adap->irq, emma_pci_handler, 0, "EMMA I2C", adap) < 0) {
			printk(KERN_ERR "emma_pci_init: Request irq%d failed\n", adap->irq);
			adap->irq = 0;
		} else
			DEB3(printk("emma_pci_init: Enabled interrupt\n"));
	}

	return 0;
}

static void emma_release(void)
{
	int i;

	for (i = 0; i < MAX_NUM_DEVS; i++) {
		struct i2c_emma *adap = (struct i2c_emma *)emma_algo_data[i]->data;
		if (adap->irq > 0) {
			disable_irq(adap->irq);
			free_irq(adap->irq, adap);
		}
		kfree(emma_algo_data[i]);
		kfree(emma_ops[i]);
	}
}

static int emma_pci_reg(struct i2c_client *client)
{
	return 0;
}

static int emma_pci_unreg(struct i2c_client *client)
{
	return 0;
}

static int __init i2c_emma_init(void)
{
	int i;

	printk(KERN_INFO "i2c-emma	: NEC EMMA i2c adapter module\n");

	for (i = 0; i < MAX_NUM_DEVS; i++) {
		emma_algo_data[i] = kmalloc(sizeof(struct i2c_algo_emma_data), GFP_KERNEL);
		if (emma_algo_data[i] == NULL)
			return -ENOMEM;
		memset(emma_algo_data[i], 0, sizeof(struct i2c_algo_emma_data));
		emma_algo_data[i]->data = i2c_emma_adaps[i];
		emma_algo_data[i]->setemma = emma_pci_setbyte;
		emma_algo_data[i]->getemma = emma_pci_getbyte;
		emma_algo_data[i]->getown = emma_pci_getown;
		emma_algo_data[i]->getclock = emma_pci_getclock;
		emma_algo_data[i]->waitforpin = emma_pci_waitforpin;
		emma_algo_data[i]->udelay = 80;
		emma_algo_data[i]->mdelay = 80;
		emma_algo_data[i]->timeout = 100;

		emma_ops[i] = kmalloc(sizeof(struct i2c_adapter), GFP_KERNEL);
		if (emma_ops[i] == NULL)
			return -ENOMEM;
		memset(emma_ops[i], 0, sizeof(struct i2c_adapter));
		/*
		 * ftr revisit: add instance # to name
		 */
		strcpy(emma_ops[i]->name, "EMMA I2C adapter");
		emma_ops[i]->id = 0x00;
		emma_ops[i]->algo = NULL;
		emma_ops[i]->algo_data = emma_algo_data[i];
		emma_ops[i]->client_register = emma_pci_reg;
		emma_ops[i]->client_unregister = emma_pci_unreg;

		init_waitqueue_head(&emma_wait[i]);
		emma_pending[i] = 0;
		spin_lock_init(&emma_lock[i]);
		if (emma_pci_init(i) == 0) {
			if (i2c_emma_add_bus(emma_ops[i]) < 0) {
				/* FIXME : when add_bus() fails at i>=1,
				 * algorithm module is remaining 'used by
				 * adapter modules', therefore couldn't rmmod.
				 * We must success in add_bus() for now.
				 */
				return -ENODEV;
			}
		} else
			return -ENODEV;

		DEB(printk(KERN_INFO "i2c_emma_init: found device at %#x.\n", i2c_emma_adaps[i]->base));
	}

	return 0;
}

static void i2c_emma_exit(void)
{
	int i;

	for (i = 0; i < MAX_NUM_DEVS; i++)
		i2c_emma_del_bus(emma_ops[i]);

	emma_release();
}

MODULE_AUTHOR("NEC Electronics Corporation <www.necel.com>");
MODULE_DESCRIPTION("I2C-Bus adapter routines for EMMA I2C bus adapter");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

module_param(i2c_debug, int, S_IRUGO | S_IWUSR);

module_init(i2c_emma_init);
module_exit(i2c_emma_exit);
