/* 2010-08-25: File added by Sony Corporation */
/*
 * omap35xdc_core.c - driver for OMAP35X Daughter Card
 *
 * Copyright (C) 2008-2009 Texas Instruments, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/module.h>
#include <linux/kernel_stat.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/mutex.h>
#include <linux/i2c.h>
#include <linux/clk.h>
#include <linux/device.h>

#include <asm/io.h>
#include <asm/arch/io.h>
#include <asm/arch/gpmc.h>
#include <asm/arch/omap35xdc.h>


#define MODULE_NAME			"omap35xdc"

#ifdef pr_err
#undef pr_err
#endif

#ifdef pr_info
#undef pr_info
#endif

/* #define OMAP35XDC_DEBUG */

#ifdef OMAP35XDC_DEBUG

#define pr_info(fmt, args...) printk(KERN_INFO MODULE_NAME ": " fmt, ## args)
#define pr_err(fmt, args...) printk(KERN_ERR MODULE_NAME ": " fmt, ## args)

#define FN_IN printk(KERN_INFO MODULE_NAME "[%s]: start\n", __FUNCTION__)
#define FN_OUT(n) printk(KERN_INFO MODULE_NAME "[%s]: end(%u)\n",__FUNCTION__,n)

#define DUMP_GPMC_REGISTERS		(1)

#else

#define pr_info(fmt, args...) printk(KERN_INFO MODULE_NAME ": " fmt, ## args)
#define pr_err(fmt, args...)

#define FN_IN
#define FN_OUT(n)

#define DUMP_GPMC_REGISTERS		(0)

#endif	/* #ifdef OMAP35XDC_DEBUG */


#if DUMP_GPMC_REGISTERS
static void dump_gpmc_regs(void);
#endif	/* #if DUMP_GPMC_REGISTERS */

/* Macro Definitions */

/* System control module register offsets */
#define REG_CONTROL_PADCONF_GPMC_NCS3	(0x480020B4u)
#define REG_CONTROL_PADCONF_I2C2_SDA	(0x480021C0u)
#define REG_CONTROL_PADCONF_I2C3_SDA	(0x480021C4u)

#define PADCONF_I2C3_SCL_MASK		(0xFFFF0000u)
#define PADCONF_I2C3_SDA_MASK		(0x0000FFFFu)
#define PADCONF_GPMC_NCS4_MASK		(0xFFFF0000u)

/* mux mode 0 (enable I2C3 SCL), pull-up enable, input enable */
#define PADCONF_I2C3_SCL_DEF		(0x01180000u)
/* mux mode 0 (enable I2C3 SDA), pull-up enable, input enable */
#define PADCONF_I2C3_SDA_DEF		(0x00000118u)
/* mux mode 0 (enable GPMC CS4), pull-up enable */
#define PADCONF_GPMC_NCS4_DEF		(0x00180000u)


/* OMAP35X Daughter Card Core internal information (BEGIN) */

#define OMAP35XDC_TVP5146_I2C_ADDR	(0x5D)
/* connected to OMAP I2C-3 on OMAP35X daughter card */
#define CONFIG_TVP5146_I2C_BUS_ID	(3)

#define OMAP35XDC_AIC23_I2C_ADDR	(0x1B)
/* connected to OMAP I2C-3 on OMAP35X daughter card */
#define CONFIG_AIC23_I2C_BUS_ID		(3)

/* GPMC chip select */
#define OMAP35XDC_GPMC_CS		(4)

/* Register offsets */
#define REG_MLC_NCS0			(0x00000000u)
#define REG_MLC_NCS1			(0x00000040u)
#define REG_F_ACSN			(0x00000080u)
#define REG_F_SCSN			(0x000000C0u)
#define REG_PATA_NCS			(0x00000100u)
#define REG_BUS_CTRL1			(0x00000180u)
#define REG_BUS_CTRL2			(0x000001C0u)

/* Bit defines for Bus Control 1 register */
#define TVP5146_EN_SHIFT		(0x0000u)
#define TVP5146_EN_MASK			(1u << TVP5146_EN_SHIFT)

#define MCBSP2_EN_SHIFT			(0x0001u)
#define MCBSP2_EN_MASK			(0x0006u)	/* two bits used */
#define MCBSP2_DISCONNECT_BIT		(0x0000u)
#define MCBSP2_AIC23_EN_BIT		(0x0001u)
#define MCBSP2_SPDIF_EN_BIT		(0x0002u)
#define MCBSP2_HDMI_EN_BIT		(0x0003u)

#define BT_MCBSP3_EN_SHIFT		(0x0003u)
#define BT_MCBSP3_EN_MASK		(1u << BT_MCBSP3_EN_SHIFT)

#define JAC_MCSPI2_EN_SHIFT		(0x0004u)
#define JAC_MCSPI2_EN_MASK		(1u << JAC_MCSPI2_EN_SHIFT)

#define BT_UART1_EN_SHIFT		(0x0005u)
#define BT_UART1_EN_MASK		(1u << BT_UART1_EN_SHIFT)

#define MLC_NAND_EN_SHIFT		(0x0006u)
#define MLC_NAND_EN_MASK		(1u << MLC_NAND_EN_SHIFT)

#define JAC_MCBSP1_EN_SHIFT		(0x0007u)
#define JAC_MCBSP1_EN_MASK		(1u << JAC_MCBSP1_EN_SHIFT)

#define IMAGE_SENSOR_EN_SHIFT		(0x0008u)
#define IMAGE_SENSOR_EN_MASK		(1u << IMAGE_SENSOR_EN_SHIFT)

#define JAC_MCBSP3_EN_SHIFT		(0x0009u)
#define JAC_MCBSP3_EN_MASK		(1u << JAC_MCBSP3_EN_SHIFT)

#define SPDIF_MCSPI2_EN_SHIFT		(0x000Au)
#define SPDIF_MCSPI2_EN_MASK		(1u << SPDIF_MCSPI2_EN_SHIFT)

/* default value for bus control registers */
#define BUS_CONTROL1_DEF		(0x0141u)	/* Disable all mux */
#define BUS_CONTROL2_DEF		(0x010Au)	/* Disable all mux */

/* OMAP35X Daughter Card Core internal information (END) */


#define OMAP35XDC_CLIENT_USED		(1)
#define OMAP35XDC_CLIENT_FREE		(0)

/* Helper functions */
static int omap35xdc_detect_client(struct i2c_adapter *adapter,
					unsigned char sid);
static int omap35xdc_attach_adapter(struct i2c_adapter *adapter);
static int omap35xdc_detach_client(struct i2c_client *client);

/* Structure to define on OMAP35XDC Slave ID */
struct omap35xdc_client {
	struct i2c_client client;
	const char client_name[I2C_NAME_SIZE];
	const unsigned char address;
	const char adapter_index;
	unsigned char inuse;

	/* Max number of i2c_msg required is for read = 2 */
	struct i2c_msg xfer_msg[2];

	/* To lock access to xfer_msg */
	struct mutex xfer_lock;
};

/* Structure to define OMAP35XDC GPMC settings */
struct omap35xdc_gpmc {
	u32 *gpmc_phy_base;		/* GPMC device physical base address */
	u32 *gpmc_vir_base;		/* GPMC device virtual base address */

	u16 bus_control1;		/* Bus control register 1 */
	u16 bus_control2;		/* Bus control register 2 */

	/* To lock access to GPMC bus control registers */
	struct mutex reg_lock;
};

/* client information. For adding new client device, add another structure
   member and accordingly add a client id in omap35xdc.h file */
static struct omap35xdc_client omap35xdc_modules[] = {
	{
		.address	= OMAP35XDC_TVP5146_I2C_ADDR,
		.client_name	= "omap35xdc_tvp5146",
		.adapter_index	= CONFIG_TVP5146_I2C_BUS_ID,
	},
	{
		.address	= OMAP35XDC_AIC23_I2C_ADDR,
		.client_name	= "omap35xdc_aic23",
		.adapter_index	= CONFIG_AIC23_I2C_BUS_ID,
	},
};

/* Number of I2C slave devices */
#define OMAP35XDC_NUM_SLAVES	(sizeof(omap35xdc_modules) \
					/ sizeof(struct omap35xdc_client))

/* One client driver, OMAP35XDC_NUM_SLAVES clients */
static struct i2c_driver omap35xdc_driver = {
	.driver	= {
		.name	= MODULE_NAME,
		.owner	= THIS_MODULE,
	},
	.attach_adapter	= omap35xdc_attach_adapter,
	.detach_client	= omap35xdc_detach_client,
};

/* GPMC setting instance */
static struct omap35xdc_gpmc gpmc_handle;


/*
 * @brief write_gpmc_reg - writes to daughter card registers through GPMC.
 *
 * @param val - value to write
 * @param reg - daughter card register offset
 *
 * @note daughter card bus control registers are write only registers. Hence the
 *       the values are updated locally in variables and are used when they are
 *       read.
 */
static __inline void write_gpmc_reg(u16 val, u32 reg)
{
	if (reg == REG_BUS_CTRL1) {
		gpmc_handle.bus_control1 = val;
	}
	else if (reg == REG_BUS_CTRL2) {
		gpmc_handle.bus_control2 = val;
	}

	__raw_writew(val, ((u32)gpmc_handle.gpmc_vir_base + reg));
}

/*
 * @brief read_gpmc_reg - read daughter card register value
 *
 * @param reg - daughter card register offset
 * @return register value
 *
 * @note daughter card bus control registers are write only registers. Hence the
 *       the values are updated locally in variables and are used when they are
 *       read. Read access to other registers are unknown and hence GPMC read
 *       operation is performed
 */
static __inline u16 read_gpmc_reg(u32 reg)
{
	if (reg == REG_BUS_CTRL1) {
		return gpmc_handle.bus_control1;
	}
	else if (reg == REG_BUS_CTRL2) {
		return gpmc_handle.bus_control2;
	}
	else {
		return __raw_readw((u32)gpmc_handle.gpmc_vir_base + reg);
	}
}

/* Global Functions */
/*
 * @brief omap35xdc_i2c_write - Writes n registers in OMAP35XDC I2C device
 *
 * @param mod_no - I2C module number
 * @param reg - register address (just offset will do)
 * @param *value - an array of num_bytes+1 containing data to write
 * IMPORTANT - Allocate value num_bytes+1 and valid data starts at
 *		 Offset 1.
 * @param num_bytes - number of bytes to transfer
 *
 * @return result of operation - 0 is success
 */
int omap35xdc_i2c_write(u8 mod_no, u8 reg, u8 *value, u8 num_bytes)
{
	int ret;
	struct omap35xdc_client *omap35xdc;
	struct i2c_msg *msg;

	if (unlikely(mod_no >= OMAP35XDC_NUM_SLAVES)) {
		pr_err("Invalid module Number\n");
		return -EPERM;
	}
	omap35xdc = &omap35xdc_modules[mod_no];

	if (unlikely(omap35xdc->inuse != OMAP35XDC_CLIENT_USED)) {
		pr_err("I2C Client[%d] is not initialized[%d]\n",
			mod_no, __LINE__);
		return -EPERM;
	}

	mutex_lock(&omap35xdc->xfer_lock);

	/*
	 * [MSG1]: fill the register address data
	 * fill the data Tx buffer
	 */
	msg = &omap35xdc->xfer_msg[0];
	msg->addr = omap35xdc->address;
	msg->len = num_bytes + 1;
	msg->flags = 0;
	msg->buf = value;
	/* over write the first byte of buffer with the register address */
	*value = reg;
	ret = i2c_transfer(omap35xdc->client.adapter, omap35xdc->xfer_msg, 1);

	mutex_unlock(&omap35xdc->xfer_lock);

	/* i2cTransfer returns num messages. Translate it */
	if (ret >= 0)
		ret = 0;
	return ret;
}

/**
 * @brief omap35xdc_i2c_read - Reads a n bit register in OMAP35XDC
 *
 * @param mod_no - module number
 * @param reg - register address (just offset will do)
 * @param *value - an array of num_bytes containing data to be read
 * @param num_bytes - number of bytes to transfer
 *
 * @return result of operation - 0 is success
 */
int omap35xdc_i2c_read(u8 mod_no, u8 reg, u8 *value, u8 num_bytes)
{
	int ret;
	u8 val;
	struct omap35xdc_client *omap35xdc;
	struct i2c_msg *msg;

	if (unlikely(mod_no >= OMAP35XDC_NUM_SLAVES)) {
		pr_err("Invalid module Number\n");
		return -EPERM;
	}
	omap35xdc = &omap35xdc_modules[mod_no];

	if (unlikely(omap35xdc->inuse != OMAP35XDC_CLIENT_USED)) {
		pr_err("I2C Client[%d] is not initialized[%d]\n", mod_no,
			__LINE__);
		return -EPERM;
	}

	mutex_lock(&omap35xdc->xfer_lock);

	/* [MSG1] fill the register address data */
	msg = &omap35xdc->xfer_msg[0];
	msg->addr = omap35xdc->address;
	msg->len = 1;
	msg->flags = 0;		/* Write the register offset */
	val = reg;
	msg->buf = &val;

	/* [MSG2] fill the data rx buffer */
	msg = &omap35xdc->xfer_msg[1];
	msg->addr = omap35xdc->address;
	msg->flags = I2C_M_RD;	/* Read the register values */
	msg->len = num_bytes;	/* only n bytes */
	msg->buf = value;
	ret = i2c_transfer(omap35xdc->client.adapter, omap35xdc->xfer_msg, 2);

	mutex_unlock(&omap35xdc->xfer_lock);

	/* i2cTransfer returns num messages. Translate it */
	if (ret >= 0)
		ret = 0;
	return ret;
}

/**
 * @brief omap35xdc_i2c_write_u8 - Writes a 8 bit register in OMAP35XDC I2C
 *					device
 *
 * @param mod_no - module number
 * @param reg - register address (just offset will do)
 * @param value - the value to be written 8 bit
 *
 * @return result of operation - 0 is success
 */
int omap35xdc_i2c_write_u8(u8 mod_no, u8 reg, u8 value)
{
	/* 2 bytes offset 1 contains the data offset 0 is used by i2c_write */
	u8 temp_buffer[2] = { 0 };
	/* offset 1 contains the data */
	temp_buffer[1] = value;
	return omap35xdc_i2c_write(mod_no, reg, temp_buffer, 1);
}

/**
 * @brief omap35xdc_i2c_read_u8 - Reads a 8 bit register from OMAP35XDC I2C
 *					device
 *
 * @param mod_no - module number
 * @param reg - register address (just offset will do)
 * @param *value - the value read 8 bit
 *
 * @return result of operation - 0 is success
 */
int omap35xdc_i2c_read_u8(u8 mod_no, u8 reg, u8 *value)
{
	return omap35xdc_i2c_read(mod_no, reg, value, 1);
}

/**
 * @brief omap35xdc_set_mux - Sets OMAP35X daughter card mux to enable/disable
 *				signal routing to different peripherals present
 * IMPORTANT - This function will take care of writing appropriate values for
 * active low signals as well
 *
 * @param mux_id - enum, mux id as defined in omap35xdc.h file
 * @param value - enum, ENABLE_MUX for enabling and DISABLE_MUX for disabling
 *
 * @return result of operation - 0 is success
 */
int omap35xdc_set_mux(enum omap35xdc_mux mux_id, enum config_mux value)
{
	int err = 0;
	u16 bus_ctrl;

	if (unlikely(mux_id >= NUM_MUX)) {
		pr_err("Invalid mux id\n");
		return -EPERM;
	}

	mutex_lock(&gpmc_handle.reg_lock);

	switch (mux_id) {
		case MUX_TVP5146:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~TVP5146_EN_MASK;

			/* active low signal. set 0 to enable, 1 to disable */
			if (DISABLE_MUX == value) {
				bus_ctrl |= (1 << TVP5146_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_AIC23_MCBSP2:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~MCBSP2_EN_MASK;

			/* active high signal. set 1 to enable, 0 to disable */
			if (ENABLE_MUX == value) {
				bus_ctrl |= (MCBSP2_AIC23_EN_BIT
						<< MCBSP2_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_SPDIF_MCBSP2:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~MCBSP2_EN_MASK;

			/* active high signal. set 1 to enable, 0 to disable */
			if (ENABLE_MUX == value) {
				bus_ctrl |= (MCBSP2_SPDIF_EN_BIT
						<< MCBSP2_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_HDMI_MCBSP2:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~MCBSP2_EN_MASK;

			/* active high signal. set 1 to enable, 0 to disable */
			if (ENABLE_MUX == value) {
				bus_ctrl |= (MCBSP2_HDMI_EN_BIT
						<< MCBSP2_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_BT_MCBSP3:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~BT_MCBSP3_EN_MASK;

			/* active high signal. set 1 to enable, 0 to disable */
			if (ENABLE_MUX == value) {
				bus_ctrl |= (1 << BT_MCBSP3_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_JAC_MCSPI2:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~JAC_MCSPI2_EN_MASK;

			/* active high signal. set 1 to enable, 0 to disable */
			if (ENABLE_MUX == value) {
				bus_ctrl |= (1 << JAC_MCSPI2_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_BT_UART1:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~BT_UART1_EN_MASK;

			/* active high signal. set 1 to enable, 0 to disable */
			if (ENABLE_MUX == value) {
				bus_ctrl |= (1 << BT_UART1_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_MLC_NAND:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~MLC_NAND_EN_MASK;

			/* active low signal. set 0 to enable, 1 to disable */
			if (DISABLE_MUX == value) {
				bus_ctrl |= (1 << MLC_NAND_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_JAC_MCBSP1:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~JAC_MCBSP1_EN_MASK;

			/* active high signal. set 1 to enable, 0 to disable */
			if (ENABLE_MUX == value) {
				bus_ctrl |= (1 << JAC_MCBSP1_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_IMAGE_SENSOR:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~IMAGE_SENSOR_EN_MASK;

			/* active low signal. set 0 to enable, 1 to disable */
			if (DISABLE_MUX == value) {
				bus_ctrl |= (1 << IMAGE_SENSOR_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_JAC_MCBSP3:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~JAC_MCBSP3_EN_MASK;

			/* active high signal. set 1 to enable, 0 to disable */
			if (ENABLE_MUX == value) {
				bus_ctrl |= (1 << JAC_MCBSP3_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case MUX_SPDIF_MCSPI2:
			bus_ctrl = read_gpmc_reg(REG_BUS_CTRL1);
			bus_ctrl &= ~SPDIF_MCSPI2_EN_MASK;

			/* active high signal. set 1 to enable, 0 to disable */
			if (ENABLE_MUX == value) {
				bus_ctrl |= (1 << SPDIF_MCSPI2_EN_SHIFT);
			}

			write_gpmc_reg(bus_ctrl, REG_BUS_CTRL1);
			break;

		case NUM_MUX:
		default:
			pr_err("Invalid mux id\n");
			err = -EPERM;
	}

	mutex_unlock(&gpmc_handle.reg_lock);

	return err;
}

/* attach a client to the adapter */
static int omap35xdc_detect_client(struct i2c_adapter *adapter, u8 mod_no)
{
	int err = 0;
	struct omap35xdc_client *omap35xdc;

	if (unlikely(mod_no >= OMAP35XDC_NUM_SLAVES)) {
		pr_err("mod_no[%d] >= MOD_LAST[%d]\n", mod_no,
			OMAP35XDC_NUM_SLAVES);
		return -EPERM;
	}

	/* Check basic functionality */
	if (!(err = i2c_check_functionality(adapter, (I2C_FUNC_SMBUS_WORD_DATA |
						I2C_FUNC_SMBUS_WRITE_BYTE)))) {
		pr_err("SlaveID=%d functionality check failed\n", mod_no);
		return err;
	}

	omap35xdc = &omap35xdc_modules[mod_no];
	if (unlikely(omap35xdc->inuse)) {
		pr_err("Client %s is already in use \n",
			omap35xdc->client_name);
		return -EPERM;
	}

	memset(&omap35xdc->client, 0, sizeof(struct i2c_client));

	omap35xdc->client.addr = omap35xdc->address;
	omap35xdc->client.adapter = adapter;
	omap35xdc->client.driver = &omap35xdc_driver;

	memcpy(&omap35xdc->client.name, omap35xdc->client_name,I2C_NAME_SIZE);

	pr_info("TRY attach Slave %s on Adapter %s \n",
		omap35xdc->client_name, adapter->name);

	if ((err = i2c_attach_client(&omap35xdc->client))) {
		pr_err("Couldn't attach Slave %s on Adapter %s [%x] \n",
			omap35xdc->client_name, adapter->name, err);
		omap35xdc->inuse = OMAP35XDC_CLIENT_FREE;
	} else {
		omap35xdc->inuse = OMAP35XDC_CLIENT_USED;
		mutex_init(&omap35xdc->xfer_lock);
	}

	return err;
}

/* adapter callback */
static int omap35xdc_attach_adapter(struct i2c_adapter *adapter)
{
	int i;
	int ret = 0;
	static int omap35xdc_i2c_adapter = 1;

	for (i = 0; i < OMAP35XDC_NUM_SLAVES; i++) {
		/* Check if I need to hook on to this adapter or not */
		if (omap35xdc_modules[i].adapter_index
				== omap35xdc_i2c_adapter) {
			if ((ret = omap35xdc_detect_client(adapter, i)))
				goto free_client;
		}
	}
	omap35xdc_i2c_adapter++;

	return 0;

free_client:
	pr_err("omap35xdc: Client[Idx=%d] registration failed[0x%x]\n",
		i, ret);

	/* ignore current slave..it never got registered */
	i--;
	while (i >= 0) {
		/* now remove all those from the current adapter... */
		if (omap35xdc_modules[i].adapter_index
				== omap35xdc_i2c_adapter)
			(void)omap35xdc_detach_client(
						&omap35xdc_modules[i].client);
		i--;
	}
	return ret;
}

/* adapter's callback */
static int omap35xdc_detach_client(struct i2c_client *client)
{
	int err;
	if ((err = i2c_detach_client(client))) {
		pr_err("Client detach failed\n");
		return err;
	}
	return 0;
}


/*
 * GPMC CS4 timing to access OMAP35X daughter card registers - values in ns.
 * Timings based on non-muxed (limited address) GPMC mode of operation
 */
static struct gpmc_timings omap35xdc_gpmc_timings = {
	.sync_clk = 20,		/* Don't care for async operation */

	.cs_on = 30,		/* Assertion time */
	.cs_rd_off = 96,	/* Read deassertion time */
	.cs_wr_off = 150,	/* Write deassertion time */

	.adv_on = 30,		/* Assertion time */
	.adv_rd_off = 12,	/* Read deassertion time */
	.adv_wr_off = 150,	/* Write deassertion time */

	.we_on = 60,		/* WE assertion time */
	.we_off = 150,		/* WE deassertion time */

	.oe_on = 30,		/* OE assertion time */
	.oe_off = 150,		/* OE deassertion time */

	.page_burst_access = 6,
	.access = 90,		/* Start-cycle to first data valid delay */

	.rd_cycle = 102,	/* Total read cycle time */
	.wr_cycle = 186,	/* Total write cycle time */
};

#if DUMP_GPMC_REGISTERS
static void dump_gpmc_regs(void)
{
    pr_info("Register Register GPMC_CONFIG (0x6E000050) = 0x%.8x\n",
    		__raw_readl(IO_ADDRESS(0x6E000050)));

    pr_info("Register Register GPMC_CONFIG1_4 (0x6E000120) = 0x%.8x\n",
    		__raw_readl(IO_ADDRESS(0x6E000120)));
    pr_info("Register Register GPMC_CONFIG2_4 (0x6E000124) = 0x%.8x\n",
    		__raw_readl(IO_ADDRESS(0x6E000124)));
    pr_info("Register Register GPMC_CONFIG3_4 (0x6E000128) = 0x%.8x\n",
    		__raw_readl(IO_ADDRESS(0x6E000128)));
    pr_info("Register Register GPMC_CONFIG4_4 (0x6E00012C) = 0x%.8x\n",
    		__raw_readl(IO_ADDRESS(0x6E00012C)));
    pr_info("Register Register GPMC_CONFIG5_4 (0x6E000130) = 0x%.8x\n",
    		__raw_readl(IO_ADDRESS(0x6E000130)));
    pr_info("Register Register GPMC_CONFIG7_4 (0x6E000138) = 0x%.8x\n",
    		__raw_readl(IO_ADDRESS(0x6E000138)));
}
#endif	/* #if DUMP_GPMC_REGISTERS */

/*
 * GPMC init function - acquires and configures OMAP35XDC_GPMC_CS GPMC chip
 * select
 */
static int omap35xdc_gpmc_init(void)
{
	int err;
	int cs = OMAP35XDC_GPMC_CS;
	u32 *gpmc_phy_base, *gpmc_vir_base;

	if ((err = gpmc_cs_set_timings(cs, &omap35xdc_gpmc_timings)) < 0) {
		pr_err("Couldn't set GPMC timing for CS4 \n");
		return err;
	}

	if ((err = gpmc_cs_request(cs, SZ_16M,
					(unsigned long *)&gpmc_phy_base)) < 0) {
		pr_err("gpmc_cs_request failed: err = %d\n\n",err);
		return err;
	}

	/* remap the GPMC memory - disable cache since we deal with registers */
	gpmc_vir_base = (u32 *)ioremap_nocache((unsigned long)gpmc_phy_base,
						SZ_16M);

	pr_info("GPMC cs%d request passed. Physical base: 0x%p "
		"Virtual base: 0x%p \n", cs, gpmc_phy_base, gpmc_vir_base);

	if (gpmc_vir_base != NULL) {
		mutex_init(&gpmc_handle.reg_lock);

		gpmc_handle.gpmc_phy_base = gpmc_phy_base;
		gpmc_handle.gpmc_vir_base = gpmc_vir_base;

		mutex_lock(&gpmc_handle.reg_lock);

		/* write default values */
		write_gpmc_reg(BUS_CONTROL1_DEF, REG_BUS_CTRL1);
		write_gpmc_reg(BUS_CONTROL2_DEF, REG_BUS_CTRL2);

		mutex_unlock(&gpmc_handle.reg_lock);
	}

#if DUMP_GPMC_REGISTERS
	dump_gpmc_regs();
#endif	/* #if DUMP_GPMC_REGISTERS */

	return err;
}

/* GPMC exit function - frees the OMAP35XDC_GPMC_CS GPMC chip select */
static void omap35xdc_gpmc_exit(void)
{
	gpmc_cs_free(OMAP35XDC_GPMC_CS);
}

/* module init function - should be called before any client driver init call */
static int __init omap35xdc_init(void)
{
	int err;

	/*
	 * I2C3 SCL pin mux settings - mux mode 0, pull-up enable, input enable
	 * Uses the MSB 16-bit of this register, retain the LSB 16-bit.
	 * This pin is shared with gpio_184 (mux mode 4)
	 */
	omap_writel(((omap_readl(REG_CONTROL_PADCONF_I2C2_SDA) &
			~PADCONF_I2C3_SCL_MASK) | PADCONF_I2C3_SCL_DEF),
			REG_CONTROL_PADCONF_I2C2_SDA);

	/*
	 * I2C3 SDA pin mux settings - mux mode 0, pull-up enable, input enable
	 * Uses the LSB 16-bit of this register, retain the MSB 16-bit.
	 * This pin is shared with gpio_185 (mux mode 4)
	 */
	omap_writel(((omap_readl(REG_CONTROL_PADCONF_I2C3_SDA) &
			~PADCONF_I2C3_SDA_MASK) | PADCONF_I2C3_SDA_DEF),
			REG_CONTROL_PADCONF_I2C3_SDA);

	/*
	 * GPMC CS4 pin mux settings - mux mode 0, pull-up enable
	 * Uses the MSB 16-bit of this register, retain the LSB 16-bit.
	 * This pin is shared with sys_ndmareq1 (mux mode 1), mcbsp4_clkx
	 * (mux mode 2), gpt9_pwm_evt (mux mode 3), gpio_55 (mux mode 4)
	 */
	omap_writel(((omap_readl(REG_CONTROL_PADCONF_GPMC_NCS3) &
			~PADCONF_GPMC_NCS4_MASK) | PADCONF_GPMC_NCS4_DEF),
			REG_CONTROL_PADCONF_GPMC_NCS3);

	if ((err = omap35xdc_gpmc_init())) {
		pr_err("GPMC init failed \n");
		return err;
	}

	if ((err = i2c_add_driver(&omap35xdc_driver))) {
		omap35xdc_gpmc_exit();
		pr_err("I2C driver registration failed \n");
		return err;
	}

	pr_info("Driver registration complete \n");

	return err;
}

/* module exit function */
static void __exit omap35xdc_exit(void)
{
	i2c_del_driver(&omap35xdc_driver);
	omap35xdc_gpmc_exit();
}

subsys_initcall(omap35xdc_init);
module_exit(omap35xdc_exit);

EXPORT_SYMBOL(omap35xdc_i2c_write_u8);
EXPORT_SYMBOL(omap35xdc_i2c_read_u8);
EXPORT_SYMBOL(omap35xdc_i2c_read);
EXPORT_SYMBOL(omap35xdc_i2c_write);
EXPORT_SYMBOL(omap35xdc_set_mux);

MODULE_AUTHOR("Texas Instruments, Inc.");
MODULE_DESCRIPTION("I2C and GPMC interface for OMAP35X DC");
MODULE_LICENSE("GPL");
