/* 2009-08-24: File added by Sony Corporation */
/**
 ** @file  elan_tp.c
 ** @brief Device driver for ClearPad(TouchPanel) manufactured
 ** by ELAN Microelectronics(http://www.emc.com.tw/eng/index.asp)
 ** @author Hiroyuki.Motousu@jp.sony.com
 **
 ** $Id:$
 **
 **  Copyright 2009 Sony Corporation.
 **
 **
 ** This program is free software; you can redistribute  it and/or modify it
 ** under  the terms of  the GNU General  Public License as published by the
 ** Free Software Foundation;  version 2 of the  License.
 **
 ** THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 ** WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 ** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 ** NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 ** INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 ** NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 ** USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 ** ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 ** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 ** THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **
 ** You should have received a copy of the  GNU General Public License along
 **  with this program; if not, write  to the Free Software Foundation, Inc.,
 **  675 Mass Ave, Cambridge, MA 02139, USA.
 **
 **/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/freezer.h>
#include <linux/slab.h>
#include <linux/ctype.h>
#include <linux/types.h>
#include <linux/device.h>
#include <linux/i2c.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/cdev.h>

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/arch/pgpf/devnum.h>
#include <asm/arch/irqs.h>
#include <asm/arch/pwc.h>
#include <asm/arch/gpio.h>

#include <asm/arch/elan_tp.h>

static struct i2c_driver elan_i2c_client_driver;

#define	ELAN_DRIVER_NAME	    (elan_i2c_client_driver.driver.name)
#define	ELAN_DRIVER_VERSION	    "29 Jul,2009"

#define X_AXIS_MIN        0
#define X_AXIS_MAX        1214
#define Y_AXIS_MIN        0
#define Y_AXIS_MAX        702
#define PRESSURE_AXIS_MIN 0
#define PRESSURE_AXIS_MAX X_AXIS_MAX

#define X_AXIS_VAAA_MIN -1
#define X_AXIS_VAAA_MAX 1215
#define Y_AXIS_VAAA_MIN -1
#define Y_AXIS_VAAA_MAX 703

#define ELAN_FINGER_LEVEL_MIN       0
#define ELAN_FINGER_LEVEL_MAX       3
#define ELAN_SENSITIVITY_RANGE_MIN  0  /* 4.8" change point */
#define ELAN_SENSITIVITY_RANGE_MAX  15 /* 4.8" change point */

#define ELAN_HELLO_PACKET_SIZE      4
#define ELAN_NORMAL_PACKET_SIZE     6
#define ELAN_READ_PACKET_SIZE       4
#define ELAN_RESPONSE_PACKET_SIZE   4
#define ELAN_WRITE_PACKET_SIZE      4

#define ELAN_USER_WRITABLE_REG_SIZE 2

#define ELAN_ABSOLUTE_COORDINATE_PACKET_SIZE  (ELAN_NORMAL_PACKET_SIZE)
#define ELAN_INDEX_COORDINATE_PACKET_SIZE     8 

#define ELAN_PACKET_REPORT_ABSOLUTE_COORDINATE 0x51
#define ELAN_PACKET_RESPONSE_READ_REGISTER     0x52
#define ELAN_PACKET_READ_REGISTER              0x53
#define ELAN_PACKET_WRITE_REGISTER             0x54
#define ELAN_PACKET_READY_TO_WORK              0x55
#define ELAN_PACKET_REPORT_INDEX_COORDINATE    0x5A
/* register id for read  */
#define ELAN_REG_NUMBER_READ_FW_VERSION         0x00
#define ELAN_REG_NUMBER_READ_BUTTON_POSITION    0x01 /* new from 4.8 inch panel */
#define ELAN_REG_NUMBER_READ_X_ABS_POSITION     0x02
#define ELAN_REG_NUMBER_READ_Y_ABS_POSITION     0x03
#define ELAN_REG_NUMBER_READ_SENSITIVITY_STATE  0x04
#define ELAN_REG_NUMBER_READ_POWER_STATE        0x05
#define ELAN_REG_NUMBER_READ_COORDINATE_TYPE    0x0B
#define ELAN_REG_NUMBER_READ_FW_ID              0x0F
/* register id for write */
#define ELAN_REG_NUMBER_WRITE_SENSITIVITY_STATE (ELAN_REG_NUMBER_READ_SENSITIVITY_STATE)
#define ELAN_REG_NUMBER_WRITE_POWER_STATE       (ELAN_REG_NUMBER_READ_POWER_STATE)
#define ELAN_REG_NUMBER_WRITE_COORDINATE_TYPE   (ELAN_REG_NUMBER_READ_COORDINATE_TYPE)    


struct elan_data {
    struct i2c_client  client;

    struct input_dev  *dev;
    struct input_dev  *key_dev;
    struct cdev        cdev;
    struct semaphore   sem;
    wait_queue_head_t  wait_queue;
    int                flag;
    struct semaphore   rdsem;
    wait_queue_head_t  rq;
    int                rqflag;
    
    /* data part */
    int                 status;
    int                 x_position; 
    int                 y_position; 
    int                 x2_position;
    int                 y2_position;
    int                 pressure;
    int                 upper_button;
    int                 lower_button;
    int                 finger_level;

    elan_finger_sens_t  sensitivity;
    elan_coordinate_t   coordinate_type;
    elan_finger_level_t fl_threshold;
    elan_power_state_t  power_state;

    elan_y_direction_t  y_axis_inversion;

    int                 used;
    u32                 clearpad_installed;

    u8                  reg_data[ELAN_RESPONSE_PACKET_SIZE];
    unsigned int        irq_type;

    elan_fw_info_t      fw_info;
};


/* Legacy Style (with assignment of bus number) */
static unsigned short normal_i2c[] = { I2C_CLIENT_END };
static unsigned short probe[]      = { 0, 0x10, I2C_CLIENT_END };
static unsigned short ignore[]     = { I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
    .normal_i2c = normal_i2c,
    .probe      = probe,
    .ignore     = ignore,
};


static int elan_major   = PGPF_BI1050_MAJOR;
static int elan_minor   = PGPF_BI1050_MINOR;
static int elan_nr_devs = 1;

static int         elan_read_fw_info(struct elan_data *data);
static int         elan_get_fw_info(struct elan_data *data, elan_fw_info_t __user *arg);
static int         elan_set_coordinate_type(struct elan_data *data, elan_coordinate_t __user *arg);
static int         elan_get_coordinate_type(struct elan_data *data, elan_coordinate_t __user *arg);
static int         elan_set_power_state(struct elan_data *data, elan_power_state_t __user *arg);
static int         elan_get_power_state(struct elan_data *data, elan_power_state_t __user *arg);
static int         elan_check_finger_level(elan_finger_level_t *arg);
static int         elan_get_fl_threshold(struct elan_data *data, elan_finger_level_t *arg);
static int         elan_set_fl_threshold(struct elan_data *data, elan_finger_level_t *arg);
static int         elan_get_sensitivity(struct elan_data *data, elan_finger_sens_t *arg);
static int         elan_read_sensitivity(struct elan_data *data);
static int         elan_set_sensitivity(struct elan_data *data, elan_finger_sens_t *arg);
static int         elan_reg_read(struct elan_data *data, u8 reg_addr, u8 *read_data);
static int         elan_reg_write(struct elan_data *data, u8 reg_addr, u8 *write_data);
static int         elan_receive_packet(struct elan_data *data);
static int         elan_thread(void *arg);
static int         elan_write_coordinate_type(struct elan_data *data);
static int         elan_write_power_state(struct elan_data *data);
static int         elan_check_sensitivity_range(elan_finger_sens_t *arg);
static int         elan_write_sensitivity(struct elan_data *data);
static int         elan_setup(struct elan_data *data);
static int         elan_check_connection(struct elan_data *data);
static void        elan_setup_cdev(struct elan_data *data, int index);
static void        elan_change_resetline(int value);
static void        elan_initialize_data(struct elan_data *data);
static void        elan_wait_for_device_is_ready(int forced);
static void        elan_align_finger1_vaaa_area_data(struct elan_data *data);
static void        elan_align_finger2_vaaa_area_data(struct elan_data *data);

/* interrupt detection sets as asynchronous. */
/* This is a tentative                       */
#define INTERRUPT_DETECTION_ASYNCHRONOUS_090407 0
spinlock_t    elan_gio_lock = SPIN_LOCK_UNLOCKED;

/*                             */
/* for finger level evaluation */
/*                             */
#define FINGER_LEVEL_EVALUATION 1

// INTERRUPT_DETECTION_ASYNCHRONOUS_090407-------------------->
#if INTERRUPT_DETECTION_ASYNCHRONOUS_090407
static int elan_set_irq_detection(struct elan_data *data)
{
    int           err  = 0;
    unsigned int  pin = PGPF_GPIO_TP_XINT;
    //unsigned int  irq = gpio_to_irq(pin);
	unsigned int  mode = 0;
	unsigned long flag;
	unsigned int  oiia;
	unsigned int  x;
	unsigned int  mask;
	unsigned int  idtshift;
	unsigned int  idt, iia, iir;

    switch (data->irq_type & IRQ_TYPE_SENSE_MASK) {
	case IRQ_TYPE_EDGE_RISING:
        mode |= 0x08;
		break;
	case IRQ_TYPE_EDGE_FALLING:
		mode |= 0x09;
		break;
	case IRQ_TYPE_LEVEL_HIGH:
		mode |= 0x0a;
		break;
	case IRQ_TYPE_LEVEL_LOW:
        mode |= 0x0b;
		break;
	case IRQ_TYPE_EDGE_BOTH:
		mode |= 0x0c;
		break;
	default:
		return -EINVAL;
	}

    mask = (1U << ((pin) & 0x1f));
	idtshift = (pin & 0x7) << 2;	/* IDT shift bit */
    
	spin_lock_irqsave(&elan_gio_lock, flag);

	idt = VA_GIO + GP_IDT0L;
    iia = VA_GIO + GP_IIAL;
    iir = VA_GIO + GP_IIRL;

	/* IIA enable -> disable */
	oiia = __raw_readl(iia);
	if ((oiia & mask) != 0) {
		__raw_writel((oiia & ~(mask)), iia);
	}

	/* set IDT */
	x = __raw_readl(idt);
	if ((x & (0xfU << idtshift)) != (mode << idtshift)) {
		x &= ~(0xfU << idtshift);
		x |= (mode << idtshift);
		__raw_writel(x, idt);
	}

	/* Interrupt clear */
	__raw_writel(mask, iir);

	/* Restore if changed */
	if ((oiia & mask) != 0) {
		__raw_writel(oiia, iia);
	}

	spin_unlock_irqrestore(&elan_gio_lock, flag);

    return err;
}
#endif // INTERRUPT_DETECTION_ASYNCHRONOUS_090407
// <--------------------INTERRUPT_DETECTION_ASYNCHRONOUS_090407

//######################################################################
/** 
 * align data of VA-AA area data for 1st. finger from application request
 * 
 * @param data : data structure
 */
static void elan_align_finger1_vaaa_area_data(struct elan_data *data)
{
    if(((data->fw_info.major_ver == 0) && (data->fw_info.minor_ver < 4))){
        if( data->status == 0 ){
            return;
        }
    }

    if( data->x_position == X_AXIS_MIN ){
        data->x_position = X_AXIS_VAAA_MIN;
    } else if( data->x_position == X_AXIS_MAX ){
        data->x_position = X_AXIS_VAAA_MAX;
    }
    
    if( data->y_position == Y_AXIS_MIN ){
        data->y_position = Y_AXIS_VAAA_MIN;
    } else if( data->y_position == Y_AXIS_MAX ){
        data->y_position = Y_AXIS_VAAA_MAX;
    }

    return;
}

//######################################################################
/** 
 * align data of VA-AA area data for 2nd. finger from application request
 * 
 * @param data : data structure
 */
static void elan_align_finger2_vaaa_area_data(struct elan_data *data)
{
    if(((data->fw_info.major_ver == 0) && (data->fw_info.minor_ver < 4))){
        if( data->status == 0 ){
            return;
        }
    }
    
    if( data->x2_position == X_AXIS_MIN ){
        data->x2_position = X_AXIS_VAAA_MIN;
    } else if( data->x2_position == X_AXIS_MAX ){
        data->x2_position = X_AXIS_VAAA_MAX;
    }
    
    if( data->y2_position == Y_AXIS_MIN ){
        data->y2_position = Y_AXIS_VAAA_MIN;
    } else if( data->y2_position == Y_AXIS_MAX ){
        data->y2_position = Y_AXIS_VAAA_MAX;
    }

    return;
}

//######################################################################
static inline struct i2c_client * elan_get_client(struct elan_data * p)
{
    return &(p->client);
}

//######################################################################
/** 
 * align x_position & y_position data when finger up.
 * Because T/P report packets as below. "0x51 0xff 0xff 0xff 0xff 0x01"
 * 
 * @param data : data 
 */
static inline void elan_align_finger_up_data_at_absolute_type(struct elan_data *data)
{
    if( data->status == 0 ){
        data->x_position = 0;
        data->y_position = 0;
    }
    return;
}

/* ################################################################## */
/** 
 * evaluate a current finger level wheather it's out of threshold range
 * or not.
 * 
 * @param data         : a sort of client data 
 * 
 * @return  1 : OK, 0 : NG 
 */
static inline int elan_event_should_be_reported(struct elan_data *data)
{
    /* finger level is inside of threshold range. */
    return ((data->finger_level >= data->fl_threshold.min) &&
            (data->finger_level <= data->fl_threshold.max));
}

//######################################################################
/** 
 * y-axis data alignment for y position
 * 
 * @param p : data 
 * 
 * @return y1 position data
 */
static inline int elan_get_aligned_y_position(struct elan_data * p)
{
    if(((p->fw_info.major_ver == 0) && (p->fw_info.minor_ver < 4))){
        /* check whether release event or not */
        if( p->status == 0 ){
            /*
             * because when f/w ver. is lower than 0.4,
             * release event packet should be (x,y)=(0,0).
             * This is "y_axis_inversion" provision.
             */
            return 0;
        }
    }

    return (p->y_axis_inversion ? (Y_AXIS_MAX - p->y_position) : p->y_position);
}

//######################################################################
/** 
 * y-axis data alignment for y2 position
 * 
 * @param p : data 
 * 
 * @return : y2 position data
 */
static inline int elan_get_aligned_y2_position(struct elan_data * p)
{
    if(((p->fw_info.major_ver == 0) && (p->fw_info.minor_ver < 4))){
        /* check whether release event or not */
        if( p->status == 0 ){
            /*
             * because when f/w ver. is lower than 0.4,
             * release event packet is sent as (x,y)=(0,0).
             * This is "y_axis_inversion" provision.
             */
            return 0;
        }
    }

    return (p->y_axis_inversion ? (Y_AXIS_MAX - p->y2_position) : p->y2_position);
}

//######################################################################
/** 
 *  get firmware version & id
 * 
 * @param data : a sort of client data
 * 
 * @return 0 : success, other : failure
 */
static int elan_read_fw_info(struct elan_data *data)
{
    int err = 0;
    u8  read_data[ELAN_READ_PACKET_SIZE];

    memset(read_data, 0, sizeof(read_data));

    err = elan_reg_read(data, ELAN_REG_NUMBER_READ_FW_VERSION, read_data);

    if(err){
        printk(KERN_ERR "%s() : failed to get data packets(%d).\n",__FUNCTION__, err);
        err = -EBUSY;
        return err;
    }
    
    if( read_data[0] == ELAN_PACKET_RESPONSE_READ_REGISTER ){
        data->fw_info.major_ver = ((read_data[1] & 0x0F) << 4) | ((read_data[2] & 0xF0) >> 4);
        data->fw_info.minor_ver = ((read_data[2] & 0x0F) << 4) | ((read_data[3] & 0xF0) >> 4);
    }
    else{
        printk(KERN_ERR "%s() : improper response packet header(0x%02x).\n",__FUNCTION__, read_data[0]);
        err = -EINVAL;
        return err;
    }

    memset(read_data, 0, sizeof(read_data));
    
    err = elan_reg_read(data, ELAN_REG_NUMBER_READ_FW_ID, read_data);

    if(err){
        printk(KERN_ERR "%s() : failed to get data packets(%d).\n",__FUNCTION__, err);
        err = -EBUSY;
        return err;
    }

    if( read_data[0] == ELAN_PACKET_RESPONSE_READ_REGISTER ){
        data->fw_info.id_high = ((read_data[1] & 0x0F) << 4) | ((read_data[2] & 0xF0) >> 4);
        data->fw_info.id_low  = ((read_data[2] & 0x0F) << 4) | ((read_data[3] & 0xF0) >> 4);
    }
    else{
        printk(KERN_ERR "%s() : improper response packet header(0x%02x).\n",__FUNCTION__, read_data[0]);
        err = -EINVAL;
        return err;
    }

    return err;
}

//######################################################################
/** 
 * get device firmware version & id by ioctl
 * 
 * @param data : a sort of client data 
 * @param arg  : f/w info structure
 * 
 * @return 0 : success, other : failure
 */
static int elan_get_fw_info(struct elan_data *data, elan_fw_info_t __user *arg)
{
    int err = 0;

    err = elan_read_fw_info(data);
    if(err){
        return err;
    }
    
    err = copy_to_user((elan_fw_info_t __user *)arg, &data->fw_info, sizeof(elan_fw_info_t));
    
    return err;
}

//######################################################################
/** 
 * get a coordinate type by ioctl
 * 
 * @param data : a sort of client data
 * @param arg  : packet coordinate type(user output data)
 * 
 * @return 0 : success, other : failure
 */
static int elan_get_coordinate_type(struct elan_data *data, elan_coordinate_t __user *arg)
{
    int err = 0;
    elan_coordinate_t type;
    u8  rx_buf[ELAN_READ_PACKET_SIZE];

    memset(rx_buf, 0, sizeof(rx_buf));

    err = elan_reg_read(data, ELAN_REG_NUMBER_READ_COORDINATE_TYPE, rx_buf);

    if(err){
        printk(KERN_ERR "%s(): failed to get data packets(%d).\n",__FUNCTION__, err);
        return -EBUSY;
    }
    
    /* check a responce packet */
    if(rx_buf[0] == ELAN_PACKET_RESPONSE_READ_REGISTER){
        /* 4.8" change point */
        type = (rx_buf[2] & 0x01);
    }
    else{
        printk(KERN_ERR "%s() : improper response packet header(0x%02x).\n",
               __FUNCTION__, rx_buf[0]);
        return -EINVAL;
    }

    data->coordinate_type = type;

    err = put_user(type, (elan_coordinate_t __user *)arg);

    return err;
}

//######################################################################
/** 
 * modify a report coordinate type & check user input data by ioctl
 * 
 * @param data : asort of client data
 * @param arg  : report coordinate type(user input data)
 * 
 * @return 0 : success, other : failure
 */
static int elan_set_coordinate_type(struct elan_data *data, elan_coordinate_t __user *arg)
{
    int err = 0;
    elan_coordinate_t type     = 0;
    elan_coordinate_t old_type = data->coordinate_type;
    u8  tx_buf[ELAN_USER_WRITABLE_REG_SIZE];

    memset(tx_buf, 0, sizeof(tx_buf));
    
    err = get_user(type, arg);
    if( err != 0 ){
        return err;  /* = EFAULT */
    }
    
    /* check a user input value */
    if( (type != ELAN_COORDINATE_TYPE_ABSOLUTE) && (type != ELAN_COORDINATE_TYPE_INDEX) ){
        printk(KERN_WARNING "%s() : undefined coordinate type(%d).\n",__FUNCTION__, type);
        return -EINVAL;
    }
    
    data->coordinate_type = type;

    err = elan_write_coordinate_type(data);

    if(!err){
        printk(KERN_INFO "%s() : report coordinate type is changed(%d->%d).\n",
               __FUNCTION__, old_type, data->coordinate_type);
    }
    else{
        printk(KERN_WARNING "%s() : change of coordinate type is failed(->%d).\n",
               __FUNCTION__, data->coordinate_type);
        data->coordinate_type = old_type;
    }
    
    return err;
}

/* ################################################################## */
/** 
 * modify a power state & check a user input data range by ioctl
 * 
 * @param data : a sort of client data
 * @param arg  : power state(user input data)
 * 
 * @return 0 : success, other : failure
 */
static int elan_set_power_state(struct elan_data *data, elan_power_state_t *arg)
{
    int err = 0;
    elan_power_state_t state;
    elan_power_state_t old_state = data->power_state;
    u8  tx_buf[ELAN_USER_WRITABLE_REG_SIZE];

    memset(tx_buf, 0, sizeof(tx_buf));
    
    err = get_user(state, arg);
    if( err != 0 ){
        return err;  /* = EFAULT */
    }

    if( (state < ELAN_POWER_STATE_DEEP_SLEEP) || (state > ELAN_POWER_STATE_NORMAL) ){
        printk(KERN_WARNING "%s() : undefined power state(%d).\n",__FUNCTION__, state);
        return -EINVAL;
    }
    data->power_state = state;

    err = elan_write_power_state(data);

    if(!err){
        printk(KERN_INFO "%s() : power state is changed(%d->%d).\n",
               __FUNCTION__, old_state, data->power_state);
    }
    else{
        printk(KERN_WARNING "%s() : change of power state is failed(->%d).\n",
               __FUNCTION__, data->power_state);
        data->power_state = old_state;
    }
    
    return err;
}

/* ################################################################## */
/** 
 * get a power state by ioctl
 * 
 * @param data : a sort of client data
 * @param arg  : power state(user output data)
 * 
 * @return 0 : success, other : failure
 */
static int elan_get_power_state(struct elan_data *data, elan_power_state_t *arg)
{
    int err = 0;
    elan_power_state_t state;
    u8  rx_buf[ELAN_READ_PACKET_SIZE];

    memset(rx_buf, 0, sizeof(rx_buf));

    err = elan_reg_read(data, ELAN_REG_NUMBER_READ_POWER_STATE , rx_buf);

    if(err){
        printk(KERN_ERR "%s(): failed to get data packets(%d).\n",__FUNCTION__, err);
        return -EBUSY;
    }
    
    if(rx_buf[0] == ELAN_PACKET_RESPONSE_READ_REGISTER){
        state = (rx_buf[1] & 0x08) >> 3;
    }
    else{
        printk(KERN_ERR "%s() : improper response packet header(0x%02x).\n",
               __FUNCTION__, rx_buf[0]);
        return -EINVAL;
    }
    
    data->power_state = state;

    err = put_user(state, (elan_power_state_t __user *)arg);

    return err;
}

/* ################################################################## */
/** 
 * check function of finger level threshold 
 * 
 * @param arg : finger level data structure
 * 
 * @return 0 : success, other : failure
 */
static int elan_check_finger_level(elan_finger_level_t *arg)
{
    int err = 0;

    /* minimum value check */
    if( (arg->min < ELAN_FINGER_LEVEL_MIN) || (arg->min > ELAN_FINGER_LEVEL_MAX) ){
        printk(KERN_ERR "%s() : Minimum finger level threshold is out of range(%d).\n",
               __FUNCTION__, arg->min);
        err = -EINVAL;
    }
    /* maximum value check */
    if( (arg->max < ELAN_FINGER_LEVEL_MIN) || (arg->max > ELAN_FINGER_LEVEL_MAX) ){
        printk(KERN_ERR "%s() : Maximum finger level threshold is out of range(%d).\n",
               __FUNCTION__, arg->max);
        err = -EINVAL;
    }
    /* magnitude correlation check */
    if( (arg->min > arg->max) ){
        printk(KERN_ERR "%s() : Magnitude correlation of finger level is incorrect(%d>%d).\n",
               __FUNCTION__, arg->min, arg->max);
        err = -EINVAL;
    }
    
    return err;
}

/* ################################################################## */
/** 
 * acquisition function of finger level threshold by ioctl
 * 
 * @param data : a sort of client data
 * @param arg  : finger level threshold data structure
 * 
 * @return 0 : success, other : failure
 */
static int elan_get_fl_threshold(struct elan_data *data, elan_finger_level_t *arg)
{
    int err = 0;

    if(copy_to_user(arg, &data->fl_threshold, sizeof(elan_finger_level_t))){
        return -EFAULT;
    }
    
    return err;

}

/* ################################################################## */
/** 
 * setup function of finger level threshold by ioctl
 * 
 * @param data : a sort of client data
 * @param arg  : finger level threshold data structure
 * 
 * @return 0 : success, other : failure
 */
static int elan_set_fl_threshold(struct elan_data *data, elan_finger_level_t *arg)
{
    int err = 0;
    elan_finger_level_t user_data;
    
    if(copy_from_user(&user_data, arg, sizeof(elan_finger_level_t))){
            return -EFAULT;
    }

    err = elan_check_finger_level(&user_data);
    if( err ){
        printk(KERN_ERR "%s() : Finger level threshold setting is failure(%d,%d).\n"
               ,__FUNCTION__, user_data.min, user_data.max);
        return err;
    }

    data->fl_threshold = user_data;

    return err;
}


/* ################################################################## */
/** 
 * check sensitivity range of X-axis, Y-axis and Button-axis
 * 
 * @param arg : sensitivity data structure 
 *
 * @return 0 : success, other : failure
 */
static int elan_check_sensitivity_range(elan_finger_sens_t *arg)
{
    int err = 0;

    if(( arg->x < ELAN_SENSITIVITY_RANGE_MIN ) || (arg->x > ELAN_SENSITIVITY_RANGE_MAX)){
        printk(KERN_ERR "%s() : X-axis sensitivity is out of range(%d).\n",
               __FUNCTION__, arg->x);
        err = -EINVAL;
    }
    if(( arg->y < ELAN_SENSITIVITY_RANGE_MIN ) || (arg->y > ELAN_SENSITIVITY_RANGE_MAX)){
        printk(KERN_ERR "%s() : Y-axis sensitivity is out of range(%d).\n",
               __FUNCTION__, arg->y);
        err = -EINVAL;
    }
    if(( arg->button < ELAN_SENSITIVITY_RANGE_MIN ) || (arg->button > ELAN_SENSITIVITY_RANGE_MAX)){
        printk(KERN_ERR "%s() : Button-axis sensitivity is out of range(%d).\n",
               __FUNCTION__, arg->button);
        err = -EINVAL;
    }
    
    return err;
}

/* ################################################################## */
/** 
 * acquisition function of sensitivity by ioctl
 * 
 * @param data : a sort of client data
 * @param arg  : sensitivity data structure
 * 
 * @return 0 : success, other : failure
 */
static int elan_get_sensitivity(struct elan_data *data, elan_finger_sens_t *arg)
{
    int err = 0;

    err = elan_read_sensitivity(data);
    if(err){
        return err;
    }
    
    if(copy_to_user(arg, &data->sensitivity, sizeof(elan_finger_sens_t))){
        return -EFAULT;
    }
    
    return err;
}

//######################################################################
/** 
 * read sensitivity from register
 * 
 * @param data : a sort of client data
 * 
 * @return 0 : success, other : failure
 */
static int elan_read_sensitivity(struct elan_data *data)
{
    int err = 0;
    u8  rx_buf[ELAN_READ_PACKET_SIZE];

    memset(rx_buf, 0, sizeof(rx_buf));

    err = elan_reg_read(data, ELAN_REG_NUMBER_READ_SENSITIVITY_STATE, rx_buf);
    if(err){
        printk(KERN_ERR "%s(): failed to get data packets(%d).\n",__FUNCTION__, err);
        return -EBUSY;
    }
    
    if(rx_buf[0] == ELAN_PACKET_RESPONSE_READ_REGISTER){
        /* for 4.8" t/p */
        data->sensitivity.x      = (rx_buf[2] & 0xF0) >> 4;
        data->sensitivity.y      = (rx_buf[2] & 0x0F);
        data->sensitivity.button = (rx_buf[3] & 0xF0) >> 4;
    }
    else{
         printk(KERN_ERR "%s() : improper response packet header(0x%02x).\n",__FUNCTION__, rx_buf[0]);
        return -EINVAL;
    }
    
    return err;
}

/* ################################################################## */
/** 
 * setup function of sensitivity by ioctl
 * 
 * @param data : a sort of client data
 * @param arg  : sensitivity data structure
 * 
 * @return 0 : success, other : failure
 */
static int elan_set_sensitivity(struct elan_data *data, elan_finger_sens_t *arg)
{
    int err = 0;
    elan_finger_sens_t user_data;
    
    if(copy_from_user(&user_data, arg, sizeof(elan_finger_sens_t))){
            return -EFAULT;
    }

    err = elan_check_sensitivity_range(&user_data);
    if( err ){
        printk(KERN_ERR "%s() : Sensitivity setting ... failure(0x%x,0x%x,0x%x).\n"
               ,__FUNCTION__, user_data.x, user_data.y, user_data.button);
        return err;
    }

    data->sensitivity = user_data;
    
    err = elan_write_sensitivity(data);

    return err;
}

/* ################################################################## */
/** 
 * the ioctl method
 * 
 * @param inode : inode pointer
 * @param filp  : file pointer
 * @param cmd   : ioctl commands
 * @param arg   : ioctl argument
 * 
 * @return 0 : success, other : failure
 */
static int elan_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    struct elan_data *elan = (struct elan_data *)filp->private_data;

    int err = 0;

    switch( cmd ){
    case XYZ5_ELAN_SET_SENSITIVITY:
        err = elan_set_sensitivity(elan, (elan_finger_sens_t __user *)arg);
        break;
    case XYZ5_ELAN_GET_SENSITIVITY:
        err = elan_get_sensitivity(elan, (elan_finger_sens_t __user *)arg);
        break;
    case XYZ5_ELAN_SET_FINGER_LEVEL_THRESHOLD:
        err = elan_set_fl_threshold(elan, (elan_finger_level_t __user *)arg);
        break;
    case XYZ5_ELAN_GET_FINGER_LEVEL_THRESHOLD:
        err = elan_get_fl_threshold(elan, (elan_finger_level_t __user *)arg);
        break;
    case XYZ5_ELAN_GET_FINGER_LEVEL:
        err = put_user(elan->finger_level, (int __user *)arg);
        break;
    case XYZ5_ELAN_SET_POWER_STATE:
        err = elan_set_power_state(elan, (elan_power_state_t __user *)arg);
        break;
    case XYZ5_ELAN_GET_POWER_STATE:
        err = elan_get_power_state(elan, (elan_power_state_t __user *)arg);
        break;
    case XYZ5_ELAN_GET_FW_INFO:
        err = elan_get_fw_info(elan, (elan_fw_info_t __user *)arg);
        break;
    case XYZ5_ELAN_GET_COORDINATE_TYPE:
        err = elan_get_coordinate_type(elan, (elan_coordinate_t __user *)arg);
        break;
    case XYZ5_ELAN_SET_COORDINATE_TYPE:
        err = elan_set_coordinate_type(elan, (elan_coordinate_t __user *)arg);
        break;
    case XYZ5_ELAN_GET_Y_AXIS_REPORT_DIRECTION:
        err = put_user(elan->y_axis_inversion, (elan_y_direction_t __user *)arg);
        break;
    case XYZ5_ELAN_SET_Y_AXIS_REPORT_DIRECTION:
        err = get_user(elan->y_axis_inversion, (elan_y_direction_t __user *)arg);
        break;
    default:
        err = -EINVAL;
        break;
    }
    
    return err;
}

/* ################################################################## */
/** 
 * the open method
 * 
 * @param inode : inode pointer
 * @param filp  : file pointer
 * 
 * @return 0 : success, other : failure
 */
static int elan_open(struct inode *inode, struct file *filp)
{
    struct elan_data *data;

    data = container_of(inode->i_cdev, struct elan_data, cdev);
    filp->private_data = data;

    if( data->used ){
        printk(KERN_ERR "%s : ELAN using.\n",__FUNCTION__);
        return -EBUSY;
    }

    if( down_interruptible(&data->sem) ){
        printk(KERN_ERR "%s : down_interruptible() is failed.\n",__FUNCTION__);
        return -ERESTARTSYS;
    }

    data->used = 1;
    up( &data->sem );
    
    return 0;
}

/* ################################################################## */
/** 
 * the release method
 * 
 * @param inode : inode pointer
 * @param filp  : file pointer
 * 
 * @return 0 : success, other : failure 
 */
static int elan_release(struct inode *inode, struct file *filp)
{
    struct elan_data *data;

    data = container_of(inode->i_cdev, struct elan_data, cdev);

    if( down_interruptible( &data->sem ) ){
        printk(KERN_ERR "%s : down_interruptible() is failed.\n",__FUNCTION__);
        return -ERESTARTSYS;
    }

    data->used = 0;
    up( &data->sem );
    
    return 0;
}


static struct file_operations elan_fops = {
    .owner   = THIS_MODULE,
    .ioctl   = elan_ioctl,
    .open    = elan_open,
    .release = elan_release,
};

//######################################################################
/** 
 * send i2c data stream function  
 * 
 * @param p             : i2c client  
 * @param stream        : sending stream 
 * @param stream_length : stream length
 * 
 * @return number of "i2c_msg" structure
 */
static int elan_send_packet(struct elan_data * p, u8 * stream, int stream_length)
{
    struct i2c_client * client = elan_get_client(p);
    struct i2c_msg message;
    
    message.addr   = client->addr;
    message.len    = stream_length;
    message.buf    = stream;
    message.flags  = 0;
    //message[0].flags &= ~I2C_M_RD;

    return i2c_transfer(client->adapter, &message, 1);
}

//######################################################################
/** 
 * register read function
 * 
 * @param data      : a sort of client data
 * @param reg_addr  : register address
 * @param read_data : read data 
 * 
 * @return ret      : number of "i2c_msg" structure 
 */
static int elan_reg_read(struct elan_data *data, u8 reg_addr, u8 *read_data)
{
    int ret;
    u8  write_data[ELAN_READ_PACKET_SIZE];
    
    if( down_interruptible(&data->rdsem) ){
        printk(KERN_ERR "%s : down_interruptible() is failed.\n",__FUNCTION__);
        return -ERESTARTSYS;
    }
    
    write_data[0] = ELAN_PACKET_READ_REGISTER;
    write_data[1] = (reg_addr << 4) | 0x00;
    write_data[2] = 0x00;
    write_data[3] = 0x01;
    
    ret = elan_send_packet(data, write_data, sizeof(write_data));
    
    if( ret != 1 ){
        up( &data->rdsem );
        return -1;
    }
    
    /* sleep until interrupt happens */
    wait_event_interruptible(data->rq, data->rqflag != 0);
    data->rqflag = 0;
    
    memcpy(read_data, data->reg_data, ELAN_RESPONSE_PACKET_SIZE);

    up( &data->rdsem );
    
    return 0;
}

//######################################################################
/** 
 * register write function
 * 
 * @param data       : a sort of client data
 * @param reg_addr   : register address
 * @param write_data : writing data
 * 
 * @return ret       : number of "i2c_msg" structure 
 */
static int elan_reg_write(struct elan_data *data, u8 reg_addr, u8 *write_data)
{
    int            ret;
    u8             send_packet[ELAN_WRITE_PACKET_SIZE];

    memset(send_packet, 0, sizeof(send_packet));
    
    send_packet[0] = ELAN_PACKET_WRITE_REGISTER;
    send_packet[1] = (( reg_addr & 0x0F ) << 4 ) | ((write_data[0] & 0xF0) >> 4);
    send_packet[2] = ((write_data[0] & 0x0F) << 4) | ((write_data[1] & 0xF0) >> 4);
    send_packet[3] = ((write_data[1] & 0x0F) << 4) | 0x01;
    
    ret = elan_send_packet(data, send_packet, sizeof(send_packet));

    return (ret == 1) ? 0 : -1;
}

//######################################################################
/** 
 * receive data packets and dispatch packet data process.
 * 
 * @param data :  
 * 
 * @return 0 : report to linux input subsystem
 *         1 : driver dispatch packets correctly
 *        -1 : error 
 */
static int elan_receive_packet(struct elan_data *data)
{
    int ret, i;
    int receive_size = ELAN_INDEX_COORDINATE_PACKET_SIZE;
    u8  rx_buf[ELAN_INDEX_COORDINATE_PACKET_SIZE];

    u16 x1_high, x1_low, y1_high, y1_low;
    u16 x2_high, x2_low, y2_high, y2_low;
    
    struct i2c_client *client;

    client = elan_get_client(data);
    
    memset(rx_buf, 0, sizeof(rx_buf));
    x1_high = x1_low = y1_high = y1_low = 0;
    x2_high = x2_low = y2_high = y2_low = 0;
    
    ret = i2c_master_recv(client, rx_buf, receive_size);
    /* 4.8" change point */
    switch(rx_buf[0]){
    case ELAN_PACKET_REPORT_INDEX_COORDINATE:
    {
        if( (ret == receive_size) && (rx_buf[0] == ELAN_PACKET_REPORT_INDEX_COORDINATE) ){
            data->status       = (rx_buf[7] & 0x06) >> 1;
            /*
            data->x_position   = ((rx_buf[1] & 0xF0) << 4) | rx_buf[2];
            data->y_position   = ((rx_buf[1] & 0x0F) << 8) | rx_buf[3];
            data->x2_position  = ((rx_buf[4] & 0xF0) << 4) | rx_buf[5];
            data->y2_position  = ((rx_buf[4] & 0x0F) << 8) | rx_buf[6];
            */
            x1_high = (u16)((rx_buf[1] & 0xF0) >> 4);
            x1_low  = (u16)rx_buf[2];
            y1_high = (u16)(rx_buf[1] & 0x0F);
            y1_low  = (u16)rx_buf[3];

            data->x_position   = (x1_high << 8) | x1_low;
            data->y_position   = (y1_high << 8) | y1_low;
            
            x2_high = (u16)((rx_buf[4] & 0xF0) >> 4);
            x2_low  = (u16)rx_buf[5];
            y2_high = (u16)(rx_buf[4] & 0x0F);
            y2_low  = (u16)rx_buf[6];

            data->x2_position   = (x2_high << 8) | x2_low;
            data->y2_position   = (y2_high << 8) | y2_low;

            data->upper_button = (rx_buf[7] & 0x80) >> 7;
            data->lower_button = (rx_buf[7] & 0x40) >> 6;
            data->finger_level = (rx_buf[7] & 0x30) >> 4;

            if( (rx_buf[7] & 0x06) != 0 ){
                data->pressure = data->x_position;
            }
            else{
                data->pressure = 0x00;
            }
            
            ret = 0;
        }
        else{
            printk(KERN_ERR "%s() : response packet error(len=%d,0x%02x).\n",__FUNCTION__, ret, rx_buf[0]);
            ret = -1;
        }
        break;
    }
    case ELAN_PACKET_REPORT_ABSOLUTE_COORDINATE:
    {
        if( (ret == receive_size) && (rx_buf[0] == ELAN_PACKET_REPORT_ABSOLUTE_COORDINATE) ){
            data->status       = (rx_buf[5] & 0x06) >> 1;
            /*
            data->x_position   = ((rx_buf[1] & 0xF0) << 4) | rx_buf[2];
            data->y_position   = ((rx_buf[1] & 0x0F) << 8) | rx_buf[3];
            */
            x1_high = (u16)((rx_buf[1] & 0xF0) >> 4);
            x1_low  = (u16)rx_buf[2];
            y1_high = (u16)(rx_buf[1] & 0x0F);
            y1_low  = (u16)rx_buf[3];

            data->x_position   = (x1_high << 8) | x1_low;
            data->y_position   = (y1_high << 8) | y1_low;

            data->upper_button = (rx_buf[5] & 0x80) >> 7;
            data->lower_button = (rx_buf[5] & 0x40) >> 6;
            data->finger_level = (rx_buf[5] & 0x30) >> 4; /* new from 4.8" */
            
            if( (rx_buf[5] & 0x06) != 0 ){
                data->pressure = data->x_position;
            }
            else{
                data->pressure = 0x00;
            }

            elan_align_finger_up_data_at_absolute_type(data);

            ret = 0;
            break;
        }
        else{
            printk(KERN_ERR "%s() : response packet error(len=%d,0x%02x).\n",__FUNCTION__, ret, rx_buf[0]);
            ret = -1;
            break;
        }
    }
    case ELAN_PACKET_RESPONSE_READ_REGISTER:
        memcpy(data->reg_data, rx_buf, ELAN_RESPONSE_PACKET_SIZE);

        data->rqflag = 1;
        wake_up_interruptible(&data->rq);
        ret = 1;
        break;
    case ELAN_PACKET_READY_TO_WORK:
    {
        for( i = 0 ; i < 4 ; i++ ){
            if(rx_buf[i] != ELAN_PACKET_READY_TO_WORK){
                printk(KERN_ERR "%s: improper hello packet[%d](0x%02x).\n", __FUNCTION__, i, rx_buf[i]);
                ret = -1;
            }
            else{
                ret = 1;
            }
        }
        break;
    }
    case ELAN_PACKET_READ_REGISTER:
    case ELAN_PACKET_WRITE_REGISTER:
    default:
        ret = -1;
        break;
    }

    return ret;

}

//######################################################################
/** 
 * interrupt handler(bottom-half handler)
 * 
 * @param arg : client data
 * 
 * @return 0 : success, other : failure
 */
static int elan_thread(void *arg)
{
    struct elan_data  *data      = (struct elan_data *)arg;
    struct input_dev  *input_dev = data->dev;
    struct input_dev  *key_dev   = data->key_dev;
    struct i2c_client *client;
    
    int previous_upper_button = 0;
    int previous_lower_button = 0;
    int ret                   = 0;

    client = elan_get_client(data);

    set_freezable();

    daemonize("elan_tp");
    while(data->clearpad_installed) {
        try_to_freeze();

        wait_event_interruptible(data->wait_queue, data->flag != 0);
        data->flag = 0;

        if(data->clearpad_installed == 0){
            return 0;
        }

        ret = elan_receive_packet(data);

        if(!ret){
            if (elan_event_should_be_reported(data)) {
                int upper_button_is_changed = ((data->upper_button != previous_upper_button));
                int lower_button_is_changed = ((data->lower_button != previous_lower_button));

                if (upper_button_is_changed || lower_button_is_changed) {
                    if (upper_button_is_changed) {
                        // upper_button is 1bit data, only 0/1 will do.
                        input_report_key(key_dev, KEY_M, data->upper_button);
                    }
                    
                    if (lower_button_is_changed) {
                        input_report_key(key_dev, KEY_A, data->lower_button);
                    }
                    input_sync(key_dev);
                }

                data->y_position = elan_get_aligned_y_position(data);
                elan_align_finger1_vaaa_area_data(data);

                input_report_abs(input_dev, ABS_X, data->x_position);
                input_report_abs(input_dev, ABS_Y, data->y_position);
                    
                if (data->coordinate_type == ELAN_COORDINATE_TYPE_INDEX) {
                    /* if t/p is touched by multi finger */
                    if (data->status > 0x1) {
                        data->y2_position = elan_get_aligned_y2_position(data);
                        elan_align_finger2_vaaa_area_data(data);
                        
                        input_report_abs(input_dev, ABS_RX, data->x2_position);
                        input_report_abs(input_dev, ABS_RY, data->y2_position);
                    }
                    else{
                        /* single finger */
                        input_report_abs(input_dev, ABS_RX, INT_MIN);
                        input_report_abs(input_dev, ABS_RY, INT_MIN);
                    }
                }

                input_report_abs(input_dev, ABS_MISC, data->status);
                input_report_abs(input_dev, ABS_PRESSURE, data->pressure);
                // 081014-------------------->
#if FINGER_LEVEL_EVALUATION
#else
                input_sync(input_dev);
#endif // FINGER_LEVEL_EVALUATION
                // <--------------------081014
            }
            
            // 081014-------------------->
#if FINGER_LEVEL_EVALUATION
            /* for evaluation */
            input_report_abs(input_dev, ABS_DISTANCE, data->finger_level);
            input_sync(input_dev);
            /*
              input_report_rel(input_dev, REL_MISC, data->finger_level);
            */
#else
#endif // FINGER_LEVEL_EVALUATION
            // <--------------------081014

            previous_upper_button = data->upper_button;
            previous_lower_button = data->lower_button;
        }
    }
    
    return 0;
}

/* ################################################################## */
/** 
 * setup coordinate type to register
 * 
 * @param data : a sort of client data 
 * 
 * @return 0 : success, other : failure
 */
static int elan_write_coordinate_type(struct elan_data *data)
{
    u8  tx_buf[ELAN_USER_WRITABLE_REG_SIZE];
    int ret = 0;
    
    memset(tx_buf, 0, sizeof(tx_buf));
    /* 4.8" change point */
    switch (data->coordinate_type){
    case ELAN_COORDINATE_TYPE_ABSOLUTE:
        /* set ABSOLUTE coordinate type */
        tx_buf[0] = tx_buf[1] = 0x00;
        break;
    case ELAN_COORDINATE_TYPE_INDEX:
        /* set INDEX coordinate type */
        tx_buf[0] = 0x00;
        tx_buf[1] = 0x10;
        break;
    default:
        return 1;
    }
    
    ret = elan_reg_write(data, ELAN_REG_NUMBER_WRITE_COORDINATE_TYPE, tx_buf);

    return ret;
}

/* ################################################################## */
/** 
 * setup sensitivitys of X-axis, Y-axis and Button-axis
 * 
 * @param data : a sort of client data 
 * 
 * @return 0 : success, other : failure
 */
static int elan_write_sensitivity(struct elan_data *data)
{
    u8  tx_buf[ELAN_USER_WRITABLE_REG_SIZE];
    int ret = 0;

    memset(tx_buf, 0, sizeof(tx_buf));
    /* 4.8" change point */
    tx_buf[0] = data->sensitivity.x & 0x0F;
    tx_buf[1] = ( data->sensitivity.y << 4 ) | (data->sensitivity.button & 0x0F);

    ret = elan_reg_write(data, ELAN_REG_NUMBER_WRITE_SENSITIVITY_STATE, tx_buf);

    return ret;
}

/* ################################################################## */
/** 
 * write a power state to device register via i2c
 * 
 * @param data : asort of client data 
 * 
 * @return 0 : success, other : failure
 */
static int elan_write_power_state(struct elan_data *data)
{
    u8  tx_buf[ELAN_USER_WRITABLE_REG_SIZE];    
    int err = 0;
    
    memset(tx_buf, 0, sizeof(tx_buf));
    
    if( (data->power_state < ELAN_POWER_STATE_DEEP_SLEEP) || (data->power_state > ELAN_POWER_STATE_NORMAL) ){
        printk(KERN_WARNING "%s() : undefined power state(%d).\n",__FUNCTION__, data->power_state);
        return -EINVAL;
    }
    else{
        tx_buf[0] = (data->power_state) << 7;
        tx_buf[1] = 0x00;
    }
    
    err = elan_reg_write(data, ELAN_REG_NUMBER_WRITE_POWER_STATE, tx_buf);

    return err;
}

//######################################################################
/** 
 * initialize function of device data structure
 * 
 * @param data : a sort of client data
 */
static void elan_initialize_data(struct elan_data *data)
{
    data->status             = 0;
    data->x_position         = 0;
    data->y_position         = 0;
    data->x2_position        = 0;
    data->y2_position        = 0;
    data->pressure           = 0;
    data->upper_button       = 0;
    data->lower_button       = 0;
    data->finger_level       = 0;

    data->fl_threshold.min   = ELAN_FINGER_LEVEL_MIN;
    data->fl_threshold.max   = ELAN_FINGER_LEVEL_MAX;

    data->sensitivity.x      = 0x5;
    data->sensitivity.y      = 0x5;
    data->sensitivity.button = 0x5;

    data->coordinate_type    = ELAN_COORDINATE_TYPE_INDEX;

    data->power_state        = ELAN_POWER_STATE_NORMAL;

    data->used               = 0;

    data->clearpad_installed = 0;

    data->irq_type           = IRQF_TRIGGER_LOW;

    data->fw_info.major_ver  = 0;
    data->fw_info.minor_ver  = 0;
    data->fw_info.id_high    = 0;
    data->fw_info.id_low     = 0;

    data->y_axis_inversion   = ELAN_Y_AXIS_REPORT_DEVICE_BACKWARD_DIRECTION;

    return ;
}

//######################################################################
/** 
 * setup a device
 * 
 * @param data  : a sort of client data 
 * 
 * @return  0 : success, other : failure
 */
static int elan_setup(struct elan_data *data)
{
    int err = 0;

    /* 4.8" change point */
    /* aquire a sensitivity */
    /*
    ret = elan_read_sensitivity(data);
    if(ret){
        printk(KERN_ERR "%s : failed to get \"Sensitivity\".\n",__FUNCTION__);
        err = -1;
    }
    */
    
    /* coordinate type init */
    data->coordinate_type = ELAN_COORDINATE_TYPE_INDEX;
    /*
    ret = elan_write_coordinate_type(data);
    if(ret){
        printk(KERN_ERR "%s() : failed to set up \"Coordinate Type\"\n",__FUNCTION__);
        err = -1;
    }
    */
    
    /* power state init */
    data->power_state = ELAN_POWER_STATE_NORMAL;
    /*
    ret = elan_write_power_state(data);
    if(ret){
        printk(KERN_ERR "%s() : failed to set up \"Power State\"\n",__FUNCTION__);
        err = ret;
    }
    */
    return err;
}

/* ################################################################## */
/** 
 * check a device connection
 * 
 * @param data : a sort of client data 
 * 
 * @return 0 : success, other : failure
 */
static int elan_check_connection(struct elan_data *data)
{
    int err = 0;
    int ret;

    ret = elan_read_fw_info(data);
    if(ret){
        err = -1;
        return err;
    }
    else{
        printk(KERN_INFO "elan_tp : f/w ver=%d.%d, id=(0x%x,0x%x)\n",
               data->fw_info.major_ver, data->fw_info.minor_ver,
               data->fw_info.id_high, data->fw_info.id_low);
    }

    return err;
}

/* ################################################################## */
/** 
 * character device setup function
 * 
 * @param data  : device structure
 * @param index : device index
 */
static void elan_setup_cdev(struct elan_data *data, int index)
{
    int err = 0;
    int devno = MKDEV(elan_major, elan_minor + index);

    cdev_init(&data->cdev, &elan_fops);

    data->cdev.owner = THIS_MODULE;
    data->cdev.ops   = &elan_fops;

    err = cdev_add(&data->cdev, devno, 1);

    if(err)
        printk(KERN_NOTICE "%s : Error %d adding elan-%d",__FUNCTION__, err, index);

    return ;
}

//######################################################################
/** 
 *  interrupt handler(top-half handler)
 * 
 * @param irq    : IRQ number
 * @param dev_id : a sort of client data
 * @param regs   : snapshot of the processor's context before the processor
 *                 entered interrupt code.
 * 
 * @return       : IRQ_HANDLED
 */
static irqreturn_t elan_interrupt(int irq, void *dev_id)
{
    struct i2c_client *client = dev_id;
    struct elan_data  *data   = i2c_get_clientdata(client);

    irqreturn_t ret = IRQ_HANDLED;
    int         err = -1;
    int         val = 0;

    disable_irq(irq);

    val = gpio_get_value(PGPF_GPIO_TP_XINT);
    rmb();
    
    if(val == 0){
        data->irq_type = IRQF_TRIGGER_RISING;
        err = set_irq_type(client->irq, data->irq_type);
        
// INTERRUPT_DETECTION_ASYNCHRONOUS_090407-------------------->
#if INTERRUPT_DETECTION_ASYNCHRONOUS_090407
        err = elan_set_irq_detection(data);
#endif // INTERRUPT_DETECTION_ASYNCHRONOUS_090407
// <--------------------INTERRUPT_DETECTION_ASYNCHRONOUS_090407        

        /* passed process to bottom-half processing */
        data->flag = 1;
        wake_up_interruptible(&data->wait_queue);
    }
    else if(val == 1){
        data->irq_type =  IRQF_TRIGGER_LOW;
        err = set_irq_type(client->irq, data->irq_type);
        // INTERRUPT_DETECTION_ASYNCHRONOUS_090407-------------------->
#if INTERRUPT_DETECTION_ASYNCHRONOUS_090407
        err = elan_set_irq_detection(data);
#endif // INTERRUPT_DETECTION_ASYNCHRONOUS_090407
// <--------------------INTERRUPT_DETECTION_ASYNCHRONOUS_090407
    }
    else{
        ret = IRQ_NONE;
    }

    enable_irq(irq);

    return ret;
}

/* ################################################################## */
/** 
 * probe function for "legacy style" i2c drivers.
 * 
 * @param adapter :  i2c adapter
 * @param address :  i2c slave address
 * @param kind    :  0 ... forced detection, -1 ... probed detection
 * 
 * @return 0 : success, other : failure
 */
static int elan_probe(struct i2c_adapter *adapter, int address, int kind)
{
    int err = 0;
    int i;
    dev_t devno = MKDEV(elan_major, elan_minor);

    struct i2c_client *client;
    struct input_dev  *input_dev;
    struct input_dev  *key_dev;

    struct elan_data  *data;

    if(!i2c_check_functionality(adapter, I2C_FUNC_I2C)){
        err = -EIO;
        goto exit;
    }
    if(!(data = kzalloc(sizeof(struct elan_data), GFP_KERNEL))){
        err = -ENOMEM;
        goto exit;
    }

    elan_initialize_data(data);

    client = &data->client;
    i2c_set_clientdata(client, data);

    client->addr    = address;
    client->adapter = adapter;
    client->driver  = &elan_i2c_client_driver;
    client->flags   = 0;
    client->irq     = gpio_to_irq(PGPF_GPIO_TP_XINT);

    strlcpy(client->name, ELAN_DRIVER_NAME, I2C_NAME_SIZE);

    if((err = i2c_attach_client(client))){
        goto exit_kfree;
    }

    /* wait queue init */
    init_waitqueue_head(&data->wait_queue);
    data->flag = 0;

    init_waitqueue_head(&data->rq);
    data->rqflag = 0;

    /* semaphore init */
    init_MUTEX(&data->sem);
    init_MUTEX(&data->rdsem);

    /*
     * Input subsystem part (touchpad parts)
     * 
     */
    input_dev = input_allocate_device();
    if(!input_dev){
        printk(KERN_ERR
               "elan_tp : not enough memory for input device.\n");
        err = -ENOMEM;
        goto exit_i2c_detach_client;
    }

    data->dev = input_dev;
    
    input_dev->name = "elan_tp";

    /*
      input_dev->evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
      input_dev->keybit[LONG(BTN_TOUCH)] |= BIT(BTN_TOUCH);
    */
    set_bit(EV_ABS, input_dev->evbit);
    /* set_bit(EV_KEY, input_dev->evbit); */
    set_bit(ABS_X, input_dev->absbit);
    set_bit(ABS_Y, input_dev->absbit);
    set_bit(ABS_PRESSURE, input_dev->absbit);
    set_bit(ABS_RX, input_dev->absbit);
    set_bit(ABS_RY, input_dev->absbit);
    set_bit(ABS_MISC, input_dev->absbit);

    // 081014-------------------->
#if FINGER_LEVEL_EVALUATION
    /*
      set_bit(EV_REL, input_dev->evbit);
      set_bit(REL_MISC, input_dev->relbit);
    */
    set_bit(ABS_DISTANCE, input_dev->absbit);
#else    
#endif // FINGER_LEVEL_EVALUATION
    // <--------------------081014

    input_set_abs_params(input_dev, ABS_X, INT_MIN, INT_MAX, 0, 0);
    input_set_abs_params(input_dev, ABS_Y, INT_MIN, INT_MAX, 0, 0);
    input_set_abs_params(input_dev, ABS_RX, INT_MIN, INT_MAX, 0, 0);
    input_set_abs_params(input_dev, ABS_RY, INT_MIN, INT_MAX, 0, 0);
    input_set_abs_params(input_dev, ABS_MISC, INT_MIN, INT_MAX, 0, 0);
    input_set_abs_params(input_dev, ABS_PRESSURE, INT_MIN, INT_MAX, 0, 0);

    // 081014-------------------->
#if FINGER_LEVEL_EVALUATION
    input_set_abs_params(input_dev, ABS_DISTANCE, 0, 5, 0, 0);
#else
#endif // FINGER_LEVEL_EVALUATION
    // <--------------------081014
 
    err = input_register_device(data->dev);

    if(err){
        printk(KERN_ERR "elan_tp : Failed to register device\n");
        input_free_device(data->dev);
        goto exit_i2c_detach_client;
    }

    /*
     * Input subsystem part (key parts)
     * 
     */
    key_dev = input_allocate_device();
    if(!key_dev){
        printk(KERN_ERR
               "elan_tp: not enough memory for key device.\n");
        err = -ENOMEM;
        goto err_free_key_dev;
    }

    data->key_dev = key_dev;
    
    key_dev->name = "elan_key";
    /*
      key_dev->evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
      key_dev->keybit[LONG(BTN_TOP)] = BIT(BTN_TOP) | BIT(BTN_BASE);
    */
    set_bit(EV_KEY, key_dev->evbit);
    set_bit(KEY_M, key_dev->keybit); /* 4.8" change point */
    set_bit(KEY_A, key_dev->keybit); /* 4.8" change point */
    
    err = input_register_device(data->key_dev);
    
    if(err){
        printk(KERN_ERR "elan_tp : Failed to register key device\n");
        input_free_device(data->key_dev);
        goto err_free_key_dev;        
    }

    /* cdev init */
    for( i = 0 ; i < elan_nr_devs ; i++ ){
        elan_setup_cdev(data, i);
    }
    
    /*
     * kernel thread init
     * 
     */
    data->clearpad_installed = 1;
    kernel_thread(elan_thread, data, CLONE_VM | CLONE_FS);

    /*
     * set interrupt line behavior
     * 
     */
    err = set_irq_type(client->irq, IRQF_TRIGGER_LOW);
    if(err){
        printk(KERN_ERR "elan_tp : Failed to set interrupt type\n");
        goto err_free_irq;
    }
    
    if(request_irq(client->irq, elan_interrupt, 0, "elan_tp", &data->client)){
        printk(KERN_ERR "elan_tp : Can't allocate irq\n");
        err = -EBUSY;
        goto err_free_irq;
    }
// INTERRUPT_DETECTION_ASYNCHRONOUS_090407-------------------->
#if INTERRUPT_DETECTION_ASYNCHRONOUS_090407
    err = elan_set_irq_detection(data);
#endif // INTERRUPT_DETECTION_ASYNCHRONOUS_090407
// <--------------------INTERRUPT_DETECTION_ASYNCHRONOUS_090407
    err = elan_check_connection(data);
    if( err ){
        err = -ENODEV;
        goto err_free_irq;
    }
    
    /* setup device register */
    err = elan_setup(data);
    if( err ){
        printk(KERN_ERR "elan_tp : Failed to initialize device registers.\n");
        goto err_free_irq;
    }

    return 0;

  err_free_irq:
    data->clearpad_installed = 0;
    data->coordinate_type    = -1;
    data->flag               = 1;
    wake_up_interruptible(&data->wait_queue);
    free_irq(client->irq, &data->client);
    cdev_del(&data->cdev);
    input_unregister_device(data->key_dev);
  err_free_key_dev:
    input_unregister_device(data->dev);
  exit_i2c_detach_client:
    i2c_detach_client(client);
  exit_kfree:
    kfree(data);
  exit:
    unregister_chrdev_region(devno, elan_nr_devs);
    elan_change_resetline(0);
    return err;
}

//######################################################################
/** 
 *  wait for device ready status by polling interrupt line
 *  from "Hi" to "Low".
 *
 *  This method is useful wether t/p is ready or not
 *  when interrupt hander is not registered. 
 */
static void elan_wait_for_device_is_ready(int forced)
{
    unsigned int sleep       = 50;  /* unit : ms */
    unsigned int timeout     = 650; /* unit : ms */
    unsigned int count       = 0;

    if(forced){
        msleep(timeout);
        return ;
    }
    
    while(gpio_get_value(PGPF_GPIO_TP_XINT)){
        msleep(sleep);
        count++;

        if( (sleep*count) >= timeout ){
            return;
        }
    }

    return;
}

//######################################################################
/** 
 * assert/negate a device reset line
 * 
 * @param value 0 : reset assert, 1 : reset negate
 */
static void elan_change_resetline(int value)
{
    gpio_direction_output(PGPF_GPIO_TP_XRST, value);
    gpio_set_value(PGPF_GPIO_TP_XRST, value);

    return ;
}

//######################################################################
/** 
 * suspend function
 * 
 * @param client : i2c client data
 * @param msg    : power state message 
 * 
 * @return 0 : success, other : failure
 */
static int elan_suspend(struct i2c_client *client, pm_message_t msg)
{
    int err = 0;

    elan_change_resetline(0);

    return err;
}
//######################################################################
/** 
 * resume function
 * 
 * @param client : i2c client data
 * 
 * @return 0 : success, other : failure
 */
static int elan_resume(struct i2c_client *client)
{
    int err = 0;

    elan_change_resetline(1);
    
    return err;
}

static int __init elan_attach_adapter(struct i2c_adapter *adapter)
{
    return i2c_probe(adapter, &addr_data, elan_probe);
}

static int elan_detach_client(struct i2c_client *client)
{
    struct elan_data *data      = i2c_get_clientdata(client);
    struct input_dev *input_dev = data->dev;
    struct input_dev *key_dev   = data->key_dev;

    int   err;

    free_irq(client->irq, &data->client);

    /* eliminate kernel thread */
    data->clearpad_installed = 0;
    data->coordinate_type = -1;
    data->flag = 1;
    wake_up_interruptible(&data->wait_queue);
    
    cdev_del(&data->cdev);
    
    input_unregister_device(key_dev);
    input_unregister_device(input_dev);

    if((err = i2c_detach_client(client))){
        return err;
    }
    kfree(data);
    return 0;
}

/* ################################################################## */
/** 
 * 
 * function for registering as character device
 * 
 * @return errno
 */
static int elan_get_devnum(void)
{
    int   res = 0;
    dev_t dev = 0;

    /*
     * Get a range of minor numbers to work with, asking for a dynamic
     * major unless directed otherwise at load time.
     */
    if(elan_major){
        dev = MKDEV(elan_major, elan_minor);
        res = register_chrdev_region(dev, elan_nr_devs, "elan_tp");
    }
    else{
        res = alloc_chrdev_region(&dev, elan_minor, elan_nr_devs, "elan_tp");
        elan_major = MAJOR(dev);
    }
    if( res < 0 ){
        printk(KERN_WARNING "%s : Can't get major %d\n", __FUNCTION__, elan_major);
        return res;
    }

    return res;
}


static struct i2c_driver  elan_i2c_client_driver = {
    .driver = {
        .owner = THIS_MODULE,
        .name  = "elan_tp"
    },
    .attach_adapter = elan_attach_adapter,
    .detach_client  = __exit_p(elan_detach_client),
    .suspend        = elan_suspend,
    .resume         = elan_resume,
};

//######################################################################
/** 
 *  initialization function
 * 
 * 
 * @return 
 */
static int __init elan_init(void)
{
    int   res;
    int   val    = 0;
    int   forced = 0;
    dev_t devno  = MKDEV(elan_major, elan_minor);

    printk(KERN_INFO "%s : version %s\n",
           ELAN_DRIVER_NAME, ELAN_DRIVER_VERSION);


    val = gpio_get_value(PGPF_GPIO_TP_XRST);

    if( val == 0 ){
        elan_change_resetline(1);
        elan_wait_for_device_is_ready(forced);
    }
    
    if( (res = elan_get_devnum()) < 0 ){
        return res;
    }
    
    if((res = i2c_add_driver(&elan_i2c_client_driver))){
        printk("%s : Driver registration failed, module not inserted.\n",__FUNCTION__);
        unregister_chrdev_region(devno, elan_nr_devs);
        elan_change_resetline(0);
        return res;
    }
    return 0;
}

//######################################################################
/** 
 * clearnup function
 * 
 * 
 * @return 
 */
static void __exit elan_exit(void)
{
    dev_t devno = MKDEV(elan_major, elan_minor);

    i2c_del_driver(&elan_i2c_client_driver);

    unregister_chrdev_region(devno, elan_nr_devs);
    elan_change_resetline(0);
}

module_init(elan_init);
module_exit(elan_exit);


MODULE_DESCRIPTION("ELAN ClearPad driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL");
MODULE_VERSION(ELAN_DRIVER_VERSION);
