/* 2010-08-25: File added by Sony Corporation */
/*
 * linux/drivers/misc/twl4030_madc.c
 *
 * TWL4030 MADC driver for Linux
 *
 * Copyright (C) 2008 Texas Instruments, Inc.
 * Author: Texas Instruments, Inc.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>

#include <asm/uaccess.h>

#include <asm/arch/twl4030.h>
#include <asm/arch/twl4030-madc.h>

/**
 * Register definitions
 */
#define REG_CTRL1			0x00
#define REG_SW1SELECT_LSB		0x06
#define REG_SW1SELECT_MSB		0x07
#define REG_SW1AVERAGE_LSB		0x08
#define REG_SW1AVERAGE_MSB		0x09
#define REG_CTRL_SW1			0x12
#define REG_GPCH0_LSB			0x37
#define REG_GPCH0_MSB			0x38
#define REG_MADC_ISR1			0x61
#define	REG_MADC_IMR1			0X62

#define REG_BCICTL1			0x023

/**
 * Bit Field defintions
 */
#define CTRL1_MADC_ON		0x01
#define MADC_IMR1_SW1IMR1	0x02
#define MADC_ISR1_SW1ISR1	0x02
#define CTRL_SW1_SW1		0x20


#define BCICTL1_MESVBUS		0x04
#define BCICTL1_MESBAT		0x02
#define BCICTL1_MESVAC		0x01

u8 madc_prescaler_ratio[16] = {
	100,	/* ADCIN0 prescaler value*/
	100,	/* ADCIN1 prescaler value*/
	60,	/* ADCIN2 prescaler value*/
	60,	/* ADCIN3 prescaler value*/
	60,	/* ADCIN4 prescaler value*/
	60,	/* ADCIN5 prescaler value*/
	60,	/* ADCIN6 prescaler value*/
	60,	/* ADCIN7 prescaler value*/
	21,	/* ADCIN8 prescaler value*/
	33,	/* ADCIN9 prescaler value*/
	100,	/* ADCIN10 prescaler value*/
	15,	/* ADCIN11 prescaler value*/
	25,	/* ADCIN12 prescaler value*/
	100,	/* ADCIN13 prescaler value.ADCIN13 is reserved internal pin
		 * prescaler value is put here so that the array is proper.
		 */
	100,	/* ADCIN14 prescaler value.ADCIN14 is reserved internal pin
		 * prescaler value is put here so that the array is proper.
		 */
	45	/* ADCIN15 prescaler value*/
};

#define CONV_VOLTAGE(value, channel)	\
	((value * 150000) / (madc_prescaler_ratio[channel] * 1023));

/*
 * Minor Number to register the driver with
 * misc framework
 */
#define MADC_MINOR		133

static struct completion sw1conv_complete;
static int swconv1_ongoing;
struct semaphore sw1conv_lock;
static struct device *dbg_dev;

static int twl4030madc_probe(struct platform_device *pdev);
static int twl4030madc_remove(struct  platform_device *dev);

#ifdef CONFIG_PM
static int twl4030madc_suspend(struct platform_device *dev, pm_message_t state);
static int twl4030madc_resume(struct platform_device *dev);
#endif

/* Driver Information */
static struct platform_driver twl4030madc_driver = {
	.probe = twl4030madc_probe,
	.remove = twl4030madc_remove,
#ifdef CONFIG_PM
	.suspend = twl4030madc_suspend,
	.resume = twl4030madc_resume,
#endif
	.driver = {
		.name = "twl4030_madc",
	},
};

static int clear_n_set(u8 mod_no, u8 clear, u8 set, u8 reg)
{
	int ret;
	u8 data = 0;

	/* Gets the initial register value */
	ret = twl4030_i2c_read_u8(mod_no, &data, reg);
	if (ret)
		return ret;
	/* Clearing all those bits to clear */
	data &= ~(clear);
	/* Setting all those bits to set */
	data |= set;
	ret = twl4030_i2c_write_u8(mod_no, data, reg);
	/* Update the register */
	if (ret)
		return ret;
	return 0;
}
/**
 * Function : twl4030madc_setup
 * Description :
 * Unmasks the interrupt,sets up prescalers for various channels
 * and turns on the module
 */
static int twl4030madc_setup(void)
{
	int ret;
	u8 data;

	dev_dbg(dbg_dev, "twl4030madc_setup+");
	/* Clearing the ISR regster in case of any previous interrupt*/
	ret = twl4030_i2c_read_u8(TWL4030_MODULE_MADC, &data, REG_MADC_ISR1);
	if (ret) {
		printk(KERN_ERR "Could not read the ISR register\n");
		return ret;
	}
	/* Setting the prescalers for vavrious channels
	 * TBD: See if this has to be moved to specific
	 * modules
	 */
	ret = clear_n_set(TWL4030_MODULE_MAIN_CHARGE, 0,
		BCICTL1_MESVBUS | BCICTL1_MESBAT | BCICTL1_MESVAC, REG_BCICTL1);
	if (ret) {
		printk(KERN_ERR "Could not set the prescalers\n");
		return ret;
	}
	/*Unmasking SW1 interrupt*/
	ret = clear_n_set(TWL4030_MODULE_MADC, MADC_IMR1_SW1IMR1, 0,
			REG_MADC_IMR1);
	if (ret) {
		printk(KERN_ERR "Could not unmask SW1IMR1\n");
		return ret;
	}
	/* turning madc on */
	ret = twl4030_i2c_write_u8(TWL4030_MODULE_MADC, CTRL1_MADC_ON,
		REG_CTRL1);
	if (ret) {
		printk(KERN_ERR "Could Not turn on MADC module\n");
		return ret;
	}
	dev_dbg(dbg_dev, "twl4030madc_setup-");
	return 0;
}

/**
 * Function : twl4030madc_convert
 * Description :
 * Starts the SW1 conversion process on the channles specified in
 * the struct ch_request.The result after conversion is put back into
 * the data field of madc_chrequest per channel.This function is exported
 * so that it can be called from other kernel modules like BCI driver.
 * Parameters:
 * ch_request : array of requests per request structure containing channel
 * 		to be converted, field specifying whether averaging has
 * 		to be done, and the data field in which the conversion
 * 		result has to be passed back
 * count      : number of requests in the array ch_request
 */
static int twl4030madc_convert(struct madc_chrequest *ch_request,
		unsigned int count)
{
	int ret = 0, i;
	u8 data;
	unsigned int rd_val;

	dev_dbg(dbg_dev, "twl4030madc_convert+");
	/*
	 * Check if a conversion is already going on.If yes return
	 * -EBUSY
	 */
	down(&sw1conv_lock);
	if (swconv1_ongoing) {
		up(&sw1conv_lock);
		return -EBUSY;
	}
	swconv1_ongoing = 1;
	up(&sw1conv_lock);

	for (i = 0; i < count; i++) {
		if (ch_request[i].channel < ADCIN8) {
			/**
			 * Set the channel on which sw1 conversion has to be
			 * peformed
			 */
			ret = clear_n_set(TWL4030_MODULE_MADC, 0,
				1 << ch_request[i].channel, REG_SW1SELECT_LSB);
			if (ret)
				goto out;
			/**
			 * If averaging feature is requested on the channel,
			 * enable it
			 */
			if (ch_request[i].is_average) {
				ret = clear_n_set(TWL4030_MODULE_MADC, 0,
					1 << ch_request[i].channel,
					REG_SW1AVERAGE_LSB);
				if (ret)
					goto out;
			}
		} else {
			ret = clear_n_set(TWL4030_MODULE_MADC, 0,
				1 << (ch_request[i].channel - 8),
				REG_SW1SELECT_MSB);
			if (ret)
				goto out;
			if (ch_request[i].is_average) {
				ret = clear_n_set(TWL4030_MODULE_MADC, 0,
					1 << (ch_request[i].channel - 8),
					REG_SW1AVERAGE_MSB);
				if (ret)
					goto out;
			}
		}
	}

	/* Start the conversion process*/
	ret = clear_n_set(TWL4030_MODULE_MADC, 0, CTRL_SW1_SW1, REG_CTRL_SW1);
	if (ret)
		goto out;
	ret = wait_for_completion_timeout(&sw1conv_complete, 200);
	if (!ret) {
		printk(KERN_ERR "timed out while waiting for conversion \
				completion\n");
		ret = -ETIMEDOUT;
		goto out;
	}
	for (i = 0; i < count; i++) {
		ret = twl4030_i2c_read_u8(TWL4030_MODULE_MADC, &data,
				(REG_GPCH0_MSB + 2 * ch_request[i].channel));
		if (ret)
			goto out;
		rd_val = data << 2;
		ret = twl4030_i2c_read_u8(TWL4030_MODULE_MADC, &data,
				(REG_GPCH0_LSB + 2 * ch_request[i].channel));
		if (ret)
			goto out;
		rd_val = rd_val | (data >> 6);
		dev_dbg(dbg_dev, "The read value is %x\n", rd_val);
		rd_val = CONV_VOLTAGE(rd_val, ch_request[i].channel);
		ch_request[i].data = rd_val;
		dev_dbg(dbg_dev, "The SW1 conv value from channel %d is %d\n",
				ch_request[i].channel, ch_request[i].data);
	}
	dev_dbg(dbg_dev, "twl4030madc_convert-");
out:
	/*
	 * Clear the channel select and averaging registers so that
	 * next time when a request comes the current channels are
	 * not converted unnecessarily
	 */
	twl4030_i2c_write_u8(TWL4030_MODULE_MADC, 0, REG_SW1SELECT_LSB);
	twl4030_i2c_write_u8(TWL4030_MODULE_MADC, 0, REG_SW1SELECT_MSB);
	twl4030_i2c_write_u8(TWL4030_MODULE_MADC, 0, REG_SW1AVERAGE_LSB);
	twl4030_i2c_write_u8(TWL4030_MODULE_MADC, 0, REG_SW1AVERAGE_MSB);
	/* The module is no longer busy doing any cconversion */
	down(&sw1conv_lock);
	swconv1_ongoing = 0;
	up(&sw1conv_lock);
	return ret;
}
EXPORT_SYMBOL(twl4030madc_convert);
/**
 * Function : twl4030madc_interrupt
 * Description :
 * interrupt handler routine.handles sw1 conversion end of
 * sequence (SW1_ISR1) interrupt.
 */
static irqreturn_t twl4030madc_interrupt(int irq, void *dev_id)
{
	int ret;
	u8 isr1_val;

	ret = twl4030_i2c_read_u8(TWL4030_MODULE_MADC, &isr1_val,
			REG_MADC_ISR1);
	if (ret)
		return IRQ_NONE;
	if (isr1_val & MADC_ISR1_SW1ISR1)
		complete(&sw1conv_complete);
	return IRQ_HANDLED;
}

/**
 * Function : twl4030madc_open
 * Description :
 * Called from user space to open the file handle
 */
static int twl4030madc_open(struct inode *inode, struct file *file)
{
	return 0;
}

/**
 * Function : twl4030madc_release
 * Description :
 * Called from user space to close the file handle
 */
static int twl4030madc_release(struct inode *inode, struct file *file)
{
	return 0;
}

/**
 * Function : twl4030madc_ioctl
 * Description :
 * Handles ioctl MADC_SW1CONVERT.Point of entry from
 * user space
 */
static int twl4030madc_ioctl(struct inode *inode, struct file *file,
				unsigned int cmd, unsigned long arg)
{
	int ret = 0;

	dev_dbg(dbg_dev, "twl4030madc_ioctl+");
	switch (cmd) {
	case MADC_SW1CONVERT:
	{
		struct 	madc_swconvrequest sw_convreq;
		struct madc_chrequest *ch_request;
		int i;

		if (copy_from_user(&sw_convreq,
			(struct madc_swconvrequest *)arg,
			sizeof(struct madc_swconvrequest))) {
				printk(KERN_ERR "copyfromuser \
					madc_swconvrequest error\n");
				return -EFAULT;
		}
		ch_request = (struct madc_chrequest *)
				kmalloc(sizeof(struct madc_chrequest) *
				sw_convreq.no_request, GFP_KERNEL);
		for (i = 0; i < sw_convreq.no_request; i++) {
			if (copy_from_user(&ch_request[i],
				&(sw_convreq.request[i]),
				sizeof(struct madc_chrequest))) {
					printk(KERN_ERR "copyfromuser \
						madc_chrequest error\n");
					ret = -EFAULT;
					goto out;
			}
		}
		ret = twl4030madc_convert(ch_request, sw_convreq.no_request);
		if (ret) {
			printk(KERN_ERR "Error in SW1 conversion\n");
			goto out;
		}
		for (i = 0; i < sw_convreq.no_request; i++) {
			if (copy_to_user(&(sw_convreq.request[i]),
				&ch_request[i],
				sizeof(struct madc_chrequest))) {
					printk(KERN_ERR "copytouser \
						madc_chrequest error\n");
					ret = -EFAULT;
					goto out;
			}
		}
out:
		kfree(ch_request);

	}
	}
	dev_dbg(dbg_dev, "twl4030madc_ioctl-");
	return ret;
}

/* FOPS */
static struct file_operations twl4030madc_fops = {
ioctl:twl4030madc_ioctl,
open:twl4030madc_open,
release:twl4030madc_release,
};

/**
 * Device structure to be registered with misc framework
 */
static struct miscdevice madc_dev = {
	MADC_MINOR, "twl4030madc", &twl4030madc_fops
};


/**
 * Function : twl4030madc_probe
 * Description:
 * Probe function for the madc driver.Registers itself with
 * the misc framework to export the ioctls.Requests for irq
 * and turns on the MADC module keeping it ready for conversions.
 */
static int twl4030madc_probe(struct platform_device *pdev)
{
	int ret;

	/* Get the debug Device */
	dbg_dev = &(pdev->dev);
	dev_dbg(dbg_dev, "twl4030madc_probe+");
	ret = request_irq(TWL4030_MODIRQ_MADC, twl4030madc_interrupt,
		IRQF_DISABLED, pdev->name, NULL);
	if (ret) {
		printk(KERN_ERR "MADC DRIVER:IRQ%d is not free.\n",
			TWL4030_MODIRQ_MADC);
		goto exit;
	}
	swconv1_ongoing = 0;
	init_completion(&sw1conv_complete);
	init_MUTEX(&sw1conv_lock);
	ret = twl4030madc_setup();
	if (ret) {
		printk(KERN_ERR "Could not turn on MADC module!!!!\n");
		goto exit1;
	}
	ret = misc_register(&madc_dev);
	if (ret) {
		printk(KERN_ERR "....MADC registration failed!!!!\n");
		goto exit1;
	}
	dev_dbg(dbg_dev, "twl4030madc_probe-");
	return 0;
exit1:
		free_irq(TWL4030_MODIRQ_MADC, NULL);
exit:
		platform_driver_unregister(&twl4030madc_driver);
		return ret;
}

/**
 * Function : twl4030madc_remove
 * Description:
 * Called in the exit path.Frees the acquired irq and
 * deregisters from the misc framework
 */
static int twl4030madc_remove(struct  platform_device *dev)
{
	int ret = 0;

	dev_dbg(dbg_dev, "twl4030madc_remove+");
	/* Turn off the madc module*/
	ret = clear_n_set(TWL4030_MODULE_MADC, CTRL1_MADC_ON, 0,
		REG_CTRL1);
	if (ret)
		printk(KERN_ERR "Could Not turn off MADC module\n");
	/*Mask SW1 interrupt which was unmasked in probe*/
	ret = clear_n_set(TWL4030_MODULE_MADC, 0, MADC_IMR1_SW1IMR1,
			REG_MADC_IMR1);
	if (ret)
		printk(KERN_ERR "Could not mask SW1IMR1\n");

	free_irq(TWL4030_MODIRQ_MADC, NULL);
	misc_deregister(&madc_dev);
	dev_dbg(dbg_dev, "twl4030madc_remove-");
	return ret;
}

#ifdef CONFIG_PM

/**
 * Function : twl4030madc_suspend
 * Description :
 * Called when system wide suspend is happening
 * Turns off the MADC module
 */
static int twl4030madc_suspend(struct platform_device *dev,
		pm_message_t state)
{
	int ret;

	ret = clear_n_set(TWL4030_MODULE_MADC, CTRL1_MADC_ON, 0,
		REG_CTRL1);
	if (ret)
		printk(KERN_ERR "Could Not turn off MADC module\n");
	return ret;
}

/**
 * Function : twl4030madc_resume
 * Description :
 * Called when system wide resume is happening
 * Turns on the MADC module
 */
static int twl4030madc_resume(struct platform_device *dev)
{
	int ret;

	ret = clear_n_set(TWL4030_MODULE_MADC, 0, CTRL1_MADC_ON,
		REG_CTRL1);
	if (ret)
		printk(KERN_ERR "Could Not turn on MADC module\n");
	return ret;
}
#endif

/**
 * Function : twl4030madc_init
 * Description :
 * Registers TWL4030 MADC driver with LDM
 */
static int __init twl4030madc_init(void)
{
	return platform_driver_register(&twl4030madc_driver);
}

/**
 * Function : twl4030madc_exit
 * Description :
 * exit routine-unregisters the driver
 */
static void __exit twl4030madc_exit(void)
{
	platform_driver_unregister(&twl4030madc_driver);
}

MODULE_AUTHOR("Texas Instruments");
MODULE_LICENSE("GPL");

module_init(twl4030madc_init);
module_exit(twl4030madc_exit);
