/* 2009-08-24: File added and changed by Sony Corporation */
/*
 * Copyright 2004-2007 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 *
 * http://www.opensource.org/licenses/gpl-license.html
 * http://www.gnu.org/copyleft/gpl.html
 */

/*!
 * @file pmic_core_spi.c
 * @brief This is the main file for the PMIC Core/Protocol driver. SPI 
 * should be providing the interface between the PMIC and the MCU.
 *
 * @ingroup PMIC_CORE
 */

/*
 * Includes
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/spi/spi.h>

#include <asm/uaccess.h>
#include <asm/arch/gpio.h>

#include <asm/arch/pmic_external.h>
#include <asm/arch/pmic_status.h>
#include "pmic.h"

/*
 * Global variables
 */
static pmic_version_t mxc_pmic_version;
unsigned int active_events[MAX_ACTIVE_EVENTS];
#ifdef CONFIG_PGPF_MXC_PMIC_WAKEUP_EVENT_HANDLE
static unsigned int wakeup_events[MAX_ACTIVE_EVENTS];
static unsigned int wakeup_events_count;
#endif /* CONFIG_PGPF_MXC_PMIC_WAKEUP_EVENT_HANDLE */
static unsigned int pmic_irq = 0;

/* 
 * Static functions 
 */
static void pmic_pdev_register(void);
static void pmic_pdev_unregister(void);
void pmic_bh_handler(struct work_struct *work);

/* 
 * Platform device structure for PMIC client drivers 
 */
static struct platform_device adc_ldm = {
	.name = "pmic_adc",
	.id = 1,
};
static struct platform_device battery_ldm = {
	.name = "pmic_battery",
	.id = 1,
};
static struct platform_device power_ldm = {
	.name = "pmic_power",
	.id = 1,
};
static struct platform_device rtc_ldm = {
	.name = "pmic_rtc",
	.id = 1,
};
static struct platform_device light_ldm = {
	.name = "pmic_light",
	.id = 1,
};

/*
 * External functions
 */
extern void pmic_event_list_init(void);
extern void pmic_event_callback(type_event event);
#ifdef CONFIG_PGPF_NAVU2
#define gpio_pmic_active() /* nothing */
#else
extern void gpio_pmic_active(void);
#endif

/*! 
 * Bottom half handler of PMIC event handling.
 */
DECLARE_WORK(pmic_ws, pmic_bh_handler);

/*! 
 * This function registers platform device structures for 
 * PMIC client drivers.
 */
static void pmic_pdev_register(void)
{
	platform_device_register(&adc_ldm);
	platform_device_register(&battery_ldm);
	platform_device_register(&rtc_ldm);
	platform_device_register(&power_ldm);
	platform_device_register(&light_ldm);
}

/*! 
 * This function unregisters platform device structures for 
 * PMIC client drivers.
 */
static void pmic_pdev_unregister(void)
{
	platform_device_unregister(&adc_ldm);
	platform_device_unregister(&battery_ldm);
	platform_device_unregister(&rtc_ldm);
	platform_device_unregister(&power_ldm);
	platform_device_unregister(&light_ldm);
}

/*!
 * This function is called when pmic interrupt occurs on the processor.
 * It is the interrupt handler for the pmic module.
 *
 * @param        irq        the irq number
 * @param        dev_id     the pointer on the device
 *
 * @return       The function returns IRQ_HANDLED when handled.
 */
static irqreturn_t pmic_irq_handler(int irq, void *dev_id)
{
#ifdef CONFIG_PGPF_MXC_PMIC_USE_LEVEL_INTERRUPT
	disable_irq(irq);
#endif /* CONFIG_PGPF_MXC_PMIC_USE_LEVEL_INTERRUPT */

	/* prepare a task */
	schedule_work(&pmic_ws);

	return IRQ_HANDLED;
}

/*!
 * This function is the bottom half handler of the PMIC interrupt.
 * It checks for active events and launches callback for the
 * active events.
 */
void pmic_bh_handler(struct work_struct *work)
{
	unsigned int loop;
	unsigned int count = 0;

	count = pmic_get_active_events(active_events);

	for (loop = 0; loop < count; loop++) {
		pmic_event_callback(active_events[loop]);
	}
#ifdef CONFIG_PGPF_MXC_PMIC_USE_LEVEL_INTERRUPT
	enable_irq(pmic_irq);
#endif /* CONFIG_PGPF_MXC_PMIC_USE_LEVEL_INTERRUPT */

	return;
}

/*!
 * This function is used to determine the PMIC type and its revision.
 *
 * @return      Returns the PMIC type and its revision.
 */

pmic_version_t pmic_get_version(void)
{
	return mxc_pmic_version;
}

EXPORT_SYMBOL(pmic_get_version);

/*!
 * This function puts the SPI slave device in low-power mode/state.
 *
 * @param	spi	the SPI slave device
 * @param	message	the power state to enter
 *
 * @return 	Returns 0 on SUCCESS and error on FAILURE. 
 */
#ifdef CONFIG_PGPF_NAVU2
static int pmic_suspend(struct platform_device *pdev, pm_message_t message)
#else
static int pmic_suspend(struct spi_device *spi, pm_message_t message)
#endif
{
#ifdef CONFIG_PGPF_MXC_PMIC_WAKEUP_EVENT_HANDLE
#ifdef CONFIG_PGPF_NAVU2
	struct device pmic_dev = pdev->dev;
	int irq = pmic_irq;
#else
	struct device pmic_dev = spi->dev;
	int irq = spi->irq;
#endif
	if (message.event == PM_EVENT_SUSPEND) {
		disable_irq(irq);
		if (device_may_wakeup(&pmic_dev)) {
			wakeup_events_count = 0;
			enable_irq_wake(irq);
		}
	}
#endif /* CONFIG_PGPF_MXC_PMIC_WAKEUP_EVENT_HANDLE */
	return PMIC_SUCCESS;
}

/*!
 * This function brings the SPI slave device back from low-power mode/state.
 *
 * @param	spi	the SPI slave device
 *
 * @return 	Returns 0 on SUCCESS and error on FAILURE. 
 */
#ifdef CONFIG_PGPF_NAVU2
static int pmic_resume(struct platform_device *pdev)
#else
static int pmic_resume(struct spi_device *spi)
#endif
{
#ifdef CONFIG_PGPF_MXC_PMIC_WAKEUP_EVENT_HANDLE
#ifdef CONFIG_PGPF_NAVU2
	struct device pmic_dev = pdev->dev;
	int irq = pmic_irq;
#else
	struct device pmic_dev = spi->dev;
	int irq = spi->irq;
#endif
	unsigned int loop;

	if (device_may_wakeup(&pmic_dev)) {
		disable_irq_wake(irq);
		wakeup_events_count = pmic_get_active_events(wakeup_events);

		for (loop = 0; loop < wakeup_events_count; loop++) {
			printk(KERN_INFO "PMIC wakeup event: %d\n", wakeup_events[loop]);
		}
	}
	enable_irq(irq);
#endif /* CONFIG_PGPF_MXC_PMIC_WAKEUP_EVENT_HANDLE */
	return PMIC_SUCCESS;
}

/*!
 * This function is called whenever the SPI slave device is detected.
 *
 * @param	spi	the SPI slave device
 *
 * @return 	Returns 0 on SUCCESS and error on FAILURE. 
 */
#ifdef CONFIG_PGPF_NAVU2
static int __devinit pmic_probe(struct platform_device *pdev)
#else
static int __devinit pmic_probe(struct spi_device *spi)
#endif
{
	int ret = 0;
	struct device pmic_dev;
	int irq;
	char *name;

#ifdef CONFIG_PGPF_NAVU2
	pmic_dev = pdev->dev;
	irq = platform_get_irq(pdev, 0);
	name = "mc13783 on MP200 SPI0";
#else
	pmic_dev = spi->dev;
	irq = spi->irq;
	name = spi->dev.bus_id;
#endif
	
#ifndef CONFIG_PGPF_NAVU2
	if (!strcmp(spi->dev.bus_id, PMIC_ARBITRATION)) {
		if (PMIC_SUCCESS != pmic_fix_arbitration(spi)) {
			dev_err((struct device *)spi,
				"Unable to fix arbitration!! Access Failed\n");
			return -EACCES;
		}
		return PMIC_SUCCESS;
	}

	/* Initialize the PMIC parameters */
	ret = pmic_spi_setup(spi);
	if (ret != PMIC_SUCCESS) {
		return PMIC_ERROR;
	}
#endif /* CONFIG_PGPF_NAVU2 */

	/* Initialize the PMIC event handling */
	pmic_event_list_init();

	/* Initialize GPIO for PMIC Interrupt */
	gpio_pmic_active();

	/* Get the PMIC Version */
	pmic_get_revision(&mxc_pmic_version);
	if (mxc_pmic_version.revision < 0) {
		dev_err(&pmic_dev,
			"PMIC not detected!!! Access Failed\n");
		return -ENODEV;
	} else {
		dev_dbg(&pmic_dev,
			"Detected pmic core IC version number is %d\n",
			mxc_pmic_version.revision);
	}

	/* Initialize the PMIC parameters */
	ret = pmic_init_registers();
	if (ret != PMIC_SUCCESS) {
		return PMIC_ERROR;
	}

	/* Set and install PMIC IRQ handler */
#ifdef CONFIG_PGPF_MXC_PMIC_USE_LEVEL_INTERRUPT
	set_irq_type(irq, IRQF_TRIGGER_HIGH);
#else /* CONFIG_PGPF_MXC_PMIC_USE_LEVEL_INTERRUPT */
	set_irq_type(irq, IRQF_TRIGGER_RISING);
#endif /* CONFIG_PGPF_MXC_PMIC_USE_LEVEL_INTERRUPT */
	ret = request_irq(irq, pmic_irq_handler, 0, "PMIC_IRQ", 0);
	if (ret) {
		dev_err(&pmic_dev, "gpio1: irq%d error.", irq);
		return ret;
	}

	pmic_pdev_register();
#ifdef CONFIG_PGPF_MXC_PMIC_WAKEUP_EVENT_HANDLE
	device_init_wakeup(&pmic_dev, 1);
	pmic_event_unmask(EVENT_ONOFD1I);
	wakeup_events_count = 0;
#endif /* CONFIG_PGPF_MXC_PMIC_WAKEUP_EVENT_HANDLE */
	pmic_irq = irq;

	printk(KERN_INFO "Device %s probed\n", name);

	return PMIC_SUCCESS;
}

/*!
 * This function is called whenever the SPI slave device is removed.
 *
 * @param	spi	the SPI slave device
 *
 * @return 	Returns 0 on SUCCESS and error on FAILURE. 
 */
#ifdef CONFIG_PGPF_NAVU2
static int __devexit pmic_remove(struct platform_device *pdev)
#else
static int __devexit pmic_remove(struct spi_device *spi)
#endif
{
	int irq;
	char *name;

#ifdef CONFIG_PGPF_NAVU2
	irq = pmic_irq;
	name = "mc13783 on mp200_spi0";
#else
	irq = spi->irq;
	name = spi->dev.bus_id;
#endif
	free_irq(irq, 0);

	pmic_pdev_unregister();

	printk(KERN_INFO "Device %s removed\n", name);

	return PMIC_SUCCESS;
}

#ifdef CONFIG_PGPF_NAVU2
static struct platform_driver pmic_driver = {
	.driver = {
		.name = "navu2_mc13783",
		.owner = THIS_MODULE,
	},
	.probe = pmic_probe,
	.remove = __devexit_p(pmic_remove),
	.suspend = pmic_suspend,
	.resume = pmic_resume,
};
#else /* CONFIG_MP200_SPI0 */
/*!
 * This structure contains pointers to the power management callback functions.
 */
static struct spi_driver pmic_driver = {
	.driver = {
		   .name = "pmic_spi",
		   .bus = &spi_bus_type,
		   .owner = THIS_MODULE,
		   },
	.probe = pmic_probe,
	.remove = __devexit_p(pmic_remove),
	.suspend = pmic_suspend,
	.resume = pmic_resume,
};
#endif /* CONFIG_MP200_SPI0 */

/*
 * Initialization and Exit
 */

/*!
 * This function implements the init function of the PMIC device.
 * This function is called when the module is loaded. It registers 
 * the PMIC Protocol driver.
 *
 * @return       This function returns 0.
 */
static int __init pmic_init(void)
{
	pr_debug("Registering the PMIC Protocol Driver\n");
#ifdef CONFIG_PGPF_NAVU2
	return platform_driver_register(&pmic_driver);
#else
	return spi_register_driver(&pmic_driver);
#endif
}

/*!
 * This function implements the exit function of the PMIC device.
 * This function is called when the module is unloaded. It unregisters 
 * the PMIC Protocol driver.
 *
 */
static void __exit pmic_exit(void)
{
	pr_debug("Unregistering the PMIC Protocol Driver\n");
#ifdef CONFIG_PGPF_NAVU2
	platform_driver_unregister(&pmic_driver);
#else
	spi_unregister_driver(&pmic_driver);
#endif
}

/*
 * Module entry points
 */
#ifndef CONFIG_PGPF_NAVU2
subsys_initcall(pmic_init);
#else
module_init(pmic_init);
#endif
module_exit(pmic_exit);

MODULE_DESCRIPTION("Core/Protocol driver for PMIC");
MODULE_AUTHOR("Freescale Semiconductor, Inc.");
MODULE_LICENSE("GPL");
