/* 2010-08-25: File added and changed by Sony Corporation */
/*
 * drivers/net/nec_candy.c
 *
 * NEC candy ethernet driver.
 *
 * Author: Jun Sun <jsun@mvista.com>
 *
 * 2001-2006 (c) MontaVista Software, Inc. This file is licensed under
 * the terms of the GNU General Public License version 2. This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 *
 * Changes:
 *  MontaVista Software Inc. <source@mvista.com>
 *  - Added MII interface library support
 *  - Added Linux ethtool package support (based on 8139cp.c driver)
 *  - Added network interface message level support
 *  - Fixed VLTP register initialisation (set to zero)
 *
 *  MontaVista Software Inc. <source@mvista.com>
 *  - Added support for VR4133 (CCR register defines additional bits).
 *
 *  MontaVista Software Inc. <source@mvista.com>
 *  - Added support for NEC VR7701.
 *
 *  MontaVista Software Inc.<asapkov@ru.mvista.com> or <source@mvista.com>
 *  - Added support for NEC CMB-VR4133.
 *
 *  MontaVista Software Inc. <source@mvista.com>
 *  - Ported to 2.6 kernel.
 *  - Added platform device support
 *  - Added NAPI support
 */

#define DRV_NAME		"nec_candy"
#define DRV_VERSION		"0.4"
#define DRV_RELDATE		"June 20, 2006"

#include <linux/init.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/param.h>
#include <linux/etherdevice.h>
#include <linux/if.h>
#include <linux/ethtool.h>
#include <linux/mii.h>
#include <linux/crc32.h>
#include <linux/spinlock.h>
#include <linux/dma-mapping.h>
#include <linux/timer.h>
#include <linux/nec_candy_pd.h>
#include <linux/platform_device.h>
#ifdef CONFIG_SNSC_MACADDR
#include <linux/snsc_macaddr.h>
#endif

#include <asm/addrspace.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>

#ifdef CONFIG_EMMA3_FAMILY
#include <asm/emma3.h>
#endif

#include "nec_candy.h"

/***********************************************************************
 * debug
 ***********************************************************************/
#define CANDY_DEF_MSG_ENABLE	(NETIF_MSG_DRV		| \
				 NETIF_MSG_PROBE 	| \
				 NETIF_MSG_LINK)

static int debug = -1;
// #define	DEBUG_NEC_CANDY
// #define	DEBUG_VERBOSE_NEC_CANDY
// #define	SHOW_BUG		/* targeted bug reporting msgs */

#ifdef DEBUG_VERBOSE_NEC_CANDY
#define DEBUG_NEC_CANDY
#endif

#ifdef DEBUG_NEC_CANDY
#define ASSERT(x) if (!(x)) { panic("%s:%d - assert failed!\n", __FILE__, __LINE__); }
#define VERIFY(x, y) ASSERT(x y)
#define DEBUG(x)  do { x; } while (0)
#else
#define ASSERT(x)
#define VERIFY(x, y) x
#define DEBUG(x)
#endif

#ifdef DEBUG_VERBOSE_NEC_CANDY
#define DEBUG_VERBOSE(x)  do { x; } while (0)
#else
#define DEBUG_VERBOSE(x)
#endif

/***********************************************************************
 * global data
 ***********************************************************************/

/* These identify the driver base version and may not be removed. */
static char version[] __devinitdata =
    DRV_NAME ": 10/100 Ethernet driver v" DRV_VERSION " (" DRV_RELDATE ")";

static struct candy_private *candy_priv_head = NULL;

static unsigned short last_tx_err_status;
static unsigned int last_tx_buff_desc;

/***********************************************************************
 * helpers
 ***********************************************************************/
#ifdef CONFIG_EMMA3_FAMILY
#define	candy_in(x)	emma_in32((u32)&(x))
#define	candy_out(x, y)	emma_out32((u32)&(x), (u32)(y))
#else
#define	candy_in(x)	readl((void __iomem *) &(x))
#define	candy_out(x, y)	writel((unsigned long)(y), (void __iomem *)&(x))
#endif

#define	candy_set_bits(x, mask)  candy_out(x, candy_in(x) | mask)
#define	candy_clear_bits(x, mask)  candy_out(x, candy_in(x) & ~mask)

#define	candy_set_macc1_bits(x, mask) \
		candy_out(x, (candy_in(x) | mask) & MACC1_RESERVED)
#define	candy_clear_macc1_bits(x, mask) \
		candy_out(x, (candy_in(x) & ~mask) & MACC1_RESERVED)

#define	candy_set_macc2_bits(x, mask) \
		candy_out(x, (candy_in(x) | mask) & MACC2_RESERVED)
#define	candy_clear_macc2_bits(x, mask) \
		candy_out(x, (candy_in(x) & ~mask) & MACC2_RESERVED)

#ifdef CONFIG_CANDY_NAPI
#define CANDY_NAPI_MASK_INTS		(RCVDN)
#define candy_unmask_napi_ints(p) \
		candy_out((p)->msr, candy_in((p)->msr) | CANDY_NAPI_MASK_INTS)
#define candy_mask_napi_ints(p) \
		candy_out((p)->msr, candy_in((p)->msr) & ~CANDY_NAPI_MASK_INTS)
#endif

#define MEMSYNC(x)	\
	do { volatile u32 *p = (volatile u32 *)x; (void)(*p); } while (0)

/***********************************************************************
 * low-level hardware functions
 ***********************************************************************/

static void
mdio_write(struct net_device *dev, int phy_id, int location, int val)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;

	/* wait for busy */
	while (candy_in(p->mind) & BUSY) ;

	/* start the write */

	candy_out(p->madr,
		  ((phy_id << FIAD_SHIFT) & FIAD_MASK) |
		  (location & RGAD_MASK));
	candy_out(p->mwtd, (ulong) val);
}

static int
mdio_read(struct net_device *dev, int phy_id, int location)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;

	/* wait for busy */
	while (candy_in(p->mind) & BUSY) ;

	/*  mac_reg->mcmd = 0; */
	candy_out(p->madr,
		  ((phy_id << FIAD_SHIFT) & FIAD_MASK) |
		  (location & RGAD_MASK));
	candy_out(p->mcmd, RSTAT);

	/* wait for busy */
	while (candy_in(p->mind) & BUSY) ;

	return (ushort) candy_in(p->mrdd);
}

static void
candy_set_media_speed(struct net_device *dev)
{
	struct candy_private *pp = dev->priv;
	struct nec_candy_platform_data *pdata = pp->dev->platform_data;
	int lpa, advertise, media, reg;
	ulong ccr;

	/*
	 * As described in specs, SDP100 works with RMII mode, and is not
	 * related to MII mode. Hence if MII selected, let's leave it ON.
	 * this will be a workaround for VR4133A rev 2.0, formerly known as
	 * WORKAROUND_E19_VR4133A.
	 *
	 * SPD100   10HALF  10FULL  100HALF 100FULL
	 * mii      1       1       1       1
	 * rmii     1       1       0       0
	 */

	/* load default settings */
	ccr = SPD100 | MII_PIN_SELECT;
	if (pdata->rmii)
		ccr |= RMII_MODE;

	reg = mdio_read(pp->mii_if.dev, pp->mii_if.phy_id, MII_BMCR);
	if (!(reg & BMCR_ANENABLE))
		lpa = advertise = 0;
	else {
		lpa = mdio_read(pp->mii_if.dev, pp->mii_if.phy_id, MII_LPA);
		advertise = mdio_read(pp->mii_if.dev,
				      pp->mii_if.phy_id, MII_ADVERTISE);
	}

	/*
	 * If auto-negotiation is disabled or something weird is there,
	 * conform to the control register.
	 */
	if (lpa == 0 || advertise == 0) {
		pp->link_status.fullduplex = (reg & BMCR_FULLDPLX) ? 1 : 0;
		if (reg & BMCR_SPEED100) {
			if (pdata->rmii)
				ccr &= ~SPD100;
			pp->link_status.speed100 = 1;
		} else {
			pp->link_status.speed100 = 0;
		}
		candy_out(pp->regs->ccr, ccr);
		return;
	}

	media = mii_nway_result(lpa & advertise);
	switch (media) {
		case LPA_10FULL:
		case LPA_10HALF:
			pp->link_status.speed100 = 0;
			break;
		default:
			if (pdata->rmii)
				ccr &= ~SPD100;
			pp->link_status.speed100 = 1;
	}

	switch (media) {
		case LPA_10HALF:
		case LPA_100HALF:
			pp->link_status.fullduplex = 0;
			break;
		default:
			pp->link_status.fullduplex = 1;
	}

	candy_out(pp->regs->ccr, ccr);
}

static int
candy_get_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
	struct candy_private *pp = netdev_priv(dev);

	spin_lock_irq(&pp->lock);
	mii_ethtool_gset(&pp->mii_if, cmd);
	if (cmd->autoneg == AUTONEG_DISABLE){
		cmd->advertising = 0;
	}
	spin_unlock_irq(&pp->lock);

	return 0;
}

static int
candy_set_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
	struct candy_private *pp = netdev_priv(dev);
	int r;

	spin_lock_irq(&pp->lock);
	r = mii_ethtool_sset(&pp->mii_if, cmd);
	spin_unlock_irq(&pp->lock);

	return 0;
}

static void
candy_get_drvinfo(struct net_device *dev, struct ethtool_drvinfo *info)
{
	strcpy(info->driver, DRV_NAME);
	strcpy(info->version, DRV_VERSION);
	strcpy(info->bus_info, "");
	info->regdump_len = CANDY_REGS_SIZE;
	info->n_stats = 0;
}

static int
candy_get_regs_len(struct net_device *dev)
{
	return CANDY_REGS_SIZE;
}

static void
candy_get_regs(struct net_device *dev, struct ethtool_regs *regs, void *p)
{
	struct candy_private *pp = netdev_priv(dev);

	regs->version = CANDY_REGS_VER;
	spin_lock_irq(&pp->lock);
	memcpy_fromio(p, pp->regs, CANDY_REGS_SIZE);
	spin_unlock_irq(&pp->lock);
}

static u32
candy_get_msglevel(struct net_device *dev)
{
	struct candy_private *pp = netdev_priv(dev);
	return pp->msg_enable;
}

static void
candy_set_msglevel(struct net_device *dev, u32 value)
{
	struct candy_private *pp = netdev_priv(dev);
	pp->msg_enable = value;
}

static int
candy_nway_reset(struct net_device *dev)
{
	struct candy_private *pp = netdev_priv(dev);
	return mii_nway_restart(&pp->mii_if);
}

static u32
candy_get_link(struct net_device *dev)
{
	struct candy_private *pp = netdev_priv(dev);
	return mii_link_ok(&pp->mii_if);
}

static u32
candy_get_rx_csum(struct net_device *dev)
{
	/* RX cksum is not supported in this driver */
	return 0;
}

static u32
candy_get_tx_csum(struct net_device *dev)
{
	return (dev->features & NETIF_F_IP_CSUM) != 0;
}

static const struct ethtool_ops candy_ethtool_ops = {
	.get_settings		= candy_get_settings,
	.set_settings		= candy_set_settings,
	.get_drvinfo		= candy_get_drvinfo,
	.get_regs_len		= candy_get_regs_len,
	.get_regs		= candy_get_regs,
	.get_msglevel		= candy_get_msglevel,
	.set_msglevel		= candy_set_msglevel,
	.nway_reset		= candy_nway_reset,
	.get_link		= candy_get_link,
	.get_rx_csum		= candy_get_rx_csum,
	.get_tx_csum		= candy_get_tx_csum,
	.set_tx_csum		= ethtool_op_set_tx_csum,
	.get_sg			= ethtool_op_get_sg,
	.set_sg			= ethtool_op_set_sg,
	.get_tso		= ethtool_op_get_tso,
	.set_tso		= ethtool_op_set_tso,
};

static int
candy_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	struct candy_private *pp = dev->priv;
	struct mii_ioctl_data *mii = (struct mii_ioctl_data *) &rq->ifr_data;
	int rc;

	spin_lock_irq(&pp->lock);
	rc = generic_mii_ioctl(&pp->mii_if, mii, cmd, NULL);
	spin_unlock_irq(&pp->lock);
	return rc;
}

static void
set_mac_addr(struct net_device *dev, u_char * addr)
{
	candy_regs *p = ((struct candy_private *) dev->priv)->regs;

	candy_out(p->lsa2, (addr[0] << 8 | addr[1]) & LSA2_MASK);
	candy_out(p->lsa1,
		  addr[2] << 24 | addr[3] << 16 | addr[4] << 8 | addr[5]);
}

static void
candy_filter_init(struct net_device *dev)
{
	candy_regs *p = ((struct candy_private *) dev->priv)->regs;

	/* don't call while processing error recovery */

	candy_out(p->ht1, 0);
	candy_out(p->ht2, 0);

#ifdef	WORKAROUND_E10_PRM_AMC
	candy_out(p->afr, ABC);
#else
	candy_out(p->afr, PRM | ABC | AMC);
#endif
}

/*
 * This is called when system boots up and/or after ether chip is reset.
 */
static void
candy_hw_init(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;
	unsigned int temp;
	u32 afr, ht1, ht2;

	/*
	 * software reset - resets all modules; DMA, FIFO, MAC and MII/RMII
	 *
	 * FIXME: save and restore the contents of AFR and its friends.
	 */

	afr = candy_in(p->afr);
	ht1 = candy_in(p->ht1);
	ht2 = candy_in(p->ht2);

	candy_out(p->ccr, SRT);
	candy_in(p->isr);

	candy_out(p->afr, afr);
	candy_out(p->ht1, ht1);
	candy_out(p->ht2, ht2);

	/*
	 * MII software reset
	 */
	/* we need to maintain CLKS[2:0] = 3'b001 during reset */
	candy_out(p->miic, MISRT | CLKS33);
	candy_out(p->miic, CLKS33);

	/*
	 * initialize MAC
	 */
	temp = TXFC | RXFC | CRCEN | PADEN;
#ifdef WORKAROUND_E13_TXFC
	temp &= ~TXFC;
#endif
#ifdef WORKAROUND_E21_PAD
	temp &= ~PADEN;
#endif
	candy_out(p->macc1, temp);
	candy_out(p->macc2, APD);
	candy_out(p->ipgt, IPGT);
	candy_out(p->ipgr, IPGR1 | IPGR2);
	candy_out(p->clrt, LCOLW | RETRY);
	candy_out(p->lmax, MAXF);
#ifdef CONFIG_CANDY_VLAN
	candy_out(p->vltp, VLTP);
#else
	candy_out(p->vltp, 0x0);
#endif
#ifdef EMMA3P
	candy_out(p->miic, CLKS166);
#else
	candy_out(p->miic, CLKS66);
#endif

	/*
	 * initialize DMA / FIFO
	 */
#ifdef EMMA3P
	candy_out(p->txcfg, DTBS16);
	candy_out(p->txfc, TPTV | TX_DRTH | TX_FLTH);
	candy_out(p->rxcfg, DRBS16);
	candy_out(p->rxfc, UWM | LWM | RX_DRTH16W);
	candy_out(p->rxpd, AL | DUWM | DLWM);
#else	/* EMMA3P */
#ifdef WORKAROUND_E7_AFCE
	candy_out(p->txcfg, DTBS8);
#else
	candy_out(p->txcfg, DTBS8 | AFCE);
#endif
	candy_out(p->txfc, TPTV | TX_DRTH | TX_FLTH);
	candy_out(p->rxcfg, DRBS8);
	candy_out(p->rxfc, UWM | LWM | RX_DRTH16W);
	candy_out(p->rxpd, AL);
#endif	/* EMMA3P */

#ifdef CONFIG_CANDY_AFCE2
	/* unfortunately AFCE1 function is useless. never enable it. */
	candy_set_bits(p->txcfg, AFCE2 | AP0TE);
#endif

#ifdef EMMA3P
	/*
	 * mode configuration
	 */
	temp = 0x00800000; /* RSTH */
	candy_out(p->mode, temp);
#endif

	/* Set pins for MII/RMII use */
	candy_set_media_speed(dev);

	netif_carrier_off(dev);
	mii_check_media(&pp->mii_if, netif_msg_link(pp), 1);

	if (pp->mii_if.full_duplex)
		candy_set_macc1_bits(pp->regs->macc1, FULLD);
	else
		candy_clear_macc1_bits(pp->regs->macc1, FULLD);

	/*
	 * MAC software reset - set/clear these bits all together
	 */
	candy_set_macc2_bits(pp->regs->macc2, (MCRST | RFRST | TFRST));
	udelay(3);
	candy_clear_macc2_bits(pp->regs->macc2, (MCRST | RFRST | TFRST));

	set_mac_addr(dev, dev->dev_addr);
	/* afr, ht1, ht2 are initialized by candy_filter_init() */

	/*
	 * disable all interrupts until dev is opened later
	 */
	candy_out(p->msr, 0);
}

static void
candy_down(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;

	DEBUG_VERBOSE(printk("candy_down() invoked.\n"));

	candy_clear_macc1_bits(p->macc1, SRXEN);

#ifdef CONFIG_CANDY_AFCE2
	candy_clear_bits(p->txcfg, (TXE | AFCE2 | AP0TE));
#else
	candy_clear_bits(p->txcfg, TXE);
#endif
	candy_clear_bits(p->rxcfg, RXE);

	/*
	 * disable all interrupts until dev is opened later
	 */
	candy_out(p->msr, 0);
}

static void
candy_up(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;

	DEBUG_VERBOSE(printk("candy_up() invoked.\n"));

	/* set number of descriptors */
	candy_out(p->rxpd, AL | DUWM | DLWM | (RX_RING_SIZE & RNOD_MASK));

	/* set the receive descriptor pointer */
	candy_out(p->rxdp, CPHYSADDR(&pp->rx_ring[pp->rx_head]));

	/* enable transmit and receive */
	candy_set_bits(p->txcfg, TXE);
	candy_set_bits(p->rxcfg, RXE);

	/* turn on interrupts */
	candy_in(p->isr);
	candy_out(p->msr, 0xffffffff & ISR_RESERVED);

	/* start receiving */
	candy_set_macc1_bits(p->macc1, SRXEN);
}

#if 0 /* replaced by ether_crc() */
static unsigned int
hashit(char *addr)
{
	int i;
	int j;
	unsigned char nibblehigh;
	unsigned char nibblelow;
	unsigned long da_crc = 0L;
	unsigned long hash_p = 0L;
	unsigned long hash_byte = 0L;
	unsigned long hash_bit = 0L;
	unsigned long crc32 = 0xffffffffL;
	unsigned long crcpoly2 = 0xedb88320L;

	for (i = 0; i < 6; i++) {
		nibblelow = addr[i] & 0x0f;
		nibblehigh = addr[i] >> 4;
		crc32 ^= nibblelow;
		for (j = 0; j < 4; j++) {
			if (crc32 & 1) {
				crc32 = (crc32 >> 1) ^ crcpoly2;
			} else {
				crc32 >>= 1;
			}
		}
		crc32 ^= nibblehigh;
		for (j = 0; j < 4; j++) {
			if (crc32 & 1) {
				crc32 = (crc32 >> 1) ^ crcpoly2;
			} else {
				crc32 >>= 1;
			}
		}
	}
	da_crc = crc32 & 0x000001f8L;
	for (j = 31; j >= 23; j--) {
		hash_p = hash_p * 2 + da_crc % 2;
		da_crc = da_crc / 2;
	}
	hash_byte = (hash_p & 0x00000038) >> 3;
	hash_bit = (hash_p & 0x00000007);
	return ((hash_byte << 16) | (hash_bit & 0xFFFF));
}
#endif /* replaced by ether_crc() */

static void
candy_set_filter(struct net_device *dev, int on)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;
	struct dev_mc_list *mclist = dev->mc_list;
	unsigned long ht[2];
	int i;
	u32 bit_nr;

	DEBUG_VERBOSE(printk
		      ("candy_set_filter() invoked - %s\n", on ? "on" : "off"));

	ht[0] = ht[1] = 0;

	if (on) {
		for (i = 0, mclist = dev->mc_list;
		     mclist && i < dev->mc_count; i++, mclist = mclist->next) {
			bit_nr = (ether_crc(ETH_ALEN, mclist->dmi_addr) >> 23) & 0x3FUL;
			__set_bit(bit_nr, ht);
		}
	}

	/* lock ints */
	candy_out(p->ht1, ht[1]);
	candy_out(p->ht2, ht[0]);
}

/*
 * Apparently candy tx can stall due to various reasons.
 * This routine will attempt to recover from tx stall.
 */
static void
candy_error_recover(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;

	spin_lock(&pp->lock);

	netif_stop_queue(dev);

	candy_down(dev);
	candy_hw_init(dev);
	candy_up(dev);

	netif_wake_queue(dev);

	/* restart transmitting */
	pp->tx_ring[pp->tx_head].status = LAST | DB_LP;
	MEMSYNC(pp->tx_ring);
#if 0
	pp->tx_stop = pp->tx_tail;
#endif
	candy_out(p->txdp, CPHYSADDR(&pp->tx_ring[pp->tx_head]));

	spin_unlock(&pp->lock);
}

/*
 * This implements the workaround described for E-8, and Tx errors that stop the transmission
 */
static void
tx_stall_recover(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;

	/* set ths status to the correct value before retransmission */
	pp->tx_ring[pp->tx_head].status = LAST | DB_LP;
	MEMSYNC(pp->tx_ring);

	last_tx_err_status = pp->tx_ring[pp->tx_head].status;
	last_tx_buff_desc = (unsigned int)&pp->tx_ring[pp->tx_head];

#if defined(WORKAROUND_E8_TX_STALL)
	/* E-8 bug only happens when receiving is on */
	if ((candy_in(p->macc1) & SRXEN) && (candy_in(p->rxcfg) & RXE)) {

		candy_clear_macc1_bits(p->macc1, SRXEN);
		candy_clear_bits(p->rxcfg, RXE);

		udelay(20);

		candy_out(p->txdp, CPHYSADDR(&pp->tx_ring[pp->tx_head]));

		candy_set_bits(p->rxcfg, RXE);
		candy_set_macc1_bits(p->macc1, SRXEN);
	} else
#endif
	{
		candy_out(p->txdp, CPHYSADDR(&pp->tx_ring[pp->tx_head]));
	}
}

/***********************************************************************
 * hardware-independent helper routine
 ***********************************************************************/
static void
candy_init_rings(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	int i;

	DEBUG_VERBOSE(printk("candy_init_rings() invoked.\n"));

	/* tx rings */
	for (i = 0; i < TX_RING_SIZE; i++) {
		pp->tx_ring[i].status = 0;
		pp->tx_ring[i].size = 0;
		pp->tx_ring[i].pointer = 0;
	}
	/* this may be not necessary, if we reset every txdpr in intr */
	pp->tx_ring[TX_RING_SIZE].status = 0;
	pp->tx_ring[TX_RING_SIZE].size = 0x0;
	pp->tx_ring[TX_RING_SIZE].pointer = CPHYSADDR(pp->tx_ring);

	pp->tx_tail = pp->tx_stop = pp->tx_head = 0;
	pp->tx_count = 0;

	/* rx rings */
	for (i = 0; i < RX_RING_SIZE; i++) {
		pp->rx_ring[i].status = DB_LP;
		pp->rx_ring[i].size = RX_BUF_SIZE;
		pp->rx_ring[i].pointer = CPHYSADDR(pp->rx_skb[i]->data);
	}
	pp->rx_ring[RX_RING_SIZE].status = 0;	/* link back to the beginning */
	pp->rx_ring[RX_RING_SIZE].size = 0xffff;
	pp->rx_ring[RX_RING_SIZE].pointer = CPHYSADDR(pp->rx_ring);

	pp->rx_head = 0;
	pp->rx_disable = 0;
}

static void
candy_check_intr(ulong isr)
{
	static const char *intr_name[32] = {
		"carry flag",
		"reserved",
		"reserved",
		"link failed",
		"overflow",
		"receive buffer desc request at zero",
		"receive buffer desc request at alert level",
		"receive done",

		"reserved",
		"reserved",
		"control frame transmit",
		"transmit aborted",
		"underrun",
		"transmit frame length exceed",
		"transmit buffer descriptor request at NULL",
		"transmit done",

		"reserved",
		"reserved",
		"reserved",
		"reserved",	/* 4 */
		"reserved",
		"reserved",
		"reserved",
		"reserved",

		"reserved",
		"reserved",
		"reserved",
		"reserved",	/* 4 */
		"reserved",
		"reserved",
		"reserved",
		"IBUS error"
	};
	ulong i, j;

	for (i = 0, j = 1; i < 32; j <<= 1, i++) {
		if (j & isr)
			printk("\t%s\n", intr_name[i]);
	}
}

static void
reclaim_one_rx_desc(struct net_device *dev, char *buf)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;
	struct candy_desc *dp = &pp->rx_ring[pp->rx_head];

	if (buf != NULL) {
		dp->pointer = CPHYSADDR(buf);
	}
	dp->status = DB_LP;	/* 1 stands for buffer vs link ptr */
	dp->size = RX_BUF_SIZE;

	/* we need to clean up cache here.  Otherwise we may have some
	 * dirty cache while ether controller is feeding fresh pkt data
	 * to the physical RAM.  Data corruption could happen.
	 */
	dma_cache_wback_inv(KSEG0ADDR(dp->pointer), RX_BUF_SIZE);

	if (++pp->rx_head == RX_RING_SIZE)
		pp->rx_head = 0;

	/* tell hardware we have one descriptor to work with */
	candy_out(p->rxpd, AL | DUWM | DLWM | 1);
}

static void
restart_rx_hw(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	candy_regs *p = pp->regs;

	if (netif_msg_rx_status(pp))
		printk(KERN_DEBUG "%s : re-enable SRXEN\n", dev->name);

	/* restart receiving */
	candy_out(p->rxdp, candy_in(p->rxdp));
	udelay(23);
	candy_set_macc1_bits(p->macc1, SRXEN);
}

static void
handle_rx_error(struct net_device *dev, struct candy_desc *dp, ulong isr)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;

	if (netif_msg_rx_err(pp))
		printk(KERN_DEBUG "%s: rx_err, rx_ring[%d] error,"
		       "status =%04x, size=%d, isr = 0x%08lx.\n",
		       dev->name, pp->rx_head, dp->status, dp->size, isr);

	/* log some errors that hardware doesn't log */
	pp->stats.rx_errors++;
	if (dp->status & OVRN)
		pp->stats.rx_over_errors++;
}

#ifdef CONFIG_CANDY_NAPI

static int candy_napi_rx(struct net_device *dev, int budget)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	candy_regs *p = pp->regs;
	struct candy_desc *dp = &pp->rx_ring[pp->rx_head];
	struct sk_buff *rxskb, *newskb;
	int pkt_len, rnod;
	int received = 0;

	while ((dp->status & OWN) && (received < budget)) {

		/* check rx_disable */
		if (pp->rx_disable) {
			rnod = candy_in(p->rxpd) & RNOD_MASK;
			if (rnod >= RX_RESTART_BREAK) {
				/* clear rx_disable, then restart_rx_hw() */
				pp->rx_disable = 0;
				restart_rx_hw(dev);
			}
		}

		if ((dp->status & RXOK) == 0) { /* handle the error case */
			handle_rx_error(dev, dp, 0);
			reclaim_one_rx_desc(dev, NULL);
			dp = &pp->rx_ring[pp->rx_head];
			received++;
			continue;
		}

		/* oversize? */
#if 0
		if (dp->size > ETH_FRAME_SIZE) {
			if (netif_msg_rx_status(pp))
				printk(KERN_DEBUG "%s: rx, oversized pkt\n",
				       dev->name);

			handle_rx_error(dev, dp, 0);
			reclaim_one_rx_desc(dev, NULL);
			dp = &pp->rx_ring[pp->rx_head];
			received++;
			continue;
#else
		if (dp->size > ETH_FRAME_LEN + 4) {
			if (netif_msg_rx_status(pp))
				printk(KERN_DEBUG "%s: rx, oversized pkt\n",
				       dev->name);
			pkt_len = ETH_FRAME_LEN + 4;
#endif
		} else {
			pkt_len = dp->size;
		}

		/*
		 * we got a good packet,
		 * use the new skb to replace the received one
		 */
		rxskb = pp->rx_skb[pp->rx_head];

		newskb = dev_alloc_skb(RX_BUF_SIZE);    /* allocate a new skb */
		if (newskb == NULL) {
			printk(KERN_ERR "%s: memory squeeze.\n", dev->name);
			reclaim_one_rx_desc(dev, NULL);
			pp->stats.rx_dropped++;
			dp = &pp->rx_ring[pp->rx_head];
			received++;
			continue;
		}

		/* TODO: maybe we also want rx_copybreak for NAPI */

		newskb->dev = dev;
		skb_reserve(newskb, 2); /* align IP pkt to 16-byte */
		pp->rx_skb[pp->rx_head] = newskb;
		reclaim_one_rx_desc(dev, newskb->data);

		skb_put(rxskb, pkt_len);
		rxskb->protocol = eth_type_trans(rxskb, dev);
		rxskb->ip_summed = CHECKSUM_UNNECESSARY;
		netif_receive_skb(rxskb);
		received++;
		dev->last_rx = jiffies;

		dp = &pp->rx_ring[pp->rx_head];
	}

	if (netif_msg_rx_status(pp))
		printk(KERN_DEBUG "%s: rx, processed %d frames.\n",
		       dev->name, received);

	/*
	 * we might fail to restart receiving due to budget.
	 */
	if (pp->rx_disable)
		printk(KERN_DEBUG "%s: couldn't restart receiving "
		       "budget=%d received=%d rxpd=%08x\n",
		       dev->name, budget, received, candy_in(pp->regs->rxpd));

	return received;
}

static int candy_napi_poll(struct net_device *dev, int *budget)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	int not_done;
	int orig_budget = *budget;
	int received = 0;

	spin_lock(&pp->rxlock);

	not_done = 0;

	if (orig_budget > dev->quota)
		orig_budget = dev->quota;

	received = candy_napi_rx(dev, orig_budget);

	*budget -= received;
	dev->quota -= received;

	if (received >= orig_budget)
		not_done = 1;

	if (!not_done) {
		netif_rx_complete(dev);
		candy_unmask_napi_ints(pp->regs);
	}

	spin_unlock(&pp->rxlock);

	return not_done;
}

#else /* CONFIG_CANDY_NAPI */

static void
candy_rx(struct net_device *dev, ulong isr)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;
	struct candy_desc *dp = &pp->rx_ring[pp->rx_head];
	int pkt_len, rnod;
	struct sk_buff *newskb;
	struct sk_buff *rxskb;
	int i;
	int skb_size;

	if (netif_msg_rx_status(pp)) {
#if 0
		printk(KERN_DEBUG "%s: rx, stat=0x%08lx\n",
		       dev->name, (ulong) candy_in(p->rxst));
#endif
		printk(KERN_DEBUG "\trx_head = %d, {0x%04x, %u, 0x%08lx}\n",
		       pp->rx_head, dp->status, dp->size, dp->pointer);
	}

	/* FIXME : we are fetching packets.  How do we know where the
	 * end is?  When OWN bit is 0 (in previous linux driver)?
	 *
	 * Checking OWN bit alone is not reliable especially when kgdb
	 * is used.  Hack to work around.  We really need to understand
	 * what entries we can consume here.
	 */
	for (i = 0;; i++) {
		dp = &pp->rx_ring[pp->rx_head];

		/* check rx_disable */
		if (pp->rx_disable) {
			/*
			 * Make sure the current RNOD is 4 or larger.
			 * This was originally the workarond for VR4133A E-20,
			 * but now it's renewed with RX_RESTART_BREAK.
			 */
			rnod = candy_in(p->rxpd) & RNOD_MASK;
			if (rnod >= RX_RESTART_BREAK) {
				pp->rx_disable = 0;
				restart_rx_hw(dev);
			}
		}

#if 0
		/*
		 * No need to hack to work around any more. The problem has
		 * been tracked down and fixed now. It wasn't working properly
		 * because of reclaim_one_rx_desc(). Checking rx_disable after
		 * reclaiming one rx desc is wrong.
		 *
		 * When NAPI disabled, rx_disable is set *only before* entering
		 * candy_rx(). Suppose RBDRU happens when we are in candy_rx().
		 * Rx DMA has already stopped at this moment and rx_diable
		 * will not be set until next candy_interrupt() dispatched,
		 * we can reclaim all rx descriptors before leaving candy_rx().
		 *
		 * Next time we enter candy_rx(), rx_disable is flagged,
		 * all rx descriptors are clean, *but* rx DMA is disabled.
		 * OWN bit could not be set unless we restart receiving, so we
		 * had to give a special treatment for ((rx_disable == 1) &&
		 * (i == 0)) case.
		 */
		if (((dp->status & OWN) == 0) &&
		    ((pp->rx_disable == 0) || (i != 0))) {
#else
		if ((dp->status & OWN) == 0) {
#endif
			/* no frame received in this descriptor yet */
			break;
		}

		/* handle the error case */
		if ((dp->status & RXOK) == 0) {
			handle_rx_error(dev, dp, isr);
			reclaim_one_rx_desc(dev, NULL);
			continue;
		}

		/* oversize? */
		if (dp->size > ETH_FRAME_LEN + 4) {
			if (netif_msg_rx_status(pp))
				printk(KERN_DEBUG "%s: rx, oversized pkt\n",
				       dev->name);
			pkt_len = ETH_FRAME_LEN + 4;
		} else {
			pkt_len = dp->size;
		}

		/* we got a good packet */

		/* STRATEGY: ether packet has 14 bytes.  So we will
		 * suffer from emulated unaligned access if we pass
		 * the skb straight to upper layer.  An alternative is
		 * to copy the buffer by offset of 2 and then pass it up.
		 * Then the overhead is copying.
		 *
		 * In general, it is more beneficial to copy if we have smaller
		 * packet.  Also, it is more beneficial to copy if we have
		 * faster machines.
		 *
		 * To keep it flexible, we will leave rx_copybreak flexible.
		 */

		if (pkt_len < rx_copybreak) {
			skb_size = pkt_len + 2;
		} else {
			skb_size = RX_BUF_SIZE;
		}

		/* allocate a new skb */
		newskb = dev_alloc_skb(skb_size);
		if (newskb == NULL) {
			printk(KERN_ERR
			       "%s: Memory squeeze, dropping packet.\n",
			       dev->name);
			reclaim_one_rx_desc(dev, NULL);
			pp->stats.rx_dropped++;
			continue;
		}
		newskb->dev = dev;
		skb_reserve(newskb, 2); /* align IP pkt to 16-byte */

		if (pkt_len < rx_copybreak) {
			/* we will copy */
			rxskb = pp->rx_skb[pp->rx_head];

			eth_copy_and_sum(newskb, rxskb->data, pkt_len, 0);

			rxskb = newskb;
			reclaim_one_rx_desc(dev, NULL);
		} else {
			/* use the new skb to replace the recived one */
			rxskb = pp->rx_skb[pp->rx_head];

			pp->rx_skb[pp->rx_head] = newskb;
			reclaim_one_rx_desc(dev, newskb->data);
		}

		skb_put(rxskb, pkt_len);
		rxskb->protocol = eth_type_trans(rxskb, dev);
		rxskb->ip_summed = CHECKSUM_UNNECESSARY;
		netif_rx(rxskb);

		dev->last_rx = jiffies;
	}

	/* rx_disable should be cleared */
	WARN_ON(pp->rx_disable);

	/*
	 * when we are out here, should rxdp be the same as
	 * &pp->rx_ring[pp->head]?
	 */
	if (netif_msg_rx_status(pp))
		printk(KERN_DEBUG "%s: rx, processed %d frames.\n",
		       dev->name, i);
#ifdef SHOW_BUG
	if ((candy_in(p->rxdp) != CPHYSADDR(&pp->rx_ring[pp->rx_head])) &&
	    (candy_in(p->rxdp) != CPHYSADDR(&pp->rx_ring[RX_RING_SIZE]))) {
		int i;
		printk
		    ("%s : unexpected out of rx - rx_ring[rx_head] = (%04x, %d)\n",
		     dev->name, dp->status, dp->size);
		for (i = 0; i < RX_RING_SIZE + 1; i++) {
			if (p->rxdp == CPHYSADDR(&pp->rx_ring[i]))
				break;
		}
		if (i == RX_RING_SIZE + 1)
			panic("cannot find current rx tail");
		printk("\trx_head = %d, rx_tail = %d\n", pp->rx_head, i);
	}
#endif
	return;
}

#endif /* CONFIG_CANDY_NAPI */

static void
append_one_tx_desc(struct net_device *dev,
		   ushort status,
		   ushort size, ulong pointer, struct sk_buff *skb)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	struct candy_desc *dp = &pp->tx_ring[pp->tx_tail];
#ifdef WORKAROUND_E10_VR4133
	int i, shift = pointer % 4;
#endif

	dp->status = status;
	dp->size = size;
	dp->pointer = pointer;

#ifdef WORKAROUND_E10_VR4133
	if (shift && (size == 30 || size == 31)) {
		if (skb_tailroom(skb) >= 4 - shift) {
			shift = 4 - shift;
			skb->data += shift;
			dp->pointer += shift;
			for (i = size + shift - 1; i >= 0; i--)
				skb->data[i] = skb->data[i - shift];
			dma_cache_wback_inv((ulong) (skb->data), skb->len);
		} else if (skb_headroom(skb) >=  shift) {
			skb->data -= shift;
			dp->pointer -= shift;
			for (i = 0; i < size; i++)
				skb->data[i] = skb->data[i + shift];
			dma_cache_wback_inv((ulong) (skb->data), skb->len);
		} else {
			panic("nec_candy.c: Unable to do a workaround for hardware bug (Restriction 10 for VR4133).\n");
		}
	}
#endif
	pp->tx_skb[pp->tx_tail] = skb;

	pp->tx_count++;
	if (++pp->tx_tail == TX_RING_SIZE)
		pp->tx_tail = 0;
}

static void
reclaim_one_tx_desc(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	struct candy_desc *dp = &pp->tx_ring[pp->tx_head];

	dp->status = 0;
	dp->size = 0;
	dp->pointer = 0;

	/* free skb */
	if (pp->tx_skb[pp->tx_head]) {
		dev_kfree_skb_irq(pp->tx_skb[pp->tx_head]);
		pp->tx_skb[pp->tx_head] = NULL;
	}

	pp->tx_count--;
	if (++pp->tx_head == TX_RING_SIZE)
		pp->tx_head = 0;
}

static void
restart_tx_hw(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;

	append_one_tx_desc(dev, 0, 0, 0, NULL);

	/*
	 * FIXME: This problem is associated with system bus.
	 *
	 * Now we've added a NULL descriptor and are about to re-start xmit.
	 * Beware of tx descs posted writes.
	 */
	MEMSYNC(pp->tx_ring);	/* read wherever, so long as uncached */

	/* start */
	pp->tx_stop = pp->tx_tail;
	candy_out(p->txdp, CPHYSADDR(&pp->tx_ring[pp->tx_head]));
}

static void
handle_tx_error(struct net_device *dev, struct candy_desc *dp, ulong isr)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;

	if (netif_msg_tx_err(pp))
		printk(KERN_DEBUG "%s: tx_ring[%d] error, "
		       "status = %04x, isr = 0x%08lx.\n",
		       dev->name, pp->tx_head, dp->status, isr);

	pp->stats.tx_errors++;
	pp->stats.tx_aborted_errors++;
	if (dp->status & TUDR)
		pp->stats.tx_fifo_errors++;
#if 0
	if (dp->status & HBF)
		pp->stats.tx_heartbeat_errors++;
#endif

	if (dp->status & (TUDR | LCOL | ECOL | EDFR))
		tx_stall_recover(dev);
}

static void
candy_tx_done(struct net_device *dev, ulong isr)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	struct candy_desc *dp;

	if (netif_msg_tx_done(pp))
		printk(KERN_DEBUG "%s: tx done, process %d frames from %d\n",
		       dev->name, pp->tx_tail > pp->tx_head ?
		       pp->tx_tail - pp->tx_head :
		       pp->tx_tail + TX_RING_SIZE - pp->tx_head, pp->tx_head);

	spin_lock(&pp->lock);	/* sync with xmit() */

	while (pp->tx_head != pp->tx_stop) {
		dp = &pp->tx_ring[pp->tx_head];

		/* deal with null descriptor, the "stop" packet */
		if ((isr & TBDR) && (dp->status == 0)) {
			reclaim_one_tx_desc(dev);
			continue;
		}

		/* how about checking OWN bit */
		if (!(dp->status & OWN)) {
			if (netif_msg_tx_done(pp))
				printk(KERN_DEBUG "%s: tx done, "
				       "found pkt being sent. Break the loop\n",
				       dev->name);
			break;
		}

		/* handle error */
		if (!(dp->status & TOK)) {
			handle_tx_error(dev, dp, isr);
			/* transmitter should have stopped */
			break;
		}

		/* reclaim the descriptor */
		if (!pp->tx_skb[pp->tx_head]) {
			printk(KERN_ERR "%s: tx_done but without skb!\n",
			       dev->name);
		}
		reclaim_one_tx_desc(dev);

		/* FIXME: The Japanese version has a tx restart under
		 * certain error conditions.  Don't understand it.
		 */
	}

	/* check if tx has stopped */
	if ((pp->tx_head == pp->tx_stop) && (pp->tx_stop != pp->tx_tail)) {
		restart_tx_hw(dev);
	}

	/* check if queue were stopped */
	if (netif_queue_stopped(dev) && (pp->tx_count < TX_RING_SIZE - 2)) {
		if (netif_msg_tx_done(pp))
			printk(KERN_DEBUG "%s: tx done, queue becomes free,"
			       "wake up net queue\n", dev->name);
		netif_wake_queue(dev);
	}

	spin_unlock(&pp->lock);
}

static void
candy_update_stats(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;

	/* some stats we get from hardware, while the rest we do
	 * counting by ourselves
	 */
	pp->stats.rx_packets = candy_in(p->rpkt);
	pp->stats.tx_packets = candy_in(p->tpct);
	pp->stats.rx_bytes = candy_in(p->rbyt);
	pp->stats.tx_bytes = candy_in(p->tbyt);

	/* we count rx_errors, tx_errors, rx_dropped, tx_dropped */

	pp->stats.multicast = candy_in(p->rmca);
	pp->stats.collisions = candy_in(p->tncl);

	pp->stats.rx_length_errors = candy_in(p->rund) +
	    candy_in(p->rovr) + candy_in(p->rfrg) + candy_in(p->rjbr);

	/* we count rx_over_errors */

	pp->stats.rx_crc_errors = candy_in(p->rfcs);
	pp->stats.rx_frame_errors = candy_in(p->raln);

	/* we count rx_fifo_errors and rx_missed_errors */

	/* we count tx_aborted_errors */

	pp->stats.tx_carrier_errors = candy_in(p->tcse);

	/* we count tx_fifo_errors */

	/* we don't have tx_heartbeat_errors */

	pp->stats.tx_window_errors = candy_in(p->tlcl);

	/* we don't have rx_compressed and tx_compressed */
}

/***********************************************************************
 * high-level linux-related functions
 ***********************************************************************/
static int candy_open(struct net_device *dev);
static int candy_close(struct net_device *dev);
static irqreturn_t
candy_interrupt(int irq, void *dev_instance)
{
	struct net_device *dev = (struct net_device *) dev_instance;
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;
	ulong isr;

	isr = candy_in(p->isr);

	if (netif_msg_intr(pp)) {
		printk(KERN_DEBUG "%s: intr, isr = 0x%08lx\n", dev->name, isr);
		candy_check_intr(isr);
	}

	if (isr & BUSERR) {
		printk(KERN_ERR "%s: bus error ... resetting\n", dev->name);
		candy_error_recover(dev);
		return IRQ_HANDLED;
	}

	if (isr & RBDRU) {
		/*
		 * disabling RX asap is a sane way not only for VR4133A E-20
		 * but also for all variants.
		 */
		candy_clear_bits(p->macc1, SRXEN);
		pp->rx_disable = 1;
		pp->stats.rx_missed_errors++;
		printk(KERN_ERR "%s : no more receive buffers. Stop receiving.\n",
		       dev->name);
	}

	if (isr & OF)
		pp->stats.rx_fifo_errors++;

#ifdef CONFIG_CANDY_NAPI
	if (isr & INT_ISR_RX_MASK) {
		if (netif_rx_schedule_prep(dev)) {
			candy_mask_napi_ints(p);
			__netif_rx_schedule(dev);
		}
	}
#else
	if (isr & INT_ISR_RX_MASK) {
		candy_rx(dev, isr);
	}
#endif

	if (isr & INT_ISR_TX_MASK) {
		candy_tx_done(dev, isr);
	}

	/* we may need to do something with other intrs too in the future */

	return IRQ_HANDLED;
}

static void
candy_get_phy_status(struct net_device *dev, int *duplex, int *linkup,
		     int *speed)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	u16 reg;

	reg = mdio_read(pp->mii_if.dev, pp->mii_if.phy_id, MII_BMCR);

	if (!(reg & BMCR_ANENABLE)) {
		/*
		 * Auto-negotiation is disabled so the full duplex bit in
		 * the control register tells us if the PHY is running
		 * half or full duplex.
		 */
		*duplex = (reg & BMCR_FULLDPLX) ? 1 : 0;
		*speed = (reg & BMCR_SPEED100) ? 1: 0;
	} else {
		/*
		 * Auto-negotiation is enabled.  Figure out what was
		 * negotiated by looking for the best mode in the union
		 * of what we and our partner advertise.
		 */
		u16 advertise, partner, negotiated;

		advertise = mdio_read(pp->mii_if.dev,
				      pp->mii_if.phy_id, MII_ADVERTISE);
		partner = mdio_read(pp->mii_if.dev, pp->mii_if.phy_id, MII_LPA);

		negotiated = advertise & partner & ADVERTISE_ALL;
		if (negotiated & ADVERTISE_100FULL) {
			*duplex = 1;
			*speed = 1;
		} else if (negotiated & ADVERTISE_100HALF) {
			*duplex = 0;
			*speed = 1;
		} else if (negotiated & ADVERTISE_10FULL) {
			*duplex = 1;
			*speed = 0;
		} else {
			*duplex = 0;
			*speed = 0;
		}
	}

	reg = mdio_read(pp->mii_if.dev, pp->mii_if.phy_id, MII_BMSR);

	*linkup = (reg & BMSR_LSTATUS) != 0;
}

static void
candy_poll_mii(unsigned long data)
{
	struct net_device *dev = (struct net_device *) data;
	struct candy_private *pp = (struct candy_private *) dev->priv;
	int phy_duplex, mac_duplex;
	int phy_carrier, netif_carrier;
	int phy_speed, mac_speed;

	/* First, find out what's going on with the PHY. */
	candy_get_phy_status(dev, &phy_duplex, &phy_carrier, &phy_speed);

	/* Second, figure out the speed/duplex of the EMAC */
	mac_duplex = pp->link_status.fullduplex;
	mac_speed = pp->link_status.speed100;

	/* Now see if there is a mismatch. */
	if ((mac_duplex != phy_duplex || phy_speed != mac_speed)
	    && phy_carrier) {
		/* reset */
		netif_stop_queue(dev);

		candy_down(dev);
		candy_hw_init(dev);
		candy_up(dev);

		netif_wake_queue(dev);
#if 0
		/*
		 * since we've already initialized hardware above, both SPD100
		 * and FULLD bits are already configured properly by hw_init().
		 * so we don't need to re-initialize ccr.SPD100, macc1,FULLD,
		 * link_status.{fullduplex,speed100} and mii_if.full_duplex.
		 *
		 * make sure that phy_{duplex,speed} no longer serve their
		 * original purposes at this point.
		 */
		printk(KERN_INFO "%s: Link has been changed: now ", dev->name);

		if (phy_duplex == 1 && phy_speed == 1) {
			candy_out(pp->regs->ccr,
				  candy_in(pp->regs->ccr) | SPD100);
			candy_set_macc1_bits(pp->regs->macc1, FULLD);
			printk("100Mbps Full Duplex\n");
		} else if (phy_duplex == 1 && phy_speed == 0) {
			candy_out(pp->regs->ccr,
				  candy_in(pp->regs->ccr) & ~SPD100);
			candy_set_macc1_bits(pp->regs->macc1, FULLD);
			printk("10Mbps Full Duplex\n");
		} else if (phy_duplex == 0 && phy_speed == 1) {
			candy_out(pp->regs->ccr,
				  candy_in(pp->regs->ccr) | SPD100);
			candy_clear_macc1_bits(pp->regs->macc1, FULLD);
			printk("100Mbps Half Duplex\n");
		} else if (phy_duplex == 0 && phy_speed == 0) {
			candy_out(pp->regs->ccr,
				  candy_in(pp->regs->ccr) & ~SPD100);
			candy_clear_macc1_bits(pp->regs->macc1, FULLD);
			printk("10Mbps Half Duplex\n");
		}

		pp->link_status.fullduplex = phy_duplex;
		pp->link_status.speed100 = phy_speed;
		pp->mii_if.full_duplex = phy_duplex;
#endif
	}

	netif_carrier = netif_carrier_ok(dev) != 0;

	if (phy_carrier != netif_carrier) {
		if (phy_carrier) {
			candy_set_media_speed(dev);
			printk(KERN_INFO "%s: Link carrier restored.\n",
			       dev->name);
			netif_carrier_on(dev);
		} else {
			printk(KERN_INFO "%s: Link carrier lost.\n", dev->name);
			netif_carrier_off(dev);
		}
	}

	/* Set up the timer so we'll get called again in 2 seconds. */
	mod_timer(&pp->phy_timer, jiffies + 2 * HZ);
}

static int
candy_open(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	int retval;
#if 0
	int count = 0;
#endif

	if (netif_msg_ifup(pp))
		printk(KERN_DEBUG "%s: enabling interface\n", dev->name);

	candy_init_rings(dev);

	candy_hw_init(dev);
	candy_filter_init(dev);
	candy_up(dev);

#if 0
	dev->flags |= IFF_RUNNING;
#endif

	/* request IRQ */
	retval = request_irq(dev->irq, &candy_interrupt, IRQF_SHARED,
			     dev->name, dev);
	if (retval) {
		printk(KERN_ERR "%s: unable to get IRQ %d\n",
		       dev->name, dev->irq);
		return retval;
	}

	init_timer(&pp->phy_timer);

	pp->phy_timer.data = (unsigned long) dev;
	pp->phy_timer.function = &candy_poll_mii;
	candy_poll_mii((unsigned long) dev);

#if 0
	/*
	 * FIXME: Why do we need some udelay()s here?
	 *
	 * Candy itself does not support "link change interrupt",
	 * so it doesn't make sense to wait here. As for detecting
	 * link status update, we rely on candy_poll_mii().
	 *
	 * If such an interrupt supported, the following routine might
	 * be good for detecting netif_carrier_on().
	 *
	 *     timeout = jiffies + 2 * HZ;
	 *     while (!netif_carrier_ok(dev) && time_before(jiffies, timeout))
	 *         msleep_interruptible(10);
	 *
	 * And as an aside, the original candy driver always used to
	 * trigger auto-negotiate on candy_open(). This code might be
	 * the vestige from then.
	 */
	while (!netif_carrier_ok(dev) && ++count < 5) {
		udelay(10);
	}
#endif

	netif_start_queue(dev);

	return 0;
}

static int
candy_close(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;

	if (netif_msg_ifdown(pp))
		printk(KERN_DEBUG "%s: disabling interface\n", dev->name);

	del_timer_sync(&pp->phy_timer);

#if 0
	dev->flags &= ~(IFF_UP | IFF_RUNNING);
#endif

	if (netif_device_present(dev)) {
		netif_stop_queue(dev);
		netif_carrier_off(dev);
		candy_down(dev);

		/* free tx skb */
		while (pp->tx_tail != pp->tx_head) {
			if (pp->tx_skb[pp->tx_head]) {
				dev_kfree_skb_irq(pp->tx_skb[pp->tx_head]);
				pp->tx_skb[pp->tx_head] = NULL;
			}

			pp->tx_count--;
			if (++pp->tx_head == TX_RING_SIZE)
				pp->tx_head = 0;
		}
	}

	free_irq(dev->irq, dev);

	return 0;
}

static int
candy_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	ulong flags;

	if (netif_msg_tx_queued(pp))
		printk(KERN_DEBUG "%s: tx queued, skblen %d\n",
		       dev->name, skb->len);

	/* check packet size */
#ifdef CONFIG_CANDY_VLAN
	if (skb->len > ETH_FRAME_LEN + 4) {
#else
	if (skb->len > ETH_FRAME_LEN) {
#endif
		printk(KERN_ERR "%s: packet size too big, %d\n", dev->name,
		       skb->len);
		pp->stats.tx_dropped++;
		return 1;
	}

	/* check to see if tx_ring is full */
	if (pp->tx_count >= TX_RING_SIZE - 1) {
		printk(KERN_ERR "%s: TX ring full, packet dropped.\n",
		       dev->name);
		pp->stats.tx_dropped++;
		/* why the queue was not stopped before we get here? */
		netif_stop_queue(dev);
		return 1;
	}

	spin_lock_irqsave(&pp->lock, flags);

	/* add the descriptor */
	{
		ushort temp = skb->len;
#ifdef WORKAROUND_E21_PAD
		if (temp < 60)
			temp = 60;
#endif
		dma_cache_wback_inv((ulong) (skb->data), skb->len);
		append_one_tx_desc(dev, LAST | DB_LP, temp,
				   CPHYSADDR(skb->data), skb);
	}

	/* logistics */
	dev->trans_start = jiffies;

	/* do we need to start sending or just append */
	if ((pp->tx_head == pp->tx_stop) && (pp->tx_stop != pp->tx_tail)) {
		restart_tx_hw(dev);
	}

	if (pp->tx_count >= TX_RING_SIZE - 2) {
		netif_stop_queue(dev);
	}

	spin_unlock_irqrestore(&pp->lock, flags);

	return 0;
}

static struct net_device_stats *
candy_get_stats(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	unsigned long flags;

	if (netif_device_present(dev)) {
		spin_lock_irqsave(&pp->lock, flags);
		candy_update_stats(dev);
		spin_unlock_irqrestore(&pp->lock, flags);
	}
	return &pp->stats;
}

static void
candy_set_rx_mode(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;
	ulong val;

	DEBUG_VERBOSE(printk("candy_set_rx_mode() invoked.\n"));

	/* TODO: need to acquire spinlock and stop receiving */

	val = candy_in(p->afr);
	val &= ~(PRO | PRM | AMC);
	if (dev->flags & IFF_PROMISC) {
		val |= PRO;
	} else if ((dev->flags & IFF_ALLMULTI) || (dev->mc_count > 64)) {
		/* disable promiscuous mode, use normal mode */
		candy_set_filter(dev, 0);
		val |= PRM;
	} else if (dev->mc_count) {
		/* walk the address list, and load the filter */
		candy_set_filter(dev, 1);
		val |= AMC;
	}
#ifdef	WORKAROUND_E10_PRM_AMC
	candy_out(p->afr, val & ABC);
#else
	candy_out(p->afr, val);
#endif
}

static void
candy_tx_timeout(struct net_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;

	printk(KERN_ERR "%s : tx_timeout.\n", dev->name);
	printk(KERN_ERR "Last Tx Err Status = 0x%04x\n", last_tx_err_status);
	printk(KERN_ERR "Last Tx Buff Desc = 0x%08x\n", last_tx_buff_desc);
	printk(KERN_ERR "TXDP = 0x%08x\n", candy_in(pp->regs->txdp));

	pp->stats.tx_errors++;

	candy_error_recover(dev);
}

#ifdef CONFIG_NET_POLL_CONTROLLER
static void candy_poll_controller(struct net_device *dev)
{
	disable_irq(dev->irq);
	candy_interrupt(dev->irq, dev);
	enable_irq(dev->irq);
}
#endif

static int nec_candy_probe(struct device *dev)
{
	int i, ret;
	struct resource *res;
	struct net_device *ndev;
	candy_regs *p;
	struct candy_private *pp;
	struct nec_candy_platform_data *pdata = dev->platform_data;
	struct platform_device *pdev = to_platform_device(dev);

	printk(KERN_INFO "%s\n", version);

	/*
	 * hardware is already initialized.  We just need do some Linux
	 * related initialization.
	 */

	/* create net_device structure */
	ndev = alloc_etherdev(sizeof(struct candy_private));
	if (!ndev) {
		printk(KERN_ERR "ether device alloc failed. aborting\n");
		ret = -EINVAL;
		goto err1;
	}

	/* init some device related data/func ptrs */

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM,
					   "nec_candy_regs");
	if (!res) {
		ret = -ENODEV;
		goto err2;
	}

	if (!request_mem_region(res->start, res->end - res->start + 1,
				"NEC_CANDY")) {
		ret = -EBUSY;
		goto err2;
	}

	ndev->base_addr = (unsigned long) ioremap(res->start,
						  res->end - res->start + 1);
	if (!ndev->base_addr) {
		ret = -ENOMEM;
		goto err3;
	}

	ndev->irq = platform_get_irq_byname(pdev, "nec_candy_irq");

#ifdef CONFIG_SNSC_MACADDR
	snsc_macaddr_get(ndev->dev_addr, "nec_candy");
#else
	for (i = 0; i < 6; i++)
		ndev->dev_addr[i] = pdata->mac_addr[i];
#endif

	ndev->open = candy_open;
	ndev->stop = candy_close;
	ndev->do_ioctl = candy_ioctl;
	ndev->hard_start_xmit = candy_xmit;
	ndev->get_stats = candy_get_stats;
	SET_ETHTOOL_OPS(ndev, &candy_ethtool_ops);
	ndev->set_multicast_list = candy_set_rx_mode;
	ndev->tx_timeout = candy_tx_timeout;
	ndev->watchdog_timeo = TX_TIMEOUT;

#ifdef CONFIG_CANDY_NAPI
	ndev->poll = candy_napi_poll;
	ndev->weight = 64;
#ifdef CONFIG_NET_POLL_CONTROLLER
	ndev->poll_controller = candy_poll_controller;
#endif
#endif

	/* init private data */
	pp = (struct candy_private *) ndev->priv;
	p = pp->regs = (candy_regs *) ndev->base_addr;
	pp->dev = dev;

	/* alloc tx/rx rings and rx buffers */

	pp->tx_ring = kmalloc(sizeof(struct candy_desc) * (TX_RING_SIZE + 1), GFP_ATOMIC);
	if (!pp->tx_ring) {
		ret = -ENOMEM;
		goto err4;
	}
	pp->rx_ring = kmalloc(sizeof(struct candy_desc) * (RX_RING_SIZE + 1), GFP_ATOMIC);
	if (!pp->rx_ring) {
		ret = -ENOMEM;
		goto err5;
	}

	dma_cache_wback_inv((ulong)pp->tx_ring,
		      sizeof(struct candy_desc) * (TX_RING_SIZE + 1));
	dma_cache_wback_inv((ulong)pp->rx_ring,
		      sizeof(struct candy_desc) * (RX_RING_SIZE + 1));

	pp->tx_ring = (void *) KSEG1ADDR(pp->tx_ring);
	pp->rx_ring = (void *) KSEG1ADDR(pp->rx_ring);

	/* allocate rx skbs */
	for (i = 0; i < RX_RING_SIZE; i++) {
		pp->rx_skb[i] = dev_alloc_skb(RX_BUF_SIZE);
		if (pp->rx_skb[i] == NULL) {
			panic("%s: failed to alloc rx skb!", ndev->name);
		}
		pp->rx_skb[i]->dev = ndev;
		skb_reserve(pp->rx_skb[i], 2); /* align IP pkt to 16-byte */
		dma_cache_wback_inv((ulong)pp->rx_skb[i]->data, RX_BUF_SIZE);
	}

	/* set up links */
	pp->ndev = ndev;
	pp->msg_enable = (debug < 0 ? CANDY_DEF_MSG_ENABLE : debug);
	pp->next = candy_priv_head;
	candy_priv_head = pp;
	dev_set_drvdata(dev, ndev);
	pp->mii_if.dev = ndev;
	pp->mii_if.mdio_read = mdio_read;
	pp->mii_if.mdio_write = mdio_write;
	pp->mii_if.phy_id = pdata->pmd_addr;
	pp->mii_if.phy_id_mask = 0x1f;
	pp->mii_if.reg_num_mask = 0x1f;

	spin_lock_init(&pp->lock);
	spin_lock_init(&pp->rxlock);

	/*==============================================================
	 * hardware initialization
	 *==============================================================
	 */

	/* TODO: maybe we want to make sure the chip is there */

	/*
	 * zero out counters
	 */
	candy_out(p->rbyt, 0);
	candy_out(p->rpkt, 0);
	candy_out(p->rfcs, 0);
	candy_out(p->rmca, 0);
	candy_out(p->rbca, 0);
	candy_out(p->rxcf, 0);
	candy_out(p->rxpf, 0);
	candy_out(p->rxuo, 0);
	candy_out(p->raln, 0);
	candy_out(p->rflr, 0);
	candy_out(p->rcde, 0);
	candy_out(p->rfcr, 0);
	candy_out(p->rund, 0);
	candy_out(p->rovr, 0);
	candy_out(p->rfrg, 0);
	candy_out(p->rjbr, 0);
	candy_out(p->r64, 0);
	candy_out(p->r127, 0);
	candy_out(p->r255, 0);
	candy_out(p->r511, 0);
	candy_out(p->r1k, 0);
	candy_out(p->rmax, 0);
	candy_out(p->rvbt, 0);

	candy_out(p->tbyt, 0);
	candy_out(p->tpct, 0);
	candy_out(p->tfcs, 0);
	candy_out(p->tmca, 0);
	candy_out(p->tbca, 0);
	candy_out(p->tuca, 0);
	candy_out(p->txpf, 0);
	candy_out(p->tdfr, 0);
	candy_out(p->txdf, 0);
	candy_out(p->tscl, 0);
	candy_out(p->tmcl, 0);
	candy_out(p->tlcl, 0);
	candy_out(p->txcl, 0);
	candy_out(p->tncl, 0);
	candy_out(p->tcse, 0);
	candy_out(p->time, 0);

	ret = register_netdev(ndev);
	if (ret != 0)
		goto err6;

	printk (KERN_INFO DRV_NAME ": Probe candy chip at "
		"0x%08lx, irq %d, Phy ID 0x%08x, "
		"MAC Addr %02x:%02x:%02x:%02x:%02x:%02x\n",
		res->start, ndev->irq, pp->mii_if.phy_id,
		ndev->dev_addr[0], ndev->dev_addr[1],
		ndev->dev_addr[2], ndev->dev_addr[3],
		ndev->dev_addr[4], ndev->dev_addr[5]);

	return 0;

err6:
	printk(KERN_ERR "Failed to register ethernet device\n");
	for (i = 0; i < RX_RING_SIZE; i++)
		dev_kfree_skb(pp->rx_skb[i]);
	kfree(pp->rx_ring);
err5:
	kfree(pp->tx_ring);
err4:
	iounmap((void *) ndev->base_addr);
err3:
	release_mem_region(res->start, res->end - res->start + 1);
err2:
	free_netdev(ndev);
err1:
	return ret;
}

static int nec_candy_remove(struct device *dev)
{
	int i;
	struct net_device *ndev = dev_get_drvdata(dev);
	struct candy_private *pp = netdev_priv(ndev);
	struct resource *res =
		platform_get_resource_byname(to_platform_device(dev),
					     IORESOURCE_MEM,
					     "nec_candy_regs");

	for (i = 0; i < RX_RING_SIZE; i++)
		dev_kfree_skb(pp->rx_skb[i]);

	kfree((const void *)KSEG0ADDR(pp->rx_ring));
	kfree((const void *)KSEG0ADDR(pp->tx_ring));

	iounmap((void *)ndev->base_addr);
	release_mem_region(res->start, res->end - res->start + 1);

	unregister_netdev(ndev);
	free_netdev(ndev);

	return 0;
}

#ifdef CONFIG_PM
static int nec_candy_suspend(struct device *dev, pm_message_t state)
{
	struct net_device *ndev = dev_get_drvdata(dev);
	struct candy_private *pp = netdev_priv(ndev);

	spin_lock(&pp->lock);

	if (!netif_running(ndev))
		goto out;

	netif_stop_queue(ndev);
	netif_device_detach(ndev);

	candy_down(ndev);
 out:
	spin_unlock(&pp->lock);

	return 0;
}

static int nec_candy_resume(struct device *dev)
{
	struct net_device *ndev = dev_get_drvdata(dev);
	struct candy_private *pp = netdev_priv(ndev);

	spin_lock(&pp->lock);

	if (!netif_running(ndev))
		goto out;

	candy_init_rings(ndev);
	candy_hw_init(ndev);
	candy_filter_init(ndev);
	candy_up(ndev);

	netif_device_attach(ndev);
	netif_wake_queue(ndev);
out:
	spin_unlock(&pp->lock);

	return 0;
}
#endif /* CONFIG_PM */

static struct device_driver nec_candy_driver = {
	.name		= "nec_candy",
	.bus		= &platform_bus_type,
	.probe		= nec_candy_probe,
	.remove		= nec_candy_remove,
#ifdef CONFIG_PM
	.suspend	= nec_candy_suspend,
	.resume		= nec_candy_resume,
#endif /* CONFIG_PM */
};

static int __init nec_candy_module_init(void)
{
	return driver_register(&nec_candy_driver);
}

static void __exit nec_candy_module_exit(void)
{
	driver_unregister(&nec_candy_driver);
}

module_init(nec_candy_module_init);
module_exit(nec_candy_module_exit);

#if defined(MODULE)
MODULE_AUTHOR("Jun Sun, jsun@mvista.com or jsun@junsun.net");
MODULE_DESCRIPTION("Ether driver for NEC Candy controller");
MODULE_LICENSE("GPL");
#endif
