/* 2009-08-24: File added and changed by Sony Corporation */
/*
 *  Copyright 2009 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/platform_device.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/freezer.h>
#include <linux/err.h>
#include <linux/param.h>
#include <asm/system.h>

#include <asm/arch/hardware.h>
#include <asm/arch/pwc.h>
#include <asm/arch/pm.h>
#include <asm/arch/pgpf/board.h>
#include <asm/arch/navufb.h>
#include "internal.h"
#include "controller.h"
#include "chargerif.h"
#include "suspend.h"
#include "event.h"
#include "chrdv.h"
#include "navu2pm_irq.h"

#define DC_IN_NUM    2
#define USB_IN_NUM   2

static int  controller_main_thread(void *arg);
static int  suspend_thread(void *arg);
static int  sensor_thread(void *arg);
static int  charger_led_on(void);
static int  charger_led_off(void);
static int  charger_led_red(void);
static int  charger_led_green(void);
static int  get_sensor_status(void);
static int  check_battery_full_to_normal(void);
static int  check_battery_normal_to_full(void);
static int  check_charge_timeout(void);
static void get_connection_status(void);
static void enter_charge_status(navu2pm_charge_status_t charge_status);
static void charge_start(void);
static void charge_stop(void);
static void discharge_start(void);
static void discharge_stop(void);
static void charge_control(struct work_struct *work);
static void charge_state_trans(void);
static void mode_trans(void);
static void enter_mode(navu2pm_mode_t mode);
static void led_control(void);
static void register_callback(void);
static void check_supply_low(void);
static void sampling_past_battery_date(void);
static void clear_past_battery_date(void);
static void battery_state_trans(void);
static void cb_dc_plug_in_out(void *arg);
static void cb_usb_plug_in_out(void *arg);
static void cb_pk_press(void *arg);
static void cb_high_temp(void *arg);
static void cb_rtc_det(void *arg);
static void cb_dc_usb_plug_out(void *arg);
static void init_pwc_xacc_in(void);
static int  lcd_back_light_on(int brightness, int tpled_mode);
static int  lcd_back_light_off(void);

static inline void init_plugin_matrix(void);
static inline void init_controller_status(void);
static inline int is_therm_charge_to_monitor(void);
static inline int is_therm_monitor_to_charge(void);
static inline int is_therm_volt_discharge_to_monitor(void);
static inline int is_supply_volt_low(void);
static inline int is_cond_normal_pseudo_suspend_to_dsm(void);
static inline int is_cond_acc_wake(void);
static inline int is_cradle_on(void);
static inline int is_det_high_temp(void);
static inline int is_charging(void);

static int suspend_ready, resume_ready;
static controller_status_t controller_status;
static struct task_struct *main_task;
static struct task_struct *suspend_task;
static struct task_struct *sensor_task;

static unsigned int t_charge_max_h;
static unsigned int t_charge_max_l;
static unsigned int t_charge_min_h;
static unsigned int t_charge_min_l;
static unsigned int t_discharge_min_h;

static DECLARE_WAIT_QUEUE_HEAD(navu2pm_suspend_it);
static DECLARE_WAIT_QUEUE_HEAD(navu2pm_resume_it);

static DECLARE_MUTEX(SuspendDetSem);
static DECLARE_MUTEX(PKPressDetSem);
static DECLARE_MUTEX(RTCDetSem);
static DECLARE_MUTEX(JudgePKPressSem);
static DECLARE_MUTEX(UsbSuspendSem);
static DECLARE_MUTEX(UsbDetSem);
static DECLARE_MUTEX(DcDetSem);

static DECLARE_WORK(charge_control_ws, charge_control);
static struct delayed_work charge_control_delay_ws;        

#ifdef CONFIG_PGPF_NAVU2_PM_APPTHERM
static int therm_fake_start;
#endif /* CONFIG_PGPF_NAVU2_PM_APPTHERM */

static plugin_status_t plugin_matrix[DC_IN_NUM][USB_IN_NUM];

static navu2pm_callback_t callback[NAVU2PM_IRQ_CB_NUM_MAX] = {
	{NAVU2PM_IRQ_CB_DC_DET,         cb_dc_plug_in_out,  (void*)NULL},
	{NAVU2PM_IRQ_CB_USB_DET,        cb_usb_plug_in_out, (void*)NULL},
	{NAVU2PM_IRQ_CB_DC_USB_OUT_DET, cb_dc_usb_plug_out, (void*)NULL},
	{NAVU2PM_IRQ_CB_POWER_KEY,      cb_pk_press,        (void*)NULL},
	{NAVU2PM_IRQ_CB_HIGH_TEMP,      cb_high_temp,       (void*)NULL},
	{NAVU2PM_IRQ_CB_RTC_DET,        cb_rtc_det,         (void*)NULL},
};

static int controller_main_thread(void *arg)
{
	while(1){
		schedule_work(&charge_control_ws);
		msleep(MAIN_THREAD_SLEEP_TIME);
	}

	return 0;
}

static int suspend_thread(void *arg)
{
	while(1){

		init_waitqueue_head(&navu2pm_suspend_it);
		suspend_ready = TRUE;
		wait_event_freezable(navu2pm_suspend_it, suspend_ready == FALSE); 

		charge_stop();
		charger_led_off();

		if(lcd_back_light_off()){
			DPRINT("## Error mp200_fb_suspend\n");
		}

		navu2pm_suspend();
	}

	return 0;
}

static int sensor_thread(void *arg)
{
	int ret;

	while(1){
		ret = get_sensor_status();
		if(ret < 0){
			DPRINT("## Error get_sensor_status\n");
			return -1;
		}

		msleep(SENSOR_THREAD_SLEEP_TIME);
	}

	return 0;
}

static int get_sensor_status(void)
{
	int ret;

	ret = chargerif_get_battery_current(&(controller_status.bc), &(controller_status.bc_hex));
	if(NAVU2PM_SUCCESS != ret){
		DPRINT("## error  get_battery_current\n");
		return -1;
	}

	ret = chargerif_get_battery_voltage(&(controller_status.bv), &(controller_status.bv_hex));
	if(NAVU2PM_SUCCESS != ret){
		DPRINT("## error  get_battery_voltage\n");
		return -1;
	}

	ret = chargerif_get_supply_voltage(&(controller_status.sv), &(controller_status.sv_hex));
	if(NAVU2PM_SUCCESS != ret){
		DPRINT("## error  get_battery_voltage\n");
		return -1;
	}

#ifdef CONFIG_PGPF_NAVU2_PM_APPTHERM
	if(!therm_fake_start){
		ret = chargerif_get_battery_temperature(&(controller_status.bt), &(controller_status.bt_hex));
		if(NAVU2PM_SUCCESS != ret){
			DPRINT("## error  get_battery_temperature\n");
			return -1;
		}
	}
#else  /* CONFIG_PGPF_NAVU2_PM_APPTHERM */
	ret = chargerif_get_battery_temperature(&(controller_status.bt), &(controller_status.bt_hex));
	if(NAVU2PM_SUCCESS != ret){
		DPRINT("## error  get_battery_temperature\n");
		return -1;
	}
#endif /* CONFIG_PGPF_NAVU2_PM_APPTHERM */

	/* Battery State Trans*/
	battery_state_trans();

	/* Check Supply Voltage is Low */
	check_supply_low();

	return 0;
}

static int charger_led_on(void)
{
	navu2pm_status_t ret;

	if(NAVU2PM_BATTERY_FULL == controller_status.battery_status){
		ret = charger_led_green();
		if(NAVU2PM_SUCCESS != ret){
			return -1;
		}
	}else{
		ret = charger_led_red();
		if(NAVU2PM_SUCCESS != ret){
			return -1;
		}
	}

	return 0;
}

static int charger_led_red(void)
{
	navu2pm_status_t ret;

	ret = chargerif_set_charger_led_green(FALSE);
	if(NAVU2PM_SUCCESS != ret){
		return -1;
	}

	ret = chargerif_set_charger_led_red(TRUE);
	if(NAVU2PM_SUCCESS != ret){
		return -1;
	}

	return 0;
}

static int charger_led_green(void)
{
	navu2pm_status_t ret;

	ret = chargerif_set_charger_led_red(FALSE);
	if(NAVU2PM_SUCCESS != ret){
		return -1;
	}

	ret = chargerif_set_charger_led_green(TRUE);
	if(NAVU2PM_SUCCESS != ret){
		return -1;
	}

	return 0;
}

static int charger_led_off(void)
{
	navu2pm_status_t ret;

	ret = chargerif_set_charger_led_red(FALSE);
	if(NAVU2PM_SUCCESS != ret){
		return -1;
	}

	ret = chargerif_set_charger_led_green(FALSE);
	if(NAVU2PM_SUCCESS != ret){
		return -1;
	}

	return 0;
}

static int  lcd_back_light_on(int brightness, int tpled_mode)
{
	int ret;

	ret = mp200_fb_resume(NULL);
	if(ret){
		return ret;
	}

	ret = mp200_fb_set_brightness(brightness);
	if(ret){
		DPRINT("## mp200_fb_set_brightness(%d)\n", brightness);
	}

	ret = mp200_fb_set_tpled(tpled_mode);
	if(ret){
		DPRINT("## mp200_fb_set_tpled(%d)\n", tpled_mode);
	}

	return ret; 
}  

static int lcd_back_light_off(void)
{
	int ret ;

	ret = mp200_fb_set_tpled(SONY_NAVU_TPLED_OFF);
	if(ret){
		DPRINT("## mp200_fb_set_tpled(%d)\n", SONY_NAVU_TPLED_OFF);
	}

	ret = mp200_fb_set_brightness(0);
	if(ret){
		DPRINT("## mp200_fb_set_brightness(%d)\n", 0);
	}
	
	return mp200_fb_suspend(NULL, PMSG_SUSPEND);
}  

static void get_connection_status(void)
{
	int dc, usb;

	dc  = controller_get_dc_connect();
	usb = controller_get_usb_connect();

	controller_status.connect.dc = dc;
	controller_status.connect.usb = usb;
	controller_status.connect.cradle = is_cradle_on();
	controller_status.plugin = plugin_matrix[dc][usb];

	return ;
}

static void charge_control(struct work_struct *work)
{
	get_connection_status();

	/* Mode Transition Executed */
 	mode_trans(); 

	/* Charge Status Transition Executed */
	charge_state_trans();

	/* Charge LED On/Off */
	led_control();

	return;
}

static void led_control(void)
{
	switch(controller_status.charge_status){
	case NAVU2PM_CHARGE_STAT_CHARGE :

		if(is_charging() || 
		   NAVU2PM_BATTERY_FULL    == controller_status.battery_status ||
		   NAVU2PM_BATTERY_TIMEOUT == controller_status.battery_status ){
			charger_led_on();
		}else{
			charger_led_off();
		}

		break;

	case NAVU2PM_CHARGE_STAT_MONITOR :

		if(PLUGIN_NONE != controller_status.plugin){
			charger_led_red();
		}else{
			charger_led_off();
		}

		break;

	default :
		charger_led_off();

		break;
	}

	return ;
}

static void mode_trans(void)
{
	down_interruptible(&SuspendDetSem);
	down_interruptible(&PKPressDetSem);
	down_interruptible(&RTCDetSem);
	down_interruptible(&DcDetSem);
	down_interruptible(&UsbDetSem);
	down_interruptible(&UsbSuspendSem);

	switch(controller_status.mode){
	case NAVU2PM_MODE_NORMAL_RUN:
		{
			switch(controller_status.plugin){
			case PLUGIN_USB_ONLY:
				enter_mode(NAVU2PM_MODE_USB);
				break;

			case PLUGIN_DC_ONLY:
				if(controller_status.det_suspend){
					enter_mode(NAVU2PM_MODE_PSEUDO_SUSPEND);
				}
				break;

			case PLUGIN_DC_USB:
				if(controller_status.det_suspend){
					controller_status.usb_suspend = TRUE;
					enter_mode(NAVU2PM_MODE_PSEUDO_SUSPEND);
				}
				break;

			case PLUGIN_NONE:
				if(controller_status.det_suspend){
					if(is_cond_normal_pseudo_suspend_to_dsm()){
						enter_mode(NAVU2PM_MODE_DSM);
					}else{
						enter_mode(NAVU2PM_MODE_PSEUDO_SUSPEND);
					}
				}
				break;

			default:
				break;
			}
		}
		break;

	case NAVU2PM_MODE_USB:

		if(PLUGIN_USB_ONLY != controller_status.plugin){
			enter_mode(NAVU2PM_MODE_NORMAL_RUN);
		}else if(controller_status.det_suspend){
			if(is_cond_normal_pseudo_suspend_to_dsm()){
				enter_mode(NAVU2PM_MODE_DSM);
			}else{
				controller_status.usb_suspend = TRUE;
				enter_mode(NAVU2PM_MODE_PSEUDO_SUSPEND);
			}
		}

		break;

	case NAVU2PM_MODE_PSEUDO_SUSPEND:
		{
			switch(controller_status.plugin){
			case PLUGIN_DC_ONLY:

				if(controller_status.det_pkpress){
					controller_status.factor = NAVU2_RESUME_FACTOR_PK;

					enter_mode(NAVU2PM_MODE_NORMAL_RUN);

				}else if(controller_status.det_rtc){
					controller_status.factor = NAVU2_RESUME_FACTOR_RTC;

					enter_mode(NAVU2PM_MODE_NORMAL_RUN);

				}else if(controller_status.det_dc && is_cond_acc_wake()){

					enter_mode(NAVU2PM_MODE_NORMAL_RUN);
				}

				break;

			case PLUGIN_USB_ONLY:

				if(controller_status.det_usb){
					controller_status.factor      = NAVU2_RESUME_FACTOR_DETUSB;
					controller_status.usb_suspend = FALSE;

					enter_mode(NAVU2PM_MODE_USB);

				}else if(controller_status.det_pkpress){
					controller_status.factor      = NAVU2_RESUME_FACTOR_PK;
					controller_status.usb_suspend = FALSE;

					enter_mode(NAVU2PM_MODE_USB);

				}else if(controller_status.det_rtc){
					controller_status.factor      = NAVU2_RESUME_FACTOR_RTC;
					controller_status.usb_suspend = FALSE;

					enter_mode(NAVU2PM_MODE_USB);

				}else if(TRUE == controller_status.usb_suspend){

					if(is_cond_normal_pseudo_suspend_to_dsm()){
						controller_status.usb_suspend = FALSE;
						enter_mode(NAVU2PM_MODE_DSM);
					}
				}

				break;

			case PLUGIN_DC_USB:

				if(controller_status.det_usb){
					controller_status.factor      = NAVU2_RESUME_FACTOR_DETUSB;
					controller_status.usb_suspend = FALSE;

					enter_mode(NAVU2PM_MODE_NORMAL_RUN);

				}else if(controller_status.det_dc && is_cond_acc_wake()){
					controller_status.factor      = NAVU2_RESUME_FACTOR_DETADP;
					controller_status.usb_suspend = FALSE;

					enter_mode(NAVU2PM_MODE_NORMAL_RUN);

				}else if(controller_status.det_pkpress){
					controller_status.factor      = NAVU2_RESUME_FACTOR_PK;
					controller_status.usb_suspend = FALSE;

					enter_mode(NAVU2PM_MODE_NORMAL_RUN);

				}else if(controller_status.det_rtc){
					controller_status.factor      = NAVU2_RESUME_FACTOR_RTC;
					controller_status.usb_suspend = FALSE;

					enter_mode(NAVU2PM_MODE_NORMAL_RUN);
				}

				break;

			case PLUGIN_NONE:

				if(controller_status.det_pkpress){
					controller_status.factor = NAVU2_RESUME_FACTOR_PK;
					enter_mode(NAVU2PM_MODE_NORMAL_RUN);
				}else if( controller_status.det_rtc){
					controller_status.factor = NAVU2_RESUME_FACTOR_RTC;
					enter_mode(NAVU2PM_MODE_NORMAL_RUN);
				}else if(is_cond_normal_pseudo_suspend_to_dsm()){
					enter_mode(NAVU2PM_MODE_DSM);
				}

				break;

			default:
				break;
			}
		}
		break;

	case NAVU2PM_MODE_DSM:
		{
			if(controller_status.dsm){
				break;
			}

			switch(controller_status.plugin){
			case PLUGIN_DC_ONLY:

				if(is_cond_acc_wake()){
					enter_mode(NAVU2PM_MODE_NORMAL_RUN);
				}else{
					enter_mode(NAVU2PM_MODE_PSEUDO_SUSPEND);
				}

				break;

			case PLUGIN_DC_USB:
				enter_mode(NAVU2PM_MODE_NORMAL_RUN);
				break;

			case PLUGIN_USB_ONLY:
				enter_mode(NAVU2PM_MODE_USB);
				break;

			case PLUGIN_NONE:

				if(is_det_high_temp()){
					enter_mode(NAVU2PM_MODE_PSEUDO_SUSPEND);
				}else if(is_supply_volt_low()){
					enter_mode(NAVU2PM_MODE_DSM);
				}else{
					enter_mode(NAVU2PM_MODE_NORMAL_RUN);
				}

				break;

			default:
				break;
			}
		}
		break;

	default:
		break;
	}
	
	controller_status.det_suspend = FALSE;
	up(&SuspendDetSem);

	controller_status.det_pkpress = FALSE;
	up(&PKPressDetSem);

	controller_status.det_rtc     = FALSE;
	up(&RTCDetSem);

	controller_status.det_dc      = FALSE;
	up(&DcDetSem);

	controller_status.det_usb     = FALSE;
	up(&UsbDetSem);

	up(&UsbSuspendSem);

	return;
}

static void charge_state_trans(void)
{
	switch(controller_status.charge_status){
	case NAVU2PM_CHARGE_STAT_BATTERY:

		if(NAVU2PM_MODE_DSM == controller_status.mode){
			enter_charge_status(NAVU2PM_CHARGE_STAT_NOSUPPLY);
		}else if(is_therm_charge_to_monitor()){
			enter_charge_status(NAVU2PM_CHARGE_STAT_MONITOR);
		}else if(PLUGIN_NONE != controller_status.plugin){
			enter_charge_status(NAVU2PM_CHARGE_STAT_CHARGE);
		}else{
			enter_charge_status(NAVU2PM_CHARGE_STAT_BATTERY);
		}

		break;

	case NAVU2PM_CHARGE_STAT_CHARGE:

		if(NAVU2PM_MODE_DSM == controller_status.mode){
			enter_charge_status(NAVU2PM_CHARGE_STAT_NOSUPPLY);
		}else if(is_therm_charge_to_monitor()){
			enter_charge_status(NAVU2PM_CHARGE_STAT_MONITOR);
		}else if(PLUGIN_NONE == controller_status.plugin){
			enter_charge_status(NAVU2PM_CHARGE_STAT_BATTERY);
		}else{
			enter_charge_status(NAVU2PM_CHARGE_STAT_CHARGE);
		}

		break;

	case NAVU2PM_CHARGE_STAT_MONITOR:

		if(is_det_high_temp()){
			enter_charge_status(NAVU2PM_CHARGE_STAT_DISCHARGE);
		}else if(NAVU2PM_MODE_DSM == controller_status.mode){
			enter_charge_status(NAVU2PM_CHARGE_STAT_NOSUPPLY);
		}else if(is_therm_monitor_to_charge()){
			if(PLUGIN_NONE == controller_status.plugin){
				enter_charge_status(NAVU2PM_CHARGE_STAT_BATTERY);
			}else{
				enter_charge_status(NAVU2PM_CHARGE_STAT_CHARGE);
			}
		}else{
			enter_charge_status(NAVU2PM_CHARGE_STAT_MONITOR);
		}
			
		break;

	case NAVU2PM_CHARGE_STAT_DISCHARGE:

		if(is_therm_volt_discharge_to_monitor()){
			enter_charge_status(NAVU2PM_CHARGE_STAT_MONITOR);
		}else{
			enter_charge_status(NAVU2PM_CHARGE_STAT_DISCHARGE);
		}

		break;

	case NAVU2PM_CHARGE_STAT_NOSUPPLY:

		if(is_det_high_temp()){
			enter_charge_status(NAVU2PM_CHARGE_STAT_DISCHARGE);
		}else if(FALSE == controller_status.dsm){
			if(PLUGIN_NONE == controller_status.plugin){
				enter_charge_status(NAVU2PM_CHARGE_STAT_BATTERY);
			}else{
				if(TRUE == is_therm_charge_to_monitor()){
					enter_charge_status(NAVU2PM_CHARGE_STAT_MONITOR);
				}else{
					enter_charge_status(NAVU2PM_CHARGE_STAT_CHARGE);
				}
			}
		}else{
			enter_charge_status(NAVU2PM_CHARGE_STAT_NOSUPPLY);
		}

		break;

	default:

		break;		
	}

	return;
}

static void enter_charge_status(navu2pm_charge_status_t charge_status)
{
	controller_status.charge_status = charge_status;

	switch(charge_status){
	case NAVU2PM_CHARGE_STAT_CHARGE:
		charge_start();
		break;

	case NAVU2PM_CHARGE_STAT_BATTERY:
		charge_stop();
		break;

	case NAVU2PM_CHARGE_STAT_MONITOR:
	case NAVU2PM_CHARGE_STAT_NOSUPPLY:
		charge_stop();
		discharge_stop();
		break;

	case NAVU2PM_CHARGE_STAT_DISCHARGE:
		charge_stop();
		discharge_start();
		break;
		
	default :
		break;
	}		

	return ;
}

static void enter_mode(navu2pm_mode_t mode)
{
	controller_status.mode = mode;

	switch(mode){
	case NAVU2PM_MODE_NORMAL_RUN:
	case NAVU2PM_MODE_USB:

 		if(resume_ready){

			if(lcd_back_light_on(0, SONY_NAVU_TPLED_AUTO)){
				DPRINT("## Error lcd_back_light_on\n");
			}

			resume_ready = FALSE;
			wake_up(&navu2pm_resume_it);
		}

		break;

	case NAVU2PM_MODE_PSEUDO_SUSPEND:

		if(lcd_back_light_off()){
			DPRINT("## Error lcd_back_light_off\n");
		}

		break;

	case NAVU2PM_MODE_DSM:

		if(suspend_ready){
			suspend_ready = FALSE;
			wake_up(&navu2pm_suspend_it);
			controller_status.dsm = TRUE;
		}

		break;

	default:
		break;
	}

	return ;	
}

static void charge_start(void)
{
	switch(controller_status.battery_status){
	case NAVU2PM_BATTERY_NORMAL:
		{
			switch(controller_status.plugin){
			case PLUGIN_DC_USB:

				chargerif_set_ichrg(ATLAS_ICHRG_443);
				break;

			case PLUGIN_DC_ONLY:

				if(NAVU2PM_MODE_NORMAL_RUN == controller_status.mode){
					chargerif_set_ichrg(ATLAS_ICHRG_443);
				}else{
					chargerif_set_ichrg(ATLAS_ICHRG_709);
				}

				break;

			case PLUGIN_USB_ONLY:
				chargerif_set_ichrg(controller_status.usb.ichrg);
				break;

			default:
				chargerif_set_ichrg(ATLAS_ICHRG_0);
				break;
			}
		}
		break;

	default:
		chargerif_set_ichrg(ATLAS_ICHRG_0);
		break;
	}

	return ;
}

static void charge_stop(void)
{
	chargerif_set_ichrg(ATLAS_ICHRG_0);
	return;
}

static void register_callback(void)
{
	int count;

	for(count = 0; count < NAVU2PM_IRQ_CB_NUM_MAX; count++){
		navu2pm_register_callback(callback[count].irq, 
								  callback[count].func, 
								  callback[count].arg);
	}

	return ;
}

static void battery_state_trans(void)
{
	sampling_past_battery_date();

	switch(controller_status.battery_status){
	case NAVU2PM_BATTERY_NORMAL:

		if(check_charge_timeout()){
			controller_status.battery_status = NAVU2PM_BATTERY_TIMEOUT;
		}else if(check_battery_normal_to_full()){
			controller_status.battery_status = NAVU2PM_BATTERY_FULL;
		}

		break;

	case NAVU2PM_BATTERY_FULL:
		
		if(check_battery_full_to_normal()){
			controller_status.battery_status = NAVU2PM_BATTERY_NORMAL;
		}

		break;

	case NAVU2PM_BATTERY_TIMEOUT:

		if(NAVU2PM_CHARGE_STAT_CHARGE != controller_status.charge_status){
			controller_status.battery_status = NAVU2PM_BATTERY_NORMAL;
		}

		break;

	default:
		break;
	}

	return ;
}

static int check_battery_full_to_normal(void)
{
	int i;

	if(NAVU2PM_CHARGE_STAT_CHARGE != controller_status.charge_status){
		goto exit;
	}

	for(i = 0; i < SAMPLE_COUNT ; i++){
		if(BATT_V_CHARGE_RESTART < controller_status.history_bv[i]){
			return FALSE;
		}
	}

  exit:
	clear_past_battery_date();

	return TRUE;
}

static int check_battery_normal_to_full(void)
{
	int i;

	if(NAVU2PM_CHARGE_STAT_CHARGE != controller_status.charge_status){
		return FALSE;
	}

	for(i = 0; i < SAMPLE_COUNT ; i++){
		if(BATT_C_FULL > controller_status.history_bc[i]){
			return FALSE;
		}
	}

	clear_past_battery_date();

	return TRUE;
}

static int check_charge_timeout(void)
{
	if(NAVU2PM_CHARGE_STAT_CHARGE != controller_status.charge_status){
		controller_status.count_charge_time = 0; 
		return FALSE;
	}

	if(is_charging()){
		controller_status.count_charge_time++;
	}else{
		controller_status.count_charge_time = 0;
	}

	if(COUNT_STOP_CHARGE == controller_status.count_charge_time){
		controller_status.count_charge_time = 0; 
		return TRUE;
	}

	return FALSE;
}

static void check_supply_low(void)
{
	if(is_supply_volt_low()){
		controller_status.supply_low = TRUE;
		complete_wait_irq(NAVU2PM_IRQ_LOW_BATTERY);	
	}else{
		controller_status.supply_low = FALSE;
	}

	return ;
}

static void sampling_past_battery_date(void)
{
	int i;

	if(NAVU2PM_CHARGE_STAT_CHARGE != controller_status.charge_status){
		clear_past_battery_date();
		return ;
	}

	for(i = 0; i < SAMPLE_COUNT - 1 ; i++){
		controller_status.history_bc[i] = controller_status.history_bc[i + 1];
		controller_status.history_bv[i] = controller_status.history_bv[i + 1];
	}
	
	controller_status.history_bc[SAMPLE_COUNT - 1] = controller_status.bc_hex;
	controller_status.history_bv[SAMPLE_COUNT - 1] = controller_status.bv_hex;

	return ;
}

static void clear_past_battery_date(void)
{
	int i;

	for(i = 0; i < SAMPLE_COUNT ; i++){
		controller_status.history_bc[i] = 0x0;
		controller_status.history_bv[i] = 0xffff;
	}

	controller_status.count_charge_time = 0;

	return ;
}

static void discharge_start(void)
{
	gpio_set_value(PGPF_GPIO_USE_BATT, 1);

	if(NAVU2PM_MODE_PSEUDO_SUSPEND == controller_status.mode){
		if(lcd_back_light_on(100, SONY_NAVU_TPLED_OFF)){
			DPRINT("## Error lcd_back_light_on()\n");
		}
	}

	return ;
}

static void discharge_stop(void)
{
	gpio_set_value(PGPF_GPIO_USE_BATT, 0);

	if(NAVU2PM_MODE_PSEUDO_SUSPEND == controller_status.mode){
		if(lcd_back_light_off()){
			DPRINT("## Error lcd_back_light_off()\n");
		}
	}

	return ;
}

static void	init_pwc_xacc_in()
{
	if(PGPF_PRODUCTION_LEVEL_RS <= pgpf_get_production_level()){
		/* GPIO08 pull down */
		pwc_write(PWC_GPPUPD3, 0x02, 0x03);

		/* GPIO08 IN select */
		pwc_write(PWC_IOSEL2, 0x00, 0x01);
	}

	return ;
}

static inline int is_therm_charge_to_monitor(void)
{
	if(controller_status.bt_hex > t_charge_max_h || 
	   controller_status.bt_hex < t_charge_min_l){
		return TRUE;
	}

	return FALSE;
}

static inline int is_therm_monitor_to_charge(void)
{
	if(controller_status.bt_hex < t_charge_max_l && 
	   controller_status.bt_hex > t_charge_min_h){
		return TRUE;
	}

	return FALSE;
}

static inline int is_therm_volt_discharge_to_monitor(void)
{
	if(controller_status.bt_hex > t_discharge_min_h || 
	   controller_status.bv_hex < BATT_V_DISCHARGE_L){
		return TRUE;
	}

	return FALSE;
}

static inline int is_supply_volt_low(void)
{
	if(controller_status.sv_hex < BATT_V_LOW){
		return TRUE;
	}

	return FALSE;
}

static inline int is_cond_normal_pseudo_suspend_to_dsm(void)
{
	if(!is_det_high_temp() && 
	   NAVU2PM_CHARGE_STAT_DISCHARGE != controller_status.charge_status){
		return TRUE;
	}
	
	return FALSE;
}

static inline int is_cond_acc_wake(void)
{
	if(controller_status.connect.cradle && 
	   TRUE == controller_status.acc_wake){
		return TRUE;
	} 

	return FALSE;
}

static inline int is_cradle_on(void)
{
	if(!gpio_get_value(PGPF_GPIO_CRADLE_XDET)){
		return TRUE;
	}

	return FALSE;
}

static inline int is_det_high_temp(void)
{
	if(!gpio_get_value(PGPF_GPIO_HIGHTEMP_XDET)){
		return TRUE;
	}

	return FALSE;
}

static inline int is_charging(void)
{
	unsigned char val;

	chargerif_get_ichrg(&val);
	if(ATLAS_ICHRG_0 != val){
		return TRUE;
	}
	
	return FALSE;
}

static inline void init_plugin_matrix(void)
{
	plugin_matrix[FALSE][FALSE] = PLUGIN_NONE;
	plugin_matrix[TRUE ][FALSE] = PLUGIN_DC_ONLY;
	plugin_matrix[FALSE][TRUE ] = PLUGIN_USB_ONLY;
	plugin_matrix[TRUE ][TRUE ] = PLUGIN_DC_USB;

	return ;
}

static inline void init_controller_status(void)
{
	if(PGPF_PRODUCTION_LEVEL_RS > pgpf_get_production_level()){
		t_charge_max_h    = THERMO_CHARGE_MAX_H_0;
		t_charge_max_l    = THERMO_CHARGE_MAX_L_0;
		t_charge_min_h    = THERMO_CHARGE_MIN_H_0;
		t_charge_min_l    = THERMO_CHARGE_MIN_L_0;
		t_discharge_min_h = THERMO_DISCHARGE_MIN_H_0;
	}else{
		t_charge_max_h    = THERMO_CHARGE_MAX_H_1;
		t_charge_max_l    = THERMO_CHARGE_MAX_L_1;
		t_charge_min_h    = THERMO_CHARGE_MIN_H_1;
		t_charge_min_l    = THERMO_CHARGE_MIN_L_1;
		t_discharge_min_h = THERMO_DISCHARGE_MIN_H_1;
	}

	controller_status.usb.flag       = FALSE;
	controller_status.usb.ichrg      = ATLAS_ICHRG_0;
	controller_status.mode           = NAVU2PM_MODE_NORMAL_RUN;
	controller_status.charge_status  = NAVU2PM_CHARGE_STAT_BATTERY;
	controller_status.battery_status = NAVU2PM_BATTERY_NORMAL;

	return ;
}

void controller_request_suspend(void)
{
	down_interruptible(&SuspendDetSem);
	controller_status.det_suspend = TRUE;
	up(&SuspendDetSem);

	schedule_work(&charge_control_ws);

	/* wait resume */
	init_waitqueue_head(&navu2pm_resume_it);
	resume_ready = TRUE;
	wait_event_freezable(navu2pm_resume_it, resume_ready == FALSE); 

	return ;
}

inline void controller_set_acc_wake(int acc_wake)
{
	controller_status.acc_wake = acc_wake;
	
	return ;
}

inline int controller_get_acc_wake(void)
{
	return controller_status.acc_wake;
}

inline int controller_judge_dc_in(void)
{
	return controller_status.connect.dc;
}

inline int controller_judge_dc_out(void)
{
	return ~(controller_status.connect.dc) & 0x1;
}

inline int controller_judge_usb_in(void)
{
	return controller_status.connect.usb;
}

inline int controller_judge_usb_out(void)
{
	return ~(controller_status.connect.usb) & 0x1;
}

inline int controller_judge_supply_low(void)
{
	return controller_status.supply_low;
}

int controller_judge_pk_press(void)
{
	int ret; 

	down_interruptible(&JudgePKPressSem);
	ret = controller_status.judge_pk_press;
	controller_status.judge_pk_press = FALSE;
	up(&JudgePKPressSem);

	return ret;	
}

navu2pm_status_t controller_init(void)
{
	int ret;

	INIT_DELAYED_WORK(&charge_control_delay_ws, charge_control);                    

	chargerif_init();
	init_plugin_matrix();
	init_pwc_xacc_in();
	init_controller_status();
	register_callback();

	ret = get_sensor_status();
	if(ret < 0){
		DPRINT("## Error get_sensor_status\n");
		return NAVU2PM_ERR_OTHER;
	}

	main_task = kthread_create(controller_main_thread, NULL, "NAVU2PM_controller");
	if(IS_ERR(main_task)){
		return NAVU2PM_ERR_THREAD;
	}
	wake_up_process(main_task);

	sensor_task = kthread_create(sensor_thread, NULL, "NAVU2PM_sensor");
	if(IS_ERR(sensor_task)){
		return NAVU2PM_ERR_THREAD;
	}
	wake_up_process(sensor_task);

	suspend_task = kthread_create(suspend_thread, NULL, "NAVU2PM_suspend");
	if(IS_ERR(suspend_task)){
		return NAVU2PM_ERR_THREAD;
	}
	wake_up_process(suspend_task);

	return NAVU2PM_SUCCESS;
}

navu2pm_status_t controller_resume(void)
{
	int ret;

	controller_status.factor = navu2_get_resume_factor();

	ret = get_sensor_status();
	if(ret < 0){
		DPRINT("## Error get_sensor_status\n");
		return NAVU2PM_ERR_OTHER;
	}

	controller_status.dsm = FALSE;

	/* Temporary Procedure */
	if(controller_status.factor & NAVU2_RESUME_FACTOR_PK){
		cb_pk_press(NULL);
	}

	if(controller_status.factor & NAVU2_RESUME_FACTOR_DETADP){
		cb_dc_plug_in_out(NULL);
	}

	if(controller_status.factor & NAVU2_RESUME_FACTOR_DETUSB){
		cb_usb_plug_in_out(NULL);
	}

	if(controller_status.factor & NAVU2_RESUME_FACTOR_RTC){
		cb_rtc_det(NULL);
	}

	if(controller_status.factor & NAVU2_RESUME_FACTOR_HITEMP){
		cb_high_temp(NULL);
	}


	return NAVU2PM_SUCCESS;
}

void cb_dc_plug_in_out(void *arg)
{
	controller_status.connect.dc = controller_get_dc_connect();

	if(controller_status.connect.dc){

		down_interruptible(&DcDetSem);
		controller_status.det_dc = TRUE;
		up(&DcDetSem);

		eventset_append(NAVU2PM_EVENT_DC_IN);
		complete_wait_irq(NAVU2PM_IRQ_DC_IN);

		schedule_delayed_work(&charge_control_delay_ws, HZ / 10);
	}else{
		eventset_append(NAVU2PM_EVENT_DC_OUT);
		complete_wait_irq(NAVU2PM_IRQ_DC_OUT);

		schedule_work(&charge_control_ws);
	}

	chrdv_notify();

	return ;
}

void cb_usb_plug_in_out(void *arg)
{
	controller_status.connect.usb = controller_get_usb_connect();
	if(controller_status.connect.usb){

		down_interruptible(&UsbDetSem);
		controller_status.det_usb = TRUE;
		up(&UsbDetSem);

		eventset_append(NAVU2PM_EVENT_USB_IN);
		complete_wait_irq(NAVU2PM_IRQ_USB_IN);
	}else{

		down_interruptible(&UsbSuspendSem);
		controller_status.usb_suspend = FALSE;
		up(&UsbSuspendSem);

		eventset_append(NAVU2PM_EVENT_USB_OUT);
		complete_wait_irq(NAVU2PM_IRQ_USB_OUT);

	}

	chrdv_notify();	

	schedule_work(&charge_control_ws);

	return ;
}

void cb_dc_usb_plug_out(void *arg)
{
	chargerif_set_ichrg(ATLAS_ICHRG_0);

	return ;
}

void cb_pk_press(void *arg)
{
	navu2pm_status_t ret; 

	eventset_append(NAVU2PM_EVENT_POWER_KEY);
	chrdv_notify();

	ret = complete_wait_irq(NAVU2PM_IRQ_POWER_KEY);
	if(NAVU2PM_ERR_NOT_READY == ret){
		down_interruptible(&JudgePKPressSem);
		controller_status.judge_pk_press = TRUE;
		up(&JudgePKPressSem);		
	}

	down_interruptible(&PKPressDetSem);
	controller_status.det_pkpress = TRUE;
	up(&PKPressDetSem);

	schedule_work(&charge_control_ws);

	return ;
}

void cb_high_temp(void *arg)
{
	if(!is_det_high_temp()){
		return ;
	}

	eventset_append(NAVU2PM_EVENT_HIGH_TEMP);
	chrdv_notify();

	schedule_work(&charge_control_ws);

	return ;
}

void cb_rtc_det(void *arg)
{
	complete_wait_irq(NAVU2PM_IRQ_RTC);

	down_interruptible(&RTCDetSem);
	controller_status.det_rtc = TRUE;
	up(&RTCDetSem);

	schedule_work(&charge_control_ws);

	return ;
}

int controller_get_dc_connect(void)
{
	int plugin;
	unsigned int val;

	if(PGPF_PRODUCTION_LEVEL_RS > pgpf_get_production_level()){
		chargerif_get_acpin(&plugin);

		return plugin;
	}else{
		pwc_read(PWC_IOIN2, &val);
		if(val & PWC_XACC_IN){
			return FALSE;
		}

		return TRUE;
	}
}

int controller_get_usb_connect(void)
{
	unsigned int val;
	int ret;

	pwc_read(PWC_PSW_STATE, &val);

	if(PWC_PSW_STATE_SDETUSB & val){
		ret = TRUE;
	}else{
		ret = FALSE;
	}

	return ret;
}

inline int controller_get_supply_low(void)
{
	return controller_status.supply_low;
}

inline navu2_factor_t controller_get_resume_factor(void)
{
	return controller_status.factor;
}

navu2pm_status_t controller_get_controller_status(controller_status_t *cs)
{
	if(NULL == cs){
		return NAVU2PM_ERR_ADDR;
	}

	cs->bv     = controller_status.bv;
	cs->bc     = controller_status.bc;
	cs->bt     = controller_status.bt;
	cs->bv_hex = controller_status.bv_hex;
	cs->bc_hex = controller_status.bc_hex;
	cs->bt_hex = controller_status.bt_hex;

	cs->connect.dc     = controller_status.connect.dc;
	cs->connect.cradle = controller_status.connect.cradle;
	cs->connect.usb    = controller_status.connect.usb;

	return NAVU2PM_SUCCESS;
}
EXPORT_SYMBOL(controller_get_controller_status);

unsigned short controller_get_battery_voltage(void)
{
	return controller_status.bv_hex;
}

unsigned short controller_get_supply_voltage(void)
{
	return controller_status.sv_hex;
}

unsigned short controller_get_battery_current(void)
{
	return controller_status.bc_hex;
}

unsigned short controller_get_battery_temperature(void)
{
	return controller_status.bt_hex;
}

navu2pm_charge_status_t controller_get_charge_status(void)
{
	return controller_status.charge_status;
}

navu2pm_battery_status_t controller_get_battery_status(void)
{
	return controller_status.battery_status;
}

navu2pm_mode_t controller_get_mode(void)
{
	return controller_status.mode;
}

navu2pm_status_t navu2pm_set_usb_enumration_done(int flag, int ilimit)
{
	controller_status.usb.flag = flag;

	if(flag){
		controller_status.usb.ichrg = chargerif_get_usb_ilimit_setting(ilimit);
	}else{
		controller_status.usb.ichrg = ATLAS_ICHRG_0;
	}

	return NAVU2PM_SUCCESS;
}

EXPORT_SYMBOL(navu2pm_set_usb_enumration_done);

#ifdef CONFIG_PGPF_NAVU2_PM_APPTHERM
void start_fake_therm(unsigned short val)
{
	therm_fake_start = TRUE;
	controller_status.bt_hex = val;

	return;
}

void stop_fake_therm(void)
{
	therm_fake_start = FALSE;

	return;
}

#endif /* CONFIG_PGPF_NAVU2_PM_APPTHERM */

