/* 2009-08-24: File added by Sony Corporation */
/*
 * pgpf_gpio.c - GPIO driver for PGPF
 *
 *     GPIO-IRQ handler and character device interface.
 *
 *  Copyright 2003-2009 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/list.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/mach/irq.h>
#include <asm/arch/gpio.h>
#include <asm/arch/smu.h>
#include <asm/arch/hardware.h>
#include <asm/arch/pgpf/devnum.h>


/****************************************************************************
 *                        Debug Print Macros
 ****************************************************************************/
#define GPIO_DEBUG

#ifdef GPIO_DEBUG
#define DPRINTK(format,...) \
    printk(KERN_DEBUG "[pgpf-gpio:%d:%s] " format "\n", \
           __LINE__, __FUNCTION__, ##__VA_ARGS__)
#else
#define DPRINTK(format,...) /*vanish*/
#endif

#define PRINTK(format, ...) \
    printk(KERN_INFO "gpio: " format "\n", ##__VA_ARGS__)

#define PRINTK_ERR(format, ...) \
    printk(KERN_ERR "gpio: " format "\n", ##__VA_ARGS__)

#define PRINTK_WARN(format, ...) \
    printk(KERN_WARNING "gpio: " format "\n", ##__VA_ARGS__)


/****************************************************************************
 *                           Device file
 ****************************************************************************/

static int DevMajorNumber = PGPF_GPIO_MAJOR;
static int DevMinorNumber = PGPF_GPIO_MINOR;
static int DevCount = 1;
static const char *DevName = "gpio";


/****************************************************************************
 *                          Global variable
 ****************************************************************************/

LIST_HEAD(IrqList);
DECLARE_MUTEX(IRQListSem);
static spinlock_t IRQListLock = SPIN_LOCK_UNLOCKED;
static unsigned long IRQListLockFlags;


/****************************************************************************
 *                        Typedef declaration
 ****************************************************************************/

typedef struct {
    struct list_head  list;
    unsigned long     port;
    unsigned int      type;
    int               irq_nr;
    wait_queue_head_t *wait_q;
    atomic_t          int_count;
    atomic_t          cancel_count; /* Used for canceling waiting threads */
    atomic_t          refcount;     /* if zero, you can remove this entry */
} pgpf_irq_entry;


/****************************************************************************
 *                        Prototype declaration
 ****************************************************************************/

static int irq_entry_type(pgpf_irq_ioctl_t* arg, unsigned int *type);
static void irq_entry_free(pgpf_irq_entry* entry);
static pgpf_irq_entry* irq_entry_alloc(void);
static int irq_entry_get(pgpf_irq_ioctl_t* arg, pgpf_irq_entry** ent_ptr);
static void irq_entry_release(unsigned long port);
static irqreturn_t gpio_int(int irq_nr, void *dev_id);


/****************************************************************************
 *                      IRQ Management Routines
 ****************************************************************************/

/**
 * irq_entry_type - Set irq entry type.
 * @arg GPIO port and trigger to register.
 * @type pointer to set irq type.
 */
static int
irq_entry_type(pgpf_irq_ioctl_t* arg, unsigned int *type)
{
    switch (arg->trigger) {
        case PGPF_GPIO_FALL_EDGE:
            *type = IRQT_FALLING;
            break;

        case PGPF_GPIO_RISE_EDGE:
            *type = IRQT_RISING;
            break;

        case PGPF_GPIO_LOW_LEVEL:
            *type = IRQT_LOW;
            break;

        case PGPF_GPIO_HIGH_LEVEL:
            *type = IRQT_HIGH;
            break;

        case PGPF_GPIO_DOUBLE_EDGE:
            *type = IRQT_BOTHEDGE;
            break;

        default:
            return -EINVAL;
    }

    return 0;
}

/**
 * irq_entry_free - free the memory region of pgpf_irq_entry*
 * @entry pointer to the memory region you want to free
 */
static void
irq_entry_free(pgpf_irq_entry* entry)
{
    if( entry != NULL ) {
        if( entry->wait_q != NULL ) {
            kfree( entry->wait_q );
            entry->wait_q = NULL;
        }
        kfree( entry );
        entry = NULL;
    }
    return;
}


/**
 * irq_entry_alloc - allocate a memory region for pgpf_irq_entry*
 */
static pgpf_irq_entry*
irq_entry_alloc(void)
{
    pgpf_irq_entry* entry;

    entry = kmalloc(sizeof(pgpf_irq_entry), GFP_KERNEL);
    if( entry == NULL ) {
        PRINTK_ERR("kmalloc() FAILED for pgpf_irq_entry");
        goto InitErr;
    }

    entry->wait_q = kmalloc(sizeof(wait_queue_head_t), GFP_KERNEL);
    if( entry->wait_q == NULL ) {
        PRINTK_ERR("kmalloc() FAILED for wait_queue_head_t ");
        goto InitErr;
    }

    return entry;

InitErr:
    irq_entry_free(entry);
    return NULL;
}


/**
 * irq_entry_get - Pick an entry for GPIO 'port' and register.
 * @arg GPIO port and trigger to register.
 */
static int 
irq_entry_get(pgpf_irq_ioctl_t* arg, pgpf_irq_entry** ent_ptr)
{
    int rc, irq_nr;
    pgpf_irq_entry* entry;
    unsigned long port;
    unsigned int type;

    port = arg->port;
    irq_nr  = gpio_to_irq(port);
    *ent_ptr = NULL;

    rc = irq_entry_type(arg, &type);
    if (rc < 0) {
        PRINTK_ERR("Invalid irq trigger type\n");
        return rc;
    }
    
    DPRINTK("Request for pgpf_irq_entry for irq_nr=%d", irq_nr);

    if ( down_interruptible(&IRQListSem) ) {
        return -EINTR;
    }

    list_for_each_entry(entry, &IrqList, list) {
        if ( entry->irq_nr == irq_nr ) {
            DPRINTK("Found already registered entry");
            if ( entry->type == type ) {
                atomic_inc(&entry->refcount);
                up(&IRQListSem);
                *ent_ptr = entry;
                return 0;
            }
            else {
                PRINTK_ERR("Irq type is different from registered entry\n");
                up(&IRQListSem);
                return -EINVAL;
            }
        }
    }

    /*
     * In case the port is new to us.
     */
    DPRINTK("No previously registered entry. calling irq_entry_alloc().");
    entry = irq_entry_alloc();
    if ( ! entry ) {
        up( &IRQListSem );
        return -ENOMEM;
    }

    init_waitqueue_head( entry->wait_q );
    entry->port = port;
    entry->type = type;
    entry->irq_nr = irq_nr;
    rc = set_irq_type(irq_nr, type);
    if ( rc < 0 ) {
        PRINTK_ERR("set_irq_type() FAILED (%d)", rc);
        up( &IRQListSem );
        irq_entry_free(entry);
        return rc;
    }

    rc = request_irq(irq_nr, gpio_int, 0, DevName, NULL);
    if ( rc < 0 ) {
        PRINTK_ERR("request_irq() FAILED (%d)", rc);
        up( &IRQListSem );
        irq_entry_free(entry);
        return rc;
    }

    atomic_set(&entry->refcount, 1);
    atomic_set(&entry->int_count, 0);
    atomic_set(&entry->cancel_count, 0);
    spin_lock_irqsave(&IRQListLock, IRQListLockFlags);
    list_add(&entry->list, &IrqList);
    spin_unlock_irqrestore(&IRQListLock, IRQListLockFlags);

    up( &IRQListSem );
    *ent_ptr = entry;
    return 0;
}


/**
 * irq_entry_release - Release a registered irq entry.
 * @port GPIO port to release.
 */
static void
irq_entry_release(unsigned long port)
{
    int irq_nr;
    pgpf_irq_entry* entry;
    pgpf_irq_entry* tmp;

    irq_nr = gpio_to_irq(port);
    DPRINTK("Release request for pgpf_irq_entry: irq_nr=%d", irq_nr);

    down(&IRQListSem);

    list_for_each_entry_safe(entry, tmp, &IrqList, list) {
        if ( entry->irq_nr == irq_nr ) {
            if ( atomic_dec_and_test(&entry->refcount) ) {
                DPRINTK("The irq entry no more refered (I'm the last one.)");
                free_irq( irq_nr, NULL );
                spin_lock_irqsave(&IRQListLock, IRQListLockFlags);
                list_del( &entry->list );
                spin_unlock_irqrestore(&IRQListLock, IRQListLockFlags);
                irq_entry_free( entry );
            }
            else {
                DPRINTK("The irq entry is still in use. Not released.");
            }
            up(&IRQListSem);
            return;
        }
    }

    PRINTK_ERR("Entry for IRQ#%d NOT FOUND "
               "get/release mismatch?", irq_nr);
    up(&IRQListSem);
    return;
}


/**
 * gpio_int - Interrupt handler for GPIO Interrupt
 */
static irqreturn_t
gpio_int(int irq_nr, void *dev_id)
{
    pgpf_irq_entry* entry;

    spin_lock_irqsave(&IRQListLock, IRQListLockFlags);
    list_for_each_entry(entry, &IrqList, list) {
        if ( entry->irq_nr == irq_nr ) {
            DPRINTK("Interrupt received for IRQ#%d", irq_nr );
            atomic_inc(&entry->int_count);
            disable_irq(irq_nr);
            wake_up_interruptible( entry->wait_q );
            spin_unlock_irqrestore(&IRQListLock, IRQListLockFlags);
            return IRQ_HANDLED;
        }
    }

    PRINTK_ERR("Unknown IRQ#%d received (ignored).", irq_nr);
    spin_unlock_irqrestore(&IRQListLock, IRQListLockFlags);
    return IRQ_HANDLED;
}


/**
 * pgpf_gpio_get_direction - Get gpio direction function.
 * @port GPIO port for get direction.
 */
static int
pgpf_gpio_get_direction(unsigned long port)
{
    int direction = 0; /* 0: input, 1: output */

    if (port <= GPIO_GPIO_LAST) {
        direction = __gpio_get_direction(port);
    } else if ((GPIO_EXT1_BASE <= port) && (port <= GPIO_EXT1_LAST)) {
        unsigned char addr, mask, base;
        if (port <= GPIO_EXT1_P7) {
            base = GPIO_EXT1_P0;
            mask = 1U << (port - base);
            addr = PWC_IOSEL1;
        } else {
            base = GPIO_EXT1_P8;
            mask = 1U << (port - base);
            addr = PWC_IOSEL2;
        }
        if (pwc_reg_read(addr, (unsigned char*)&direction)) {
            return -1;
        }
        direction = (direction & mask) >> (port - base);
    } else {
        return -1;
    }

    return direction;
}

/**
 * pgpf_gpio_get_level - IOCTL Backend function for get data.
 * @port GPIO port for get.
 */
int
pgpf_gpio_get_level(unsigned long port)
{
    return gpio_get_value(port);
}

/**
 * pgpf_gpio_set_level - IOCTL Backend function for set data.
 * @port  GPIO port for set.
 * @level Port level for set (0 or 1).
 */
int
pgpf_gpio_set_level(unsigned long port, int level)
{
    int stat = 0;

    if (pgpf_gpio_get_direction(port) != 1) {
        stat = -EINVAL;
    }

    if (stat == 0 && (level == 0 || level == 1))
    {
        switch (port) {
            case PGPF_GPIO_BT_PWRON:
                if (level) {
                    gpio_set_value(port, level);
                    msleep(1);
                    /* set BT_XRTS / BT_XCTS / BT_TXD / BT_RXD to UART */
                    __raw_writel(__raw_readl(CHG_PINSEL_G96) | 0x55000000,
                                 CHG_PINSEL_G96);
                } else {
                    /* set BT_XRTS / BT_XCTS / BT_TXD / BT_RXD to GPIO */
                    __raw_writel(__raw_readl(CHG_PINSEL_G96) & ~0x55000000,
                                 CHG_PINSEL_G96);
                    gpio_set_value(port, level);
                }
                break;

            default:
                gpio_set_value(port, level);
                break;
        }
    }
    else
    {
        stat = -EINVAL;
    }

    return stat;
}

/**
 * pgpf_gpio_irq_wait - IOCTL Backend function for IRQ request.
 * @arg GPIO port and trigger for wait.
 */
static int
pgpf_gpio_irq_wait(pgpf_irq_ioctl_t* arg)
{
    int rc, irq_nr, portlevel;
    pgpf_irq_entry* entry;
    unsigned long port;
    int int_count_before,  cancel_count_before;
    unsigned int trigger;

    port = arg->port;
    trigger = arg->trigger;
    irq_nr  = gpio_to_irq(port);
    portlevel = gpio_get_value(port);

    DPRINTK("Waiting for irq_nr=%d, trigger=0x%X", irq_nr, trigger);

    if( ( (trigger == PGPF_GPIO_LOW_LEVEL)  && (portlevel == 0) )
     || ( (trigger == PGPF_GPIO_HIGH_LEVEL) && (portlevel == 1) ) )
    {
        DPRINTK("Level trigger; returning before wait_event_interruptible().");
        return portlevel;
    }

    rc = irq_entry_get(arg, &entry);
    if( rc < 0 ) {
        return rc;
    }

    /*
     * The interrupt handler increments these values accourding to the
     * refering GPIO status just after an interrupt.
     * If we found these values
     */
    int_count_before    = atomic_read(&entry->int_count);
    cancel_count_before = atomic_read(&entry->cancel_count);

    switch (trigger) {
        case PGPF_GPIO_LOW_LEVEL:
        case PGPF_GPIO_FALL_EDGE:
        case PGPF_GPIO_HIGH_LEVEL:
        case PGPF_GPIO_RISE_EDGE:
        case PGPF_GPIO_DOUBLE_EDGE:
            rc = wait_event_interruptible(
                    *(entry->wait_q),
                    (  int_count_before < atomic_read(&entry->int_count)
                    || cancel_count_before < atomic_read(&entry->cancel_count) )
                 );
            break;

        default:
            PRINTK_ERR("Invalid trigger type %d", trigger);
            irq_entry_release(port);
            return -EINVAL;
    }

    irq_entry_release(port);

    if( rc != 0 ) {
        DPRINTK("IRQ wake awaken by signal.");
        return -ERESTARTSYS;
    }
    else
    if( cancel_count_before < atomic_read(&entry->cancel_count) ) {
        /* There might be a chance that &entry->cancel_count have been
         * incremented just after comparison in wait_event_interruptible()
         * and before the if() above. But we leave this window since
         * it won't break semantics.
         */
        DPRINTK("IRQ wait canceled.");
        return -ECANCELED;
    } else {
        return pgpf_gpio_get_level(port);
    }
}


/**
 * pgpf_gpio_irq_cancel - IOCTL Backend function for IRQ cancel.
 * @port port name for cancel.
 */
static int
pgpf_gpio_irq_cancel(unsigned long port)
{
    pgpf_irq_entry* entry;
    int irq_nr = gpio_to_irq(port);

    if ( down_interruptible(&IRQListSem) ) {
        return -EINTR;
    }
    list_for_each_entry(entry, &IrqList, list) {
        if ( entry->irq_nr == irq_nr ) {
            atomic_inc(&entry->cancel_count);
            wake_up_interruptible( entry->wait_q );
            up(&IRQListSem);
            return 0;
        }
    }
    DPRINTK("Attempted thread cancel but no waiting thread for IRQ#%d.",
            irq_nr);
    up(&IRQListSem);
    return 0;
}


/****************************************************************************
                     Kernel File Operations Interface
 ****************************************************************************/

static int gpio_ioctl(struct inode *inode, struct file *filp,
                      unsigned int cmd, unsigned long arg)
{
    int stat = 0;

    switch( cmd )
    {
        case PGPF_GPIO_GET_LEVEL:
            stat = pgpf_gpio_get_level(arg);
            break;
        case PGPF_GPIO_SET_HIGH:
            stat = pgpf_gpio_set_level(arg, 1);
            break;
        case PGPF_GPIO_SET_LOW:
            stat = pgpf_gpio_set_level(arg, 0);
            break;
        case PGPF_GPIO_IRQ_WAIT:
            {
                pgpf_irq_ioctl_t wait_arg;
                if (copy_from_user(&wait_arg, (pgpf_irq_ioctl_t*)arg, 
                                   sizeof(pgpf_irq_ioctl_t))) {
                        stat = -EFAULT;
                        break;
                }
                stat = pgpf_gpio_irq_wait(&wait_arg);
                break;
            }
        case PGPF_GPIO_IRQ_CANCEL:
            stat = pgpf_gpio_irq_cancel(arg);
            break;
        default:
            stat = -EINVAL;
    }
    return stat;
}


static struct file_operations gpio_fops = {
    owner:   THIS_MODULE,
    ioctl:   gpio_ioctl,
};


/****************************************************************************
                    Kernel Power Management Interface
 ****************************************************************************/

/*
 *  Power manager Support Routines
 */
static int __init gpio_probe(struct platform_device *dev)
{
    return 0;
}

static int gpio_remove(struct platform_device *dev)
{
    return 0;
}

static int gpio_suspend(struct platform_device *dev, pm_message_t state)
{
    return 0;
}

static int gpio_resume(struct platform_device *dev)
{
    return 0;
}


static struct platform_driver pgpf_gpio_driver = {
    .probe      = gpio_probe,
    .remove     = gpio_remove,
    .suspend    = gpio_suspend,
    .resume     = gpio_resume,
    .driver     = {
        .name   = "pgpf_gpio",
    },
};


/****************************************************************************
                    Kernel Init / Cleanup Routines
 ****************************************************************************/

static struct cdev gpio_cdev;
static int __init gpio_init( void )
{
    int result;
    int err;
    int devno;
    dev_t dev = 0;

    DPRINTK("PGPF GPIO driver started (%s %s)", __DATE__, __TIME__);

    /* Append this device driver to system. */
    if (DevMajorNumber) {
        dev = MKDEV(DevMajorNumber, DevMinorNumber);
        result = register_chrdev_region(dev, DevCount, DevName);
    }
    else {
        result = alloc_chrdev_region(&dev, DevMinorNumber, DevCount, DevName);
        DevMajorNumber = MAJOR(dev);
    }
    if (result < 0) {
        PRINTK_ERR("can't get major %d", DevMinorNumber);
        return result;
    }

    devno = MKDEV(DevMajorNumber, DevMinorNumber);
    cdev_init(&gpio_cdev, &gpio_fops);
    gpio_cdev.owner = THIS_MODULE;
    gpio_cdev.ops = &gpio_fops;

    err = cdev_add(&gpio_cdev, devno, DevCount);
    if( err ) {
        PRINTK_ERR("Error %d while cdev_add()", err);
    }

    result = platform_driver_register(&pgpf_gpio_driver);
    if( result != 0 ) {
        PRINTK_ERR("driver_register() failed. (%d)", result);
    }

    return result;
}


static void __exit gpio_cleanup(void)
{
    dev_t devno = MKDEV(DevMajorNumber, DevMinorNumber);

    if ( IrqList.prev != IrqList.next ) {
        PRINTK_ERR("Oops, IrqList is not empty!");
    }

    platform_driver_unregister(&pgpf_gpio_driver);

    cdev_del(&gpio_cdev);

    unregister_chrdev_region(devno, DevCount);
}


module_init(gpio_init);
module_exit(gpio_cleanup);

MODULE_AUTHOR("Sony");
MODULE_DESCRIPTION("GPIO-IRQ handler and character device interface for PGPF");
MODULE_LICENSE("GPL");
