/* 2008-08-12: File added and changed by Sony Corporation */
/*
 * OHCI HCD (Host Controller Driver) for USB.
 * Bus Glue for NEC EMMA3PF
 * Based on "ohci-lh7a404.c"
 *
 * Author: NEC Electronics Corporation
 *
 * Copyright (C) 2008 NEC Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/platform_device.h>
#include <asm/emma3.h>

extern int usb_disabled(void);

static void emma_pseudo_pci_init(struct platform_device *dev)
{
	struct resource *res;
	int		irq;
	u32		d;

	res = platform_get_resource_byname(dev, IORESOURCE_MEM, "ohci_mem");
	irq = platform_get_irq_byname(dev, "ohci_irq");

	/* PCI BAR */
	emma_out32(EMMA_OHCI_CONF_BASE + PCI_BASE_ADDRESS_0,
		   CPHYSADDR(res->start));

	/* PCI interrupt line */
	d = emma_in32(EMMA_OHCI_CONF_BASE + PCI_INTERRUPT_LINE);
	d &= ~0xff;
	d |= irq;
	emma_out32(EMMA_OHCI_CONF_BASE + PCI_INTERRUPT_LINE, d);

	/* PCI command status */
	emma_out32(EMMA_OHCI_CONF_BASE + PCI_COMMAND,
		   PCI_COMMAND_MASTER|PCI_COMMAND_MEMORY);
}

static void emma_start_hc(struct platform_device *dev)
{
	printk(KERN_DEBUG __FILE__
	       ": starting EMMA OHCI USB Controller\n");

	emma_pseudo_pci_init(dev);

	printk(KERN_DEBUG __FILE__
		   ": Clock to USB host has been enabled \n");
}

static void emma_stop_hc(struct platform_device *dev)
{
	printk(KERN_DEBUG __FILE__
	       ": stopping EMMA OHCI USB Controller\n");
}

/* configure so an HC device and id are always provided */
/* always called with process context; sleeping is OK */


/**
 * usb_hcd_emma_probe - initialize EMMA-based HCDs
 * Context: !in_interrupt()
 *
 * Allocates basic resources for this USB host controller, and
 * then invokes the start() method for the HCD associated with it
 * through the hotplug entry's driver_data.
 *
 */
int usb_hcd_emma_probe(const struct hc_driver *driver,
			  struct platform_device *dev)
{
	int retval;
	struct usb_hcd	*hcd = 0;
	struct resource *res;
	int		irq;

	res = platform_get_resource_byname(
			dev, IORESOURCE_MEM, "ohci_mem");
	if (!res) {
		pr_debug("Found EMMA_HC with no MEM. Check %s setup!\n",
			dev->dev.bus_id);
		retval = -ENODEV;
		goto err1;
	}

	irq = platform_get_irq_byname(dev, "ohci_irq");
	if (irq == -ENXIO) {
		pr_debug("Found EMMA_HC with no IRQ. Check %s setup!\n",
				dev->dev.bus_id);
		retval = -ENODEV;
		goto err1;
	}

	hcd = usb_create_hcd(driver, &dev->dev, "emma-ohci");
	if (!hcd)
		return -ENOMEM;
	hcd->rsrc_start = (u64)res->start;
	hcd->rsrc_len = (u64)(res->end - res->start + 1);

	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, hcd_name)) {
		pr_debug("request_mem_region failed");
		retval = -EBUSY;
		goto err1;
	}

	hcd->regs = (void __iomem *)ioremap(hcd->rsrc_start, hcd->rsrc_len);

	emma_start_hc(dev);

	retval = usb_add_hcd(hcd, irq, IRQF_SHARED);
	if (retval == 0)
		return retval;

	emma_stop_hc(dev);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
 err1:
	usb_put_hcd(hcd);
	return retval;
}

/* may be called without controller electrically present */
/* may be called with controller, bus, and devices active */

/**
 * usb_hcd_emma_remove - shutdown processing for EMMA-based HCDs
 * @dev: USB Host Controller being removed
 * Context: !in_interrupt()
 *
 * Reverses the effect of usb_hcd_emma_probe(), first invoking
 * the HCD's stop() method.  It is always called from a thread
 * context, normally "rmmod", "apmd", or something similar.
 *
 */
void usb_hcd_emma_remove(struct usb_hcd *hcd, struct platform_device *dev)
{
	usb_remove_hcd(hcd);
	emma_stop_hc(dev);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
	usb_put_hcd(hcd);
}

static int __devinit
ohci_emma_start(struct usb_hcd *hcd)
{
	struct ohci_hcd	*ohci = hcd_to_ohci(hcd);
	int		ret;

	ohci_dbg(ohci, "ohci_emma_start, ohci:%p", ohci);

	ret = ohci_run(ohci);
	if (ret < 0) {
		err("can't start %s", hcd->self.bus_name);
		ohci_stop(hcd);
		return ret;
	}
	return 0;
}

static int ohci_emma_reset (struct usb_hcd *hcd)
{
       struct ohci_hcd *ohci = hcd_to_ohci (hcd);

       ohci_hcd_init (ohci);
       return ohci_init (ohci);
}

#if defined(CONFIG_SNSC_USB_RH_PORT_POWER_CONTROL_WITH_COMPANION)
static int ohci_emma_match_companion(struct usb_hcd *hcd, struct usb_hcd *chcd)
{
	return (strcmp(chcd->product_desc, "EMMA EHCI"))? 0: 1;
}
#endif /* CONFIG_SNSC_USB_RH_PORT_POWER_CONTROL_WITH_COMPANION */

static const struct hc_driver ohci_emma_hc_driver = {
	.description =		hcd_name,
	.product_desc =		"EMMA OHCI",
	.hcd_priv_size =	sizeof(struct ohci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq =			ohci_irq,
	.flags =		HCD_USB11 | HCD_MEMORY,

	/*
	 * basic lifecycle operations
	 */
	.reset =                ohci_emma_reset,
	.start =		ohci_emma_start,
#ifdef	CONFIG_PM
	/* suspend:		ohci_lh7a404_suspend,  -- tbd */
	/* resume:		ohci_lh7a404_resume,   -- tbd */
#endif /*CONFIG_PM*/
	.stop =			ohci_stop,
	.shutdown = 		ohci_shutdown,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue =		ohci_urb_enqueue,
	.urb_dequeue =		ohci_urb_dequeue,
	.endpoint_disable =	ohci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number =	ohci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data =	ohci_hub_status_data,
	.hub_control =		ohci_hub_control,
#if defined(CONFIG_SNSC_USB_RH_PORT_POWER_CONTROL_WITH_COMPANION)
	.match_companion	= ohci_emma_match_companion,
	.hub_control_pp_by_companion =	ohci_hub_control_pp_by_companion,
#endif /* CONFIG_SNSC_USB_RH_PORT_POWER_CONTROL_WITH_COMPANION */
	.hub_irq_enable =	ohci_rhsc_enable,
#ifdef	CONFIG_PM
	.bus_suspend =		ohci_bus_suspend,
	.bus_resume =		ohci_bus_resume,
#endif
	.start_port_reset =	ohci_start_port_reset,
};

static int ohci_hcd_emma_drv_probe(struct platform_device *pdev)
{
	int ret;

	pr_debug("In ohci_hcd_emma_drv_probe");

	if (usb_disabled())
		return -ENODEV;

	ret = usb_hcd_emma_probe(&ohci_emma_hc_driver, pdev);
	return ret;
}

static int ohci_hcd_emma_drv_remove(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);

	usb_hcd_emma_remove(hcd, pdev);
	return 0;
}

#ifdef CONFIG_PM
static int ohci_hcd_emma_drv_suspend(struct platform_device *dev, pm_message_t message)
{
	struct usb_hcd *hcd = platform_get_drvdata(dev);
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);
	unsigned long	flags;
	int		rc = 0;

	spin_lock_irqsave (&ohci->lock, flags);
	if (hcd->state != HC_STATE_SUSPENDED) {
		rc = -EINVAL;
		goto bail;
	}

	dev->dev.power.power_state = PMSG_SUSPEND;

	ohci_writel(ohci, OHCI_INTR_MIE, &ohci->regs->intrdisable);
	(void)ohci_readl(ohci, &ohci->regs->intrdisable);

	clear_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);
bail:
	spin_unlock_irqrestore (&ohci->lock, flags);

	return rc;
}

static int ohci_hcd_emma_drv_resume(struct platform_device *dev)
{
	struct usb_hcd *hcd = platform_get_drvdata(dev);
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);
	int		port;

	emma_pseudo_pci_init(dev);

	set_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);
	dev->dev.power.power_state = PMSG_ON;

	/* Here we "know" root ports should always stay powered */
	ohci_dbg(ohci, "powerup ports\n");
	for (port = 0; port < ohci->num_ports; port++)
		ohci_writel(ohci, RH_PS_PPS,
			    &ohci->regs->roothub.portstatus[port]);

	/* Flush those writes */
	ohci_readl(ohci, &ohci->regs->control);
	msleep(20);

	return 0;
}
#else
#define ohci_hcd_emma_drv_suspend NULL
#define ohci_hcd_emma_drv_resume NULL
#endif

static struct platform_driver ohci_hcd_emma_driver = {
	.probe		= ohci_hcd_emma_drv_probe,
	.remove		= ohci_hcd_emma_drv_remove,
	.suspend	= ohci_hcd_emma_drv_suspend,
	.resume		= ohci_hcd_emma_drv_resume,
	.driver		= {
		.name	= "emma-ohci",
		.owner	= THIS_MODULE,
	},
};
