/* 2010-08-25: File added and changed by Sony Corporation */
/*
 * File Name       : /drivers/video/mp200/nts/mp200_nts.c
 * Function        : NTSC Driver corresponding to V4L2
 * Release Version : Ver 1.00
 * Release Date    : 2008/11/13
 *
 * Copyright (C) NEC Electronics Corporation 2008
 *
 *
 * This program is free software;you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by Free Softwere
 * Foundation; either version 2 of License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warrnty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; If not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


/********************************************************
 *  Include Files                                       *
 *******************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/video_encoder.h>

#if defined(CONFIG_PM) || defined(CONFIG_DPM)
#include <asm/arch/pm.h>
#endif /* CONFIG_PM || CONFIG_DPM */

#include <asm/uaccess.h>
#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>
#include <asm/arch/pmu.h>
#include <asm/arch/em1_mem.h>

#include "../mp200_common.h"
#include "../mp200_lcd_common.h"
#include "../mp200_fb.h"
#include "ntsc.h"
#include "mp200_nts_common.h"
#include "mp200_ntshw.h"
#include "mp200_ntsipu.h"
#include "mp200_adv7179.h"


/********************************************************
 *  debug parameters                                    *
 *******************************************************/
#define _DEBUG_NTS  0x00 /* 00008421(bit) */
                         /* 0x01: debug function in
                          * 0x02: debug function out
                          * 0x80: debug semafore
                          */


#define printk_err(fmt, arg...) \
	do{                     \
		printk(KERN_ERR DEV_NAME ": %s: " fmt, __func__, ## arg); \
	}while(0)

#define printk_wrn(fmt, arg...) \
	do{                     \
		printk(KERN_WARNING DEV_NAME ": %s: " fmt, __func__, ## arg); \
	}while(0)

#define printk_info(fmt, arg...) \
	do{                      \
		printk(KERN_INFO DEV_NAME ": " fmt, ## arg); \
	}while(0)

#if _DEBUG_NTS
#define printk_dbg(level, fmt, arg...) \
	do{                            \
		if(level > 0) printk(KERN_DEBUG DEV_NAME ": %s: " fmt, __func__, ## arg); \
	}while(0)
#else
#define printk_dbg(level, fmt, arg...) \
	;
#endif


/********************************************************
 *  Definitions                                         *
 *******************************************************/
#define DEV_NAME "mp200_nts"

/* V4L2 mix ON/OFF */
#define V4L2_MIX_OFF		0
#define V4L2_MIX_ON		1
/* V4L2 pause status */
#define V4L2_PAUSE_OFF		0
#define V4L2_PAUSE_ON		1
/* callback to V4L2 status */
#define V4L2_CALLBACK_OFF	0
#define V4L2_CALLBACK_ON	1

/* waite timer status */
#define MIX_V4L2_TIMEOUT	10		/* (MIX_V4L2_TIMEOUT * 10ms) is waittime */
#define NTS_TIMER_STOP		0
#define NTS_TIMER_START		1

/* nts_probe() */
#define NTS_SUCCESS		0
#define NTS_FAILED_REQUESTIRQ	-1


/********************************************************
 *  Macros                                              *
 *******************************************************/


/********************************************************
 *  Structure                                           *
 *******************************************************/


/********************************************************
 *  Variables                                           *
 *******************************************************/
struct mp200_nts_dev *ntsc;


/********************************************************
 *  Prototype declarations of function                  *
 *******************************************************/
/* Public function */
       int          mp200_nts_reserve(int iActiveDevice, int iSetOutmode);
       int          mp200_nts_release(int iActiveDevice);
       int          mp200_nts_getmode(void);
       int          mp200_nts_set_v4l2_image(V4L2_IMAGE_INFO *v4l2_info);
       int          mp200_nts_set_fb_image(FB_IMAGE_INFO *fb_info);
#if defined(CONFIG_PM) || defined(CONFIG_DPM)
       int          mp200_nts_suspend(struct platform_device *dev, pm_message_t state);
       int          mp200_nts_resume(struct platform_device *dev);

static int          DPM_suspend_flg = 0;
static int          ctrl_func       = 0;
#endif /* CONFIG_PM || CONFIG_DPM */

/* Private function */
static int          nts_set_v4l2_image(V4L2_IMAGE_INFO *v4l2_info);
static int          nts_set_fb_image(FB_IMAGE_INFO *fb_image_info);
static void         nts_mix_image(int iMixMode, int iCallbackV4L2);
static irqreturn_t  nts_irq_handler(int irq, void *dev_id);	/* VSYNC interrupt handler */
static void         nts_irq_handler_sub(int irq, void *dev_id);	/* VSYNC interrupt handler */
       void         nts_return_callback(void);
       void         nts_return_callback_refresh(void);
       void         nts_return_callback_ready(void);

static void         nts_init_flags(struct mp200_nts_dev *ntsc);
static void         nts_init_val(struct mp200_nts_dev *ntsc);
static int          nts_probe(void);
static void         nts_tmr_init(void);
static void         nts_tmr_add(void);
static void         nts_tmr_timeout(unsigned long data);
static void         nts_tmr_timeout_bottom_do(struct work_struct *num);

static int  __init  mp200_nts_init_module(void);
#ifdef MODULE
static void __exit  mp200_nts_exit_module(void);
static void         mp200_nts_remove(void);
#endif /* MODULE */

/*
 * extern V4L2 Driver function
 */
extern void         mp200_v4l2_lcd_callback(FRAME_DATA frame_data);
extern void         mp200_v4l2_lcd_refresh_callback(FRAME_DATA frame_data);
extern void         mp200_v4l2_ntsc_ready_callback(FRAME_DATA frame_data);


/******************************************************************************
* MODULE   : mp200_nts_reserve
* FUNCTION : 
* RETURN   :       0 : success
*          : -EINVAL : misstype argument
*          :    -EIO : 
* NOTE     : none
* CREATE   : 2008.11.13 M.Haguro
* UPDATE   :
******************************************************************************/
int mp200_nts_reserve(int iActiveDevice, int iSetOutmode)
{
	int iRet = 0;
	printk_dbg((_DEBUG_NTS & 0x01), "%d: iActiveDevice(%x)  iSetOutmode(%x)\n", __LINE__, iActiveDevice, iSetOutmode);

	/* check argument */
	if(iActiveDevice < NTS_ACTIVE_FB || iActiveDevice > NTS_ACTIVE_V4L2){
		iRet = -EINVAL;
		goto failed;
	}
	if(iSetOutmode < NTS_OUTPUT_NTSC || iSetOutmode > NTS_OUTPUT_PAL){
		iRet = -EINVAL;
		goto failed;
	}

	down(&ntsc->sem_img_data);
	printk_dbg((_DEBUG_NTS & 0x80), "down(A)\n");

	if(!ntsc->iNtsActive || ntsc->iOutmodeFlg == iSetOutmode){
		if(!ntsc->iNtsActive){
			/* NTSC initialize */
			iRet = ntshw_reserve(iSetOutmode);
			if(iRet || !ntsc->encoder->hw_init){
				iRet = -EIO;
			}
			else{
				int dummy;
				/* ADV7179 initialize */
				iRet = ntsc->encoder->hw_init(&dummy);
				if(iRet){
					ntshw_release();
					iRet = -EIO;
				}
				else{
					/* install NTSC irq handler */
					iRet = request_irq(INT_NTSC, nts_irq_handler, 0, DEV_NAME, NULL);
					if(iRet){
						if(ntsc->encoder->hw_shutdown){ 
							int dummy;
							ntsc->encoder->hw_shutdown(&dummy);
						}
						ntshw_release();
						iRet = -EIO;
					}
					else{
						int iZero = 0;
						ntsc->encoder->hw_command(ENCODER_SET_NORM,  &iSetOutmode);
						ntsc->encoder->hw_command(ENCODER_SET_INPUT, &iZero);

						/* start NTSC output */
						ntsc->iOutmodeFlg = iSetOutmode;
						ntshw_start(NTSHW_START);
					}
				}
			}
		}

		if(!iRet){
			if(iActiveDevice == NTS_ACTIVE_FB){
				memset(&ntsc->from_fb, 0, sizeof(FB_IMAGE_INFO));
			}
			ntsc->iNtsActive |= iActiveDevice;
			printk_dbg((_DEBUG_NTS & 0x01), "%d: iNtsActive(%04x)\n", __LINE__, ntsc->iNtsActive);
		}
	}
	else{
		iRet = -EIO;
	}

	up(&ntsc->sem_img_data);
	printk_dbg((_DEBUG_NTS & 0x80), "up(A)\n");

failed:
	return iRet;
}
EXPORT_SYMBOL(mp200_nts_reserve);


/******************************************************************************
* MODULE   : mp200_nts_release
* FUNCTION : 
* RETURN   :       0 : success
*          : -EINVAL : misstype argument
* NOTE     : none
* CREATE   : 2008.11.13 M.Haguro
* UPDATE   :
******************************************************************************/
int mp200_nts_release(int iActiveDevice)
{
	int iRet = 0;
	printk_dbg((_DEBUG_NTS & 0x01), "%d: iActiveDevice(%x)\n", __LINE__, iActiveDevice);

	/* check argument */
	if(iActiveDevice < NTS_ACTIVE_FB || iActiveDevice > NTS_ACTIVE_V4L2){
		iRet = -EINVAL;
		goto failed;
	}

	down(&ntsc->sem_img_data);
	printk_dbg((_DEBUG_NTS & 0x80), "down(A)\n");
	down(&ntsc->sem_img_flag);
	printk_dbg((_DEBUG_NTS & 0x80), "down(B)\n");

	if(ntsc->iNtsActive & iActiveDevice){
		if(ntsc->iNtsActive == (NTS_ACTIVE_FB | NTS_ACTIVE_V4L2)){
			/* fb1 & v4l2 mixing mode */
			if(iActiveDevice == NTS_ACTIVE_FB){
				/* fb1 disable */
				ntsc->iMixImageMode = (ntsc->iMixImageMode & ~FB_BIT) | FB_OFF;
				memset(&ntsc->from_fb, 0, sizeof(FB_IMAGE_INFO));

				if(ntsc->iMixImageMode != FB_V4L2_OFF){
					if(ntsc->iFrameNoNow == NTS_DISP_FRAME_A){
						ntshw_clr_framebuf(NTS_DISP_FRAME_B);
						ntsc->img_complete = 0;
						ntsipu_img_request();
						wait_event_interruptible(ntsc->wait_img, ntsc->img_complete);
						ntshw_clr_framebuf(NTS_DISP_FRAME_A);
					}
					else{	/* NTS_DISP_FRAME_B */
						ntshw_clr_framebuf(NTS_DISP_FRAME_A);
						ntsc->img_complete = 0;
						ntsipu_img_request();
						wait_event_interruptible(ntsc->wait_img, ntsc->img_complete);
						ntshw_clr_framebuf(NTS_DISP_FRAME_B);
					}
				}
				else{
					up(&ntsc->sem_img_flag);
					printk_dbg((_DEBUG_NTS & 0x80), "up(B)\n");
				}
			}
			else if(iActiveDevice == NTS_ACTIVE_V4L2){
				/* v4l2 disable */
				ntsc->iMixImageMode = (ntsc->iMixImageMode & ~V4L2_BIT) | V4L2_OFF;
				memset(&ntsc->from_v4l2, 0, sizeof(V4L2_IMAGE_INFO));

				if(ntsc->iMixImageMode != FB_V4L2_OFF){
					ntsc->img_complete = 0;
					ntsipu_img_request();
					wait_event_interruptible(ntsc->wait_img, ntsc->img_complete);
				}
				else{
					up(&ntsc->sem_img_flag);
					printk_dbg((_DEBUG_NTS & 0x80), "up(B)\n");
				}
			}
			else{
				up(&ntsc->sem_img_flag);
				printk_dbg((_DEBUG_NTS & 0x80), "up(B)\n");
			}
		}
		else
		{
			up(&ntsc->sem_img_flag);
			printk_dbg((_DEBUG_NTS & 0x80), "up(B)\n");
		}

		ntsc->iNtsActive &= ~iActiveDevice;
		printk_dbg((_DEBUG_NTS & 0x01), "%d: iNtsActive(%04x)\n", __LINE__, ntsc->iNtsActive);

		if(!ntsc->iNtsActive){
			ntsc->iOutmodeFlg = NTS_OUTPUT_DISABLE;

			ntshw_start(NTSHW_STOP);
			free_irq(INT_NTSC, NULL);
			ntshw_release();

			if(ntsc->encoder->hw_shutdown){ 
				int dummy;
				ntsc->encoder->hw_shutdown(&dummy);
			}
			nts_init_flags(ntsc);
		}
	}
	else
	{
		up(&ntsc->sem_img_flag);
		printk_dbg((_DEBUG_NTS & 0x80), "up(B)\n");
	}

	up(&ntsc->sem_img_data);
	printk_dbg((_DEBUG_NTS & 0x80), "up(A)\n");

failed:
	return iRet;
}
EXPORT_SYMBOL(mp200_nts_release);


/******************************************************************************
* MODULE   : mp200_nts_getmode
* FUNCTION : 
* RETURN   : 0 : non active
*          : 1 : active & NTSC
*          : 2 : active & PAL
* NOTE     : none
* CREATE   : 2008.11.13 M.Haguro
* UPDATE   :
******************************************************************************/
int mp200_nts_getmode(void)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");
	return ntsc->iOutmodeFlg;
}
EXPORT_SYMBOL(mp200_nts_getmode);


/******************************************************************************
* MODULE   : mp200_nts_set_v4l2_image
* FUNCTION : set image data from v4l2 to NTSC local
* RETURN   :  0 : success
*            -1 : faile
* NOTE     : none
* CREATE   : 2008.11.20 M.Haguro
* UPDATE   :
******************************************************************************/
int mp200_nts_set_v4l2_image(V4L2_IMAGE_INFO *v4l2_info)
{
	unsigned long flags;
	int ret = 0;

	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	/* Stop pause timer */
	spin_lock_irqsave(&ntsc->lock_tmr, flags);
	del_timer(&ntsc->nts_timer);
	ntsc->iTimerFlag    = NTS_TIMER_STOP;
	spin_unlock_irqrestore(&ntsc->lock_tmr, flags);
	ntsc->iV4L2PauseFlg = V4L2_PAUSE_OFF;

	down(&ntsc->sem_img_data);	/* get data semafore */
	printk_dbg((_DEBUG_NTS & 0x80), "down(A)\n");

	if(ntsc->iNtsActive & NTS_ACTIVE_V4L2){
		/* V4L2 active */
		nts_set_v4l2_image(v4l2_info);
	}
	else{
		/* V4L2 not active */
		ret = -EACCES;
	}
	up(&ntsc->sem_img_data);	/* release data semafore */
	printk_dbg((_DEBUG_NTS & 0x80), "up(A)\n");
	return ret;
}


/******************************************************************************
* MODULE   : mp200_nts_set_fb_image
* FUNCTION : set image data from Frame Buffer to NTSC local
* RETURN   :  0 : success
*            -1 : faile
* NOTE     : none
* CREATE   : 2008.11.20 M.Haguroq
* UPDATE   :
******************************************************************************/
int mp200_nts_set_fb_image(FB_IMAGE_INFO *fb_info)
{
	int ret = 0;
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	down(&ntsc->sem_img_data);	/* get data semafore */
	printk_dbg((_DEBUG_NTS & 0x80), "down(A)\n");

	if(ntsc->iNtsActive & NTS_ACTIVE_FB){
		/* fb1 active */
		nts_set_fb_image(fb_info);
	}
	else{
		/* fb1 not active */
		ret = -EACCES;
	}
	up(&ntsc->sem_img_data);	/* release data semafore */
	printk_dbg((_DEBUG_NTS & 0x80), "up(A)\n");
	return ret;
}


/******************************************************************************
* MODULE   : nts_set_v4l2_image
* FUNCTION : set image data from v4l2 to NTSC local
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2008.11.13 M.Haguro
* UPDATE   :
******************************************************************************/
static int nts_set_v4l2_image(V4L2_IMAGE_INFO *v4l2_info)
{
	unsigned long flags;

	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	/* DSP stop */
	if(v4l2_info->image_data.yrgbaddr == 0){
		printk_dbg((_DEBUG_NTS & 0x01), "<stop V4L2>\n");
		/* movie off */
		ntsc->iMixDSPFlg = V4L2_MIX_OFF;
		/* update Layer data */
		nts_mix_image(V4L2_OFF, V4L2_CALLBACK_OFF);
		return 0;
	}

	/* movie on */
	ntsc->iMixDSPFlg = V4L2_MIX_ON;

	memcpy(&ntsc->from_v4l2, v4l2_info, sizeof(V4L2_IMAGE_INFO));

	/* mix - image data from v4l2 and image data from fb driver (DSP and 2D) */
	if(ntsc->from_v4l2.screen_data.hsize == 0){
		printk_dbg((_DEBUG_NTS & 0x01), "<V4L2_OFF>\n");
		nts_mix_image(V4L2_OFF, V4L2_CALLBACK_ON);
	}
	else{
		printk_dbg((_DEBUG_NTS & 0x01), "<V4L2_ON>\n");
		nts_mix_image(V4L2_ON,  V4L2_CALLBACK_ON);
	}

	/* PAUSE check timer start */
	spin_lock_irqsave(&ntsc->lock_tmr, flags);
	if(ntsc->iTimerFlag == NTS_TIMER_STOP
#if defined(CONFIG_PM) || defined(CONFIG_DPM)
			&& !DPM_suspend_flg
#endif /* CONFIG_PM || CONFIG_DPM */
	){
		nts_tmr_add();
		ntsc->iTimerFlag = NTS_TIMER_START;
	}
	spin_unlock_irqrestore(&ntsc->lock_tmr, flags);
	return 0;
}


/******************************************************************************
* MODULE   : nts_set_fb_image
* FUNCTION : set image data from Frame Buffer to NTSC local
* RETURN   :  0 : success
*            -1 : faile
* NOTE     : none
* CREATE   : 2008.11.20 M.Haguro
* UPDATE   :
******************************************************************************/
static int nts_set_fb_image(FB_IMAGE_INFO *fb_info)
{
	int iRet = 0;

	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	memcpy(&ntsc->from_fb, fb_info, sizeof(FB_IMAGE_INFO));

	/* check update flag */
	switch(ntsc->from_fb.update_flag){
	case FB_UPDATE_ON:  /*    ioctl(MP200_FB_UPDATE_SCRN) & update = 0 */
	                    /* or interval update                          */
		printk_dbg((_DEBUG_NTS & 0x02), "FB_UPDATE_ON\n");

		if(ntsc->iMixDSPFlg == V4L2_MIX_ON && ntsc->iV4L2PauseFlg == V4L2_PAUSE_OFF){
			/* movie on.  and not pause status */
			iRet = -1;
		}
		else{
			/* movie off. or movie pause status */
			nts_mix_image(FB_ON, V4L2_CALLBACK_OFF);
		}
		break;
	case FB_ABSOLUTERY_UPDATE: /* ioctl(MP200_FB_UPDATE_SCRN) & update = 1 */
		printk_dbg((_DEBUG_NTS & 0x02), "FB_ABSOLUTERY_UPDATE\n");
		nts_mix_image(FB_ON, V4L2_CALLBACK_OFF);
		break;
	case FB_UPDATE_OFF:  /* ioctl(MP200_FB_SET_MODES) */
	default:             /* other */
		printk_dbg((_DEBUG_NTS & 0x02), "FB_UPDATE_OFF\n");
		iRet = -1;
		break;
	}
	return iRet;
}


/******************************************************************************
* MODULE   : nts_mix_image
* FUNCTION : By using IMG, mix 2D (and DSP) image
* RETURN   :  0 : success
*          : -1 : failed
* NOTE     : none
* CREATE   : 2008.11.14 M.Haguro
* UPDATE   :
******************************************************************************/
static void nts_mix_image(int iMixMode, int iCallbackV4L2)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	/* get flag semafore */
	down(&ntsc->sem_img_flag);
	printk_dbg((_DEBUG_NTS & 0x80), "down(B)\n");

	/* set flag and temp data */
	ntsc->iV4L2CallbackFlg     = iCallbackV4L2;
	ntsc->uiMixFrameBufferPage = ntsc->from_fb.mix_buf_page;
	ntsc->iMixDSPFlg_toFb      = ntsc->iMixDSPFlg;

	if(iMixMode & FB_BIT){
		ntsc->iMixImageMode = (ntsc->iMixImageMode & ~FB_BIT)   | iMixMode;
		if((ntsc->iNtsActive & NTS_ACTIVE_V4L2) && ntsc->from_v4l2.image_data.yrgbaddr){
			ntsc->iMixImageMode = (ntsc->iMixImageMode & ~V4L2_BIT) | V4L2_ON;
		}
		else{
			ntsc->iMixImageMode = (ntsc->iMixImageMode & ~V4L2_BIT) | V4L2_OFF;
		}
	}
	else{
		ntsc->iMixImageMode = (ntsc->iMixImageMode & ~V4L2_BIT) | iMixMode;
		if((ntsc->iNtsActive & NTS_ACTIVE_FB) && ntsc->from_fb.image_data.yrgbaddr){
			ntsc->iMixImageMode = (ntsc->iMixImageMode & ~FB_BIT) | FB_ON;
		}
		else{
			ntsc->iMixImageMode = (ntsc->iMixImageMode & ~FB_BIT) | FB_OFF;
		}

		if(ntsc->iV4L2CallbackFlg == V4L2_CALLBACK_ON){
			/* set frame data from V4L2 */
			memcpy(&ntsc->v4l2_callback_data, &ntsc->from_v4l2.frame_data, sizeof(struct _FRAME_DATA));
		}
	}

	if(ntsc->iMixImageMode == FB_V4L2_OFF){
		printk_dbg((_DEBUG_NTS & 0x02), "iMixImageMode(%d) -> Black screen\n", ntsc->iMixImageMode);

		/* Black Screen */
		ntshw_set_ntsout(NTS_NTSOUT_ENABLE_BLACK);
		nts_return_callback_refresh();
		nts_return_callback();
	}
	else{
		printk_dbg((_DEBUG_NTS & 0x02), "iMixImageMode(%d)\n", ntsc->iMixImageMode);
		printk_dbg((_DEBUG_NTS & 0x02), "ntsipu_img_request() call\n");
		/* regist img request to IPU Driver */
		if(ntsipu_img_request()){
			printk_dbg((_DEBUG_NTS & 0x02), "ntsipu_img_request() failed\n");
			nts_return_callback_refresh();
			nts_return_callback();
		}
	}
}


/*****************************************************************************
* MODULE   : nts_irq_handler
* FUNCTION : NTSC interrupt handler
* RETURN   : IRQ_HANDLED
* NOTE     : none
* CREATE   : 2008/11/17 M.Haguro
* UPDATE   :
******************************************************************************/
static irqreturn_t nts_irq_handler(int irq, void *dev_id)
{
	unsigned long ulRegVal32, flags;

	printk_dbg((_DEBUG_NTS & 0x01), "\n");
	spin_lock_irqsave(&ntsc->lock_tmr, flags);

	/*  check Interuppt Status */
	ulRegVal32 = ntshw_chk_intstatus();

	/* Interrupt Status Clear */
	ntshw_set_interruput(NTSHW_INTFFCLR);

	if(ulRegVal32 & NTS_NTSVS_BIT){
		nts_irq_handler_sub(irq, dev_id);
	}

	spin_unlock_irqrestore(&ntsc->lock_tmr, flags);
	printk_dbg((_DEBUG_NTS & 0x02), "<end>\n");

	return IRQ_HANDLED;
}


/*****************************************************************************
* MODULE   : nts_irq_handler_sub
* FUNCTION : NTSC interrupt handler
* RETURN   : 
* NOTE     : none
* CREATE   : 2008/11/17 M.Haguro
* UPDATE   :
******************************************************************************/
static void nts_irq_handler_sub(int irq, void *dev_id)
{
	unsigned long ulRegVal32;

	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	/* check AREASTATUS */
	ulRegVal32 = ntshw_chk_framesel();

	if (((ntsc->iFrameNoNext == NTS_DISP_FRAME_A) && ((ulRegVal32 & NTS_AREASTATUS_BIT) == NTS_AREASTATUS_BUFA)) ||
	    ((ntsc->iFrameNoNext == NTS_DISP_FRAME_B) && ((ulRegVal32 & NTS_AREASTATUS_BIT) == NTS_AREASTATUS_BUFB)))
	{
		/* disable NTSC interrupts */
		ntshw_set_interruput(NTSHW_INTENCLR);

		ntsc->iFrameNoNow = ntsc->iFrameNoNext;

		nts_return_callback();
	}
	printk_dbg((_DEBUG_NTS & 0x02), "<end>\n");
}


/*****************************************************************************
* MODULE   : nts_return_callback
* FUNCTION : NTSC interrupt handler
* RETURN   : 
* NOTE     : none
* CREATE   : 2008/11/17 M.Haguro
* UPDATE   :
******************************************************************************/
void nts_return_callback(void)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	/* when kick nts_mix_image by V4L2 */
	if(ntsc->iV4L2CallbackFlg == V4L2_CALLBACK_ON){
		ntsc->iV4L2CallbackFlg = V4L2_CALLBACK_OFF;

		printk_dbg((_DEBUG_NTS & 0x02), "-> v4l2\n");
		/* call v4l2 callback function */
		mp200_v4l2_lcd_callback(ntsc->v4l2_callback_data);
		memset(&ntsc->v4l2_callback_data, 0, sizeof(struct _FRAME_DATA));
	}

#if defined(CONFIG_PM) || defined(CONFIG_DPM)
	ctrl_func = 0;
#endif /* CONFIG_PM || CONFIG_DPM */

	ntsc->img_complete = 1;
	wake_up_interruptible(&ntsc->wait_img);

	/* release semafore */
	up(&ntsc->sem_img_flag);
	printk_dbg((_DEBUG_NTS & 0x80), "up(B)\n");
}


/*******************************************************************************
* MODULE   : nts_return_callback_refresh
* FUNCTION : 
* RETURN   : none
* NOTE     : none
* CREATE   : 2008/12/05 M.Haguro
* UPDATE   :
********************************************************************************/
void nts_return_callback_refresh(void)
{
	unsigned long flags;
	spin_lock_irqsave(&ntsc->lock_callback, flags);
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	if(ntsc->iNtsActive & NTS_ACTIVE_FB){
		/* call FB Driver function "mp200_fb_callback" */
		mp200_fb_callback((int)ntsc->uiMixFrameBufferPage, ntsc->iMixDSPFlg_toFb, EM1_FB_DEVICE_NTS);
	}

	/* when kick nts_mix_image by V4L2 */
	if(ntsc->iV4L2CallbackFlg == CALLBACK_V4L2_ON){
		printk_dbg((_DEBUG_NTS & 0x02), "-> v4l2\n");
		/* call v4l2 callback function */
		mp200_v4l2_lcd_refresh_callback(ntsc->v4l2_callback_data);
	}

#if defined(CONFIG_PM) || defined(CONFIG_DPM)
	ctrl_func = 1;
#endif /* CONFIG_PM || CONFIG_DPM */
	spin_unlock_irqrestore(&ntsc->lock_callback, flags);
}


/*******************************************************************************
* MODULE   : nts_return_callback_ready
* FUNCTION : 
* RETURN   : none
* NOTE     : none
* CREATE   : 2009/04/06 M.Haguro
* UPDATE   :
********************************************************************************/
void nts_return_callback_ready(void)
{
	unsigned long flags;
	spin_lock_irqsave(&ntsc->lock_callback, flags);
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	/* when kick nts_mix_image by V4L2 */
	if(ntsc->iV4L2CallbackFlg == CALLBACK_V4L2_ON){
		printk_dbg((_DEBUG_NTS & 0x02), "-> v4l2\n");
		/* call v4l2 callback function */
		mp200_v4l2_ntsc_ready_callback(ntsc->v4l2_callback_data);
	}
	spin_unlock_irqrestore(&ntsc->lock_callback, flags);
}


/********************************************************
 *  Timer Function Definitions                          *
 *******************************************************/
/******************************************************************************
* MODULE   : nts_tmr_init
* FUNCTION : timer function init
* RETURN   : none
* NOTE     : private function
* CREATE   : 2008.11.13 M.Haguro
* UPDATE   :
******************************************************************************/
static void nts_tmr_init(void)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	init_timer(&ntsc->nts_timer);
	ntsc->nts_timer.function = nts_tmr_timeout;
	ntsc->nts_timer.data     = 0;
	ntsc->iTimerFlag         = NTS_TIMER_STOP;
}


/******************************************************************************
* MODULE   : nts_tmr_add
* FUNCTION : timer function add_timer
* RETURN   : none
* NOTE     : private function
* CREATE   : 2008.11.13 M.Haguro
* UPDATE   :
******************************************************************************/
static void nts_tmr_add(void)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	ntsc->nts_timer.expires = jiffies + (MIX_V4L2_TIMEOUT * 10 * HZ / 1000);
	add_timer(&ntsc->nts_timer);
}


/******************************************************************************
* MODULE   : nts_tmr_timeout
* FUNCTION : when call mp200_fb_call_timer
* RETURN   : none
* NOTE     : public function
* CREATE   : 2008.11.13 M.Haguro
* UPDATE   :
******************************************************************************/
static void nts_tmr_timeout(unsigned long data)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	/* change NTSC status to PAUSE */
	ntsc->iV4L2PauseFlg = V4L2_PAUSE_ON;
	queue_work(ntsc->mp200_nts_workqueue, &ntsc->wk_timeout_bottom_half);
}


/******************************************************************************
* MODULE   : nts_tmr_timeout_bottom_do
* FUNCTION : when call mp200_fb_call_timer
* RETURN   : none
* NOTE     : public function
* CREATE   : 2008.11.13 M.Haguro
* UPDATE   :
******************************************************************************/
static void nts_tmr_timeout_bottom_do(struct work_struct *num)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	down(&ntsc->sem_img_data);	/* get semafore */
	printk_dbg((_DEBUG_NTS & 0x80), "down(A)\n");

	/* mix image */
	if(ntsc->from_v4l2.screen_data.hsize == 0){
		printk_dbg((_DEBUG_NTS & 0x01), "<V4L2_OFF>\n");
		nts_mix_image(V4L2_OFF, V4L2_CALLBACK_OFF);
	}
	else{
		printk_dbg((_DEBUG_NTS & 0x01), "<V4L2_ON>\n");
		nts_mix_image(V4L2_ON, V4L2_CALLBACK_OFF);
	}

	up(&ntsc->sem_img_data);	/* release data semafore */
	printk_dbg((_DEBUG_NTS & 0x80), "up(A)\n");
}


#if defined(CONFIG_PM) || defined(CONFIG_DPM)
/********************************************************
 *  Suspend/Resume Function Definitions                 *
 *******************************************************/
/******************************************************************************
* MODULE   : mp200_nts_suspend
* FUNCTION : suspend NTSC driver
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2006.08.15 H.Nagashima
* UPDATE   :
******************************************************************************/
int mp200_nts_suspend(struct platform_device *dev, pm_message_t state)
{
	if(ntsc->iOutmodeFlg == NTS_OUTPUT_DISABLE){
		return 0;
	}
	else{
		if(state.event == DEV_SUSPEND_IDLE_1){
			return -EBUSY;
		}
		else{	/* state.event == PM_EVENT_SUSPEND */
			if(!mp200_sleep_while_idle){
				DPM_suspend_flg = 1;

				/* NTSC timer stop */
				del_timer(&ntsc->nts_timer);

				/* NTSC POWER OFF */
				ntshw_standby(NTSHW_STANDBY);

				/* Reset NTSC */
				ntshw_reset(NTSHW_RESET);
			}
		}
		return 0;
	}
}


/******************************************************************************
* MODULE   : mp200_nts_resume
* FUNCTION : suspend NTSC driver
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2006.08.15 H.Nagashima
* UPDATE   :
******************************************************************************/
int mp200_nts_resume(struct platform_device *dev)
{
	if(ntsc->iOutmodeFlg == NTS_OUTPUT_DISABLE){
		return 0;
	}
	else{
		if(DPM_suspend_flg){
			DPM_suspend_flg = 0;

			/* UnReset NTSC */
			ntshw_reset(NTSHW_UNRESET);

			/* NTSC POWER ON */
			ntshw_standby(NTSHW_WAKEUP);

			/* NTSC timer start */
			if(ntsc->iTimerFlag == NTS_TIMER_START){
				nts_tmr_add();
			}

			switch (ctrl_func){
//			case 2:
//				nts_return_callback_refresh();
//				/* FALL THROUGH */
			case 1:
				nts_return_callback();
				break;
			case 0:
			default:
				break;
			}
		}
		return 0;
	}
}
#endif /* CONFIG_PM || CONFIG_DPM */


/********************************************************
 *  Init Function Definitions                           *
 *******************************************************/
/******************************************************************************
* MODULE   : nts_probe
* FUNCTION : NTSC Driver probe
* RETURN   : 0     : success
*            other : fail
* NOTE     : none
* CREATE   : 2008.11.17 M.Haguro
* UPDATE   :
******************************************************************************/
static int nts_probe(void)
{
	int iRet;
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	/* initiarize ipu request */
	iRet = ntsipu_initialize(ntsc);
	if(iRet){
		goto err_ret;
	}

	/* initialize semafore */
	sema_init(&ntsc->sem_img_data, 1);
	sema_init(&ntsc->sem_img_flag, 1);

	/* initialize spin lock */
	spin_lock_init(&ntsc->lock_tmr);
	spin_lock_init(&ntsc->lock_callback);

	/* initiarize work queue */
	ntsc->mp200_nts_workqueue = create_singlethread_workqueue(DEV_NAME);
	INIT_WORK(&ntsc->wk_timeout_bottom_half, nts_tmr_timeout_bottom_do);

	init_waitqueue_head(&ntsc->wait_img);
err_ret:
	return iRet;
}


/******************************************************************************
* MODULE   : nts_init_flags
* FUNCTION : variable default set
* RETURN   : none
* NOTE     : none
* CREATE   : 2009.1.6 M.Haguro
* UPDATE   :
******************************************************************************/
static void nts_init_flags(struct mp200_nts_dev *ntsc)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");
	ntsc->iNtsActive           = 0;

	/* input image data from fb driver */
	ntsc->uiMixFrameBufferPage = NTS_DISP_FRAME_A;		/* next frame buffer (2D)         */
	ntsc->uiAbsolutelyUpFlag   = FB_UPDATE_OFF;		/* absolutely update flag         */

	/* flags */
	ntsc->iMixDSPFlg           = V4L2_MIX_OFF;		/* movie & UI mix ON/OFF          */
	ntsc->iMixDSPFlg_toFb      = V4L2_MIX_OFF;		/* movie & UI mix ON/OFF (for FB) */

	ntsc->iV4L2PauseFlg        = V4L2_PAUSE_OFF;		/* V4L2 pause status flag         */
	ntsc->iV4L2CallbackFlg     = V4L2_CALLBACK_OFF;		/* need to callback to V4L2       */
	memset(&ntsc->v4l2_callback_data, 0, sizeof(struct _FRAME_DATA));

	ntsc->iFrameNoNow          = NTS_DISP_INIT;		/* now display frame no           */
	ntsc->iFrameNoNext         = NTS_DISP_INIT;		/* set change frame no            */

	ntsc->iOutmodeFlg          = NTS_OUTPUT_DISABLE;	/* output i/f                     */
	ntsc->iMixImageMode        = FB_OFF | V4L2_OFF;		/*  */

	memset(&ntsc->from_fb,   0, sizeof(FB_IMAGE_INFO));
	memset(&ntsc->from_v4l2, 0, sizeof(V4L2_IMAGE_INFO));
}


/******************************************************************************
* MODULE   : nts_init_val
* FUNCTION : variable default set
* RETURN   : none
* NOTE     : none
* CREATE   : 2008.11.17 M.Haguro
* UPDATE   :
******************************************************************************/
static void nts_init_val(struct mp200_nts_dev *ntsc)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");
	memset(ntsc, 0, sizeof(struct mp200_nts_dev));

	/* input image data from fb driver */
	nts_init_flags(ntsc);
}


/******************************************************************************
* MODULE   : nts_alloc_dev
* FUNCTION : variable default set
* RETURN   : none
* NOTE     : none
* CREATE   : 2008.11.17 M.Haguro
* UPDATE   :
******************************************************************************/
static void *nts_alloc_dev(void)
{
	void *alloc_dev;

	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	alloc_dev = kmalloc(sizeof(struct mp200_nts_dev), GFP_KERNEL);
	if(!alloc_dev){
		;
	}
	else{
		/* variable initialized */
		nts_init_val((struct mp200_nts_dev *)alloc_dev);
	}
	return alloc_dev;
}


/******************************************************************************
* MODULE   : nts_free_dev
* FUNCTION : variable default set
* RETURN   : none
* NOTE     : none
* CREATE   : 2008.11.17 M.Haguro
* UPDATE   :
******************************************************************************/
static void nts_free_dev(void)
{
	if(ntsc){
		memset(ntsc, 0, sizeof(struct mp200_nts_dev));
		kfree(ntsc);
	}
}


/******************************************************************************
* MODULE   : mp200_nts_init_module
* FUNCTION : initialize NTSC Driver
* RETURN   : 0     : success
*            other : fail
* NOTE     : none
* CREATE   : 2008.11.17 M.Haguro
* UPDATE   :
******************************************************************************/
static int __init mp200_nts_init_module(void)
{
	int iRet;

	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	/* nts_device memset */
	ntsc = (struct mp200_nts_dev *)nts_alloc_dev();
	if(!ntsc){
		printk_err("memory allocate error\n");
		iRet = -ENOMEM;
		goto fail_nts_alloc_dev;
	}

	/* initialize ADV7179 */
	iRet = adv7179_init(ntsc);
	if(iRet){
		printk_err("adv7179 initalize error\n");
		goto fail_adv7179_init;
	}

	/* initialize NTSC */
	iRet = ntshw_init();
	if(iRet){
		printk_err("ntshw initalize error\n");
		goto fail_ntshw_init;
	}

	/* call nts_probe */
	nts_probe();
	if(iRet){
		printk_err("nts probe error\n");
		goto fail_nts_probe;
	}

	/* timer initiarize */
	nts_tmr_init();

	printk_info("registered device ntsc\n");
	iRet = 0;
	goto success;

fail_nts_probe:
	ntshw_exit();
fail_ntshw_init:
	adv7179_exit(ntsc);
fail_adv7179_init:
	nts_free_dev();
fail_nts_alloc_dev:
success:
	return iRet;
}


/********************************************************
 *  Exit Function Definitions                           *
 *******************************************************/
#ifdef MODULE
/******************************************************************************
* MODULE   : mp200_nts_exit_module
* FUNCTION : cleanup NTSC module
* RETURN   : none
* NOTE     : none
* CREATE   : 2008.11.17 M.Haguro
* UPDATE   :
******************************************************************************/
static void __exit mp200_nts_exit_module(void)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	mp200_nts_remove();

	/* disable NTSC */
	ntshw_exit();

	adv7179_exit(ntsc);
	nts_free_dev();
}


/******************************************************************************
* MODULE   : mp200_nts_remove
* FUNCTION : release NTSC resource
* RETURN   : none
* NOTE     : none
* CREATE   : 2008.11.17 M.Haguro
* UPDATE   :
******************************************************************************/
static void mp200_nts_remove(void)
{
	printk_dbg((_DEBUG_NTS & 0x01), "\n");

	/* workqueue uninitialize */
	destroy_workqueue(ntsc->mp200_nts_workqueue);

	ntsipu_finitialize();
}
#endif /* MODULE */


/* module alias */
#ifdef MODULE
module_init(mp200_nts_init_module);
module_exit(mp200_nts_exit_module);
#else
__initcall(mp200_nts_init_module);
#endif

/* EXPORT_SYMBOL */
EXPORT_SYMBOL(mp200_nts_set_v4l2_image);
EXPORT_SYMBOL(mp200_nts_set_fb_image);

