/* 2010-08-25: File added and changed by Sony Corporation */
/*
 * File Name       : /drivers/video/mp200/nts/mp200_ntsipu.c
 * Function        : NTSC Driver corresponding to V4L2 (IPU Control)
 * Release Version : Ver 1.00
 * Release Date    : 2008/11/13
 *
 * Copyright (C) NEC Electronics Corporation 2008
 *
 *
 * This program is free software;you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by Free Softwere
 * Foundation; either version 2 of License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warrnty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; If not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


/********************************************************
 *  Include Files                                       *
 *******************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/device.h>
#include <linux/platform_device.h>

#if defined(CONFIG_PM) || defined(CONFIG_DPM)
#include <asm/arch/pm.h>
#endif /* CONFIG_PM || CONFIG_DPM */

#include <asm/uaccess.h>
#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>
#include <asm/arch/pmu.h>
#include <asm/arch/em1_mem.h>
#include <asm/arch/ipu.h>
#include <asm/arch/img.h>
#include <asm/arch/gdma.h>

#include "../mp200_common.h"
#include "ntsc.h"
#include "mp200_nts_common.h"
#include "mp200_nts.h"
#include "mp200_ntshw.h"
#include "mp200_ntsipu.h"
#include "mp200_nts_debug.h"


/********************************************************
 *  Definitions                                         *
 *******************************************************/
#define DEV_NAME "mp200_ntsipu"
#define NTSC_ERRATA 1


/********************************************************
 *  debug parameters                                    *
 *******************************************************/
#define _DEBUG_NTSIPU  0x00 /* 00008421(bit) */
                            /* 0x01: debug function in
                             * 0x02: debug function out
                             * 0x80: debug semafore
                             */


#define printk_err(fmt, arg...) \
	do{                     \
		printk(KERN_ERR DEV_NAME ": %s: " fmt, __func__, ## arg); \
	}while(0)

#define printk_wrn(fmt, arg...) \
	do{                     \
		printk(KERN_WARNING DEV_NAME ": %s: " fmt, __func__, ## arg); \
	}while(0)

#define printk_info(fmt, arg...) \
	do{                      \
		printk(KERN_INFO DEV_NAME ": " fmt, ## arg); \
	}while(0)

#if _DEBUG_NTSIPU
#define printk_dbg(level, fmt, arg...) \
	do{                            \
		if(level > 0) printk(KERN_DEBUG DEV_NAME ": %s: " fmt, __func__, ## arg); \
	}while(0)
#else
#define printk_dbg(level, fmt, arg...) \
	;
#endif


/********************************************************
 *  Macros                                              *
 *******************************************************/
/* request data. for IPU driver */
#define IPU_REQUEST_MAX 2

#define IpuRequest(DATA, REG)                                              \
do{                                                                        \
	ntsipu->ipu_request[index]->ipu_reg_info[index_req].reg  = (REG);  \
	ntsipu->ipu_request[index]->ipu_reg_info[index_req].data = (DATA); \
	index_req++;                                                       \
}while(0)


/********************************************************
 *  Structure                                           *
 *******************************************************/
enum old_frame_max {
	OLD_FRAME_A = 0,
	OLD_FRAME_B = 1,
	OLD_MAX     = 2,
};

/* request data. from fb/V4L2 driver */
enum img_max {
	IMG_SRC_B = 0,
	IMG_SRC_F = 1,
	IMG_DST   = 2,
	IMG_MAX   = 3,
};

enum gdma_max {
	GDMA_FILL1 = 0,
	GDMA_MAX   = 1,
};

/* request data. for IPU driver */
struct mp200_ntsipu {
	struct mp200_nts_dev *ntsc;
	wait_queue_head_t     wait_gdma;
	int                   gdma_complete;

	struct image_data    *old[OLD_MAX];
	struct image_data    *img[IMG_MAX];
	struct image_data    *gdma[GDMA_MAX];
	struct mp200_ipu_req *ipu_request[IPU_REQUEST_MAX]; /* queue info for IPU */
};
static struct mp200_ntsipu *ntsipu;


/********************************************************
 *  Prototype declarations of function                  *
 *******************************************************/
/* ------------------ public function ------------------------------------------- */
       int           ntsipu_initialize(struct mp200_nts_dev *ntsc);
       void          ntsipu_finitialize(void);
       int           ntsipu_img_request(void);
       int           ntsipu_clr_framebuf(struct image_data *buff);
/* ------------------ private function < IMG > ---------------------------------- */
static void          ntsipu_callback_yuv422(unsigned long sequence, int flag);
static void          ntsipu_callback_yuv422_bottom(void);
static void          ntsipu_callback_yuv422_return_ntsc(void);
static void          ntsipu_callback_rgb565(unsigned long sequence, int flag);
static int           ntsipu_chk_ipuimg(void);
static int           ntsipu_chk_ipuimg_bottom(void);
static unsigned long ntsipu_set_imgmode(void);
static unsigned long ntsipu_set_format(void);
static int           ntsipu_set_ipuimg(int index, int img_mode);
/* ------------------ private function < GDMA > --------------------------------- */
static void          ntsipu_callback_ipugdma(unsigned long sequence, int flag);
static int           ntsipu_chk_ipugdma(void);
static int           ntsipu_gdma_request(int index);
static int           ntsipu_set_ipugdma(int index);


#ifdef NTSC_ERRATA
#include <asm/arch/dma.h>
#define NTSC_DMA_CALLBACK_COMP 0x01

struct mp200_ntsc_errata{
	int               ntsc_upscale;
	int               ntsc_dma_stat;
	wait_queue_head_t queue_p;

	dma_regs_t       *dma_regs;
	dma_addr_t        srcV;
	dma_addr_t        destV;

	struct workqueue_struct *ntsipu_workqueue;
	struct work_struct       wk_dma_request_bottom;
};
static struct mp200_ntsc_errata *errata;

static int  ntsipu_dma_request(void);
static void ntsipu_dma_request_bottom(struct work_struct *num);
static void ntsipu_dma_callback(void *data, int intsts, int intrawsts)
{
	errata->ntsc_dma_stat = NTSC_DMA_CALLBACK_COMP;
	wake_up_interruptible(&errata->queue_p);
}
#endif /* NTSC_ERRATA */


/* ------------------ public function ------------------------------------------- */
/******************************************************************************
* MODULE   : ntsipu_initialize
* FUNCTION : 
* RETURN   :        0 : success
           : -ENOEMEM : failed
* NOTE     : none
* CREATE   : 2008.11.17 M.Haguro
* UPDATE   :
******************************************************************************/
int ntsipu_initialize(struct mp200_nts_dev *ntsc)
{
	int i, iRet;

	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	/* allocate request data. from fb/V4L2 driver */
	ntsipu = (struct mp200_ntsipu *)kmalloc(sizeof(struct mp200_ntsipu), GFP_KERNEL);
	if(!ntsipu){
		goto fail_alloc_mem;
	}
	memset(ntsipu, 0, sizeof(struct mp200_ntsipu));
	ntsipu->ntsc = ntsc;

	/* allocate old request data. from fb/V4L2 driver */
	for(i = 0; i < OLD_MAX; i++){
		ntsipu->old[i] = (struct image_data *)kmalloc(sizeof(struct image_data), GFP_KERNEL);
		if(!ntsipu->old[i]){
			goto fail_alloc_mem;
		}
		memset(ntsipu->old[i], 0, sizeof(struct image_data));
	}
	/* allocate request data. from fb/V4L2 driver */
	for(i = IMG_SRC_B; i < IMG_MAX; i++){
		ntsipu->img[i] = (struct image_data *)kmalloc(sizeof(struct image_data), GFP_KERNEL);
		if(!ntsipu->img[i]){
			goto fail_alloc_mem;
		}
		memset(ntsipu->img[i], 0, sizeof(struct image_data));
	}
	/* allocate request data. from fb/V4L2 driver */
	for(i = GDMA_FILL1; i < GDMA_MAX; i++){
		ntsipu->gdma[i] = (struct image_data *)kmalloc(sizeof(struct image_data), GFP_KERNEL);
		if(!ntsipu->gdma[i]){
			goto fail_alloc_mem;
		}
		memset(ntsipu->gdma[i], 0, sizeof(struct image_data));
	}

	/* allocate request data. for IPU-IMG driver */
	for(i = 0; i < IPU_REQUEST_MAX; i++){
		ntsipu->ipu_request[i] = (struct mp200_ipu_req *)kmalloc(MAX_SIZE_MP200_IPU_REQ, GFP_KERNEL);
		if(!ntsipu->ipu_request[i]){
			goto fail_alloc_mem;
		}
		memset(ntsipu->ipu_request[i], 0, MAX_SIZE_MP200_IPU_REQ);
	}

#ifdef NTSC_ERRATA
	errata = (struct mp200_ntsc_errata *)kmalloc(sizeof(struct mp200_ntsc_errata), GFP_KERNEL);
	if(!errata){
		goto fail_alloc_mem;
	}
	memset(errata, 0, sizeof(struct mp200_ntsc_errata));

	mp200_request_dma(MP200_DMAC_M2M_ACPU_LCH3, "NTSC M2M", ntsipu_dma_callback, (void *)0, &errata->dma_regs);
	init_waitqueue_head(&errata->queue_p);

	/* workqueue initialize */
	errata->ntsipu_workqueue = create_singlethread_workqueue(DEV_NAME);
	INIT_WORK(&errata->wk_dma_request_bottom, ntsipu_dma_request_bottom);
#endif /* NTSC_ERRATA */

	init_waitqueue_head(&ntsipu->wait_gdma);
	iRet = 0;
	goto success;

fail_alloc_mem:
	printk_wrn("memory allocate err");
	ntsipu_finitialize();
	iRet = -ENOMEM;
success:
	return iRet;
}


/******************************************************************************
* MODULE   : ntsipu_finitialize
* FUNCTION : 
* RETURN   : none
* NOTE     : none
* CREATE   : 2008.11.17 M.Haguro
* UPDATE   :
******************************************************************************/
void ntsipu_finitialize(void)
{
	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

#ifdef NTSC_ERRATA
	if(errata){
		/* workqueue uninitialize */
		destroy_workqueue(errata->ntsipu_workqueue);
		memset(errata, 0, sizeof(struct mp200_ntsc_errata));
		kfree(errata);
	}
#endif /* NTSC_ERRATA */

	if(ntsipu){
		int i;

		/* free request data. for IPU driver */
		for(i = 0; i < IPU_REQUEST_MAX; i++){
			if(ntsipu->ipu_request[i]){
				memset(ntsipu->ipu_request[i], 0, sizeof(MAX_SIZE_MP200_IPU_REQ));
				kfree(ntsipu->ipu_request[i]);
			}
		}
		/* free request data. from fb/V4L2 driver */
		for(i = GDMA_FILL1; i < GDMA_MAX; i++){
			if(ntsipu->gdma[i]){
				memset(ntsipu->gdma[i], 0, sizeof(struct image_data));
				kfree(ntsipu->gdma[i]);
			}
		}
		/* free request data. from fb/V4L2 driver */
		for(i = IMG_SRC_B; i < IMG_MAX; i++){
			if(ntsipu->img[i]){
				memset(ntsipu->img[i], 0, sizeof(struct image_data));
				kfree(ntsipu->img[i]);
			}
		}
		/* allocate old request data. from fb/V4L2 driver */
		for(i = 0; i < OLD_MAX; i++){
			if(ntsipu->old[i]){
				memset(ntsipu->old[i], 0, sizeof(struct image_data));
				kfree(ntsipu->old[i]);
			}
		}
		memset(ntsipu, 0, sizeof(struct mp200_ntsipu));
		kfree(ntsipu);
	}
}


/*****************************************************************************
* MODULE   : ntsipu_clr_framebuf
* FUNCTION : 
* RETURN   : -
* NOTE     : 
* CREATE   : 2009.1.22 M.Haguro
* UPDATE   : 
******************************************************************************/
int ntsipu_clr_framebuf(struct image_data *buff)
{
	int index = 0;

	memcpy(ntsipu->gdma[index], buff, sizeof(struct image_data));

	return ntsipu_gdma_request(index);
}


/******************************************************************************
* MODULE   : ntsipu_img_request
* FUNCTION : changing frame and callback to fb and v4l2
* RETURN   :       0 : success
           : -EINVAL : failed
* NOTE     : none
* CREATE   : 2008.12.08 M.Haguro
* UPDATE   :
******************************************************************************/
int ntsipu_img_request(void)
{
	int i, index = 0;
	int iRet;

	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");
	ntsipu->ntsc->img_complete = 0;

#ifdef NTSC_ERRATA
	errata->ntsc_upscale = (int)(ntshw_chk_control() & NTS_UPSCALE_BIT);
#endif /* NTSC_ERRATA */

	for(i = IMG_SRC_B; i < IMG_MAX; i++){
		memset(ntsipu->img[i], 0, sizeof(struct image_data));
	}
	for(i = 0; i < IPU_REQUEST_MAX; i++){
		memset(ntsipu->ipu_request[i], 0, sizeof(MAX_SIZE_MP200_IPU_REQ));
	}

	/* 1st request */
	iRet = ntsipu_chk_ipuimg();
	if(iRet != -EINVAL){
		iRet = ntsipu_set_ipuimg(index, iRet);
		if(iRet){
			printk_dbg((_DEBUG_NTSIPU & 0x02), "(%d) error\n", __LINE__);
			goto failed;
		}

		ntsipu->ipu_request[index]->callback    = ntsipu_callback_yuv422;
		ntsipu->ipu_request[index]->ipu_hw_func = IPU_HW_FUNC_IMG;
		ntsipu->ipu_request[index]->data        = 0l;
		index++;

		/* 2nd request */
		iRet = ntsipu_chk_ipuimg_bottom();
		if(iRet != -EINVAL){
			iRet = ntsipu_set_ipuimg(index, iRet);
			if(iRet){
				printk_dbg((_DEBUG_NTSIPU & 0x02), "(%d) error\n", __LINE__);
				goto failed;
			}

			ntsipu->ipu_request[index -1]->callback = ntsipu_callback_rgb565;
			ntsipu->ipu_request[index]->callback    = ntsipu_callback_yuv422;
			ntsipu->ipu_request[index]->ipu_hw_func = IPU_HW_FUNC_IMG;
			ntsipu->ipu_request[index]->data        = 0l;
			index++;
		}
		iRet = 0;

		for(i = 0; i < index; i++){
			/* request IPU */
			iRet = mp200_ipu_regist((struct mp200_ipu_req *)(ntsipu->ipu_request[i]));
			if(iRet){
				printk_dbg((_DEBUG_NTSIPU & 0x02), "ipu_regist(%d)\n", iRet);
				goto failed;
			}
		}
	}
failed:
	return iRet;
}


/* ------------------ private function < IMG > ---------------------------------- */
#ifdef NTSC_ERRATA
/******************************************************************************
* MODULE   : ntsipu_dma_request
* FUNCTION : 
* RETURN   : 
* NOTE     : 
* CREATE   : 2009.3.24 M.Haguro
* UPDATE   :
******************************************************************************/
static int ntsipu_dma_request(void)
{
	int ret;
	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	if(ntsipu->ntsc->iFrameNoNow == NTS_DISP_FRAME_A){
		errata->srcV = errata->destV = SMEM_START + NTSCFRAME_B_OFFSET + (ntsipu->img[IMG_DST]->uiY) * ntsipu->img[IMG_DST]->uiScreenWidth;
	}
	else{
		errata->srcV = errata->destV = SMEM_START + NTSCFRAME_A_OFFSET + (ntsipu->img[IMG_DST]->uiY) * ntsipu->img[IMG_DST]->uiScreenWidth;
	}
	if(ntsipu->img[IMG_DST]->uiY % 2){
		errata->srcV  += NTSC_WIDTH;
		errata->destV -= NTSC_WIDTH;
	}
	else{
		errata->srcV  += NTSC_WIDTH*2;
		errata->destV += 0;
	}

	/* copy Y plane data */
	errata->dma_regs->aadd = errata->srcV;
	errata->dma_regs->badd = errata->destV;
	errata->dma_regs->aoff = NTSC_WIDTH;
	errata->dma_regs->boff = NTSC_WIDTH;
	errata->dma_regs->size = NTSC_WIDTH;
	errata->dma_regs->leng = NTSC_WIDTH * (ntsipu->img[IMG_DST]->uiHeight/2 +1);
	errata->dma_regs->mode = MP200_DMAC_DEFMODE_32BIT;
	errata->ntsc_dma_stat  = 0;
	ret = mp200_start_dma( MP200_DMAC_M2M_ACPU_LCH3, errata->srcV, 0, errata->destV,
							MP200_DMAC_INT_ERROR_EN | MP200_DMAC_INT_LENG_EN );
	if( ret != 0 ) {
		unsigned long flags;
		spin_lock_irqsave(&ntsipu->ntsc->lock_callback, flags);
		printk(KERN_INFO "[Error]ntsipu_callback_yuv422 : mp200_start_dma(Y Plane)\n");
		ntsipu_callback_yuv422_return_ntsc();
		spin_unlock_irqrestore(&ntsipu->ntsc->lock_callback, flags);
	} else {
		queue_work(errata->ntsipu_workqueue, &errata->wk_dma_request_bottom);
	}
	return ret;
}


static void ntsipu_dma_request_bottom(struct work_struct *num)
{
	int ret;
	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	wait_event_interruptible(errata->queue_p, (errata->ntsc_dma_stat == NTSC_DMA_CALLBACK_COMP));

	/* copy UV interleave data */
	errata->srcV           = errata->srcV  + NTS_WIDTH * NTS_HEIGHT;
	errata->destV          = errata->destV + NTS_WIDTH * NTS_HEIGHT;
	errata->dma_regs->aadd = errata->srcV;
	errata->dma_regs->badd = errata->destV;
	errata->ntsc_dma_stat  = 0;
	ret = mp200_start_dma( MP200_DMAC_M2M_ACPU_LCH3, errata->srcV, 0, errata->destV,
							MP200_DMAC_INT_ERROR_EN | MP200_DMAC_INT_LENG_EN );
	if( ret != 0 ) {
		unsigned long flags;
		spin_lock_irqsave(&ntsipu->ntsc->lock_callback, flags);
		printk(KERN_INFO "[Error]ntsipu_callback_yuv422 : mp200_start_dma(UV Plane)\n");
		ntsipu_callback_yuv422_return_ntsc();
		spin_unlock_irqrestore(&ntsipu->ntsc->lock_callback, flags);
	} else {
		wait_event_interruptible(errata->queue_p, (errata->ntsc_dma_stat == NTSC_DMA_CALLBACK_COMP));

		ntsipu_callback_yuv422_bottom();
	}
}
#endif /* NTSC_ERRATA */


/******************************************************************************
* MODULE   : ntsipu_callback_yuv422
* FUNCTION : changing frame and callback to fb and v4l2
* RETURN   : none
* NOTE     : none
* CREATE   : 2008.11.14 M.Haguro
* UPDATE   :
******************************************************************************/
static void ntsipu_callback_yuv422(unsigned long sequence, int flag)
{
	unsigned long flags;
	spin_lock_irqsave(&ntsipu->ntsc->lock_callback, flags);
	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	/* if IMG success, change frame execute */
	if(flag){
		/* IMG error */
		ntsipu_callback_yuv422_return_ntsc();
	}
	else{
#ifdef NTSC_ERRATA
		if(ntsipu->ntsc->iMixImageMode == V4L2_ONLY || ntsipu->ntsc->iMixImageMode == FB_ONLY){
			nts_return_callback_refresh(); /* refresh callback (informed v4l2 or fb1 of budder free) */
		}

		if(ntsipu->ntsc->iOutmodeFlg == NTS_OUTPUT_NTSC && errata->ntsc_upscale == NTS_UPSCALE_OFF){
			ntsipu_dma_request();
		}
		else{
			ntsipu_callback_yuv422_bottom();
		}
#else
		ntsipu_callback_yuv422_bottom();
#endif /* NTSC_ERRATA */
	}
	spin_unlock_irqrestore(&ntsipu->ntsc->lock_callback, flags);
}


static void ntsipu_callback_yuv422_bottom(void)
{
	unsigned long flags;
	spin_lock_irqsave(&ntsipu->ntsc->lock_callback, flags);
	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	/* frame change */
	if(ntsipu->ntsc->iFrameNoNow == NTS_DISP_FRAME_A){
		ntsipu->ntsc->iFrameNoNext = NTS_DISP_FRAME_B;
		printk_dbg((_DEBUG_NTSIPU & 0x02), "now(A) -> next(B)\n");
	}
	else{
		ntsipu->ntsc->iFrameNoNext = NTS_DISP_FRAME_A;
		printk_dbg((_DEBUG_NTSIPU & 0x02), "now(B) -> next(A)\n");
	}
	ntshw_set_framesel(ntsipu->ntsc->iFrameNoNext);
	ntshw_set_ntsout(NTS_NTSOUT_ENABLE);

	/* enable NTSC VSYNC interrupt */
	ntshw_set_interruput(NTSHW_INTENSET);

	if(ntsipu->ntsc->iMixImageMode == V4L2_ONLY || ntsipu->ntsc->iMixImageMode == FB_V4L2_ON){
		nts_return_callback_ready();   /* ready callback (informed v4l2 of IPU complete time) */
	}
#ifndef NTSC_ERRATA
	if(ntsipu->ntsc->iMixImageMode == V4L2_ONLY || ntsipu->ntsc->iMixImageMode == FB_ONLY){
		nts_return_callback_refresh(); /* refresh callback (informed v4l2 or fb1 of budder free) */
	}
#endif /* NTSC_ERRATA */

	spin_unlock_irqrestore(&ntsipu->ntsc->lock_callback, flags);
}


static void ntsipu_callback_yuv422_return_ntsc(void)
{
	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	if(ntsipu->ntsc->iMixImageMode == V4L2_ONLY || ntsipu->ntsc->iMixImageMode == FB_ONLY){
		nts_return_callback_refresh(); /* refresh callback (informed v4l2 or fb1 of budder free) */
		nts_return_callback();
	}
	else{ /* ntsipu->ntsc->iMixImageMode == FB_V4L2_ON */
		if(ntsipu->ntsc->img_complete == 0){
			nts_return_callback();
		}
	}
}


/******************************************************************************
* MODULE   : ntsipu_callback_rgb565
* FUNCTION : changing frame and callback to fb and v4l2
* RETURN   : none
* NOTE     : none
* CREATE   : 2008.12.08 M.Haguro
* UPDATE   :
******************************************************************************/
static void ntsipu_callback_rgb565(unsigned long sequence, int flag)
{
	unsigned long flags;
	spin_lock_irqsave(&ntsipu->ntsc->lock_callback, flags);
	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	nts_return_callback_refresh(); /* refresh callback (informed v4l2 or fb1 of budder free) */
	if(flag){
		/* IMG error */
		nts_return_callback();
	}
	spin_unlock_irqrestore(&ntsipu->ntsc->lock_callback, flags);
}


/*****************************************************************************
* MODULE   : ntsipu_chk_ipuimg
* FUNCTION : 
* RETURN   : -
* NOTE     : 
* CREATE   : 2008.11.14 M.Haguro
* UPDATE   : 
******************************************************************************/
static int ntsipu_chk_ipuimg(void)
{
	int iRet;

	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	switch(ntsipu->ntsc->iMixImageMode){
	case V4L2_ONLY:
		/* copy Src(Back) data */
		ntsipu->img[IMG_SRC_B]->uiWidth        = ntsipu->ntsc->from_v4l2.image_data.hsize;
		ntsipu->img[IMG_SRC_B]->uiHeight       = ntsipu->ntsc->from_v4l2.image_data.vsize;
		ntsipu->img[IMG_SRC_B]->uiX            = ntsipu->ntsc->from_v4l2.image_data.x;
		ntsipu->img[IMG_SRC_B]->uiY            = ntsipu->ntsc->from_v4l2.image_data.y;
		ntsipu->img[IMG_SRC_B]->uiScreenWidth  = ntsipu->ntsc->from_v4l2.image_data.size;
		ntsipu->img[IMG_SRC_B]->uiFormat       = ntsipu->ntsc->from_v4l2.yuvfmt;
		ntsipu->img[IMG_SRC_B]->uiEndian       = ntsipu->ntsc->from_v4l2.endian << IMG_PEL_ENDIAN_IN_ENDI_SFT;

		ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB = ntsipu->ntsc->from_v4l2.image_data.yrgbaddr;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrUV   = ntsipu->ntsc->from_v4l2.image_data.uvaddr;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrV    = ntsipu->ntsc->from_v4l2.image_data.vaddr;

		/* copy Destination data */
		ntsipu->img[IMG_DST]->uiWidth          = ntsipu->ntsc->from_v4l2.screen_data.hsize;
		ntsipu->img[IMG_DST]->uiHeight         = ntsipu->ntsc->from_v4l2.screen_data.vsize;
		ntsipu->img[IMG_DST]->uiX              = ntsipu->ntsc->from_v4l2.screen_data.x;
		ntsipu->img[IMG_DST]->uiY              = ntsipu->ntsc->from_v4l2.screen_data.y;
		ntsipu->img[IMG_DST]->uiScreenWidth    = NTS_WIDTH;
		ntsipu->img[IMG_DST]->uiFormat         = IMG_FORMAT_YUV422_SEMI_PLANAR;
		ntsipu->img[IMG_DST]->uiEndian         = IMG_PEL_ENDIAN_OUT_ENDI_LITTLE;

		if(ntsipu->ntsc->iFrameNoNow == NTS_DISP_FRAME_A){
			ntsipu->img[IMG_DST]->ulPhysAddrYRGB = SMEM_START + NTSCFRAME_B_OFFSET;
		}
		else{
			ntsipu->img[IMG_DST]->ulPhysAddrYRGB = SMEM_START + NTSCFRAME_A_OFFSET;
		}
		ntsipu->img[IMG_DST]->ulPhysAddrUV     = ntsipu->img[IMG_DST]->ulPhysAddrYRGB + NTS_WIDTH * NTS_HEIGHT;
		ntsipu->img[IMG_DST]->ulPhysAddrV      = 0;

		if((ntsipu->img[IMG_SRC_B]->uiWidth == ntsipu->img[IMG_DST]->uiWidth) && (ntsipu->img[IMG_SRC_B]->uiHeight == ntsipu->img[IMG_DST]->uiHeight)){
			/* IMG -- Color Format Convert */
			printk_dbg((_DEBUG_NTSIPU & 0x02), "V4L2 -- Color Format Convert\n");
			iRet = IMG_OP_MODE_01_BIT;
		}
		else{
			/* IMG -- Resize & Color Format Convert */
			printk_dbg((_DEBUG_NTSIPU & 0x02), "V4L2 -- Resize & Color Format Convert\n");
			iRet = IMG_OP_MODE_03_BIT;
		}
		printk_dbg((_DEBUG_NTSIPU & 0x02), "=== ntsipu_chk_ipuimg(V4L2_ONLY) =============================\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " SRC_B | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB, ntsipu->img[IMG_SRC_B]->ulPhysAddrUV, ntsipu->img[IMG_SRC_B]->ulPhysAddrV);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_SRC_B]->uiScreenWidth, ntsipu->img[IMG_SRC_B]->uiWidth, ntsipu->img[IMG_SRC_B]->uiHeight, ntsipu->img[IMG_SRC_B]->uiX, ntsipu->img[IMG_SRC_B]->uiY);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "-------+------------------------------------------------------\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " DST   | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_DST]->ulPhysAddrYRGB, ntsipu->img[IMG_DST]->ulPhysAddrUV, ntsipu->img[IMG_DST]->ulPhysAddrV);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_DST]->uiScreenWidth, ntsipu->img[IMG_DST]->uiWidth, ntsipu->img[IMG_DST]->uiHeight, ntsipu->img[IMG_DST]->uiX, ntsipu->img[IMG_DST]->uiY);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "==============================================================\n\n");

		/* check framebuffer fillrect to GDMA */
		ntsipu_chk_ipugdma();
		break;
#ifdef CONFIG_FB_MP200_NTS
	case FB_ONLY:
		/* copy Src(Back) data */
		ntsipu->img[IMG_SRC_B]->uiWidth        = ntsipu->ntsc->from_fb.image_data.hsize;
		ntsipu->img[IMG_SRC_B]->uiHeight       = ntsipu->ntsc->from_fb.image_data.vsize;
		ntsipu->img[IMG_SRC_B]->uiX            = ntsipu->ntsc->from_fb.image_data.x;
		ntsipu->img[IMG_SRC_B]->uiY            = ntsipu->ntsc->from_fb.image_data.y;
		ntsipu->img[IMG_SRC_B]->uiScreenWidth  = ntsipu->ntsc->from_fb.image_data.size;
		ntsipu->img[IMG_SRC_B]->uiFormat       = IMG_FORMAT_RGB565;
		ntsipu->img[IMG_SRC_B]->uiEndian       = IMG_PEL_ENDIAN_IN_ENDI_LITTLE;

		ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB = ntsipu->ntsc->from_fb.image_data.yrgbaddr;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrUV   = 0;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrV    = 0;

		/* copy Destination data */
		ntsipu->img[IMG_DST]->uiWidth          = ntsipu->ntsc->from_fb.image_data.hsize;
		ntsipu->img[IMG_DST]->uiHeight         = ntsipu->ntsc->from_fb.image_data.vsize;
		ntsipu->img[IMG_DST]->uiX              = 0;
		ntsipu->img[IMG_DST]->uiY              = 0;
		ntsipu->img[IMG_DST]->uiScreenWidth    = NTS_WIDTH;
		ntsipu->img[IMG_DST]->uiFormat         = IMG_FORMAT_YUV422_SEMI_PLANAR;
		ntsipu->img[IMG_DST]->uiEndian         = IMG_PEL_ENDIAN_OUT_ENDI_LITTLE;

		if(ntsipu->ntsc->iFrameNoNow == NTS_DISP_FRAME_A){
			ntsipu->img[IMG_DST]->ulPhysAddrYRGB = SMEM_START + NTSCFRAME_B_OFFSET;
		}
		else{
			ntsipu->img[IMG_DST]->ulPhysAddrYRGB = SMEM_START + NTSCFRAME_A_OFFSET;
		}
		ntsipu->img[IMG_DST]->ulPhysAddrUV     = ntsipu->img[IMG_DST]->ulPhysAddrYRGB + NTS_WIDTH * NTS_HEIGHT;
		ntsipu->img[IMG_DST]->ulPhysAddrV      = 0;

		/* IMG -- Color Format Convert */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "FB -- Color Format Conver\n");
		iRet = IMG_OP_MODE_06_BIT;

		printk_dbg((_DEBUG_NTSIPU & 0x02), "=== ntsipu_chk_ipuimg(FB_ONLY) ===============================\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " SRC_B | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB, ntsipu->img[IMG_SRC_B]->ulPhysAddrUV, ntsipu->img[IMG_SRC_B]->ulPhysAddrV);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_SRC_B]->uiScreenWidth, ntsipu->img[IMG_SRC_B]->uiWidth, ntsipu->img[IMG_SRC_B]->uiHeight, ntsipu->img[IMG_SRC_B]->uiX, ntsipu->img[IMG_SRC_B]->uiY);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "-------+------------------------------------------------------\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " DST   | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_DST]->ulPhysAddrYRGB, ntsipu->img[IMG_DST]->ulPhysAddrUV, ntsipu->img[IMG_DST]->ulPhysAddrV);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_DST]->uiScreenWidth, ntsipu->img[IMG_DST]->uiWidth, ntsipu->img[IMG_DST]->uiHeight, ntsipu->img[IMG_DST]->uiX, ntsipu->img[IMG_DST]->uiY);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "==============================================================\n\n");
		break;
	case FB_V4L2_ON:
		/* copy Src(Front) data */
		ntsipu->img[IMG_SRC_F]->uiWidth        = ntsipu->ntsc->from_fb.image_data.hsize;
		ntsipu->img[IMG_SRC_F]->uiHeight       = ntsipu->ntsc->from_fb.image_data.vsize;
		ntsipu->img[IMG_SRC_F]->uiX            = ntsipu->ntsc->from_fb.image_data.x;
		ntsipu->img[IMG_SRC_F]->uiY            = ntsipu->ntsc->from_fb.image_data.y;
		ntsipu->img[IMG_SRC_F]->uiScreenWidth  = ntsipu->ntsc->from_fb.image_data.size;
		ntsipu->img[IMG_SRC_F]->uiFormat       = IMG_FORMAT_RGB565;
		ntsipu->img[IMG_SRC_F]->uiEndian       = IMG_PEL_ENDIAN_IN_ENDI_LITTLE;

		ntsipu->img[IMG_SRC_F]->ulPhysAddrYRGB = ntsipu->ntsc->from_fb.image_data.yrgbaddr;
		ntsipu->img[IMG_SRC_F]->ulPhysAddrUV   = 0;
		ntsipu->img[IMG_SRC_F]->ulPhysAddrV    = 0;

		/* copy Src(Back) data */
		ntsipu->img[IMG_SRC_B]->uiWidth        = ntsipu->ntsc->from_v4l2.image_data.hsize;
		ntsipu->img[IMG_SRC_B]->uiHeight       = ntsipu->ntsc->from_v4l2.image_data.vsize;
		ntsipu->img[IMG_SRC_B]->uiX            = ntsipu->ntsc->from_v4l2.image_data.x;
		ntsipu->img[IMG_SRC_B]->uiY            = ntsipu->ntsc->from_v4l2.image_data.y;
		ntsipu->img[IMG_SRC_B]->uiScreenWidth  = ntsipu->ntsc->from_v4l2.image_data.size;
		ntsipu->img[IMG_SRC_B]->uiFormat       = ntsipu->ntsc->from_v4l2.yuvfmt;
		ntsipu->img[IMG_SRC_B]->uiEndian       = ntsipu->ntsc->from_v4l2.endian << IMG_PEL_ENDIAN_IN_ENDI_SFT;

		ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB = ntsipu->ntsc->from_v4l2.image_data.yrgbaddr;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrUV   = ntsipu->ntsc->from_v4l2.image_data.uvaddr;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrV    = ntsipu->ntsc->from_v4l2.image_data.vaddr;

		/* copy Destination data */
		ntsipu->img[IMG_DST]->uiWidth          = ntsipu->ntsc->from_v4l2.screen_data.hsize;
		ntsipu->img[IMG_DST]->uiHeight         = ntsipu->ntsc->from_v4l2.screen_data.vsize;
		ntsipu->img[IMG_DST]->uiX              = ntsipu->ntsc->from_v4l2.screen_data.x;
		ntsipu->img[IMG_DST]->uiY              = ntsipu->ntsc->from_v4l2.screen_data.y;
#if 0
//@		ntsipu->img[IMG_DST]->uiScreenWidth    = NTS_WIDTH*3;
//@		ntsipu->img[IMG_DST]->uiFormat         = IMG_FORMAT_RGB888;
#else
		ntsipu->img[IMG_DST]->uiScreenWidth    = NTS_WIDTH*2;
		ntsipu->img[IMG_DST]->uiFormat         = IMG_FORMAT_RGB565;
#endif
		ntsipu->img[IMG_DST]->uiEndian         = IMG_PEL_ENDIAN_OUT_ENDI_LITTLE;

		ntsipu->img[IMG_DST]->ulPhysAddrYRGB   = SMEM_START + NTSCWORK_OFFSET;
		ntsipu->img[IMG_DST]->ulPhysAddrUV     = 0;
		ntsipu->img[IMG_DST]->ulPhysAddrV      = 0;

		/* IMG -- Resize & Color Format Convert & Alpha bleading */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "FB & V4L2 -- Resize & Color Format Convert & Alpha bleading\n");
		iRet = IMG_OP_MODE_18_BIT;

		printk_dbg((_DEBUG_NTSIPU & 0x02), "=== ntsipu_chk_ipuimg(FB_V4L2_ON) ============================\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " SRC_F | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_SRC_F]->ulPhysAddrYRGB, ntsipu->img[IMG_SRC_F]->ulPhysAddrUV, ntsipu->img[IMG_SRC_F]->ulPhysAddrV);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_SRC_F]->uiScreenWidth, ntsipu->img[IMG_SRC_F]->uiWidth, ntsipu->img[IMG_SRC_F]->uiHeight, ntsipu->img[IMG_SRC_F]->uiX, ntsipu->img[IMG_SRC_F]->uiY);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "-------+------------------------------------------------------\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " SRC_B | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB, ntsipu->img[IMG_SRC_B]->ulPhysAddrUV, ntsipu->img[IMG_SRC_B]->ulPhysAddrV);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_SRC_B]->uiScreenWidth, ntsipu->img[IMG_SRC_B]->uiWidth, ntsipu->img[IMG_SRC_B]->uiHeight, ntsipu->img[IMG_SRC_B]->uiX, ntsipu->img[IMG_SRC_B]->uiY);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "-------+------------------------------------------------------\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " DST   | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_DST]->ulPhysAddrYRGB, ntsipu->img[IMG_DST]->ulPhysAddrUV, ntsipu->img[IMG_DST]->ulPhysAddrV);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_DST]->uiScreenWidth, ntsipu->img[IMG_DST]->uiWidth, ntsipu->img[IMG_DST]->uiHeight, ntsipu->img[IMG_DST]->uiX, ntsipu->img[IMG_DST]->uiY);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "==============================================================\n\n");
		break;
#endif /* CONFIG_FB_MP200_NTS */
	default:
		iRet = -EINVAL;
		break;
	}
	printk_dbg((_DEBUG_NTSIPU & 0x02), "mode(%d)  return(%x)\n", ntsipu->ntsc->iMixImageMode, iRet);
	return iRet;
}


/*****************************************************************************
* MODULE   : ntsipu_chk_ipuimg_bottom
* FUNCTION : 
* RETURN   : -
* NOTE     : 
* CREATE   : 2008.12.04 M.Haguro
* UPDATE   : 
******************************************************************************/
static int ntsipu_chk_ipuimg_bottom(void)
{
	int iRet;
	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	switch(ntsipu->ntsc->iMixImageMode){
#ifdef CONFIG_FB_MP200_NTS
	case FB_V4L2_ON:
		/* copy Src(Back) data */
		ntsipu->img[IMG_SRC_B]->uiWidth        = ntsipu->ntsc->from_fb.image_data.hsize;
		ntsipu->img[IMG_SRC_B]->uiHeight       = ntsipu->ntsc->from_fb.image_data.vsize;
		ntsipu->img[IMG_SRC_B]->uiX            = 0;
		ntsipu->img[IMG_SRC_B]->uiY            = 0;
		ntsipu->img[IMG_SRC_B]->uiScreenWidth  = ntsipu->img[IMG_DST]->uiScreenWidth;
		ntsipu->img[IMG_SRC_B]->uiFormat       = ntsipu->img[IMG_DST]->uiFormat;
		ntsipu->img[IMG_SRC_B]->uiEndian       = ntsipu->img[IMG_DST]->uiEndian;

		ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB = ntsipu->img[IMG_DST]->ulPhysAddrYRGB;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrUV   = ntsipu->img[IMG_DST]->ulPhysAddrUV;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrV    = ntsipu->img[IMG_DST]->ulPhysAddrV;

		/* copy Destination data */
		ntsipu->img[IMG_DST]->uiWidth          = ntsipu->ntsc->from_fb.image_data.hsize;
		ntsipu->img[IMG_DST]->uiHeight         = ntsipu->ntsc->from_fb.image_data.vsize;
		ntsipu->img[IMG_DST]->uiX              = 0;
		ntsipu->img[IMG_DST]->uiY              = 0;
		ntsipu->img[IMG_DST]->uiScreenWidth    = NTS_WIDTH;
		ntsipu->img[IMG_DST]->uiFormat         = IMG_FORMAT_YUV422_SEMI_PLANAR;
		ntsipu->img[IMG_DST]->uiEndian         = IMG_PEL_ENDIAN_OUT_ENDI_LITTLE;

		if(ntsipu->ntsc->iFrameNoNow == NTS_DISP_FRAME_A){
			ntsipu->img[IMG_DST]->ulPhysAddrYRGB = SMEM_START + NTSCFRAME_B_OFFSET;
		}
		else{
			ntsipu->img[IMG_DST]->ulPhysAddrYRGB = SMEM_START + NTSCFRAME_A_OFFSET;
		}
		ntsipu->img[IMG_DST]->ulPhysAddrUV     = ntsipu->img[IMG_DST]->ulPhysAddrYRGB + NTS_WIDTH * NTS_HEIGHT;
		ntsipu->img[IMG_DST]->ulPhysAddrV      = 0;

		/* IMG -- Color Format Convert */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "FB & V4L2 -- Color Format Conver\n");
		iRet = IMG_OP_MODE_06_BIT;

		printk_dbg((_DEBUG_NTSIPU & 0x02), "=== ntsipu_chk_ipuimg() ======================================\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " SRC_B | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB, ntsipu->img[IMG_SRC_B]->ulPhysAddrUV, ntsipu->img[IMG_SRC_B]->ulPhysAddrV);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_SRC_B]->uiScreenWidth, ntsipu->img[IMG_SRC_B]->uiWidth, ntsipu->img[IMG_SRC_B]->uiHeight, ntsipu->img[IMG_SRC_B]->uiX, ntsipu->img[IMG_SRC_B]->uiY);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "-------+------------------------------------------------------\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " DST   | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_DST]->ulPhysAddrYRGB, ntsipu->img[IMG_DST]->ulPhysAddrUV, ntsipu->img[IMG_DST]->ulPhysAddrV);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_DST]->uiScreenWidth, ntsipu->img[IMG_DST]->uiWidth, ntsipu->img[IMG_DST]->uiHeight, ntsipu->img[IMG_DST]->uiX, ntsipu->img[IMG_DST]->uiY);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "==============================================================\n\n");
		break;
#endif /* CONFIG_FB_MP200_NTS */
	default:
		iRet = -EINVAL;
		break;
	}
	printk_dbg((_DEBUG_NTSIPU & 0x02), "mode(%d)  return(%x)\n", ntsipu->ntsc->iMixImageMode, iRet);
	return iRet;
}


/******************************************************************************
* MODULE   : ntsipu_set_imgmode
* FUNCTION : set IMG_MODE
* RETURN   : none
* NOTE     : none
* CREATE   : 2008.11.14 M.Haguro
* UPDATE   :
******************************************************************************/
static unsigned long ntsipu_set_imgmode(void)
{
	unsigned long ulImgMode;

	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	ulImgMode = IMG_NEWEN_BIT;

	/* vanish mask color */
	if(ntsipu->ntsc->from_fb.maskcolrflg == FB_MASK_COLOR_DISP_OFF){
		ulImgMode |= IMG_MASK_COL_BIT;
	}

	/* up/down and right/left inverse (180 rotate) */
	if(ntsipu->ntsc->from_fb.invflg == UDRL_INVERSE){
		ulImgMode |= (IMG_UDINV_BIT | IMG_RLINV_BIT);
	}
	return ulImgMode;
}


/******************************************************************************
* MODULE   : ntsipu_set_format
* FUNCTION : set IMG_MODE
* RETURN   : none
* NOTE     : none
* CREATE   : 2008.11.14 M.Haguro
* UPDATE   :
******************************************************************************/
static unsigned long ntsipu_set_format(void)
{
	int bpp_y, bpp_uv, bpp_v; /* bit par pixel  */
	int ppl_y, ppl_uv, ppl_v; /* pixel par line */
	unsigned int uiScreenPixcelWidthY, uiScreenPixcelWidthUV, uiScreenPixcelWidthV;

	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	/*
	 * Source(Back) image
	 */
	switch(ntsipu->img[IMG_SRC_B]->uiFormat){
	case IMG_FORMAT_YUV420_SEMI_PLANAR: /* YUV420 Semi-Planar */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "src_b_format(YUV420Pl2)\n");
		bpp_y = 8;  bpp_uv = 8; bpp_v = 0; ppl_y = 1; ppl_uv = 2; ppl_v = 0;
		break;
	case IMG_FORMAT_YUV420_PLANAR:      /* YUV420 Planar      */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "src_b_format(YUV420Pl)\n");
		bpp_y = 8;  bpp_uv = 4; bpp_v = 4; ppl_y = 1; ppl_uv = 2; ppl_v = 2;
		break;
	case IMG_FORMAT_YUV422_SEMI_PLANAR: /* YUV422 Semi-Planar */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "src_b_format(YUV422Pl2)\n");
		bpp_y = 8;  bpp_uv = 8; bpp_v = 0; ppl_y = 1; ppl_uv = 1; ppl_v = 0;
		break;
	case IMG_FORMAT_YUV422_PLANAR:      /* YUV422 Planar      */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "src_b_format(YUV422Pl)\n");
		bpp_y = 8;  bpp_uv = 8; bpp_v = 8; ppl_y = 1; ppl_uv = 2; ppl_v = 2;
		break;
	case IMG_FORMAT_YUV422_INTERLEAVE:  /* YUV422 Interleave  */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "src_b_format(YUV422Px)\n");
		bpp_y = 16; bpp_uv = 0; bpp_v = 0; ppl_y = 1; ppl_uv = 0; ppl_v = 0;
		break;
	case IMG_FORMAT_RGB565:             /* RGB565             */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "src_b_format(RGB565)\n");
		bpp_y = 16; bpp_uv = 0; bpp_v = 0; ppl_y = 1; ppl_uv = 0; ppl_v = 0;
		break;
	case IMG_FORMAT_RGB888:             /* RGB888             */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "src_b_format(RGB888)\n");
		bpp_y = 24; bpp_uv = 0; bpp_v = 0; ppl_y = 1; ppl_uv = 0; ppl_v = 0;
		break;
	default:
		printk_dbg((_DEBUG_NTSIPU & 0x02), "src_b_format(%d)\n", ntsipu->img[IMG_SRC_B]->uiFormat);
		return (unsigned long)-EINVAL;
	}
	uiScreenPixcelWidthY = ntsipu->img[IMG_SRC_B]->uiScreenWidth / bpp_y  * 8;
	if(bpp_uv) uiScreenPixcelWidthUV = ntsipu->img[IMG_SRC_B]->uiScreenWidth / bpp_uv * 8;
	else       uiScreenPixcelWidthUV = 0;
	if(bpp_v)  uiScreenPixcelWidthV  = ntsipu->img[IMG_SRC_B]->uiScreenWidth / bpp_v  * 8;
	else       uiScreenPixcelWidthV  = 0;

	switch(ntsipu->img[IMG_SRC_B]->uiFormat){
	case IMG_FORMAT_RGB565:             /* RGB565             */
	case IMG_FORMAT_RGB888:             /* RGB888             */
	case IMG_FORMAT_YUV422_INTERLEAVE:  /* YUV422 Interleave  */
		ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB = ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB + ntsipu->img[IMG_SRC_B]->uiY * uiScreenPixcelWidthY  / ppl_y  + ntsipu->img[IMG_SRC_B]->uiX * bpp_y  / 8;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrUV   = 0;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrV    = 0;
		break;
	case IMG_FORMAT_YUV420_SEMI_PLANAR: /* YUV420 Semi-Planar */
	case IMG_FORMAT_YUV422_SEMI_PLANAR: /* YUV422 Semi-Planar */
		ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB = ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB + ntsipu->img[IMG_SRC_B]->uiY * uiScreenPixcelWidthY  / ppl_y  + ntsipu->img[IMG_SRC_B]->uiX * bpp_y  / 8;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrUV   = ntsipu->img[IMG_SRC_B]->ulPhysAddrUV   + ntsipu->img[IMG_SRC_B]->uiY * uiScreenPixcelWidthUV / ppl_uv + ntsipu->img[IMG_SRC_B]->uiX * bpp_uv / 8;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrV    = 0;
		break;
	case IMG_FORMAT_YUV420_PLANAR:      /* YUV420 Planar      */
	case IMG_FORMAT_YUV422_PLANAR:      /* YUV422 Planar      */
		ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB = ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB + ntsipu->img[IMG_SRC_B]->uiY * uiScreenPixcelWidthY  / ppl_y  + ntsipu->img[IMG_SRC_B]->uiX * bpp_y  / 8;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrUV   = ntsipu->img[IMG_SRC_B]->ulPhysAddrUV   + ntsipu->img[IMG_SRC_B]->uiY * uiScreenPixcelWidthUV / ppl_uv + ntsipu->img[IMG_SRC_B]->uiX * bpp_uv / 8;
		ntsipu->img[IMG_SRC_B]->ulPhysAddrV    = ntsipu->img[IMG_SRC_B]->ulPhysAddrV    + ntsipu->img[IMG_SRC_B]->uiY * uiScreenPixcelWidthV / ppl_v  + ntsipu->img[IMG_SRC_B]->uiX * bpp_v  / 8;
		break;
	default:
		printk_dbg((_DEBUG_NTSIPU & 0x02), "src_b_format(%d)\n", ntsipu->img[IMG_SRC_B]->uiFormat);
		return (unsigned long)-EINVAL;
	}

	/*
	 * Source(Front) image
	 */
	switch(ntsipu->ntsc->iMixImageMode){
	case FB_V4L2_ON:
		switch(ntsipu->img[IMG_SRC_F]->uiFormat){
		case IMG_FORMAT_RGB565:             /* RGB565             */
			printk_dbg((_DEBUG_NTSIPU & 0x02), "src_f_format(RGB565)\n");
			bpp_y = 16; bpp_uv = 0; bpp_v = 0; ppl_y = 1; ppl_uv = 0; ppl_v = 0;
			break;
		default:
			printk_dbg((_DEBUG_NTSIPU & 0x02), "src_f_format(%d)\n", ntsipu->img[IMG_SRC_F]->uiFormat);
			return (unsigned long)-EINVAL;
		}
		uiScreenPixcelWidthY  = ntsipu->img[IMG_SRC_F]->uiScreenWidth / bpp_y  * 8;
		if(bpp_uv) uiScreenPixcelWidthUV = ntsipu->img[IMG_SRC_F]->uiScreenWidth / bpp_uv * 8;
		else       uiScreenPixcelWidthUV = 0;
		if(bpp_y)  uiScreenPixcelWidthV  = ntsipu->img[IMG_SRC_F]->uiScreenWidth / bpp_v  * 8;
		else       uiScreenPixcelWidthV  = 0;

		switch(ntsipu->img[IMG_SRC_F]->uiFormat){
		case IMG_FORMAT_RGB565:             /* RGB565             */
			ntsipu->img[IMG_SRC_F]->ulPhysAddrYRGB = ntsipu->img[IMG_SRC_F]->ulPhysAddrYRGB + ntsipu->img[IMG_SRC_F]->uiY * uiScreenPixcelWidthY / ppl_y  + ntsipu->img[IMG_SRC_F]->uiX * bpp_y  / 8;
			ntsipu->img[IMG_SRC_F]->ulPhysAddrUV   = 0;
			ntsipu->img[IMG_SRC_F]->ulPhysAddrV    = 0;
			break;
		default:
			printk_dbg((_DEBUG_NTSIPU & 0x02), "src_f_format(%d)\n", ntsipu->img[IMG_SRC_F]->uiFormat);
			return (unsigned long)-EINVAL;
		}
		break;
	case V4L2_ONLY:
	case FB_ONLY:
		break;
	default:
		return (unsigned long)-EINVAL;
	}

	/*
	 * Destination image
	 */
	switch(ntsipu->ntsc->iMixImageMode){
	case V4L2_ONLY:
	case FB_ONLY:
		switch(ntsipu->img[IMG_DST]->uiFormat){
		case IMG_FORMAT_YUV422_SEMI_PLANAR: /* YUV422 Semi-Planar */
			printk_dbg((_DEBUG_NTSIPU & 0x02), "dst_format(YUV422Pl2)\n");
			bpp_y = 8;  bpp_uv = 8; bpp_v = 0; ppl_y = 1; ppl_uv = 1; ppl_v = 0;
			break;
		default:
			printk_dbg((_DEBUG_NTSIPU & 0x02), "dst_format(%d)\n", ntsipu->img[IMG_DST]->uiFormat);
			return (unsigned long)-EINVAL;
		}
		uiScreenPixcelWidthY = ntsipu->img[IMG_DST]->uiScreenWidth / bpp_y  * 8;
		if(bpp_uv) uiScreenPixcelWidthUV = ntsipu->img[IMG_DST]->uiScreenWidth / bpp_uv * 8;
		else       uiScreenPixcelWidthUV = 0;
		if(bpp_v)  uiScreenPixcelWidthV  = ntsipu->img[IMG_DST]->uiScreenWidth / bpp_v  * 8;
		else       uiScreenPixcelWidthV  = 0;

		switch(ntsipu->img[IMG_DST]->uiFormat){
		case IMG_FORMAT_YUV422_SEMI_PLANAR: /* YUV422 Semi-Planar */
			ntsipu->img[IMG_DST]->ulPhysAddrYRGB = ntsipu->img[IMG_DST]->ulPhysAddrYRGB + ntsipu->img[IMG_DST]->uiY * uiScreenPixcelWidthY  / ppl_y  + ntsipu->img[IMG_DST]->uiX * bpp_y  / 8;
			ntsipu->img[IMG_DST]->ulPhysAddrUV   = ntsipu->img[IMG_DST]->ulPhysAddrUV   + ntsipu->img[IMG_DST]->uiY * uiScreenPixcelWidthUV / ppl_uv + ntsipu->img[IMG_DST]->uiX * bpp_uv / 8;
			ntsipu->img[IMG_DST]->ulPhysAddrV    = 0;
			break;
		default:
			printk_dbg((_DEBUG_NTSIPU & 0x02), "dst_format(%d)\n", ntsipu->img[IMG_DST]->uiFormat);
			return (unsigned long)-EINVAL;
		}
		break;
	case FB_V4L2_ON:
		switch(ntsipu->img[IMG_DST]->uiFormat){
		case IMG_FORMAT_YUV422_SEMI_PLANAR: /* YUV422 Semi-Planar */
			printk_dbg((_DEBUG_NTSIPU & 0x02), "dst_format(YUV422Pl2)\n");
			break;
		case IMG_FORMAT_RGB888:             /* RGB888             */
			printk_dbg((_DEBUG_NTSIPU & 0x02), "dst_format(RGB888)\n");
			break;
		case IMG_FORMAT_RGB565:             /* RGB565             */
			printk_dbg((_DEBUG_NTSIPU & 0x02), "dst_format(RGB565)\n");
			break;
		default:
			printk_dbg((_DEBUG_NTSIPU & 0x02), "dst_format(%d)\n", ntsipu->img[IMG_DST]->uiFormat);
			return (unsigned long)-EINVAL;
		}
		break;
	default:
		return (unsigned long)-EINVAL;
	}

	return (unsigned long)(ntsipu->img[IMG_DST]->uiFormat << IMG_FORMAT_DSTFMT_SFT) | (ntsipu->img[IMG_SRC_F]->uiFormat << IMG_FORMAT_SRCFMT_F_SFT) | (ntsipu->img[IMG_SRC_B]->uiFormat << IMG_FORMAT_SRCFMT_B_SFT);
}


/*****************************************************************************
* MODULE   : ntsipu_set_ipuimg
* FUNCTION : make rotate request data to IPU-IMG driver.
* RETURN   :  0: success
*          : -1: error
* NOTE     : The register information table make, requested to the IMG driver.
* CREATE   : 2008.11.13 M.Haguro
* UPDATE   : 
******************************************************************************/
static int ntsipu_set_ipuimg(int index, int img_op_mode)
{
	unsigned long ulImgMode, ulFormat, ulEndian;
	int index_req = 0;

	printk_dbg((_DEBUG_NTSIPU & 0x01), "index(%d)\n", index);

	/*******************************
	    set IMG_MODE
	********************************/
	ulImgMode = (unsigned long)img_op_mode | ntsipu_set_imgmode();

	/*******************************
	    set ENDIAN
	********************************/
	ulEndian = ntsipu->img[IMG_SRC_B]->uiEndian | ntsipu->img[IMG_DST]->uiEndian;

	/*******************************
	    set IMG_FORMAT
	********************************/
	ulFormat = ntsipu_set_format();
	if(ulFormat == (unsigned long)-EINVAL){
		printk_dbg((_DEBUG_NTSIPU & 0x02), "(%d) error\n", __LINE__);
		return (int)ulFormat;
	}

	/*******************************
	    set IMG request data
	********************************/
	/*==============================
	    IMG_MODE
	================================*/
	IpuRequest(ulImgMode,                                      IMG_MODE);
	IpuRequest(ulFormat,                                       IMG_FORMAT);
	IpuRequest(ulEndian,                                       IMG_PEL_ENDIAN);

	/*==============================
	    source Back image
	================================*/
	IpuRequest(ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB,         IMG_SRCYRGBADR_B);
	IpuRequest(ntsipu->img[IMG_SRC_B]->ulPhysAddrUV,           IMG_SRCUVADR_B);
	IpuRequest(ntsipu->img[IMG_SRC_B]->ulPhysAddrV,            IMG_SRCVADR_B);
	IpuRequest(ntsipu->img[IMG_SRC_B]->uiScreenWidth,          IMG_SRCSIZE_B);
	IpuRequest(ntsipu->img[IMG_SRC_B]->uiWidth,                IMG_SRCHSIZE_B);
	IpuRequest(ntsipu->img[IMG_SRC_B]->uiHeight,               IMG_SRCVSIZE_B);

	/*==============================
	    destination image
	================================*/
#ifdef NTSC_ERRATA
	if(ntsipu->ntsc->iOutmodeFlg == NTS_OUTPUT_NTSC && errata->ntsc_upscale == NTS_UPSCALE_OFF && ntsipu->img[IMG_DST]->uiFormat == IMG_FORMAT_YUV422_SEMI_PLANAR){
		IpuRequest(ntsipu->img[IMG_DST]->ulPhysAddrYRGB + NTSC_WIDTH, IMG_DSTYRGBADR);
		IpuRequest(ntsipu->img[IMG_DST]->ulPhysAddrUV   + NTSC_WIDTH, IMG_DSTUVADR);
		IpuRequest(ntsipu->img[IMG_DST]->ulPhysAddrV,                 IMG_DSTVADR);
	}
	else{
		IpuRequest(ntsipu->img[IMG_DST]->ulPhysAddrYRGB,   IMG_DSTYRGBADR);
		IpuRequest(ntsipu->img[IMG_DST]->ulPhysAddrUV,     IMG_DSTUVADR);
		IpuRequest(ntsipu->img[IMG_DST]->ulPhysAddrV,      IMG_DSTVADR);
	}
#else
	IpuRequest(ntsipu->img[IMG_DST]->ulPhysAddrYRGB,           IMG_DSTYRGBADR);
	IpuRequest(ntsipu->img[IMG_DST]->ulPhysAddrUV,             IMG_DSTUVADR);
	IpuRequest(ntsipu->img[IMG_DST]->ulPhysAddrV,              IMG_DSTVADR);
#endif /* NTSC_ERRATA */
	IpuRequest(ntsipu->img[IMG_DST]->uiScreenWidth,            IMG_DSTSIZE);

	if(((ulImgMode & IMG_OP_MODE_BIT) == IMG_OP_MODE_03_BIT) || ((ulImgMode & IMG_OP_MODE_BIT) == IMG_OP_MODE_18_BIT)){
		unsigned long step_h, step_v, fold_h, fold_v;

		/* resize */
		IpuRequest(ntsipu->img[IMG_DST]->uiWidth,          IMG_DSTHSIZE);
		IpuRequest(ntsipu->img[IMG_DST]->uiHeight,         IMG_DSTVSIZE);

		/* step */
		step_h = ntsipu->img[IMG_SRC_B]->uiWidth  * 256 / ntsipu->img[IMG_DST]->uiWidth;
		step_v = ntsipu->img[IMG_SRC_B]->uiHeight * 256 / ntsipu->img[IMG_DST]->uiHeight;
		IpuRequest(step_h,                                 IMG_HSTEP);
		IpuRequest(step_v,                                 IMG_VSTEP);

		/* fold */
		fold_h = ntsipu->img[IMG_DST]->uiWidth  * 64 / ntsipu->img[IMG_SRC_B]->uiWidth;
		fold_v = ntsipu->img[IMG_DST]->uiHeight * 64 / ntsipu->img[IMG_SRC_B]->uiHeight;
		IpuRequest(fold_h,                                 IMG_HFOLD);
		IpuRequest(fold_v,                                 IMG_VFOLD);
	}

	if(ulImgMode & IMG_OVERLAY_BIT){
		/*==============================
		    source Front image
		================================*/
		IpuRequest(ntsipu->img[IMG_SRC_F]->ulPhysAddrYRGB, IMG_SRCYRGBADR_F);
		IpuRequest(ntsipu->img[IMG_SRC_F]->ulPhysAddrUV,   IMG_SRCUVADR_F);
		IpuRequest(ntsipu->img[IMG_SRC_F]->ulPhysAddrV,    IMG_SRCVADR_F);
		IpuRequest(ntsipu->img[IMG_SRC_F]->uiScreenWidth,  IMG_SRCSIZE_F);
		IpuRequest(ntsipu->img[IMG_SRC_F]->uiWidth,        IMG_SRCHSIZE_F);
		IpuRequest(ntsipu->img[IMG_SRC_F]->uiHeight,       IMG_SRCVSIZE_F);

		/* overlay */
		IpuRequest(ntsipu->img[IMG_DST]->uiX,              IMG_OFFSETX);
		IpuRequest(ntsipu->img[IMG_DST]->uiY,              IMG_OFFSETY);

		if(((ulImgMode & IMG_OP_MODE_BIT) == IMG_OP_MODE_12_BIT) || ((ulImgMode & IMG_OP_MODE_BIT) == IMG_OP_MODE_18_BIT)){
			/* maskcolor */
			IpuRequest(ntsipu->ntsc->from_fb.maskcolr, IMG_MASKCOLR);
			IpuRequest(ntsipu->ntsc->from_fb.alpha,    IMG_ALPHA);
		}
	}

	/*==============================
	    Start IMG
	================================*/
	IpuRequest(IMG_REQ_BIT,                                    IMG_REQ);

	/*==============================
	    end mark
	================================*/
	IpuRequest(IPU_REG_INFO_END,                               IPU_REG_INFO_END);

#if _DEBUG_NTSIPU
	printk_dbg((_DEBUG_NTSIPU & 0x02), "=== mp200_ipu_regist() =======================================\n");
	printk_dbg((_DEBUG_NTSIPU & 0x02), " IMG_MODE(0x%08lx)\n", ulImgMode);
	if(ulImgMode & IMG_OVERLAY_BIT){
	printk_dbg((_DEBUG_NTSIPU & 0x02), "-------+------------------------------------------------------\n");
	printk_dbg((_DEBUG_NTSIPU & 0x02), " SRC_F | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_SRC_F]->ulPhysAddrYRGB, ntsipu->img[IMG_SRC_F]->ulPhysAddrUV, ntsipu->img[IMG_SRC_F]->ulPhysAddrV);
	printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_SRC_F]->uiScreenWidth, ntsipu->img[IMG_SRC_F]->uiWidth, ntsipu->img[IMG_SRC_F]->uiHeight, ntsipu->img[IMG_SRC_F]->uiX, ntsipu->img[IMG_SRC_F]->uiY);
	}
	printk_dbg((_DEBUG_NTSIPU & 0x02), "-------+------------------------------------------------------\n");
	printk_dbg((_DEBUG_NTSIPU & 0x02), " SRC_B | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_SRC_B]->ulPhysAddrYRGB, ntsipu->img[IMG_SRC_B]->ulPhysAddrUV, ntsipu->img[IMG_SRC_B]->ulPhysAddrV);
	printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_SRC_B]->uiScreenWidth, ntsipu->img[IMG_SRC_B]->uiWidth, ntsipu->img[IMG_SRC_B]->uiHeight, ntsipu->img[IMG_SRC_B]->uiX, ntsipu->img[IMG_SRC_B]->uiY);
	printk_dbg((_DEBUG_NTSIPU & 0x02), "-------+------------------------------------------------------\n");
	printk_dbg((_DEBUG_NTSIPU & 0x02), " DST   | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->img[IMG_DST]->ulPhysAddrYRGB, ntsipu->img[IMG_DST]->ulPhysAddrUV, ntsipu->img[IMG_DST]->ulPhysAddrV);
	printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->img[IMG_DST]->uiScreenWidth, ntsipu->img[IMG_DST]->uiWidth, ntsipu->img[IMG_DST]->uiHeight, ntsipu->img[IMG_DST]->uiX, ntsipu->img[IMG_DST]->uiY);
	printk_dbg((_DEBUG_NTSIPU & 0x02), "==============================================================\n\n");
#endif
	return 0;
}


/* ------------------ private function < GDMA > --------------------------------- */
/******************************************************************************
* MODULE   : ntsipu_callback_ipugdma
* FUNCTION : changing frame and callback to fb and v4l2
* RETURN   : none
* NOTE     : none
* CREATE   : 2009.1.15 M.Haguro
* UPDATE   :
******************************************************************************/
static void ntsipu_callback_ipugdma(unsigned long sequence, int flag)
{
	unsigned long flags;
	spin_lock_irqsave(&ntsipu->ntsc->lock_callback, flags);
	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	/* wakeup mainhandle */
	ntsipu->gdma_complete = 1;
	wake_up_interruptible(&ntsipu->wait_gdma);

	spin_unlock_irqrestore(&ntsipu->ntsc->lock_callback, flags);
}


/*****************************************************************************
* MODULE   : ntsipu_chk_ipugdma
* FUNCTION : 
* RETURN   : -
* NOTE     : 
* CREATE   : 2009.1.13 M.Haguro
* UPDATE   : 
******************************************************************************/
static int ntsipu_chk_ipugdma(void)
{
	struct image_data *old = NULL;
	int index = 0, iRet = 0;

	printk_dbg((_DEBUG_NTSIPU & 0x01), "\n");

	if(ntsipu->ntsc->iFrameNoNow == NTS_DISP_FRAME_A){
		old = ntsipu->old[OLD_FRAME_B];
		if(old->uiWidth == 0){
			goto failed;
		}
	}
	else{ /* NTS_DISP_FRAME_B */
		old = ntsipu->old[OLD_FRAME_A];
		if(old->uiWidth == 0){
			goto failed;
		}
	}

	if(ntsipu->gdma[index]->uiX == old->uiX && ntsipu->gdma[index]->uiWidth  == old->uiWidth &&
	   ntsipu->gdma[index]->uiY == old->uiY && ntsipu->gdma[index]->uiHeight == old->uiHeight)
	{
		goto failed;
	}

	if(ntsipu->img[IMG_DST]->uiY > old->uiY){
		iRet = 1;	/* need fillrect */
	}
	if(ntsipu->img[IMG_DST]->uiX > old->uiX){
		iRet = 1;	/* need fillrect */
	}
	if(ntsipu->img[IMG_DST]->uiX + ntsipu->img[IMG_DST]->uiWidth  < old->uiX + old->uiWidth){
		iRet = 1;	/* need fillrect */
	}
	if(ntsipu->img[IMG_DST]->uiY + ntsipu->img[IMG_DST]->uiHeight < old->uiY + old->uiHeight){
		iRet = 1;	/* need fillrect */
	}

	if(iRet){
		printk_dbg((_DEBUG_NTSIPU & 0x02), "all fill\n");
		memcpy(ntsipu->gdma[index], ntsipu->img[IMG_DST], sizeof(struct image_data));

		ntsipu->gdma[index]->uiX      = old->uiX;
		ntsipu->gdma[index]->uiY      = old->uiY;
		ntsipu->gdma[index]->uiWidth  = old->uiWidth;
		ntsipu->gdma[index]->uiHeight = old->uiHeight;

		/* request IPU-GDMA */
		iRet = ntsipu_gdma_request(index);
		if(iRet){
			printk_dbg((_DEBUG_NTSIPU & 0x02), "gdma_request(%d)\n", iRet);
		}
	}

failed:
	memcpy(old, ntsipu->img[IMG_DST], sizeof(struct image_data));

	return iRet;
}


/*****************************************************************************
* MODULE   : ntsipu_gdma_request
* FUNCTION : 
* RETURN   : -
* NOTE     : 
* CREATE   : 2009.1.22 M.Haguro
* UPDATE   : 
******************************************************************************/
static int ntsipu_gdma_request(int index)
{
	int iRet;

	iRet = ntsipu_set_ipugdma(index);
	if(iRet){
		printk_dbg((_DEBUG_NTSIPU & 0x02), "(%d) error\n", __LINE__);
		goto failed;
	}
	ntsipu->ipu_request[index]->callback    = ntsipu_callback_ipugdma;
	ntsipu->ipu_request[index]->ipu_hw_func = IPU_HW_FUNC_GDMA;
	ntsipu->ipu_request[index]->data        = 0l;

	/* request IPU */
	ntsipu->gdma_complete = 0;
	iRet = mp200_ipu_regist((struct mp200_ipu_req *)(ntsipu->ipu_request[index]));
	if(iRet){
		printk_dbg((_DEBUG_NTSIPU & 0x02), "ipu_regist(%d)\n", iRet);
		goto failed;
	}
	wait_event_interruptible(ntsipu->wait_gdma, ntsipu->gdma_complete);

failed:
	memset(ntsipu->gdma[index],        0, sizeof(struct image_data));
	memset(ntsipu->ipu_request[index], 0, sizeof(MAX_SIZE_MP200_IPU_REQ));
	return iRet;
}


/*****************************************************************************
* MODULE   : ntsipu_set_ipugdma
* FUNCTION : 
* RETURN   : -
* NOTE     : 
* CREATE   : 2009.1.14 M.Haguro
* UPDATE   : 
******************************************************************************/
static int ntsipu_set_ipugdma(int index)
{
	unsigned long ulGdmaMode, ulFillData;
	unsigned int  uiScreenPixcelWidthY, uiScreenPixcelWidthUV, uiScreenPixcelWidthV;
	int bpp_y, bpp_uv, bpp_v; /* bit par pixel  */
	int ppl_y, ppl_uv, ppl_v; /* pixel par line */
	int index_req = 0;

	printk_dbg((_DEBUG_NTSIPU & 0x01), "index(%d)\n", index);

	/*******************************
	    set GDMA register
	********************************/
	ulGdmaMode = DMA_OP_FILL;
	ulFillData = 0x80008000; /* black */


	switch(ntsipu->gdma[index]->uiFormat){
	case DMA_FORMAT_YUV422_SEMI_PLANAR: /* YUV422 Semi-Planar */
		printk_dbg((_DEBUG_NTSIPU & 0x02), "src_b_format(YUV422Pl2)\n");
		bpp_y = 8;  bpp_uv = 8; bpp_v = 0; ppl_y = 1; ppl_uv = 1; ppl_v = 0;
		uiScreenPixcelWidthY  = ntsipu->gdma[index]->uiScreenWidth / bpp_y  * 8;
		uiScreenPixcelWidthUV = ntsipu->gdma[index]->uiScreenWidth / bpp_uv * 8;
		uiScreenPixcelWidthV  = 0;
		ntsipu->gdma[index]->ulPhysAddrYRGB = ntsipu->gdma[index]->ulPhysAddrYRGB + ntsipu->gdma[index]->uiY * uiScreenPixcelWidthY  / ppl_y  + ntsipu->gdma[index]->uiX * bpp_y  / 8;
		ntsipu->gdma[index]->ulPhysAddrUV   = ntsipu->gdma[index]->ulPhysAddrUV   + ntsipu->gdma[index]->uiY * uiScreenPixcelWidthUV / ppl_uv + ntsipu->gdma[index]->uiX * bpp_uv / 8;
		ntsipu->gdma[index]->ulPhysAddrYRGB &= ~0x03;
		ntsipu->gdma[index]->ulPhysAddrUV   &= ~0x03;
		ntsipu->gdma[index]->ulPhysAddrV    = 0;


		printk_dbg((_DEBUG_NTSIPU & 0x02), "=== mp200_ipu_regist() =======================================\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " DMA_MODE(0x%08lx)\n", ulGdmaMode);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "-------+------------------------------------------------------\n");
		printk_dbg((_DEBUG_NTSIPU & 0x02), " DST   | Y(%08lx)  U(%08lx)  V(%08lx)\n", ntsipu->gdma[index]->ulPhysAddrYRGB, ntsipu->gdma[index]->ulPhysAddrUV, ntsipu->gdma[index]->ulPhysAddrV);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "       | p(%3d)  w(%3d)  h(%3d)   x(%3d)  y(%3d)\n", ntsipu->gdma[index]->uiScreenWidth, ntsipu->gdma[index]->uiWidth, ntsipu->gdma[index]->uiHeight, ntsipu->gdma[index]->uiX, ntsipu->gdma[index]->uiY);
		printk_dbg((_DEBUG_NTSIPU & 0x02), "==============================================================\n\n");
		break;
	default:
		return -1;
	}


	/*==============================
	    GDMA_MODE
	================================*/
	IpuRequest(ulGdmaMode,                          DMA_MODE);
	IpuRequest(ulFillData,                          DMA_FILLDATA);

	/*==============================
	    destination image
	================================*/
	IpuRequest(ntsipu->gdma[index]->uiFormat,       DMA_FORMAT);
	IpuRequest(ntsipu->gdma[index]->uiWidth,        DMA_HSIZE);
#ifdef NTSC_ERRATA
	if(ntsipu->ntsc->iOutmodeFlg == NTS_OUTPUT_NTSC && errata->ntsc_upscale == NTS_UPSCALE_OFF){
		if(ntsipu->gdma[index]->uiY % 2){

			IpuRequest(ntsipu->gdma[index]->uiHeight +2, DMA_VSIZE);
			ntsipu->gdma[index]->ulPhysAddrYRGB -= ntsipu->gdma[index]->uiScreenWidth;
			ntsipu->gdma[index]->ulPhysAddrUV   -= ntsipu->gdma[index]->uiScreenWidth;
		}else{
			IpuRequest(ntsipu->gdma[index]->uiHeight,    DMA_VSIZE);
		}
	}
	else{
		IpuRequest(ntsipu->gdma[index]->uiHeight, DMA_VSIZE);
	}
#else
	IpuRequest(ntsipu->gdma[index]->uiHeight,       DMA_VSIZE);
#endif /* NTSC_ERRATA */
	IpuRequest(ntsipu->gdma[index]->uiScreenWidth,  DMA_DSTSIZE);
	IpuRequest(ntsipu->gdma[index]->ulPhysAddrYRGB, DMA_DSTYRGBADR);
	IpuRequest(ntsipu->gdma[index]->ulPhysAddrUV,   DMA_DSTUVADR);
	IpuRequest(ntsipu->gdma[index]->ulPhysAddrV,    DMA_DSTVADR);

	/*==============================
	    Start GDMA
	================================*/
	IpuRequest(DMA_REQ_BIT,                         DMA_REQ);

	/*==============================
	    end mark
	================================*/
	IpuRequest(IPU_REG_INFO_END,                    IPU_REG_INFO_END);

	return 0;
}


