/* 2009-08-24: File added by Sony Corporation */
/*
 *  LB035Q04 LCD Support for PGPF
 *
 *  Copyright 2009 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>

#include <asm/arch/hardware.h>
#include <asm/arch/smu.h>
#include <asm/arch/spi.h>

#include "lcdc.h"
#include "mp200_lcdhw.h"
#include <asm/arch/pgpf/lcdhw.h>

#if defined(CONFIG_PLL3_238MHZ)
#define LCD_LCLK_DIV 40
#define LCD_LB035Q04_DIVLCDLCLK (SMU_PLLSEL_PLL3 | SMU_DIV40)
#elif defined(CONFIG_PLL3_230MHZ)
#define LCD_LCLK_DIV 40
#define LCD_LB035Q04_DIVLCDLCLK (SMU_PLLSEL_PLL3 | SMU_DIV40)
#endif /* CONFIG_PLL3_* */
#define LCD_LB035Q04_PIXCLOCK (1.0 / (PGPF_PLL3_CLK / LCD_LCLK_DIV) * 1000000 * 1000000)

#define LCD_LB035Q04_CONTROL_BASE (LCD_CLKPOL_NEG | LCD_HPOL_POS | LCD_VPOL_POS | LCD_ENPOL_LOW)
#if defined(CONFIG_MP200_LCD_RGB666)
#define LCD_LB035Q04_CONTROL (LCD_OFORMAT_RGB666 | LCD_LB035Q04_CONTROL_BASE)
#else /* CONFIG_MP200_LCD_RGB666 */
#define LCD_LB035Q04_CONTROL (LCD_OFORMAT_RGB565 | LCD_LB035Q04_CONTROL_BASE)
#endif /* CONFIG_MP200_LCD_RGB666 */

#define LCD_LB035Q04_HTOTAL    (372)
#define LCD_LB035Q04_HAREA_ACT (320)
#define LCD_LB035Q04_HAREA     (LCD_LB035Q04_HAREA_ACT)
#define LCD_LB035Q04_HEDGE1    (0)
#define LCD_LB035Q04_HEDGE2    (LCD_LB035Q04_HTOTAL - LCD_LB035Q04_HAREA)

#define LCD_LB035Q04_VTOTAL    (248)
#define LCD_LB035Q04_VAREA_ACT (240)
#define LCD_LB035Q04_VAREA     (LCD_LB035Q04_VAREA_ACT)
#define LCD_LB035Q04_VEDGE1    (0)
#define LCD_LB035Q04_VEDGE2    (LCD_LB035Q04_VTOTAL - LCD_LB035Q04_VAREA)

#define _DEBUG_LCD  0x00

#define DEV_NAME "pgpf_lcd_lb035q04"

#define printk_err(fmt, arg...) \
	do{                     \
		printk(KERN_ERR DEV_NAME ": %s: " fmt, __func__, ## arg); \
	}while(0)

#define printk_wrn(fmt, arg...) \
	do{                     \
		printk(KERN_WARNING DEV_NAME ": %s: " fmt, __func__, ## arg); \
	}while(0)

#define printk_info(fmt, arg...) \
	do{                      \
		printk(KERN_INFO DEV_NAME ": " fmt, ## arg); \
	}while(0)

#if _DEBUG_LCD
#define printk_dbg(level, fmt, arg...) \
	do{                            \
		if(level > 0) printk(KERN_DEBUG DEV_NAME ": %s: " fmt, __func__, ## arg); \
	}while(0)
#else
#define printk_dbg(level, fmt, arg...) \
	;
#endif

#define G64_PINSEL_MASK_G71   (0x0000c000)
#define G71_PINSEL_LCD_ENABLE (0x00004000)
#define G71_PINSEL_GIO        (0x00000000)

#define REQUIRE_INITIAL_CODE

#ifdef REQUIRE_INITIAL_CODE
struct lb035q04_reg {
	uint8_t addr;
	uint16_t val;
};

#define LB035Q04_REG_TERMINATOR 0xFF

struct lb035q04_reg init_code[] = {
	{0x01, 0x6300}, /* Driver Output */
	{0x02, 0x0200}, /* LCD Driving Waveform Control */
	{0x03, 0x041C}, /* Power Control 1 */
	{0x04, 0x04C7}, /* Input Data & C/F Control */
	{0x05, 0xF7C0}, /* Function Control */
	{0x06, 0xE806}, /* - */
	{0x0A, 0x4008}, /* Contrast/Brightness Control */
	{0x0B, 0x0000}, /* Frame Cycle Control */
	{0x0D, 0x0024}, /* Power Control 2 */
	{0x0E, 0x29C0}, /* Power Control 3 */
	{0x0F, 0x0000}, /* Gate Scan Position */
	
	/* Porch */
	{0x16, 0x9F80}, /* Horizon Porch */
	{0x17, 0x0A0F}, /* Vertical Porch */
	{0x1E, 0x00DD}, /* Power Control 4 */

	/* Gamma Collection */
	{0x30, 0x0700},
	{0x31, 0x0207},
	{0x32, 0x0000},
	{0x33, 0x0105},
	{0x34, 0x0007},
	{0x35, 0x0101},
	{0x36, 0x0707},
	{0x37, 0x0504},
	{0x3A, 0x140F},
	{0x3B, 0x0509},

	/* End */
	{LB035Q04_REG_TERMINATOR, LB035Q04_REG_TERMINATOR},
};

static int pgpf_lcdhw_lb035q04_spiw(uint8_t address, uint16_t param)
{
	int ret = -1;
	SPI_CONFIG spi_conf;
	uint8_t tx_data[3];

	spi_conf.dev = SPI_DEV_SPI1;
	spi_conf.pol = (SPI_POL_CSW_4SCLK | SPI_POL_CK0_DLY_ON | SPI_POL_CK0_POL_POS | SPI_POL_CS0_POL_NEG);
	spi_conf.sclk = SPI_SCLK_6MHZ;
	spi_conf.mode = SPI_MODE_NB_24BIT | SPI_MODE_CS_SEL_CS0 | SPI_MODE_DMA_OFF | SPI_MODE_M_S_MASTER;
	spi_conf.size = 3;

	tx_data[0] = 0x70 | 0x00;
	tx_data[1] = (uint8_t)0x00;
	tx_data[2] = (uint8_t)address;
	ret = spi_write_without_intr(&spi_conf, (const char*)tx_data, SPI_RESET_OFF);

	if (ret) {
		printk_err("spi write error!\n");
		goto end;
	}

	tx_data[0] = 0x70 | 0x02;
	tx_data[1] = (uint8_t)((param & 0xff00) >> 8);
	tx_data[2] = (uint8_t)(param & 0x00ff);
	ret = spi_write_without_intr(&spi_conf, (const char*)tx_data, SPI_RESET_OFF);

	if (ret) {
		printk_err("spi write error!\n");
		goto end;
	}

  end:
	return ret;
}
#endif /* REQUIRE_INITIAL_CODE */

static int pgpf_lcdhw_lb035q04_resume(int restore, struct pgpf_bl_device* bl_dev)
{
	int i;

	/* set to pwm */
	if (bl_dev && bl_dev->prep) {
		bl_dev->prep();
	}

	/* LCD_PWRON on */
	pgpf_set_lcd_pwron(1);

	/* wait 10ms(min) */
	mdelay(20);

	if (restore) {
		lcd_hw_imc_hw_init();
		lcd_controller_hw_init();
		pgpf_lcdhw_controller_hw_init();
	}

	pgpf_lcdhw_change_gio_settings(1);

	/* change GIO_P71(LCD_ENABLE) to GIO Low */
	gpio_direction_output(GPIO_P71, 0);
	outl((inl(CHG_PINSEL_G64) & ~G64_PINSEL_MASK_G71) | G71_PINSEL_GIO, CHG_PINSEL_G64);

	/* LCD start (PXCLK, DATA, ...) */
	lcd_hw_start();

	/* wait 10us(min) */
	udelay(10);

	/* LCD_PON on */
	pgpf_set_lcd_pon(1);

#ifdef REQUIRE_INITIAL_CODE
	/* wait 10us(min) */
	udelay(10);
	
	/* initial code */
	for (i = 0; init_code[i].addr != LB035Q04_REG_TERMINATOR; i++) {
		pgpf_lcdhw_lb035q04_spiw(init_code[i].addr, init_code[i].val);
	}
#endif REQUIRE_INITIAL_CODE

	/* wait 11 frame(min) */
	mdelay(250);

	/* change GIO_P71(LCD_ENABLE) to LCD_ENABLE */
	outl((inl(CHG_PINSEL_G64) & ~G64_PINSEL_MASK_G71) | G71_PINSEL_LCD_ENABLE, CHG_PINSEL_G64);

	/* wait 160ms(min) */
	mdelay(160);

	/* Backlight on */
	if (bl_dev && bl_dev->resume) {
		bl_dev->resume();
	}

	return 0;
}

static int pgpf_lcdhw_lb035q04_suspend(int save, struct pgpf_bl_device* bl_dev)
{
	/* wait 1 frame (max) */

	/* Backlight off */
	if (bl_dev && bl_dev->suspend) {
		bl_dev->suspend();
	}

	/* wait 160ms(min) */
	mdelay(160);

	/* LCD stop (PXCLK, DATA, ...) */
	lcd_hw_stop();

	pgpf_lcdhw_change_gio_settings(0);

	/* wait 10ms(min) */
	mdelay(10);

	/* LCD_PON off */
	pgpf_set_lcd_pon(0);

	/* LCD_PWRON off */
	pgpf_set_lcd_pwron(0);

	/* set to gpio */
	if (bl_dev && bl_dev->post) {
		bl_dev->post();
	}

	return 0;
}


static int pgpf_lcdhw_lb035q04_init(struct pgpf_bl_device* bl_dev)
{
	if (!(lcd_hw_chk_status() & LCD_LCDOUT_ENABLE)) {
		/* LCD Module initialize */
		pgpf_lcdhw_change_gio_settings(0);
		pgpf_lcdhw_setup_clock();
		pgpf_lcdhw_lb035q04_resume(1, bl_dev);
	} else {
		lcd_hw_imc_hw_init();
		lcd_controller_hw_init();
		pgpf_lcdhw_controller_hw_init();
	}

	return 0;
}

static void pgpf_lcdhw_lb035q04_exit(struct pgpf_bl_device* bl_dev)
{
	pgpf_lcdhw_lb035q04_suspend(0, bl_dev);
}

static struct pgpf_lcdhw_device pgpf_lcdhw_lb035q04_dev = {
	.init = pgpf_lcdhw_lb035q04_init,
	.exit = pgpf_lcdhw_lb035q04_exit,
	.suspend = pgpf_lcdhw_lb035q04_suspend,
	.resume = pgpf_lcdhw_lb035q04_resume,

	.config.xres = LCD_LB035Q04_HAREA_ACT,
	.config.yres = LCD_LB035Q04_VAREA_ACT,

	.config.pixclock = LCD_LB035Q04_PIXCLOCK,
	.config.divlcdlclk = LCD_LB035Q04_DIVLCDLCLK,

	.config.control = LCD_LB035Q04_CONTROL,

	.config.htotal = LCD_LB035Q04_HTOTAL,
	.config.harea  = LCD_LB035Q04_HAREA,
	.config.hedge1 = LCD_LB035Q04_HEDGE1,
	.config.hedge2 = LCD_LB035Q04_HEDGE2,
	.config.vtotal = LCD_LB035Q04_VTOTAL,
	.config.varea  = LCD_LB035Q04_VAREA,
	.config.vedge1 = LCD_LB035Q04_VEDGE1,
	.config.vedge2 = LCD_LB035Q04_VEDGE2,
};

int pgpf_lcdhw_register(struct pgpf_lcdhw_device** lcd_dev, struct platform_device* pf_dev)
{
	printk_dbg(_DEBUG_LCD, "pgpf_lcdhw_register() called\n");
	*lcd_dev = &pgpf_lcdhw_lb035q04_dev;
	return 0;
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("LB035Q04 LCD Support for PGPF");
