/* 2009-09-03: File added and changed by Sony Corporation */
/*
 *  fs/fat/snsc_fileio_trace.c
 *
 *  Copyright 2009 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/major.h>

#define FATTRACE_PFX "FAT_TRACE: "

#ifdef CONFIG_SNSC_FS_FAT_FILEIO_TRACE_SUPPRESS_LOG_IF_ON_LOOPBACK
static unsigned int
get_target_dev_major(struct dentry *dentry)
{

	if (dentry->d_sb != NULL)
		return MAJOR(dentry->d_sb->s_dev);
	else
		return 0;
}
#endif

static void
print_filename_size(char *operation, struct dentry *dentry, ssize_t size)
{
	struct dentry *dent;

	dent = dget(dentry);

	if (dent != NULL) {
#ifdef CONFIG_SNSC_FS_FAT_FILEIO_TRACE_SUPPRESS_LOG_IF_ON_LOOPBACK
		if (get_target_dev_major(dent) == LOOP_MAJOR)
			goto out;
#endif

		if (size >= 0) {
			pr_info(FATTRACE_PFX "%s %s (%zdbytes)\n",
				dent->d_name.name,
				operation,
				size);

		} else {
			pr_info(FATTRACE_PFX "%s %s error (errno = %d)\n",
				dent->d_name.name,
				operation,
				(int)size);
		}

	} else {
		pr_err(FATTRACE_PFX "dentry is null\n");
	}

#ifdef CONFIG_SNSC_FS_FAT_FILEIO_TRACE_SUPPRESS_LOG_IF_ON_LOOPBACK
 out:
#endif

	dput(dent);
}

static void
write_probe(const struct marker *mdata, void *private, const char *fmt, ...)
{
	va_list args;
	struct dentry *dentry;
	ssize_t ret;

	va_start(args, fmt);
	dentry = va_arg(args, typeof(dentry));
	ret    = va_arg(args, typeof(ret));
	va_end(args);

	print_filename_size("write", dentry, ret);
}

static void
unlink_probe(const struct marker *mdata, void *private, const char *fmt, ...)
{
	va_list args;
	struct dentry *dentry;

	va_start(args, fmt);
	dentry = va_arg(args, typeof(dentry));
	va_end(args);

	print_filename_size("unlink", dentry, 0);
}

struct probe_data {
	const char *name;
	const char *format;
	marker_probe_func *probe_func;
};

static struct probe_data probe_array[] =
{
	{
		.name       = "fat_write",
		.format     = "%p %zd",
		.probe_func = write_probe,
	},
	{
		.name       = "fat_unlink",
		.format     = "%p",
		.probe_func = unlink_probe,
	},
};

static int __init
init_fat_fileio_trace(void)
{
	int i, result, ret;

	ret = 0;

	for (i = 0; i < ARRAY_SIZE(probe_array); i++) {

		result = marker_probe_register(
				probe_array[i].name,
				probe_array[i].format,
				probe_array[i].probe_func,
				NULL);
		if (result) {
			pr_err(FATTRACE_PFX "unable to register probe %s\n",
			       probe_array[i].name);
			ret = result;
			continue;
		}

		result = marker_arm(probe_array[i].name);
		if (result) {
			pr_err(FATTRACE_PFX "unable to arm probe %s\n",
			       probe_array[i].name);
			ret = result;
		}
	}

	if (!ret)
		pr_info(FATTRACE_PFX "installed\n");
	else
		pr_err(FATTRACE_PFX "unable to install\n");

	return ret;
}

static void __exit
exit_fat_fileio_trace(void)
{
	int i, result;

	for (i = ARRAY_SIZE(probe_array) - 1; i >= 0; i--) {

		result = marker_disarm(probe_array[i].name);

		if (result) {
			pr_warning(FATTRACE_PFX "unable to unarm probe %s\n",
				   probe_array[i].name);
			continue;
		}

		marker_probe_unregister(probe_array[i].name);
	}
}

module_init(init_fat_fileio_trace);
module_exit(exit_fat_fileio_trace);

MODULE_LICENSE("GPL");
