/* 2009-02-24: File added and changed by Sony Corporation */
/*
 *  common backtrace interface
 *
 *  Copyright 2009 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/ctype.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/uaccess.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/snsc_backtrace.h>
#include <asm/mmu_context.h>

#define BT_PARAM_LEN 128
char bt_param[BT_PARAM_LEN] = "kstack";

#define PROC_PARAM_NAME            "backtrace"
#define PROC_UIF_NAME              "bt"

static int __init bt_setup(char *str)
{
	/* Rely on implicit final '\0' for static data */
	strncpy(bt_param, str, BT_PARAM_LEN - 1);
	return 0;
}
__setup("backtrace=", bt_setup);

static struct proc_dir_entry *proc_bt = NULL;

static int bt_proc_read(char *buf, char **start, off_t off,
			int count, int *eof, void *data)
{
	int len;

	/* Rely on implicit final '\0' for static data */
	strncpy(buf, bt_param, BT_PARAM_LEN - 1);

	len = strlen(buf);
	if (len <= off + count)
		*eof = 1;
	*start = buf + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

static int bt_proc_write(struct file *file, const char __user *buffer,
			 unsigned long count, void *data)
{
	if (count > BT_PARAM_LEN - 1)
		return -EINVAL;
	if (copy_from_user(bt_param, buffer, count))
		return -EFAULT;
	bt_param[count] = '\0';
	return count;
}

static void *btuif_seq_start(struct seq_file *m, loff_t *pos)
{
	return 0;
}

static void btuif_seq_stop(struct seq_file *m, void *p)
{
}

static void *btuif_seq_next(struct seq_file *m, void *p, loff_t *pos)
{
	return 0;
}

static int btuif_seq_show(struct seq_file *m, void *v)
{
	return 0;
}

static struct seq_operations btuif_seq_op = {
	.start	= btuif_seq_start,
	.next	= btuif_seq_next,
	.stop	= btuif_seq_stop,
	.show	= btuif_seq_show
};

static int btuif_file_open(struct inode *inode, struct file *file)
{
	struct bt_session *priv;
	struct seq_file *sfile;
	int ret;
	priv = kmalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;
	ret = seq_open(file, &btuif_seq_op);
	if (ret) {
		kfree(priv);
		return ret;
	}
	bt_init_session(priv, bt_param, GFP_KERNEL);
	sfile = file->private_data;
	sfile->private = priv;
	return ret;
}

static int btuif_file_release(struct inode *inode, struct file *file)
{
	struct seq_file *seq = file->private_data;
	struct bt_session *priv = seq->private;
	bt_release_session(priv);
	return seq_release_private(inode, file);
}

static int btuif_ioctl_addrs(struct bt_session *session,
			     struct bt_ioctl_addrs __user *u_arg)
{
	struct bt_ioctl_addrs arg;

	if (copy_from_user(&arg, u_arg, sizeof(arg))) {
		return -EFAULT;
	}
	return bt_ustack_user(session, &arg);
}

/* return 0 for success, 1 for failure. */
/* number pointed to by cur is incremented by the length written, */
/* excluding terminating '\0', dest always null terminated. */
static int bt_copy_str_to_user(char __user *u_dest, const char *src,
			       size_t max, int *cur)
{
	size_t len = strnlen(src, max - 1);
	if (copy_to_user(u_dest, src, len))
		return 1;
	if (put_user(0, u_dest + len))
		return 1;
	*cur += len;
	return 0;
}

static int btuif_ioctl_symbol(struct bt_session *session,
			      struct bt_ioctl_symbol __user *u_arg)
{
	struct bt_ioctl_symbol arg;
	int ret, cur;
	char buf[256];
	int buflen = sizeof(buf);
	struct bt_arch_callback_arg cbarg;
	const char *libname;

	if (copy_from_user(&arg, u_arg, sizeof(arg))) {
		return -EFAULT;
	}
	if (arg.bs_size == 0)
		return 0;
	ret = bt_find_symbol((unsigned long)arg.bs_addr, session,
			     buf, buflen, &cbarg);
	if (ret < 0) {
		/* no symbol found, return "" */
		put_user(0, arg.bs_buf);
		return 1;
	}
	cur = 0;
	if (cbarg.ba_file) {
		libname = bt_file_name(cbarg.ba_file);
		if (bt_copy_str_to_user(arg.bs_buf, libname, arg.bs_size, &cur))
			return -EFAULT;
	}
	if (bt_copy_str_to_user(arg.bs_buf + cur, "(", arg.bs_size - cur, &cur))
		return -EFAULT;
	if (bt_copy_str_to_user(arg.bs_buf + cur, buf, arg.bs_size - cur, &cur))
		return -EFAULT;
	snprintf(buf, buflen, "+0x%lx", (unsigned long)arg.bs_addr
		                      - cbarg.ba_sym_start);
	if (bt_copy_str_to_user(arg.bs_buf + cur, buf, arg.bs_size - cur, &cur))
		return -EFAULT;
	if (bt_copy_str_to_user(arg.bs_buf + cur, ")", arg.bs_size - cur, &cur))
		return -EFAULT;
	return cur + 1;
}

static int btuif_ioctl(struct inode *inode, struct file *filp,
			unsigned int num, unsigned long arg)
{
	struct seq_file *seq = filp->private_data;
	struct bt_session *priv = seq->private;

	switch (num) {
	case BT_IOCTL_ADDRS:
		return btuif_ioctl_addrs(priv, (struct bt_ioctl_addrs
					       __user *)arg);
	case BT_IOCTL_SYMBOL:
		return btuif_ioctl_symbol(priv, (struct bt_ioctl_symbol
						 __user *)arg);
	default:
		return -EINVAL;
	}
}

static struct file_operations btuif_file_operations = {
	.open           = btuif_file_open,
	.read           = seq_read,
	.ioctl          = btuif_ioctl,
	.release        = btuif_file_release,
};

int __init bt_proc_init(void)
{
	struct proc_dir_entry *btuif_entry;

	if ((proc_bt = create_proc_entry(PROC_PARAM_NAME, 0600, &proc_root)) == NULL) {
		printk("cannot create proc entry for %s\n", PROC_PARAM_NAME);
		return 1;
	}
	proc_bt->read_proc  = &bt_proc_read;
	proc_bt->write_proc  = &bt_proc_write;

	btuif_entry = create_proc_entry(PROC_UIF_NAME, S_IWUSR|S_IRUSR, NULL);
	if (btuif_entry)
		btuif_entry->proc_fops = &btuif_file_operations;
	return 0;
}

const char *bt_param_match(const char *param, const char *name)
{
	const char *from = strstr(param, name);
	const char *to;

	if (from == NULL)
		return NULL;
	if (from > param && isalnum(from[-1])) /* suffix match */
		return NULL;
	to = from + strlen(name);
	if (isalnum(*to))
		return NULL; /* prefix match */
	return to;
}

int bt_ustack_enabled(void)
{
	return (int)bt_param_match(bt_param, "ustack");
}

/* call bt_ustack with regs of specified task */
void bt_ustack_task(struct task_struct *task,
		    const char *mode, int is_atomic,
		    bt_callback_t *cb, void *uarg)
{
	struct pt_regs *regs = task_pt_regs(task);
	struct task_struct *cur = current;
	struct mm_struct *saved_mm, *saved_active_mm, *new_mm;

	if (cur == task) {
		bt_ustack(mode, is_atomic, regs, cb, uarg);
		return;
	}

	new_mm = get_task_mm(task); /* need mmput() below */
	task_lock(cur);
	saved_mm = cur->mm;
	saved_active_mm = cur->active_mm;
	cur->mm = new_mm;
	cur->active_mm = new_mm;
	switch_mm(saved_mm, new_mm, cur);
	task_unlock(cur);

	bt_ustack(mode, is_atomic, regs, cb, uarg);

	task_lock(cur);
	cur->mm = saved_mm;
	cur->active_mm = saved_active_mm;
	switch_mm(new_mm, saved_mm, cur);
	task_unlock(cur);
	mmput(new_mm);
}
EXPORT_SYMBOL(bt_ustack_task);

struct bt_bufpos {
	char *buf;
	size_t size;
	int ofs;
};

static int bt_ustack_buf_cb(struct bt_arch_callback_arg *cbarg, void *user)
{
	struct bt_bufpos *p = user;
	p->ofs += scnprintf(p->buf + p->ofs, p->size - p->ofs,
			    "[0x%08lx] ", cbarg->ba_addr);
	if (bt_status_is_error(cbarg->ba_status)) {
		p->ofs += scnprintf(p->buf + p->ofs, p->size - p->ofs,
				    "stop backtracing: %s\n", cbarg->ba_str);
		return 0;
	}
	if (cbarg->ba_str[0]) {
		p->ofs += scnprintf(p->buf + p->ofs, p->size - p->ofs,
				    "%s", cbarg->ba_str);
	}
	else {
		p->ofs += scnprintf(p->buf + p->ofs, p->size - p->ofs,
				    "0x%08lx", cbarg->ba_sym_start);
	}
	if (bt_hash_valid(cbarg)) {
		/* by symlist section */
		const unsigned char *hash = cbarg->ba_hash;
		p->ofs += scnprintf(p->buf + p->ofs, p->size - p->ofs,
				    "+%#lx (%s hash:%02x%02x%02x%02x adj:%ld)\n",
				    cbarg->ba_addr - cbarg->ba_sym_start,
				    bt_file_name(cbarg->ba_file),
				    hash[0], hash[1], hash[2], hash[3],
				    cbarg->ba_adjust);
	}
	else {
		/* by symtab section */
		p->ofs += scnprintf(p->buf + p->ofs, p->size - p->ofs,
				    "+%#lx/%#lx (%s)\n",
				    cbarg->ba_addr - cbarg->ba_sym_start,
				    cbarg->ba_sym_size,
				    bt_file_name(cbarg->ba_file));
	}
	return 0;
}

/*
 * bt_ustack_task_buf - Format and write result of bt_ustack_task into a buffer
 *
 *  size: The size of the buffer, including the trailing null space
 *
 * The return value is the number of characters written into @buf not including
 * the trailing '\0'.
 */
int bt_ustack_task_buf(char *buf, size_t size,
		       struct task_struct *task,
		       const char *mode, int is_atomic)
{
	struct bt_bufpos p = {
		.buf = buf,
		.size = size,
		.ofs = 0,
	};
	bt_ustack_task(task, mode, is_atomic, bt_ustack_buf_cb, &p);
	return p.ofs;
}
