/* 2010-08-25: File added and changed by Sony Corporation */
/*
 * linux/sound/arm/mp200-pcm.c -- ALSA PCM interface for mp200 chip
 *
 * Created:	Oct 09, 2006
 * update :     Sep 17, 2009
 *
 */

#include <linux/errno.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/arch/pwc.h>
#include <asm/arch/pmu.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>
#include <sound/hwdep.h>

#ifdef CONFIG_PM
#include <asm/arch/pm.h>
#endif

#include <linux/irq.h>

#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#include <asm/arch/gpio.h>

#include "mp200-pcm.h"
#define MP200_PCM_PACK_PELIOD

#include "mp200-mixer.h"

#ifdef MP200_PCM_MMAP_ENABLE
#include <linux/dma-mapping.h>
#endif

#define PCM0_TX(s)	(s->dma_ch == MP200_DMAC_M2P_LCH9)

#ifdef MP200_PCM_USE_PDMA

#include <asm/arch/pcm_irq.h>
#include <asm/arch/timer.h>

#define IS_PCM0(p)	(p.irq	   == INT_PCM0)

#define SRAM_BASE		MP200_SSRAM_BASE
#define SRAM_SIZE		MP200_SSRAM_SIZE /* 128k */

/****  Prototype definition  ****/
#define PDMA_BASE	IO_ADDRESS(MP200_PDMA_BASE)

#define PDMA_DMA_SEL		(PDMA_BASE + 0x00)
#define PDMA_CONT			(PDMA_BASE + 0x04)
#define PDMA_STATUS			(PDMA_BASE + 0x08)
#define PDMA_RCV_CANCEL		(PDMA_BASE + 0x0C)
#define PDMA_END			(PDMA_BASE + 0x10)
#define PDMA_RSV_ADD		(PDMA_BASE + 0x20)
#define PDMA_RSV_LENG		(PDMA_BASE + 0x24)
#define PDMA_RUN_ADD		(PDMA_BASE + 0x28)
#define PDMA_RUN_LENG		(PDMA_BASE + 0x2C)
#define PDMA_INT_STATUS		(PDMA_BASE + 0x30)
#define PDMA_INT_RAW_STATUS	(PDMA_BASE + 0x34)
#define PDMA_INT_ENABLE		(PDMA_BASE + 0x38)
#define PDMA_INT_ENABLE_CL	(PDMA_BASE + 0x3C)
#define PDMA_INT_REQ_CL		(PDMA_BASE + 0x40)
#define PDMA_RUN_ADP		(PDMA_BASE + 0x50)
#define PDMA_ERR			(PDMA_BASE + 0x54)
#define PDMA_TMP			(PDMA_BASE + 0x58)

#endif

/*** debug code by the making ->*/
#ifdef AUDIO_MAKING_DEBUG
#include <linux/moduleparam.h>
#define FNC_ENTRY	\
	if (debug == 1 || debug >= 9) {	\
		printk(KERN_DEBUG "entry:%s\n", __FUNCTION__); \
	}
#define FNC_EXIT	\
	if (debug == 1 || debug >= 9) {	\
		printk(KERN_DEBUG "exit:%s:%d\n", __FUNCTION__ , __LINE__); \
	}
#define d0b(fmt, args...)	\
		printk(KERN_DEBUG "%s:%d: " fmt , __FUNCTION__ , __LINE__ , ## args);
#define d1b(fmt, args...)	\
	if (debug == 1 || debug >= 9) {	\
		printk(KERN_DEBUG "%s:%d: " fmt , __FUNCTION__ , __LINE__ , ## args); \
	}
#define d2b(fmt, args...)	\
	if (debug == 2 || debug >= 9) {	\
		printk(KERN_DEBUG "%s:%d: " fmt , __FUNCTION__ , __LINE__ , ## args); \
	}
#define d3b(fmt, args...)	\
	if (debug == 3 || debug >= 9) {	\
		printk(KERN_DEBUG " --M%d-- "fmt , debug , ## args); \
	}
#define d4b(fmt, args...)	\
	if (debug == 4 || debug >= 9) {	\
		printk(KERN_DEBUG " --M%d-- "fmt , debug , ## args); \
	}
#define d5b(fmt, args...)	\
	if (debug == 5 || debug >= 9) {	\
		printk(KERN_DEBUG " --M%d-- "fmt , debug , ## args); \
	}
#define d6b(fmt, args...)	\
	if (debug == 6 || debug >= 9) {	\
		printk(KERN_DEBUG " --M%d-- "fmt , debug , ## args); \
	}
static unsigned int ddata;
int debug = 2;
EXPORT_SYMBOL(debug);
module_param(debug,  int, 0644);
#else
#define FNC_ENTRY
#define FNC_EXIT
#define d0b(fmt, args...)
#define d1b(fmt, args...)
#define d2b(fmt, args...)
#define d3b(fmt, args...)
#define d4b(fmt, args...)
#define d5b(fmt, args...)
#define d6b(fmt, args...)
#endif
/*<- debug code by the making ***/

/* Type definitions */

/* PMx_TXRX_EN Masks */
#define PCM_TX_EN		0x01
#define PCM_RX_EN		0x02

static int power_flag = 0;

/*
 * Data for managing PCM channels
 */


struct mp200_pcm {
	char *name;		/* irq name. */
	u_int base;		/* base address */
	u_int clock_s;		/* SCLK */
	u_int clock_p;		/* PCLK */
	u_int clkctl;		/* CLKCTRL */
	u_int reset;		/* reset */
#ifdef MP200_PCM_INTERRUPT_ENABLE
	u_int irq;		/* irq */
	struct snd_pcm *pcm;		/* arg */
#endif
	u_int8_t in_use;	/* PCM in use flag. */
};

static struct mp200_pcm pcm_ch[MP200_PCM_MAX_CHANNELS] = {
	{
		.name = "mp200-pcm0",
		.base = MP200_PCM0_BASE,
		.clock_s = MP200_CLOCKGATE_PM0_SCLK,
		.clock_p = MP200_CLOCKGATE_PM0_CLK,
		.clkctl = MP200_CLKCTRL_PM0PCLK,
		.reset = MP200_RESETDEVICE_PM0,
#ifdef MP200_PCM_INTERRUPT_ENABLE
		.irq = INT_PCM0,
#endif
		.in_use = 0,
	},
	{
		.name = "mp200-pcm1",
		.base = MP200_PCM1_BASE,
		.clock_s = MP200_CLOCKGATE_PM1_SCLK,
		.clock_p = MP200_CLOCKGATE_PM1_CLK,
		.clkctl = MP200_CLKCTRL_PM1PCLK,
		.reset = MP200_RESETDEVICE_PM1,
#ifdef MP200_PCM_INTERRUPT_ENABLE
		.irq = INT_PCM1,
#endif
		.in_use = 0,
	},
};

/*
 * prototypes
 */
static int mp200_pcm_hwdep_open(struct snd_hwdep * hw, struct file *file);
static int mp200_pcm_hwdep_ioctl(struct snd_hwdep * hw, struct file *file,
				 unsigned int cmd, unsigned long arg);
static int mp200_pcm_hwdep_release(struct snd_hwdep * hw, struct file *file);

static snd_pcm_uframes_t output_ptr = 0;

/* PDMA Source */
#ifdef MP200_PCM_USE_PDMA

static pcm_callback_t alsa_pdma_callback = NULL;
static void *alsa_pdma_data = NULL;
static int pdma_init = 0;
static int pdma_in_use = 0;

#ifdef MP200_DAC_HANGUP_PDMA
static char *tg_name = "tg5_pdma";
static int timer_usec = 1000000; /* 1s */

/* work queue */
static struct workqueue_struct *mp200_pdma_workqueue;
static struct work_struct mp200_pdma_dac_work;
static void mp200_dac_work_do(struct work_struct *work);
#endif

static void mp200_stop_pdma(void)
{

#ifdef MP200_DAC_HANGUP_PDMA
	/* for DAC hang up */
	mp200_timer_stop(DAC_HANG_TG);
#endif

	writel(1, PDMA_END);
	mp200_pm_pdma_suspend_disable();
	pdma_init = 0;
}

static irqreturn_t pdma_callback(int irq, void *data)
{

#ifdef MP200_DAC_HANGUP_PDMA
	/* for DAC hang up */
	mp200_timer_stop(DAC_HANG_TG);
#endif

	writel(1, PDMA_INT_REQ_CL);

	return alsa_pdma_callback(irq, alsa_pdma_data);
}

static void mp200_pdma_deinit(void)
{
	mp200_pmu_reset_device(MP200_RESETDEVICE_PDMA);

	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PDMA_PCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PDMA_HCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PDMA_ACLK);
}

static void mp200_pdma_init(void)
{
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PDMA_PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PDMA_HCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PDMA_ACLK);

	mp200_pmu_unreset_device(MP200_RESETDEVICE_PDMA);

	mp200_pmu_clkctrl_on(MP200_CLKCTRL_PDMAPCLK);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_PDMAHCK);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_PDMAACK);
}

static int mp200_request_pdma(const char *device_id, pcm_callback_t callback, void *data)
{
	int err;

	if (pdma_in_use == 1) {
		return -EBUSY;
	}

	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SRC);
	err = mp200_pdma_request_irq(pdma_callback, IRQF_DISABLED, device_id, data, MP200_PCMMODE_ALSA);
	if (err != 0) {
		printk(KERN_INFO "%s(): unable to request IRQ %d for DMA channel (%s)\n",
			  __FUNCTION__, INT_PDMA, device_id);
		return err;
	}

	alsa_pdma_callback = callback;
	alsa_pdma_data = data;

	pdma_in_use = 1;

	writel(1, PDMA_DMA_SEL);
	writel(1, PDMA_INT_ENABLE);
	pdma_init = 1;

	return 0;
}

/*
 * Release DMA channel
 */
static void mp200_free_pdma(void)
{
	mp200_stop_pdma();
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SRC);

	mp200_pdma_free_irq(alsa_pdma_data, MP200_PCMMODE_ALSA);

	pdma_in_use = 0;
	alsa_pdma_callback = NULL;
	alsa_pdma_data = NULL;
}


/*
 * Setup and enable DMA.
 */
static int mp200_start_pdma(dma_addr_t src_ptr, u_int size)
{
	if (pdma_init == 0) {
		writel(1, PDMA_DMA_SEL);
		writel(1, PDMA_INT_ENABLE);
	}

	mp200_pm_pdma_suspend_enable();

	writel(src_ptr, PDMA_RSV_ADD);
	writel(size / 4, PDMA_RSV_LENG);

#ifdef MP200_DAC_HANGUP_PDMA
	/* for DAC hang up */
	mp200_timer_start(DAC_HANG_TG);
#endif

	writel(1, PDMA_CONT);

	return 0;
}

/*
 * Acquire DMA channel status.
 */
static int mp200_pdma_status(void)
{
	unsigned int status;

	status = readl(PDMA_STATUS);

	switch (status) {
	case 1:
	case 2:
		return 1;
	default:
		return 0;
	}
}

/*
 * Acquire DMA channel status (register writable flag).
 */
static int mp200_pdma_writable_flag(void)
{
	unsigned int status;

	status = readl(PDMA_STATUS);

	switch (status) {
	case 2:
		return 1;
	default:
		return 0;
	}
}

static void mp200_clear_pdma(void)
{
	mp200_stop_pdma();
}

dma_addr_t mp200_get_pdma_pos(void)
{
#if 0
	return readl(PDMA_RUN_ADP);
#else
	unsigned int pos;
	unsigned int addr;
	unsigned int leng;
	unsigned int state;

	state = readl(PDMA_STATUS);
	pos = readl(PDMA_RUN_ADP);
	addr = readl(PDMA_RUN_ADD);
	leng = readl(PDMA_RUN_LENG) * 4;

	if( (addr <= pos) && (pos <(addr+leng)) ) {
		;
	}else{
		pos = addr;
	}

	return( pos );
#endif
}

#ifdef MP200_DAC_HANGUP_PDMA
static irqreturn_t audio_pdma_callback(int irq, void *data);


/* for DAC hang up */
static void mp200_dac_work_do(struct work_struct *work)
{

	mp200_codec_reset();

	return;
}

static irqreturn_t timer_callback(int irq, void *dev_id)
{
	mp200_timer_stop(DAC_HANG_TG);

	queue_work(mp200_pdma_workqueue, &mp200_pdma_dac_work);

	return IRQ_HANDLED;
}
#endif /* MP200_DAC_HANGUP_PDMA */

#endif

/* DMA staff */

static int audio_dma_request(struct audio_stream *s, void (*callback)(void*, int, int))
{
	int ret;
	FNC_ENTRY

	output_ptr = 0;

#ifdef MP200_PCM_USE_PDMA
	if (PCM0_TX(s)) {
		ret = mp200_request_pdma(s->id, audio_pdma_callback, s);
	} else {
		ret = mp200_request_dma(s->dma_ch, s->id, (dma_callback_t)callback, s, &s->dma_regs);
	}
#else
	ret = mp200_request_dma(s->dma_ch, s->id, (dma_callback_t)callback, s, &s->dma_regs);
#endif
	if (ret < 0)
		printk(KERN_ERR "unable to grab audio dma 0x%x\n", s->dma_ch);

	FNC_EXIT return ret;
}

static void audio_dma_free(struct audio_stream *s)
{
	FNC_ENTRY
#ifdef MP200_PCM_USE_PDMA
	if (PCM0_TX(s)) {
		mp200_free_pdma();
	} else {
		mp200_free_dma(s->dma_ch);
	}
#else
	mp200_free_dma(s->dma_ch);
#endif
	s->dma_regs = 0;
	FNC_EXIT return;
}

static u_int audio_get_dma_pos(struct audio_stream *s)
{
	struct snd_pcm_substream * substream = s->stream;
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned int offset;
	unsigned int pos;
	unsigned long flags;
	dma_addr_t addr;
	FNC_ENTRY

	spin_lock_irqsave(&s->dma_lock, flags);
#ifdef MP200_PCM_USE_PDMA
	if (PCM0_TX(s)) {
		addr = mp200_get_pdma_pos();
		if( addr == 0 ) {
			/* already stopped. */
			addr = runtime->dma_addr;
		}
	} else {
		addr = mp200_get_dma_pos(s->dma_ch);
	}
#else
	addr = mp200_get_dma_pos(s->dma_ch);
#endif
	offset = addr - runtime->dma_addr;
	spin_unlock_irqrestore(&s->dma_lock, flags);

	/* d6b("offset = %d : bit = %d\n", offset, runtime->frame_bits); */
	pos = bytes_to_frames(runtime,offset);
	if (pos >= runtime->buffer_size)
		pos = 0;

#ifdef MP200_PCM_USE_PDMA
	if( runtime->status->state == SNDRV_PCM_STATE_DRAINING ) {
		/* special fake code, for stop. */
		/* warning : if returned pos that it is buffer_size, not stopped. */
		pos = frames_to_bytes(runtime, runtime->control->appl_ptr - runtime->hw_ptr_base);
	}
#endif

	/* d3b("offset = 0x%08x - 0x%08x\n", addr, runtime->dma_addr); */

	FNC_EXIT return pos;
}

/*
 * this stops the dma and clears the dma ptrs
 */
static void audio_stop_dma(struct audio_stream *s)
{
	unsigned long flags;
	FNC_ENTRY

	*s->txrx_dis = s->txrx_id;
	d2b("PCM TXRX_DIS(0x%08x) : 0x%08x\n",
		(u_int)s->txrx_dis, s->txrx_id);

	d3b("%s stop dma ch : %d, pos : %d\n",((PCM_TX_EN == s->txrx_id) ? "TX" : "RX"), s->dma_ch, audio_get_dma_pos(s));
	spin_lock_irqsave(&s->dma_lock, flags);
	s->active = 0;
	s->period = 0;
	output_ptr = 0;
	/* this stops the dma channel and clears the buffer ptrs */
#ifdef MP200_PCM_USE_PDMA
	if (PCM0_TX(s)) {
		mp200_clear_pdma();
	} else {
		mp200_clear_dma(s->dma_ch);
	}
#else
	mp200_clear_dma(s->dma_ch);
#endif
	spin_unlock_irqrestore(&s->dma_lock, flags);

	FNC_EXIT return;
}

static void audio_process_dma(struct audio_stream *s)
{
	struct snd_pcm_substream *substream = s->stream;
	struct snd_pcm_runtime *runtime;
	unsigned int dma_size;
	unsigned int offset;
	unsigned int kick = 0;
	int ret;
#ifdef MP200_PCM_USE_PDMA
	unsigned int writable_flag = 0;
#endif
	FNC_ENTRY

#if 0
	/* we are requested to process synchronization DMA transfer */
	if (s->tx_spin) {
		snd_assert(s->stream_id == SNDRV_PCM_STREAM_PLAYBACK, return);
		/* fill the xmit dma buffers and return */
		while (1) {
			ret = mp200_start_dma(s->dma_ch, FORCE_CLOCK_ADDR, FORCE_CLOCK_SIZE,0,0);
			if (ret) {
				FNC_EXIT return;
			}
		}
		FNC_EXIT return;
	}
#endif

#ifdef MP200_PCM_USE_PDMA
	if (PCM0_TX(s)) {
		kick = mp200_pdma_status();
	} else {
		kick = mp200_dma_status(s->dma_ch);
	}
#else
	kick = mp200_dma_status(s->dma_ch);
#endif
	s->add_period = 0;

	/* must be set here - only valid for running streams, not for forced_clock dma fills  */
	runtime = substream->runtime;
	while (s->active && (s->periods < runtime->periods)) {
		if (!PCM0_TX(s)) {
			dma_size = frames_to_bytes(runtime, runtime->period_size);
			s->add_period = 1;
		}else{
			if( (output_ptr != 0)
				&& (output_ptr >= runtime->control->appl_ptr )) {
				/* new data not yet transfered from user area. */
				/* or data finished. */
				return;
			}
			dma_size = runtime->control->appl_ptr - output_ptr;

#ifdef MP200_PCM_PACK_PELIOD
#ifdef MP200_PCM_USE_PDMA
			if( (kick == 1) && 
				((runtime->status->state == SNDRV_PCM_STATE_PREPARED)
				|| (runtime->status->state == SNDRV_PCM_STATE_RUNNING)) ) {
				if( dma_size < runtime->buffer_size/2 ){
					// pdma if stop then not break?
					break;
				}
			}
#endif
			if( dma_size >= (runtime->buffer_size/2)) {
				s->add_period = (runtime->buffer_size/2)/runtime->period_size;
				if( s->add_period == 0 ) {
					s->add_period = 1;
				}
				dma_size = runtime->period_size * s->add_period;
			}else{
				s->add_period = (dma_size+(runtime->period_size-1))/runtime->period_size;
				if( s->add_period == 0 ) {
					s->add_period = 1;
				}
			}
#else
			if( dma_size > runtime->period_size ) {
				dma_size = runtime->period_size;
			}
			s->add_period = 1;
#endif
			dma_size = frames_to_bytes(runtime, dma_size);
		}

		if (s->old_offset) {
			/* a little trick, we need resume from old position */
			offset = frames_to_bytes(runtime, s->old_offset - 1);
			s->old_offset = 0;
			s->periods = 0;
			s->period = offset / dma_size;
			offset %= dma_size;
			dma_size = dma_size - offset;
			if (!dma_size)
				continue;		/* special case */
		} else {
			offset = frames_to_bytes(runtime, runtime->period_size) * s->period;
			/* snd_assert(dma_size <= DMA_BUF_SIZE, ); */
		}



#ifdef MP200_PCM_USE_PDMA
		if (PCM0_TX(s)){
			writable_flag = mp200_pdma_writable_flag();
		} else {
			writable_flag = mp200_dma_writable_flag(s->dma_ch);
		}

		if(0 == writable_flag){
#else
		if(0 == mp200_dma_writable_flag(s->dma_ch)) {
#endif
			dma_addr_t dma_ptr;
			dma_addr_t dst_ptr;
			if(s->txrx_id == PCM_TX_EN) {
				dma_ptr = runtime->dma_addr + offset;
				dst_ptr = s->phys_xq;
#ifdef MP200_PCM_USE_PDMA
				if(!PCM0_TX(s)){
					s->dma_regs->aoff = 0;
					s->dma_regs->asize = dma_size;
					s->dma_regs->leng = dma_size;
					s->dma_regs->mode = (MP200_DMAC_AMODE_BIT32
							   | MP200_DMAC_ENDI_R1032
							   | MP200_DMAC_ENDI_W3210);
				}
#else
				s->dma_regs->aoff = 0;
				s->dma_regs->asize = dma_size;
				s->dma_regs->leng = dma_size;
				s->dma_regs->mode = (MP200_DMAC_AMODE_BIT32
						   | MP200_DMAC_ENDI_R1032
						   | MP200_DMAC_ENDI_W3210);
#endif
			}
			else {
				dma_ptr = s->phys_xq;
				dst_ptr = runtime->dma_addr + offset;
				s->dma_regs->boff = 0;
				s->dma_regs->bsize = dma_size;
				s->dma_regs->leng = dma_size;
#if 0
				s->dma_regs->mode = MP200_DMAC_DEFMODE_16BIT;
#else
				s->dma_regs->mode = (MP200_DMAC_AMODE_BIT32
						   | MP200_DMAC_ENDI_R3210
						   | MP200_DMAC_ENDI_W1032);
#endif
			}
			d1b("dma ch   : %d\n",s->dma_ch);
			d1b("dma_ptr  : 0x%08x\n",dma_ptr);
			d1b("dst_ptr  : 0x%08x\n",dst_ptr);
			d1b("dma_size : %d\n",dma_size);
			d1b("bk count : %d\n",runtime->periods);

#ifdef MP200_PCM_USE_PDMA
			if (PCM0_TX(s)) {
				ret = mp200_start_pdma(dma_ptr, dma_size);
			} else {
				ret = mp200_start_dma(s->dma_ch,
							dma_ptr,
							0,
							dst_ptr,
							MP200_DMAC_INT_LENG_EN);
			}
#else
			ret = mp200_start_dma(s->dma_ch,
						dma_ptr,
						0,
						dst_ptr,
						MP200_DMAC_INT_LENG_EN);
#endif
			if (ret) {
				s->add_period = 0;
				printk(KERN_ERR "audio_process_dma: cannot queue DMA buffer (%i)\n", ret);
				FNC_EXIT return;
			}
 			output_ptr += bytes_to_frames(runtime, dma_size);
			if(0 == kick) {
				*s->txrx_en = s->txrx_id;
				d2b("PCM TXRX_EN(0x%08x) : 0x%08x\n",
					(u_int)s->txrx_en, s->txrx_id);
				udelay(50);	/* LRCLK1 22u: x2+a = 50 */
				kick = 1;
			}

		}
		else {
			s->add_period = 0;
			break;
		}

		d3b("%s%02d\n", ((PCM_TX_EN == s->txrx_id) ? "t" : "r"), s->period);
		d4b("dma ptr : 0x%08x\n", runtime->dma_addr + offset);

#ifdef MP200_PCM_USE_PDMA
		if (PCM0_TX(s)) {
			s->period += s->add_period;
			s->period %= runtime->periods;
			s->periods += s->add_period;
		} else {
			s->period++;
			s->period %= runtime->periods;
			s->periods++;
		}
#else
		s->period += s->add_period;
		s->period %= runtime->periods;
		s->periods += s->add_period;
#endif
	}

	FNC_EXIT return;
}

static void audio_dma_callback(void *data, int intsts, int intrawsts)
{
	struct audio_stream *s = data;
	FNC_ENTRY

	/*
	 * If we are getting a callback for an active stream then we inform
	 * the PCM middle layer we've finished a period
	 */
 	if (s->active) {
		snd_pcm_period_elapsed(s->stream);
		d6b("pos : %d\n", audio_get_dma_pos(s));
		d3b("pos : %d\n", audio_get_dma_pos(s));
	}

	spin_lock(&s->dma_lock);
	if (!s->tx_spin && s->periods > 0)
		s->periods -= s->add_period;

	audio_process_dma(s);
	spin_unlock(&s->dma_lock);

	FNC_EXIT return;
}

#ifdef MP200_PCM_USE_PDMA
static irqreturn_t audio_pdma_callback(int irq, void *data)
{
	struct audio_stream *s = data;
	struct snd_pcm_substream *substream = s->stream;
	struct snd_pcm_runtime *runtime = substream->runtime;
	FNC_ENTRY

	/*
	 * If we are getting a callback for an active stream then we inform
	 * the PCM middle layer we've finished a period
	 */
 	if (s->active) {
		snd_pcm_period_elapsed(s->stream);
		d6b("pos : %d\n", audio_get_dma_pos(s));
		d3b("pos : %d\n", audio_get_dma_pos(s));
	}

	spin_lock(&s->dma_lock);
	if( output_ptr != runtime->control->appl_ptr ) {
		if (!s->tx_spin && s->periods > 0)
			s->periods -= s->add_period;
		audio_process_dma(s);
	}
	spin_unlock(&s->dma_lock);

	FNC_EXIT return IRQ_HANDLED;
}
#endif

/* PCM setting */

/* HW params & free */

static const struct snd_pcm_hardware mp200_pcm_hardware = {
	.info			= (SNDRV_PCM_INFO_INTERLEAVED |
				   SNDRV_PCM_INFO_BLOCK_TRANSFER |
				   SNDRV_PCM_INFO_MMAP |
				   SNDRV_PCM_INFO_MMAP_VALID |
				   SNDRV_PCM_INFO_PAUSE |
				   SNDRV_PCM_INFO_RESUME),
#if defined(MP200_PCM_USE_PDMA) && defined(MP200_PCM_FMT16)
	.formats = SNDRV_PCM_FMTBIT_S32_LE,			/* ilegal setting */
#else
	.formats = SNDRV_PCM_FMTBIT_S16_LE,
#endif

#ifdef MP200_PCM_USE_PDMA
	.buffer_bytes_max	= 128*1024,
	.period_bytes_min	= 64,
	.period_bytes_max	= 0x10000,
	.periods_min		= 2,
	.periods_max		= 32,
#else
	.buffer_bytes_max	= 0xFFFE*16,
	.period_bytes_min	= 64,
	.period_bytes_max	= 0xFFFE,
	.periods_min		= 2,
	.periods_max		= 16,
#endif
	.fifo_size		= 32,
};

#ifdef MP200_PCM_INTERRUPT_ENABLE

#include <linux/interrupt.h>

/* PMx_INT Masks */
#define PCM_INT_TX_FRE		0x01
#define PCM_INT_TX_URE		0x02
#define PCM_INT_TX_ORE		0x04
#define PCM_INT_TX_WEN		0x08
#define PCM_INT_RX_FRE		0x10
#define PCM_INT_RX_URE		0x20
#define PCM_INT_RX_ORE		0x40
#define PCM_INT_RX_REN		0x80
#define PCM_INT_TX_STP		(1 << 8)
#define PCM_INT_RX_STP		(1 << 12)

static void mp200_pcm_kick(unsigned long inData)
{
	struct audio_stream *s = (struct audio_stream *)inData;
	struct mp200_pcm_client *chip = snd_pcm_substream_chip(s->stream);
	int wait, mode, val, status;
	FNC_ENTRY

	mode = *s->func_sel &  (MP200_PCM_FUNC_MODE0 |
				MP200_PCM_FUNC_MODE1 |
				MP200_PCM_FUNC_MODE2 |
				MP200_PCM_FUNC_MODE3 |
				MP200_PCM_FUNC_MODE4 |
				MP200_PCM_FUNC_MODE5 |
				MP200_PCM_FUNC_MODE6);

	switch(mode) {
	case MP200_PCM_FUNC_MODE5:
	case MP200_PCM_FUNC_MODE6:
		val = (((chip->pcm_regs->cycle & 0xff) *
			(((chip->pcm_regs->cycle >> 8) & 0x1f) + 1)) +
		       ((chip->pcm_regs->cycle2 & 0xff) *
			(((chip->pcm_regs->cycle >> 8) & 0x1f) + 1)));
		break;
	/* mode0-4 */
	default:
		val = (chip->pcm_regs->cycle & 0xff) + 1;
		break;
	}

	wait = (((1000000/s->stream->runtime->rate) * val) +
		((1000000%s->stream->runtime->rate)?1:0));

	udelay(wait);

#ifdef MP200_PCM_USE_PDMA
	if (PCM0_TX(s)) {
		if( s->stream->runtime->status->state == SNDRV_PCM_STATE_RUNNING ) {
			status = 1;
		}else{
			status = 0;
		}
	} else {
		status = mp200_dma_status(s->dma_ch);
	}
#else
	status = mp200_dma_status(s->dma_ch);
#endif

	if(1 == status){
		*s->txrx_en = s->txrx_id;
	} else {
		unsigned long	flags;

		local_irq_save(flags);
		audio_stop_dma(s);
		audio_process_dma(s);
		local_irq_restore(flags);
	}

	FNC_EXIT return;
}

DECLARE_TASKLET(mp200_pcm_tx_tasklet, mp200_pcm_kick, 0);
DECLARE_TASKLET(mp200_pcm_rx_tasklet, mp200_pcm_kick, 0);

static irqreturn_t audio_pcm_callback(int irq, void *dev_id)
{
	struct snd_pcm *pcm = (struct snd_pcm *) dev_id;
	struct mp200_pcm_client *chip = pcm->private_data;
	static irqreturn_t ret = IRQ_NONE;
	FNC_ENTRY

	/* printk("%s\n", __FUNCTION__); */

	d2b("PCM STATUS(0x%08x) : 0x%08x\n",
		(u_int)&chip->pcm_regs->status, chip->pcm_regs->status);

	if(chip->pcm_regs->status & ~(PCM_INT_TX_FRE | PCM_INT_TX_URE |
				      PCM_INT_TX_ORE | PCM_INT_RX_FRE |
				      PCM_INT_RX_URE | PCM_INT_RX_ORE)) {

		chip->pcm_regs->clear = chip->pcm_regs->status;
		chip->pcm_regs->txrx_dis = PCM_TX_EN | PCM_RX_EN;

		printk(KERN_INFO "%s(): unable to pcm status 0x%08x\n",
					       __FUNCTION__, chip->pcm_regs->status);
		return ret;
	}

	if(chip->pcm_regs->status & (PCM_INT_TX_FRE | PCM_INT_TX_URE | PCM_INT_TX_ORE)) {

		chip->pcm_regs->clear = PCM_INT_TX_FRE | PCM_INT_TX_URE |
					PCM_INT_TX_STP;
		chip->pcm_regs->txrx_dis = PCM_TX_EN;
		d2b("PCM TXRX_EN(0x%08x) : 0x%08x\n",
			(u_int)&chip->pcm_regs->txrx_en, chip->pcm_regs->txrx_en);

		if (chip->s[SNDRV_PCM_STREAM_PLAYBACK]) {
			struct audio_stream *s = chip->s[SNDRV_PCM_STREAM_PLAYBACK];
			if (s->active) {
				mp200_pcm_tx_tasklet.data = (unsigned long)s;
				tasklet_schedule(&mp200_pcm_tx_tasklet);
			}
			ret = IRQ_HANDLED;
		}
		else {
			printk(KERN_INFO "%s(): unable to pcm tx status 0x%08x\n",
					       __FUNCTION__, chip->pcm_regs->status);
			return ret;
		}
	}
	if(chip->pcm_regs->status & (PCM_INT_RX_FRE | PCM_INT_RX_URE | PCM_INT_RX_ORE)) {

		chip->pcm_regs->clear = PCM_INT_RX_FRE | PCM_INT_RX_STP;
		chip->pcm_regs->txrx_dis = PCM_RX_EN;
		d2b("PCM TXRX_EN(0x%08x) : 0x%08x\n",
			(u_int)&chip->pcm_regs->txrx_en, chip->pcm_regs->txrx_en);

		if (chip->s[SNDRV_PCM_STREAM_CAPTURE]) {
			struct audio_stream *s = chip->s[SNDRV_PCM_STREAM_CAPTURE];
			if (s->active) {
				mp200_pcm_rx_tasklet.data = (unsigned long)s;
				tasklet_schedule(&mp200_pcm_rx_tasklet);
			}
			ret = IRQ_HANDLED;
		}
		else {
			printk(KERN_INFO "%s(): unable to pcm rx status 0x%08x\n",
					       __FUNCTION__, chip->pcm_regs->status);
			return ret;
		}
	}

	FNC_EXIT return ret;
}
#endif

static int mp200_pcm_hw_params(struct snd_pcm_substream * substream,
					struct snd_pcm_hw_params * hw_params)
{
#ifdef MP200_PCM_USE_PDMA
	int ret = 0;
	FNC_ENTRY

	struct mp200_pcm_client *chip = snd_pcm_substream_chip(substream);
	if(PCM0_TX(chip->s[substream->pstr->stream])) {
		struct snd_pcm_runtime *runtime = substream->runtime;

		snd_pcm_set_runtime_buffer(substream, &substream->dma_buffer);
		runtime->dma_bytes = params_buffer_bytes(hw_params);
	}
	else {
		ret = snd_pcm_lib_malloc_pages(substream,
			params_buffer_bytes(hw_params));
	}
	FNC_EXIT return ret;
#else
	int ret;
	FNC_ENTRY
	ret = snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params));
	FNC_EXIT return ret;
#endif
}

static int mp200_pcm_hw_free(struct snd_pcm_substream * substream)
{
#ifdef MP200_PCM_USE_PDMA
	int ret = 0;
	FNC_ENTRY

	struct mp200_pcm_client *chip = snd_pcm_substream_chip(substream);
	if(PCM0_TX(chip->s[substream->pstr->stream])) {
		snd_pcm_set_runtime_buffer(substream, NULL);
	}
	else {
		ret = snd_pcm_lib_free_pages(substream);
	}
	FNC_EXIT return ret;
#else
	int ret;
	FNC_ENTRY
	ret = snd_pcm_lib_free_pages(substream);
	FNC_EXIT return ret;
#endif
}

/* trigger & timer */

static int mp200_pcm_trigger(struct snd_pcm_substream * substream, int cmd)
{
	struct mp200_pcm_client *chip = snd_pcm_substream_chip(substream);
	int stream_id = substream->pstr->stream;
	struct audio_stream *s = chip->s[stream_id];
	int err = 0;
	FNC_ENTRY

	/* note local interrupts are already disabled in the midlevel code */
	spin_lock(&s->dma_lock);
	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		d1b("SNDRV_PCM_TRIGGER_START\n");
		/* requested stream startup */
		s->active = 1;
		audio_process_dma(s);
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		d1b("SNDRV_PCM_TRIGGER_STOP\n");
		/* requested stream shutdown */
		audio_stop_dma(s);
		break;
	case SNDRV_PCM_TRIGGER_SUSPEND:
		d1b("SNDRV_PCM_TRIGGER_SUSPEND\n");
		s->active = 0;

#ifdef MP200_PCM_USE_PDMA
		if (PCM0_TX(s)) {
			mp200_stop_pdma();
		} else {
			mp200_stop_dma(s->dma_ch);
		}
#else
		mp200_stop_dma(s->dma_ch);
#endif
		s->old_offset = audio_get_dma_pos(s) + 1;
		s->periods = 0;
		break;
	case SNDRV_PCM_TRIGGER_RESUME:
		d1b("SNDRV_PCM_TRIGGER_RESUME\n");
		s->active = 1;
		s->tx_spin = 0;
		audio_process_dma(s);
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
		d1b("SNDRV_PCM_TRIGGER_PAUSE_PUSH\n");
#ifdef MP200_PCM_USE_PDMA
		if (PCM0_TX(s)) {
			mp200_stop_pdma();
		} else {
			mp200_stop_dma(s->dma_ch);
		}
#else
		mp200_stop_dma(s->dma_ch);
#endif
		s->active = 0;
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
		d1b("SNDRV_PCM_TRIGGER_PAUSE_RELEASE\n");
		s->active = 1;
		if (s->old_offset) {
			s->tx_spin = 0;
			audio_process_dma(s);
			break;
		}
		break;
	default:
		err = -EINVAL;
		break;
	}
	spin_unlock(&s->dma_lock);
	FNC_EXIT return err;
}

static int mp200_pcm_prepare(struct snd_pcm_substream * substream)
{
	struct mp200_pcm_client *chip = snd_pcm_substream_chip(substream);
	struct audio_stream *s = chip->s[substream->pstr->stream];
	FNC_ENTRY

	s->period = 0;
	s->periods = 0;

	/* set requested samplerate */
	FNC_EXIT return chip->prepare(substream);
}

static snd_pcm_uframes_t mp200_pcm_pointer(struct snd_pcm_substream * substream)
{
	struct mp200_pcm_client *chip = snd_pcm_substream_chip(substream);
	struct audio_stream *s = chip->s[substream->pstr->stream];
	struct snd_pcm_runtime *runtime = substream->runtime;
	snd_pcm_uframes_t ret;
	FNC_ENTRY
#if 1
	spin_lock(&s->dma_lock);
	if( output_ptr != runtime->control->appl_ptr ) {
		if (!s->tx_spin && s->periods > 0)
			s->periods -= s->add_period;
		audio_process_dma(s);
	}
	spin_unlock(&s->dma_lock);
#endif

	ret = audio_get_dma_pos(chip->s[substream->pstr->stream]);
	FNC_EXIT return ret;
}

#ifdef MP200_PCM_MMAP_ENABLE
static int mp200_pcm_mmap(struct snd_pcm_substream *substream, struct vm_area_struct *vma)
{
	int ret;
	struct snd_pcm_runtime *runtime = substream->runtime;
	FNC_ENTRY

#ifdef MP200_PCM_USE_PDMA
	struct mp200_pcm_client *chip = snd_pcm_substream_chip(substream);

	if(PCM0_TX(chip->s[substream->pstr->stream])) {
		unsigned long user_size;
		user_size = (vma->vm_end - vma->vm_start) >> PAGE_SHIFT;

		vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
		ret = remap_pfn_range(vma, vma->vm_start,
			(runtime->dma_addr >> PAGE_SHIFT) + vma->vm_pgoff,
				runtime->dma_bytes, vma->vm_page_prot);
	} else {
		ret = dma_mmap_writecombine(substream->pcm->card->dev, vma,
				     runtime->dma_area,
				     runtime->dma_addr,
				     runtime->dma_bytes);
	}
#else
	ret = dma_mmap_writecombine(substream->pcm->card->dev, vma,
				     runtime->dma_area,
				     runtime->dma_addr,
				     runtime->dma_bytes);
#endif

/*	printk("%s: ret=%d\n", __FUNCTION__, ret); */
	FNC_EXIT return ret;
}
#endif

static int mp200_pcm_open(struct snd_pcm_substream * substream)
{
	struct mp200_pcm_client *chip = snd_pcm_substream_chip(substream);
	int ret;
	unsigned long flags;
	FNC_ENTRY

#ifdef MP200_PCM_USE_PDMA
	if (IS_PCM0(pcm_ch[chip->pcm_ch])) {
		ret  = mp200_pcm0_request_irq(audio_pcm_callback, IRQF_DISABLED,
				pcm_ch[chip->pcm_ch].name,
				(void *)pcm_ch[chip->pcm_ch].pcm, MP200_PCMMODE_ALSA);

		if(ret != 0){
			printk("error request_irq = %d\n",ret);
			FNC_EXIT return ret;
		}
		disable_irq(pcm_ch[chip->pcm_ch].irq);

		if (PCM0_TX(chip->s[substream->pstr->stream])) {
			mp200_pdma_deinit();
			mp200_pdma_init();
		} else {
			writel(0, PDMA_DMA_SEL);
		}
	}
#endif

	chip->s[substream->pstr->stream]->stream = substream;

	substream->runtime->hw = mp200_pcm_hardware;

	local_irq_save(flags);
	power_flag++;
	local_irq_restore(flags);

	ret = chip->startup(substream);
	if (ret < 0) {
		goto err1;
	}

	ret = audio_dma_request(chip->s[substream->pstr->stream],
				audio_dma_callback);
	if (ret < 0) {
		goto err2;
	}

#ifdef MP200_PCM_INTERRUPT_ENABLE
	if(chip->s[substream->pstr->stream]->txrx_id == PCM_TX_EN) {
		chip->pcm_regs->enset = PCM_INT_TX_FRE | PCM_INT_TX_URE | PCM_INT_TX_ORE;
	}
	else {
		chip->pcm_regs->enset = PCM_INT_RX_FRE | PCM_INT_RX_URE | PCM_INT_RX_ORE;
	}
	enable_irq(pcm_ch[chip->pcm_ch].irq);
#endif

#ifdef MP200_DAC_HANGUP_PDMA
	ret = mp200_timer_register_cb(DAC_HANG_TG, tg_name, timer_callback, (void *)tg_name);
	ret = mp200_timer_set_period(DAC_HANG_TG, timer_usec);
#endif

	FNC_EXIT return ret;
err2:
	chip->shutdown(substream);
err1:
	local_irq_save(flags);
	power_flag--;
	local_irq_restore(flags);

	chip->s[substream->pstr->stream]->stream = NULL;
#ifdef MP200_PCM_USE_PDMA
	if (IS_PCM0(pcm_ch[chip->pcm_ch])){
		mp200_pcm0_free_irq((void *)pcm_ch[chip->pcm_ch].pcm,
			MP200_PCMMODE_ALSA);
	}
#endif
	FNC_EXIT return ret;
}

static int mp200_pcm_close(struct snd_pcm_substream * substream)
{
	struct mp200_pcm_client *chip = snd_pcm_substream_chip(substream);
	unsigned long flags;
	FNC_ENTRY

	audio_dma_free(chip->s[substream->pstr->stream]);

#ifdef MP200_PCM_INTERRUPT_ENABLE
	disable_irq(pcm_ch[chip->pcm_ch].irq);
#ifdef MP200_PCM_USE_PDMA
	if (IS_PCM0(pcm_ch[chip->pcm_ch])) {
		mp200_pdma_deinit();
		mp200_pdma_init();
		mp200_pcm0_free_irq((void *)pcm_ch[chip->pcm_ch].pcm,
					MP200_PCMMODE_ALSA);
	}
#endif
	if(chip->s[substream->pstr->stream]->txrx_id == PCM_TX_EN) {
		chip->pcm_regs->enclr = PCM_INT_TX_FRE | PCM_INT_TX_URE |
					PCM_INT_TX_ORE | PCM_INT_TX_WEN |
					PCM_INT_TX_STP;
		chip->pcm_regs->clear = PCM_INT_TX_FRE | PCM_INT_TX_URE |
					PCM_INT_TX_STP;
	}
	else {
		chip->pcm_regs->enclr = PCM_INT_RX_FRE | PCM_INT_RX_URE |
					PCM_INT_RX_ORE | PCM_INT_RX_REN |
					PCM_INT_RX_STP;
		chip->pcm_regs->clear = PCM_INT_RX_FRE | PCM_INT_RX_STP;
	}
	d2b("PCM TXRX_EN(0x%08x) : 0x%08x\n",
		(u_int)&chip->pcm_regs->txrx_en, chip->pcm_regs->txrx_en);
#endif

	chip->shutdown(substream);

	local_irq_save(flags);
	power_flag--;
	local_irq_restore(flags);

	chip->s[substream->pstr->stream]->stream = NULL;

#ifdef MP200_DAC_HANGUP_PDMA
	mp200_timer_unregister_cb(DAC_HANG_TG, (void *)tg_name);
#endif

	FNC_EXIT return 0;
}

static struct snd_pcm_ops mp200_pcm_ops = {
	.open		= mp200_pcm_open,
	.close		= mp200_pcm_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= mp200_pcm_hw_params,
	.hw_free	= mp200_pcm_hw_free,
	.prepare	= mp200_pcm_prepare,
	.trigger	= mp200_pcm_trigger,
	.pointer	= mp200_pcm_pointer,
#ifdef MP200_PCM_MMAP_ENABLE
	.mmap		= mp200_pcm_mmap,
#endif
};

#ifdef CONFIG_PM
int mp200_pcm_suspend(struct platform_device *dev, pm_message_t state)
{
	FNC_ENTRY

	switch (state.event) {
	case DEV_SUSPEND_IDLE_1:
	case PM_EVENT_SUSPEND:
		if (power_flag != 0) {
#ifdef MP200_PCM_USE_PDMA
			if((pdma_in_use) && (power_flag == 1))
				break;
#endif
			printk("busy\n");
			FNC_EXIT return -EBUSY;
		}
		break;
	default:
		break;
	}

	FNC_EXIT return 0;
}

int mp200_pcm_resume(struct platform_device *dev)
{
	FNC_ENTRY

	FNC_EXIT return 0;
}
#endif

int mp200_pcm_free(struct mp200_pcm_client *client)
{
	int i;
	FNC_ENTRY

	if (pcm_ch[client->pcm_ch].in_use) {
#ifdef MP200_PCM_INTERRUPT_ENABLE
#ifdef MP200_PCM_USE_PDMA
		if(!IS_PCM0(pcm_ch[client->pcm_ch])){
			free_irq(pcm_ch[client->pcm_ch].irq,
				(void *)pcm_ch[client->pcm_ch].pcm);
		}
#else
		free_irq(pcm_ch[client->pcm_ch].irq,
				(void *)pcm_ch[client->pcm_ch].pcm);
#endif
#endif
		mp200_pmu_clkctrl_off(pcm_ch[client->pcm_ch].clkctl);
		mp200_pmu_reset_device(pcm_ch[client->pcm_ch].reset);
		d2b("mp200_pmu_reset_device(0x%08x)\n",
				pcm_ch[client->pcm_ch].reset);
		mp200_pmu_close_clockgate(pcm_ch[client->pcm_ch].clock_s);
		mp200_pmu_close_clockgate(pcm_ch[client->pcm_ch].clock_p);
		pcm_ch[client->pcm_ch].in_use = 0;
	}

	for (i=0; i<MP200_PCM_MAX_CHANNELS; i++) {
		if (pcm_ch[i].in_use) {
			break;
		}
	}

	if (i == MP200_PCM_MAX_CHANNELS) {
		mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_PCLK);
	}

	FNC_EXIT return 0;
}

#ifdef MP200_PCM_USE_PDMA

static int mp200_pcm_preallocate_dma_buffer(struct snd_pcm *pcm, int stream,
	unsigned long addr, size_t size)
{
	struct snd_pcm_substream *substream = pcm->streams[stream].substream;
	struct snd_dma_buffer *buf = &substream->dma_buffer;
	buf->dev.type = SNDRV_DMA_TYPE_DEV;
	buf->dev.dev = pcm->card->dev;
	buf->private_data = NULL;

	buf->addr = addr;
	buf->area = ioremap_nocache(buf->addr, size);
	if (!buf->area)
		return -ENOMEM;
	buf->bytes = size;

	return 0;
}
#endif

int mp200_pcm_new(struct snd_card *card, struct mp200_pcm_client *client, struct snd_pcm **rpcm)
{
	struct snd_hwdep *hw;
	struct snd_pcm *pcm;
	int play = client->s[SNDRV_PCM_STREAM_PLAYBACK] ? 1 : 0;
	int capt = client->s[SNDRV_PCM_STREAM_CAPTURE] ? 1 : 0;
	int ret = -EBUSY;
	FNC_ENTRY

	d2b("play : %d, capt : %d\n",play, capt);

	if (pcm_ch[client->pcm_ch].in_use)
		goto out;

	ret = snd_hwdep_new(card, pcm_ch[client->pcm_ch].name, client->pcm_ch, &hw);

	if (ret)
		goto out;

	hw->private_data = client;

	hw->ops.open = mp200_pcm_hwdep_open;
	hw->ops.ioctl = mp200_pcm_hwdep_ioctl;
	hw->ops.release = mp200_pcm_hwdep_release;

	ret = snd_pcm_new(card, pcm_ch[client->pcm_ch].name, client->pcm_ch, play, capt, &pcm);

	if (ret)
		goto out;

#ifdef MP200_PCM_USE_PDMA
	if (!IS_PCM0(pcm_ch[client->pcm_ch])) {
		snd_pcm_lib_preallocate_pages_for_all(pcm,
			SNDRV_DMA_TYPE_DEV, NULL, 128*1024, 128*1024);
	}
#else
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,
					      /* snd_pcm_dma_flags(0)*/ NULL,
					      128*1024, 128*1024);
#endif
	pcm->private_data = client;

	client->pcm_regs = (struct pcm_regs *)IO_ADDRESS(pcm_ch[client->pcm_ch].base);

	d2b("pcm base : 0x%08x\n",pcm_ch[client->pcm_ch].base);
	d2b("pcm reg  : 0x%08x\n",IO_ADDRESS(pcm_ch[client->pcm_ch].base));
	d2b("pcm txq  : 0x%08x\n",(dma_addr_t)&((struct pcm_regs*)pcm_ch[client->pcm_ch].base)->txq);

	if (play) {
		int stream = SNDRV_PCM_STREAM_PLAYBACK;
		struct snd_pcm_substream * substream;
		struct audio_stream *s = client->s[stream];

		snd_pcm_set_ops(pcm, stream, &mp200_pcm_ops);
		substream = pcm->streams[stream].substream;
#ifdef MP200_PCM_USE_PDMA
		if (IS_PCM0(pcm_ch[client->pcm_ch])) {
			/* BANK 10 -- SRAM */
			ret = mp200_pcm_preallocate_dma_buffer(pcm,
				 stream, SRAM_BASE, SRAM_SIZE);
			if (ret)
				goto out;
		}
#endif
		strncpy(substream->name, client->s[stream]->id, sizeof(substream->name));
		d2b("name : %s \n", substream->name);
		s->phys_xq = (dma_addr_t)&((struct pcm_regs*)pcm_ch[client->pcm_ch].base)->txq;
		s->txrx_id = PCM_TX_EN;
		s->txrx_en = &client->pcm_regs->txrx_en;
		s->txrx_dis = &client->pcm_regs->txrx_dis;
#ifdef MP200_PCM_INTERRUPT_ENABLE
		s->func_sel = &client->pcm_regs->func_sel;
#endif
		spin_lock_init(&s->dma_lock);
	}
	if (capt) {
		int stream = SNDRV_PCM_STREAM_CAPTURE;
		struct snd_pcm_substream * substream;
		struct audio_stream *s = client->s[stream];

		snd_pcm_set_ops(pcm, stream, &mp200_pcm_ops);
		substream = pcm->streams[stream].substream;
#ifdef MP200_PCM_USE_PDMA
		if (IS_PCM0(pcm_ch[client->pcm_ch])) {
			ret = snd_pcm_lib_preallocate_pages(substream,
				SNDRV_DMA_TYPE_DEV, NULL, 64*1024, 64*1024);
			if (ret)
				goto out;
		}
#endif
		strncpy(substream->name, client->s[stream]->id, sizeof(substream->name));
		d2b("name : %s \n", substream->name);
		s->phys_xq = (dma_addr_t)&((struct pcm_regs*)pcm_ch[client->pcm_ch].base)->rxq;
		s->txrx_id = PCM_RX_EN;
		s->txrx_en = &client->pcm_regs->txrx_en;
		s->txrx_dis = &client->pcm_regs->txrx_dis;
#ifdef MP200_PCM_INTERRUPT_ENABLE
		s->func_sel = &client->pcm_regs->func_sel;
#endif
		spin_lock_init(&s->dma_lock);
	}

	if (rpcm)
		*rpcm = pcm;

	d2b("mp200_pmu_unreset_device(0x%08x)\n",pcm_ch[client->pcm_ch].reset);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_PCLK);
	mp200_pmu_open_clockgate(pcm_ch[client->pcm_ch].clock_p);
	mp200_pmu_open_clockgate(pcm_ch[client->pcm_ch].clock_s);
	mp200_pmu_clkctrl_off(pcm_ch[client->pcm_ch].clkctl);
	mp200_pmu_unreset_device(pcm_ch[client->pcm_ch].reset);
	mp200_pmu_clkctrl_on(pcm_ch[client->pcm_ch].clkctl);
	mp200_pmu_close_clockgate(pcm_ch[client->pcm_ch].clock_s);
#ifdef MP200_PCM_INTERRUPT_ENABLE
	client->pcm_regs->enclr = PCM_INT_TX_FRE | PCM_INT_TX_URE |
				  PCM_INT_TX_ORE | PCM_INT_TX_WEN |
				  PCM_INT_RX_FRE | PCM_INT_RX_URE |
				  PCM_INT_RX_ORE | PCM_INT_RX_REN |
				  PCM_INT_TX_STP | PCM_INT_RX_STP;
	client->pcm_regs->clear = PCM_INT_TX_FRE | PCM_INT_TX_URE |
				  PCM_INT_RX_FRE |
				  PCM_INT_TX_STP | PCM_INT_RX_STP;
	client->pcm_regs->txrx_dis = PCM_TX_EN | PCM_RX_EN;
	d2b("PCM TXRX_EN(0x%08x) : 0x%08x\n",
		(u_int)&client->pcm_regs->txrx_en, client->pcm_regs->txrx_en);

#ifdef MP200_PCM_USE_PDMA
	if (!IS_PCM0(pcm_ch[client->pcm_ch])) {
		ret = request_irq(pcm_ch[client->pcm_ch].irq,
			audio_pcm_callback, IRQF_DISABLED,
			pcm_ch[client->pcm_ch].name, (void *)pcm);

		disable_irq(pcm_ch[client->pcm_ch].irq);
	}
#else
	ret = request_irq(pcm_ch[client->pcm_ch].irq,
			audio_pcm_callback, IRQF_DISABLED,
			pcm_ch[client->pcm_ch].name, (void *)pcm);

	disable_irq(pcm_ch[client->pcm_ch].irq);
#endif
	pcm_ch[client->pcm_ch].pcm = pcm;
#endif

	if (0 == ret) {
		pcm_ch[client->pcm_ch].in_use = 1;
	}
#ifdef MP200_DAC_HANGUP_PDMA
	if(IS_PCM0(pcm_ch[client->pcm_ch])){
	    mp200_pdma_workqueue = create_singlethread_workqueue(pcm_ch[client->pcm_ch].name);
	    INIT_WORK(&mp200_pdma_dac_work, mp200_dac_work_do);
	}
#endif

 out:
	FNC_EXIT return ret;
}

static int mp200_pcm_hwdep_open(struct snd_hwdep * hw, struct file *file)
{
	FNC_ENTRY

	FNC_EXIT return 0;
}

static int mp200_pcm_hwdep_ioctl(struct snd_hwdep * hw, struct file *file,
				 unsigned int cmd, unsigned long arg)
{
	struct mp200_pcm_client *chip = hw->private_data;
	pcm_ctrl_t sett;
	int ret = 0;

	FNC_ENTRY
	/* dispatch based on command */
	switch (cmd) {
	case SNDRV_MP200_IOCTL_PCM_SET_CTRL:
		d2b("SNDRV_MP200_IOCTL_PCM_SET_CTRL cmd=0x%08x\n",cmd);
		if (copy_from_user(&sett, (pcm_ctrl_t *) arg, sizeof(sett)) != 0) {
			FNC_EXIT return -EFAULT;
		}

		ret = mp200_pcm_set_ctrl(chip->pcm_regs, &sett);

		if (ret < 0) {
			FNC_EXIT return ret;
		}

		*chip->sett = sett;

		FNC_EXIT return ret;

	case SNDRV_MP200_IOCTL_PCM_GET_CTRL:
		d2b("SNDRV_MP200_IOCTL_PCM_GET_CTRL cmd=0x%08x\n",cmd);
		sett = *chip->sett;
		FNC_EXIT return copy_to_user((void *)arg, &sett, sizeof(sett));

	default:
		d2b("not supported cmd=0x%08x\n",cmd);
		FNC_EXIT return -EINVAL;
	}

	return 0;
}

static int mp200_pcm_hwdep_release(struct snd_hwdep * hw, struct file *file)
{
	FNC_ENTRY

	FNC_EXIT return 0;
}

int mp200_pcm_set_ctrl(struct pcm_regs *regs, pcm_ctrl_t * s)
{
	pcm_ctrl_t *sett = s;
	unsigned int func_val = 0;
	unsigned int cycle_val = 0;
	unsigned int cycle2_val = 0;

	if (sett == NULL) {
		return -EINVAL;
	}

	/* mode_sel : Operation mode */
	switch (sett->func.mode_sel) {
	case PCM_MODE_0:	/* mode 0 */
		func_val |= MP200_PCM_FUNC_MODE0;
		break;
	case PCM_MODE_1:	/* mode 1 */
		func_val |= MP200_PCM_FUNC_MODE1;
		break;
	case PCM_MODE_2:	/* mode 2 (I2S Format) */
		func_val |= MP200_PCM_FUNC_MODE2;
		break;
	case PCM_MODE_3:	/* mode 3 (MSB First)  */
		func_val |= MP200_PCM_FUNC_MODE3;
		break;
	case PCM_MODE_4:	/* mode 4 (LSB First)  */
		func_val = MP200_PCM_FUNC_MODE4;
		break;
	case PCM_MODE_5:	/* mode 5 (Multi Channel Mode) */
		func_val |= MP200_PCM_FUNC_MODE5;
		break;
	case PCM_MODE_6:	/* mode 6 (Multi Channel Mode) */
		func_val |= MP200_PCM_FUNC_MODE6;
		break;
	default:
		return -EINVAL;
	}

	/* m_s : Master/Slave */
	switch (sett->func.m_s) {
	case PCM_MASTER_MODE:	/* Master */
		func_val |= MP200_PCM_FUNC_MASTER;
		break;
	case PCM_SLAVE_MODE:	/* Slave  */
		func_val |= MP200_PCM_FUNC_SLAVE;
		break;
	default:
		return -EINVAL;
	}

	/* tx_tim : Transmission timing */
	switch (sett->func.tx_tim) {
	case PCM_TX_30_WORD:	/* 30word */
		func_val |= MP200_PCM_FUNC_30WORD;
		break;
	case PCM_TX_16_WORD:	/* 16word */
		func_val |= MP200_PCM_FUNC_16WORD;
		break;
	case PCM_TX_08_WORD:	/*  8word */
		func_val |= MP200_PCM_FUNC_08WORD;
		break;
	case PCM_TX_04_WORD:	/*  4word */
		func_val |= MP200_PCM_FUNC_04WORD;
		break;
	default:
		return -EINVAL;
	}

	switch (sett->func.mode_sel) {
	case PCM_MODE_0:	/* mode 0 */
	case PCM_MODE_1:	/* mode 1 */
	case PCM_MODE_2:	/* mode 2 (I2S Format) */
	case PCM_MODE_3:	/* mode 3 (MSB First)  */
	case PCM_MODE_4:	/* mode 4 (LSB First)  */
		/* cyc_val : Frame length */
		if ((0x3f < sett->cyc.cyc_val) || (sett->cyc.cyc_val < 0x07)) {
			return -EINVAL;
		}
		if ((sett->cyc.cyc_val < sett->cyc.sib)
		    || (sett->cyc.cyc_val < sett->cyc.sob)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.cyc_val;

		/* sib : Data bit length (PMx_SI) */
		if ((0x1f < sett->cyc.sib) || (sett->cyc.sib < 0x07)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.sib << 8;

		/* rx_pd : Data padding (RX) */
		if ((sett->cyc.rx_pd != PCM_PADDING_OFF)
		    && (sett->cyc.rx_pd != PCM_PADDING_ON)) {
			return -EINVAL;
		}
		if (sett->cyc.rx_pd == PCM_PADDING_ON) {
			if ((sett->cyc.sib != 0x0f)
			    && (sett->cyc.sib != 0x07)) {
				d2b("[warning]PADDING_ON\tSIB : 0x%08x\n",
					sett->cyc.sib);
			}
			cycle_val |= sett->cyc.rx_pd << 15;
		}

		/* sob : Data bit length (PMx_SO) */
		if ((0x1f < sett->cyc.sob) || (sett->cyc.sob < 0x07)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.sob << 16;

		/* tx_pd : Data padding (TX) */
		if ((sett->cyc.tx_pd != PCM_PADDING_OFF)
		    && (sett->cyc.tx_pd != PCM_PADDING_ON)) {
			return -EINVAL;
		}
		if (sett->cyc.tx_pd == PCM_PADDING_ON) {
			if ((sett->cyc.sob != 0x0f)
			    && (sett->cyc.sob != 0x07)) {
				d2b("[warning]PADDING_ON\tSOB : 0x%08x\n",
					sett->cyc.sob);
			}
			cycle_val |= sett->cyc.tx_pd << 23;
		}
		break;
	case PCM_MODE_5:	/* mode 5 (Multi Channel Mode) */
	case PCM_MODE_6:	/* mode 6 (Multi Channel Mode) */
		/* cyc_val : Frame length */
		if (0x80 < sett->cyc.cyc_val) {
			return -EINVAL;
		}
		if ((sett->cyc.cyc_val == 0) && (sett->cyc2.cyc_val2 == 0)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.cyc_val;

		/* sib : Data bit length (PMx_SI) */
		if ((0x1f < sett->cyc.sib) || (sett->cyc.sib < 0x07)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.sib << 8;

		/* rx_pd : Data padding (RX) */
		if (sett->cyc.rx_pd != PCM_PADDING_OFF) {
			return -EINVAL;
		}

		/* sob : Data bit length (PMx_SO) */
		if ((0x1f < sett->cyc.sob) || (sett->cyc.sob < 0x07)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.sob << 16;

		/* tx_pd : Data padding (TX) */
		if (sett->cyc.tx_pd != PCM_PADDING_OFF) {
			return -EINVAL;
		}

		/* cyc_val2 : Frame length */
		if (0x80 < sett->cyc2.cyc_val2) {
			return -EINVAL;
		}
		cycle2_val |= sett->cyc2.cyc_val2;

		/* sib2 : Data bit length (PMx_SI) */
		if ((0x1f < sett->cyc2.sib2) || (sett->cyc2.sib2 < 0x07)) {
			return -EINVAL;
		}
		cycle2_val |= sett->cyc2.sib2 << 8;

		/* sob2 : Data bit length (PMx_SO) */
		if ((0x1f < sett->cyc2.sob2) || (sett->cyc2.sob2 < 0x07)) {
			return -EINVAL;
		}
		cycle2_val |= sett->cyc2.sob2 << 16;
		break;
	default:
		return -EINVAL;
	}

	if(regs != NULL) {
		/* set PMx_FUNC_SEL */
		regs->func_sel = func_val;

		/* set PMx_CYCLE1 */
		regs->cycle = cycle_val;

		if ((sett->func.mode_sel == PCM_MODE_5)
		    || (sett->func.mode_sel == PCM_MODE_6)) {
			/* set PMx_CYCLE2 */
			regs->cycle2 = cycle2_val;
		}
	}

	return 0;
}

EXPORT_SYMBOL(mp200_pcm_new);
EXPORT_SYMBOL(mp200_pcm_free);
EXPORT_SYMBOL(mp200_pcm_set_ctrl);
#ifdef CONFIG_PM
EXPORT_SYMBOL(mp200_pcm_suspend);
EXPORT_SYMBOL(mp200_pcm_resume);
#endif

MODULE_DESCRIPTION("mp200 PCM DMA module");
MODULE_LICENSE("GPL");

