/* 2009-08-24: File added and changed by Sony Corporation */
/*
 * file_storage_navu.c -- File-backed USB Storage Gadget, for USB development
 *
 * Copyright (C) 2003-2005 Alan Stern
 * Copyright (C) 2007, Ryoichi KATO, Sony Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the above-listed copyright holders may not be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*
 * The File-backed Storage Gadget acts as a USB Mass Storage device,
 * appearing to the host as a disk drive.  In addition to providing an
 * example of a genuinely useful gadget driver for a USB device, it also
 * illustrates a technique of double-buffering for increased throughput.
 * Last but not least, it gives an easy way to probe the behavior of the
 * Mass Storage drivers in a USB host.
 *
 * Backing storage is provided by a regular file or a block device, specified
 * by the "file" module parameter.  Access can be limited to read-only by
 * setting the optional "ro" module parameter.  The gadget will indicate that
 * it has removable media if the optional "removable" module parameter is set.
 *
 * The gadget supports the Control-Bulk (CB), Control-Bulk-Interrupt (CBI),
 * and Bulk-Only (also known as Bulk-Bulk-Bulk or BBB) transports, selected
 * by the optional "transport" module parameter.  It also supports the
 * following protocols: RBC (0x01), ATAPI or SFF-8020i (0x02), QIC-157 (0c03),
 * UFI (0x04), SFF-8070i (0x05), and transparent SCSI (0x06), selected by
 * the optional "protocol" module parameter.  In addition, the default
 * Vendor ID, Product ID, and release number can be overridden.
 *
 * There is support for multiple logical units (LUNs), each of which has
 * its own backing file.  The number of LUNs can be set using the optional
 * "luns" module parameter (anywhere from 1 to 8), and the corresponding
 * files are specified using comma-separated lists for "file" and "ro".
 * The default number of LUNs is taken from the number of "file" elements;
 * it is 1 if "file" is not given.  If "removable" is not set then a backing
 * file must be specified for each LUN.  If it is set, then an unspecified
 * or empty backing filename means the LUN's medium is not loaded.
 *
 * Requirements are modest; only a bulk-in and a bulk-out endpoint are
 * needed (an interrupt-out endpoint is also needed for CBI).  The memory
 * requirement amounts to two 16K buffers, size configurable by a parameter.
 * Support is included for both full-speed and high-speed operation.
 *
 * Note that the driver is slightly non-portable in that it assumes a
 * single memory/DMA buffer will be useable for bulk-in, bulk-out, and
 * interrupt-in endpoints.  With most device controllers this isn't an
 * issue, but there may be some with hardware restrictions that prevent
 * a buffer from being used by more than one endpoint.
 *
 * Module options:
 *
 *	file=filename[,filename...]
 *				Required if "removable" is not set, names of
 *					the files or block devices used for
 *					backing storage
 *	ro=b[,b...]		Default false, booleans for read-only access
 *	removable		Default false, boolean for removable media
 *	luns=N			Default N = number of filenames, number of
 *					LUNs to support
 *	stall			Default determined according to the type of
 *					USB device controller (usually true),
 *					boolean to permit the driver to halt
 *					bulk endpoints
 *	transport=XXX		Default BBB, transport name (CB, CBI, or BBB)
 *	protocol=YYY		Default SCSI, protocol name (RBC, 8020 or
 *					ATAPI, QIC, UFI, 8070, or SCSI;
 *					also 1 - 6)
 *	vendor=0xVVVV		Default 0x0525 (NetChip), USB Vendor ID
 *	product=0xPPPP		Default 0xa4a5 (FSG), USB Product ID
 *	release=0xRRRR		Override the USB release number (bcdDevice)
 *	buflen=N		Default N=16384, buffer size used (will be
 *					rounded down to a multiple of
 *					PAGE_CACHE_SIZE)
 *
 * If CONFIG_USB_FILE_STORAGE_TEST is not set, only the "file", "ro",
 * "removable", "luns", and "stall" options are available; default values
 * are used for everything else.
 *
 * The pathnames of the backing files and the ro settings are available in
 * the attribute files "file" and "ro" in the lun<n> subdirectory of the
 * gadget's sysfs directory.  If the "removable" option is set, writing to
 * these files will simulate ejecting/loading the medium (writing an empty
 * line means eject) and adjusting a write-enable tab.  Changes to the ro
 * setting are not allowed when the medium is loaded.
 *
 * This gadget driver is heavily based on "Gadget Zero" by David Brownell.
 * The driver's SCSI command interface was based on the "Information
 * technology - Small Computer System Interface - 2" document from
 * X3T9.2 Project 375D, Revision 10L, 7-SEP-93, available at
 * <http://www.t10.org/ftp/t10/drafts/s2/s2-r10l.pdf>.  The single exception
 * is opcode 0x23 (READ FORMAT CAPACITIES), which was based on the
 * "Universal Serial Bus Mass Storage Class UFI Command Specification"
 * document, Revision 1.0, December 14, 1998, available at
 * <http://www.usb.org/developers/devclass_docs/usbmass-ufi10.pdf>.
 */


/*
 *				Driver Design
 *
 * The FSG driver is fairly straightforward.  There is a main kernel
 * thread that handles most of the work.  Interrupt routines field
 * callbacks from the controller driver: bulk- and interrupt-request
 * completion notifications, endpoint-0 events, and disconnect events.
 * Completion events are passed to the main thread by wakeup calls.  Many
 * ep0 requests are handled at interrupt time, but SetInterface,
 * SetConfiguration, and device reset requests are forwarded to the
 * thread in the form of "exceptions" using SIGUSR1 signals (since they
 * should interrupt any ongoing file I/O operations).
 *
 * The thread's main routine implements the standard command/data/status
 * parts of a SCSI interaction.  It and its subroutines are full of tests
 * for pending signals/exceptions -- all this polling is necessary since
 * the kernel has no setjmp/longjmp equivalents.  (Maybe this is an
 * indication that the driver really wants to be running in userspace.)
 * An important point is that so long as the thread is alive it keeps an
 * open reference to the backing file.  This will prevent unmounting
 * the backing file's underlying filesystem and could cause problems
 * during system shutdown, for example.  To prevent such problems, the
 * thread catches INT, TERM, and KILL signals and converts them into
 * an EXIT exception.
 *
 * In normal operation the main thread is started during the gadget's
 * fsg_bind() callback and stopped during fsg_unbind().  But it can also
 * exit when it receives a signal, and there's no point leaving the
 * gadget running when the thread is dead.  So just before the thread
 * exits, it deregisters the gadget driver.  This makes things a little
 * tricky: The driver is deregistered at two places, and the exiting
 * thread can indirectly call fsg_unbind() which in turn can tell the
 * thread to exit.  The first problem is resolved through the use of the
 * REGISTERED atomic bitflag; the driver will only be deregistered once.
 * The second problem is resolved by having fsg_unbind() check
 * fsg->state; it won't try to stop the thread if the state is already
 * FSG_STATE_TERMINATED.
 *
 * To provide maximum throughput, the driver uses a circular pipeline of
 * buffer heads (struct fsg_buffhd).  In principle the pipeline can be
 * arbitrarily long; in practice the benefits don't justify having more
 * than 2 stages (i.e., double buffering).  But it helps to think of the
 * pipeline as being a long one.  Each buffer head contains a bulk-in and
 * a bulk-out request pointer (since the buffer can be used for both
 * output and input -- directions always are given from the host's
 * point of view) as well as a pointer to the buffer and various state
 * variables.
 *
 * Use of the pipeline follows a simple protocol.  There is a variable
 * (fsg->next_buffhd_to_fill) that points to the next buffer head to use.
 * At any time that buffer head may still be in use from an earlier
 * request, so each buffer head has a state variable indicating whether
 * it is EMPTY, FULL, or BUSY.  Typical use involves waiting for the
 * buffer head to be EMPTY, filling the buffer either by file I/O or by
 * USB I/O (during which the buffer head is BUSY), and marking the buffer
 * head FULL when the I/O is complete.  Then the buffer will be emptied
 * (again possibly by USB I/O, during which it is marked BUSY) and
 * finally marked EMPTY again (possibly by a completion routine).
 *
 * A module parameter tells the driver to avoid stalling the bulk
 * endpoints wherever the transport specification allows.  This is
 * necessary for some UDCs like the SuperH, which cannot reliably clear a
 * halt on a bulk endpoint.  However, under certain circumstances the
 * Bulk-only specification requires a stall.  In such cases the driver
 * will halt the endpoint and set a flag indicating that it should clear
 * the halt in software during the next device reset.  Hopefully this
 * will permit everything to work correctly.  Furthermore, although the
 * specification allows the bulk-out endpoint to halt when the host sends
 * too much data, implementing this would cause an unavoidable race.
 * The driver will always use the "no-stall" approach for OUT transfers.
 *
 * One subtle point concerns sending status-stage responses for ep0
 * requests.  Some of these requests, such as device reset, can involve
 * interrupting an ongoing file I/O operation, which might take an
 * arbitrarily long time.  During that delay the host might give up on
 * the original ep0 request and issue a new one.  When that happens the
 * driver should not notify the host about completion of the original
 * request, as the host will no longer be waiting for it.  So the driver
 * assigns to each ep0 request a unique tag, and it keeps track of the
 * tag value of the request associated with a long-running exception
 * (device-reset, interface-change, or configuration-change).  When the
 * exception handler is finished, the status-stage response is submitted
 * only if the current ep0 request tag is equal to the exception request
 * tag.  Thus only the most recently received ep0 request will get a
 * status-stage response.
 *
 * Warning: This driver source file is too long.  It ought to be split up
 * into a header file plus about 3 separate .c files, to handle the details
 * of the Gadget, USB Mass Storage, and SCSI protocols.
 */


/*====================================================================
   If you want to make a file storage driver for PND( NAVU ),
   need to define the following USB_FILE_STORAGE_NAVU labels.
   When this label is not defined,
   this driver is sama as a normal file storage driver.

   When we get a file storage driver source,
   there were already the following 2 labels on the source.
     + USB_FILE_STORAGE_NAVU
     + MACH_SONY_NAVU

   In this source, MACH_SONY_NAVU is instead of USB_FILE_STORAGE_NAVU.
  =====================================================================*/
#define USB_FILE_STORAGE_NAVU


#ifdef USB_FILE_STORAGE_NAVU
/* Release Info
 * -------------------------------------------------------------------------
 * version : 		SDEC05-018-USBDrv_1.0.00_20060417
 * change point : 	add the timeout process to ioctl_wait_event function
 * author :             Richard.Zhou@sony.com.cn
 * date :               2006.04.17
 * -------------------------------------------------------------------------
 */
/* -------------------------------------------------------------------------
 * version : 		SDEC05-018-USBDrv_1.0.00_20060429
 * change point : 	prevent kernel fsg_main_thread exit, but ioctl pending
 * author :             Richard.Zhou@sony.com.cn
 * date :               2006.04.29
 * -------------------------------------------------------------------------
 */
#endif // USB_FILE_STORAGE_NAVU

#undef DEBUG
#undef VERBOSE
#undef DUMP_MSGS


#include <asm/system.h>
#include <asm/uaccess.h>

#include <linux/bitops.h>
#include <linux/blkdev.h>
#include <linux/compiler.h>
#include <linux/completion.h>
#include <linux/dcache.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/fcntl.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/kref.h>
#include <linux/kthread.h>
#include <linux/limits.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/pagemap.h>
#include <linux/rwsem.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/string.h>
#include <linux/freezer.h>
#include <linux/utsname.h>

#include <linux/usb/ch9.h>
#include <linux/usb/gadget.h>

#include "gadget_chips.h"

#ifdef USB_FILE_STORAGE_NAVU
//Richard 2006/2/28 add here
#include "msgchdev.h"
#include <linux/buffer_head.h>
#include <asm/arch/fsg_handler.h>
#endif

#include <linux/k3d.h>  // aaded for improving to check media insertion. 20080626.

/*-------------------------------------------------------------------------*/

#ifdef USB_FILE_STORAGE_NAVU
#define DRIVER_DESC		"SONYCAR_PND"
#define DRIVER_NAME		"g_file_storage_navu"
#define MS_MAJOR		8
#define MS_MINOR		0
#else
#define DRIVER_DESC		"File-backed Storage Gadget"
#define DRIVER_NAME		"g_file_storage"
#endif

#define DRIVER_VERSION		"12 Oct 2007 "

static const char longname[] = DRIVER_DESC;
static const char shortname[] = DRIVER_NAME;

MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_AUTHOR("Alan Stern");
MODULE_LICENSE("Dual BSD/GPL");

/* Thanks to NetChip Technologies for donating this product ID.
 *
 * DO NOT REUSE THESE IDs with any other driver!!  Ever!!
 * Instead:  allocate your own, using normal USB-IF procedures. */
#ifdef USB_FILE_STORAGE_NAVU
#define DRIVER_VENDOR_ID	0x054C	// Sony
#define DRIVER_PRODUCT_ID	0x0412	// changed on 2009/06/08
#else
#define DRIVER_VENDOR_ID	0x0525	// NetChip
#define DRIVER_PRODUCT_ID	0xa4a5	// Linux-USB File-backed Storage Gadget
#endif


/*
 * This driver assumes self-powered hardware and has no way for users to
 * trigger remote wakeup.  It uses autoconfiguration to select endpoints
 * and endpoint addresses.
 */
#ifdef USB_FILE_STORAGE_NAVU
/* Debug switch */
//#define DEBUG
#endif
/*-------------------------------------------------------------------------*/

#define xprintk(f,level,fmt,args...) \
	dev_printk(level , &(f)->gadget->dev , fmt , ## args)
#define yprintk(l,level,fmt,args...) \
	dev_printk(level , &(l)->dev , fmt , ## args)

#ifdef DEBUG
#define DBG(fsg,fmt,args...) \
	xprintk(fsg , KERN_DEBUG , fmt , ## args)
#define LDBG(lun,fmt,args...) \
	yprintk(lun , KERN_DEBUG , fmt , ## args)
#define MDBG(fmt,args...) \
	printk(KERN_DEBUG DRIVER_NAME ": " fmt , ## args)
#else
#define DBG(fsg,fmt,args...) \
	do { } while (0)
#define LDBG(lun,fmt,args...) \
	do { } while (0)
#define MDBG(fmt,args...) \
	do { } while (0)
#undef VERBOSE
#undef DUMP_MSGS
#endif /* DEBUG */

#ifdef VERBOSE
#define VDBG	DBG
#define VLDBG	LDBG
#else
#define VDBG(fsg,fmt,args...) \
	do { } while (0)
#define VLDBG(lun,fmt,args...) \
	do { } while (0)
#endif /* VERBOSE */

#define ERROR(fsg,fmt,args...) \
	xprintk(fsg , KERN_ERR , fmt , ## args)
#define LERROR(lun,fmt,args...) \
	yprintk(lun , KERN_ERR , fmt , ## args)

#define WARN(fsg,fmt,args...) \
	xprintk(fsg , KERN_WARNING , fmt , ## args)
#define LWARN(lun,fmt,args...) \
	yprintk(lun , KERN_WARNING , fmt , ## args)

#define INFO(fsg,fmt,args...) \
	xprintk(fsg , KERN_INFO , fmt , ## args)
#define LINFO(lun,fmt,args...) \
	yprintk(lun , KERN_INFO , fmt , ## args)

#define MINFO(fmt,args...) \
	printk(KERN_INFO DRIVER_NAME ": " fmt , ## args)


/*-------------------------------------------------------------------------*/

#ifdef USB_FILE_STORAGE_NAVU
struct media_type_handler_info{
	int lun;
	int (*handler)(int, void *, char *);
	void *handler_id;
};

#define MAX_PARM_INFO_HANDLER_NS (32)
struct parm_info_handler_info{
	int lun;
	int page_code;
	int (*handler)(int, int, void *, char *);
	int parm_page_max_len;
	void *handler_id;
};

#define MAX_VC_HANDLER_NS (10)
struct vc_handler_info{
	int lun;
	int cmd;
	int (*handler)(int, int, void *, char *, int, char *, u32 *, u32 *);
	int send_data_max_len;
	void *handler_id;
};

#define MAX_USB_EVENT_HANDLER_NS (10)
struct usb_event_handler_info{
	void (*handler)(char, void *);
	void *handler_id;
};
#endif /* USB_FILE_STORAGE_NAVU */

/*-------------------------------------------------------------------------*/

/* Encapsulate the module parameter settings */

#define MAX_LUNS	8

static struct {
	char		*file[MAX_LUNS];
	int		ro[MAX_LUNS];
	int		num_filenames;
	int		num_ros;
	unsigned int	nluns;

	int		removable;
	int		can_stall;

	char		*transport_parm;
	char		*protocol_parm;
	unsigned short	vendor;
	unsigned short	product;
	unsigned short	release;
	unsigned int	buflen;

#ifdef USB_FILE_STORAGE_NAVU
	char		*vendor_string;
	char		*product_string;
	char		*serial_string;
#endif

	int		transport_type;
	char		*transport_name;
	int		protocol_type;
	char		*protocol_name;

} mod_data = {					// Default values
#ifdef USB_FILE_STORAGE_NAVU
	.file			= {"dev/ram",},
	.ro			= {0, },
#endif
	.transport_parm		= "BBB",
	.protocol_parm		= "SCSI",
	.removable		= 0,
#ifdef USB_FILE_STORAGE_NAVU
	.can_stall		= 0,
#else
	.can_stall		= 1,
#endif
	.vendor			= DRIVER_VENDOR_ID,
	.product		= DRIVER_PRODUCT_ID,
	.release		= 0xffff,	// Use controller chip type
	.buflen			= 16384,	// Decided to use a kernal original buffer size on 2008/05/27.
#ifdef USB_FILE_STORAGE_NAVU
	.vendor_string  = "Sony   ",
	.product_string = "nav-u",
	.serial_string  = "000000000000",
#endif
	};

#ifdef USB_FILE_STORAGE_NAVU
static int Mount_disable_flag  = 0;	// 1=Disable    0:Enable
static int Usb_connection_flag = 0;	// 1=Connection 0:No connection

module_param_named(serial_string, mod_data.serial_string, charp, S_IRUGO);
MODULE_PARM_DESC(serial_string, "Serial Number string");

module_param_named(auto_mount_off, Mount_disable_flag, bool, S_IRUGO);
MODULE_PARM_DESC(auto_mount_off, "Auto Mount Off");
#endif

module_param_array_named(file, mod_data.file, charp, &mod_data.num_filenames, S_IRUGO);
MODULE_PARM_DESC(file, "names of backing files or devices");

module_param_array_named(ro, mod_data.ro, bool, &mod_data.num_ros, S_IRUGO);
MODULE_PARM_DESC(ro, "true to force read-only");

module_param_named(luns, mod_data.nluns, uint, S_IRUGO);
MODULE_PARM_DESC(luns, "number of LUNs");

module_param_named(removable, mod_data.removable, bool, S_IRUGO);
MODULE_PARM_DESC(removable, "true to simulate removable media");

module_param_named(vendor, mod_data.vendor, ushort, S_IRUGO);
MODULE_PARM_DESC(vendor, "USB Vendor ID");

module_param_named(product, mod_data.product, ushort, S_IRUGO);
MODULE_PARM_DESC(product, "USB Product ID");

module_param_named(release, mod_data.release, ushort, S_IRUGO);
MODULE_PARM_DESC(release, "USB release number");

module_param_named(buflen, mod_data.buflen, uint, S_IRUGO);
MODULE_PARM_DESC(buflen, "I/O buffer size");

#ifdef USB_FILE_STORAGE_NAVU
module_param_named(vendor_string, mod_data.vendor_string, charp, S_IRUGO);
MODULE_PARM_DESC(vendor_string, "USB vendor ID string");

module_param_named(product_string, mod_data.product_string, charp, S_IRUGO);
MODULE_PARM_DESC(product_string, "USB product ID string");
#endif


/* In the non-TEST version, only the module parameters listed above
 * are available. */
#ifdef CONFIG_USB_FILE_STORAGE_TEST

module_param_named(transport, mod_data.transport_parm, charp, S_IRUGO);
MODULE_PARM_DESC(transport, "type of transport (BBB, CBI, or CB)");

module_param_named(protocol, mod_data.protocol_parm, charp, S_IRUGO);
MODULE_PARM_DESC(protocol, "type of protocol (RBC, 8020, QIC, UFI, "
		"8070, or SCSI)");

module_param_named(stall, mod_data.can_stall, bool, S_IRUGO);
MODULE_PARM_DESC(stall, "false to prevent bulk stalls");

#endif /* CONFIG_USB_FILE_STORAGE_TEST */


/*-------------------------------------------------------------------------*/

/* USB protocol value = the transport method */
#define USB_PR_CBI	0x00		// Control/Bulk/Interrupt
#define USB_PR_CB	0x01		// Control/Bulk w/o interrupt
#define USB_PR_BULK	0x50		// Bulk-only

/* USB subclass value = the protocol encapsulation */
#define USB_SC_RBC	0x01		// Reduced Block Commands (flash)
#define USB_SC_8020	0x02		// SFF-8020i, MMC-2, ATAPI (CD-ROM)
#define USB_SC_QIC	0x03		// QIC-157 (tape)
#define USB_SC_UFI	0x04		// UFI (floppy)
#define USB_SC_8070	0x05		// SFF-8070i (removable)
#define USB_SC_SCSI	0x06		// Transparent SCSI

/* Bulk-only data structures */

/* Command Block Wrapper */
struct bulk_cb_wrap {
#ifdef CONFIG_USB_FILE_STORAGE_TEST
	u32	Signature;		// Contains 'USBC'
#else
	__le32	Signature;		// Contains 'USBC'
#endif
	u32	Tag;			// Unique per command id
#ifdef CONFIG_USB_FILE_STORAGE_TEST
	u32	DataTransferLength;	// Size of the data
#else
	__le32	DataTransferLength;	// Size of the data
#endif
	u8	Flags;			// Direction in bit 7
	u8	Lun;			// LUN (normally 0)
	u8	Length;			// Of the CDB, <= MAX_COMMAND_SIZE
	u8	CDB[16];		// Command Data Block
};

#define USB_BULK_CB_WRAP_LEN	31
#define USB_BULK_CB_SIG		0x43425355	// Spells out USBC
#define USB_BULK_IN_FLAG	0x80

/* Command Status Wrapper */
struct bulk_cs_wrap {
#ifdef CONFIG_USB_FILE_STORAGE_TEST
	u32	Signature;		// Should = 'USBS'
#else
	__le32	Signature;		// Should = 'USBS'
#endif
	u32	Tag;			// Same as original command
#ifdef CONFIG_USB_FILE_STORAGE_TEST
	u32	Residue;		// Amount not transferred
#else
	__le32	Residue;		// Amount not transferred
#endif
	u8	Status;			// See below
};

#define USB_BULK_CS_WRAP_LEN	13
#define USB_BULK_CS_SIG		0x53425355	// Spells out 'USBS'
#define USB_STATUS_PASS		0
#define USB_STATUS_FAIL		1
#define USB_STATUS_PHASE_ERROR	2

/* Bulk-only class specific requests */
#define USB_BULK_RESET_REQUEST		0xff
#define USB_BULK_GET_MAX_LUN_REQUEST	0xfe


/* CBI Interrupt data structure */
struct interrupt_data {
	u8	bType;
	u8	bValue;
};

#define CBI_INTERRUPT_DATA_LEN		2

/* CBI Accept Device-Specific Command request */
#define USB_CBI_ADSC_REQUEST		0x00


#define MAX_COMMAND_SIZE	16	// Length of a SCSI Command Data Block

/* SCSI commands that we recognize */
#define SC_FORMAT_UNIT			0x04
#define SC_INQUIRY			0x12
#define SC_MODE_SELECT_6		0x15
#define SC_MODE_SELECT_10		0x55
#define SC_MODE_SENSE_6			0x1a
#define SC_MODE_SENSE_10		0x5a
#define SC_PREVENT_ALLOW_MEDIUM_REMOVAL	0x1e
#define SC_READ_6			0x08
#define SC_READ_10			0x28
#define SC_READ_12			0xa8
#define SC_READ_CAPACITY		0x25
#define SC_READ_FORMAT_CAPACITIES	0x23
#define SC_RELEASE			0x17
#define SC_REQUEST_SENSE		0x03
#define SC_RESERVE			0x16
#define SC_SEND_DIAGNOSTIC		0x1d
#define SC_START_STOP_UNIT		0x1b
#define SC_SYNCHRONIZE_CACHE		0x35
#define SC_TEST_UNIT_READY		0x00
#define SC_VERIFY			0x2f
#define SC_WRITE_6			0x0a
#define SC_WRITE_10			0x2a
#define SC_WRITE_12			0xaa

/* SCSI Sense Key/Additional Sense Code/ASC Qualifier values */
#define SS_NO_SENSE				0
#define SS_COMMUNICATION_FAILURE		0x040800
#define SS_INVALID_COMMAND			0x052000
#define SS_INVALID_FIELD_IN_CDB			0x052400
#define SS_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE	0x052100
#define SS_LOGICAL_UNIT_NOT_SUPPORTED		0x052500
#define SS_MEDIUM_NOT_PRESENT			0x023a00
#define SS_MEDIUM_REMOVAL_PREVENTED		0x055302
#define SS_NOT_READY_TO_READY_TRANSITION	0x062800
#define SS_RESET_OCCURRED			0x062900
#define SS_SAVING_PARAMETERS_NOT_SUPPORTED	0x053900
#define SS_UNRECOVERED_READ_ERROR		0x031100
#define SS_WRITE_ERROR				0x030c02
#define SS_WRITE_PROTECTED			0x072700

#define SK(x)		((u8) ((x) >> 16))	// Sense Key byte, etc.
#define ASC(x)		((u8) ((x) >> 8))
#define ASCQ(x)		((u8) (x))


/*-------------------------------------------------------------------------*/

/*
 * These definitions will permit the compiler to avoid generating code for
 * parts of the driver that aren't used in the non-TEST version.  Even gcc
 * can recognize when a test of a constant expression yields a dead code
 * path.
 */

#ifdef CONFIG_USB_FILE_STORAGE_TEST

#define transport_is_bbb()	(mod_data.transport_type == USB_PR_BULK)
#define transport_is_cbi()	(mod_data.transport_type == USB_PR_CBI)
#define protocol_is_scsi()	(mod_data.protocol_type == USB_SC_SCSI)

#else

#define transport_is_bbb()	1
#define transport_is_cbi()	0
#define protocol_is_scsi()	1

#endif /* CONFIG_USB_FILE_STORAGE_TEST */


struct lun {
	struct file	*filp;
	loff_t		file_length;
	loff_t		num_sectors;

	unsigned int	ro : 1;
	unsigned int	prevent_medium_removal : 1;
	unsigned int	registered : 1;
	unsigned int	info_valid : 1;

	u32		sense_data;
	u32		sense_data_info;
	u32		unit_attention_data;
#ifdef USB_FILE_STORAGE_NAVU
	bool	is_fake_sense_data;
#endif /* USB_FILE_STORAGE_NAVU */

	struct device	dev;
};

#define backing_file_is_open(curlun)	((curlun)->filp != NULL)

#ifdef USB_FILE_STORAGE_NAVU
static int open_backing_file(struct lun *curlun, const char *filename);
#endif
static inline struct lun *dev_to_lun(struct device *dev)
{
	return container_of(dev, struct lun, dev);
}


/* Big enough to hold our biggest descriptor */
#define EP0_BUFSIZE	256
#define DELAYED_STATUS	(EP0_BUFSIZE + 999)	// An impossibly large value

/* Number of buffers we will use.  2 is enough for double-buffering */
#define NUM_BUFFERS	2

enum fsg_buffer_state {
	BUF_STATE_EMPTY = 0,
	BUF_STATE_FULL,
	BUF_STATE_BUSY
};

struct fsg_buffhd {
	void				*buf;
	enum fsg_buffer_state		state;
	struct fsg_buffhd		*next;

	/* The NetChip 2280 is faster, and handles some protocol faults
	 * better, if we don't submit any short bulk-out read requests.
	 * So we will record the intended request length here. */
	unsigned int			bulk_out_intended_length;

	struct usb_request		*inreq;
	int				inreq_busy;
	struct usb_request		*outreq;
	int				outreq_busy;
};

enum fsg_state {
	FSG_STATE_COMMAND_PHASE = -10,		// This one isn't used anywhere
	FSG_STATE_DATA_PHASE,
	FSG_STATE_STATUS_PHASE,

	FSG_STATE_IDLE = 0,
	FSG_STATE_ABORT_BULK_OUT,
	FSG_STATE_RESET,
	FSG_STATE_INTERFACE_CHANGE,
	FSG_STATE_CONFIG_CHANGE,
	FSG_STATE_DISCONNECT,
	FSG_STATE_EXIT,
	FSG_STATE_TERMINATED
};

enum data_direction {
	DATA_DIR_UNKNOWN = 0,
	DATA_DIR_FROM_HOST,
	DATA_DIR_TO_HOST,
	DATA_DIR_NONE
};

struct fsg_dev {
	/* lock protects: state, all the req_busy's, and cbbuf_cmnd */
	spinlock_t		lock;
	struct usb_gadget	*gadget;

	/* filesem protects: backing files in use */
	struct rw_semaphore	filesem;

	/* reference counting: wait until all LUNs are released */
	struct kref		ref;

	struct usb_ep		*ep0;		// Handy copy of gadget->ep0
	struct usb_request	*ep0req;	// For control responses
	unsigned int		ep0_req_tag;
	const char		*ep0req_name;

	struct usb_request	*intreq;	// For interrupt responses
	int			intreq_busy;
	struct fsg_buffhd	*intr_buffhd;

 	unsigned int		bulk_out_maxpacket;
	enum fsg_state		state;		// For exception handling
	unsigned int		exception_req_tag;

	u8			config, new_config;

	unsigned int		running : 1;
	unsigned int		bulk_in_enabled : 1;
	unsigned int		bulk_out_enabled : 1;
	unsigned int		intr_in_enabled : 1;
	unsigned int		phase_error : 1;
	unsigned int		short_packet_received : 1;
	unsigned int		bad_lun_okay : 1;

	unsigned long		atomic_bitflags;
#define REGISTERED		0
#define CLEAR_BULK_HALTS	1
#define SUSPENDED		2
#ifdef USB_FILE_STORAGE_NAVU
#define RESETED			3
#endif

	struct usb_ep		*bulk_in;
	struct usb_ep		*bulk_out;
	struct usb_ep		*intr_in;

	struct fsg_buffhd	*next_buffhd_to_fill;
	struct fsg_buffhd	*next_buffhd_to_drain;
	struct fsg_buffhd	buffhds[NUM_BUFFERS];

	int			thread_wakeup_needed;
	struct completion	thread_notifier;
	struct task_struct	*thread_task;
	int			thread_pid;

	int			cmnd_size;
	u8			cmnd[MAX_COMMAND_SIZE];
	enum data_direction	data_dir;
	u32			data_size;
	u32			data_size_from_cmnd;
	u32			tag;
	unsigned int		lun;
	u32			residue;
	u32			usb_amount_left;

	/* The CB protocol offers no way for a host to know when a command
	 * has completed.  As a result the next command may arrive early,
	 * and we will still have to handle it.  For that reason we need
	 * a buffer to store new commands when using CB (or CBI, which
	 * does not oblige a host to wait for command completion either). */
	int			cbbuf_cmnd_size;
	u8			cbbuf_cmnd[MAX_COMMAND_SIZE];

	unsigned int		nluns;
	struct lun		*luns;
	struct lun		*curlun;

#ifdef USB_FILE_STORAGE_NAVU
	unsigned int	last_lun;
#endif /* USB_FILE_STORAGE_NAVU */

	//struct semaphore
#ifdef USB_FILE_STORAGE_NAVU
	struct media_type_handler_info *media_type_handler_info;
	struct parm_info_handler_info *parm_info_handler_table[MAX_PARM_INFO_HANDLER_NS];
	struct vc_handler_info *vc_handler_table[MAX_VC_HANDLER_NS];
	struct usb_event_handler_info *usb_event_handler_table[MAX_USB_EVENT_HANDLER_NS];
#endif /* USB_FILE_STORAGE_NAVU */
};

typedef void (*fsg_routine_t)(struct fsg_dev *);

static int inline exception_in_progress(struct fsg_dev *fsg)
{
	return (fsg->state > FSG_STATE_IDLE);
}

/* Make bulk-out requests be divisible by the maxpacket size */
static void inline set_bulk_out_req_length(struct fsg_dev *fsg,
		struct fsg_buffhd *bh, unsigned int length)
{
	unsigned int	rem;

	bh->bulk_out_intended_length = length;
	rem = length % fsg->bulk_out_maxpacket;
	if (rem > 0)
		length += fsg->bulk_out_maxpacket - rem;
	bh->outreq->length = length;
}

static struct fsg_dev			*the_fsg;
static struct usb_gadget_driver		fsg_driver;

static void	close_backing_file(struct lun *curlun);
static void	close_all_backing_files(struct fsg_dev *fsg);

#ifdef USB_FILE_STORAGE_NAVU
#define SCSI_TEST_APP         0

//********************Richard Add
// #ifdef MACH_SONY_NAVU is old label.

#define SPECIAL_MALLOC

#ifdef  SPECIAL_MALLOC
unsigned long alloc_size = 0;
void * malloc_address[100] = {NULL};
unsigned long malloc_size[100] = {0};
void* special_malloc(size_t size, unsigned int __nocast flags) {
  void* p = NULL;
  int i = 0;
  for(i = 0; i < 100; i++) {
    if(NULL == malloc_address[i]) {
      p = kmalloc(size, flags);
      malloc_size[i] = size;
      malloc_address[i] = p;
      break;
    }
  }
  alloc_size += size;
  return p;
}
void special_free(const void* p) {
  int i = 0;
  for(i = 0; i < 100; i++) {
    if(malloc_address[i] == p) {
      alloc_size -= malloc_size[i];
      kfree(p);
      malloc_address[i] = NULL;
      break;
    }
  }
}
#define KMalloc(size, option)  special_malloc(size, option)
#define KFree(p)               special_free(p)
#define MEM_SIZE_INFO  \
do { \
  printk(KERN_INFO "file_storage_navu.c: alloc_size is %d \n", (int)alloc_size); \
}while(0);
#define INIT_MEM_SIZE \
do { \
  memset(&malloc_size[0], 0x00, sizeof(malloc_size));          \
  memset(&malloc_address[0], 0x00, sizeof(malloc_address));    \
  alloc_size = 0;                  \
}while(0);


#else

#define INIT_MEM_SIZE           0
#define MEM_SIZE_INFO           0
#define KMalloc(size, option)  kmalloc(size, option)
#define KFree(p)               kfree(p)

#endif /* SPECIAL_MALLOC */


//#define SONY_DEBUG

#ifdef SONY_DEBUG
#define sonydbg(fsg,fmt,args...) \
	xprintk(fsg , KERN_ALERT , fmt , ## args)
#define sonydef(x)         x

#else
#define sonydbg(fsg,fmt,args...)
#define sonydef(x)
#endif /* SONY_DEBUG */


//#define NEW_DEBUG

#ifdef NEW_DEBUG
#define newdbg(fsg,fmt,args...) \
	xprintk(fsg , KERN_ALERT , fmt , ## args)
#else
#define newdbg(fsg,fmt,args...)
#endif /* NEW_DEBUG */

//Declaration
//static int do_set_config(struct fsg_dev *fsg, u8 new_config);
//static void fsync_all(struct fsg_dev *fsg);
//static void handle_exception(struct fsg_dev *fsg);
static void start_transfer(struct fsg_dev *fsg, struct usb_ep *ep,
			   struct usb_request *req, int *pbusy,
			   enum fsg_buffer_state *state);
static void wakeup_thread(struct fsg_dev *fsg);

static int check_vendor_command(struct fsg_dev *fsg, int cmnd_size,
		enum data_direction data_dir,
		int needs_medium, const char *name);
static enum data_direction parse_command_dir(struct fsg_dev* fsg);

static int do_send_data(struct fsg_dev* fsg);
static int do_receive_data(struct fsg_dev *fsg);
static int do_no_data(struct fsg_dev *fsg);

#ifdef USB_FILE_STORAGE_NAVU
static void fake_sense_data(struct fsg_dev *fsg);
#endif /* USB_FILE_STORAGE_NAVU */

static int sleep_thread(struct fsg_dev *fsg);
static void wakeup_fsg(struct fsg_dev *fsg);
static int lock_fsg(struct fsg_dev *fsg);
//define wait queue to block the ioctl operation if more data transfer
#define SUCCESS            0
static DECLARE_WAIT_QUEUE_HEAD(WaitDataQue);
static DECLARE_WAIT_QUEUE_HEAD(SendDataQue);

//Added Steven 0704 start
static DECLARE_WAIT_QUEUE_HEAD(ResponseQue);
static atomic_t data_response_control = ATOMIC_INIT(0);    /*Wait queue control var */
//Added Steven 0704 end

//static DECLARE_WAIT_QUEUE_HEAD(IoctlQue);             /*IOCTL lock */


static atomic_t data_get_control = ATOMIC_INIT(0);    /*Wait queue control var */
/* 0 -> wait , 1->wake up */
static atomic_t data_send_control = ATOMIC_INIT(0);

static atomic_t atm_left_to_write = ATOMIC_INIT(0);
static atomic_t atm_left_to_read  = ATOMIC_INIT(0);

//static atomic_t ioctl_lock = ATOMIC_INIT(0);

static atomic_t req_fsg_wakeup = ATOMIC_INIT(0);  // modified on 2008.03.20

static int device_open = 0;

static int g_time_out = 0;

//ljzhou 2006/4/29 modify to prevent fsg_main_thread exit, but ioctl pending
static int g_exit_flag = 0;
//static int g_ioctl_open = 0;
// ljzhou 2006/6/20 start
//static int g_vendor_command = 0;
// ljzhou 2006/07/11 start
static int g_usb_config_cplt = 0;
// ljzhou 2006/07/11 end
//static int g_vendor_short = 0;
static int g_first_wait = 1;
#if 0
static struct semaphore g_vendor_sem;
#endif
// ljzhou 2006/6/20 end

static int transfer_ioctl(struct inode* pnode, struct file* pfile,
			  unsigned int cmd, unsigned long arg);
static int transfer_ioctl_ms(struct inode* pnode, struct file* pfile,
			  unsigned int cmd, unsigned long arg);
static int transfer_open(struct inode *inode, struct file *file);
static int transfer_release(struct inode *inode, struct file *file);

struct file_operations transfer_ops = {
	.ioctl = transfer_ioctl,
	.open  = transfer_open,
	.release = transfer_release
};

#ifdef CONFIG_MACH_PGPF
extern void pgpf_usb_notify_supplied_power(bool flag, int power);
#endif /* CONFIG_MACH_PGPF */

/* Tempory code to test if the cable is plug in or out */
#define TEMP_SUS_RES
#ifdef  TEMP_SUS_RES
/* Tempory code */
static unsigned long g_suspend = 0;
static unsigned long g_resume = 0;
/* end  Tempory */
static void clear_status(void)
{
	g_suspend = 0;
	g_resume = 0;
}
static void set_resume(void)
{
	g_resume = 1;
}
static void set_suspend(void)
{
	g_suspend = 1;
}
static int get_status(void)
{
  return (int)(g_suspend && g_resume);
}
#define CLEAR_STATUS   do { clear_status(); }while(0);
#define SET_RESUME     do { set_resume(); }  while(0);
#define SET_SUSPEND    do { set_suspend(); } while(0);
#define GET_STATUS     get_status()
/* Tempory code to test if the cable is plug in or out */
#else   /* NO DEFINE TEMP_SUS_RES */
#define CLEAR_STATUS
#define SET_RESUME
#define SET_SUSPEND
#define GET_STATUS     0

#endif /* TEMP_SUS_RES */


#ifdef USB_FILE_STORAGE_NAVU

int fsg_register_media_type_handler( int lun,
									 int (*handler)(int, void *, char * ),
									 void *handler_id )
{
	struct fsg_dev *fsg = the_fsg;
	int rc = -1;
	struct media_type_handler_info *buf;

	if ( ( handler == NULL )
		|| ( handler_id == NULL ) ) {
		return rc;
	}

	if ( fsg->media_type_handler_info != NULL ) {
		printk("Could not register cmd handler. a handler already register.\n");
		return rc;
	}

	buf = KMalloc( sizeof( struct media_type_handler_info ), GFP_KERNEL );

	if ( buf == NULL ) {
		printk("Could not register a handler. Because a memory allocation was failure.\n");
	} else {
		buf->lun = lun;
		buf->handler = handler;
		buf->handler_id = handler_id;

		fsg->media_type_handler_info = (struct media_type_handler_info *)buf;

		rc = 0;
	}

	return rc;
}

void fsg_unregister_media_type_handler( int lun, void *handler_id )
{
	struct fsg_dev *fsg = the_fsg;

	if ( fsg->media_type_handler_info != NULL ) {
		if ( ( fsg->media_type_handler_info->lun == lun )
			&& ( fsg->media_type_handler_info->handler_id == handler_id ) ) {
			KFree( fsg->media_type_handler_info );
			fsg->media_type_handler_info = NULL;
		}
	}
}

static bool check_media_type_handler(struct fsg_dev *fsg)
{
	bool rc = false;

	if ( ( fsg->media_type_handler_info != NULL )
		&& ( fsg->media_type_handler_info->lun == fsg->lun ) ) {
		rc = true;
	}

	return rc;
}

static bool get_media_type(struct fsg_dev *fsg, char *media_type)
{
	int lun;
	void *handler_id;
	int ret;
	bool rc = false;

	lun = fsg->media_type_handler_info->lun;
	handler_id = fsg->media_type_handler_info->handler_id;

	ret = fsg->media_type_handler_info->handler( lun, handler_id, media_type );
	if ( ret == 0 ) {
		rc = true;
	}

	return rc;
}


static bool is_vendor_page_code( int page_code )
{
	bool ret = false;

	if ( ( page_code == 0x00 )
		|| ( ( 0x20 <= page_code ) && ( page_code <= 0x3e ) ) ) {
		ret = true;
	}
	return ret;
}

int fsg_register_parm_info_handler( int lun,
									int page_code,
									int (*handler)(int, int, void *, char * ),
									int parm_page_max_length,
									void *handler_id )
{
	struct fsg_dev *fsg = the_fsg;
	int i;
	int empty_index = -1;
	int rc = -1;
	struct parm_info_handler_info *buf;

	if ( ( handler == NULL )
		|| ( handler_id == NULL ) ) {
		return rc;
	}

	if ( parm_page_max_length <= 0 ) {
		return rc;
	}

	if ( !is_vendor_page_code( page_code ) ){
		return rc;
	}

	for ( i = 0 ; i < MAX_PARM_INFO_HANDLER_NS ; i++ ) {
		if ( fsg->parm_info_handler_table[i] != NULL ) {
			if ( ( fsg->parm_info_handler_table[i]->lun == lun )
				&& ( fsg->parm_info_handler_table[i]->page_code == page_code ) ) {
				printk("Could not register cmd handler. A handler already has been registered to requested cmd.\n");
				return rc;
			}
		}
		if ( ( fsg->parm_info_handler_table[i] == NULL ) && ( empty_index < 0 ) ) {
			empty_index = i;
		}
	}

	if ( empty_index < 0) {
		printk("Could not register cmd handler. No empty for a handler registration\n");
		return rc;
	}

	buf = KMalloc( sizeof( struct parm_info_handler_info ), GFP_KERNEL );

	if ( buf == NULL ) {
		printk("Could not register a cmd handler. Because a memory allocation was failure.\n");
	} else {
		buf->lun = lun;
		buf->page_code = page_code;
		buf->handler = handler;
		buf->parm_page_max_len = parm_page_max_length;
		buf->handler_id = handler_id;

		fsg->parm_info_handler_table[empty_index] = (struct parm_info_handler_info *)buf;

		rc = 0;
	}

	return rc;
}

void fsg_unregister_parm_info_handler( int lun, int page_code, void *handler_id )
{
	struct fsg_dev *fsg = the_fsg;
	int i;

	for ( i = 0 ; i < MAX_PARM_INFO_HANDLER_NS ; i++ ) {
		if ( fsg->parm_info_handler_table[i] != NULL ) {
			if ( ( fsg->parm_info_handler_table[i]->lun == lun )
				&& ( fsg->parm_info_handler_table[i]->page_code == page_code )
				&& ( fsg->parm_info_handler_table[i]->handler_id == handler_id ) ) {
				break;
			}
		}
	}

	if ( MAX_PARM_INFO_HANDLER_NS <= i) {
		printk("Could not unregist cmd handler. Because requested cmd and handler_id were not match.\n");
	} else {
		KFree( fsg->parm_info_handler_table[i] );
		fsg->parm_info_handler_table[i] = NULL;
	}
}

static int check_parm_info_handler(struct fsg_dev *fsg, int vendor_page_code)
{
	int i;
	int rc = -1;

	for ( i = 0 ; i < MAX_PARM_INFO_HANDLER_NS ; i++ ) {
		if( fsg->parm_info_handler_table[i] != NULL ) {
			if ( ( fsg->parm_info_handler_table[i]->lun == fsg->lun )
				&& ( fsg->parm_info_handler_table[i]->page_code == vendor_page_code ) ) {
				rc = i;
				break;
			}
		}
	}

	return rc;
}

static int get_parm_info_data(struct fsg_dev *fsg, int index, u8 *page_data)
{
	int lun;
	int page_code;
	void *handler_id;
	int page_data_len = -1;
	int page_data_max_len;
	char *page_data_buf;

	page_data_max_len = fsg->parm_info_handler_table[index]->parm_page_max_len;

	if ( 0 < page_data_max_len ) {
		page_data_buf = KMalloc( sizeof( char ) * page_data_max_len, GFP_KERNEL );
		if ( page_data_buf == NULL ) {
			return page_data_len;
		}
	} else {
		return page_data_len;
	}

	lun = fsg->parm_info_handler_table[index]->lun;
	page_code = fsg->parm_info_handler_table[index]->page_code;
	handler_id = fsg->parm_info_handler_table[index]->handler_id;

	page_data_len
	  = fsg->parm_info_handler_table[index]->handler( lun,
													  page_code,
													  handler_id,
													  page_data_buf );

	if ( 0 < page_data_len ) {
		memcpy(page_data, page_data_buf, page_data_len);
	}
	KFree( page_data_buf );

	return page_data_len;
}


int fsg_register_vc_handler( int lun,
							 int cmd,
							 int (*handler)(int, int, void *, char *, int, char *, u32 *, u32 * ),
							 int send_data_max_length,
							 void *handler_id )
{
	struct fsg_dev *fsg = the_fsg;
	int i;
	int empty_index = -1;
	int rc = -1;
	struct vc_handler_info *buf;

	if ( ( handler == NULL )
		|| ( handler_id == NULL ) ) {
		return rc;
	}

	for ( i = 0 ; i < MAX_VC_HANDLER_NS ; i++ ) {
		if ( fsg->vc_handler_table[i] != NULL ) {
			if ( ( fsg->vc_handler_table[i]->lun == lun )
				&& ( fsg->vc_handler_table[i]->cmd == cmd ) ) {
				printk("Could not register vc handler. A handler already has been registered to requested cmd.\n");
				return rc;
			}
		}
		if ( ( fsg->vc_handler_table[i] == NULL ) && ( empty_index < 0 ) ) {
			empty_index = i;
		}
	}

	if ( empty_index < 0) {
		printk("Could not register vc handler. No empty for a handler registration\n");
		return rc;
	}

	buf = KMalloc( sizeof( struct vc_handler_info ), GFP_KERNEL );

	if ( buf == NULL ) {
		printk("Could not register a vc handler. Because a memory allocation was failure.\n");
	} else {
		buf->lun = lun;
		buf->cmd = cmd;
		buf->handler = handler;
		buf->send_data_max_len = send_data_max_length;
		buf->handler_id = handler_id;

		fsg->vc_handler_table[empty_index] = (struct vc_handler_info *)buf;

		rc = 0;
	}

	return rc;
}

void fsg_unregister_vc_handler( int lun, int cmd, void *handler_id )
{
	struct fsg_dev *fsg = the_fsg;
	int i;

	for ( i = 0 ; i < MAX_VC_HANDLER_NS ; i++ ) {
		if ( fsg->vc_handler_table[i] != NULL ) {
			if ( ( fsg->vc_handler_table[i]->lun == lun )
				&& ( fsg->vc_handler_table[i]->cmd == cmd )
				&& ( fsg->vc_handler_table[i]->handler_id == handler_id ) ) {
				break;
			}
		}
	}

	if ( MAX_VC_HANDLER_NS <= i) {
		printk("Could not unregist vc handler. Because requested cmd and handler_id were not match.\n");
	} else {
		KFree( fsg->vc_handler_table[i] );
		fsg->vc_handler_table[i] = NULL;
	}
}

static int check_vc_handler(struct fsg_dev *fsg)
{
	int i;
	int rc = -1;

	for ( i = 0 ; i < MAX_VC_HANDLER_NS ; i++ ) {
		if( fsg->vc_handler_table[i] != NULL ) {
			if ( ( fsg->vc_handler_table[i]->lun == fsg->lun )
				&& ( fsg->vc_handler_table[i]->cmd == fsg->cmnd[0] ) ) {
				rc = i;
				break;
			}
		}
	}

	return rc;
}

static int do_recieve_vc_data(struct fsg_dev *fsg, char *buf)
{
	struct lun		*curlun = fsg->curlun;
	struct fsg_buffhd	*bh;
	int			get_some_more;
	u32			amount_left_to_req, amount_left_to_write;
	unsigned int		amount;
	int			nwritten;
	int			rc;

	if (curlun->ro) {
		curlun->sense_data = SS_WRITE_PROTECTED;
		return -EINVAL;
	}

	/* Carry out the file writes */
	get_some_more = 1;
	amount_left_to_req = amount_left_to_write = amount =fsg->data_size;

	while (amount_left_to_write > 0) {

		/* Queue a request for more data from the host */
		bh = fsg->next_buffhd_to_fill;
		if (bh->state == BUF_STATE_EMPTY && get_some_more) {

			/* Figure out how much we want to get:
			 * Try to get the remaining amount.
			 * But don't get more than the buffer size.
			 * And don't try to go past the end of the file. */
			amount = min(amount_left_to_req, mod_data.buflen);

			if (amount == 0) {
				get_some_more = 0;
				continue;
			}

			/* Get the next buffer */
			fsg->usb_amount_left -= amount;
			amount_left_to_req -= amount;
			if (amount_left_to_req == 0)
				get_some_more = 0;

			/* amount is always divisible by 512, hence by
			 * the bulk-out maxpacket size */
			bh->outreq->length = bh->bulk_out_intended_length =
					amount;
			bh->outreq->short_not_ok = 1;
			start_transfer(fsg, fsg->bulk_out, bh->outreq,
					&bh->outreq_busy, &bh->state);
			fsg->next_buffhd_to_fill = bh->next;
			continue;
		}

		/* Write the received data to the backing file */
		bh = fsg->next_buffhd_to_drain;
		if (bh->state == BUF_STATE_EMPTY && !get_some_more)
			break;			// We stopped early
		if (bh->state == BUF_STATE_FULL) {
			smp_rmb();
			fsg->next_buffhd_to_drain = bh->next;
			bh->state = BUF_STATE_EMPTY;

			/* Did something go wrong with the transfer? */
			if (bh->outreq->status != 0) {
				//recover its origin status
				break;
			}

			if ( bh->outreq->actual <= amount ) {
				amount = bh->outreq->actual;
			}

			/* Perform the memcpy */
			memcpy( buf, bh->buf, amount );
			buf += amount;
			nwritten = amount;
			if (signal_pending(current))
				return -EINTR;		// Interrupted!

			amount_left_to_write -= nwritten;
			fsg->residue -= nwritten;

			/* Did the host decide to stop early? */
			if (bh->outreq->actual != bh->outreq->length) {
				fsg->short_packet_received = 1;
				break;
			}
			continue;
		}
		/* Wait for something to happen */
		if ((rc = sleep_thread(fsg)) != 0)
			return rc;
	}

	//return ( fsg->data_size - fsg->residue );		// No default reply
	return -EIO;		// No default reply
}

static int do_send_vc_data(struct fsg_dev *fsg, char *buf)
{
	struct lun		*curlun = fsg->curlun;
	struct fsg_buffhd	*bh;
	int			rc;
	u32			amount_left;
	unsigned int		amount, nread;

	/* Carry out the file reads */
	amount_left = fsg->data_size_from_cmnd;
	if (unlikely(amount_left == 0)) {
		return -EIO;
	}

	for (;;) {

		/* Figure out how much we need to read:
		 * Try to read the remaining amount.
		 * But don't read more than the buffer size.
		 * And don't try to read past the end of the file.
		 * Finally, if we're not at a page boundary, don't read past
		 *	the next page.
		 * If this means reading 0 then we were asked to read past
		 *	the end of file. */
		amount = min((unsigned int) amount_left, mod_data.buflen);

		/* Wait for the next buffer to become available */
		bh = fsg->next_buffhd_to_fill;
		while (bh->state != BUF_STATE_EMPTY) {
			if ((rc = sleep_thread(fsg)) != 0) {
				return rc;
			}
		}

		/* Perform the read */
		memcpy( bh->buf, buf, amount );

		buf += amount;
		nread = amount;
		if (signal_pending(current)) {
			return -EINTR;
		}

		amount_left  -= nread;
		fsg->residue -= nread;
		bh->inreq->length = nread;
		bh->state = BUF_STATE_FULL;

		/* If an error occurred, report it and its position */
		if (nread < amount) {
			curlun->sense_data = SS_UNRECOVERED_READ_ERROR;
			curlun->info_valid = 1;
			close_backing_file(curlun);
			return -EIO;
		}

		if (amount_left == 0)
			break;		// No more left to read

		/* Send this buffer and go read some more */
		bh->inreq->zero = 0;
		start_transfer(fsg, fsg->bulk_in, bh->inreq,
				&bh->inreq_busy, &bh->state);
		fsg->next_buffhd_to_fill = bh->next;
	}

	return -EIO;		// No default reply
}

static int do_send_data_vc_handler(struct fsg_dev *fsg, int index, struct fsg_buffhd *bh)
{
	int send_buf_len, receive_buf_len;
	char *send_buf, *receive_buf;
	static u32 sense_data;
	static u32 sense_info;

	send_buf_len = -1;
	send_buf = receive_buf = NULL;

	receive_buf_len = fsg->cmnd_size;

	send_buf_len = fsg->vc_handler_table[index]->send_data_max_len;

	if ( 0 < receive_buf_len ) {
		receive_buf = KMalloc( sizeof( char ) * receive_buf_len, GFP_KERNEL );
		if ( receive_buf == NULL ) {
			INFO(fsg, "cmd = %x : receive_buf allocation was failure.\n", fsg->cmnd[0]);
			return -EINVAL;
		}
	} else {
		INFO(fsg, "No command data.\n");
		return -EINVAL;
	}

	if ( 0 < send_buf_len ) {
		send_buf = KMalloc( sizeof( char ) * send_buf_len, GFP_KERNEL );
		if ( send_buf == NULL ) {
			INFO(fsg, "cmd = %x : send_buf allocation was failure.\n", fsg->cmnd[0]);
			if( receive_buf != NULL ) {
				KFree( receive_buf );
			}
			return -EINVAL;
		}
	}

	memcpy(receive_buf, fsg->cmnd, fsg->cmnd_size);

	send_buf_len
		= (*fsg->vc_handler_table[index]->handler)( fsg->lun,
												 fsg->cmnd[0],
												 fsg->vc_handler_table[index]->handler_id,
												 receive_buf,
												 receive_buf_len,
												 send_buf,
												 &sense_data,
												 &sense_info );

	if( ( send_buf != NULL ) && ( send_buf_len != 0 ) ) {
		memcpy(bh->buf, send_buf, send_buf_len);
		send_buf_len = do_send_vc_data( fsg, send_buf );
	}

	fsg->curlun->sense_data = sense_data;
	fsg->curlun->sense_data_info = sense_info;

	KFree( receive_buf );
	if( send_buf != NULL ) {
		KFree( send_buf );
	}

	return send_buf_len;
}

static int do_receive_data_vc_handler(struct fsg_dev *fsg, int index, struct fsg_buffhd *bh)
{
	int send_buf_len, receive_buf_len, recieve_data_len;
	char *send_buf, *receive_buf, *buf;
	static u32 sense_data;
	static u32 sense_info;

	send_buf_len = -1;
	send_buf = receive_buf = NULL;

	receive_buf_len = fsg->cmnd_size;
	if ( fsg->data_size != 0 ) {
		receive_buf_len += fsg->data_size;
	}

	send_buf_len = fsg->vc_handler_table[index]->send_data_max_len;

	if ( 0 < receive_buf_len ) {
		receive_buf = KMalloc( sizeof( char ) * receive_buf_len, GFP_KERNEL );
		if ( receive_buf == NULL ) {
			INFO(fsg, "cmd = %x : receive_buf allocation was failure.\n", fsg->cmnd[0]);
			return -EINVAL;
		}
	} else {
		INFO(fsg, "No command data.\n");
		return -EINVAL;
	}

	if ( 0 < send_buf_len ) {
		send_buf = KMalloc( sizeof( char ) * send_buf_len, GFP_KERNEL );
		if ( send_buf == NULL ) {
			INFO(fsg, "cmd = %x : send_buf allocation was failure.\n", fsg->cmnd[0]);
			if( receive_buf != NULL ) {
				KFree( receive_buf );
			}
			return -EINVAL;
		}
	}

	memcpy(receive_buf, fsg->cmnd, fsg->cmnd_size);

	if ( fsg->data_size != 0 ) {
		buf = receive_buf + fsg->cmnd_size;
		recieve_data_len = do_recieve_vc_data( fsg, buf );
	}

	send_buf_len
		= (*fsg->vc_handler_table[index]->handler)( fsg->lun,
												 fsg->cmnd[0],
												 fsg->vc_handler_table[index]->handler_id,
												 receive_buf,
												 receive_buf_len,
												 send_buf,
												 &sense_data,
												 &sense_info );

	fsg->curlun->sense_data = sense_data;
	fsg->curlun->sense_data_info = sense_info;

	KFree( receive_buf );
	if( send_buf != NULL ) {
		KFree( send_buf );
	}

	return send_buf_len;
}

static int do_no_data_vc_handler(struct fsg_dev *fsg, int index, struct fsg_buffhd *bh)
{
	int send_buf_len, receive_buf_len;
	char *send_buf, *receive_buf;
	static u32 sense_data;
	static u32 sense_info;

	send_buf_len = -1;
	send_buf = receive_buf = NULL;

	receive_buf_len = fsg->cmnd_size;

	send_buf_len = fsg->vc_handler_table[index]->send_data_max_len;

	if ( 0 < receive_buf_len ) {
		receive_buf = KMalloc( sizeof( char ) * receive_buf_len, GFP_KERNEL );
		if ( receive_buf == NULL ) {
			INFO(fsg, "cmd = %x : receive_buf allocation was failure.\n", fsg->cmnd[0]);
			return -EINVAL;
		}
	} else {
		INFO(fsg, "No command data.\n");
		return -EINVAL;
	}

	if ( 0 < send_buf_len ) {
		send_buf = KMalloc( sizeof( char ) * send_buf_len, GFP_KERNEL );
		if ( send_buf == NULL ) {
			INFO(fsg, "cmd = %x : send_buf allocation was failure.\n", fsg->cmnd[0]);
			if( receive_buf != NULL ) {
				KFree( receive_buf );
			}
			return -EINVAL;
		}
	}

	memcpy(receive_buf, fsg->cmnd, fsg->cmnd_size);

	send_buf_len
		= (*fsg->vc_handler_table[index]->handler)( fsg->lun,
												 fsg->cmnd[0],
												 fsg->vc_handler_table[index]->handler_id,
												 receive_buf,
												 receive_buf_len,
												 send_buf,
												 &sense_data,
												 &sense_info );

	fsg->curlun->sense_data = sense_data;
	fsg->curlun->sense_data_info = sense_info;

	KFree( receive_buf );
	if( send_buf != NULL ) {
		KFree( send_buf );
	}

	return send_buf_len;
}

int fsg_register_usb_event_handler( void (*handler)(char, void *),
									void *handler_id )
{
	struct fsg_dev *fsg = the_fsg;
	int i;
	int empty_index = -1;
	int rc = -1;
	struct usb_event_handler_info *buf;

	if ( ( handler == NULL )
		|| ( handler_id == NULL ) ) {
		return rc;
	}

	for ( i = 0 ; i < MAX_USB_EVENT_HANDLER_NS ; i++ ) {
		if ( fsg->usb_event_handler_table[i] == NULL  ) {
			empty_index = i;
			break;
		}
	}

	if ( empty_index < 0) {
		printk("Could not register a usb event handler. No empty for a handler registration\n");
		return rc;
	}

	buf = KMalloc( sizeof( struct usb_event_handler_info ), GFP_KERNEL );

	if ( buf == NULL ) {
		printk("Could not register a usb event handler. Because a memory allocation was failure.\n");
	} else {
		buf->handler = handler;
		buf->handler_id = handler_id;

		fsg->usb_event_handler_table[empty_index] = (struct usb_event_handler_info *)buf;

		rc = 0;
	}

	return rc;
}

void fsg_unregister_usb_event_handler( void *handler_id )
{
	struct fsg_dev *fsg = the_fsg;
	int i;

	for ( i = 0 ; i < MAX_USB_EVENT_HANDLER_NS ; i++ ) {
		if ( fsg->usb_event_handler_table[i] != NULL ) {
			if ( fsg->usb_event_handler_table[i]->handler_id == handler_id ) {
				break;
			}
		}
	}

	if ( MAX_USB_EVENT_HANDLER_NS <= i) {
		printk("Could not unregist a usb event handler. Because requested handler_id was not match.\n");
	} else {
		KFree( fsg->usb_event_handler_table[i] );
		fsg->usb_event_handler_table[i] = NULL;
	}
}

static void do_callback_usb_event_handler(struct fsg_dev *fsg, char status)
{
	int i;

	for ( i = 0 ; i < MAX_USB_EVENT_HANDLER_NS ; i++ ) {
		if ( fsg->usb_event_handler_table[i] != NULL ) {
			(*fsg->usb_event_handler_table[i]->handler)( status,
														 fsg->usb_event_handler_table[i]->handler_id);
		}
	}
}
#endif /* USB_FILE_STORAGE_NAVU */


#define PROTECT_TIME_OUT_VALUE    3000
static int lock_fsg(struct fsg_dev *fsg)
{
	int	rc;

	// Modified on 2008.03.20.
	// Because if the wakeup_fsg() is performed before lock_fsg(),
	// locked fsg is not released.

	rc = msecs_to_jiffies( PROTECT_TIME_OUT_VALUE );
	if ( atomic_read(&req_fsg_wakeup) ) {
		sonydbg(fsg, "already request wakeup!!!\n" ); // added on 2008.03.20
	} else {
		sonydbg(fsg, "request wakeup yet!!!\n" ); // added on 2008.03.20

		/* Wait until a signal arrives or we are woken up */
		for (;;) {
			try_to_freeze();
			set_current_state(TASK_INTERRUPTIBLE);
			if ( atomic_read(&req_fsg_wakeup) ) {
				sonydbg(fsg, "locked fsg wakeup\n");
				break;
			}
			if ( signal_pending(current) ) {
				rc = -ERESTARTSYS;
				break;
			} else {
				rc = schedule_timeout( rc );
                		if ( !rc ) {
					sonydbg(fsg, "lock_fsg timeout\n");
                        		break;
                		}
			}
        	}
		__set_current_state(TASK_RUNNING);
	}
	atomic_set(&req_fsg_wakeup, 0);
	return rc;
}

static void flush_all_waitqueue(struct fsg_dev* fsg)
{
	atomic_set(&data_get_control, 1);
	atomic_set(&data_send_control, 1);
	wake_up_all(&WaitDataQue);
	wake_up_all(&SendDataQue);
	wakeup_fsg(fsg);

//	atomic_set(&data_get_control, 0);
//	atomic_set(&data_send_control, 0);
}

#if 0
static void flush_usb_fifo(struct fsg_dev* fsg)
{
        struct fsg_buffhd	*bh;
	int    i;
	int    num_active;
	/* Cancel all the pending transfers */
	if (fsg->intreq_busy)
		usb_ep_dequeue(fsg->intr_in, fsg->intreq);
	for (i = 0; i < NUM_BUFFERS; ++i) {
		bh = &fsg->buffhds[i];
		if (bh->inreq_busy)
			usb_ep_dequeue(fsg->bulk_in, bh->inreq);
		if (bh->outreq_busy)
			usb_ep_dequeue(fsg->bulk_out, bh->outreq);
	}

	/* Wait until everything is idle */
	for (;;) {
		num_active = fsg->intreq_busy;
		for (i = 0; i < NUM_BUFFERS; ++i) {
			bh = &fsg->buffhds[i];
			num_active += bh->inreq_busy + bh->outreq_busy;
		}
		if (num_active == 0)
			break;
		if (sleep_thread(fsg))
			return;
	}

	/* Reset the I/O buffer states and pointers, the SCSI
	 * state, and the exception.  Then invoke the handler. */
	spin_lock_irq(&fsg->lock);

	for (i = 0; i < NUM_BUFFERS; ++i) {
		bh = &fsg->buffhds[i];
		bh->state = BUF_STATE_EMPTY;
	}
	fsg->next_buffhd_to_fill = fsg->next_buffhd_to_drain =
			&fsg->buffhds[0];

	/* Clear out the controller's fifos */
	if (fsg->bulk_in_enabled)
		usb_ep_fifo_flush(fsg->bulk_in);
	if (fsg->bulk_out_enabled)
		usb_ep_fifo_flush(fsg->bulk_out);
	if (fsg->intr_in_enabled)
		usb_ep_fifo_flush(fsg->intr_in);
	spin_unlock_irq(&fsg->lock);
}
#endif

static void wakeup_fsg(struct fsg_dev *fsg)
{
	/* Tell the main thread that something has happened */
	// atomic_set(&fsg_lock, 1);  // removed on 2008.03.20
	if ( fsg->thread_task )
		wake_up_process(fsg->thread_task);
}

static int transfer_open(struct inode *inode, struct file *file)
{
	int minor = iminor( inode );

	if( minor == 0 )
	{
		if(device_open) {
			return -EBUSY;
		}
		device_open++;
		try_module_get(THIS_MODULE);
		return SUCCESS;
	}
	else if( minor == 1 ) {
		return SUCCESS;
	}
	return -ENXIO;
}

static int transfer_release(struct inode *inode, struct file *file)
{
	int minor = iminor( inode );

	if( minor == 0 ) {
		device_open--;
		atomic_set(&data_get_control, 0);    /*Wait queue control var */
		atomic_set(&data_send_control, 0);
		atomic_set(&atm_left_to_write, 0);
		atomic_set(&atm_left_to_read, 0);
		atomic_set(&req_fsg_wakeup, 0);  // modified on 2008.03.20
		atomic_set(&data_response_control, 0);
		g_time_out = 0;
		flush_all_waitqueue(the_fsg);
		g_exit_flag = 0;
		g_first_wait = 1;

		module_put(THIS_MODULE);
		return 0;
	}
	else if( minor == 1 )
	{
		return 0;
	}
	return -ENXIO;
}

#define IOCTL_OK             (0)
#define IOCTL_ERR_INVAL      (-1)
#define IOCTL_ERR_SUSPEND    (-2)
#define IOCTL_ERR_TIMEOUT    (-3)
#define IOCTL_ERR_EXIT       (-4)
#define MAX_TRANS_SIZE       (64*1024)
//65536 is the max transfering size

static int transfer_ioctl(struct inode* pnode, struct file* pfile,
			  unsigned int cmd, unsigned long arg)
{
	int minor = iminor( pnode );

	if( minor == 0 ) {
		return transfer_ioctl_ms( pnode, pfile, cmd, arg);
	}
	else if( minor == 1 ) {
#ifdef USB_FILE_STORAGE_NAVU  //m33
		int ret_val         = 0;
		switch(cmd) {
			case IOCTL_MS_GET_CONNECTION:
				ret_val = put_user(Usb_connection_flag, (unsigned long*)arg);
				break;
			case IOCTL_MS_SET_MOUNT:
				if( arg == 1 ) {
					Mount_disable_flag = 0;	// Mount enable.
				}
			ret_val = 0;
				break;
		}
		return ret_val;
#endif  //m33
	}
	return -ENXIO;
}

static int transfer_ioctl_ms(struct inode* pnode, struct file* pfile,
			  unsigned int cmd, unsigned long arg)
{
	int ret_val         = 0;
	unsigned long* parg = (unsigned long*)arg;
	struct fsg_dev* fsg;
#ifdef SONY_DEBUG
	int i = 0;
#endif	/* SONY_DEBUG */

	if(device_open <= 0) {
		return IOCTL_ERR_INVAL;
	}

	if(NULL == the_fsg) {
		return IOCTL_ERR_INVAL;  /* fsg_dev struct can't be null */
	}
	fsg = the_fsg;
#if 0
	if( test_bit(SUSPENDED,&fsg->atomic_bitflags) || GET_STATUS ) {
		return IOCTL_ERR_SUSPEND;  /* usb is suspend */
	}
#endif
	sonydbg(fsg, ">>>>>>>>>>>>>>>>>>>bit flag is %d \n", (int)fsg->atomic_bitflags);


#if 1
	if( test_bit(RESETED,&fsg->atomic_bitflags)) {
		clear_bit(RESETED, &fsg->atomic_bitflags);
	//	return IOCTL_ERR_SUSPEND;  /* usb is suspend */
	}
#endif



	switch(cmd) {
	case IOCTL_MS_WAITEVENT:
		{
		  /* data copy stage */
		  struct fsg_buffhd	*bh;
		  //struct vendor_wait_event_t* vendor_wait;
		  //vendor_wait = (struct vendor_wait_event_t*)parg;
		  unsigned long amount  = 0;
		  unsigned long nwritten= 0;
		  long	nWaitTime = 0;

		  g_time_out = 0;

		  sonydbg(fsg, "this is wait_event \n");

		  if( NULL == (parg+4) ) {
		          return IOCTL_ERR_INVAL;
		  }


		  //4/12 add wait_timeout function
		  if( get_user(nWaitTime, (long __user *)(parg+4)) ) {
			ret_val = -EFAULT;
			goto label_exit;
		  }

		  /* set for wait status */
		  //atomic_set(&data_get_control, 0);
#if 0
		  if( test_bit(SUSPENDED,&fsg->atomic_bitflags) || GET_STATUS) {
			 sonydbg(fsg, "~~~~~~~~~~~~~~~~file storage driver is suspended \n");
			 return IOCTL_ERR_SUSPEND;  /* usb is suspend */
		  }
		  sonydbg(fsg, "bit flag is %d \n", fsg->atomic_bitflags);
#endif
		  //atomic_set(&ioctl_lock, 1);
		  /* sleep itself, wait for next data */
		  //if( DATA_DIR_FROM_HOST == fsg->data_dir) {
		  /* Wait , untill the kernel thread has data buffer */
		  if(1 == g_first_wait) {
		    g_first_wait = 0;
		    if( 1 != g_exit_flag ) {
		      atomic_set(&data_get_control, 0);
		    }
		  }

		  if(nWaitTime < 0)
		  {
		  	ret_val = wait_event_interruptible(WaitDataQue, atomic_read(&data_get_control));
			sonydbg(fsg, "wake up from fsg thread ...\n");

			if( ret_val < 0 ) {
				goto label_exit;
			}
		  }
		  else
                  {
		        sonydbg(fsg, "nWaitTime = %ld\n", nWaitTime);

	                ret_val = wait_event_interruptible_timeout( WaitDataQue,
						                    atomic_read(&data_get_control),
			                                            msecs_to_jiffies(nWaitTime) );

			sonydbg(fsg, "jiffies is = %ld\n", msecs_to_jiffies(nWaitTime));

			if(0 == ret_val)
			{
				ret_val = IOCTL_ERR_TIMEOUT;
				atomic_set(&data_get_control, 0);

				g_time_out = 1; //ensure ioctl SEND_RESPONSE not error

				sonydbg(fsg, "goto function%d\n", ret_val);
				goto label_exit;
			} else if( ret_val < 0 ) {
				goto label_exit;
		    }
		  }

		  //ljzhou 2006/4/29 ensure fsg_main_thread exit -> ioctl terminated
		  if( 1 == g_exit_flag ) {
		    g_exit_flag = 0;
		    ret_val = IOCTL_ERR_EXIT;
		    goto label_exit;
		  }

		  sonydbg(fsg, "here can go ? xxxxxxxxxxxx \n");
		  //}
		  //else if( DATA_DIR_TO_HOST == fsg->data_dir) {

		  //}
		  //Protected Invalid Parameters

		  if( 	(NULL == parg) || (NULL == (parg+1))
			|| (NULL == (parg+2)) || (NULL == (parg+3))
			|| (NULL == (void*)parg[1]) )
			//Delete by Steven 060710 start
//			|| (NULL == (void*)parg[3]) )
			//Delete by Steven 060710 end
		  {
//		    printk("<1> error argument in wait event\n");
			return IOCTL_ERR_INVAL;
                  }

		  if( put_user(fsg->data_size, (unsigned long*)parg+2) ){	/* put the data size */
				ret_val= -EFAULT;
				goto label_exit;
          }
		  if( put_user(fsg->cmnd_size, (unsigned long*)parg) ) {	/* put the command size */
				ret_val= -EFAULT;
				goto label_exit;
          }
		  if( copy_to_user( (void*)(*(parg+1)), fsg->cmnd, fsg->cmnd_size) ){
				ret_val= -EFAULT;
				goto label_exit;
          }

#ifdef SONY_DEBUG
		  sonydbg(fsg, "cmnd_size = %d, data_size = %d \n", fsg->cmnd_size, fsg->data_size);
		  for( i = 0; i < fsg->cmnd_size; i++) {
			sonydbg(fsg, "0x%2x ", fsg->cmnd[i]);
		  }
		  sonydbg(fsg, "\n");
#endif	/* SONY_DEBUG */

		  if( DATA_DIR_FROM_HOST == fsg->data_dir) {
			/* copy the command buffer */
			/* Now process the data buffer */
			//Added by Steven 060710 start
			if( NULL == (void*)parg[3] )
			{
				return IOCTL_ERR_INVAL;
			}
			//Added by Steven 060710 end


			while( atomic_read(&atm_left_to_write) > 0 ) {
				/* Write the received data to the backing file */
				bh = fsg->next_buffhd_to_drain;
				//if (bh->state == BUF_STATE_EMPTY)
				//	break;			// We stopped early

				if (bh->state == BUF_STATE_FULL) {
					fsg->next_buffhd_to_drain = bh->next;
					bh->state = BUF_STATE_EMPTY;

					amount = bh->outreq->actual;
					/* Perform the write */
					if( copy_to_user( ( ((unsigned char*)parg[3] )+nwritten ),
						      bh->buf, sizeof(char)*amount) ){
                                            ret_val = -EFAULT;
                                            goto label_exit;
                    }
					nwritten += amount;

					sonydbg(fsg, "befor fsg->residue = %d \n", fsg->residue);
					fsg->residue -= amount;

					amount = atomic_read(&atm_left_to_write) - amount;
					atomic_set(&atm_left_to_write, amount);


					sonydbg(fsg, "end fsg->residue = %d \n", fsg->residue);
					continue;   /* check if the next buffer is valid */
				}
				/* set for wait status */
				atomic_set(&data_get_control, 0);
				/* wake up the kernel thread */
				//wakeup_thread(fsg);
				atomic_set(&req_fsg_wakeup, 1); // added on 2008.03.20
				wakeup_fsg(fsg);
				sonydbg(fsg, "Wake up fsg thread \n");

				if( test_bit(SUSPENDED,&fsg->atomic_bitflags) || GET_STATUS ) {
					return IOCTL_ERR_SUSPEND;
				}
				sonydbg(fsg, ">>>>>>>>bit flag is %d \n", (int)fsg->atomic_bitflags);
				if(atomic_read(&atm_left_to_write) <= 0)
					break;
				/* sleep itself, wait for next data */
				if( 1 == g_exit_flag ) {
				  g_exit_flag = 0;
				  ret_val = IOCTL_ERR_EXIT;
				  goto label_exit;
				}

				ret_val = wait_event_interruptible(WaitDataQue, atomic_read(&data_get_control));
				if( ret_val < 0 ) {
					goto label_exit;
				}
			} /* While( atomic_read(&atm_left_to_write) > 0 ) */
			sonydbg(fsg, "IOCTL_MS_WAITEVENT function call wakeup_thread \n");
			//wakeup_thread(fsg); //ensure to wake up fsg thread
			atomic_set(&req_fsg_wakeup, 1); // added on 2008.03.20
			wakeup_fsg(fsg);
			sonydbg(fsg, "print out fsg->state = %d \n", fsg->state);
		  } /* if( DATA_DIR_FROM_HOST == fsg->data_dir) */
		  else if( DATA_DIR_NONE == fsg->data_dir ){
//		  	printk(KERN_DEBUG "IOCTL_MS_WAITEVENT DATA_DIR_NONE!\n");
			atomic_set(&req_fsg_wakeup, 1); // added on 2008.03.20
		  	wakeup_fsg(fsg);
		  	if( test_bit(SUSPENDED,&fsg->atomic_bitflags) || GET_STATUS ) {
					return IOCTL_ERR_SUSPEND;
			}
		  }/* if( DATA_DIR_NONE == fsg->data_dir) */
		}
 		//atomic_set(&ioctl_lock, 0);
		ret_val = fsg->data_dir; // fsg->cmnd[0];
		atomic_set(&data_get_control, 0);
		break;
	case IOCTL_MS_SENDRESPONSE:
		{
		  struct fsg_buffhd	*bh = NULL;
		  unsigned long sen_code = 0;
		  unsigned long data_dir = 0;
		  unsigned long buf_len  = 0;
		  unsigned char* buffer  = NULL;
		  unsigned long amount_to_send = 0;
		  unsigned long amount   = 0;
		  unsigned long nread    = 0;

		  //add by Richard 06.04.12
		  if(1 == g_time_out ) {
		        g_time_out = 0;
			ret_val = IOCTL_ERR_TIMEOUT;
			goto label_exit;
		  }

		  if( test_bit(SUSPENDED,&fsg->atomic_bitflags) || GET_STATUS )  {
			 sonydbg(fsg, "~~~~~~~~~~~~~~~~file storage driver is suspended \n");
			 return IOCTL_ERR_SUSPEND;  /* usb is suspend */
		  }

                  //Protected Invalid Parameter
		  if( 	(NULL == parg) || (NULL == (parg+1))
                	|| (NULL == (parg+2)) || (NULL == (parg+3))
			|| ((parg[2] > MAX_TRANS_SIZE) || (parg[2] == 0))
			|| (NULL == (void*)parg[3]) )
		  {
//		    printk("<1> error argument in send response command\n");
			return IOCTL_ERR_INVAL;
                  }
		  //printk("<1> driver ... this is begin buffer \n");

		  //spin_lock_irq(&fsg->lock);
		  if( get_user(sen_code, (unsigned long __user*)parg ) ) {
			ret_val = -EFAULT;
			goto label_exit;
		  }
		  if( get_user(data_dir, (unsigned long __user *)(parg+1)) ){
			ret_val = -EFAULT;
			goto label_exit;
		  }
		  if( get_user(buf_len, (unsigned long __user *)(parg+2)) ) {
			ret_val = -EFAULT;
			goto label_exit;
		  }
		  buffer = (unsigned char*)parg[3];
		  //spin_lock_irq(&fsg->lock);

		  //printk("<1> driver ... this is end the buffer \n");

		  sonydbg(fsg, "sen_code = 0x%x, data_dir = %d, buf_len = %d \n",
			  (unsigned int)sen_code, (unsigned int)data_dir, (unsigned int)buf_len);
		  /* send sens_code from device to host */
		  spin_lock_irq(&fsg->lock);
		  if(0 != fsg->curlun) {
		    fsg->curlun->sense_data = sen_code;
		  }
		  /* set total size */
		  // ljzhou 2006.6.19 start
		  if(buf_len >= fsg->data_size) {
		    amount_to_send = fsg->data_size;
		    sonydbg(fsg, "2006.6.19 amount_to_send buf_len >= fsg->data_size is %ld\n", amount_to_send);
		    fsg->residue = fsg->data_size;
		  }
		  else { // buf_len < fsg->data_size
		    amount_to_send = buf_len;
		    sonydbg(fsg, "2006.6.19 buf_len < fsg->data_size %ld\n", amount_to_send);
		    fsg->curlun->sense_data      = SS_COMMUNICATION_FAILURE;
		    fsg->curlun->sense_data_info = buf_len;
		    fsg->residue = fsg->data_size;
		    //printk("<1> before fsg->residue = %d, buf_len = %d", fsg->residue, buf_len);
		    //fsg->residue = buf_len;
		    //printk("<1> after fsg->residue = %d, buf_len = %d", fsg->residue, buf_len);

		    //if(0 == (buf_len % 64)) {
		    //  g_vendor_short = 1;
		    //}
		  }
		  spin_unlock_irq(&fsg->lock);

		  atomic_set(&atm_left_to_read, amount_to_send);
		  // ljzhou 2006.6.19 end
		  sonydbg(fsg, "fsg->data_size = %d *********\n", fsg->data_size);
		  sonydbg(fsg, "IOCTL_MS_SENDRESPONSE \n");
		  /* send buffer from device to host */

		  //steven 0704 start
		  //atomic_set(&data_response_control, 0);
		  //steven 0704 end
		  while(  amount_to_send > 0 ) {
			/* Write the received data to the backing file */
			bh = fsg->next_buffhd_to_fill;

			//printk("<1> driver ... this is loop \n");

			amount = min((unsigned int) amount_to_send, mod_data.buflen);
			/* Wait for the next buffer to become available */
			bh = fsg->next_buffhd_to_fill;
			if(NULL == bh) {
				//printk("file_storage_navu.c: next_bufhd_to_fill == NULL\n");
			}
			while (bh->state != BUF_STATE_EMPTY) {
				/*
				printk(KERN_ALERT "buffer state empty ?\n");
				ret_val = sleep_thread(fsg);
				printk(KERN_ALERT "Sleep Status %d!!!!\n",ret_val);
				if(ret_val != 0) {
					printk(KERN_ALERT "Sleep failed !!!!\n");
					ret_val = IOCTL_ERR_INVAL;
					//return ret_val;
					goto label_exit;
				}*/
				//steven 0704 start

				//printk(KERN_ALERT "Buffer is empty? !!!!\n");
#if 1
				if(0 == wait_event_interruptible_timeout(ResponseQue, atomic_read(&data_response_control),
						    	    msecs_to_jiffies(2000)) ) {
					//printk(KERN_ALERT "Sleep failed !!!!\n");
					ret_val = IOCTL_ERR_INVAL;
					//return ret_val;
					goto label_exit;
				}
				atomic_set(&data_response_control, 0);
				if( test_bit(RESETED,&fsg->atomic_bitflags)) {
				    //printk(KERN_ALERT "Usb has reseted!!!!\n");
				    ret_val = IOCTL_ERR_INVAL;
				    //return ret_val;
				    goto label_exit;
				}

#endif
#if 0
				ret_val = wait_event_interruptible(ResponseQue, atomic_read(&data_response_control));
				atomic_set(&data_response_control, 0);
				if(ret_val != 0 || test_bit(SUSPENDED,&fsg->atomic_bitflags ))
				{
					printk(KERN_ALERT "Sleep failed !!!!\n");
					ret_val = IOCTL_ERR_INVAL;
					//return ret_val;
					goto label_exit;
				}
#endif
				//printk(KERN_ALERT "Test Buffer end!!!!\n");
				//steven 0704 end

				//lock_fsg(fsg);
				sonydbg(fsg, "IOCTL_MS_SENDRESPONSE : sleep fsg thread \n");
			}

			/* If we were asked to read past the end of file,
			 * end with an empty buffer. */
			if (amount == 0) {
			  if(NULL == bh->inreq) {
			    ret_val = IOCTL_ERR_INVAL;
			    goto label_exit;
			  }
			  bh->inreq->length = 0;
			  bh->state = BUF_STATE_FULL;
			  break;
			}

			if(NULL == bh->inreq) {
			  ret_val = IOCTL_ERR_INVAL;
			  //printk("<1>bh->inreq == NULL\n");
			  goto label_exit;
			}
			/* Perform the read */
			if( copy_from_user( bh->buf, (unsigned char*)buffer+nread, amount) ){
				ret_val = -EFAULT;
				goto label_exit;
            }
			nread += amount;

			sonydbg(fsg, "before fsg->residue = %d \n", fsg->residue);
			fsg->residue -= amount;
			sonydbg(fsg, "end fsg->residue = %d \n", fsg->residue);

			if(NULL == bh->inreq) {
			  ret_val = IOCTL_ERR_INVAL;
			  goto label_exit;
			}
			bh->inreq->length = amount;
			sonydbg(fsg, "can type the bh->inreq->length \n");



                        bh->state = BUF_STATE_FULL;
			amount_to_send  -= amount;
			/* Set correct data */
			atomic_set(&atm_left_to_read, amount_to_send);

			if(0 == amount_to_send) {
				sonydbg(fsg, "IOCTL_MS_SENDRESPONSE : 0 == amount_to_send \n");
				break;
			}

			/* Sleep itself , let the data buffer to send */
			//Modified 060705 Steven Start.
			atomic_set(&data_send_control, 0);
			//Modified 060705 Steven End.
			/* sleep itself, wake up main thread */
			sonydbg(fsg, "Wake up fsg thread : IOCTL_MS_SENDRESPONSE \n");
			//wakeup_thread(fsg);
			//printk(KERN_ALERT "Wake up fsg thread : IOCTL_MS_SENDRESPONSE \n");
			atomic_set(&req_fsg_wakeup, 1); // added on 2008.03.20
			wakeup_fsg(fsg);

			if( test_bit(SUSPENDED,&fsg->atomic_bitflags) || GET_STATUS) {
				 sonydbg(fsg, "~~~~~~~~~~~~~~~~file storage driver is suspended \n");
				 return IOCTL_ERR_SUSPEND;  /* usb is suspend */
			}
			if( test_bit(RESETED,&fsg->atomic_bitflags)) {
				 //printk(KERN_ALERT "Usb has reseted!!!!\n");
				 ret_val = IOCTL_ERR_INVAL;
				 //return ret_val;
				 goto label_exit;
			}


			sonydbg(fsg, "Sleep SendDataQueue : IOCTL_MS_SENDRESPONSE \n");

			//printk(KERN_ALERT "Sleep SendDataQueue : IOCTL_MS_SENDRESPONSE \n");
			//wait_event_interruptible(SendDataQue, atomic_read(&data_send_control));
			//Modified 060705 Steven Start.
			if(0 == wait_event_interruptible_timeout(SendDataQue, atomic_read(&data_send_control),
						    	    msecs_to_jiffies(PROTECT_TIME_OUT_VALUE)) ) {
				//printk("<1> this is timeout in the send data ioctl function...\n");
				 ret_val = IOCTL_ERR_INVAL;
			  	//printk("<1>bh->inreq == NULL\n");
			  	goto label_exit;
			}
			//atomic_set(&data_send_control, 0);
			//Modified 060705 Steven End.
			//wait_event(SendDataQue, atomic_read(&data_send_control));
			//ljzhou 2006/4/29 ensure fsg_main_thread exit -> ioctl terminated
			if( 1 == g_exit_flag ) {
				g_exit_flag = 0;
				ret_val = IOCTL_ERR_TIMEOUT;
				goto label_exit;
			}

			fsg->next_buffhd_to_fill = bh->next;
			//printk(KERN_ALERT "After Sleep : IOCTL_MS_SENDRESPONSE \n");
		  } /* while( atomic_read(&atm_left_to_read) > 0 ) */
		  sonydbg(fsg, "FSG : fsg->residue = %d \n", fsg->residue);
		  //wakeup_thread(fsg);  //ensure that main fsg is waken up
		  atomic_set(&req_fsg_wakeup, 1); // added on 2008.03.20
		  wakeup_fsg(fsg);
		}
		sonydbg(fsg, "IOCTL_MS_RESP: fsg->next_buffhd_to_fill->inreq->length = %d \n",
				fsg->next_buffhd_to_fill->inreq->length);
		ret_val = 0;
		break;
	default:
		sonydbg(fsg, "Unknown ioctl command, please check the application \n");
		ret_val = IOCTL_ERR_INVAL;
		break;
	}
label_exit:
	sonydbg(fsg, "label_exit : = %d\n", ret_val);
	return ret_val;
}

static enum data_direction parse_command_dir(struct fsg_dev* fsg)
{
	return fsg->data_dir;
}

static int do_send_data(struct fsg_dev* fsg)
{
	struct fsg_buffhd	*bh ;
	u32			amount_left;
	//	unsigned int		amount;
	int                     ret_value = -EIO;
	/* read the total bytes to send to host */
	//amount_left = atomic_read(&atm_left_to_read);

	/* Wait for Command pass to the application
	   and expect the next ioctl response comming */
	atomic_set(&data_get_control, 1);
	wake_up(&WaitDataQue);

	sonydbg(fsg, "Wake up WaitDataQue : do_send_data \n");

	sonydbg(fsg, "Sleep the fsg thread \n");
	//sleep_thread(fsg);
	if(0 == lock_fsg(fsg)) {
		flush_all_waitqueue(fsg);
		//flush_usb_fifo(fsg);
		//Added by Steven 060707 start
		//printk(KERN_DEBUG "SS_COMMUNICATION_FAILURE Cur LUN is %d!\n",fsg->curlun);
		if(0 != fsg->curlun){
		   fsg->curlun->sense_data = SS_COMMUNICATION_FAILURE;
		}
		//Added by Steven 060707 End
		return -EINVAL;
	}
//	atomic_set(&data_get_control, 0);
	sonydbg(fsg, "fsg thread is waken up \n");

	//atomic_set(&data_send_control, 0);
	//wait_event_interruptible_timeout( SendDataQue,
	//				  atomic_read(&data_send_control),
	//				  200000 );
	for (;;) {

		/* Figure out how much we need to read:
		 * Try to read the remaining amount.
		 */
		bh = fsg->next_buffhd_to_fill;
		amount_left = atomic_read(&atm_left_to_read);
		if (amount_left == 0) {
			sonydbg(fsg, "do_send_data : amount_left == 0, break \n");
			//ret_value = SUCCESS;
			break;		// No more left to read
		}

		//amount = min((unsigned int) amount_left, mod_data.buflen);
		/* Wait for the next buffer to become available */
		//wait_event_interruptible(SendDataQue, (BUF_STATE_EMPTY == bh->state));
		//sonydbg(fsg, "bh->state = %d \n", bh->state);

		/* Send this buffer and go read some more */
		bh->inreq->zero = 0;
		start_transfer(fsg, fsg->bulk_in, bh->inreq,
				&bh->inreq_busy, &bh->state);

		if (signal_pending(current)) {
			sonydbg(fsg, "do_send_data : signal_pending(current) function \n");
			return -EINTR;
		}

		atomic_set(&data_send_control, 1);
		wake_up(&SendDataQue);

		sonydbg(fsg, "do_send_data, Wake up the SendDataQue , sleep fsg thread \n");
		//sleep_thread(fsg);
		if(0 == lock_fsg(fsg)) {
			flush_all_waitqueue(fsg);
			//flush_usb_fifo(fsg);
			//Added by Steven 060707 start
			//printk(KERN_DEBUG "SS_COMMUNICATION_FAILURE Cur LUN is %d!\n",fsg->curlun);
			if(0 != fsg->curlun){
		   		fsg->curlun->sense_data = SS_COMMUNICATION_FAILURE;
			}
			//Added by Steven 060707 End
			return -EINVAL;
		}
	}
	return ret_value;		// No default reply
}


/*-------------------------------------------------------------------------*/

static int do_receive_data(struct fsg_dev *fsg)
{
	struct fsg_buffhd	*bh;
	int			get_some_more;
	u32			amount_left_to_req; // amount_left_to_write;
	unsigned int		amount;
	int			rc;

	/* Carry out the file writes */
	get_some_more = 1;
	amount_left_to_req = fsg->data_size;
	//Lock the variable due to this var must used in two thread
	atomic_set(&atm_left_to_write, amount_left_to_req);

	sonydbg(fsg, "amount_left_to_req is %d \n", amount_left_to_req);
	while ( atomic_read(&atm_left_to_write) > 0) {
		/* Queue a request for more data from the host */
		bh = fsg->next_buffhd_to_fill;
		if (bh->state == BUF_STATE_EMPTY && get_some_more) {
		/* Figure out how much we want to get:
		* Try to get the remaining amount.
		* But don't get more than the buffer size.
		* If this means getting 0, then we were asked
			* Finally, round down to a block boundary. */
			amount = min(amount_left_to_req, mod_data.buflen);
			if (amount == 0) {
				get_some_more = 0;
				continue;
			}

			sonydbg(fsg, "amount = %d\n", amount);
			sonydbg(fsg, "before fsg->usb_amount_left = %d\n", fsg->usb_amount_left);
			/* Get the next buffer */
			fsg->usb_amount_left -= amount;
			amount_left_to_req -= amount;
			if (amount_left_to_req == 0)
				get_some_more = 0;

			sonydbg(fsg, "after fsg->usb_amount_left = %d\n", fsg->usb_amount_left);
			sonydbg(fsg, "amount_left_to_req = %d\n", amount_left_to_req);
			/* amount is always divisible by 512, hence by
			* the bulk-out maxpacket size */
			bh->outreq->length = bh->bulk_out_intended_length = amount;

			sonydbg(fsg, "bh->out->length = %d\n", bh->outreq->length);

			start_transfer(fsg, fsg->bulk_out, bh->outreq,
				&bh->outreq_busy, &bh->state);
			fsg->next_buffhd_to_fill = bh->next;

			//FIX ??
			//bh->state = BUF_STATE_FULL;
			sonydbg(fsg, "do_receive_data: read buffer from usb host\n");
			continue;
		}
		/* Write the received data to the backing file */
		bh = fsg->next_buffhd_to_drain;
		if (bh->state == BUF_STATE_EMPTY && !get_some_more) {
			sonydbg(fsg, "next_buffhd_to_drain->state == EMPTY && !get_some_more \n");
			break;			// We stopped early
		}
		if (bh->state == BUF_STATE_FULL) {
			/* Did something go wrong with the transfer? */
			if (bh->outreq->status != 0) {
				sonydbg(fsg, "bh->outreq->status != 0 , do_receive_data function \n");
				bh->state = BUF_STATE_EMPTY;  //recover its origin status
				break;
			}

			sonydbg(fsg, "common : buffer state = %d \n", bh->state);
			sonydbg(fsg, "common : atm_left_to_write = %d \n", atomic_read(&atm_left_to_write));
			sonydbg(fsg, "common : fsg->usb_amount_left = %d \n", fsg->usb_amount_left);
			sonydbg(fsg, "common : amount_left_to_req = %d \n", amount_left_to_req);

			//wait_event_interruptible(IoctlQue, atomic_read(&ioctl_lock));
			//sonydbg(fsg, "open ioctl lock , ioctl_lock == 1 \n");
			/* wake up WaitDataQueue to process ioctl command
			and copy usb data to user space */
			atomic_set(&data_get_control, 1);
			//wake_up(&WaitDataQue);
			wake_up_interruptible_sync(&WaitDataQue);

			sonydbg(fsg, "Wake up WaitDataQueue, and wait for fsg thread \n");

			/* sleep itself, ioctl function will wake up this thread */
			//sleep_thread(fsg);
			if(0 == lock_fsg(fsg)) {
				flush_all_waitqueue(fsg);
				//flush_usb_fifo(fsg);
				//Added by Steven 060707 start
				//printk(KERN_DEBUG "SS_COMMUNICATION_FAILURE Cur LUN is %d!\n",fsg->curlun);
				if(0 != fsg->curlun){
				   fsg->curlun->sense_data = SS_COMMUNICATION_FAILURE;
				}
				//Added by Steven 060707 End
				return -EINVAL;
			}
			atomic_set(&data_get_control, 0);

			sonydbg(fsg, "Fsg thread sleep itself \n");

			/* be woken up by ioctl function */
			if (signal_pending(current))
			{
				siginfo_t info;
				int sig = dequeue_signal_lock(current,
					&current->blocked,
					&info);
				if(SIGUSR1 != sig) {
					sonydbg(fsg, "print out the current signal = %x \n",
						sig);
					return -EINTR;		// Interrupted!
				}
				else {
					sonydbg(fsg, "the current signal is SIGUSR1 \n");
				}
			}
			/* Did the host decide to stop early? */
			if (bh->outreq->actual != bh->outreq->length) {
				fsg->short_packet_received = 1;
				sonydbg(fsg, "Did the host decide to stop early ? \n");
				break;
			}
			continue;
		}
		else {
			sonydbg(fsg, "here is unexpected situation \n");
			sonydbg(fsg, "buffer state = %d \n", bh->state);
			sonydbg(fsg, "atm_left_to_write = %d \n", atomic_read(&atm_left_to_write));
			sonydbg(fsg, "fsg->usb_amount_left = %d \n", fsg->usb_amount_left);
			sonydbg(fsg, "amount_left_to_req = %d \n", amount_left_to_req);
		}
		/* Wait for something to happen */
		sonydbg(fsg, "do_receive_data : before sleep function\n");
		if ((rc = sleep_thread(fsg)) != 0)
			return rc;
		sonydbg(fsg, "do_receive_data : end sleep function\n");
	}
	return SUCCESS;		// No default reply
}


/*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
static int do_no_data(struct fsg_dev *fsg)
{
	atomic_set(&data_get_control, 1);
	//wake_up(&WaitDataQue);
	wake_up_interruptible_sync(&WaitDataQue);
	if(0 == lock_fsg(fsg)) {
		flush_all_waitqueue(fsg);
		//flush_usb_fifo(fsg);
		//Added by Steven 060707 start
		//printk(KERN_DEBUG "do_no_data : SS_COMMUNICATION_FAILURE Cur LUN is %d!\n",fsg->curlun);
		if(0 != fsg->curlun){
	        	fsg->curlun->sense_data = SS_COMMUNICATION_FAILURE;
		}
		//Added by Steven 060707 End
		return -EINVAL;
	}
	atomic_set(&data_get_control, 0);

	printk(KERN_DEBUG "do_no_data: end do_no_data function!\n");
	fsg->residue = 0;
	return SUCCESS;		// No default reply

}
/*-------------------------------------------------------------------------*/


/* Check whether the command is properly formed and whether its data size
 * and direction agree with the values we already have. */
static int check_vendor_command(struct fsg_dev *fsg, int cmnd_size,
		enum data_direction data_dir,
		int needs_medium, const char *name)
{
	int			lun = fsg->cmnd[1] >> 5;
	static const char	dirletter[4] = {'u', 'o', 'i', 'n'};
	char			hdlen[20];
	struct lun		*curlun;

	/* Adjust the expected cmnd_size for protocol encapsulation padding.
	 * Transparent SCSI doesn't pad. */
	if (protocol_is_scsi())
		;

	/* There's some disagreement as to whether RBC pads commands or not.
	 * We'll play it safe and accept either form. */
	else if (mod_data.protocol_type == USB_SC_RBC) {
		if (fsg->cmnd_size == 12)
			cmnd_size = 12;

	/* All the other protocols pad to 12 bytes */
	} else
		cmnd_size = 12;

	hdlen[0] = 0;
	if (fsg->data_dir != DATA_DIR_UNKNOWN)
		snprintf(hdlen, sizeof(hdlen), ", H%c=%u", dirletter[(int) fsg->data_dir],
				 fsg->data_size);
	VDBG(fsg, "SCSI command: %s;  Dc=%d, D%c=%u;  Hc=%d%s\n",
			name, cmnd_size, dirletter[(int) data_dir],
			fsg->data_size_from_cmnd, fsg->cmnd_size, hdlen);

	/* We can't reply at all until we know the correct data direction
	 * and size. */
	if (fsg->data_size_from_cmnd == 0)
		data_dir = DATA_DIR_NONE;
	if (fsg->data_dir == DATA_DIR_UNKNOWN) {	// CB or CBI
		fsg->data_dir = data_dir;
		fsg->data_size = fsg->data_size_from_cmnd;

	} else {					// Bulk-only
		if (fsg->data_size < fsg->data_size_from_cmnd) {

			/* Host data size < Device data size is a phase error.
			 * Carry out the command, but only transfer as much
			 * as we are allowed. */
			fsg->data_size_from_cmnd = fsg->data_size;
			fsg->phase_error = 1;
		}
	}
	fsg->residue = fsg->usb_amount_left = fsg->data_size;

	/* Conflicting data directions is a phase error */
	if (fsg->data_dir != data_dir && fsg->data_size_from_cmnd > 0) {
		fsg->phase_error = 1;
		sonydbg(fsg, "Check Vendor Command ------------------Conflicting data directions\n");
		return -EINVAL;
	}

	/* Verify the length of the command itself */
	if (cmnd_size != fsg->cmnd_size) {

		/* Special case workaround: MS-Windows issues REQUEST SENSE
		 * with cbw->Length == 12 (it should be 6). */
		if (fsg->cmnd[0] == SC_REQUEST_SENSE && fsg->cmnd_size == 12)
			cmnd_size = fsg->cmnd_size;
		else {
			fsg->phase_error = 1;
			sonydbg(fsg, "Check Vendor Command ------------------- cmnd_size != fsg->cmnd_size\n");
			return -EINVAL;
		}
	}

	/* Check that the LUN values are oonsistent */
	if (transport_is_bbb()) {
		if (fsg->lun != lun)
			DBG(fsg, "using LUN %d from CBW, "
					"not LUN %d from CDB\n",
					fsg->lun, lun);
	} else
		fsg->lun = lun;		// Use LUN from the command

	/* Check the LUN */
	if (fsg->lun >= 0 && fsg->lun < fsg->nluns) {
		fsg->curlun = curlun = &fsg->luns[fsg->lun];
		if (fsg->cmnd[0] != SC_REQUEST_SENSE) {
			curlun->sense_data = SS_NO_SENSE;
			curlun->sense_data_info = 0;
		}
	} else {
		fsg->curlun = curlun = NULL;
		fsg->bad_lun_okay = 0;

		/* INQUIRY and REQUEST SENSE commands are explicitly allowed
		 * to use unsupported LUNs; all others may not. */
		if (fsg->cmnd[0] != SC_INQUIRY &&
				fsg->cmnd[0] != SC_REQUEST_SENSE) {
			DBG(fsg, "unsupported LUN %d\n", fsg->lun);
			sonydbg(fsg, "Check Vendor Command ----------- unsupported LUN\n");
			return -EINVAL;
		}
	}

	/* If a unit attention condition exists, only INQUIRY and
	 * REQUEST SENSE commands are allowed; anything else must fail. */
	if (curlun && curlun->unit_attention_data != SS_NO_SENSE &&
			fsg->cmnd[0] != SC_INQUIRY &&
			fsg->cmnd[0] != SC_REQUEST_SENSE) {
		curlun->sense_data = curlun->unit_attention_data;
		curlun->unit_attention_data = SS_NO_SENSE;
		sonydbg(fsg, "Check Vendor Command ------------- Unit _attention\n");
		//return -EINVAL;
	}

	/* If the medium isn't mounted and the command needs to access
	 * it, try to mount it. */
	if (curlun && !backing_file_is_open(curlun) && needs_medium) {
		if (open_backing_file(curlun, mod_data.file[fsg->lun])){
			curlun->sense_data = SS_MEDIUM_NOT_PRESENT;
			sonydbg(fsg, "Check Vendor Command -------------- Media Not Presented\n");
			return -EINVAL;
		}
	}

	return 0;
}
// #endif /* MACH_SONY_NAVU is old label. */
#endif // USB_FILE_STORAGE_NAVU


/*-------------------------------------------------------------------------*/

#ifdef DUMP_MSGS

static void dump_msg(struct fsg_dev *fsg, const char *label,
		const u8 *buf, unsigned int length)
{
	unsigned int	start, num, i, offset;
	char		line[52], *p;

	if (length >= 512)
		return;
	DBG(fsg, "%s, length %u:\n", label, length);

	start = 0;
	while (length > 0) {
		num = min(length, 16u);
		p = line;
		offset = 0;
		for (i = 0; i < num; ++i) {
			if (i == 8){
				*p++ = ' ';
				offset++;
			}
			if( sizeof(line) < offset )
				break;
			snprintf(p, sizeof(line)-offset, " %02x", buf[i]);
			p += 3;
			offset += 3;
		}
		*p = 0;
		printk(KERN_DEBUG "%6x: %s\n", start, line);
		buf += num;
		start += num;
		length -= num;
	}
}

static void inline dump_cdb(struct fsg_dev *fsg)
{}

#else

static void inline dump_msg(struct fsg_dev *fsg, const char *label,
		const u8 *buf, unsigned int length)
{}

static void inline dump_cdb(struct fsg_dev *fsg)
{
	int	i;
	unsigned int num;
	char	cmdbuf[3*MAX_COMMAND_SIZE + 1];

	num = min(MAX_COMMAND_SIZE, fsg->cmnd_size);
	for (i = 0; i < num; ++i)
		snprintf(cmdbuf + i*3, sizeof(cmdbuf) - i*3, " %02x", fsg->cmnd[i]);
	VDBG(fsg, "SCSI CDB: %s\n", cmdbuf);
}

#endif /* DUMP_MSGS */


static int fsg_set_halt(struct fsg_dev *fsg, struct usb_ep *ep)
{
	const char	*name;

	if (ep == fsg->bulk_in)
		name = "bulk-in";
	else if (ep == fsg->bulk_out)
		name = "bulk-out";
	else
		name = ep->name;
	DBG(fsg, "%s set halt\n", name);
	return usb_ep_set_halt(ep);
}


/*-------------------------------------------------------------------------*/

/* Routines for unaligned data access */

static u16 inline get_be16(u8 *buf)
{
	return ((u16) buf[0] << 8) | ((u16) buf[1]);
}

static u32 inline get_be32(u8 *buf)
{
	return ((u32) buf[0] << 24) | ((u32) buf[1] << 16) |
			((u32) buf[2] << 8) | ((u32) buf[3]);
}

static void inline put_be16(u8 *buf, u16 val)
{
	buf[0] = val >> 8;
	buf[1] = val;
}

static void inline put_be32(u8 *buf, u32 val)
{
	buf[0] = val >> 24;
	buf[1] = val >> 16;
	buf[2] = val >> 8;
#ifdef USB_FILE_STORAGE_NAVU
	buf[3] = val;
#else
	buf[3] = val & 0xff;
#endif
}


/*-------------------------------------------------------------------------*/

/*
 * DESCRIPTORS ... most are static, but strings and (full) configuration
 * descriptors are built on demand.  Also the (static) config and interface
 * descriptors are adjusted during fsg_bind().
 */
#define STRING_MANUFACTURER	1
#define STRING_PRODUCT		2
#define STRING_SERIAL		3
#ifdef USB_FILE_STORAGE_NAVU
#else
#define STRING_CONFIG		4
#define STRING_INTERFACE	5
#endif

/* There is only one configuration. */
#define	CONFIG_VALUE		1

static struct usb_device_descriptor
device_desc = {
	.bLength =		sizeof device_desc,
	.bDescriptorType =	USB_DT_DEVICE,

	.bcdUSB =		__constant_cpu_to_le16(0x0200),
	.bDeviceClass =		USB_CLASS_PER_INTERFACE,

	/* The next three values can be overridden by module parameters */
	.idVendor =		__constant_cpu_to_le16(DRIVER_VENDOR_ID),
	.idProduct =		__constant_cpu_to_le16(DRIVER_PRODUCT_ID),
	.bcdDevice =		__constant_cpu_to_le16(0xffff),

	.iManufacturer =	STRING_MANUFACTURER,
	.iProduct =		STRING_PRODUCT,
	.iSerialNumber =	STRING_SERIAL,
	.bNumConfigurations =	1,
};

static struct usb_config_descriptor
config_desc = {
	.bLength =		sizeof config_desc,
	.bDescriptorType =	USB_DT_CONFIG,

	/* wTotalLength computed by usb_gadget_config_buf() */
	.bNumInterfaces =	1,
	.bConfigurationValue =	CONFIG_VALUE,
#ifdef USB_FILE_STORAGE_NAVU
#else
	.iConfiguration =	STRING_CONFIG,
#endif
#ifdef USB_FILE_STORAGE_NAVU
	.bmAttributes =		USB_CONFIG_ATT_ONE,
	.bMaxPower =		250, // bus-powered
#else
	.bmAttributes =		USB_CONFIG_ATT_ONE | USB_CONFIG_ATT_SELFPOWER,
	.bMaxPower =		1,	// self-powered
#endif
};

static struct usb_otg_descriptor
otg_desc = {
	.bLength =		sizeof(otg_desc),
	.bDescriptorType =	USB_DT_OTG,

	.bmAttributes =		USB_OTG_SRP,
};

/* There is only one interface. */

static struct usb_interface_descriptor
intf_desc = {
	.bLength =		sizeof intf_desc,
	.bDescriptorType =	USB_DT_INTERFACE,

	.bNumEndpoints =	2,		// Adjusted during fsg_bind()
	.bInterfaceClass =	USB_CLASS_MASS_STORAGE,
	.bInterfaceSubClass =	USB_SC_SCSI,	// Adjusted during fsg_bind()
	.bInterfaceProtocol =	USB_PR_BULK,	// Adjusted during fsg_bind()
#ifdef USB_FILE_STORAGE_NAVU
#else
	.iInterface =		STRING_INTERFACE,
#endif
};

/* Three full-speed endpoint descriptors: bulk-in, bulk-out,
 * and interrupt-in. */

static struct usb_endpoint_descriptor
fs_bulk_in_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	.bEndpointAddress =	USB_DIR_IN,
	.bmAttributes =		USB_ENDPOINT_XFER_BULK,
	/* wMaxPacketSize set by autoconfiguration */
};

static struct usb_endpoint_descriptor
fs_bulk_out_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	.bEndpointAddress =	USB_DIR_OUT,
	.bmAttributes =		USB_ENDPOINT_XFER_BULK,
	/* wMaxPacketSize set by autoconfiguration */
};

static struct usb_endpoint_descriptor
fs_intr_in_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	.bEndpointAddress =	USB_DIR_IN,
	.bmAttributes =		USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize =	__constant_cpu_to_le16(2),
	.bInterval =		32,	// frames -> 32 ms
};

static const struct usb_descriptor_header *fs_function[] = {
	(struct usb_descriptor_header *) &otg_desc,
	(struct usb_descriptor_header *) &intf_desc,
	(struct usb_descriptor_header *) &fs_bulk_in_desc,
	(struct usb_descriptor_header *) &fs_bulk_out_desc,
	(struct usb_descriptor_header *) &fs_intr_in_desc,
	NULL,
};
#define FS_FUNCTION_PRE_EP_ENTRIES	2


#ifdef	CONFIG_USB_GADGET_DUALSPEED

/*
 * USB 2.0 devices need to expose both high speed and full speed
 * descriptors, unless they only run at full speed.
 *
 * That means alternate endpoint descriptors (bigger packets)
 * and a "device qualifier" ... plus more construction options
 * for the config descriptor.
 */
static struct usb_qualifier_descriptor
dev_qualifier = {
	.bLength =		sizeof dev_qualifier,
	.bDescriptorType =	USB_DT_DEVICE_QUALIFIER,

	.bcdUSB =		__constant_cpu_to_le16(0x0200),
	.bDeviceClass =		USB_CLASS_PER_INTERFACE,

	.bNumConfigurations =	1,
};

static struct usb_endpoint_descriptor
hs_bulk_in_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	/* bEndpointAddress copied from fs_bulk_in_desc during fsg_bind() */
	.bmAttributes =		USB_ENDPOINT_XFER_BULK,
	.wMaxPacketSize =	__constant_cpu_to_le16(512),
};

static struct usb_endpoint_descriptor
hs_bulk_out_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	/* bEndpointAddress copied from fs_bulk_out_desc during fsg_bind() */
	.bmAttributes =		USB_ENDPOINT_XFER_BULK,
	.wMaxPacketSize =	__constant_cpu_to_le16(512),
	.bInterval =		1,	// NAK every 1 uframe
};

static struct usb_endpoint_descriptor
hs_intr_in_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	/* bEndpointAddress copied from fs_intr_in_desc during fsg_bind() */
	.bmAttributes =		USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize =	__constant_cpu_to_le16(2),
	.bInterval =		9,	// 2**(9-1) = 256 uframes -> 32 ms
};

static const struct usb_descriptor_header *hs_function[] = {
	(struct usb_descriptor_header *) &otg_desc,
	(struct usb_descriptor_header *) &intf_desc,
	(struct usb_descriptor_header *) &hs_bulk_in_desc,
	(struct usb_descriptor_header *) &hs_bulk_out_desc,
	(struct usb_descriptor_header *) &hs_intr_in_desc,
	NULL,
};
#define HS_FUNCTION_PRE_EP_ENTRIES	2

/* Maxpacket and other transfer characteristics vary by speed. */
#define ep_desc(g,fs,hs)	(((g)->speed==USB_SPEED_HIGH) ? (hs) : (fs))

#else

/* If there's no high speed support, always use the full-speed descriptor. */
#define ep_desc(g,fs,hs)	fs

#endif	/* !CONFIG_USB_GADGET_DUALSPEED */


/* The CBI specification limits the serial string to 12 uppercase hexadecimal
 * characters. */
#ifdef USB_FILE_STORAGE_NAVU
static char				manufacturer[40];
#else
static char				manufacturer[64];
#endif
static char				serial[13];

/* Static strings, in UTF-8 (for simplicity we use only ASCII characters) */
static struct usb_string		strings[] = {
	{STRING_MANUFACTURER,	manufacturer},
	{STRING_PRODUCT,	longname},
	{STRING_SERIAL,		serial},
#ifdef USB_FILE_STORAGE_NAVU
#else
	{STRING_CONFIG,		"Self-powered"},
	{STRING_INTERFACE,	"Mass Storage"},
#endif
	{}
};

static struct usb_gadget_strings	stringtab = {
	.language	= 0x0409,		// en-us
	.strings	= strings,
};


#define DEFAULT_SUPPLIED_POWER 100
static void notify_supplied_power_enable(bool flag)
{
	int power = DEFAULT_SUPPLIED_POWER;
	bool is_self_powerd;

	is_self_powerd = config_desc.bmAttributes & USB_CONFIG_ATT_SELFPOWER;

	if( !is_self_powerd ){
		if( flag ){
			power = config_desc.bMaxPower * 2;
		}else{
			power = DEFAULT_SUPPLIED_POWER;
		}
#ifdef CONFIG_MACH_PGPF
		pgpf_usb_notify_supplied_power( flag, power );
#endif /* CONFIG_MACH_PGPF */
	}
}


/*
 * Config descriptors must agree with the code that sets configurations
 * and with code managing interfaces and their altsettings.  They must
 * also handle different speeds and other-speed requests.
 */
static int populate_config_buf(struct usb_gadget *gadget,
		u8 *buf, u8 type, unsigned index)
{
#ifdef CONFIG_USB_GADGET_DUALSPEED
	enum usb_device_speed			speed = gadget->speed;
#endif
	int					len;
	const struct usb_descriptor_header	**function;

	if (index > 0)
		return -EINVAL;

#ifdef CONFIG_USB_GADGET_DUALSPEED
	if (type == USB_DT_OTHER_SPEED_CONFIG)
		speed = (USB_SPEED_FULL + USB_SPEED_HIGH) - speed;
	if (speed == USB_SPEED_HIGH)
		function = hs_function;
	else
#endif
		function = fs_function;

	/* for now, don't advertise srp-only devices */
	if (!gadget->is_otg)
		function++;

	len = usb_gadget_config_buf(&config_desc, buf, EP0_BUFSIZE, function);
	((struct usb_config_descriptor *) buf)->bDescriptorType = type;
	return len;
}


/*-------------------------------------------------------------------------*/

/* These routines may be called in process context or in_irq */

/* Caller must hold fsg->lock */
static void wakeup_thread(struct fsg_dev *fsg)
{
	/* Tell the main thread that something has happened */
	fsg->thread_wakeup_needed = 1;
	if (fsg->thread_task)
		wake_up_process(fsg->thread_task);
}


static void raise_exception(struct fsg_dev *fsg, enum fsg_state new_state)
{
	unsigned long		flags;

	/* Do nothing if a higher-priority exception is already in progress.
	 * If a lower-or-equal priority exception is in progress, preempt it
	 * and notify the main thread by sending it a signal. */
	spin_lock_irqsave(&fsg->lock, flags);
	if (fsg->state <= new_state) {
		fsg->exception_req_tag = fsg->ep0_req_tag;
		fsg->state = new_state;
		if (fsg->thread_task)
#ifdef USB_FILE_STORAGE_NAVU
			//printk(KERN_INFO "file_storage_navu.c: raise_exception(): \n");
			sonydbg(fsg, "raise_exception : Send SIGUSR1 signal \n");
#endif
			send_sig_info(SIGUSR1, SEND_SIG_FORCED,
					fsg->thread_task);
	}
	spin_unlock_irqrestore(&fsg->lock, flags);
}


/*-------------------------------------------------------------------------*/

/* The disconnect callback and ep0 routines.  These always run in_irq,
 * except that ep0_queue() is called in the main thread to acknowledge
 * completion of various requests: set config, set interface, and
 * Bulk-only device reset. */

static void fsg_disconnect(struct usb_gadget *gadget)
{
	struct fsg_dev		*fsg = get_gadget_data(gadget);

#ifdef USB_FILE_STORAGE_NAVU
	CLEAR_STATUS;
	// ljzhou 2006/07/11
	if( 1 == g_usb_config_cplt) {
		flush_all_waitqueue(fsg);
		wakeup_thread(fsg);
		g_exit_flag = 1;
		//atomic_set(&data_send_control, 1);
		//wake_up(&SendDataQue);
		//printk(KERN_ALERT "reset the wait queue end\n");

		//Added steven 0704 start
		set_bit(RESETED, &fsg->atomic_bitflags);
		atomic_set(&data_response_control, 1);
		wake_up_all(&ResponseQue);
		atomic_set(&data_send_control, 1);
		wake_up_all(&SendDataQue);
		//Added steven 0704 end
	}


//	g_usb_config_cplt = 0;
	// ljzhou 2006/07/11
	sonydbg(fsg, "disconnect or port reset\n");
#else
	DBG(fsg, "disconnect or port reset\n");
#endif
	raise_exception(fsg, FSG_STATE_DISCONNECT);
}


static int ep0_queue(struct fsg_dev *fsg)
{
	int	rc;

	rc = usb_ep_queue(fsg->ep0, fsg->ep0req, GFP_ATOMIC);
	if (rc != 0 && rc != -ESHUTDOWN) {

		/* We can't do much more than wait for a reset */
		WARN(fsg, "error in submission: %s --> %d\n",
				fsg->ep0->name, rc);
	}
	return rc;
}

static void ep0_complete(struct usb_ep *ep, struct usb_request *req)
{
	struct fsg_dev		*fsg = ep->driver_data;

	if (req->actual > 0)
		dump_msg(fsg, fsg->ep0req_name, req->buf, req->actual);
	if (req->status || req->actual != req->length)
		DBG(fsg, "%s --> %d, %u/%u\n", __FUNCTION__,
				req->status, req->actual, req->length);
	if (req->status == -ECONNRESET)		// Request was cancelled
		usb_ep_fifo_flush(ep);

	if (req->status == 0 && req->context)
		((fsg_routine_t) (req->context))(fsg);
}


/*-------------------------------------------------------------------------*/

/* Bulk and interrupt endpoint completion handlers.
 * These always run in_irq. */

static void bulk_in_complete(struct usb_ep *ep, struct usb_request *req)
{
	struct fsg_dev		*fsg = ep->driver_data;
	struct fsg_buffhd	*bh = req->context;

	if (req->status || req->actual != req->length)
		DBG(fsg, "%s --> %d, %u/%u\n", __FUNCTION__,
				req->status, req->actual, req->length);
	if (req->status == -ECONNRESET)		// Request was cancelled
		usb_ep_fifo_flush(ep);

	/* Hold the lock while we update the request and buffer states */
	smp_wmb();
	spin_lock(&fsg->lock);
	bh->inreq_busy = 0;
	bh->state = BUF_STATE_EMPTY;
#ifdef USB_FILE_STORAGE_NAVU
	/* Richard Zhou */
	//wake_up(&SendDataQue);
	/* ??? is it right */
	newdbg(fsg, "bulk_in_complete req.len=%d\n", req->length);
	// ljzhou 2006/6/21 start
#if 0
	if(1 == g_vendor_command) {
	  up(&g_vendor_sem);
	  //printk(KERN_ALERT "up complete\n");
	}
#endif
	// ljzhou 2006/6/21 end
	wakeup_thread(fsg);
	spin_unlock(&fsg->lock);

	//Added steven 0704 start
	atomic_set(&data_response_control, 1);
	wake_up_all(&ResponseQue);
	//Added steven 0704 end
#else
	wakeup_thread(fsg);
	spin_unlock(&fsg->lock);
#endif
}

static void bulk_out_complete(struct usb_ep *ep, struct usb_request *req)
{
	struct fsg_dev		*fsg = ep->driver_data;
	struct fsg_buffhd	*bh = req->context;

	dump_msg(fsg, "bulk-out", req->buf, req->actual);
	if (req->status || req->actual != bh->bulk_out_intended_length)
		DBG(fsg, "%s --> %d, %u/%u\n", __FUNCTION__,
				req->status, req->actual,
				bh->bulk_out_intended_length);
	if (req->status == -ECONNRESET)		// Request was cancelled
		usb_ep_fifo_flush(ep);

	/* Hold the lock while we update the request and buffer states */
	smp_wmb();
	spin_lock(&fsg->lock);
	bh->outreq_busy = 0;
	bh->state = BUF_STATE_FULL;
	wakeup_thread(fsg);
	spin_unlock(&fsg->lock);
#ifdef USB_FILE_STORAGE_NAVU
	/* Richard Zhou */
	//wake_up(&SendDataQue);
	/* ??? is it right */
	sonydbg(fsg, "bulk_out_complete function call wakeup_thread \n");
#endif
}


#ifdef CONFIG_USB_FILE_STORAGE_TEST
static void intr_in_complete(struct usb_ep *ep, struct usb_request *req)
{
	struct fsg_dev		*fsg = ep->driver_data;
	struct fsg_buffhd	*bh = req->context;

	if (req->status || req->actual != req->length)
		DBG(fsg, "%s --> %d, %u/%u\n", __FUNCTION__,
				req->status, req->actual, req->length);
	if (req->status == -ECONNRESET)		// Request was cancelled
		usb_ep_fifo_flush(ep);

	/* Hold the lock while we update the request and buffer states */
	smp_wmb();
	spin_lock(&fsg->lock);
	fsg->intreq_busy = 0;
	bh->state = BUF_STATE_EMPTY;
	wakeup_thread(fsg);
	spin_unlock(&fsg->lock);
}

#else
static void intr_in_complete(struct usb_ep *ep, struct usb_request *req)
{}
#endif /* CONFIG_USB_FILE_STORAGE_TEST */


/*-------------------------------------------------------------------------*/

/* Ep0 class-specific handlers.  These always run in_irq. */

#ifdef CONFIG_USB_FILE_STORAGE_TEST
static void received_cbi_adsc(struct fsg_dev *fsg, struct fsg_buffhd *bh)
{
	struct usb_request	*req = fsg->ep0req;
	static u8		cbi_reset_cmnd[6] = {
			SC_SEND_DIAGNOSTIC, 4, 0xff, 0xff, 0xff, 0xff};

	/* Error in command transfer? */
	if (req->status || req->length != req->actual ||
			req->actual < 6 || req->actual > MAX_COMMAND_SIZE) {

		/* Not all controllers allow a protocol stall after
		 * receiving control-out data, but we'll try anyway. */
		fsg_set_halt(fsg, fsg->ep0);
		return;			// Wait for reset
	}

	/* Is it the special reset command? */
	if (req->actual >= sizeof cbi_reset_cmnd &&
			memcmp(req->buf, cbi_reset_cmnd,
				sizeof cbi_reset_cmnd) == 0) {

		/* Raise an exception to stop the current operation
		 * and reinitialize our state. */
		DBG(fsg, "cbi reset request\n");
		raise_exception(fsg, FSG_STATE_RESET);
		return;
	}

	VDBG(fsg, "CB[I] accept device-specific command\n");
	spin_lock(&fsg->lock);

	/* Save the command for later */
	if (fsg->cbbuf_cmnd_size)
		WARN(fsg, "CB[I] overwriting previous command\n");
	fsg->cbbuf_cmnd_size = req->actual;
	memcpy(fsg->cbbuf_cmnd, req->buf, fsg->cbbuf_cmnd_size);

	wakeup_thread(fsg);
	spin_unlock(&fsg->lock);
}

#else
static void received_cbi_adsc(struct fsg_dev *fsg, struct fsg_buffhd *bh)
{}
#endif /* CONFIG_USB_FILE_STORAGE_TEST */


static int class_setup_req(struct fsg_dev *fsg,
		const struct usb_ctrlrequest *ctrl)
{
	struct usb_request	*req = fsg->ep0req;
	int			value = -EOPNOTSUPP;
	u16			w_index = le16_to_cpu(ctrl->wIndex);
	u16                     w_value = le16_to_cpu(ctrl->wValue);
	u16			w_length = le16_to_cpu(ctrl->wLength);

	if (!fsg->config)
		return value;

	/* Handle Bulk-only class-specific requests */
	if (transport_is_bbb()) {
		switch (ctrl->bRequest) {

		case USB_BULK_RESET_REQUEST:
			if (ctrl->bRequestType != (USB_DIR_OUT |
					USB_TYPE_CLASS | USB_RECIP_INTERFACE))
				break;
			if (w_index != 0 || w_value != 0) {
				value = -EDOM;
				break;
			}

			/* Raise an exception to stop the current operation
			 * and reinitialize our state. */
#ifdef USB_FILE_STORAGE_NAVU
			//printk(KERN_ALERT "Here is reset command\n");
			sonydbg(fsg, "bulk reset request\n");
#endif
			raise_exception(fsg, FSG_STATE_RESET);
			value = DELAYED_STATUS;
			break;

		case USB_BULK_GET_MAX_LUN_REQUEST:
			if (ctrl->bRequestType != (USB_DIR_IN |
					USB_TYPE_CLASS | USB_RECIP_INTERFACE))
				break;
			if (w_index != 0 || w_value != 0) {
				value = -EDOM;
				break;
			}
			VDBG(fsg, "get max LUN\n");
			*(u8 *) req->buf = fsg->nluns - 1;
#ifdef USB_FILE_STORAGE_NAVU
			value = min(w_length, (u16) 1);
#else
			value = 1;
#endif
			break;
		}
	}

	/* Handle CBI class-specific requests */
	else {
		switch (ctrl->bRequest) {

		case USB_CBI_ADSC_REQUEST:
			if (ctrl->bRequestType != (USB_DIR_OUT |
					USB_TYPE_CLASS | USB_RECIP_INTERFACE))
				break;
			if (w_index != 0 || w_value != 0) {
				value = -EDOM;
				break;
			}
			if (w_length > MAX_COMMAND_SIZE) {
				value = -EOVERFLOW;
				break;
			}
			value = w_length;
			fsg->ep0req->context = received_cbi_adsc;
			break;
		}
	}

	if (value == -EOPNOTSUPP)
		VDBG(fsg,
			"unknown class-specific control req "
			"%02x.%02x v%04x i%04x l%u\n",
			ctrl->bRequestType, ctrl->bRequest,
			w_value, w_index, w_length);
	return value;
}


/*-------------------------------------------------------------------------*/

/* Ep0 standard request handlers.  These always run in_irq. */

static int standard_setup_req(struct fsg_dev *fsg,
		const struct usb_ctrlrequest *ctrl)
{
	struct usb_request	*req = fsg->ep0req;
	int			value = -EOPNOTSUPP;
	u16			w_index = le16_to_cpu(ctrl->wIndex);
	u16			w_value = le16_to_cpu(ctrl->wValue);
	u16			w_length = le16_to_cpu(ctrl->wLength);

#ifdef USB_FILE_STORAGE_NAVU
	if( FSG_STATE_DISCONNECT <= fsg->state ){
		return value;
	}
#endif

	/* Usually this just stores reply data in the pre-allocated ep0 buffer,
	 * but config change events will also reconfigure hardware. */
	switch (ctrl->bRequest) {

	case USB_REQ_GET_DESCRIPTOR:
		if (ctrl->bRequestType != (USB_DIR_IN | USB_TYPE_STANDARD |
				USB_RECIP_DEVICE))
			break;
		switch (w_value >> 8) {

		case USB_DT_DEVICE:
			VDBG(fsg, "get device descriptor\n");
#ifdef USB_FILE_STORAGE_NAVU
			value = min(w_length, (u16) sizeof device_desc);
#else
			value = sizeof device_desc;
#endif
			memcpy(req->buf, &device_desc, value);
			break;
#ifdef CONFIG_USB_GADGET_DUALSPEED
		case USB_DT_DEVICE_QUALIFIER:
			VDBG(fsg, "get device qualifier\n");
			if (!fsg->gadget->is_dualspeed)
				break;
#ifdef USB_FILE_STORAGE_NAVU
			value = min(w_length, (u16) sizeof dev_qualifier);
#else
			value = sizeof dev_qualifier;
#endif
			memcpy(req->buf, &dev_qualifier, value);
			break;

		case USB_DT_OTHER_SPEED_CONFIG:
			VDBG(fsg, "get other-speed config descriptor\n");
			if (!fsg->gadget->is_dualspeed)
				break;
			goto get_config;
#endif
		case USB_DT_CONFIG:
			VDBG(fsg, "get configuration descriptor\n");
#ifdef CONFIG_USB_GADGET_DUALSPEED
		get_config:
#endif
			value = populate_config_buf(fsg->gadget,
					req->buf,
					w_value >> 8,
						    w_value & 0xff);
#ifdef USB_FILE_STORAGE_NAVU
			if (value >= 0)
				value = min(w_length, (u16) value);
#endif
			break;

		case USB_DT_STRING:
			VDBG(fsg, "get string descriptor\n");

			/* wIndex == language code */
			value = usb_gadget_get_string(&stringtab,
					w_value & 0xff, req->buf);
#ifdef USB_FILE_STORAGE_NAVU
			if (value >= 0)
				value = min(w_length, (u16) value);
#endif
			break;
		}
		break;

	/* One config, two speeds */
	case USB_REQ_SET_CONFIGURATION:
		if (ctrl->bRequestType != (USB_DIR_OUT | USB_TYPE_STANDARD |
				USB_RECIP_DEVICE))
			break;
		VDBG(fsg, "set configuration\n");
		if (w_value == CONFIG_VALUE || w_value == 0) {
			fsg->new_config = w_value;

			/* Raise an exception to wipe out previous transaction
			 * state (queued bufs, etc) and set the new config. */
			raise_exception(fsg, FSG_STATE_CONFIG_CHANGE);
			value = DELAYED_STATUS;
		}
		break;
	case USB_REQ_GET_CONFIGURATION:
		if (ctrl->bRequestType != (USB_DIR_IN | USB_TYPE_STANDARD |
				USB_RECIP_DEVICE))
			break;
		VDBG(fsg, "get configuration\n");
		*(u8 *) req->buf = fsg->config;
#ifdef USB_FILE_STORAGE_NAVU
		value = min(w_length, (u16) 1);
#else
		value = 1;
#endif
		break;

	case USB_REQ_SET_INTERFACE:
		if (ctrl->bRequestType != (USB_DIR_OUT| USB_TYPE_STANDARD |
				USB_RECIP_INTERFACE))
			break;
		if (fsg->config && w_index == 0) {

			/* Raise an exception to wipe out previous transaction
			 * state (queued bufs, etc) and install the new
			 * interface altsetting. */
			raise_exception(fsg, FSG_STATE_INTERFACE_CHANGE);
			value = DELAYED_STATUS;
		}
		break;
	case USB_REQ_GET_INTERFACE:
		if (ctrl->bRequestType != (USB_DIR_IN | USB_TYPE_STANDARD |
				USB_RECIP_INTERFACE))
			break;
		if (!fsg->config)
			break;
		if (w_index != 0) {
			value = -EDOM;
			break;
		}
		VDBG(fsg, "get interface\n");
		*(u8 *) req->buf = 0;
#ifdef USB_FILE_STORAGE_NAVU
		value = min(w_length, (u16) 1);
#else
		value = 1;
#endif
		break;

	default:
		VDBG(fsg,
			"unknown control req %02x.%02x v%04x i%04x l%u\n",
			ctrl->bRequestType, ctrl->bRequest,
			w_value, w_index, w_length);
	}

	return value;
}


static int fsg_setup(struct usb_gadget *gadget,
		const struct usb_ctrlrequest *ctrl)
{
	struct fsg_dev		*fsg = get_gadget_data(gadget);
	int			rc;
	int			w_length = le16_to_cpu(ctrl->wLength);

	++fsg->ep0_req_tag;		// Record arrival of a new request
	fsg->ep0req->context = NULL;
	fsg->ep0req->length = 0;
	dump_msg(fsg, "ep0-setup", (u8 *) ctrl, sizeof(*ctrl));

	if ((ctrl->bRequestType & USB_TYPE_MASK) == USB_TYPE_CLASS)
		rc = class_setup_req(fsg, ctrl);
	else
		rc = standard_setup_req(fsg, ctrl);

	/* Respond with data/status or defer until later? */
	if (rc >= 0 && rc != DELAYED_STATUS) {
		rc = min(rc, w_length);
		fsg->ep0req->length = rc;
#ifdef USB_FILE_STORAGE_NAVU
		fsg->ep0req->zero = (rc < w_length &&
				(rc % gadget->ep0->maxpacket) == 0);
#else
		fsg->ep0req->zero = rc < w_length;
#endif
		fsg->ep0req_name = (ctrl->bRequestType & USB_DIR_IN ?
				"ep0-in" : "ep0-out");
		rc = ep0_queue(fsg);
	}

	/* Device either stalls (rc < 0) or reports success */
	return rc;
}


/*-------------------------------------------------------------------------*/

/* All the following routines run in process context */


/* Use this for bulk or interrupt transfers, not ep0 */
static void start_transfer(struct fsg_dev *fsg, struct usb_ep *ep,
		struct usb_request *req, int *pbusy,
		enum fsg_buffer_state *state)
{
	int	rc;

	if (ep == fsg->bulk_in)
		dump_msg(fsg, "bulk-in", req->buf, req->length);
	else if (ep == fsg->intr_in)
		dump_msg(fsg, "intr-in", req->buf, req->length);

	spin_lock_irq(&fsg->lock);
	*pbusy = 1;
	*state = BUF_STATE_BUSY;
	spin_unlock_irq(&fsg->lock);
	rc = usb_ep_queue(ep, req, GFP_KERNEL);
	if (rc != 0) {
		*pbusy = 0;
		*state = BUF_STATE_EMPTY;

		/* We can't do much more than wait for a reset */

		/* Note: currently the net2280 driver fails zero-length
		 * submissions if DMA is enabled. */
		if (rc != -ESHUTDOWN && !(rc == -EOPNOTSUPP &&
						req->length == 0))
#ifdef USB_FILE_STORAGE_NAVU
			newdbg(fsg, "error in submission: %s --> %d\n",
					ep->name, rc);
	}
	else {
		newdbg(fsg, "succ start_transfer req.len=%d\n", req->length);
	}
#else
			WARN(fsg, "error in submission: %s --> %d\n",
					ep->name, rc);
	}
#endif

}


static int sleep_thread(struct fsg_dev *fsg)
{
	int	rc = 0;

	/* Wait until a signal arrives or we are woken up */
	for (;;) {
		try_to_freeze();
		set_current_state(TASK_INTERRUPTIBLE);
		if (signal_pending(current)) {
			rc = -EINTR;
			break;
		}
		if (fsg->thread_wakeup_needed)
			break;
		schedule();
	}
	__set_current_state(TASK_RUNNING);
	fsg->thread_wakeup_needed = 0;
	/* if a suspend happened return non zero let wait function return */
#ifdef USB_FILE_STORAGE_NAVU
	if (test_bit(SUSPENDED,&fsg->atomic_bitflags ))
		return -EINTR;
#endif
	return rc;
}


#ifdef USB_FILE_STORAGE_NAVU
// function name has been changed 2008/06/10
static int check_media_memorystick(struct fsg_dev *fsg)
{
	if( fsg == NULL ) {
		return -EINVAL;
	}

	if( strncmp( mod_data.file[fsg->lun], "/dev/mshc", 10 ) == 0 ) {
		return 0;	// found MemoryStick
	}
	return -ENODEV;	// not found.
}

static int fsync_sub(struct lun *curlun);
static void truncate_sub(struct lun *curlun);
static int check_media_insertion(struct fsg_dev *fsg)
{
	int rc = 0;
	static struct gendisk *disk;  // aaded for improving to check media insertion. 20080626.

	if( Mount_disable_flag ) {
		if( !( (fsg->curlun)
		    && (fsg->curlun->filp)
		    && (fsg->curlun->filp->f_dentry)
		    && (fsg->curlun->filp->f_dentry->d_inode)
		    && (fsg->curlun->filp->f_dentry->d_inode->i_bdev) ) ) {
			return 0;	// Nothing to do here.
		}

		fsg->curlun->prevent_medium_removal = 0;
		fsg->curlun->sense_data = SS_MEDIUM_NOT_PRESENT;
		return -EINVAL;
	}

	if( check_media_memorystick(fsg) == 0 ) {
		// when media is MemoryStick.
		if( !( (fsg->curlun)
		    && (fsg->curlun->filp)
		    && (fsg->curlun->filp->f_dentry)
		    && (fsg->curlun->filp->f_dentry->d_inode)
		    && (fsg->curlun->filp->f_dentry->d_inode->i_bdev) ) ) {
			return 0;	// Nothing to do here.
		}


#if 1   // changed for improving to check media insertion. 20080626.
		disk = fsg->curlun->filp->f_dentry->d_inode->i_bdev->bd_disk;
		if( k3d_get_disk_insert(disk, 0) != 1 ) {
#else
		if( check_disk_change(fsg->curlun->filp->f_dentry->d_inode->i_bdev) != 0 ) {
#endif
			// when media removed.
			fsg->curlun->prevent_medium_removal = 0;
			fsg->curlun->ro = mod_data.ro[fsg->lun];
			fsg->curlun->sense_data = SS_MEDIUM_NOT_PRESENT;

			truncate_sub(fsg->curlun);

			filp_close(fsg->curlun->filp, current->files);
			fsg->curlun->filp = NULL;

			return -EINVAL;
		}
	}
	return rc;
}

static int check_media_change(struct fsg_dev *fsg)
{
	struct gendisk *disk;

	if (check_media_memorystick(fsg) == 0) {
		// when media is MemoryStick.
		if (!((fsg->curlun)
		    && (fsg->curlun->filp)
		    && (fsg->curlun->filp->f_path.dentry)
		    && (fsg->curlun->filp->f_path.dentry->d_inode)
		    && (fsg->curlun->filp->f_path.dentry->d_inode->i_bdev)
		    && (fsg->curlun->filp->f_path.dentry->d_inode->i_bdev->bd_disk))) {
			return 0;
		}

		disk = fsg->curlun->filp->f_path.dentry->d_inode->i_bdev->bd_disk;
		return k3d_get_disk_change(disk, 0);
	}

	return 0;
}
#endif // USB_FILE_STORAGE_NAVU

/*-------------------------------------------------------------------------*/

static int do_read(struct fsg_dev *fsg)
{
	struct lun		*curlun = fsg->curlun;
	u32			lba;
	struct fsg_buffhd	*bh;
	int			rc;
	u32			amount_left;
	loff_t			file_offset, file_offset_tmp;
	unsigned int		amount;
	unsigned int		partial_page;
	ssize_t			nread;

	/* Get the starting Logical Block Address and check that it's
	 * not too big */
	if (fsg->cmnd[0] == SC_READ_6)
		lba = (fsg->cmnd[1] << 16) | get_be16(&fsg->cmnd[2]);
	else {
		lba = get_be32(&fsg->cmnd[2]);

		/* We allow DPO (Disable Page Out = don't save data in the
		 * cache) and FUA (Force Unit Access = don't read from the
		 * cache), but we don't implement them. */
		if ((fsg->cmnd[1] & ~0x18) != 0) {
			curlun->sense_data = SS_INVALID_FIELD_IN_CDB;
			return -EINVAL;
		}
	}
	if (lba >= curlun->num_sectors) {
		curlun->sense_data = SS_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE;
		return -EINVAL;
	}
	file_offset = ((loff_t) lba) << 9;

	/* Carry out the file reads */
	amount_left = fsg->data_size_from_cmnd;
	if (unlikely(amount_left == 0))
		return -EIO;		// No default reply

	for (;;) {

		/* Figure out how much we need to read:
		 * Try to read the remaining amount.
		 * But don't read more than the buffer size.
		 * And don't try to read past the end of the file.
		 * Finally, if we're not at a page boundary, don't read past
		 *	the next page.
		 * If this means reading 0 then we were asked to read past
		 *	the end of file. */
		amount = min((unsigned int) amount_left, mod_data.buflen);
		amount = min((loff_t) amount,
				curlun->file_length - file_offset);
		partial_page = file_offset & (PAGE_CACHE_SIZE - 1);
		if (partial_page > 0)
			amount = min(amount, (unsigned int) PAGE_CACHE_SIZE -
					partial_page);

		/* Wait for the next buffer to become available */
		bh = fsg->next_buffhd_to_fill;
		while (bh->state != BUF_STATE_EMPTY) {
			if ((rc = sleep_thread(fsg)) != 0)
				return rc;
		}

		/* If we were asked to read past the end of file,
		 * end with an empty buffer. */
		if (amount == 0) {
			curlun->sense_data =
					SS_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE;
			curlun->sense_data_info = file_offset >> 9;
			curlun->info_valid = 1;
			bh->inreq->length = 0;
			bh->state = BUF_STATE_FULL;
			break;
		}

		/* Perform the read */
		file_offset_tmp = file_offset;
		nread = vfs_read(curlun->filp,
				(char __user *) bh->buf,
				amount, &file_offset_tmp);
		VLDBG(curlun, "file read %u @ %llu -> %d\n", amount,
				(unsigned long long) file_offset,
				(int) nread);
		if (signal_pending(current))
			return -EINTR;

		if (nread < 0) {
			LDBG(curlun, "error in file read: %d\n",
					(int) nread);
			nread = 0;
		} else if (nread < amount) {
			LDBG(curlun, "partial file read: %d/%u\n",
					(int) nread, amount);
			nread -= (nread & 511);	// Round down to a block
		}
		file_offset  += nread;
		amount_left  -= nread;
		fsg->residue -= nread;
		bh->inreq->length = nread;
		bh->state = BUF_STATE_FULL;

		/* If an error occurred, report it and its position */
		if (nread < amount) {
			curlun->sense_data = SS_UNRECOVERED_READ_ERROR;
			curlun->sense_data_info = file_offset >> 9;
			curlun->info_valid = 1;
#ifdef USB_FILE_STORAGE_NAVU
			close_backing_file(curlun);
#endif
			break;
		}

		if (amount_left == 0)
			break;		// No more left to read

		/* Send this buffer and go read some more */
		bh->inreq->zero = 0;
		start_transfer(fsg, fsg->bulk_in, bh->inreq,
				&bh->inreq_busy, &bh->state);
		fsg->next_buffhd_to_fill = bh->next;
	}

	return -EIO;		// No default reply
}


/*-------------------------------------------------------------------------*/

static int do_write(struct fsg_dev *fsg)
{
	struct lun		*curlun = fsg->curlun;
	u32			lba;
	struct fsg_buffhd	*bh;
	int			get_some_more;
	u32			amount_left_to_req, amount_left_to_write;
	loff_t			usb_offset, file_offset, file_offset_tmp;
	unsigned int		amount;
	unsigned int		partial_page;
	ssize_t			nwritten;
	int			rc;

	if (curlun->ro) {
		curlun->sense_data = SS_WRITE_PROTECTED;
		return -EINVAL;
	}
	curlun->filp->f_flags &= ~O_SYNC;	// Default is not to wait

	/* Get the starting Logical Block Address and check that it's
	 * not too big */
	if (fsg->cmnd[0] == SC_WRITE_6)
		lba = (fsg->cmnd[1] << 16) | get_be16(&fsg->cmnd[2]);
	else {
		lba = get_be32(&fsg->cmnd[2]);

		/* We allow DPO (Disable Page Out = don't save data in the
		 * cache) and FUA (Force Unit Access = write directly to the
		 * medium).  We don't implement DPO; we implement FUA by
		 * performing synchronous output. */
		if ((fsg->cmnd[1] & ~0x18) != 0) {
			curlun->sense_data = SS_INVALID_FIELD_IN_CDB;
			return -EINVAL;
		}
		if (fsg->cmnd[1] & 0x08)	// FUA
			curlun->filp->f_flags |= O_SYNC;
	}
	if (lba >= curlun->num_sectors) {
		curlun->sense_data = SS_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE;
		return -EINVAL;
	}

	/* Carry out the file writes */
	get_some_more = 1;
	file_offset = usb_offset = ((loff_t) lba) << 9;
	amount_left_to_req = amount_left_to_write = fsg->data_size_from_cmnd;

#ifdef USB_FILE_STORAGE_NAVU
	sonydbg(fsg, "do_write function begin \n");
#endif

	while (amount_left_to_write > 0) {

		/* Queue a request for more data from the host */
		bh = fsg->next_buffhd_to_fill;
		if (bh->state == BUF_STATE_EMPTY && get_some_more) {

			/* Figure out how much we want to get:
			 * Try to get the remaining amount.
			 * But don't get more than the buffer size.
			 * And don't try to go past the end of the file.
			 * If we're not at a page boundary,
			 *	don't go past the next page.
			 * If this means getting 0, then we were asked
			 *	to write past the end of file.
			 * Finally, round down to a block boundary. */
			amount = min(amount_left_to_req, mod_data.buflen);
			amount = min((loff_t) amount, curlun->file_length -
					usb_offset);
			partial_page = usb_offset & (PAGE_CACHE_SIZE - 1);
			if (partial_page > 0)
				amount = min(amount,
	(unsigned int) PAGE_CACHE_SIZE - partial_page);

			if (amount == 0) {
				get_some_more = 0;
				curlun->sense_data =
					SS_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE;
				curlun->sense_data_info = usb_offset >> 9;
				curlun->info_valid = 1;
				continue;
			}
			amount -= (amount & 511);
			if (amount == 0) {

				/* Why were we were asked to transfer a
				 * partial block? */
				get_some_more = 0;
				continue;
			}

			/* Get the next buffer */
			usb_offset += amount;
			fsg->usb_amount_left -= amount;
			amount_left_to_req -= amount;
			if (amount_left_to_req == 0)
				get_some_more = 0;

			/* amount is always divisible by 512, hence by
			 * the bulk-out maxpacket size */
			bh->outreq->length = bh->bulk_out_intended_length =
					amount;
			bh->outreq->short_not_ok = 1;
			start_transfer(fsg, fsg->bulk_out, bh->outreq,
					&bh->outreq_busy, &bh->state);
			fsg->next_buffhd_to_fill = bh->next;
			continue;
		}

		/* Write the received data to the backing file */
		bh = fsg->next_buffhd_to_drain;
		if (bh->state == BUF_STATE_EMPTY && !get_some_more)
			break;			// We stopped early
		if (bh->state == BUF_STATE_FULL) {
			smp_rmb();
			fsg->next_buffhd_to_drain = bh->next;
			bh->state = BUF_STATE_EMPTY;

#ifdef USB_FILE_STORAGE_NAVU
			sonydbg(fsg, "COMMON : buffer state = %d \n", bh->state);
			sonydbg(fsg, "COMMON : atm_left_to_write = %d \n", atomic_read(&atm_left_to_write));
			sonydbg(fsg, "COMMON :fsg->usb_amount_left = %d \n", fsg->usb_amount_left);
			sonydbg(fsg, "COMMON :amount_left_to_req = %d \n", amount_left_to_req);
#endif

			/* Did something go wrong with the transfer? */
			if (bh->outreq->status != 0) {
				curlun->sense_data = SS_COMMUNICATION_FAILURE;
				curlun->sense_data_info = file_offset >> 9;
				curlun->info_valid = 1;
				break;
			}

			amount = bh->outreq->actual;
			if (curlun->file_length - file_offset < amount) {
				LERROR(curlun,
	"write %u @ %llu beyond end %llu\n",
	amount, (unsigned long long) file_offset,
	(unsigned long long) curlun->file_length);
				amount = curlun->file_length - file_offset;
			}

			/* Perform the write */
			file_offset_tmp = file_offset;
			nwritten = vfs_write(curlun->filp,
					(char __user *) bh->buf,
					amount, &file_offset_tmp);
			VLDBG(curlun, "file write %u @ %llu -> %d\n", amount,
					(unsigned long long) file_offset,
					(int) nwritten);
			if (signal_pending(current))
				return -EINTR;		// Interrupted!

			if (nwritten < 0) {
				LDBG(curlun, "error in file write: %d\n",
						(int) nwritten);
				nwritten = 0;
			} else if (nwritten < amount) {
				LDBG(curlun, "partial file write: %d/%u\n",
						(int) nwritten, amount);
				nwritten -= (nwritten & 511);
						// Round down to a block
			}
			file_offset += nwritten;
			amount_left_to_write -= nwritten;
			fsg->residue -= nwritten;

			/* If an error occurred, report it and its position */
			if (nwritten < amount) {
				curlun->sense_data = SS_WRITE_ERROR;
				curlun->sense_data_info = file_offset >> 9;
				curlun->info_valid = 1;
#ifdef USB_FILE_STORAGE_NAVU
				close_backing_file(curlun);
#endif
				break;
			}

			/* Did the host decide to stop early? */
			if (bh->outreq->actual != bh->outreq->length) {
				fsg->short_packet_received = 1;
				break;
			}
			continue;
		}
#ifdef USB_FILE_STORAGE_NAVU
#ifdef SONY_DEBUG
		else {
			sonydbg(fsg, "here is unexpected situation \n");
			sonydbg(fsg, "buffer state = %d \n", bh->state);
			sonydbg(fsg, "atm_left_to_write = %d \n", atomic_read(&atm_left_to_write));
			sonydbg(fsg, "fsg->usb_amount_left = %d \n", fsg->usb_amount_left);
			sonydbg(fsg, "amount_left_to_req = %d \n", amount_left_to_req);
		}
#endif /* SONY_DEBUG */
#endif
		/* Wait for something to happen */
		if ((rc = sleep_thread(fsg)) != 0)
			return rc;
	}
#ifdef USB_FILE_STORAGE_NAVU
    sonydbg(fsg, "exit the do_write function begin \n");
#endif
	if( fsg->data_size_from_cmnd <= 4096 ){
		fsync_sub(curlun);
	}
	return -EIO;		// No default reply
}


/*-------------------------------------------------------------------------*/

/* Sync the file data, don't bother with the metadata.
 * This code was copied from fs/buffer.c:sys_fdatasync(). */
static int fsync_sub(struct lun *curlun)
{
	struct file	*filp = curlun->filp;
	struct inode	*inode;
	int		rc, err;

	if (curlun->ro || !filp)
		return 0;
	if (!filp->f_op->fsync)
		return -EINVAL;

	inode = filp->f_path.dentry->d_inode;
	mutex_lock(&inode->i_mutex);
	rc = filemap_fdatawrite(inode->i_mapping);
	err = filp->f_op->fsync(filp, filp->f_path.dentry, 1);
	if (!rc)
		rc = err;
	err = filemap_fdatawait(inode->i_mapping);
	if (!rc)
		rc = err;
	mutex_unlock(&inode->i_mutex);
	VLDBG(curlun, "fdatasync -> %d\n", rc);
	return rc;
}

static void fsync_all(struct fsg_dev *fsg)
{
	int	i;

	for (i = 0; i < fsg->nluns; ++i)
		fsync_sub(&fsg->luns[i]);
}

static int do_synchronize_cache(struct fsg_dev *fsg)
{
	struct lun	*curlun = fsg->curlun;
	int		rc;

	/* We ignore the requested LBA and write out all file's
	 * dirty data buffers. */
	rc = fsync_sub(curlun);
	if (rc)
		curlun->sense_data = SS_WRITE_ERROR;
	return 0;
}


/*-------------------------------------------------------------------------*/

static void invalidate_sub(struct lun *curlun)
{
	struct file	*filp = curlun->filp;
	struct inode	*inode = filp->f_path.dentry->d_inode;
	unsigned long	rc;

	rc = invalidate_mapping_pages(inode->i_mapping, 0, -1);
	VLDBG(curlun, "invalidate_inode_pages -> %ld\n", rc);
}

static int do_verify(struct fsg_dev *fsg)
{
	struct lun		*curlun = fsg->curlun;
	u32			lba;
	u32			verification_length;
	struct fsg_buffhd	*bh = fsg->next_buffhd_to_fill;
	loff_t			file_offset, file_offset_tmp;
	u32			amount_left;
	unsigned int		amount;
	ssize_t			nread;

	/* Get the starting Logical Block Address and check that it's
	 * not too big */
	lba = get_be32(&fsg->cmnd[2]);
	if (lba >= curlun->num_sectors) {
		curlun->sense_data = SS_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE;
		return -EINVAL;
	}

	/* We allow DPO (Disable Page Out = don't save data in the
	 * cache) but we don't implement it. */
	if ((fsg->cmnd[1] & ~0x10) != 0) {
		curlun->sense_data = SS_INVALID_FIELD_IN_CDB;
		return -EINVAL;
	}

	verification_length = get_be16(&fsg->cmnd[7]);
	if (unlikely(verification_length == 0))
		return -EIO;		// No default reply

	/* Prepare to carry out the file verify */
	amount_left = verification_length << 9;
	file_offset = ((loff_t) lba) << 9;

	/* Write out all the dirty buffers before invalidating them */
	fsync_sub(curlun);
	if (signal_pending(current))
		return -EINTR;

	invalidate_sub(curlun);
	if (signal_pending(current))
		return -EINTR;

	/* Just try to read the requested blocks */
	while (amount_left > 0) {

		/* Figure out how much we need to read:
		 * Try to read the remaining amount, but not more than
		 * the buffer size.
		 * And don't try to read past the end of the file.
		 * If this means reading 0 then we were asked to read
		 * past the end of file. */
		amount = min((unsigned int) amount_left, mod_data.buflen);
		amount = min((loff_t) amount,
				curlun->file_length - file_offset);
		if (amount == 0) {
			curlun->sense_data =
					SS_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE;
			curlun->sense_data_info = file_offset >> 9;
			curlun->info_valid = 1;
			break;
		}

		/* Perform the read */
		file_offset_tmp = file_offset;
		nread = vfs_read(curlun->filp,
				(char __user *) bh->buf,
				amount, &file_offset_tmp);
		VLDBG(curlun, "file read %u @ %llu -> %d\n", amount,
				(unsigned long long) file_offset,
				(int) nread);
		if (signal_pending(current))
			return -EINTR;

		if (nread < 0) {
			LDBG(curlun, "error in file verify: %d\n",
					(int) nread);
			nread = 0;
		} else if (nread < amount) {
			LDBG(curlun, "partial file verify: %d/%u\n",
					(int) nread, amount);
			nread -= (nread & 511);	// Round down to a sector
		}
		if (nread == 0) {
			curlun->sense_data = SS_UNRECOVERED_READ_ERROR;
			curlun->sense_data_info = file_offset >> 9;
			curlun->info_valid = 1;
			break;
		}
		file_offset += nread;
		amount_left -= nread;
	}
	return 0;
}


/*-------------------------------------------------------------------------*/

#ifdef USB_FILE_STORAGE_NAVU
u8 ms_vendor_specific_inquiry_data[] =
{
	0x4D, 0x45, 0x4D, 0x4F, 0x52, 0x59, 0x53, 0x54, 0x49, 0x43,
	0x4B, 0x2D, 0x4D, 0x47, 0x03, 0x00, 0x00, 0x20, 0x20, 0x20
};
#endif
static int do_inquiry(struct fsg_dev *fsg, struct fsg_buffhd *bh)
{
	u8	*buf = (u8 *) bh->buf;
#ifdef USB_FILE_STORAGE_NAVU
	char product[32];
	int  length;
	u32 vender_data_field_size;
#endif

#ifdef USB_FILE_STORAGE_NAVU
#else
	static char vendor_id[] = "Linux   ";
	static char product_id[] = "File-Stor Gadget";
#endif

	if (!fsg->curlun) {		// Unsupported LUNs are okay
		fsg->bad_lun_okay = 1;
		memset(buf, 0, 36);
		buf[0] = 0x7f;		// Unsupported, no device-type
		return 36;
	}

	memset(buf, 0, 8);	// Non-removable, direct-access device
	if (mod_data.removable)
		buf[1] = 0x80;
	buf[2] = 2;		// ANSI SCSI level 2
	buf[3] = 2;		// SCSI-2 INQUIRY data format
	buf[4] = 31;		// Additional length
				// No special options

#ifdef USB_FILE_STORAGE_NAVU
#define START_INDEX_VENDOR_ID 8

	memset(product, 0, sizeof(product));
	length = strlen(mod_data.product_string);
	length = (length > 10) ? 10 : length;
	memcpy(product, mod_data.product_string, length);

	if( check_media_memorystick(fsg) == 0 ) {
		snprintf(product+length, sizeof(product)-length, "(M.S.)");
	}
	product[16] = 0x00;	// only use upper 16byte;

	snprintf(buf + START_INDEX_VENDOR_ID, 36 - START_INDEX_VENDOR_ID, "%-8s%-16s%04x",
			mod_data.vendor_string,
			product,
			mod_data.release);

	if( check_media_memorystick(fsg) == 0 ) {
		vender_data_field_size = fsg->data_size_from_cmnd - 36;
		buf[4] = 31 + sizeof(ms_vendor_specific_inquiry_data);		// Additional length
		if( vender_data_field_size < sizeof(ms_vendor_specific_inquiry_data) ) {
			memcpy(&buf[36], ms_vendor_specific_inquiry_data, vender_data_field_size);
		}else{
			memcpy(&buf[36], ms_vendor_specific_inquiry_data, sizeof(ms_vendor_specific_inquiry_data));
		}
	}

	return buf[4] + 5;
#else
	sprintf(buf + 8, "%-8s%-16s%04x",
			vendor_id,
			product_id,
			mod_data.release);

	return 36;
#endif
}


static int do_request_sense(struct fsg_dev *fsg, struct fsg_buffhd *bh)
{
	struct lun	*curlun = fsg->curlun;
	u8		*buf = (u8 *) bh->buf;
	u32		sd, sdinfo;
	int		valid;

	/*
	 * From the SCSI-2 spec., section 7.9 (Unit attention condition):
	 *
	 * If a REQUEST SENSE command is received from an initiator
	 * with a pending unit attention condition (before the target
	 * generates the contingent allegiance condition), then the
	 * target shall either:
	 *   a) report any pending sense data and preserve the unit
	 *	attention condition on the logical unit, or,
	 *   b) report the unit attention condition, may discard any
	 *	pending sense data, and clear the unit attention
	 *	condition on the logical unit for that initiator.
	 *
	 * FSG normally uses option a); enable this code to use option b).
	 */
#if 0
	if (curlun && curlun->unit_attention_data != SS_NO_SENSE) {
		curlun->sense_data = curlun->unit_attention_data;
		curlun->unit_attention_data = SS_NO_SENSE;
	}
#endif

	if (!curlun) {		// Unsupported LUNs are okay
		fsg->bad_lun_okay = 1;
		sd = SS_LOGICAL_UNIT_NOT_SUPPORTED;
		sdinfo = 0;
		valid = 0;
	} else {
#ifdef USB_FILE_STORAGE_NAVU
		if( curlun->is_fake_sense_data ){
			sd = SS_NOT_READY_TO_READY_TRANSITION;
			curlun->is_fake_sense_data = false;
		}else{
			sd = curlun->sense_data;
		}
#else /* USB_FILE_STORAGE_NAVU */
		sd = curlun->sense_data;
#endif /* USB_FILE_STORAGE_NAVU */
		sdinfo = curlun->sense_data_info;
		valid = curlun->info_valid << 7;
		curlun->sense_data = SS_NO_SENSE;
		curlun->sense_data_info = 0;
		curlun->info_valid = 0;
	}

	memset(buf, 0, 18);
	buf[0] = valid | 0x70;			// Valid, current error
	buf[2] = SK(sd);
	put_be32(&buf[3], sdinfo);		// Sense information
	buf[7] = 18 - 8;			// Additional sense length
	buf[12] = ASC(sd);
	buf[13] = ASCQ(sd);
	return 18;
}


static int do_read_capacity(struct fsg_dev *fsg, struct fsg_buffhd *bh)
{
	struct lun	*curlun = fsg->curlun;
	u32		lba = get_be32(&fsg->cmnd[2]);
	int		pmi = fsg->cmnd[8];
	u8		*buf = (u8 *) bh->buf;

	/* Check the PMI and LBA fields */
	if (pmi > 1 || (pmi == 0 && lba != 0)) {
		curlun->sense_data = SS_INVALID_FIELD_IN_CDB;
		return -EINVAL;
	}

	put_be32(&buf[0], curlun->num_sectors - 1);	// Max logical block
	put_be32(&buf[4], 512);				// Block length

#ifdef USB_FILE_STORAGE_NAVU
	if( check_media_memorystick(fsg) == 0 ) {
		if( check_media_insertion(fsg) != 0 ) {
			put_be32(&buf[0], 0);	// Max logical block
			put_be32(&buf[4], 0);	// Block length
		}
	}
#endif
	return 8;
}


static int do_mode_sense(struct fsg_dev *fsg, struct fsg_buffhd *bh)
{
	struct lun	*curlun = fsg->curlun;
	int		mscmnd = fsg->cmnd[0];
	u8		*buf = (u8 *) bh->buf;
	u8		*buf0 = buf;
	int		pc, page_code, req_page_code;
	int		changeable_values, all_pages;
	int		valid_page = 0;
	int		len, limit;
#ifdef USB_FILE_STORAGE_NAVU
	static char	media_type;
	bool	is_get_media_type = false;
	int		index;
	int		data_len;
#endif /* USB_FILE_STORAGE_NAVU */

	if ((fsg->cmnd[1] & ~0x08) != 0) {		// Mask away DBD
		curlun->sense_data = SS_INVALID_FIELD_IN_CDB;
		return -EINVAL;
	}
	pc = fsg->cmnd[2] >> 6;
	page_code = fsg->cmnd[2] & 0x3f;
	if (pc == 3) {
		curlun->sense_data = SS_SAVING_PARAMETERS_NOT_SUPPORTED;
		return -EINVAL;
	}
	changeable_values = (pc == 1);
	all_pages = (page_code == 0x3f);

#ifdef USB_FILE_STORAGE_NAVU
	if ( ( page_code == 0x00 )
		|| ( ( 0x20 <= page_code ) && ( page_code < 0x3f ) )
		|| all_pages ) {
		if ( check_media_type_handler(fsg) ) {
			is_get_media_type = get_media_type( fsg, &media_type );
		}
	}
#endif /* USB_FILE_STORAGE_NAVU */

	/* Write the mode parameter header.  Fixed values are: default
	 * medium type, no cache control (DPOFUA), and no block descriptors.
	 * The only variable value is the WriteProtect bit.  We will fill in
	 * the mode data length later. */
	memset(buf, 0, 8);
	if (mscmnd == SC_MODE_SENSE_6) {
#ifdef USB_FILE_STORAGE_NAVU
		if ( is_get_media_type ) {
			buf[1] = (u8)media_type;
		}
#endif /* USB_FILE_STORAGE_NAVU */
		buf[2] = (curlun->ro ? 0x80 : 0x00);		// WP, DPOFUA
		buf += 4;
		limit = 255;
	} else {			// SC_MODE_SENSE_10
#ifdef USB_FILE_STORAGE_NAVU
		if ( is_get_media_type ) {
			buf[2] = (u8)media_type;
		}
#endif /* USB_FILE_STORAGE_NAVU */
		buf[3] = (curlun->ro ? 0x80 : 0x00);		// WP, DPOFUA
		buf += 8;
#ifdef USB_FILE_STORAGE_NAVU
		limit = mod_data.buflen;		// Should really be mod_data.buflen
#else
		limit = 65535;		// Should really be mod_data.buflen
#endif
	}

	/* No block descriptors */

#ifdef USB_FILE_STORAGE_NAVU
	/* The mode pages, in numerical order.
	 * The following page has been supprted.
	 *   Caching page         : 0x08
	 *   Vender specific page : 0x20 */
	if (page_code == 0x00 || all_pages) {
		req_page_code = 0x00;
		index = check_parm_info_handler(fsg, req_page_code);
		if ( ( 0 <= index ) && ( index < MAX_PARM_INFO_HANDLER_NS ) ) {
			data_len = get_parm_info_data(fsg, index, &buf[2]);
			if ( 0 < data_len ) {
				buf[0] = req_page_code;		// Page code
				buf[1] = data_len;	// Page length
				buf += data_len + 2;
				valid_page = 1;
			}
		}
	}

	if (page_code == 0x08 || all_pages) {
		valid_page = 1;
		buf[0] = 0x08;		// Page code
		buf[1] = 10;		// Page length
		memset(buf+2, 0, 10);	// None of the fields are changeable

		if (!changeable_values) {
			buf[2] = 0x04;	// Write cache enable,
					// Read cache not disabled
					// No cache retention priorities
			put_be16(&buf[4], 0xffff);  // Don't disable prefetch
					// Minimum prefetch = 0
			put_be16(&buf[8], 0xffff);  // Maximum prefetch
			put_be16(&buf[10], 0xffff); // Maximum prefetch ceiling
		}
		buf += 12;
	}

	if ( ( 0x20 <= page_code ) && ( page_code < 0x3f ) ) {
		req_page_code = page_code;
		index = check_parm_info_handler(fsg, req_page_code);
		if ( ( 0 <= index ) && ( index < MAX_PARM_INFO_HANDLER_NS ) ) {
			data_len = get_parm_info_data(fsg, index, &buf[2]);
			if ( 0 < data_len ) {
				buf[0] = req_page_code;		// Page code
				buf[1] = data_len;	// Page length
				buf += data_len + 2;
				valid_page = 1;
			}
		}
	} else if ( all_pages ) {
		for ( req_page_code = 0x20 ; req_page_code < 0x3f ; req_page_code++ ){
			index = check_parm_info_handler(fsg, req_page_code);
			if ( ( 0 <= index ) && ( index < MAX_PARM_INFO_HANDLER_NS ) ) {
				data_len = get_parm_info_data(fsg, index, &buf[2]);
				if ( 0 < data_len ) {
					buf[0] = req_page_code;		// Page code
					buf[1] = data_len;	// Page length
					buf += data_len + 2;
					valid_page = 1;
				}
			}
		}
	}
#else
	/* The mode pages, in numerical order.  The only page we support
	 * is the Caching page. */
	if (page_code == 0x08 || all_pages) {
		valid_page = 1;
		buf[0] = 0x08;		// Page code
		buf[1] = 10;		// Page length
		memset(buf+2, 0, 10);	// None of the fields are changeable

		if (!changeable_values) {
			buf[2] = 0x04;	// Write cache enable,
					// Read cache not disabled
					// No cache retention priorities
			put_be16(&buf[4], 0xffff);  // Don't disable prefetch
					// Minimum prefetch = 0
			put_be16(&buf[8], 0xffff);  // Maximum prefetch
			put_be16(&buf[10], 0xffff); // Maximum prefetch ceiling
		}
		buf += 12;
	}
#endif /* USB_FILE_STORAGE_NAVU */

	/* Check that a valid page was requested and the mode data length
	 * isn't too long. */
	len = buf - buf0;
	if (!valid_page || len > limit) {
		curlun->sense_data = SS_INVALID_FIELD_IN_CDB;
		return -EINVAL;
	}

	/*  Store the mode data length */
	if (mscmnd == SC_MODE_SENSE_6)
		buf0[0] = len - 1;
	else
		put_be16(buf0, len - 2);
	return len;
}


static int do_start_stop(struct fsg_dev *fsg)
{
	struct lun	*curlun = fsg->curlun;
	int		loej, start;

	if (!mod_data.removable) {
		curlun->sense_data = SS_INVALID_COMMAND;
		return -EINVAL;
	}

	// int immed = fsg->cmnd[1] & 0x01;
	loej = fsg->cmnd[4] & 0x02;
	start = fsg->cmnd[4] & 0x01;

#ifdef CONFIG_USB_FILE_STORAGE_TEST
	if ((fsg->cmnd[1] & ~0x01) != 0 ||		// Mask away Immed
			(fsg->cmnd[4] & ~0x03) != 0) {	// Mask LoEj, Start
		curlun->sense_data = SS_INVALID_FIELD_IN_CDB;
		return -EINVAL;
	}

	if (!start) {

		/* Are we allowed to unload the media? */
		if (curlun->prevent_medium_removal) {
			LDBG(curlun, "unload attempt prevented\n");
			curlun->sense_data = SS_MEDIUM_REMOVAL_PREVENTED;
			return -EINVAL;
		}
		if (loej) {		// Simulate an unload/eject
			up_read(&fsg->filesem);
			down_write(&fsg->filesem);
			close_backing_file(curlun);
			up_write(&fsg->filesem);
			down_read(&fsg->filesem);
		}
	} else {

		/* Our emulation doesn't support mounting; the medium is
		 * available for use as soon as it is loaded. */
		if (!backing_file_is_open(curlun)) {
			curlun->sense_data = SS_MEDIUM_NOT_PRESENT;
			return -EINVAL;
		}
	}
#endif
	return 0;
}


static int do_prevent_allow(struct fsg_dev *fsg)
{
	struct lun	*curlun = fsg->curlun;
	int		prevent;

	if (!mod_data.removable) {
		curlun->sense_data = SS_INVALID_COMMAND;
		return -EINVAL;
	}

	prevent = fsg->cmnd[4] & 0x01;
	if ((fsg->cmnd[4] & ~0x01) != 0) {		// Mask away Prevent
		curlun->sense_data = SS_INVALID_FIELD_IN_CDB;
		return -EINVAL;
	}

	if (curlun->prevent_medium_removal && !prevent)
		fsync_sub(curlun);
	curlun->prevent_medium_removal = prevent;
	return 0;
}


static int do_read_format_capacities(struct fsg_dev *fsg,
			struct fsg_buffhd *bh)
{
	struct lun	*curlun = fsg->curlun;
	u8		*buf = (u8 *) bh->buf;

	buf[0] = buf[1] = buf[2] = 0;
	buf[3] = 8;		// Only the Current/Maximum Capacity Descriptor
	buf += 4;

	put_be32(&buf[0], curlun->num_sectors);		// Number of blocks
	put_be32(&buf[4], 512);				// Block length
	buf[4] = 0x02;					// Current capacity
	return 12;
}


static int do_mode_select(struct fsg_dev *fsg, struct fsg_buffhd *bh)
{
	struct lun	*curlun = fsg->curlun;

	/* We don't support MODE SELECT */
	curlun->sense_data = SS_INVALID_COMMAND;
	return -EINVAL;
}


/*-------------------------------------------------------------------------*/

static int halt_bulk_in_endpoint(struct fsg_dev *fsg)
{
	int	rc;

	rc = fsg_set_halt(fsg, fsg->bulk_in);
#ifdef USB_FILE_STORAGE_NAVU
	newdbg(fsg, "bulk-in endpoint halt rc=%d\n", rc);
	if (rc == -EAGAIN) {
		VDBG(fsg, "delayed bulk-in endpoint halt\n");
	  //newdbg(fsg, "delayed bulk-in endpoint halt\n");
	}
	while (rc != 0) {
		if (rc != -EAGAIN) {
			WARN(fsg, "usb_ep_set_halt -> %d\n", rc);
		  //newdbg(fsg, "usb_ep_set_halt -> %d\n", rc);
			rc = 0;
			break;
		}
		//newdbg(fsg, "usb_ep_set_halt again? = %d\n", rc);

		/* Wait for a short time and then try again */
		set_current_state(TASK_INTERRUPTIBLE);
		if (schedule_timeout(HZ / 10) != 0)
			return -EINTR;
		rc = usb_ep_set_halt(fsg->bulk_in);
	}
#else
	if (rc == -EAGAIN)
		VDBG(fsg, "delayed bulk-in endpoint halt\n");
	while (rc != 0) {
		if (rc != -EAGAIN) {
			WARN(fsg, "usb_ep_set_halt -> %d\n", rc);
			rc = 0;
			break;
		}

		/* Wait for a short time and then try again */
		if (msleep_interruptible(100) != 0)
			return -EINTR;
		rc = usb_ep_set_halt(fsg->bulk_in);
	}
#endif
	return rc;
}

static int pad_with_zeros(struct fsg_dev *fsg)
{
	struct fsg_buffhd	*bh = fsg->next_buffhd_to_fill;
	u32			nkeep = bh->inreq->length;
	u32			nsend;
	int			rc;

	bh->state = BUF_STATE_EMPTY;		// For the first iteration
	fsg->usb_amount_left = nkeep + fsg->residue;
	while (fsg->usb_amount_left > 0) {

		/* Wait for the next buffer to be free */
		while (bh->state != BUF_STATE_EMPTY) {
			if ((rc = sleep_thread(fsg)) != 0)
				return rc;
		}

		nsend = min(fsg->usb_amount_left, (u32) mod_data.buflen);
		memset(bh->buf + nkeep, 0, nsend - nkeep);
		bh->inreq->length = nsend;
		bh->inreq->zero = 0;
		start_transfer(fsg, fsg->bulk_in, bh->inreq,
				&bh->inreq_busy, &bh->state);
		bh = fsg->next_buffhd_to_fill = bh->next;
		fsg->usb_amount_left -= nsend;
		nkeep = 0;
	}
	return 0;
}

static int throw_away_data(struct fsg_dev *fsg)
{
	struct fsg_buffhd	*bh;
	u32			amount;
	int			rc;

	while ((bh = fsg->next_buffhd_to_drain)->state != BUF_STATE_EMPTY ||
			fsg->usb_amount_left > 0) {

		/* Throw away the data in a filled buffer */
		if (bh->state == BUF_STATE_FULL) {
			smp_rmb();
			bh->state = BUF_STATE_EMPTY;
			fsg->next_buffhd_to_drain = bh->next;

			/* A short packet or an error ends everything */
			if (bh->outreq->actual != bh->outreq->length ||
					bh->outreq->status != 0) {
#ifdef USB_FILE_STORAGE_NAVU
				sonydbg(fsg, "throw_away_data: raise_exception \n");
#endif
				raise_exception(fsg, FSG_STATE_ABORT_BULK_OUT);
				return -EINTR;
			}
			continue;
		}

		/* Try to submit another request if we need one */
		bh = fsg->next_buffhd_to_fill;
		if (bh->state == BUF_STATE_EMPTY && fsg->usb_amount_left > 0) {
			amount = min(fsg->usb_amount_left,
					(u32) mod_data.buflen);

			/* amount is always divisible by 512, hence by
			 * the bulk-out maxpacket size */
			bh->outreq->length = bh->bulk_out_intended_length =
					amount;
			bh->outreq->short_not_ok = 1;
			start_transfer(fsg, fsg->bulk_out, bh->outreq,
					&bh->outreq_busy, &bh->state);
			fsg->next_buffhd_to_fill = bh->next;
			fsg->usb_amount_left -= amount;
			continue;
		}

		/* Otherwise wait for something to happen */
		if ((rc = sleep_thread(fsg)) != 0)
			return rc;
	}
	return 0;
}


static int finish_reply(struct fsg_dev *fsg)
{
	struct fsg_buffhd	*bh = fsg->next_buffhd_to_fill;
	int			rc = 0;

#ifdef USB_FILE_STORAGE_NAVU
	sonydbg(fsg, "in the finish_reply function \n");
#endif
	switch (fsg->data_dir) {
	case DATA_DIR_NONE:
#ifdef USB_FILE_STORAGE_NAVU
	  sonydbg(fsg,"DATA_DIR_NONE state\n");
#endif
		break;			// Nothing to send

	/* If we don't know whether the host wants to read or write,
	 * this must be CB or CBI with an unknown command.  We mustn't
	 * try to send or receive any data.  So stall both bulk pipes
	 * if we can and wait for a reset. */
	case DATA_DIR_UNKNOWN:
		if (mod_data.can_stall) {
			fsg_set_halt(fsg, fsg->bulk_out);
			rc = halt_bulk_in_endpoint(fsg);
		}
#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "DATA_DIR_UNKNOWN state \n");
#endif
		break;

	/* All but the last buffer of data must have already been sent */
	case DATA_DIR_TO_HOST:
		if (fsg->data_size == 0) {
#ifdef USB_FILE_STORAGE_NAVU
			sonydbg(fsg, "finish_reply : DATA_DIR_TO_HOST :fsg->data_size = 0\n");
#endif
			// Nothing to send
		}
		/* If there's no residue, simply send the last buffer */
		else if (fsg->residue == 0) {
			bh->inreq->zero = 0;
			start_transfer(fsg, fsg->bulk_in, bh->inreq,
					&bh->inreq_busy, &bh->state);
#ifdef USB_FILE_STORAGE_NAVU
			sonydbg(fsg, "DATA_DIR_TO_HOST : send the last buffer \n");
			sonydbg(fsg, "show me the bh->inreq->length = %d \n", bh->inreq->length);
			//printk("<1> show me the bh->inreq->length = %d \n", bh->inreq->length);
#endif
			fsg->next_buffhd_to_fill = bh->next;
#ifdef USB_FILE_STORAGE_NAVU
#if 0
			if(1 == g_vendor_short) {
			  //printk("<1> vendor short packet\n");
			  struct fsg_buffhd	*bh_short = fsg->next_buffhd_to_fill;
			  bh_short->inreq->zero = 1;
			  bh_short->inreq->length = 0;
			  bh_short->inreq->actual = 0;
			  start_transfer(fsg, fsg->bulk_in, bh_short->inreq,
					 &bh_short->inreq_busy, &bh_short->state);
			  //printk("<1> show me the bh short 's length = %d \n", bh_short->inreq->length);
			  fsg->next_buffhd_to_fill = bh_short->next;
			  g_vendor_short = 0;
			}
#endif
#endif
		}

		/* There is a residue.  For CB and CBI, simply mark the end
		 * of the data with a short packet.  However, if we are
		 * allowed to stall, there was no data at all (residue ==
		 * data_size), and the command failed (invalid LUN or
		 * sense data is set), then halt the bulk-in endpoint
		 * instead. */
		else if (!transport_is_bbb()) {
#ifdef USB_FILE_STORAGE_NAVU
			sonydbg(fsg, "To Host: finish_reply, is not bbb \n");
#endif
			if (mod_data.can_stall &&
					fsg->residue == fsg->data_size &&
	(!fsg->curlun || fsg->curlun->sense_data != SS_NO_SENSE)) {
				bh->state = BUF_STATE_EMPTY;
				rc = halt_bulk_in_endpoint(fsg);
			} else {
				bh->inreq->zero = 1;
				start_transfer(fsg, fsg->bulk_in, bh->inreq,
						&bh->inreq_busy, &bh->state);
				fsg->next_buffhd_to_fill = bh->next;
			}
		}

		/* For Bulk-only, if we're allowed to stall then send the
		 * short packet and halt the bulk-in endpoint.  If we can't
		 * stall, pad out the remaining data with 0's. */
		else {
#ifdef USB_FILE_STORAGE_NAVU
			sonydbg(fsg, "For Bulk only-----, pad_with_zeros \n");
#endif
			if (mod_data.can_stall) {
				bh->inreq->zero = 1;
				start_transfer(fsg, fsg->bulk_in, bh->inreq,
						&bh->inreq_busy, &bh->state);
				fsg->next_buffhd_to_fill = bh->next;
				rc = halt_bulk_in_endpoint(fsg);
			}
			else {
			  rc = pad_with_zeros(fsg);
#ifdef USB_FILE_STORAGE_NAVU
			  //printk("<1> pad_with_zero\n");
			  // ljzhou 2006/6/20 start
			  // if not vendor command, keep the original process method
#if 0
			  if(0 == g_vendor_command) {
			    rc = pad_with_zeros(fsg);
			  }
			  else {
			    // if vendor command, send the short packet, and halt the
			    // endpoint

			    sema_init(&g_vendor_sem, 0);

			    bh->inreq->zero = 1;
			    newdbg(fsg, "send the short packet bh->inreq.len=%d\n", bh->inreq->length);
			    start_transfer(fsg, fsg->bulk_in, bh->inreq,
			    	   &bh->inreq_busy, &bh->state);
			    fsg->next_buffhd_to_fill = bh->next;


			    down(&g_vendor_sem);

			    //printk(KERN_ALERT " down complete\n");

			    int reg_s;
			    reg_s = pxa27x_ep_ctrl_status(fsg->bulk_in);
			    newdbg(fsg, "reg status is = %d \n", reg_s);
			    printk(KERN_ALERT "reg status is = %d \n", reg_s);

			    rc = halt_bulk_in_endpoint(fsg);
			    g_vendor_command = 0;
			  }
#endif
#endif
			}
		}
		break;

	/* We have processed all we want from the data the host has sent.
	 * There may still be outstanding bulk-out requests. */
	case DATA_DIR_FROM_HOST:
		if (fsg->residue == 0) {
#ifdef USB_FILE_STORAGE_NAVU
			sonydbg(fsg, "fsg->residu = 0\n");
#endif
			// Nothing to receive
		}

		/* Did the host stop sending unexpectedly early? */
		else if (fsg->short_packet_received) {
#ifdef USB_FILE_STORAGE_NAVU
			sonydbg(fsg, "DATA_DIR_FROM_HOST raise_excepiton\n");
#endif
			raise_exception(fsg, FSG_STATE_ABORT_BULK_OUT);
			rc = -EINTR;
		}

		/* We haven't processed all the incoming data.  Even though
		 * we may be allowed to stall, doing so would cause a race.
		 * The controller may already have ACK'ed all the remaining
		 * bulk-out packets, in which case the host wouldn't see a
		 * STALL.  Not realizing the endpoint was halted, it wouldn't
		 * clear the halt -- leading to problems later on. */
#if 0
		else if (mod_data.can_stall) {
			fsg_set_halt(fsg, fsg->bulk_out);
			raise_exception(fsg, FSG_STATE_ABORT_BULK_OUT);
			rc = -EINTR;
#ifdef USB_FILE_STORAGE_NAVU
			sonydbg(fsg, "mod_data.can_stall \n");
#endif
		}
#endif

		/* We can't stall.  Read in the excess data and throw it
		 * all away. */
		else {
#ifdef USB_FILE_STORAGE_NAVU
			sonydbg(fsg, "throw away data \n");
#endif
			rc = throw_away_data(fsg);
		}
		break;
	}
	return rc;
}


static int send_status(struct fsg_dev *fsg)
{
	struct lun		*curlun = fsg->curlun;
	struct fsg_buffhd	*bh;
	int			rc;
	u8			status = USB_STATUS_PASS;
	u32			sd, sdinfo = 0;

#ifdef USB_FILE_STORAGE_NAVU
	sonydbg(fsg, "send status \n");
#endif
	/* Wait for the next buffer to become available */
	bh = fsg->next_buffhd_to_fill;
	while (bh->state != BUF_STATE_EMPTY) {
		if ((rc = sleep_thread(fsg)) != 0)
			return rc;
	}

	if (curlun) {
		sd = curlun->sense_data;
		sdinfo = curlun->sense_data_info;
	} else if (fsg->bad_lun_okay)
		sd = SS_NO_SENSE;
	else
		sd = SS_LOGICAL_UNIT_NOT_SUPPORTED;

	if (fsg->phase_error) {
		DBG(fsg, "sending phase-error status\n");
		status = USB_STATUS_PHASE_ERROR;
		sd = SS_INVALID_COMMAND;
	} else if (sd != SS_NO_SENSE) {
#ifdef USB_FILE_STORAGE_NAVU
		//printk(KERN_DEBUG "sending command-failure status!\n");
		sonydbg(fsg, "sending command-failure status\n");
		status = USB_STATUS_FAIL;
		sonydbg(fsg, "  sense data: SK x%02x, ASC x%02x, ASCQ x%02x;"
				"  info x%x\n",
				SK(sd), ASC(sd), ASCQ(sd), sdinfo);
#else
		DBG(fsg, "sending command-failure status\n");
		status = USB_STATUS_FAIL;
		VDBG(fsg, "  sense data: SK x%02x, ASC x%02x, ASCQ x%02x;"
				"  info x%x\n",
				SK(sd), ASC(sd), ASCQ(sd), sdinfo);
#endif
	}

	if (transport_is_bbb()) {
		struct bulk_cs_wrap	*csw = bh->buf;

		/* Store and send the Bulk-only CSW */
		csw->Signature = __constant_cpu_to_le32(USB_BULK_CS_SIG);
		csw->Tag = fsg->tag;
		csw->Residue = cpu_to_le32(fsg->residue);
		csw->Status = status;

		bh->inreq->length = USB_BULK_CS_WRAP_LEN;
		bh->inreq->zero = 0;
		start_transfer(fsg, fsg->bulk_in, bh->inreq,
				&bh->inreq_busy, &bh->state);

#ifdef USB_FILE_STORAGE_NAVU
		newdbg(fsg, "send status , is bbb inq.len=%d, inq.zero=%d, req.state=%d \n",
		       bh->inreq->length, bh->inreq->zero, bh->state);
#endif
	} else if (mod_data.transport_type == USB_PR_CB) {

		/* Control-Bulk transport has no status phase! */
		return 0;

	} else {			// USB_PR_CBI
		struct interrupt_data	*buf = bh->buf;

		/* Store and send the Interrupt data.  UFI sends the ASC
		 * and ASCQ bytes.  Everything else sends a Type (which
		 * is always 0) and the status Value. */
		if (mod_data.protocol_type == USB_SC_UFI) {
			buf->bType = ASC(sd);
			buf->bValue = ASCQ(sd);
		} else {
			buf->bType = 0;
			buf->bValue = status;
		}
		fsg->intreq->length = CBI_INTERRUPT_DATA_LEN;

		fsg->intr_buffhd = bh;		// Point to the right buffhd
		fsg->intreq->buf = bh->inreq->buf;
		fsg->intreq->context = bh;
		start_transfer(fsg, fsg->intr_in, fsg->intreq,
				&fsg->intreq_busy, &bh->state);
	}

	fsg->next_buffhd_to_fill = bh->next;
	return 0;
}


/*-------------------------------------------------------------------------*/

#ifdef USB_FILE_STORAGE_NAVU
static void truncate_sub(struct lun *curlun)
{
	struct file		*filp = curlun->filp;
	struct block_device	*bdev = filp->f_path.dentry->d_inode->i_bdev;

	mutex_lock(&bdev->bd_inode->i_mutex);
	invalidate_bdev(bdev);
	truncate_inode_pages(bdev->bd_inode->i_mapping, 0);
	mutex_unlock(&bdev->bd_inode->i_mutex);
}
#endif

/* Check whether the command is properly formed and whether its data size
 * and direction agree with the values we already have. */
static int check_command(struct fsg_dev *fsg, int cmnd_size,
		enum data_direction data_dir, unsigned int mask,
		int needs_medium, const char *name)
{
	int			i;
	int			lun = fsg->cmnd[1] >> 5;
	static const char	dirletter[4] = {'u', 'o', 'i', 'n'};
	char			hdlen[20];
	struct lun		*curlun;

	/* Adjust the expected cmnd_size for protocol encapsulation padding.
	 * Transparent SCSI doesn't pad. */
	if (protocol_is_scsi())
		;

	/* There's some disagreement as to whether RBC pads commands or not.
	 * We'll play it safe and accept either form. */
	else if (mod_data.protocol_type == USB_SC_RBC) {
		if (fsg->cmnd_size == 12)
			cmnd_size = 12;

	/* All the other protocols pad to 12 bytes */
	} else
		cmnd_size = 12;

	hdlen[0] = 0;
	if (fsg->data_dir != DATA_DIR_UNKNOWN)
		sprintf(hdlen, ", H%c=%u", dirletter[(int) fsg->data_dir],
				fsg->data_size);
	VDBG(fsg, "SCSI command: %s;  Dc=%d, D%c=%u;  Hc=%d%s\n",
			name, cmnd_size, dirletter[(int) data_dir],
			fsg->data_size_from_cmnd, fsg->cmnd_size, hdlen);

	/* We can't reply at all until we know the correct data direction
	 * and size. */
	if (fsg->data_size_from_cmnd == 0)
		data_dir = DATA_DIR_NONE;
	if (fsg->data_dir == DATA_DIR_UNKNOWN) {	// CB or CBI
		fsg->data_dir = data_dir;
		fsg->data_size = fsg->data_size_from_cmnd;

	} else {					// Bulk-only
		if (fsg->data_size < fsg->data_size_from_cmnd) {

			/* Host data size < Device data size is a phase error.
			 * Carry out the command, but only transfer as much
			 * as we are allowed. */
			fsg->data_size_from_cmnd = fsg->data_size;
			fsg->phase_error = 1;
		}
	}
	fsg->residue = fsg->usb_amount_left = fsg->data_size;

	/* Conflicting data directions is a phase error */
	if (fsg->data_dir != data_dir && fsg->data_size_from_cmnd > 0) {
		fsg->phase_error = 1;
		return -EINVAL;
	}

	/* Verify the length of the command itself */
	if (cmnd_size != fsg->cmnd_size) {

		/* Special case workaround: MS-Windows issues REQUEST SENSE
		 * with cbw->Length == 12 (it should be 6). */
		if (fsg->cmnd[0] == SC_REQUEST_SENSE && fsg->cmnd_size == 12) {
			cmnd_size = fsg->cmnd_size;
		}
#ifdef USB_FILE_STORAGE_NAVU
#if 1
		/* To accept SCSI and SFF-8070i, a check condition has beem changed. */
		else if(fsg->cmnd_size == 12) {
#else
		else if(fsg->cmnd[0] == SC_INQUIRY && fsg->cmnd_size == 12) {
#endif
			;
		}
#endif
		else {
			fsg->phase_error = 1;
			return -EINVAL;
		}
	}

	/* Check that the LUN values are consistent */
	if (transport_is_bbb()) {
		if (fsg->lun != lun)
			DBG(fsg, "using LUN %d from CBW, "
					"not LUN %d from CDB\n",
					fsg->lun, lun);
	} else
		fsg->lun = lun;		// Use LUN from the command

	/* Check the LUN */
	if (fsg->lun >= 0 && fsg->lun < fsg->nluns) {
		fsg->curlun = curlun = &fsg->luns[fsg->lun];
		if (fsg->cmnd[0] != SC_REQUEST_SENSE) {
			curlun->sense_data = SS_NO_SENSE;
			curlun->sense_data_info = 0;
			curlun->info_valid = 0;
#ifdef USB_FILE_STORAGE_NAVU
			curlun->is_fake_sense_data = false;
#endif /* USB_FILE_STORAGE_NAVU */
		}
	} else {
		fsg->curlun = curlun = NULL;
		fsg->bad_lun_okay = 0;

		/* INQUIRY and REQUEST SENSE commands are explicitly allowed
		 * to use unsupported LUNs; all others may not. */
		if (fsg->cmnd[0] != SC_INQUIRY &&
				fsg->cmnd[0] != SC_REQUEST_SENSE) {
			DBG(fsg, "unsupported LUN %d\n", fsg->lun);
			return -EINVAL;
		}
	}

	/* If a unit attention condition exists, only INQUIRY and
	 * REQUEST SENSE commands are allowed; anything else must fail. */
	if (curlun && curlun->unit_attention_data != SS_NO_SENSE &&
			fsg->cmnd[0] != SC_INQUIRY &&
			fsg->cmnd[0] != SC_REQUEST_SENSE) {
		curlun->sense_data = curlun->unit_attention_data;
		curlun->unit_attention_data = SS_NO_SENSE;
		return -EINVAL;
	}

	/* Check that only command bytes listed in the mask are non-zero */
	fsg->cmnd[1] &= 0x1f;			// Mask away the LUN
	for (i = 1; i < cmnd_size; ++i) {
		if (fsg->cmnd[i] && !(mask & (1 << i))) {
			if (curlun)
				curlun->sense_data = SS_INVALID_FIELD_IN_CDB;
			return -EINVAL;
		}
	}

	/* If the medium isn't mounted and the command needs to access
	 * it, try to mount it. */
	if (curlun && !backing_file_is_open(curlun) && needs_medium) {
#ifdef USB_FILE_STORAGE_NAVU
		if (open_backing_file(curlun, mod_data.file[fsg->lun])){
			curlun->sense_data = SS_MEDIUM_NOT_PRESENT;
			return -EINVAL;
		}
		else {
			curlun->sense_data = SS_NOT_READY_TO_READY_TRANSITION;
			return -EINVAL;
		}
#else
		curlun->sense_data = SS_MEDIUM_NOT_PRESENT;
		return -EINVAL;
#endif
	}

#ifdef USB_FILE_STORAGE_NAVU
	if (curlun && backing_file_is_open(curlun) && needs_medium &&
	    mod_data.removable && check_media_change(fsg)) {

		truncate_sub(curlun);

		close_backing_file(curlun);

		if (open_backing_file(curlun, mod_data.file[fsg->lun])){
			curlun->sense_data = SS_MEDIUM_NOT_PRESENT;
			return -EINVAL;
		}
		else {
			curlun->sense_data = SS_NOT_READY_TO_READY_TRANSITION;
			return -EINVAL;
		}
	}
#endif
	return 0;
}

#ifdef USB_FILE_STORAGE_NAVU
static void fake_sense_data(struct fsg_dev *fsg)
{
	u32		sd;
	int		cur_lun;

	sd = fsg->curlun->sense_data;
	cur_lun = fsg->lun;

	if( SK(sd) != 6 ){
		if( cur_lun <= fsg->last_lun ){
			if( cur_lun == fsg->last_lun ){
				fsg->last_lun = cur_lun + 1;
			}
		}else{
			fsg->curlun->is_fake_sense_data = true;
			fsg->curlun->sense_data = SS_NOT_READY_TO_READY_TRANSITION;
		}
	}
}
#endif /* USB_FILE_STORAGE_NAVU */

static int do_scsi_command(struct fsg_dev *fsg)
{
	struct fsg_buffhd	*bh;
	int			rc;
	int			reply = -EINVAL;
	int			i;
#ifdef USB_FILE_STORAGE_NAVU
	int			index;
	//static char		unknown[16];
#else
	static char		unknown[16];
#endif

	dump_cdb(fsg);

	/* Wait for the next buffer to become available for data or status */
	bh = fsg->next_buffhd_to_drain = fsg->next_buffhd_to_fill;
	while (bh->state != BUF_STATE_EMPTY) {
		if ((rc = sleep_thread(fsg)) != 0)
			return rc;
		}
	fsg->phase_error = 0;
	fsg->short_packet_received = 0;

#ifdef USB_FILE_STORAGE_NAVU
	//ljzhou 2006/6/20 start
	//g_vendor_command = 0;
	//ljzhou 2006/6/20 end
#endif

	down_read(&fsg->filesem);	// We're using the backing file
#ifdef USB_FILE_STORAGE_NAVU
	Usb_connection_flag = 1;
#endif
	switch (fsg->cmnd[0]) {

	case SC_INQUIRY:
		fsg->data_size_from_cmnd = fsg->cmnd[4];
		if ((reply = check_command(fsg, 6, DATA_DIR_TO_HOST,
				(1<<4), 0,
				"INQUIRY")) == 0)
			reply = do_inquiry(fsg, bh);
		break;

	case SC_MODE_SELECT_6:
		fsg->data_size_from_cmnd = fsg->cmnd[4];
		if ((reply = check_command(fsg, 6, DATA_DIR_FROM_HOST,
				(1<<1) | (1<<4), 0,
				"MODE SELECT(6)")) == 0)
			reply = do_mode_select(fsg, bh);
		break;

	case SC_MODE_SELECT_10:
		fsg->data_size_from_cmnd = get_be16(&fsg->cmnd[7]);
		if ((reply = check_command(fsg, 10, DATA_DIR_FROM_HOST,
				(1<<1) | (3<<7), 0,
				"MODE SELECT(10)")) == 0)
			reply = do_mode_select(fsg, bh);
		break;

	case SC_MODE_SENSE_6:
		fsg->data_size_from_cmnd = fsg->cmnd[4];
		if ((reply = check_command(fsg, 6, DATA_DIR_TO_HOST,
				(1<<1) | (1<<2) | (1<<4), 0,
				"MODE SENSE(6)")) == 0)
			reply = do_mode_sense(fsg, bh);
		break;

	case SC_MODE_SENSE_10:
		fsg->data_size_from_cmnd = get_be16(&fsg->cmnd[7]);
#ifdef USB_FILE_STORAGE_NAVU
		/* Special case:
		   About MODE SENSE(10), some USB host driver sends incorrect bulk-transfer data size.
		   The data size is not same
		   between cbw->DataTransferLength( a dCBWDataTransferLength of CBW )
		   and fsg->cmnd[7]( an allocation length of CDB ). */
		if ((reply = check_command(fsg, 10, DATA_DIR_TO_HOST,
				(1<<1) | (1<<2) | (3<<7), 0,
				"MODE SENSE(10)")) == 0) {
			reply = do_mode_sense(fsg, bh);
			fsg->phase_error = 0;
		}
#else
		if ((reply = check_command(fsg, 10, DATA_DIR_TO_HOST,
				(1<<1) | (1<<2) | (3<<7), 0,
				"MODE SENSE(10)")) == 0)
			reply = do_mode_sense(fsg, bh);
#endif
		break;

	case SC_PREVENT_ALLOW_MEDIUM_REMOVAL:
		fsg->data_size_from_cmnd = 0;
		if ((reply = check_command(fsg, 6, DATA_DIR_NONE,
				(1<<4), 0,
				"PREVENT-ALLOW MEDIUM REMOVAL")) == 0)
			reply = do_prevent_allow(fsg);
		break;

	case SC_READ_6:
		i = fsg->cmnd[4];
		fsg->data_size_from_cmnd = (i == 0 ? 256 : i) << 9;
		if ((reply = check_command(fsg, 6, DATA_DIR_TO_HOST,
				(7<<1) | (1<<4), 1,
				"READ(6)")) == 0)
			reply = do_read(fsg);
		break;

	case SC_READ_10:
		fsg->data_size_from_cmnd = get_be16(&fsg->cmnd[7]) << 9;
		if ((reply = check_command(fsg, 10, DATA_DIR_TO_HOST,
				(1<<1) | (0xf<<2) | (3<<7), 1,
				"READ(10)")) == 0)
			reply = do_read(fsg);
		break;

	case SC_READ_12:
		fsg->data_size_from_cmnd = get_be32(&fsg->cmnd[6]) << 9;
		if ((reply = check_command(fsg, 12, DATA_DIR_TO_HOST,
				(1<<1) | (0xf<<2) | (0xf<<6), 1,
				"READ(12)")) == 0)
			reply = do_read(fsg);
		break;

	case SC_READ_CAPACITY:
		fsg->data_size_from_cmnd = 8;
		if ((reply = check_command(fsg, 10, DATA_DIR_TO_HOST,
				(0xf<<2) | (1<<8), 1,
				"READ CAPACITY")) == 0)
			reply = do_read_capacity(fsg, bh);
		break;

	case SC_READ_FORMAT_CAPACITIES:
		fsg->data_size_from_cmnd = get_be16(&fsg->cmnd[7]);
		if ((reply = check_command(fsg, 10, DATA_DIR_TO_HOST,
				(3<<7), 1,
				"READ FORMAT CAPACITIES")) == 0)
			reply = do_read_format_capacities(fsg, bh);
		break;

	case SC_REQUEST_SENSE:
		fsg->data_size_from_cmnd = fsg->cmnd[4];
		if ((reply = check_command(fsg, 6, DATA_DIR_TO_HOST,
				(1<<4), 0,
				"REQUEST SENSE")) == 0)
			reply = do_request_sense(fsg, bh);
		break;

	case SC_START_STOP_UNIT:
		fsg->data_size_from_cmnd = 0;
		if ((reply = check_command(fsg, 6, DATA_DIR_NONE,
				(1<<1) | (1<<4), 0,
				"START-STOP UNIT")) == 0)
			reply = do_start_stop(fsg);
		break;

	case SC_SYNCHRONIZE_CACHE:
		fsg->data_size_from_cmnd = 0;
		if ((reply = check_command(fsg, 10, DATA_DIR_NONE,
				(0xf<<2) | (3<<7), 1,
				"SYNCHRONIZE CACHE")) == 0)
			reply = do_synchronize_cache(fsg);
		break;

	case SC_TEST_UNIT_READY:
		fsg->data_size_from_cmnd = 0;
		reply = check_command(fsg, 6, DATA_DIR_NONE,
				0, 1,
				"TEST UNIT READY");
#ifdef USB_FILE_STORAGE_NAVU
		check_media_insertion(fsg);
		fake_sense_data(fsg);
#endif
		break;

	/* Although optional, this command is used by MS-Windows.  We
	 * support a minimal version: BytChk must be 0. */
	case SC_VERIFY:
		fsg->data_size_from_cmnd = 0;
		if ((reply = check_command(fsg, 10, DATA_DIR_NONE,
				(1<<1) | (0xf<<2) | (3<<7), 1,
				"VERIFY")) == 0)
			reply = do_verify(fsg);
		break;

	case SC_WRITE_6:
		i = fsg->cmnd[4];
		fsg->data_size_from_cmnd = (i == 0 ? 256 : i) << 9;
		if ((reply = check_command(fsg, 6, DATA_DIR_FROM_HOST,
				(7<<1) | (1<<4), 1,
				"WRITE(6)")) == 0)
			reply = do_write(fsg);
		break;

	case SC_WRITE_10:
		fsg->data_size_from_cmnd = get_be16(&fsg->cmnd[7]) << 9;
		if ((reply = check_command(fsg, 10, DATA_DIR_FROM_HOST,
				(1<<1) | (0xf<<2) | (3<<7), 1,
				"WRITE(10)")) == 0)
			reply = do_write(fsg);
		break;

	case SC_WRITE_12:
		fsg->data_size_from_cmnd = get_be32(&fsg->cmnd[6]) << 9;
		if ((reply = check_command(fsg, 12, DATA_DIR_FROM_HOST,
				(1<<1) | (0xf<<2) | (0xf<<6), 1,
				"WRITE(12)")) == 0)
			reply = do_write(fsg);
		break;

	/* Some mandatory commands that we recognize but don't implement.
	 * They don't mean much in this setting.  It's left as an exercise
	 * for anyone interested to implement RESERVE and RELEASE in terms
	 * of Posix locks. */
	case SC_FORMAT_UNIT:
	case SC_RELEASE:
	case SC_RESERVE:
	case SC_SEND_DIAGNOSTIC:
		// Fall through

	default:
#ifdef USB_FILE_STORAGE_NAVU  // MACH_SONY_NAVU is old label.
		index = check_vc_handler(fsg);
	  //fsg->data_size_from_cmnd = get_be32(&fsg->cmd[6]);
		/* call do_receive_data function to retrive the data
		   from usb register, then pass them to device */
		//sonydbg(fsg, "here is vendor unique command \n");
		if( DATA_DIR_TO_HOST == parse_command_dir(fsg) ) {
			newdbg(fsg, "Data direction is from device to Host \n");
			fsg->data_size_from_cmnd = fsg->data_size;
			if ((reply = check_vendor_command(fsg, fsg->cmnd_size, DATA_DIR_TO_HOST,
				1, "do_send_data")) == 0) {
				if ( ( 0 <= index ) && ( index < MAX_VC_HANDLER_NS ) ) {
					reply = do_send_data_vc_handler(fsg, index, bh);
				} else {
                                //ljzhou 2006/6/20
                                //g_vendor_command = 1;
					reply = do_send_data(fsg);
				}
			}
		}
		else if( DATA_DIR_FROM_HOST == parse_command_dir(fsg) ) {
			sonydbg(fsg, "Data direction is from Host to device \n");
			fsg->data_size_from_cmnd = fsg->data_size;
			if ((reply = check_vendor_command(fsg, fsg->cmnd_size, DATA_DIR_FROM_HOST,
				1, "do_receive_data")) == 0) {
				if ( ( 0 <= index ) && ( index < MAX_VC_HANDLER_NS ) ) {
					reply = do_receive_data_vc_handler(fsg, index, bh);
				} else {
			        //ljzhou 2006/6/20
			        //g_vendor_command = 1;
			        reply = do_receive_data(fsg);
			    }
			}
		}
		//Added by Steven 060707 Start.
		else if( DATA_DIR_NONE == parse_command_dir(fsg) ) {
			sonydbg(fsg, "Data direction is NONE\n");
			fsg->data_size_from_cmnd = fsg->data_size;
			if ((reply = check_vendor_command(fsg, fsg->cmnd_size, DATA_DIR_NONE,
				1, "do_no_data")) == 0) {
				if ( ( 0 <= index ) && ( index < MAX_VC_HANDLER_NS ) ) {
					reply = do_no_data_vc_handler(fsg, index, bh);
				} else {
			        reply = do_no_data(fsg);
			    }
			}
		}
		//Added by Steven 060707 End.
		else {
			newdbg(fsg, "Don't process this direction \n");
			reply = -EINVAL;
		}
		break;

#ifdef SCSI_TEST_APP
		printk("MassStorage:Unknown len=%d ", fsg->cmnd_size);
		{
			int iii;
			for(iii = 0; iii < fsg->cmnd_size; iii++) {
				printk(" %d:0x%02x", iii, fsg->cmnd[iii]);
			}
			printk("\n");
		}
		fsg->data_size_from_cmnd = 0;
#endif

#else /* USB_FILE_STORAGE_NAVU */ /* MACH_SONY_NAVU is old label. */

		fsg->data_size_from_cmnd = 0;
		sprintf(unknown, "Unknown x%02x", fsg->cmnd[0]);
		if ((reply = check_command(fsg, fsg->cmnd_size,
				DATA_DIR_UNKNOWN, 0xff, 0, unknown)) == 0) {
			fsg->curlun->sense_data = SS_INVALID_COMMAND;
			reply = -EINVAL;
		}
#endif /* USB_FILE_STORAGE_NAVU */ /* MACH_SONY_NAVU is old label. */
		break;
	}
	up_read(&fsg->filesem);

	if (reply == -EINTR || signal_pending(current)) {
#ifdef USB_FILE_STORAGE_NAVU
		if(reply == -EINTR)
			sonydbg(fsg, "reply == -EINTR \n");
		else
			sonydbg(fsg, "signal_pending(current)) \n");
#endif
		return -EINTR;
	}

#ifdef USB_FILE_STORAGE_NAVU
	newdbg(fsg, "printout the reply info = %d \n", reply);
#endif
	/* Set up the single reply buffer for finish_reply() */
	if (reply == -EINVAL)
		reply = 0;		// Error reply length
	if (reply >= 0 && fsg->data_dir == DATA_DIR_TO_HOST) {
		reply = min((u32) reply, fsg->data_size_from_cmnd);
		bh->inreq->length = reply;
		bh->state = BUF_STATE_FULL;
		fsg->residue -= reply;

	}				// Otherwise it's already set

#ifdef USB_FILE_STORAGE_NAVU
	newdbg(fsg, "printout the reply code = %d, fsg->residue = %d \n",
		reply, fsg->residue);

        /* Richard add */
	/* for wakeup the wait queue */
	//wake_up(&SendDataQueue);
#endif
	return 0;
}


/*-------------------------------------------------------------------------*/

static int received_cbw(struct fsg_dev *fsg, struct fsg_buffhd *bh)
{
	struct usb_request	*req = bh->outreq;
	struct bulk_cb_wrap	*cbw = req->buf;

	/* Was this a real packet? */
	if (req->status)
		return -EINVAL;

	/* Is the CBW valid? */
	if (req->actual != USB_BULK_CB_WRAP_LEN ||
			cbw->Signature != __constant_cpu_to_le32(
				USB_BULK_CB_SIG)) {
#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "invalid CBW: len %u sig 0x%x\n",
				req->actual,
				le32_to_cpu(cbw->Signature));
#else
		DBG(fsg, "invalid CBW: len %u sig 0x%x\n",
				req->actual,
				le32_to_cpu(cbw->Signature));
#endif

		/* The Bulk-only spec says we MUST stall the bulk pipes!
		 * If we want to avoid stalls, set a flag so that we will
		 * clear the endpoint halts at the next reset. */
		if (!mod_data.can_stall)
			set_bit(CLEAR_BULK_HALTS, &fsg->atomic_bitflags);
		fsg_set_halt(fsg, fsg->bulk_out);
		halt_bulk_in_endpoint(fsg);
		return -EINVAL;
	}

	/* Is the CBW meaningful? */
	if (cbw->Lun >= MAX_LUNS || cbw->Flags & ~USB_BULK_IN_FLAG ||
			cbw->Length < 6 || cbw->Length > MAX_COMMAND_SIZE) {
#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "non-meaningful CBW: lun = %u, flags = 0x%x, "
				"cmdlen %u\n",
				cbw->Lun, cbw->Flags, cbw->Length);
#else
		DBG(fsg, "non-meaningful CBW: lun = %u, flags = 0x%x, "
				"cmdlen %u\n",
				cbw->Lun, cbw->Flags, cbw->Length);
#endif

		/* We can do anything we want here, so let's stall the
		 * bulk pipes if we are allowed to. */
		if (mod_data.can_stall) {
			fsg_set_halt(fsg, fsg->bulk_out);
			halt_bulk_in_endpoint(fsg);
		}
		return -EINVAL;
	}

	/* Save the command for later */
	fsg->cmnd_size = cbw->Length;
	memcpy(fsg->cmnd, cbw->CDB, fsg->cmnd_size);
	if (cbw->Flags & USB_BULK_IN_FLAG)
		fsg->data_dir = DATA_DIR_TO_HOST;
	else
		fsg->data_dir = DATA_DIR_FROM_HOST;
	fsg->data_size = le32_to_cpu(cbw->DataTransferLength);
	if (fsg->data_size == 0)
		fsg->data_dir = DATA_DIR_NONE;
	fsg->lun = cbw->Lun;
	fsg->tag = cbw->Tag;
	return 0;
}


static int get_next_command(struct fsg_dev *fsg)
{
	struct fsg_buffhd	*bh;
	int			rc = 0;

	if (transport_is_bbb()) {

		/* Wait for the next buffer to become available */
		bh = fsg->next_buffhd_to_fill;

#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "get next command : start BBB\n");
#endif
		while (bh->state != BUF_STATE_EMPTY) {
			if ((rc = sleep_thread(fsg)) != 0)
				return rc;
			}
#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "get next command : start Transfer in get next command \n");
#endif
		/* Queue a request to read a Bulk-only CBW */
		set_bulk_out_req_length(fsg, bh, USB_BULK_CB_WRAP_LEN);
		bh->outreq->short_not_ok = 1;
		start_transfer(fsg, fsg->bulk_out, bh->outreq,
				&bh->outreq_busy, &bh->state);

		/* We will drain the buffer in software, which means we
		 * can reuse it for the next filling.  No need to advance
		 * next_buffhd_to_fill. */

		/* Wait for the CBW to arrive */
		while (bh->state != BUF_STATE_FULL) {
			if ((rc = sleep_thread(fsg)) != 0)
				return rc;
			}
		smp_rmb();
		rc = received_cbw(fsg, bh);
		bh->state = BUF_STATE_EMPTY;
#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "get next command : finish BBB \n");
#endif

	} else {		// USB_PR_CB or USB_PR_CBI

		/* Wait for the next command to arrive */
		while (fsg->cbbuf_cmnd_size == 0) {
			if ((rc = sleep_thread(fsg)) != 0)
				return rc;
			}

		/* Is the previous status interrupt request still busy?
		 * The host is allowed to skip reading the status,
		 * so we must cancel it. */
		if (fsg->intreq_busy)
			usb_ep_dequeue(fsg->intr_in, fsg->intreq);

		/* Copy the command and mark the buffer empty */
		fsg->data_dir = DATA_DIR_UNKNOWN;
		spin_lock_irq(&fsg->lock);
		fsg->cmnd_size = fsg->cbbuf_cmnd_size;
		memcpy(fsg->cmnd, fsg->cbbuf_cmnd, fsg->cmnd_size);
		fsg->cbbuf_cmnd_size = 0;
		spin_unlock_irq(&fsg->lock);
	}
	return rc;
}


/*-------------------------------------------------------------------------*/

static int enable_endpoint(struct fsg_dev *fsg, struct usb_ep *ep,
		const struct usb_endpoint_descriptor *d)
{
	int	rc;

	ep->driver_data = fsg;
	rc = usb_ep_enable(ep, d);
	if (rc)
		ERROR(fsg, "can't enable %s, result %d\n", ep->name, rc);
	return rc;
}

static int alloc_request(struct fsg_dev *fsg, struct usb_ep *ep,
		struct usb_request **preq)
{
	*preq = usb_ep_alloc_request(ep, GFP_ATOMIC);
	if (*preq)
		return 0;
	ERROR(fsg, "can't allocate request for %s\n", ep->name);
	return -ENOMEM;
}

/*
 * Reset interface setting and re-init endpoint state (toggle etc).
 * Call with altsetting < 0 to disable the interface.  The only other
 * available altsetting is 0, which enables the interface.
 */
static int do_set_interface(struct fsg_dev *fsg, int altsetting)
{
	int	rc = 0;
	int	i;
	const struct usb_endpoint_descriptor	*d;

	if (fsg->running)
		DBG(fsg, "reset interface\n");
#ifdef USB_FILE_STORAGE_NAVU
	sonydbg(fsg, "do_set_interface \n");
#endif
reset:
	/* Deallocate the requests */
	for (i = 0; i < NUM_BUFFERS; ++i) {
		struct fsg_buffhd *bh = &fsg->buffhds[i];

		if (bh->inreq) {
			usb_ep_free_request(fsg->bulk_in, bh->inreq);
			bh->inreq = NULL;
		}
		if (bh->outreq) {
			usb_ep_free_request(fsg->bulk_out, bh->outreq);
			bh->outreq = NULL;
		}
	}
	if (fsg->intreq) {
		usb_ep_free_request(fsg->intr_in, fsg->intreq);
		fsg->intreq = NULL;
	}

	/* Disable the endpoints */
	if (fsg->bulk_in_enabled) {
		usb_ep_disable(fsg->bulk_in);
		fsg->bulk_in_enabled = 0;
	}
	if (fsg->bulk_out_enabled) {
		usb_ep_disable(fsg->bulk_out);
		fsg->bulk_out_enabled = 0;
	}
	if (fsg->intr_in_enabled) {
		usb_ep_disable(fsg->intr_in);
		fsg->intr_in_enabled = 0;
	}

	fsg->running = 0;
	if (altsetting < 0 || rc != 0)
		return rc;

	DBG(fsg, "set interface %d\n", altsetting);

	/* Enable the endpoints */
	d = ep_desc(fsg->gadget, &fs_bulk_in_desc, &hs_bulk_in_desc);
	if ((rc = enable_endpoint(fsg, fsg->bulk_in, d)) != 0)
		goto reset;
	fsg->bulk_in_enabled = 1;

	d = ep_desc(fsg->gadget, &fs_bulk_out_desc, &hs_bulk_out_desc);
	if ((rc = enable_endpoint(fsg, fsg->bulk_out, d)) != 0)
		goto reset;
	fsg->bulk_out_enabled = 1;
	fsg->bulk_out_maxpacket = le16_to_cpu(d->wMaxPacketSize);

	if (transport_is_cbi()) {
		d = ep_desc(fsg->gadget, &fs_intr_in_desc, &hs_intr_in_desc);
		if ((rc = enable_endpoint(fsg, fsg->intr_in, d)) != 0)
			goto reset;
		fsg->intr_in_enabled = 1;
	}

	/* Allocate the requests */
	for (i = 0; i < NUM_BUFFERS; ++i) {
		struct fsg_buffhd	*bh = &fsg->buffhds[i];

		if ((rc = alloc_request(fsg, fsg->bulk_in, &bh->inreq)) != 0)
			goto reset;
		if ((rc = alloc_request(fsg, fsg->bulk_out, &bh->outreq)) != 0)
			goto reset;
		bh->inreq->buf = bh->outreq->buf = bh->buf;
		bh->inreq->context = bh->outreq->context = bh;
		bh->inreq->complete = bulk_in_complete;
		bh->outreq->complete = bulk_out_complete;
	}
	if (transport_is_cbi()) {
		if ((rc = alloc_request(fsg, fsg->intr_in, &fsg->intreq)) != 0)
			goto reset;
		fsg->intreq->complete = intr_in_complete;
	}

	fsg->running = 1;
	for (i = 0; i < fsg->nluns; ++i)
		fsg->luns[i].unit_attention_data = SS_RESET_OCCURRED;
	return rc;
}


/*
 * Change our operational configuration.  This code must agree with the code
 * that returns config descriptors, and with interface altsetting code.
 *
 * It's also responsible for power management interactions.  Some
 * configurations might not work with our current power sources.
 * For now we just assume the gadget is always self-powered.
 */
static int do_set_config(struct fsg_dev *fsg, u8 new_config)
{
	int	rc = 0;
#ifdef USB_FILE_STORAGE_NAVU
	int i;
#endif /* USB_FILE_STORAGE_NAVU */

	/* Disable the single interface */
	if (fsg->config != 0) {
		DBG(fsg, "reset config\n");
		fsg->config = 0;
		rc = do_set_interface(fsg, -1);
	}

	/* Enable the interface */
	if (new_config != 0) {
		fsg->config = new_config;
		if ((rc = do_set_interface(fsg, 0)) != 0)
			fsg->config = 0;	// Reset on errors
		else {
			char *speed;

			switch (fsg->gadget->speed) {
			case USB_SPEED_LOW:	speed = "low";	break;
			case USB_SPEED_FULL:	speed = "full";	break;
			case USB_SPEED_HIGH:	speed = "high";	break;
			default: 		speed = "?";	break;
			}
			INFO(fsg, "%s speed config #%d\n", speed, fsg->config);
#ifdef USB_FILE_STORAGE_NAVU
			// ljzhou 2006/07/11 start
			g_usb_config_cplt = 1;
			/* Initial the status */
			g_exit_flag = 0;
			fsg->data_size = 0;
			fsg->cmnd_size = 0;
			atomic_set(&data_get_control , 0);
			atomic_set(&data_send_control, 0);
			atomic_set(&atm_left_to_write, 0);
			atomic_set(&atm_left_to_read, 0);
			atomic_set(&data_response_control, 0);
			atomic_set(&req_fsg_wakeup, 0); // modified on 2008.03.20
			g_time_out = 0;
			g_first_wait = 1;
#ifdef USB_FILE_STORAGE_NAVU
			fsg->last_lun = 0;
			for (i = 0; i < fsg->nluns; ++i){
				fsg->luns[i].is_fake_sense_data = false;
			}
#endif /* USB_FILE_STORAGE_NAVU */
			// ljzhou 2006/07/11 end
			if ( new_config == CONFIG_VALUE ) {
				fsg->gadget->ops->set_selfpowered(fsg->gadget, 0);
				notify_supplied_power_enable(true);
				do_callback_usb_event_handler( fsg, FSG_USB_CONNECTED );
			} else {
				notify_supplied_power_enable(false);
				do_callback_usb_event_handler( fsg, FSG_USB_DISCONNECTED );
			}
#endif
		}
	}
#ifdef USB_FILE_STORAGE_NAVU
	else {
		notify_supplied_power_enable(false);
		do_callback_usb_event_handler( fsg, FSG_USB_DISCONNECTED );
	}
#endif
	return rc;
}


/*-------------------------------------------------------------------------*/

static void handle_exception(struct fsg_dev *fsg)
{
	siginfo_t		info;
	int			sig;
	int			i;
	int			num_active;
	struct fsg_buffhd	*bh;
	enum fsg_state		old_state;
	u8			new_config;
	struct lun		*curlun;
	unsigned int		exception_req_tag;
	int			rc;

	/* Clear the existing signals.  Anything but SIGUSR1 is converted
	 * into a high-priority EXIT exception. */
	for (;;) {
		sig = dequeue_signal_lock(current, &current->blocked, &info);
		if (!sig)
			break;

#ifdef USB_FILE_STORAGE_NAVU
		if( SIGKILL == sig) {
			//ljzhou 2006/4/29 modify
			g_exit_flag = 1;
			atomic_set(&data_get_control, 1);
			atomic_set(&data_send_control, 1);
			wake_up_all(&WaitDataQue);
			wake_up_all(&SendDataQue);
			//ljzhou 2006/4/29 ensure the wait queue exit
		}
		else if (sig != SIGUSR1) {
			if (fsg->state < FSG_STATE_EXIT)
				DBG(fsg, "Main thread exiting on signal\n");
			sonydbg(fsg, "handle_exception : raise_exception \n");
			if(sig == SIGSTOP) {
			  //printk("<1> SIGSTOP occurs\n");
			}
			if(sig == SIGCONT) {
			  //printk("<1> SIGCONT occurs\n");
			}
			raise_exception(fsg, FSG_STATE_EXIT);
		}
#else // USB_FILE_STORAGE_NAVU
		if (sig != SIGUSR1) {
			if (fsg->state < FSG_STATE_EXIT)
				DBG(fsg, "Main thread exiting on signal\n");
			raise_exception(fsg, FSG_STATE_EXIT);
		}
#endif // USB_FILE_STORAGE_NAVU
	}

	/* Cancel all the pending transfers */
	if (fsg->intreq_busy)
		usb_ep_dequeue(fsg->intr_in, fsg->intreq);
	for (i = 0; i < NUM_BUFFERS; ++i) {
		bh = &fsg->buffhds[i];
		if (bh->inreq_busy)
			usb_ep_dequeue(fsg->bulk_in, bh->inreq);
		if (bh->outreq_busy)
			usb_ep_dequeue(fsg->bulk_out, bh->outreq);
	}

	/* Wait until everything is idle */
	for (;;) {
		num_active = fsg->intreq_busy;
		for (i = 0; i < NUM_BUFFERS; ++i) {
			bh = &fsg->buffhds[i];
			num_active += bh->inreq_busy + bh->outreq_busy;
		}
		if (num_active == 0)
			break;
		if (sleep_thread(fsg))
			return;
	}

	/* Clear out the controller's fifos */
	if (fsg->bulk_in_enabled)
		usb_ep_fifo_flush(fsg->bulk_in);
	if (fsg->bulk_out_enabled)
		usb_ep_fifo_flush(fsg->bulk_out);
	if (fsg->intr_in_enabled)
		usb_ep_fifo_flush(fsg->intr_in);

	/* Reset the I/O buffer states and pointers, the SCSI
	 * state, and the exception.  Then invoke the handler. */
	spin_lock_irq(&fsg->lock);

	for (i = 0; i < NUM_BUFFERS; ++i) {
		bh = &fsg->buffhds[i];
		bh->state = BUF_STATE_EMPTY;
	}
	fsg->next_buffhd_to_fill = fsg->next_buffhd_to_drain =
			&fsg->buffhds[0];

	exception_req_tag = fsg->exception_req_tag;
	new_config = fsg->new_config;
	old_state = fsg->state;

	if (old_state == FSG_STATE_ABORT_BULK_OUT)
		fsg->state = FSG_STATE_STATUS_PHASE;
	else {
		for (i = 0; i < fsg->nluns; ++i) {
			curlun = &fsg->luns[i];
			curlun->prevent_medium_removal = 0;
			curlun->sense_data = curlun->unit_attention_data =
					SS_NO_SENSE;
			curlun->sense_data_info = 0;
			curlun->info_valid = 0;
		}
		fsg->state = FSG_STATE_IDLE;
	}
	spin_unlock_irq(&fsg->lock);

	/* Carry out any extra actions required for the exception */
	switch (old_state) {
	default:
		break;

	case FSG_STATE_ABORT_BULK_OUT:
		send_status(fsg);
		spin_lock_irq(&fsg->lock);
		if (fsg->state == FSG_STATE_STATUS_PHASE)
			fsg->state = FSG_STATE_IDLE;
		spin_unlock_irq(&fsg->lock);
		break;

	case FSG_STATE_RESET:
		/* In case we were forced against our will to halt a
		 * bulk endpoint, clear the halt now.  (The SuperH UDC
		 * requires this.) */
		if (test_and_clear_bit(CLEAR_BULK_HALTS,
				&fsg->atomic_bitflags)) {
			usb_ep_clear_halt(fsg->bulk_in);
			usb_ep_clear_halt(fsg->bulk_out);
		}

		if (transport_is_bbb()) {
			if (fsg->ep0_req_tag == exception_req_tag)
				ep0_queue(fsg);	// Complete the status stage

		} else if (transport_is_cbi())
			send_status(fsg);	// Status by interrupt pipe

		/* Technically this should go here, but it would only be
		 * a waste of time.  Ditto for the INTERFACE_CHANGE and
		 * CONFIG_CHANGE cases. */
		// for (i = 0; i < fsg->nluns; ++i)
		//	fsg->luns[i].unit_attention_data = SS_RESET_OCCURRED;
		break;

	case FSG_STATE_INTERFACE_CHANGE:
		rc = do_set_interface(fsg, 0);
		if (fsg->ep0_req_tag != exception_req_tag)
			break;
		if (rc != 0)			// STALL on errors
			fsg_set_halt(fsg, fsg->ep0);
		else				// Complete the status stage
			ep0_queue(fsg);
		break;

	case FSG_STATE_CONFIG_CHANGE:
		rc = do_set_config(fsg, new_config);
#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "where to set interface!!!! I think this is the last time \n");
#endif
		if (fsg->ep0_req_tag != exception_req_tag)
			break;
		if (rc != 0)			// STALL on errors
			fsg_set_halt(fsg, fsg->ep0);
		else				// Complete the status stage
			ep0_queue(fsg);
		break;

	case FSG_STATE_DISCONNECT:
		fsync_all(fsg);
		close_all_backing_files(fsg);
		do_set_config(fsg, 0);		// Unconfigured state
#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "where to set interface ??? \n");
#endif
		break;

	case FSG_STATE_EXIT:
	case FSG_STATE_TERMINATED:
		do_set_config(fsg, 0);			// Free resources
		spin_lock_irq(&fsg->lock);
		fsg->state = FSG_STATE_TERMINATED;	// Stop the thread
		spin_unlock_irq(&fsg->lock);
		break;
	}
}


/*-------------------------------------------------------------------------*/

static int fsg_main_thread(void *fsg_)
{
	struct fsg_dev		*fsg = fsg_;

#ifdef USB_FILE_STORAGE_NAVU
	fsg->thread_task = current;

	/* Release all our userspace resources */
	daemonize("file-storage-gadget");
#endif

	/* Allow the thread to be killed by a signal, but set the signal mask
	 * to block everything but INT, TERM, KILL, and USR1. */
	allow_signal(SIGINT);
	allow_signal(SIGTERM);
	allow_signal(SIGKILL);
	allow_signal(SIGUSR1);

	/* Allow the thread to be frozen */
	set_freezable();

	/* Arrange for userspace references to be interpreted as kernel
	 * pointers.  That way we can pass a kernel pointer to a routine
	 * that expects a __user pointer and it will work okay. */
	set_fs(get_ds());

#ifdef USB_FILE_STORAGE_NAVU
	/* Wait for the gadget registration to finish up */
	wait_for_completion(&fsg->thread_notifier);
#endif

	/* The main loop */
	while (fsg->state != FSG_STATE_TERMINATED) {
#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "here is thread\n");
		if (test_bit(SUSPENDED,&fsg->atomic_bitflags )){
			close_all_backing_files(fsg);
		}
		if (exception_in_progress(fsg) || signal_pending(current)) {
			sonydbg(fsg, "here is main loop \n");
			handle_exception(fsg);
			continue;
		}
#else
		if (exception_in_progress(fsg) || signal_pending(current)) {
			handle_exception(fsg);
			continue;
		}
#endif

		if (!fsg->running) {
			sleep_thread(fsg);
			continue;
		}

#ifdef USB_FILE_STORAGE_NAVU
		//ljzhou 2006/6/20 start
		//g_vendor_command = 0;
		//ljzhou 2006/6/20 end
#endif
		if (get_next_command(fsg))
			continue;

#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "break out the function : get_next_command \n");
#endif

		spin_lock_irq(&fsg->lock);
		if (!exception_in_progress(fsg))
			fsg->state = FSG_STATE_DATA_PHASE;
		spin_unlock_irq(&fsg->lock);

		if (do_scsi_command(fsg) || finish_reply(fsg))
			continue;

#ifdef USB_FILE_STORAGE_NAVU
		newdbg(fsg, "finish_do_scsi_command \n ");
#endif

		spin_lock_irq(&fsg->lock);
		if (!exception_in_progress(fsg))
			fsg->state = FSG_STATE_STATUS_PHASE;
		spin_unlock_irq(&fsg->lock);

#ifdef USB_FILE_STORAGE_NAVU
		newdbg(fsg, "here is send_status \n");
#endif
		if (send_status(fsg))
			continue;

#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "end of send_status \n");
#endif

		spin_lock_irq(&fsg->lock);
		if (!exception_in_progress(fsg))
			fsg->state = FSG_STATE_IDLE;
		spin_unlock_irq(&fsg->lock);
		}

	fsg->thread_task = NULL;
	flush_signals(current);

	/* In case we are exiting because of a signal, unregister the
	 * gadget driver and close the backing file. */
	if (test_and_clear_bit(REGISTERED, &fsg->atomic_bitflags)) {
		usb_gadget_unregister_driver(&fsg_driver);
		close_all_backing_files(fsg);
#ifdef USB_FILE_STORAGE_NAVU
		sonydbg(fsg, "usb_gadget_unregister_driver\n");
#endif
	}

#ifdef USB_FILE_STORAGE_NAVU
	sonydbg(fsg, "befor complete_and_exit(&fsg->thread_notifier, 0)\n");
#endif
	/* Let the unbind and cleanup routines know the thread has exited */
	complete_and_exit(&fsg->thread_notifier, 0);
#ifdef USB_FILE_STORAGE_NAVU
	sonydbg(fsg, "end complete_and_exit fsg thread\n");
#endif
}


/*-------------------------------------------------------------------------*/

/* If the next two routines are called while the gadget is registered,
 * the caller must own fsg->filesem for writing. */

static int open_backing_file(struct lun *curlun, const char *filename)
{
	int				ro;
	struct file			*filp = NULL;
	int				rc = -EINVAL;
	struct inode			*inode = NULL;
	loff_t				size;
	loff_t				num_sectors;

#ifdef USB_FILE_STORAGE_NAVU
	if (backing_file_is_open(curlun)){//already open
		printk(KERN_INFO "already has a file opened\n");
		return 0;
	}
#endif

#ifdef USB_FILE_STORAGE_NAVU  //m33
	if( Mount_disable_flag )
	{
		return -EINVAL;
	}
#endif
	/* R/W if we can, R/O if we must */
	ro = curlun->ro;
	if (!ro) {
		filp = filp_open(filename, O_RDWR | O_LARGEFILE, 0);
		if (-EROFS == PTR_ERR(filp))
			ro = 1;
	}
	if (ro) {
		filp = filp_open(filename, O_RDONLY | O_LARGEFILE, 0);
	}
	if (IS_ERR(filp)) {
#ifdef USB_FILE_STORAGE_NAVU
		//modified 2008/06/10
		if ( -ENOENT == PTR_ERR(filp) ) {
			/* If the device node doesn't exist, that's a problem.
			 * Other will be OK for removable devices,
			 * as we can use 'auto_mount_off' for nav-u. */
			LINFO(curlun, "unable to open backing file: %s\n", filename);
		}
#else
		LINFO(curlun, "unable to open backing file: %s\n", filename);
#endif
		return PTR_ERR(filp);
	}

	if (!(filp->f_mode & FMODE_WRITE))
		ro = 1;

	if (filp->f_path.dentry)
		inode = filp->f_path.dentry->d_inode;
	if (inode && S_ISBLK(inode->i_mode)) {
		if (bdev_read_only(inode->i_bdev))
			ro = 1;
	} else if (!inode || !S_ISREG(inode->i_mode)) {
		LINFO(curlun, "invalid file type: %s\n", filename);
		goto out;
	}

	/* If we can't read the file, it's no good.
	 * If we can't write the file, use it read-only. */
	if (!filp->f_op || !(filp->f_op->read || filp->f_op->aio_read)) {
		LINFO(curlun, "file not readable: %s\n", filename);
		goto out;
	}
	if (!(filp->f_op->write || filp->f_op->aio_write))
		ro = 1;

	size = i_size_read(inode->i_mapping->host);
	if (size < 0) {
		LINFO(curlun, "unable to find file size: %s\n", filename);
		rc = (int) size;
		goto out;
	}
	num_sectors = size >> 9;	// File size in 512-byte sectors
	if (num_sectors == 0) {
		LINFO(curlun, "file too small: %s\n", filename);
		rc = -ETOOSMALL;
		goto out;
	}

	get_file(filp);
	curlun->ro = ro;
	curlun->filp = filp;
	curlun->file_length = size;
	curlun->num_sectors = num_sectors;
	LDBG(curlun, "open backing file: %s\n", filename);
	rc = 0;

out:
	filp_close(filp, current->files);
	return rc;
}


static void close_backing_file(struct lun *curlun)
{
	if (curlun->filp) {
		LDBG(curlun, "close backing file\n");
		fput(curlun->filp);
		curlun->filp = NULL;
	}
}

static void close_all_backing_files(struct fsg_dev *fsg)
{
	int	i;

	for (i = 0; i < fsg->nluns; ++i)
		close_backing_file(&fsg->luns[i]);
}


static ssize_t show_ro(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct lun	*curlun = dev_to_lun(dev);

	return sprintf(buf, "%d\n", curlun->ro);
}

static ssize_t show_file(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct lun	*curlun = dev_to_lun(dev);
	struct fsg_dev	*fsg = dev_get_drvdata(dev);
	char		*p;
	ssize_t		rc;

	down_read(&fsg->filesem);
	if (backing_file_is_open(curlun)) {	// Get the complete pathname
		p = d_path(curlun->filp->f_path.dentry, curlun->filp->f_path.mnt,
				buf, PAGE_SIZE - 1);
		if (IS_ERR(p))
			rc = PTR_ERR(p);
		else {
			rc = strlen(p);
			memmove(buf, p, rc);
			buf[rc] = '\n';		// Add a newline
			buf[++rc] = 0;
		}
	} else {				// No file, return 0 bytes
		*buf = 0;
		rc = 0;
	}
	up_read(&fsg->filesem);
	return rc;
}


static ssize_t store_ro(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	ssize_t		rc = count;
	struct lun	*curlun = dev_to_lun(dev);
	struct fsg_dev	*fsg = dev_get_drvdata(dev);
	int		i;

	if (sscanf(buf, "%d", &i) != 1)
		return -EINVAL;

	/* Allow the write-enable status to change only while the backing file
	 * is closed. */
	down_read(&fsg->filesem);
	if (backing_file_is_open(curlun)) {
		LDBG(curlun, "read-only status change prevented\n");
		rc = -EBUSY;
	} else {
		curlun->ro = !!i;
		LDBG(curlun, "read-only status set to %d\n", curlun->ro);
	}
	up_read(&fsg->filesem);
	return rc;
}

static ssize_t store_file(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	struct lun	*curlun = dev_to_lun(dev);
	struct fsg_dev	*fsg = dev_get_drvdata(dev);
	int		rc = 0;

	if (curlun->prevent_medium_removal && backing_file_is_open(curlun)) {
		LDBG(curlun, "eject attempt prevented\n");
		return -EBUSY;				// "Door is locked"
	}

	/* Remove a trailing newline */
	if (count > 0 && buf[count-1] == '\n')
		((char *) buf)[count-1] = 0;		// Ugh!

	/* Eject current medium */
	down_write(&fsg->filesem);
	if (backing_file_is_open(curlun)) {
		close_backing_file(curlun);
		curlun->unit_attention_data = SS_MEDIUM_NOT_PRESENT;
	}

	/* Load new medium */
	if (count > 0 && buf[0]) {
		rc = open_backing_file(curlun, buf);
		if (rc == 0)
			curlun->unit_attention_data =
					SS_NOT_READY_TO_READY_TRANSITION;
	}
	up_write(&fsg->filesem);
	return (rc < 0 ? rc : count);
}


/* The write permissions and store_xxx pointers are set in fsg_bind() */
static DEVICE_ATTR(ro, 0444, show_ro, NULL);
static DEVICE_ATTR(file, 0444, show_file, NULL);


/*-------------------------------------------------------------------------*/

static void fsg_release(struct kref *ref)
{
	struct fsg_dev	*fsg = container_of(ref, struct fsg_dev, ref);

#ifdef USB_FILE_STORAGE_NAVU
	KFree(fsg->luns);
	KFree(fsg);
#else
	kfree(fsg->luns);
	kfree(fsg);
#endif
}

static void lun_release(struct device *dev)
{
	struct fsg_dev	*fsg = dev_get_drvdata(dev);

	kref_put(&fsg->ref, fsg_release);
}

static void /* __init_or_exit */ fsg_unbind(struct usb_gadget *gadget)
{
	struct fsg_dev		*fsg = get_gadget_data(gadget);
	int			i;
	struct lun		*curlun;
	struct usb_request	*req = fsg->ep0req;

	DBG(fsg, "unbind\n");
	clear_bit(REGISTERED, &fsg->atomic_bitflags);

	/* Unregister the sysfs attribute files and the LUNs */
	for (i = 0; i < fsg->nluns; ++i) {
		curlun = &fsg->luns[i];
		if (curlun->registered) {
			device_remove_file(&curlun->dev, &dev_attr_ro);
			device_remove_file(&curlun->dev, &dev_attr_file);
			device_unregister(&curlun->dev);
			curlun->registered = 0;
		}
	}

	/* If the thread isn't already dead, tell it to exit now */
	if (fsg->state != FSG_STATE_TERMINATED) {
		raise_exception(fsg, FSG_STATE_EXIT);
		wait_for_completion(&fsg->thread_notifier);

		/* The cleanup routine waits for this completion also */
		complete(&fsg->thread_notifier);
	}

	/* Free the data buffers */
	for (i = 0; i < NUM_BUFFERS; ++i)
		kfree(fsg->buffhds[i].buf);

	/* Free the request and buffer for endpoint 0 */
	if (req) {
		kfree(req->buf);
		usb_ep_free_request(fsg->ep0, req);
	}

	set_gadget_data(gadget, NULL);
}


static int __init check_parameters(struct fsg_dev *fsg)
{
	int	prot;
	int	gcnum;

	/* Store the default values */
	mod_data.transport_type = USB_PR_BULK;
	mod_data.transport_name = "Bulk-only";
	mod_data.protocol_type = USB_SC_SCSI;
	mod_data.protocol_name = "Transparent SCSI";

	if (mod_data.release == 0xffff) {	// Parameter wasn't set
		gcnum = usb_gadget_controller_number(fsg->gadget);
		if (gcnum >= 0)
			mod_data.release = 0x0300 + gcnum;
		else {
			WARN(fsg, "controller '%s' not recognized\n",
				fsg->gadget->name);
			mod_data.release = 0x0399;
		}
	}

	prot = simple_strtol(mod_data.protocol_parm, NULL, 0);

#ifdef CONFIG_USB_FILE_STORAGE_TEST
	if (strnicmp(mod_data.transport_parm, "BBB", 10) == 0) {
		;		// Use default setting
	} else if (strnicmp(mod_data.transport_parm, "CB", 10) == 0) {
		mod_data.transport_type = USB_PR_CB;
		mod_data.transport_name = "Control-Bulk";
	} else if (strnicmp(mod_data.transport_parm, "CBI", 10) == 0) {
		mod_data.transport_type = USB_PR_CBI;
		mod_data.transport_name = "Control-Bulk-Interrupt";
	} else {
		ERROR(fsg, "invalid transport: %s\n", mod_data.transport_parm);
		return -EINVAL;
	}

	if (strnicmp(mod_data.protocol_parm, "SCSI", 10) == 0 ||
			prot == USB_SC_SCSI) {
		;		// Use default setting
	} else if (strnicmp(mod_data.protocol_parm, "RBC", 10) == 0 ||
			prot == USB_SC_RBC) {
		mod_data.protocol_type = USB_SC_RBC;
		mod_data.protocol_name = "RBC";
	} else if (strnicmp(mod_data.protocol_parm, "8020", 4) == 0 ||
			strnicmp(mod_data.protocol_parm, "ATAPI", 10) == 0 ||
			prot == USB_SC_8020) {
		mod_data.protocol_type = USB_SC_8020;
		mod_data.protocol_name = "8020i (ATAPI)";
	} else if (strnicmp(mod_data.protocol_parm, "QIC", 3) == 0 ||
			prot == USB_SC_QIC) {
		mod_data.protocol_type = USB_SC_QIC;
		mod_data.protocol_name = "QIC-157";
	} else if (strnicmp(mod_data.protocol_parm, "UFI", 10) == 0 ||
			prot == USB_SC_UFI) {
		mod_data.protocol_type = USB_SC_UFI;
		mod_data.protocol_name = "UFI";
	} else if (strnicmp(mod_data.protocol_parm, "8070", 4) == 0 ||
			prot == USB_SC_8070) {
		mod_data.protocol_type = USB_SC_8070;
		mod_data.protocol_name = "8070i";
	} else {
		ERROR(fsg, "invalid protocol: %s\n", mod_data.protocol_parm);
		return -EINVAL;
	}

	mod_data.buflen &= PAGE_CACHE_MASK;
	if (mod_data.buflen <= 0) {
		ERROR(fsg, "invalid buflen\n");
		return -ETOOSMALL;
	}
#endif /* CONFIG_USB_FILE_STORAGE_TEST */

	return 0;
}


static int __init fsg_bind(struct usb_gadget *gadget)
{
	struct fsg_dev		*fsg = the_fsg;
	int			rc;
	int			i;
	struct lun		*curlun;
	struct usb_ep		*ep;
	struct usb_request	*req;
	char			*pathbuf, *p;

	fsg->gadget = gadget;
	set_gadget_data(gadget, fsg);
	fsg->ep0 = gadget->ep0;
	fsg->ep0->driver_data = fsg;

	if ((rc = check_parameters(fsg)) != 0)
		goto out;

	if (mod_data.removable) {	// Enable the store_xxx attributes
		dev_attr_ro.attr.mode = dev_attr_file.attr.mode = 0644;
		dev_attr_ro.store = store_ro;
		dev_attr_file.store = store_file;
	}

	/* Find out how many LUNs there should be */
	i = mod_data.nluns;
	if (i == 0)
		i = max(mod_data.num_filenames, 1);
	if (i > MAX_LUNS) {
		ERROR(fsg, "invalid number of LUNs: %d\n", i);
		rc = -EINVAL;
		goto out;
	}

	/* Create the LUNs, open their backing files, and register the
	 * LUN devices in sysfs. */
#ifdef USB_FILE_STORAGE_NAVU
	//fsg->luns = kmalloc(i * sizeof(struct lun), GFP_KERNEL);
	fsg->luns = KMalloc(i * sizeof(struct lun), GFP_KERNEL);
#else
	fsg->luns = kzalloc(i * sizeof(struct lun), GFP_KERNEL);
#endif
	if (!fsg->luns) {
		rc = -ENOMEM;
		goto out;
	}
#ifdef USB_FILE_STORAGE_NAVU
	memset(fsg->luns, 0, i * sizeof(struct lun));
#endif
	fsg->nluns = i;

	for (i = 0; i < fsg->nluns; ++i) {
		curlun = &fsg->luns[i];
		curlun->ro = mod_data.ro[i];
		curlun->dev.release = lun_release;
		curlun->dev.parent = &gadget->dev;
		curlun->dev.driver = &fsg_driver.driver;
		dev_set_drvdata(&curlun->dev, fsg);
		snprintf(curlun->dev.bus_id, BUS_ID_SIZE,
				"%s-lun%d", gadget->dev.bus_id, i);

		if ((rc = device_register(&curlun->dev)) != 0) {
			INFO(fsg, "failed to register LUN%d: %d\n", i, rc);
			goto out;
		}
		if ((rc = device_create_file(&curlun->dev,
					&dev_attr_ro)) != 0 ||
				(rc = device_create_file(&curlun->dev,
					&dev_attr_file)) != 0) {
			device_unregister(&curlun->dev);
			goto out;
		}
		curlun->registered = 1;
		kref_get(&fsg->ref);

		if (mod_data.file[i] && *mod_data.file[i]) {
			rc = open_backing_file(curlun, mod_data.file[i]);
#if !defined(USB_FILE_STORAGE_NAVU) // improved auto_mount_off 2008/06/10
			if (rc != 0)
#else
			if (rc != 0 && rc == ENOENT)
#endif
			{
				goto out;
			}
		} else
#if !defined(USB_FILE_STORAGE_NAVU)  //modified 2008/06/10
		if (!mod_data.removable)
#endif
		{
			ERROR(fsg, "no file given for LUN%d\n", i);
			rc = -EINVAL;
			goto out;
		}
	}

	/* Find all the endpoints we will use */
	usb_ep_autoconfig_reset(gadget);
	ep = usb_ep_autoconfig(gadget, &fs_bulk_in_desc);
	if (!ep)
		goto autoconf_fail;
	ep->driver_data = fsg;		// claim the endpoint
	fsg->bulk_in = ep;

	ep = usb_ep_autoconfig(gadget, &fs_bulk_out_desc);
	if (!ep)
		goto autoconf_fail;
	ep->driver_data = fsg;		// claim the endpoint
	fsg->bulk_out = ep;

	if (transport_is_cbi()) {
		ep = usb_ep_autoconfig(gadget, &fs_intr_in_desc);
		if (!ep)
			goto autoconf_fail;
		ep->driver_data = fsg;		// claim the endpoint
		fsg->intr_in = ep;
	}

	/* Fix up the descriptors */
	device_desc.bMaxPacketSize0 = fsg->ep0->maxpacket;
	device_desc.idVendor = cpu_to_le16(mod_data.vendor);
	device_desc.idProduct = cpu_to_le16(mod_data.product);
	device_desc.bcdDevice = cpu_to_le16(mod_data.release);

	i = (transport_is_cbi() ? 3 : 2);	// Number of endpoints
	intf_desc.bNumEndpoints = i;
	intf_desc.bInterfaceSubClass = mod_data.protocol_type;
	intf_desc.bInterfaceProtocol = mod_data.transport_type;
	fs_function[i + FS_FUNCTION_PRE_EP_ENTRIES] = NULL;

#ifdef CONFIG_USB_GADGET_DUALSPEED
	hs_function[i + HS_FUNCTION_PRE_EP_ENTRIES] = NULL;

	/* Assume ep0 uses the same maxpacket value for both speeds */
	dev_qualifier.bMaxPacketSize0 = fsg->ep0->maxpacket;

	/* Assume that all endpoint addresses are the same for both speeds */
	hs_bulk_in_desc.bEndpointAddress = fs_bulk_in_desc.bEndpointAddress;
	hs_bulk_out_desc.bEndpointAddress = fs_bulk_out_desc.bEndpointAddress;
	hs_intr_in_desc.bEndpointAddress = fs_intr_in_desc.bEndpointAddress;
#endif

	if (gadget->is_otg) {
		otg_desc.bmAttributes |= USB_OTG_HNP;
	}

	rc = -ENOMEM;

	/* Allocate the request and buffer for endpoint 0 */
	fsg->ep0req = req = usb_ep_alloc_request(fsg->ep0, GFP_KERNEL);
	if (!req)
		goto out;
	req->buf = kmalloc(EP0_BUFSIZE, GFP_KERNEL);
	if (!req->buf)
		goto out;
	req->complete = ep0_complete;

	/* Allocate the data buffers */
	for (i = 0; i < NUM_BUFFERS; ++i) {
		struct fsg_buffhd	*bh = &fsg->buffhds[i];

		/* Allocate for the bulk-in endpoint.  We assume that
		 * the buffer will also work with the bulk-out (and
		 * interrupt-in) endpoint. */
		bh->buf = kmalloc(mod_data.buflen, GFP_KERNEL);
		if (!bh->buf)
			goto out;
		bh->next = bh + 1;
	}
	fsg->buffhds[NUM_BUFFERS - 1].next = &fsg->buffhds[0];

	/* This should reflect the actual gadget power source */
	usb_gadget_set_selfpowered(gadget);

#ifdef USB_FILE_STORAGE_NAVU
	snprintf(manufacturer, sizeof manufacturer,
			"Sony nav-u");
#else
	snprintf(manufacturer, sizeof manufacturer, "%s %s with %s",
			init_utsname()->sysname, init_utsname()->release,
			gadget->name);
#endif

	/* On a real device, serial[] would be loaded from permanent
	 * storage.  We just encode it from the driver version string. */
#ifdef USB_FILE_STORAGE_NAVU
	for (i = 0; i < sizeof(serial) - 1; i++) {
		unsigned char	c = mod_data.serial_string[i];
		if (!c)
			break;
		serial[i] = c;
	}
#else
	for (i = 0; i < sizeof(serial) - 2; i += 2) {
		unsigned char		c = DRIVER_VERSION[i / 2];

		if (!c)
			break;
		sprintf(&serial[i], "%02X", c);
	}
#endif

#ifdef USB_FILE_STORAGE_NAVU
	if ((rc = kernel_thread(fsg_main_thread, fsg, (CLONE_VM | CLONE_FS |
			CLONE_FILES))) < 0)
		goto out;
	fsg->thread_pid = rc;
#else
	fsg->thread_task = kthread_create(fsg_main_thread, fsg,
			"file-storage-gadget");
	if (IS_ERR(fsg->thread_task)) {
		rc = PTR_ERR(fsg->thread_task);
		goto out;
	}
#endif

#ifdef USB_FILE_STORAGE_NAVU
	INFO(fsg, DRIVER_DESC ", version: %s\n", serial);
#else
	INFO(fsg, DRIVER_DESC ", version: " DRIVER_VERSION "\n");
#endif
	INFO(fsg, "Number of LUNs=%d\n", fsg->nluns);

#ifdef USB_FILE_STORAGE_NAVU
	//pathbuf = kmalloc(PATH_MAX, GFP_KERNEL);
	pathbuf = KMalloc(PATH_MAX, GFP_KERNEL);
#else
	pathbuf = kmalloc(PATH_MAX, GFP_KERNEL);
#endif
	for (i = 0; i < fsg->nluns; ++i) {
		curlun = &fsg->luns[i];
		if (backing_file_is_open(curlun)) {
			p = NULL;
			if (pathbuf) {
				p = d_path(curlun->filp->f_path.dentry,
					curlun->filp->f_path.mnt,
					pathbuf, PATH_MAX);
				if (IS_ERR(p))
					p = NULL;
			}
			LINFO(curlun, "ro=%d, file: %s\n",
					curlun->ro, (p ? p : "(error)"));
		}
	}
#ifdef USB_FILE_STORAGE_NAVU
	KFree(pathbuf);
	//kfree(pathbuf);
#else
	kfree(pathbuf);
#endif

	DBG(fsg, "transport=%s (x%02x)\n",
			mod_data.transport_name, mod_data.transport_type);
	DBG(fsg, "protocol=%s (x%02x)\n",
			mod_data.protocol_name, mod_data.protocol_type);
	DBG(fsg, "VendorID=x%04x, ProductID=x%04x, Release=x%04x\n",
			mod_data.vendor, mod_data.product, mod_data.release);
	DBG(fsg, "removable=%d, stall=%d, buflen=%u\n",
			mod_data.removable, mod_data.can_stall,
			mod_data.buflen);
#ifdef USB_FILE_STORAGE_NAVU
	DBG(fsg, "I/O thread pid: %d\n", fsg->thread_pid);
#else
	DBG(fsg, "I/O thread pid: %d\n", fsg->thread_task->pid);

	set_bit(REGISTERED, &fsg->atomic_bitflags);

	/* Tell the thread to start working */
	wake_up_process(fsg->thread_task);
#endif
	return 0;

autoconf_fail:
	ERROR(fsg, "unable to autoconfigure all endpoints\n");
	rc = -ENOTSUPP;

out:
	fsg->state = FSG_STATE_TERMINATED;	// The thread is dead
	fsg_unbind(gadget);
	close_all_backing_files(fsg);
	return rc;
}


/*-------------------------------------------------------------------------*/

static void fsg_suspend(struct usb_gadget *gadget)
{
	struct fsg_dev		*fsg = get_gadget_data(gadget);
	int i;
	struct fsg_buffhd	*bh;

	DBG(fsg, "suspend\n");
	set_bit(SUSPENDED, &fsg->atomic_bitflags);
#ifdef USB_FILE_STORAGE_NAVU
	sonydbg(fsg, "fsg_suspend function to call wakeup_up fsg ------+++++++++++++---------------\n");

	if ( fsg->new_config == CONFIG_VALUE ) {
		notify_supplied_power_enable(false);
		do_callback_usb_event_handler( fsg, FSG_USB_DISCONNECTED );
	}
	SET_SUSPEND;  /* Tempory code */

	g_exit_flag = 1;
	atomic_set(&data_get_control, 1);
	atomic_set(&data_send_control, 1);
	atomic_set(&data_response_control, 1);
	atomic_set(&atm_left_to_write, 0);
	atomic_set(&atm_left_to_read, 0);
	wake_up_all(&WaitDataQue);
	wake_up_all(&SendDataQue);
	wake_up_all(&ResponseQue);
	wakeup_thread(fsg);

	if (fsg->intreq_busy){
		usb_ep_dequeue(fsg->intr_in, fsg->intreq);
	}
	for (i = 0; i < NUM_BUFFERS; ++i) {
		bh = fsg->next_buffhd_to_fill;
		if (bh->inreq_busy){
			usb_ep_dequeue(fsg->bulk_in, bh->inreq);
		}
		if (bh->outreq_busy){
			usb_ep_dequeue(fsg->bulk_out, bh->outreq);
		}
	}
#endif
}

static void fsg_resume(struct usb_gadget *gadget)
{
	struct fsg_dev		*fsg = get_gadget_data(gadget);
	int i;
	struct fsg_buffhd	*bh;

#ifdef USB_FILE_STORAGE_NAVU
	SET_RESUME;  /* Tempory code */

	if(1 == g_usb_config_cplt) {
		printk(KERN_INFO "file_storage_navu.c: USB cable connected\n");

		if (fsg->intreq_busy){
			usb_ep_queue(fsg->intr_in, fsg->intreq, GFP_KERNEL);
			fsg->intreq_busy = 0;
		}
		for (i = 0; i < NUM_BUFFERS; ++i) {
			bh = fsg->next_buffhd_to_fill;
			if (bh->inreq_busy){
				usb_ep_queue(fsg->bulk_in, bh->inreq, GFP_KERNEL);
				bh->inreq_busy = 0;
			}
			if (bh->outreq_busy){
				usb_ep_queue(fsg->bulk_out, bh->outreq, GFP_KERNEL);
				bh->outreq_busy = 0;
			}
		}

		g_exit_flag = 0;
		g_time_out = 0;
		g_first_wait = 1;
		fsg->data_size = 0;
		fsg->cmnd_size = 0;
		atomic_set(&data_get_control , 0);
		atomic_set(&data_send_control, 0);
		atomic_set(&atm_left_to_write, 0);
		atomic_set(&atm_left_to_read, 0);
		atomic_set(&data_response_control, 0);
		atomic_set(&req_fsg_wakeup, 0);
		fsg->next_buffhd_to_fill->state = BUF_STATE_EMPTY;
		wakeup_thread(fsg);

		fsg->gadget->ops->set_selfpowered(fsg->gadget, 0);
		notify_supplied_power_enable(true);
		do_callback_usb_event_handler( fsg, FSG_USB_CONNECTED );
	}
	else {
		printk(KERN_INFO "file_storage_navu.c: USB cable connected w/o config_cplt\n");
	}

	sonydbg(fsg, "*******************************************************resume\n");
#else
	DBG(fsg, "resume\n");
#endif
	clear_bit(SUSPENDED, &fsg->atomic_bitflags);
}


/*-------------------------------------------------------------------------*/

static struct usb_gadget_driver		fsg_driver = {
#ifdef CONFIG_USB_GADGET_DUALSPEED
	.speed		= USB_SPEED_HIGH,
#else
	.speed		= USB_SPEED_FULL,
#endif
	.function	= (char *) longname,
	.bind		= fsg_bind,
	.unbind		= fsg_unbind,
	.disconnect	= fsg_disconnect,
	.setup		= fsg_setup,
	.suspend	= fsg_suspend,
	.resume		= fsg_resume,

	.driver		= {
		.name		= (char *) shortname,
#ifdef USB_FILE_STORAGE_NAVU
#else
		.owner		= THIS_MODULE,
#endif
		// .release = ...
		// .suspend = ...
		// .resume = ...
	},
};


static int __init fsg_alloc(void)
{
	struct fsg_dev		*fsg;

#ifdef USB_FILE_STORAGE_NAVU
	//fsg = kmalloc(sizeof *fsg, GFP_KERNEL);
	fsg = KMalloc(sizeof *fsg, GFP_KERNEL);
#else
	fsg = kzalloc(sizeof *fsg, GFP_KERNEL);
#endif
	if (!fsg)
		return -ENOMEM;
#ifdef USB_FILE_STORAGE_NAVU
	memset(fsg, 0, sizeof *fsg);
#endif
	spin_lock_init(&fsg->lock);
	init_rwsem(&fsg->filesem);
	kref_init(&fsg->ref);
	init_completion(&fsg->thread_notifier);

	the_fsg = fsg;
	return 0;
}

static int __init fsg_init(void)
{
	int		rc;
	struct fsg_dev	*fsg;
#ifdef USB_FILE_STORAGE_NAVU
	int		i;
#endif /* USB_FILE_STORAGE_NAVU */

	if ((rc = fsg_alloc()) != 0)
		return rc;
	fsg = the_fsg;
	if ((rc = usb_gadget_register_driver(&fsg_driver)) != 0) {
		kref_put(&fsg->ref, fsg_release);
		return rc;
	}
#ifdef USB_FILE_STORAGE_NAVU
	set_bit(REGISTERED, &fsg->atomic_bitflags);
	//Richard 2006/2/28 add start for vendor command transfer
	rc = register_chrdev(MAJOR_NUM, DEVICE_MS_NAME, &transfer_ops);
	if( 0 != rc ) {
		kref_put(&fsg->ref, fsg_release);
		return rc;
	}

	fsg->media_type_handler_info = NULL;

	for ( i = 0 ; i < MAX_PARM_INFO_HANDLER_NS ; i++ ) {
		fsg->parm_info_handler_table[i] = NULL;
	}

	for ( i = 0 ; i < MAX_VC_HANDLER_NS ; i++ ) {
		fsg->vc_handler_table[i] = NULL;
	}

	for ( i = 0 ; i < MAX_USB_EVENT_HANDLER_NS ; i++ ) {
		fsg->usb_event_handler_table[i] = NULL;
	}

	//Richard 2006/2/28 add end
	/* Tell the thread to start working */
	complete(&fsg->thread_notifier);
#endif
	return 0;
}
module_init(fsg_init);


static void __exit fsg_cleanup(void)
{
	struct fsg_dev	*fsg = the_fsg;
#ifdef USB_FILE_STORAGE_NAVU
	int		i;
#endif /* USB_FILE_STORAGE_NAVU */

#ifdef USB_FILE_STORAGE_NAVU
	notify_supplied_power_enable(false);
	fsg->gadget->ops->set_selfpowered(fsg->gadget, 1);
	do_callback_usb_event_handler( fsg, FSG_USB_DISCONNECTED );

	if ( fsg->media_type_handler_info != NULL ) {
		KFree( fsg->media_type_handler_info );
		fsg->media_type_handler_info = NULL;
	}

	for ( i = 0 ; i < MAX_PARM_INFO_HANDLER_NS ; i++ ) {
		if ( fsg->parm_info_handler_table[i] != NULL ) {
			KFree( fsg->parm_info_handler_table[i] );
			fsg->parm_info_handler_table[i] = NULL;
		}
	}

	for ( i = 0 ; i < MAX_VC_HANDLER_NS ; i++ ) {
		if ( fsg->vc_handler_table[i] != NULL ) {
			KFree( fsg->vc_handler_table[i] );
			fsg->vc_handler_table[i] = NULL;
		}
	}

	for ( i = 0 ; i < MAX_USB_EVENT_HANDLER_NS ; i++ ) {
		if ( fsg->usb_event_handler_table[i] != NULL ) {
			KFree( fsg->usb_event_handler_table[i] );
			fsg->usb_event_handler_table[i] = NULL;
		}
	}

	//Richard 2006/2/28 add start

	//exit_files(current);    /* free file resources */
	//flush_all_waitqueue();
	unregister_chrdev(MAJOR_NUM, DEVICE_MS_NAME);
#endif
	/* Unregister the driver iff the thread hasn't already done so */
	if (test_and_clear_bit(REGISTERED, &fsg->atomic_bitflags))
		usb_gadget_unregister_driver(&fsg_driver);

	/* Wait for the thread to finish up */
	wait_for_completion(&fsg->thread_notifier);

	close_all_backing_files(fsg);
	kref_put(&fsg->ref, fsg_release);
#ifdef USB_FILE_STORAGE_NAVU
	the_fsg = NULL;
	//memory check function
	MEM_SIZE_INFO;
#endif
}
module_exit(fsg_cleanup);

#ifdef USB_FILE_STORAGE_NAVU
EXPORT_SYMBOL(fsg_register_media_type_handler);
EXPORT_SYMBOL(fsg_unregister_media_type_handler);
EXPORT_SYMBOL(fsg_register_parm_info_handler);
EXPORT_SYMBOL(fsg_unregister_parm_info_handler);
EXPORT_SYMBOL(fsg_register_vc_handler);
EXPORT_SYMBOL(fsg_unregister_vc_handler);
EXPORT_SYMBOL(fsg_register_usb_event_handler);
EXPORT_SYMBOL(fsg_unregister_usb_event_handler);
#endif /* USB_FILE_STORAGE_NAVU */

